Package: mlr3
Title: Machine Learning in R - Next Generation
Version: 1.7.1
Authors@R: 
  c(
    person("Michel", "Lang", , "michellang@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-9754-0393")),
    person("Bernd", "Bischl", , "bernd_bischl@gmx.net", role = "aut",
           comment = c(ORCID = "0000-0001-6002-6980")),
    person("Jakob", "Richter", , "jakob1richter@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0003-4481-5554")),
    person("Patrick", "Schratz", , "patrick.schratz@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0003-0748-6624")),
    person("Giuseppe", "Casalicchio", , "giuseppe.casalicchio@stat.uni-muenchen.de", role = "ctb",
           comment = c(ORCID = "0000-0001-5324-5966")),
    person("Stefan", "Coors", , "mail@stefancoors.de", role = "ctb",
           comment = c(ORCID = "0000-0002-7465-2146")),
    person("Quay", "Au", , "quayau@gmail.com", role = "ctb",
           comment = c(ORCID = "0000-0002-5252-8902")),
    person("Martin", "Binder", , "mlr.developer@mb706.com", role = "aut"),
    person("Florian", "Pfisterer", , "pfistererf@googlemail.com", role = "aut",
           comment = c(ORCID = "0000-0001-8867-762X")),
    person("Raphael", "Sonabend", , "raphaelsonabend@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-9225-4654")),
    person("Lennart", "Schneider", , "lennart.sch@web.de", role = "ctb",
           comment = c(ORCID = "0000-0003-4152-5308")),
    person("Marc", "Becker", , "marcbecker@posteo.de", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0002-8115-0400")),
    person("Sebastian", "Fischer", , "sebf.fischer@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0002-9609-3197")),
    person("Lona", "Koers", , "lona.koers@gmail.com", role = "ctb"),
    person("John", "Zobolas", , "bblodfon@gmail.com", role = "ctb",
           comment = c(ORCID = "0000-0002-3609-8674")),
    person("Maximilian", "Mücke", , "muecke.maximilian@gmail.com", role = "ctb",
           comment = c(ORCID = "0009-0000-9432-9795")),
    person("Keno", "Mersmann", , "keno.mersmann@gmail.com", role = "ctb")
  )
Description: Efficient, object-oriented programming on the
    building blocks of machine learning. Provides 'R6' objects for tasks,
    learners, resamplings, and measures. The package is geared towards
    scalability and larger datasets by supporting parallelization and
    out-of-memory data-backends like databases. While 'mlr3' focuses on
    the core computational operations, add-on packages provide additional
    functionality.
License: LGPL-3
URL: https://mlr3.mlr-org.com, https://github.com/mlr-org/mlr3
BugReports: https://github.com/mlr-org/mlr3/issues
Depends: R (>= 3.5.0)
Imports: R6 (>= 2.4.1), backports (>= 1.5.0), checkmate (>= 2.0.0),
        cli, data.table (>= 1.15.0), evaluate (>= 1.0.4), future,
        future.apply (>= 1.5.0), lgr (>= 0.3.4), mirai (>= 2.4.1),
        methods, mlbench, mlr3measures (>= 1.2.0), mlr3misc (>=
        0.21.0), parallelly, palmerpenguins, paradox (>= 1.0.1), uuid
Suggests: callr, codetools, datasets, future.callr, mlr3data,
        progressr, remotes, RhpcBLASctl, rpart, testthat (>= 3.3.0)
Encoding: UTF-8
Config/testthat/edition: 3
Config/testthat/parallel: false
NeedsCompilation: no
Collate: 'mlr_reflections.R' 'BenchmarkResult.R' 'CallbackResample.R'
        'ContextResample.R' 'warn_deprecated.R' 'DataBackend.R'
        'DataBackendCbind.R' 'DataBackendDataTable.R'
        'DataBackendRbind.R' 'DataBackendRename.R' 'HotstartStack.R'
        'Learner.R' 'LearnerClassif.R' 'mlr_learners.R'
        'LearnerClassifDebug.R' 'LearnerClassifFeatureless.R'
        'LearnerClassifRpart.R' 'LearnerRegr.R' 'LearnerRegrDebug.R'
        'LearnerRegrFeatureless.R' 'LearnerRegrRpart.R' 'Measure.R'
        'mlr_measures.R' 'MeasureAIC.R' 'MeasureBIC.R'
        'MeasureClassif.R' 'MeasureClassifCosts.R' 'MeasureDebug.R'
        'MeasureElapsedTime.R' 'MeasureInternalValidScore.R'
        'MeasureOOBError.R' 'MeasureRegr.R' 'MeasureRegrPinball.R'
        'MeasureRegrRQR.R' 'MeasureRegrRSQ.R'
        'MeasureSelectedFeatures.R' 'MeasureSimilarity.R'
        'MeasureSimple.R' 'Prediction.R' 'PredictionClassif.R'
        'PredictionData.R' 'PredictionDataClassif.R'
        'PredictionDataRegr.R' 'PredictionRegr.R' 'ResampleResult.R'
        'Resampling.R' 'mlr_resamplings.R' 'ResamplingBootstrap.R'
        'ResamplingCV.R' 'ResamplingCustom.R' 'ResamplingCustomCV.R'
        'ResamplingHoldout.R' 'ResamplingInsample.R' 'ResamplingLOO.R'
        'ResamplingRepeatedCV.R' 'ResamplingSubsampling.R'
        'ResultData.R' 'Task.R' 'TaskSupervised.R' 'TaskClassif.R'
        'mlr_tasks.R' 'TaskClassif_breast_cancer.R'
        'TaskClassif_german_credit.R' 'TaskClassif_iris.R'
        'TaskClassif_penguins.R' 'TaskClassif_pima.R'
        'TaskClassif_sonar.R' 'TaskClassif_spam.R' 'TaskClassif_wine.R'
        'TaskClassif_zoo.R' 'TaskGenerator.R' 'mlr_task_generators.R'
        'TaskGenerator2DNormals.R' 'TaskGeneratorCassini.R'
        'TaskGeneratorCircle.R' 'TaskGeneratorFriedman1.R'
        'TaskGeneratorMoons.R' 'TaskGeneratorPeak.R'
        'TaskGeneratorSimplex.R' 'TaskGeneratorSmiley.R'
        'TaskGeneratorSpirals.R' 'TaskGeneratorXor.R' 'TaskRegr.R'
        'TaskRegr_california_housing.R' 'TaskRegr_mtcars.R'
        'TaskUnsupervised.R' 'as_benchmark_result.R'
        'as_data_backend.R' 'as_learner.R' 'as_measure.R'
        'as_prediction.R' 'as_prediction_classif.R'
        'as_prediction_data.R' 'as_prediction_regr.R'
        'as_resample_result.R' 'as_resampling.R' 'as_result_data.R'
        'as_task.R' 'as_task_classif.R' 'as_task_regr.R'
        'as_task_unsupervised.R' 'assertions.R' 'auto_convert.R'
        'benchmark.R' 'benchmark_grid.R' 'bibentries.R'
        'default_fallback.R' 'default_measures.R' 'fix_factor_levels.R'
        'helper.R' 'helper_data_table.R' 'helper_exec.R'
        'helper_hashes.R' 'install_pkgs.R' 'marshal.R'
        'mlr_callbacks.R' 'mlr_sugar.R' 'mlr_test_helpers.R'
        'partition.R' 'predict.R' 'reexports.R' 'resample.R'
        'score_roc_measures.R' 'set_threads.R' 'set_validate.R'
        'task_converters.R' 'worker.R' 'zzz.R'
Config/roxygen2/version: 8.0.0
Packaged: 2026-06-11 11:56:25 UTC; marc
Author: Michel Lang [aut] (ORCID: <https://orcid.org/0000-0001-9754-0393>),
  Bernd Bischl [aut] (ORCID: <https://orcid.org/0000-0001-6002-6980>),
  Jakob Richter [aut] (ORCID: <https://orcid.org/0000-0003-4481-5554>),
  Patrick Schratz [aut] (ORCID: <https://orcid.org/0000-0003-0748-6624>),
  Giuseppe Casalicchio [ctb] (ORCID:
    <https://orcid.org/0000-0001-5324-5966>),
  Stefan Coors [ctb] (ORCID: <https://orcid.org/0000-0002-7465-2146>),
  Quay Au [ctb] (ORCID: <https://orcid.org/0000-0002-5252-8902>),
  Martin Binder [aut],
  Florian Pfisterer [aut] (ORCID:
    <https://orcid.org/0000-0001-8867-762X>),
  Raphael Sonabend [aut] (ORCID: <https://orcid.org/0000-0001-9225-4654>),
  Lennart Schneider [ctb] (ORCID:
    <https://orcid.org/0000-0003-4152-5308>),
  Marc Becker [cre, aut] (ORCID: <https://orcid.org/0000-0002-8115-0400>),
  Sebastian Fischer [aut] (ORCID:
    <https://orcid.org/0000-0002-9609-3197>),
  Lona Koers [ctb],
  John Zobolas [ctb] (ORCID: <https://orcid.org/0000-0002-3609-8674>),
  Maximilian Mücke [ctb] (ORCID: <https://orcid.org/0009-0000-9432-9795>),
  Keno Mersmann [ctb]
Maintainer: Marc Becker <marcbecker@posteo.de>
Repository: CRAN
Date/Publication: 2026-06-11 14:00:20 UTC
Built: R 4.5.2; ; 2026-06-11 17:07:27 UTC; unix

Package: mlr3
Title: Machine Learning in R - Next Generation
Version: 1.7.1
Authors@R: 
  c(
    person("Michel", "Lang", , "michellang@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-9754-0393")),
    person("Bernd", "Bischl", , "bernd_bischl@gmx.net", role = "aut",
           comment = c(ORCID = "0000-0001-6002-6980")),
    person("Jakob", "Richter", , "jakob1richter@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0003-4481-5554")),
    person("Patrick", "Schratz", , "patrick.schratz@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0003-0748-6624")),
    person("Giuseppe", "Casalicchio", , "giuseppe.casalicchio@stat.uni-muenchen.de", role = "ctb",
           comment = c(ORCID = "0000-0001-5324-5966")),
    person("Stefan", "Coors", , "mail@stefancoors.de", role = "ctb",
           comment = c(ORCID = "0000-0002-7465-2146")),
    person("Quay", "Au", , "quayau@gmail.com", role = "ctb",
           comment = c(ORCID = "0000-0002-5252-8902")),
    person("Martin", "Binder", , "mlr.developer@mb706.com", role = "aut"),
    person("Florian", "Pfisterer", , "pfistererf@googlemail.com", role = "aut",
           comment = c(ORCID = "0000-0001-8867-762X")),
    person("Raphael", "Sonabend", , "raphaelsonabend@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-9225-4654")),
    person("Lennart", "Schneider", , "lennart.sch@web.de", role = "ctb",
           comment = c(ORCID = "0000-0003-4152-5308")),
    person("Marc", "Becker", , "marcbecker@posteo.de", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0002-8115-0400")),
    person("Sebastian", "Fischer", , "sebf.fischer@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0002-9609-3197")),
    person("Lona", "Koers", , "lona.koers@gmail.com", role = "ctb"),
    person("John", "Zobolas", , "bblodfon@gmail.com", role = "ctb",
           comment = c(ORCID = "0000-0002-3609-8674")),
    person("Maximilian", "Mücke", , "muecke.maximilian@gmail.com", role = "ctb",
           comment = c(ORCID = "0009-0000-9432-9795")),
    person("Keno", "Mersmann", , "keno.mersmann@gmail.com", role = "ctb")
  )
Description: Efficient, object-oriented programming on the
    building blocks of machine learning. Provides 'R6' objects for tasks,
    learners, resamplings, and measures. The package is geared towards
    scalability and larger datasets by supporting parallelization and
    out-of-memory data-backends like databases. While 'mlr3' focuses on
    the core computational operations, add-on packages provide additional
    functionality.
License: LGPL-3
URL: https://mlr3.mlr-org.com, https://github.com/mlr-org/mlr3
BugReports: https://github.com/mlr-org/mlr3/issues
Depends: R (>= 3.5.0)
Imports: R6 (>= 2.4.1), backports (>= 1.5.0), checkmate (>= 2.0.0),
        cli, data.table (>= 1.15.0), evaluate (>= 1.0.4), future,
        future.apply (>= 1.5.0), lgr (>= 0.3.4), mirai (>= 2.4.1),
        methods, mlbench, mlr3measures (>= 1.2.0), mlr3misc (>=
        0.21.0), parallelly, palmerpenguins, paradox (>= 1.0.1), uuid
Suggests: callr, codetools, datasets, future.callr, mlr3data,
        progressr, remotes, RhpcBLASctl, rpart, testthat (>= 3.3.0)
Encoding: UTF-8
Config/testthat/edition: 3
Config/testthat/parallel: false
NeedsCompilation: no
Collate: 'mlr_reflections.R' 'BenchmarkResult.R' 'CallbackResample.R'
        'ContextResample.R' 'warn_deprecated.R' 'DataBackend.R'
        'DataBackendCbind.R' 'DataBackendDataTable.R'
        'DataBackendRbind.R' 'DataBackendRename.R' 'HotstartStack.R'
        'Learner.R' 'LearnerClassif.R' 'mlr_learners.R'
        'LearnerClassifDebug.R' 'LearnerClassifFeatureless.R'
        'LearnerClassifRpart.R' 'LearnerRegr.R' 'LearnerRegrDebug.R'
        'LearnerRegrFeatureless.R' 'LearnerRegrRpart.R' 'Measure.R'
        'mlr_measures.R' 'MeasureAIC.R' 'MeasureBIC.R'
        'MeasureClassif.R' 'MeasureClassifCosts.R' 'MeasureDebug.R'
        'MeasureElapsedTime.R' 'MeasureInternalValidScore.R'
        'MeasureOOBError.R' 'MeasureRegr.R' 'MeasureRegrPinball.R'
        'MeasureRegrRQR.R' 'MeasureRegrRSQ.R'
        'MeasureSelectedFeatures.R' 'MeasureSimilarity.R'
        'MeasureSimple.R' 'Prediction.R' 'PredictionClassif.R'
        'PredictionData.R' 'PredictionDataClassif.R'
        'PredictionDataRegr.R' 'PredictionRegr.R' 'ResampleResult.R'
        'Resampling.R' 'mlr_resamplings.R' 'ResamplingBootstrap.R'
        'ResamplingCV.R' 'ResamplingCustom.R' 'ResamplingCustomCV.R'
        'ResamplingHoldout.R' 'ResamplingInsample.R' 'ResamplingLOO.R'
        'ResamplingRepeatedCV.R' 'ResamplingSubsampling.R'
        'ResultData.R' 'Task.R' 'TaskSupervised.R' 'TaskClassif.R'
        'mlr_tasks.R' 'TaskClassif_breast_cancer.R'
        'TaskClassif_german_credit.R' 'TaskClassif_iris.R'
        'TaskClassif_penguins.R' 'TaskClassif_pima.R'
        'TaskClassif_sonar.R' 'TaskClassif_spam.R' 'TaskClassif_wine.R'
        'TaskClassif_zoo.R' 'TaskGenerator.R' 'mlr_task_generators.R'
        'TaskGenerator2DNormals.R' 'TaskGeneratorCassini.R'
        'TaskGeneratorCircle.R' 'TaskGeneratorFriedman1.R'
        'TaskGeneratorMoons.R' 'TaskGeneratorPeak.R'
        'TaskGeneratorSimplex.R' 'TaskGeneratorSmiley.R'
        'TaskGeneratorSpirals.R' 'TaskGeneratorXor.R' 'TaskRegr.R'
        'TaskRegr_california_housing.R' 'TaskRegr_mtcars.R'
        'TaskUnsupervised.R' 'as_benchmark_result.R'
        'as_data_backend.R' 'as_learner.R' 'as_measure.R'
        'as_prediction.R' 'as_prediction_classif.R'
        'as_prediction_data.R' 'as_prediction_regr.R'
        'as_resample_result.R' 'as_resampling.R' 'as_result_data.R'
        'as_task.R' 'as_task_classif.R' 'as_task_regr.R'
        'as_task_unsupervised.R' 'assertions.R' 'auto_convert.R'
        'benchmark.R' 'benchmark_grid.R' 'bibentries.R'
        'default_fallback.R' 'default_measures.R' 'fix_factor_levels.R'
        'helper.R' 'helper_data_table.R' 'helper_exec.R'
        'helper_hashes.R' 'install_pkgs.R' 'marshal.R'
        'mlr_callbacks.R' 'mlr_sugar.R' 'mlr_test_helpers.R'
        'partition.R' 'predict.R' 'reexports.R' 'resample.R'
        'score_roc_measures.R' 'set_threads.R' 'set_validate.R'
        'task_converters.R' 'worker.R' 'zzz.R'
Config/roxygen2/version: 8.0.0
Packaged: 2026-06-11 11:56:25 UTC; marc
Author: Michel Lang [aut] (ORCID: <https://orcid.org/0000-0001-9754-0393>),
  Bernd Bischl [aut] (ORCID: <https://orcid.org/0000-0001-6002-6980>),
  Jakob Richter [aut] (ORCID: <https://orcid.org/0000-0003-4481-5554>),
  Patrick Schratz [aut] (ORCID: <https://orcid.org/0000-0003-0748-6624>),
  Giuseppe Casalicchio [ctb] (ORCID:
    <https://orcid.org/0000-0001-5324-5966>),
  Stefan Coors [ctb] (ORCID: <https://orcid.org/0000-0002-7465-2146>),
  Quay Au [ctb] (ORCID: <https://orcid.org/0000-0002-5252-8902>),
  Martin Binder [aut],
  Florian Pfisterer [aut] (ORCID:
    <https://orcid.org/0000-0001-8867-762X>),
  Raphael Sonabend [aut] (ORCID: <https://orcid.org/0000-0001-9225-4654>),
  Lennart Schneider [ctb] (ORCID:
    <https://orcid.org/0000-0003-4152-5308>),
  Marc Becker [cre, aut] (ORCID: <https://orcid.org/0000-0002-8115-0400>),
  Sebastian Fischer [aut] (ORCID:
    <https://orcid.org/0000-0002-9609-3197>),
  Lona Koers [ctb],
  John Zobolas [ctb] (ORCID: <https://orcid.org/0000-0002-3609-8674>),
  Maximilian Mücke [ctb] (ORCID: <https://orcid.org/0009-0000-9432-9795>),
  Keno Mersmann [ctb]
Maintainer: Marc Becker <marcbecker@posteo.de>
Repository: CRAN
Date/Publication: 2026-06-11 14:00:20 UTC
Built: R 4.6.0; ; 2026-06-11 15:53:42 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.data.table,BenchmarkResult)
S3method(as.data.table,DictionaryLearner)
S3method(as.data.table,DictionaryMeasure)
S3method(as.data.table,DictionaryResampling)
S3method(as.data.table,DictionaryTask)
S3method(as.data.table,DictionaryTaskGenerator)
S3method(as.data.table,PredictionClassif)
S3method(as.data.table,PredictionRegr)
S3method(as.data.table,ResampleResult)
S3method(as.data.table,Resampling)
S3method(as.data.table,Task)
S3method(as_benchmark_result,BenchmarkResult)
S3method(as_benchmark_result,ResampleResult)
S3method(as_data_backend,DataBackend)
S3method(as_data_backend,data.frame)
S3method(as_learner,Learner)
S3method(as_learners,default)
S3method(as_learners,list)
S3method(as_measure,"NULL")
S3method(as_measure,Measure)
S3method(as_measures,"NULL")
S3method(as_measures,default)
S3method(as_measures,list)
S3method(as_prediction,Prediction)
S3method(as_prediction,PredictionDataClassif)
S3method(as_prediction,PredictionDataRegr)
S3method(as_prediction_classif,PredictionClassif)
S3method(as_prediction_classif,data.frame)
S3method(as_prediction_data,Prediction)
S3method(as_prediction_data,PredictionData)
S3method(as_prediction_data,list)
S3method(as_prediction_regr,PredictionRegr)
S3method(as_prediction_regr,data.frame)
S3method(as_predictions,list)
S3method(as_resample_result,ResampleResult)
S3method(as_resample_result,ResultData)
S3method(as_resample_result,list)
S3method(as_resampling,Resampling)
S3method(as_resamplings,default)
S3method(as_resamplings,list)
S3method(as_task,Task)
S3method(as_task,default)
S3method(as_task_classif,DataBackend)
S3method(as_task_classif,TaskClassif)
S3method(as_task_classif,TaskRegr)
S3method(as_task_classif,data.frame)
S3method(as_task_classif,formula)
S3method(as_task_classif,matrix)
S3method(as_task_regr,DataBackend)
S3method(as_task_regr,TaskClassif)
S3method(as_task_regr,TaskRegr)
S3method(as_task_regr,data.frame)
S3method(as_task_regr,formula)
S3method(as_task_regr,matrix)
S3method(as_task_unsupervised,DataBackend)
S3method(as_task_unsupervised,Task)
S3method(as_task_unsupervised,data.frame)
S3method(as_tasks,default)
S3method(as_tasks,list)
S3method(as_tasks_unsupervised,Task)
S3method(as_tasks_unsupervised,list)
S3method(c,BenchmarkResult)
S3method(c,Prediction)
S3method(c,PredictionDataClassif)
S3method(c,PredictionDataRegr)
S3method(c,ResampleResult)
S3method(check_prediction_data,PredictionDataClassif)
S3method(check_prediction_data,PredictionDataRegr)
S3method(col_info,DataBackend)
S3method(col_info,data.table)
S3method(create_empty_prediction_data,TaskClassif)
S3method(create_empty_prediction_data,TaskRegr)
S3method(data.table::format_col,condition)
S3method(data.table::format_list_item,condition)
S3method(default_fallback,Learner)
S3method(default_fallback,LearnerClassif)
S3method(default_fallback,LearnerRegr)
S3method(default_values,Learner)
S3method(default_values,LearnerClassifRpart)
S3method(default_values,LearnerRegrRpart)
S3method(extract_pkgs,BenchmarkResult)
S3method(extract_pkgs,R6)
S3method(extract_pkgs,ResampleResult)
S3method(extract_pkgs,character)
S3method(extract_pkgs,list)
S3method(filter_prediction_data,PredictionDataClassif)
S3method(filter_prediction_data,PredictionDataRegr)
S3method(fix_factor_levels,data.table)
S3method(head,Task)
S3method(is_missing_prediction_data,PredictionDataClassif)
S3method(is_missing_prediction_data,PredictionDataRegr)
S3method(marshal_model,classif.debug_model)
S3method(marshal_model,default)
S3method(marshal_model,learner_state)
S3method(partition,Task)
S3method(predict,Learner)
S3method(print,PredictionData)
S3method(print,benchmark_grid)
S3method(print,bmr_aggregate)
S3method(print,bmr_score)
S3method(print,marshaled)
S3method(print,roc_measures)
S3method(print,rr_score)
S3method(rd_info,Learner)
S3method(rd_info,Measure)
S3method(rd_info,Task)
S3method(set_threads,R6)
S3method(set_threads,default)
S3method(set_threads,list)
S3method(set_validate,Learner)
S3method(summary,Task)
S3method(tail,Task)
S3method(task_check_col_roles,Task)
S3method(task_check_col_roles,TaskClassif)
S3method(task_check_col_roles,TaskRegr)
S3method(task_check_col_roles,TaskSupervised)
S3method(task_check_col_roles,TaskUnsupervised)
S3method(unmarshal_model,classif.debug_model_marshaled)
S3method(unmarshal_model,default)
S3method(unmarshal_model,learner_state_marshaled)
export(BenchmarkResult)
export(CallbackResample)
export(ContextResample)
export(DataBackend)
export(DataBackendDataTable)
export(HotstartStack)
export(Learner)
export(LearnerClassif)
export(LearnerClassifDebug)
export(LearnerClassifFeatureless)
export(LearnerClassifRpart)
export(LearnerRegr)
export(LearnerRegrDebug)
export(LearnerRegrFeatureless)
export(LearnerRegrRpart)
export(Measure)
export(MeasureAIC)
export(MeasureBIC)
export(MeasureClassif)
export(MeasureClassifCosts)
export(MeasureDebugClassif)
export(MeasureElapsedTime)
export(MeasureInternalValidScore)
export(MeasureOOBError)
export(MeasureRegr)
export(MeasureRegrPinball)
export(MeasureRegrRQR)
export(MeasureRegrRSQ)
export(MeasureSelectedFeatures)
export(MeasureSimilarity)
export(Prediction)
export(PredictionClassif)
export(PredictionRegr)
export(ResampleResult)
export(Resampling)
export(ResamplingBootstrap)
export(ResamplingCV)
export(ResamplingCustom)
export(ResamplingCustomCV)
export(ResamplingHoldout)
export(ResamplingInsample)
export(ResamplingLOO)
export(ResamplingRepeatedCV)
export(ResamplingSubsampling)
export(ResultData)
export(Task)
export(TaskClassif)
export(TaskGenerator)
export(TaskGenerator2DNormals)
export(TaskGeneratorCassini)
export(TaskGeneratorCircle)
export(TaskGeneratorFriedman1)
export(TaskGeneratorMoons)
export(TaskGeneratorPeak)
export(TaskGeneratorSimplex)
export(TaskGeneratorSmiley)
export(TaskGeneratorSpirals)
export(TaskGeneratorXor)
export(TaskRegr)
export(TaskSupervised)
export(TaskUnsupervised)
export(as.data.table)
export(as_benchmark_result)
export(as_data_backend)
export(as_learner)
export(as_learners)
export(as_measure)
export(as_measures)
export(as_prediction)
export(as_prediction_classif)
export(as_prediction_data)
export(as_prediction_regr)
export(as_predictions)
export(as_resample_result)
export(as_resampling)
export(as_resamplings)
export(as_result_data)
export(as_task)
export(as_task_classif)
export(as_task_regr)
export(as_task_unsupervised)
export(as_tasks)
export(as_tasks_unsupervised)
export(assert_backend)
export(assert_benchmark_result)
export(assert_empty_ellipsis)
export(assert_has_backend)
export(assert_learnable)
export(assert_learner)
export(assert_learners)
export(assert_measure)
export(assert_measures)
export(assert_predictable)
export(assert_prediction)
export(assert_quantiles)
export(assert_resample_callback)
export(assert_resample_callbacks)
export(assert_resample_result)
export(assert_resampling)
export(assert_resamplings)
export(assert_row_ids)
export(assert_scorable)
export(assert_task)
export(assert_tasks)
export(assert_validate)
export(auto_convert)
export(benchmark)
export(benchmark_grid)
export(callback_resample)
export(check_prediction_data)
export(clbk)
export(clbks)
export(col_info)
export(convert_task)
export(create_empty_prediction_data)
export(data.table)
export(default_fallback)
export(default_measures)
export(deprecated_binding_old)
export(extract_pkgs)
export(filter_prediction_data)
export(install_pkgs)
export(is_marshaled_model)
export(is_missing_prediction_data)
export(learner_marshal)
export(learner_marshaled)
export(learner_unmarshal)
export(lrn)
export(lrns)
export(marshal_model)
export(mlr_callbacks)
export(mlr_learners)
export(mlr_measures)
export(mlr_reflections)
export(mlr_resamplings)
export(mlr_task_generators)
export(mlr_tasks)
export(msr)
export(msrs)
export(partition)
export(resample)
export(rsmp)
export(rsmps)
export(score_roc_measures)
export(set_threads)
export(set_validate)
export(task_check_col_roles)
export(tgen)
export(tgens)
export(tsk)
export(tsks)
export(uhash)
export(uhashes)
export(unmarshal_model)
export(warn_deprecated_old)
import(checkmate)
import(cli)
import(data.table)
import(mlr3misc)
import(palmerpenguins)
import(paradox)
importFrom(R6,R6Class)
importFrom(R6,is.R6)
importFrom(data.table,as.data.table)
importFrom(data.table,data.table)
importFrom(future,nbrOfWorkers)
importFrom(future,plan)
importFrom(graphics,plot)
importFrom(methods,formalArgs)
importFrom(mlr3misc,clbk)
importFrom(mlr3misc,clbks)
importFrom(mlr3misc,mlr_callbacks)
importFrom(parallelly,availableCores)
importFrom(stats,contr.treatment)
importFrom(stats,model.frame)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,terms)
importFrom(stats,weighted.mean)
importFrom(utils,bibentry)
importFrom(utils,data)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,packageVersion)
importFrom(utils,tail)
importFrom(uuid,UUIDgenerate)

# mlr3 <img src="man/figures/logo.png" align="right" width = "120" />

Package website: [release](https://mlr3.mlr-org.com/) \|
[dev](https://mlr3.mlr-org.com/dev/)

Efficient, object-oriented programming on the building blocks of machine
learning. Successor of [mlr](https://github.com/mlr-org/mlr).

<!-- badges: start -->

[![r-cmd-check](https://github.com/mlr-org/mlr3/actions/workflows/r-cmd-check.yml/badge.svg)](https://github.com/mlr-org/mlr3/actions/workflows/r-cmd-check.yml)
[![DOI](https://joss.theoj.org/papers/10.21105/joss.01903/status.svg)](https://doi.org/10.21105/joss.01903)
[![CRAN
Status](https://www.r-pkg.org/badges/version-ago/mlr3)](https://cran.r-project.org/package=mlr3)
[![Mattermost](https://img.shields.io/badge/chat-mattermost-orange.svg)](https://lmmisld-lmu-stats-slds.srv.mwn.de/mlr_invite/)
<!-- badges: end -->

## Resources (for users and developers)

- We have written a [book](https://mlr3book.mlr-org.com/). This should
  be the central entry point to the package.
- The [mlr-org website](https://mlr-org.com/) includes for example a
  [gallery](https://mlr-org.com/gallery.html) with case studies.
- [Reference manual](https://mlr3.mlr-org.com/reference/)
- [FAQ](https://mlr-org.com/faq.html)
- Ask questions on Stackoverflow (tag mlr3)
- **Extension Learners**
  - Recommended core regression, classification, and survival learners
    are in [mlr3learners](https://github.com/mlr-org/mlr3learners)
  - All others are in
    [mlr3extralearners](https://github.com/mlr-org/mlr3extralearners)
  - Use the [learner search](https://mlr-org.com/learners.html) to get a
    simple overview
- **Cheatsheets**
  - [Overview of cheatsheets](https://cheatsheets.mlr-org.com)
  - [mlr3](https://cheatsheets.mlr-org.com/mlr3.pdf)
  - [mlr3tuning](https://cheatsheets.mlr-org.com/mlr3tuning.pdf)
  - [mlr3pipelines](https://cheatsheets.mlr-org.com/mlr3pipelines.pdf)
- **Videos**:
  - [useR2019 talk on mlr3](https://www.youtube.com/watch?v=wsP2hiFnDQs)
  - [useR2019 talk on mlr3pipelines and
    mlr3tuning](https://www.youtube.com/watch?v=gEW5RxkbQuQ)
  - [useR2020 tutorial on mlr3, mlr3tuning and
    mlr3pipelines](https://www.youtube.com/watch?v=T43hO2o_nZw)
    <!--   - [Recorded talk about mlr3spatiotempcv and mlr3spatial at OpenDataScience Europe Conference 2021 in Wageningen, NL](https://av.tib.eu/media/55271) -->
- **Courses/Lectures**
  - The course [Introduction to Machine learning
    (I2ML)](https://slds-lmu.github.io/i2ml/) is a free and open flipped
    classroom course on the basics of machine learning. `mlr3` is used
    in the
    [demos](https://github.com/slds-lmu/lecture_i2ml/tree/master/code-demos-pdf)
    and
    [exercises](https://github.com/slds-lmu/lecture_i2ml/tree/master/exercises).
- **Templates/Tutorials**
  - [mlr3-targets](https://github.com/mlr-org/mlr3-targets): Tutorial
    showcasing how to use {mlr3} with
    [targets](https://docs.ropensci.org/targets/) for reproducible ML
    workflow automation.
  - Using `mlr3viz` and `animint2` for [interactive
    visualizations](https://animint-manual-en.netlify.app/ch20/).
- [List of extension packages](https://mlr-org.com/ecosystem.html)
- [mlr-outreach](https://github.com/mlr-org/mlr-outreach) contains
  public talks and slides resources.
- [Wiki](https://github.com/mlr-org/mlr3/wiki): Contains mainly
  information for developers.

## Installation

Install the last release from CRAN:

``` r
install.packages("mlr3")
```

Install the development version from GitHub:

``` r
# install.packages("pak")
pak::pak("mlr-org/mlr3")
```

If you want to get started with `mlr3`, we recommend installing the
[mlr3verse](https://mlr3verse.mlr-org.com/) meta-package which installs
`mlr3` and some of the most important extension packages:

``` r
install.packages("mlr3verse")
```

## Example

### Constructing Learners and Tasks

``` r
library(mlr3)

# create learning task
task_penguins = as_task_classif(species ~ ., data = palmerpenguins::penguins)
task_penguins
```

    ## 
    ## ── <TaskClassif> (344x8) ───────────────────────────────────────────────────────
    ## • Target: species
    ## • Target classes: Adelie (44%), Gentoo (36%), Chinstrap (20%)
    ## • Properties: multiclass
    ## • Features (7):
    ##   • int (3): body_mass_g, flipper_length_mm, year
    ##   • dbl (2): bill_depth_mm, bill_length_mm
    ##   • fct (2): island, sex

``` r
# load learner and set hyperparameter
learner = lrn("classif.rpart", cp = .01)
```

### Basic train + predict

``` r
# train/test split
split = partition(task_penguins, ratio = 0.67)

# train the model
learner$train(task_penguins, split$train_set)

# predict data
prediction = learner$predict(task_penguins, split$test_set)

# calculate performance
prediction$confusion
```

    ##            truth
    ## response    Adelie Chinstrap Gentoo
    ##   Adelie       146         5      0
    ##   Chinstrap      6        63      1
    ##   Gentoo         0         0    123

``` r
measure = msr("classif.acc")
prediction$score(measure)
```

    ## classif.acc 
    ##   0.9651163

### Resample

``` r
# 3-fold cross validation
resampling = rsmp("cv", folds = 3L)

# run experiments
rr = resample(task_penguins, learner, resampling)

# access results
rr$score(measure)[, .(task_id, learner_id, iteration, classif.acc)]
```

    ##                     task_id    learner_id iteration classif.acc
    ## 1: palmerpenguins::penguins classif.rpart         1   0.8956522
    ## 2: palmerpenguins::penguins classif.rpart         2   0.9478261
    ## 3: palmerpenguins::penguins classif.rpart         3   0.9649123

``` r
rr$aggregate(measure)
```

    ## classif.acc 
    ##   0.9361302

## Extension Packages

<a href="https://raw.githubusercontent.com/mlr-org/mlr3/main/man/figures/mlr3verse.svg?sanitize=true"><img src="man/figures/mlr3verse.svg" /></a>

Consult the
[wiki](https://github.com/mlr-org/mlr3/wiki/Extension-Packages) for
short descriptions and links to the respective repositories.

For beginners, we strongly recommend to install and load the
[mlr3verse](https://mlr3verse.mlr-org.com/) package for a better user
experience.

## Why a rewrite?

[mlr](https://github.com/mlr-org/mlr) was first released to
[CRAN](https://cran.r-project.org/package=mlr) in 2013. Its core design
and architecture date back even further. The addition of many features
has led to a [feature
creep](https://en.wikipedia.org/wiki/Feature_creep) which makes
[mlr](https://github.com/mlr-org/mlr) hard to maintain and hard to
extend. We also think that while mlr was nicely extensible in some parts
(learners, measures, etc.), other parts were less easy to extend from
the outside. Also, many helpful R libraries did not exist at the time
[mlr](https://github.com/mlr-org/mlr) was created, and their inclusion
would result in non-trivial API changes.

## Design principles

- Only the basic building blocks for machine learning are implemented in
  this package.
- Focus on computation here. No visualization or other stuff. That can
  go in extra packages.
- Overcome the limitations of R’s [S3
  classes](https://adv-r.hadley.nz/s3.html) with the help of
  [R6](https://cran.r-project.org/package=R6).
- Embrace [R6](https://cran.r-project.org/package=R6) for a clean
  OO-design, object state-changes and reference semantics. This might be
  less “traditional R”, but seems to fit `mlr` nicely.
- Embrace [`data.table`](https://cran.r-project.org/package=data.table)
  for fast and convenient data frame computations.
- Combine `data.table` and `R6`, for this we will make heavy use of list
  columns in data.tables.
- Defensive programming and type safety. All user input is checked with
  [`checkmate`](https://cran.r-project.org/package=checkmate). Return
  types are documented, and mechanisms popular in base R which
  “simplify” the result unpredictably (e.g., `sapply()` or `drop`
  argument in `[.data.frame`) are avoided.
- Be light on dependencies. `mlr3` requires the following packages at
  runtime:
  - [`parallelly`](https://cran.r-project.org/package=parallelly):
    Helper functions for parallelization. No extra recursive
    dependencies.
  - [`future.apply`](https://cran.r-project.org/package=future.apply):
    Resampling and benchmarking is parallelized with the
    [`future`](https://cran.r-project.org/package=future) abstraction
    interfacing many parallel backends.
  - [`backports`](https://cran.r-project.org/package=backports): Ensures
    backward compatibility with older R releases. Developed by members
    of the `mlr` team. No recursive dependencies.
  - [`checkmate`](https://cran.r-project.org/package=checkmate): Fast
    argument checks. Developed by members of the `mlr` team. No extra
    recursive dependencies.
  - [`mlr3misc`](https://cran.r-project.org/package=mlr3misc):
    Miscellaneous functions used in multiple mlr3 [extension
    packages](https://mlr-org.com/ecosystem.html). Developed by the
    `mlr` team.
  - [`paradox`](https://cran.r-project.org/package=paradox):
    Descriptions for parameters and parameter sets. Developed by the
    `mlr` team. No extra recursive dependencies.
  - [`R6`](https://cran.r-project.org/package=R6): Reference class
    objects. No recursive dependencies.
  - [`data.table`](https://cran.r-project.org/package=data.table):
    Extension of R’s `data.frame`. No recursive dependencies.
  - [`digest`](https://cran.r-project.org/package=digest) (via
    `mlr3misc`): Hash digests. No recursive dependencies.
  - [`uuid`](https://cran.r-project.org/package=uuid): Create unique
    string identifiers. No recursive dependencies.
  - [`lgr`](https://cran.r-project.org/package=lgr): Logging facility.
    No extra recursive dependencies.
  - [`mlr3measures`](https://cran.r-project.org/package=mlr3measures):
    Performance measures. No extra recursive dependencies.
  - [`mlbench`](https://cran.r-project.org/package=mlbench): A
    collection of machine learning data sets. No dependencies.
  - [`palmerpenguins`](https://cran.r-project.org/package=palmerpenguins):
    A classification data set about penguins, used on examples and
    provided as a toy task. No dependencies.
- [Reflections](https://en.wikipedia.org/wiki/Reflection_%28computer_programming%29):
  Objects are queryable for properties and capabilities, allowing you to
  program on them.
- Additional functionality that comes with extra dependencies:
  - To capture output, warnings and exceptions,
    [`evaluate`](https://cran.r-project.org/package=evaluate) and
    [`callr`](https://cran.r-project.org/package=callr) can be used.

## Contributing to mlr3

This R package is licensed under the
[LGPL-3](https://www.gnu.org/licenses/lgpl-3.0.html.en). If you
encounter problems using this software (lack of documentation,
misleading or wrong documentation, unexpected behavior, bugs, …) or just
want to suggest features, please open an issue in the [issue
tracker](https://github.com/mlr-org/mlr3/issues). Pull requests are
welcome and will be included at the discretion of the maintainers.

Please consult the [wiki](https://github.com/mlr-org/mlr3/wiki/) for a
[style guide](https://github.com/mlr-org/mlr3/wiki/Style-Guide), a
[roxygen guide](https://github.com/mlr-org/mlr3/wiki/Roxygen-Guide) and
a [pull request
guide](https://github.com/mlr-org/mlr3/wiki/PR-Guidelines).

## Citing mlr3

If you use mlr3, please cite our [JOSS
article](https://doi.org/10.21105/joss.01903):

    @Article{mlr3,
      title = {{mlr3}: A modern object-oriented machine learning framework in {R}},
      author = {Michel Lang and Martin Binder and Jakob Richter and Patrick Schratz and Florian Pfisterer and Stefan Coors and Quay Au and Giuseppe Casalicchio and Lars Kotthoff and Bernd Bischl},
      journal = {Journal of Open Source Software},
      year = {2019},
      month = {dec},
      doi = {10.21105/joss.01903},
      url = {https://joss.theoj.org/papers/10.21105/joss.01903},
    }

`[[.R6` = function(x, i, ...) {
  if (exists(i, envir = x, inherits = FALSE)) {
    return(get(i, envir = x))
  }
  stop("R6 class ", paste0(class(x), collapse = "/"), " does not have slot '", i, "'!")
}

`$.R6` = function(x, name) {
  if (exists(name, envir = x, inherits = FALSE)) {
    return(get(name, envir = x))
  }
  stop("R6 class ", paste0(class(x), collapse = "/"), " does not have slot '", name, "'!")
}

#' @title Sonar Classification Task
#'
#' @name mlr_tasks_sonar
#' @format [R6::R6Class] inheriting from [TaskClassif].
#' @include mlr_tasks.R
#'
#' @description
#' A classification task for the [mlbench::Sonar] data set.
#' Positive class is set to "M" (Mine).
#'
#' @templateVar id sonar
#' @template task
#'
#' @template seealso_task
NULL

load_task_sonar = function(id = "sonar") {
  b = as_data_backend(load_dataset("Sonar", "mlbench"))
  task = TaskClassif$new(id, b, target = "Class", positive = "M", label = "Sonar: Mines vs. Rocks")
  b$hash = task$man = "mlr3::mlr_tasks_sonar"
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("sonar", load_task_sonar)

#' @title Custom Resampling
#'
#' @name mlr_resamplings_custom
#' @include Resampling.R
#'
#' @description
#' Splits data into training and test sets using manually provided indices.
#'
#' @templateVar id custom
#' @template resampling
#'
#' @template seealso_resampling
#' @export
#' @examples
#' # Create a task with 10 observations
#' task = tsk("penguins")
#' task$filter(1:10)
#'
#' # Instantiate Resampling
#' custom = rsmp("custom")
#' train_sets = list(1:5, 5:10)
#' test_sets = list(5:10, 1:5)
#' custom$instantiate(task, train_sets, test_sets)
#'
#' custom$train_set(1)
#' custom$test_set(1)
ResamplingCustom = R6Class(
  "ResamplingCustom",
  inherit = Resampling,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      super$initialize(
        id = "custom",
        duplicated_ids = TRUE,
        label = "Custom Splits",
        man = "mlr3::mlr_resamplings_custom"
      )
    },

    #' @description
    #' Instantiate this [Resampling] with custom splits into training and test set.
    #'
    #' @param task [Task]\cr
    #'   Mainly used to check if `train_sets` and `test_sets` are feasible.
    #'
    #' @param train_sets (list of `integer()`)\cr
    #'   List with row ids for training, one list element per iteration.
    #'   Must have the same length as `test_sets`.
    #'
    #' @param test_sets (list of `integer()`)\cr
    #'   List with row ids for testing, one list element per iteration.
    #'   Must have the same length as `train_sets`.
    instantiate = function(task, train_sets, test_sets) {
      task = assert_task(as_task(task))
      private$.hash = NULL
      assert_list(train_sets, types = "atomicvector", any.missing = FALSE)
      assert_list(test_sets, types = "atomicvector", len = length(train_sets), any.missing = FALSE, null.ok = TRUE)
      assert_subset(unlist(train_sets, use.names = FALSE), task$row_ids)
      assert_subset(unlist(test_sets, use.names = FALSE), task$row_ids)
      self$instance = list(train = train_sets, test = test_sets)
      self$task_hash = task$hash
      self$task_nrow = task$nrow
      self$task_row_hash = task$row_hash
      invisible(self)
    }
  ),

  active = list(
    #' @template field_iters
    iters = function(rhs) {
      assert_ro_binding(rhs)
      if (self$is_instantiated) length(self$instance$train) else NA_integer_
    }
  ),

  private = list(
    .get_train = function(i) {
      self$instance$train[[i]]
    },

    .get_test = function(i) {
      self$instance$test[[i]]
    }
  )
)

#' @include mlr_resamplings.R
mlr_resamplings$add("custom", function() ResamplingCustom$new())

#' @title Featureless Regression Learner
#'
#' @name mlr_learners_regr.featureless
#' @include LearnerRegr.R
#'
#' @description
#' A simple [LearnerRegr] which only analyzes the response during train, ignoring all features.
#' If hyperparameter `robust` is `FALSE` (default), constantly predicts `mean(y)` as response
#' and `sd(y)` as standard error.
#' If `robust` is `TRUE`, [median()] and [mad()] are used instead of [mean()] and [sd()],
#' respectively.
#'
#' For weighted data, the response is the weighted mean (weighted median for robust regression).
#' The predicted standard error is the square root of the weighted variance estimator with bias correction
#' based on effective degrees of freedom:
#' ```
#' sd(y, weights) = sqrt(
#'   sum(weights * (y - weighted.mean(y, weights))^2) /
#'     (sum(weights) - sum(weights ^2) / sum(weights))
#' )
#' ```
#' If `robust` is `TRUE`, the weighted median absolute deviation is used, adjusted by a factor of 1.4826
#' for consistency with [mad()].
#'
#' @templateVar id regr.featureless
#' @template learner
#'
#' @template seealso_learner
#' @export
LearnerRegrFeatureless = R6Class(
  "LearnerRegrFeatureless",
  inherit = LearnerRegr,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        robust = p_lgl(default = TRUE, tags = "train")
      )
      ps$set_values(robust = FALSE)

      super$initialize(
        id = "regr.featureless",
        feature_types = unname(mlr_reflections$task_feature_types),
        predict_types = c("response", "se", "quantiles"),
        param_set = ps,
        properties = c("featureless", "missings", "importance", "selected_features", "weights"),
        packages = "stats",
        label = "Featureless Regression Learner",
        man = "mlr3::mlr_learners_regr.featureless"
      )
    },

    #' @description
    #' All features have a score of `0` for this learner.
    #' @return Named `numeric()`.
    importance = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      fn = self$model$features
      named_vector(fn, 0)
    },

    #' @description
    #' Selected features are always the empty set for this learner.
    #' @return `character(0)`.
    selected_features = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      character()
    }
  ),

  private = list(
    .train = function(task) {
      pv = self$param_set$get_values(tags = "train")
      x = task$data(cols = task$target_names)[[1L]]
      weights = private$.get_weights(task)
      quantiles = if (self$predict_type == "quantiles") {
        assert_quantiles(self, quantile_response = TRUE)
        quantile_weighted(x, probs = private$.quantiles, weights = weights)
      }

      if (isFALSE(pv$robust)) {
        wmd = weighted_mean_sd(x, weights)
        location = wmd$mean
        dispersion = wmd$sd
      } else {
        location = quantile_weighted(x, probs = 0.5, weights = weights, continuous = FALSE)
        dispersion = quantile_weighted(abs(x - location), probs = 0.5, weights = weights, continuous = FALSE) * 1.4826
      }

      set_class(
        list(
          location = location,
          dispersion = dispersion,
          quantiles = quantiles,
          features = task$feature_names
        ),
        "regr.featureless_model"
      )
    },

    .predict = function(task) {
      n = task$nrow

      if (self$predict_type == "quantiles") {
        assert_quantiles(self, quantile_response = TRUE)
        quantiles = matrix(rep(self$model$quantiles, n), nrow = n, byrow = TRUE)
        setattr(quantiles, "probs", private$.quantiles)
        setattr(quantiles, "response", private$.quantile_response)
        return(list(quantiles = quantiles))
      }

      response = rep(self$model$location, n)
      se = if (self$predict_type == "se") rep(self$model$dispersion, n) else NULL
      list(response = response, se = se)
    }
  )
)

#' @include mlr_learners.R
mlr_learners$add("regr.featureless", function() LearnerRegrFeatureless$new())

test_that("autotest catches error in train", {
  learner = lrn("classif.debug", error_train = 1)
  task = tsk("spam")

  result = run_experiment(task, learner)
  expect_false(result$ok)
  expect_integer(result$seed)
  expect_task(result$task)
  expect_learner(result$learner)
  expect_null(result$prediction)
  expect_null(result$score)
  expect_string(result$error)
})

test_that("autotest catches error in predict", {
  learner = lrn("classif.debug", error_predict = 1)
  task = tsk("spam")

  result = run_experiment(task, learner)
  expect_false(result$ok)
  expect_integer(result$seed)
  expect_task(result$task)
  expect_learner(result$learner)
  expect_null(result$prediction)
  expect_null(result$score)
  expect_string(result$error)
})

test_that("autotest recognizes learner that does not use .get_weights()", {
  learner = R6Class(
    "learner",
    inherit = LearnerClassifDebug,
    private = list(
      .train = function(task) {
        structure(
          list(response = as.character(sample(task$truth(), 1L))),
          class = "classif.debug_model"
        )
      }
    )
  )$new()

  task = tsk("spam")

  result = run_experiment(task, learner)
  expect_false(result$ok)
  expect_string(result$error, pattern = "get_weights was not called")
})

test_that("autotest configure_learner works", {
  task = tsk("spam")
  learner = lrn("classif.debug", error_train = 1)
  # keep only feat_all_* tasks to save time
  exclude = "^feat_single|^missings|^weights|^utf8|^sanity"
  result = run_autotest(learner, 20, exclude)
  expect_false(result$ok)

  # example changes the learner's parameter during the autotest
  cfg_lrn = function(learner, task) {
    if (task$n_features > 1) {
      learner$param_set$set_values(error_train = 0)
    }
  }
  result = run_autotest(learner, 10, exclude = exclude, configure_learner = cfg_lrn)
  expect_true(result)
  expect_equal(learner$param_set$values$error_train, 1)
})

test_that("autotest on marshal / unmarshal", {
  learner = R6Class(
    "learner_broken_marshal",
    inherit = LearnerClassifDebug,
    public = list(
      marshal = function(...) {
        # do nothing
        invisible(self)
      }
    )
  )$new()

  task = tsk("spam")$clone(deep = TRUE)
  task$id = "feat_all_spam"
  result = run_experiment(task, learner)
  expect_false(result$ok)
  expect_string(result$error, pattern = "marshal")

  learner = R6Class(
    "learner_broken_marshal_class",
    inherit = LearnerClassifDebug,
    public = list(
      marshal = function(...) {
        # miss to set class "marshaled" in the model
        self$model = structure(
          list(
            marshaled = self$model,
            packages = "mlr3"
          ),
          class = c("classif.debug_model_marshaled")
        )
      }
    )
  )$new()

  task = tsk("spam")$clone(deep = TRUE)
  task$id = "feat_all_spam"
  result = run_experiment(task, learner)
  expect_false(result$ok)
  expect_string(result$error, pattern = "marshal")

  learner = R6Class(
    "learner_broken_marshal",
    inherit = LearnerClassifDebug,
    public = list(
      unmarshal = function(...) {
        # do nothing
        invisible(self)
      }
    )
  )$new()

  task = tsk("spam")$clone(deep = TRUE)
  task$id = "feat_all_spam"
  result = run_experiment(task, learner)
  expect_false(result$ok)
  expect_string(result$error, pattern = "unmarshal")

  learner = R6Class(
    "learner_broken_marshal",
    inherit = LearnerClassifDebug,
    private = list(
      .predict = function(task) {
        # simulate a broken unmarshaled model
        if (!is.null(self$model$marshal_count) && self$model$marshal_count) {
          stop("unmarshaled model is broken")
        }
        super$.predict(task)
      }
    )
  )$new()
  learner$param_set$set_values(count_marshaling = TRUE)

  task = tsk("spam")$clone(deep = TRUE)
  task$id = "feat_all_spam"
  result = run_experiment(task, learner)
  expect_false(result$ok)
  expect_string(result$error, pattern = "unmarshaled model is broken")
})

test_that("autotest on encapsulation", {
  # error in train
  learner1 = R6Class(
    "learner_broken_marshal",
    inherit = LearnerClassifDebug,
    private = list(
      .train = function(task) {
        if (Sys.getenv("in_mirai") == "TRUE") {
          stop("Error in mirai process in train")
        }
        super$.train(task)
      }
    )
  )$new()
  task = tsk("spam")
  task$id = "feat_all_spam"

  with_mirai(
    {
      mirai::everywhere(
        {
          Sys.setenv(in_mirai = "TRUE")
        },
        .compute = "mlr3_encapsulation"
      )
      result = run_experiment(task, learner1)
    },
    compute = "mlr3_encapsulation"
  )

  expect_false(result$ok)
  expect_string(result$error, pattern = "Error in mirai process in train")

  # error in predict
  learner2 = R6Class(
    "learner_broken_marshal",
    inherit = LearnerClassifDebug,
    private = list(
      .predict = function(task) {
        if (Sys.getenv("in_mirai") == "TRUE") {
          stop("Error in mirai process in predict")
        }
        super$.predict(task)
      }
    )
  )$new()

  task = tsk("spam")
  task$id = "feat_all_spam"

  with_mirai(
    {
      mirai::everywhere(
        {
          Sys.setenv(in_mirai = "TRUE")
        },
        .compute = "mlr3_encapsulation"
      )
      result = run_experiment(task, learner2)
    },
    compute = "mlr3_encapsulation"
  )

  expect_false(result$ok)
  expect_string(result$error, pattern = "Error in mirai process in predict")

  # access on state
  learner3 = R6Class(
    "learner_broken_marshal",
    inherit = LearnerClassifDebug,
    private = list(
      .predict = function(task) {
        if (Sys.getenv("in_mirai") == "TRUE") {
          self$state$train_task$levels()
          super$.predict(task)
        }
        super$.predict(task)
      }
    )
  )$new()

  task = tsk("spam")
  task$id = "feat_all_spam"

  with_mirai(
    {
      mirai::everywhere(
        {
          Sys.setenv(in_mirai = "TRUE")
        },
        .compute = "mlr3_encapsulation"
      )
      result = run_experiment(task, learner3)
    },
    compute = "mlr3_encapsulation"
  )

  expect_false(result$ok)
  expect_string(result$error, pattern = "attempt to apply non-function")
})

#' @title Circle Classification Task Generator
#'
#' @name mlr_task_generators_circle
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] for the circle binary classification task in [mlbench::mlbench.circle()].
#' Creates a large circle containing a smaller circle.
#'
#' @templateVar id circle
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("circle")
#' plot(generator, n = 200)
#'
#' task = generator$generate(200)
#' str(task$data())
TaskGeneratorCircle = R6Class(
  "TaskGeneratorCircle",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        d = p_int(2L, default = 2L)
      )

      super$initialize(
        id = "circle",
        "classif",
        "mlbench",
        ps,
        label = "Circle Classification",
        man = "mlr3::mlr_task_generators_circle"
      )
    },

    #' @description
    #' Creates a simple plot of generated data.
    #' @param n (`integer(1)`)\cr
    #'   Number of samples to draw for the plot. Default is 200.
    #' @param pch (`integer(1)`)\cr
    #'   Point char. Passed to [plot()].
    #' @param ... (any)\cr
    #'   Additional arguments passed to [plot()].
    plot = function(n = 200L, pch = 19L, ...) {
      plot(private$.generate_obj(n), pch = pch, ...)
    }
  ),

  private = list(
    .generate_obj = function(n) {
      invoke(mlbench::mlbench.circle, n = n, .args = self$param_set$values, .opts = allow_partial_matching)
    },

    .generate = function(n) {
      obj = private$.generate_obj(n)
      TaskClassif$new(sprintf("%s_%i", self$id, n), convert_mlbench(obj), target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("circle", function() TaskGeneratorCircle$new())

#' @include DataBackend.R
DataBackendRbind = R6Class(
  "DataBackendRbind",
  inherit = DataBackend,
  cloneable = FALSE,
  public = list(
    initialize = function(b1, b2) {
      assert_backend(b1)
      assert_backend(b2)
      pk = b1$primary_key

      if (pk != b2$primary_key) {
        error_input("All backends to rbind must have the primary_key '%s'", pk)
      }

      super$initialize(list(b1 = b1, b2 = b2), pk)
    },

    data = function(rows, cols) {
      pk = self$primary_key
      qrows = unique(assert_numeric(rows))
      qcols = union(assert_names(cols, type = "unique"), pk)

      data = private$.data$b2$data(qrows, qcols)
      if (nrow(data) < length(qrows)) {
        qrows = setdiff(rows, data[[pk]])
        tmp = private$.data$b1$data(qrows, qcols)
        data = rbindlist(list(data, tmp), use.names = TRUE, fill = TRUE)
      }

      # duplicate rows / reorder columns
      ijoin(data, rows, intersect(cols, names(data)), pk)
    },

    head = function(n = 6L) {
      rows = head(self$rownames, n)
      self$data(rows, cols = self$colnames)
    },

    distinct = function(rows, cols, na_rm = TRUE) {
      cols = intersect(cols, self$colnames)
      d1 = private$.data$b1$distinct(rows, cols, na_rm = na_rm)
      d2 = private$.data$b2$distinct(rows, cols, na_rm = na_rm)
      set_names(map(cols, function(nn) union(d1[[nn]], d2[[nn]])), cols)
    },

    missings = function(rows, cols) {
      cols = intersect(cols, self$colnames)
      m1 = as.list(private$.data$b1$missings(rows, cols))
      m2 = as.list(private$.data$b2$missings(rows, cols))
      set_names(map_int(cols, function(nn) m1[[nn]] %??% 0L + m2[[nn]] %??% 0L), cols)
    }
  ),

  active = list(
    rownames = function() {
      union(private$.data$b1$rownames, private$.data$b2$rownames)
    },

    colnames = function() {
      union(private$.data$b1$colnames, private$.data$b2$colnames)
    },

    nrow = function() {
      uniqueN(c(private$.data$b1$rownames, private$.data$b2$rownames))
    },

    ncol = function() {
      uniqueN(c(private$.data$b1$colnames, private$.data$b2$colnames))
    }
  ),

  private = list(
    .calculate_hash = function() {
      data = private$.data
      calculate_hash(data$b1$hash, data$b2$hash)
    }
  )
)

test_that("DataBackendRbind", {
  data = as.data.table(iris)
  data$Petal.Length[91:120] = NA
  data$id = 1:150

  b1 = as_data_backend(data[1:100, ], primary_key = "id")
  b2 = as_data_backend(data[101:150, ], primary_key = "id")
  b = DataBackendRbind$new(b1, b2)
  expect_backend(b)
  expect_iris_backend(b, n_missing = 30L)

  expect_set_equal(b$rownames, 1:150)
  expect_set_equal(b$colnames, names(data))
  expect_data_table(b$data(b$rownames, b$colnames), nrows = 150, ncols = 6)
  expect_set_equal(b$distinct(b$rownames, "Species")$Species, distinct_values(iris$Species, drop = FALSE))

  x = b$missings(b$rownames, c("Petal.Width", "Petal.Length"))
  expect_equal(x, set_names(c(0L, 30L), c("Petal.Width", "Petal.Length")))

  # all col-hashes are mutually disjoint
  expect_length(intersect(b1$col_hashes, b2$col_hashes), 0)
  expect_length(intersect(b$col_hashes, b1$col_hashes), 0)
  expect_length(intersect(b$col_hashes, b2$col_hashes), 0)
})

test_that("Backends with different cols", {
  data = as.data.table(iris)
  data$id = 1:150
  fn = c("Sepal.Width", "Species", "id")
  b1 = as_data_backend(data[1:100, c("Sepal.Length", "Sepal.Width", "Species", "id"), with = FALSE], primary_key = "id")
  b2 = as_data_backend(data[101:150, fn, with = FALSE], primary_key = "id")
  b = DataBackendRbind$new(b1, b2)

  expect_equal(b$ncol, 4)

  h = b$head(60)
  expect_data_table(h, ncols = 4, nrows = 60)

  h = b$head(120)
  expect_data_table(h, ncols = 4, nrows = 120)

  expect_data_table(b$data(rows = 1:120, cols = b$colnames), nrows = 120, ncols = 4)
})

test_that("Backends with same rows", {
  data = as.data.table(iris)
  data$id = 1:150
  data1 = data[1:20]
  data2 = data[6:10][, Petal.Length := NA]
  data2 = rbind(data2, data[101])

  b1 = as_data_backend(data1, primary_key = "id")
  b2 = as_data_backend(data2, primary_key = "id")

  b = DataBackendRbind$new(b1, b2)
  expect_backend(b)
  expect_set_equal(b$rownames, c(1:20, 101))

  data = b$data(b$rownames[1:20], b$colnames)
  expect_false(anyMissing(data$Petal.Length[1:5]))
  expect_false(anyMissing(data$Petal.Length[11:20]))
  expect_true(allMissing(data$Petal.Length[6:10]))
})

test_that("LearnerClassif predict_newdata_fast response works", {
  learner = lrn("classif.debug")
  task = tsk("pima")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "prob"))
  expect_character(pred$response)
})

test_that("LearnerClassif predict_newdata_fast prob works", {
  learner = lrn("classif.debug", predict_type = "prob")
  task = tsk("pima")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "prob"))
  expect_matrix(pred$prob, nrows = nrow(newdata), ncols = length(task$class_names))
})

test_that("LearnerClassif predict_newdata_fast works with missing values", {
  learner = lrn("classif.debug", predict_missing = 0.5)
  learner$encapsulate("evaluate", fallback = lrn("classif.featureless"))
  task = tsk("pima")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "prob"))
  expect_character(pred$response, any.missing = FALSE)

  learner = lrn("classif.debug", predict_missing = 0.5, predict_type = "prob")
  learner$encapsulate("evaluate", fallback = lrn("classif.featureless", predict_type = "prob"))
  task = tsk("pima")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "prob"))
  expect_matrix(pred$prob, nrows = nrow(newdata), ncols = length(task$class_names), any.missing = FALSE)
})

test_that("LearnerClassif predict_newdata_fast works with failed train", {
  learner = lrn("classif.debug", predict_missing = 0.5, error_train = 1)
  learner$encapsulate("evaluate", fallback = lrn("classif.featureless"))
  task = tsk("pima")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "prob"))
  expect_character(pred$response, any.missing = FALSE)

  learner = lrn("classif.debug", predict_missing = 0.5, predict_type = "prob", error_train = 1)
  learner$encapsulate("evaluate", fallback = lrn("classif.featureless", predict_type = "prob"))
  task = tsk("pima")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "prob"))
  expect_matrix(pred$prob, nrows = nrow(newdata), ncols = length(task$class_names), any.missing = FALSE)
})

test_that("Nested backends", {
  data = as.data.table(iris)
  data$Petal.Length[91:120] = NA
  data$id = 1:150

  b1 = as_data_backend(data[1:100, -"Sepal.Length"], primary_key = "id")
  b2 = as_data_backend(data[101:130, -"Sepal.Length"], primary_key = "id")
  b3 = DataBackendRbind$new(b1, b2)
  expect_backend(b3)

  b4 = as_data_backend(data[131:150, -"Sepal.Length"], primary_key = "id")
  b5 = DataBackendRbind$new(b3, b4)
  expect_backend(b5)

  b6 = as_data_backend(data[, c("id", "Sepal.Length")], primary_key = "id")
  b7 = DataBackendCbind$new(b5, b6)
  expect_backend(b7)

  expect_iris_backend(b7, n_missing = 30L)

  x = b7$missings(b7$rownames, c("Petal.Width", "Petal.Length"))
  expect_equal(x, set_names(c(0L, 30L), c("Petal.Width", "Petal.Length")))
})

test_that("convert_task - Regr -> Regr", {
  task = tsk("california_housing")
  result = convert_task(task, target = "households", drop_original_target = TRUE)

  expect_class(result, "TaskRegr")
  expect_task(result)
  expect_true(result$col_roles$target == "households")
  expect_true(all(result$feature_names != "households"))
  expect_true(all(result$feature_names != "median_house_value"))
  expect_true(all(unlist(imap(result$row_roles, .f = function(z, x) {
    all(result$row_roles[[x]] == task$row_roles[[x]])
  }))))
  expect_true(
    every(c("weights_learner", "groups", "strata", "nrow"), function(x) {
      all(result[[x]] == task[[x]])
    })
  )
})

test_that("convert_task - Regr -> Classif", {
  task = tsk("california_housing")
  result = convert_task(task, target = "ocean_proximity", new_type = "classif", drop_original_target = TRUE)

  expect_class(result, "TaskClassif")
  expect_task(result)
  expect_true(result$col_roles$target == "ocean_proximity")
  expect_true(all(result$feature_names != "ocean_proximity"))
  expect_true(all(result$feature_names != "median_house_value"))
  expect_true(all(unlist(imap(result$row_roles, .f = function(z, x) {
    all(result$row_roles[[x]] == task$row_roles[[x]])
  }))))
  expect_true(
    every(c("weights_learner", "groups", "strata", "nrow"), function(x) {
      all(result[[x]] == task[[x]])
    })
  )
})

test_that("convert_task - Classif -> Regr", {
  task = tsk("iris")
  result = convert_task(task, target = "Sepal.Width", new_type = "regr", drop_original_target = TRUE)

  expect_class(result, "TaskRegr")
  expect_task(result)
  expect_true(result$col_roles$target == "Sepal.Width")
  expect_true(all(result$feature_names != "Sepal.Width"))
  expect_true(all(result$feature_names != "Species"))
  expect_true(all(unlist(imap(result$row_roles, .f = function(z, x) {
    all(result$row_roles[[x]] == task$row_roles[[x]])
  }))))
  expect_true(
    every(c("weights_learner", "groups", "strata", "nrow"), function(x) {
      all(result[[x]] == task[[x]])
    })
  )
})

test_that("convert_task - same target", {
  task = tsk("california_housing")
  task$col_roles$feature = setdiff(task$col_roles$feature, "latitue")

  results = list(
    convert_task(task, target = "median_house_value", new_type = "regr", drop_original_target = TRUE),
    convert_task(task, target = "median_house_value", new_type = "regr", drop_original_target = FALSE)
  )

  for (result in results) {
    expect_class(result, "TaskRegr")
    expect_task(result)
    expect_true(result$col_roles$target == "median_house_value")
    expect_true(all(unlist(imap(result$row_roles, .f = function(z, x) {
      all(result$row_roles[[x]] == task$row_roles[[x]])
    }))))
    expect_true(
      every(
        c("weights_learner", "groups", "strata", "nrow", "ncol", "feature_names", "target_names", "task_type"),
        function(x) {
          all(result[[x]] == task[[x]])
        }
      )
    )
  }
})

test_that("convert task - general checks", {
  btask = tsk("california_housing")
  itask = tsk("iris")

  # target does not exist
  expect_error(convert_task(btask, target = "median_house_value2"))

  # target class does not match
  expect_error(convert_task(btask, target = "latitude", new_type = "classif"))
  expect_error(convert_task(itask, target = "Sepal.Width", new_type = "classif"))
})

test_that("convert_task reconstructs task", {
  task = tsk("iris")
  tsk = convert_task(task)
  tsk$man = "mlr3::mlr_tasks_iris"
  # TODO: re-enable after task$weights has been removed
  # expect_equal(task, tsk, ignore_attr = TRUE)

  task2 = task$filter(1:100)
  tsk2 = convert_task(task2)
  # TODO: re-enable after task$weights has been removed
  # expect_equal(task2$nrow, tsk2$nrow)
  # expect_equal(task2$ncol, tsk2$ncol)
  expect_true("twoclass" %chin% tsk2$properties)

  task3 = task2
  task3$row_roles$use = 1:150
  tsk3 = convert_task(task3)
  tsk3$man = "mlr3::mlr_tasks_iris"
  # TODO: re-enable after task$weights has been removed
  # expect_equal(task3$nrow, tsk3$nrow)
  # expect_equal(task3$ncol, tsk3$ncol)
  expect_true("multiclass" %chin% tsk3$properties)
  # TODO: re-enable after task$weights has been removed
  # expect_equal(task, tsk3, ignore_attr = TRUE)
})

test_that("extra args survive the roundtrip", {
  mytask = tsk("sonar")
  expect_equal(mytask$extra_args, list(positive = "M"))

  mytask = convert_task(mytask, target = "V1", new_type = "regr")
  expect_equal(mytask$extra_args, list(positive = "M"))

  mytask = convert_task(mytask, target = "Class", new_type = "classif")
  expect_equal(mytask$extra_args, list(positive = "M"))
  expect_equal(mytask$positive, "M")

  mytask$positive = mytask$negative
  expect_equal(mytask$extra_args, list(positive = "R"))

  mytask = convert_task(mytask, target = "V1", new_type = "regr")
  expect_equal(mytask$extra_args, list(positive = "R"))

  mytask = convert_task(mytask, target = "Class", new_type = "classif")
  expect_equal(mytask$extra_args, list(positive = "R"))
  expect_equal(mytask$positive, "R")
})

test_that("convert_task preserves internal_valid_task", {
  task = tsk("california_housing")
  task$internal_valid_task = sample(task$nrow, 10)
  expect_true(!is.null(task$internal_valid_task))

  result = convert_task(task, target = "households")
  expect_equal(result$internal_valid_task$target_names, "households")

  # also works with type conversion
  result2 = convert_task(task, target = "ocean_proximity", new_type = "classif")
  expect_equal(result2$internal_valid_task$target_names, "ocean_proximity")
  expect_class(result2$internal_valid_task, "TaskClassif")
})

test_that("as_task_classif.TaskRegr forwards drop_original_target and drop_levels", {
  task = tsk("california_housing")

  keep = as_task_classif(task, target = "ocean_proximity", drop_original_target = FALSE)
  expect_class(keep, "TaskClassif")
  expect_true("median_house_value" %chin% keep$feature_names)

  drop = as_task_classif(task, target = "ocean_proximity", drop_original_target = TRUE)
  expect_true("median_house_value" %nin% drop$feature_names)
})

test_that("as_task_regr.TaskClassif forwards drop_original_target and drop_levels", {
  task = tsk("iris")

  keep = as_task_regr(task, target = "Sepal.Width", drop_original_target = FALSE)
  expect_class(keep, "TaskRegr")
  expect_true("Species" %chin% keep$feature_names)

  drop = as_task_regr(task, target = "Sepal.Width", drop_original_target = TRUE)
  expect_true("Species" %nin% drop$feature_names)
})

test_that("data.frame converters", {
  data("mtcars", package = "datasets")
  task = as_task_regr(mtcars, "mpg")
  expect_task_regr(task)
  expect_equal(task$id, "mtcars")

  data("iris", package = "datasets")
  task = as_task_classif(iris, "Species")
  expect_task_classif(task)
  expect_equal(task$id, "iris")

  task = as_task_unsupervised(mtcars)
  expect_task_unsupervised(task)
})

test_that("formula converters", {
  data("mtcars", package = "datasets")
  task = as_task_regr(mpg ~ am + carb, mtcars)
  expect_task_regr(task)
  expect_equal(task$id, "mtcars")
  expect_set_equal(task$feature_names, c("am", "carb"))
  expect_equal(task$target_names, "mpg")

  my_mtcars = mtcars
  my_mtcars$hp[5] = NA
  task = as_task_regr(mpg ~ ., my_mtcars)
  expect_task_regr(task)
  expect_equal(task$id, "my_mtcars")
  expect_set_equal(task$feature_names, setdiff(names(my_mtcars), "mpg"))
  expect_equal(task$target_names, "mpg")
  expect_equal(task$nrow, nrow(my_mtcars))

  expect_error(as_task_regr(mpg ~ am + x, mtcars), "Assertion on 'formula' failed")
  expect_error(as_task_regr(~ am + carb, mtcars), "is missing a response")

  data("iris", package = "datasets")
  task = as_task_classif(Species ~ Sepal.Length + Sepal.Width, iris)
  expect_task_classif(task)
  expect_equal(task$id, "iris")
  expect_set_equal(task$feature_names, c("Sepal.Length", "Sepal.Width"))
  expect_equal(task$target_names, "Species")

  data("penguins", package = "palmerpenguins")
  task = as_task_classif(species ~ ., penguins)
  expect_task_classif(task)
  expect_equal(task$id, "penguins")
  expect_set_equal(task$feature_names, setdiff(names(penguins), "species"))
  expect_equal(task$target_names, "species")
  expect_equal(task$nrow, nrow(penguins))

  expect_error(as_task_classif(Species ~ Sepal.Length + x, iris), "Assertion on 'formula' failed")
  expect_error(as_task_classif(~ Sepal.Length + Sepal.Width, iris), "is missing a response")
})

test_that("matrix converters", {
  X = matrix(1:9, nrow = 3)
  colnames(X) = letters[1:3]
  expect_task(as_task_regr(X, target = "a"))
  expect_task(as_task_classif(X, target = "a"))
})

test_that("convert_task - Regr -> Regr with weights", {
  task = cars_weights_learner
  result = convert_task(task, target = "speed", drop_original_target = FALSE)

  expect_equal(result$weights_learner, task$weights_learner)

  task = cars_weights_measure
  result = convert_task(task, target = "speed", drop_original_target = FALSE)

  expect_equal(result$weights_measure, task$weights_measure)
})

test_that("LearnerRegr predict_newdata_fast response works", {
  learner = lrn("regr.debug")
  task = tsk("mtcars")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "se", "quantiles"))
  expect_numeric(pred$response)
})

test_that("LearnerRegr predict_newdata_fast se works", {
  learner = lrn("regr.debug", predict_type = "se")
  task = tsk("mtcars")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "se", "quantiles"))
  expect_numeric(pred$se)
})

test_that("LearnerRegr predict_newdata_fast quantiles works", {
  learner = lrn("regr.debug", predict_type = "quantiles")
  learner$quantiles = c(0.1, 0.5, 0.9)
  learner$quantile_response = 0.5
  task = tsk("mtcars")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "se", "quantiles"))
  expect_matrix(pred$quantiles, nrows = nrow(newdata), ncols = length(learner$quantiles))
})

test_that("LearnerRegr predict_newdata_fast works with missing values", {
  learner = lrn("regr.debug", predict_missing = 0.5)
  learner$encapsulate("evaluate", fallback = lrn("regr.featureless"))
  task = tsk("mtcars")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "se", "quantiles"))
  expect_numeric(pred$response, any.missing = FALSE)

  learner = lrn("regr.debug", predict_missing = 0.5, predict_type = "se")
  learner$encapsulate("evaluate", fallback = lrn("regr.featureless", predict_type = "se"))
  task = tsk("mtcars")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "se", "quantiles"))
  expect_numeric(pred$se, any.missing = FALSE)

  learner = lrn("regr.debug", predict_missing = 0.5, predict_type = "quantiles")
  learner$quantiles = c(0.1, 0.5, 0.9)
  learner$quantile_response = 0.5
  fallback = lrn("regr.featureless", predict_type = "quantiles")
  fallback$quantiles = c(0.1, 0.5, 0.9)
  fallback$quantile_response = 0.5
  learner$encapsulate("evaluate", fallback = fallback)
  task = tsk("mtcars")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "se", "quantiles"))
  expect_matrix(pred$quantiles, nrows = nrow(newdata), ncols = length(learner$quantiles), any.missing = FALSE)
})

test_that("LearnerRegr predict_newdata_fast works with failed train", {
  learner = lrn("regr.debug", predict_missing = 0.5, error_train = 1)
  learner$encapsulate("evaluate", fallback = lrn("regr.featureless"))
  task = tsk("mtcars")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "se", "quantiles"))
  expect_numeric(pred$response, any.missing = FALSE)

  learner = lrn("regr.debug", predict_missing = 0.5, predict_type = "se", error_train = 1)
  learner$encapsulate("evaluate", fallback = lrn("regr.featureless", predict_type = "se"))
  task = tsk("mtcars")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)

  expect_list(pred)
  expect_names(names(pred), subset.of = c("response", "se", "quantiles"))
  expect_numeric(pred$se, any.missing = FALSE)

  learner = lrn("regr.debug", predict_missing = 0.5, predict_type = "quantiles", error_train = 1)
  learner$quantiles = c(0.1, 0.5, 0.9)
  learner$quantile_response = 0.5
  fallback = lrn("regr.featureless", predict_type = "quantiles")
  fallback$quantiles = c(0.1, 0.5, 0.9)
  fallback$quantile_response = 0.5
  learner$encapsulate("evaluate", fallback = fallback)
  task = tsk("mtcars")
  newdata = task$data()
  learner$train(task)
  pred = learner$predict_newdata_fast(newdata)
})

test_that("Classification measures", {
  keys = mlr_measures$keys("^classif\\.")
  task = tsk("sonar")
  learner = lrn("classif.rpart", predict_type = "prob")
  learner$train(task)
  p = learner$predict(task)

  for (key in keys) {
    m = mlr_measures$get(key)
    if (is.na(m$task_type) || m$task_type == "classif") {
      if (key == "classif.costs") {
        costs = 1 - diag(length(task$class_names))
        rownames(costs) = colnames(costs) = task$class_names
        m$costs = costs
      }
      perf = m$score(prediction = p, task = task, learner = learner)
      expect_number(perf, na.ok = FALSE, lower = m$range[1], upper = m$range[2])
    }
  }
})

test_that("DataBackendDataTable construction", {
  b = as_data_backend(iris)
  expect_backend(b)
  expect_iris_backend(b)

  i = 1:30
  data = iris[i, ]
  data$id = i
  b = as_data_backend(data, primary_key = "id")
  expect_backend(b)
  expect_set_equal(b$rownames, i)

  rownames(data) = sprintf("rn_%i", data$id)
  data$id = NULL
  b = as_data_backend(data, keep_rownames = TRUE)
  expect_backend(b)
  expect_set_equal(b$data(i, "..rownames")[[1]], sprintf("rn_%i", i))

  data$Petal.Length[21:30] = NA
  b = as_data_backend(data)
  x = b$missings(b$rownames, c("Petal.Width", "Petal.Length"))
  expect_equal(x, set_names(c(0L, 10L), c("Petal.Width", "Petal.Length")))

  b = as_data_backend(iris, primary_key = 151:300)
  expect_equal(b$rownames, 151:300)
})

test_that("DataBackendDataTable with 0 rows", {
  b = as_data_backend(iris[integer(), ])
  expect_backend(b)
})

test_that("DataBackendRename", {
  old = names(iris)
  new = paste0("xx_", old)
  b1 = as_data_backend(iris)
  b = DataBackendRename$new(b1, old, new)
  expect_backend(b)

  expect_set_equal(b$colnames, c(new, "..row_id"))
  expect_names(names(b$data(b$rownames, b$colnames)), identical.to = c(new, "..row_id"))
  expect_names(names(b$head()), identical.to = c(new, "..row_id"))

  expect_data_table(b$data(rows = b$rownames, cols = old), ncols = 0)
  expect_data_table(b$data(rows = b$rownames, cols = new), ncols = 5)

  expect_equal(unname(b1$col_hashes), unname(b$col_hashes))
})

test_that("DataBackendRename / partial rename", {
  old = names(iris)[1:2]
  new = paste0("xx_", old)
  b1 = as_data_backend(iris)
  b = DataBackendRename$new(b1, old, new)
  expect_backend(b)

  expected = c(map_values(names(iris), old, new), "..row_id")
  expect_set_equal(b$colnames, expected)

  expect_names(names(b$data(b$rownames, b$colnames)), identical.to = expected)
  expect_names(names(b$head()), identical.to = expected)

  expect_data_table(b$data(rows = b$rownames, cols = expected), ncols = 6)

  expect_equal(unname(b1$col_hashes), unname(b$col_hashes))
})

test_that("nested backends", {
  data = as.data.table(iris)
  b1 = as_data_backend(data)
  b2 = DataBackendRename$new(as_data_backend(data), old = names(iris), new = paste0("new_", names(iris)))
  b = DataBackendCbind$new(b1, b2)
  expect_backend(b1)
  expect_backend(b2)
  expect_backend(b)

  expect_set_equal(b$rownames, 1:150)
  expect_set_equal(b$colnames, c(b1$colnames, b2$colnames))

  expect_equal(b$col_hashes, c(b1$col_hashes, b2$col_hashes)[setdiff(b$colnames, b$primary_key)])

  expect_set_equal(b$col_hashes, b1$col_hashes) # b2 has the same hashes, so cbind should have duplicate hashes.
})

test_that("rename does not yield duplicated columns (#701)", {
  db = DataBackendDataTable$new(data.table(id = 1L, a = 1, b = 2), primary_key = "id")
  expect_error(DataBackendRename$new(db, "a", "b"), "duplicated", ignore.case = TRUE)
})

#' @title Iris Classification Task
#'
#' @name mlr_tasks_iris
#' @format [R6::R6Class] inheriting from [TaskClassif].
#' @include mlr_tasks.R
#'
#' @description
#' A classification task for the popular [datasets::iris] data set.
#'
#' @templateVar id iris
#' @template task
#'
#' @source
#' \url{https://en.wikipedia.org/wiki/Iris_flower_data_set}
#'
#' `r format_bib("anderson_1936")`
#'
#' @template seealso_task
NULL

load_task_iris = function(id = "iris") {
  b = as_data_backend(load_dataset("iris", "datasets"))
  task = TaskClassif$new(id, b, target = "Species", label = "Iris Flowers")
  b$hash = task$man = "mlr3::mlr_tasks_iris"
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("iris", load_task_iris)

#' @title Convert to a Classification Prediction
#'
#' @description
#' Convert object to a [PredictionClassif].
#'
#' @inheritParams as_prediction
#'
#' @return [PredictionClassif].
#' @export
#' @examples
#' # create a prediction object
#' task = tsk("penguins")
#' learner = lrn("classif.rpart", predict_type = "prob")
#' learner$train(task)
#' p = learner$predict(task)
#'
#' # convert to a data.table
#' tab = as.data.table(p)
#'
#' # convert back to a Prediction
#' as_prediction_classif(tab)
#'
#' # split data.table into a list of data.tables
#' tabs = split(tab, tab$truth)
#'
#' # convert back to list of predictions
#' preds = lapply(tabs, as_prediction_classif)
#'
#' # calculate performance in each group
#' sapply(preds, function(p) p$score())
as_prediction_classif = function(x, ...) {
  UseMethod("as_prediction_classif")
}


#' @rdname as_prediction_classif
#' @export
# nolint next
as_prediction_classif.PredictionClassif = function(x, ...) {
  x
}


#' @rdname as_prediction_classif
#' @export
# nolint next
as_prediction_classif.data.frame = function(x, ...) {
  assert_names(names(x), must.include = c("row_ids", "truth", "response"))
  prob_cols = setdiff(names(x), c("row_ids", "truth", "response", "weights"))
  if (!all(startsWith(prob_cols, "prob."))) {
    error_input(
      # nolint next
      "Table may only contain columns 'row_ids', 'truth', 'response', 'weights' as well as columns prefixed with 'prob.' for class probabilities"
    )
  }

  x = as.data.table(x)
  if (length(prob_cols)) {
    prob = as.matrix(x[, prob_cols, with = FALSE])
    colnames(prob) = substr(colnames(prob), 6L, nchar(colnames(prob)))
  } else {
    prob = NULL
  }

  invoke(PredictionClassif$new, prob = prob, .args = x[, -prob_cols, with = FALSE])
}

#' @title Stack for Hot Start Learners
#'
#' @description
#' This class stores learners for hot starting training, i.e. resuming or
#' continuing from an already fitted model.
#' We assume that hot starting is only possible if a single hyperparameter
#' (also called the fidelity parameter, usually controlling the complexity or
#' expensiveness) is altered and all other hyperparameters are identical.
#'
#' The `HotstartStack` stores trained learners which can be potentially used to
#' hot start a learner. Learner automatically hot start while training if a
#' stack is attached to the `$hotstart_stack` field and the stack contains a
#' suitable learner.
#'
#' For example, if you want to train a random forest learner with 1000 trees but
#' already have a random forest learner with 500 trees (hot start learner),
#' you can add the hot start learner to the `HotstartStack` of the expensive learner
#' with 1000 trees. If you now call the `train()` method (or [resample()] or
#' [benchmark()]), a random forest with 500 trees will be fitted and combined
#' with the 500 trees of the hotstart learner, effectively saving you to
#' fit 500 trees.
#'
#' Hot starting is only supported by learners which have the property
#' `"hotstart_forward"` or `"hotstart_backward"`. For example, an `xgboost` model
#' (in \CRANpkg{mlr3learners}) can hot start forward by adding more boosting
#' iterations, and a random forest can go backwards by removing trees.
#' The fidelity parameters are tagged with `"hotstart"` in learner's parameter set.
#'
#' @export
#' @examples
#' # train learner on pima task
#' task = tsk("pima")
#' learner = lrn("classif.debug", iter = 1)
#' learner$train(task)
#'
#' # initialize stack with previously fitted learner
#' hot = HotstartStack$new(list(learner))
#'
#' # retrieve learner with increased fidelity parameter
#' learner = lrn("classif.debug", iter = 2)
#'
#' # calculate cost of hot starting
#' hot$start_cost(learner, task$hash)
#'
#' # add stack with hot start learner
#' learner$hotstart_stack = hot
#'
#' # train automatically uses hot start learner while fitting the model
#' learner$train(task)
HotstartStack = R6Class(
  "HotstartStack",
  public = list(
    #' @field stack [data.table::data.table()]\cr
    #' Stores hot start learners.
    stack = NULL,

    #' @field hotstart_threshold (named `numeric(1)`)\cr
    #' Threshold for storing learners in the stack.
    #' If the value of the hotstart parameter is below this threshold, the learner is not added to the stack.
    hotstart_threshold = NULL,

    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param learners (List of [Learner]s)\cr
    #'   Learners are added to the hotstart stack. If `NULL` (default), empty
    #'   stack is created.
    #' @param hotstart_threshold (named `numeric(1)`)\cr
    #'   Threshold for storing learners in the stack.
    initialize = function(learners = NULL, hotstart_threshold = NULL) {
      self$stack = data.table()
      self$hotstart_threshold = assert_numeric(hotstart_threshold, names = "named", null.ok = TRUE)

      # add learners to stack
      if (!is.null(learners)) self$add(learners)
    },

    #' @description
    #' Add learners to hot start stack.
    #'
    #' @param learners (List of [Learner]s).
    #'   Learners are added to the hotstart stack.
    #'
    #' @return self (invisibly).
    add = function(learners) {
      learners = assert_learners(as_learners(learners))

      walk(learners, function(learner) {
        if (!is.null(get0("validate", learner))) {
          error_input("Hotstart learners that did validation is currently not supported.")
        } else if (is.null(learner$model)) {
          error_input("Learners must be trained before adding them to the hotstart stack.")
        } else if (is_marshaled_model(learner$model)) {
          error_input("Learners must be unmarshaled before adding them to the hotstart stack.")
        }
      })

      if (!is.null(self$hotstart_threshold)) {
        learners = keep(learners, function(learner) {
          hotstart_id = learner$param_set$ids(tags = "hotstart")
          learner$param_set$values[[hotstart_id]] >= self$hotstart_threshold[hotstart_id]
        })
        if (!length(learners)) return(invisible(self))
      }

      # hashes
      task_hash = map_chr(learners, function(learner) learner$state$task_hash)
      learner_hash = map_chr(learners, learner_hotstart_hash)

      lg$debug("Adding %s learner(s) to the hotstart stack.", length(learners))

      self$stack = rbindlist(list(self$stack, data.table(start_learner = learners, task_hash, learner_hash)))
      setkeyv(self$stack, c("task_hash", "learner_hash"))

      invisible(self)
    },

    #' @description
    #' Calculates the cost for each learner of the stack to hot start the target `learner`.
    #'
    #' The following cost values can be returned:
    #'
    #' * `NA_real_`: Learner is unsuitable to hot start target `learner`.
    #' * `-1`: Hotstart learner in the stack and target `learner` are identical.
    #' * `0` Cost for hot starting backwards is always 0.
    #' * `> 0` Cost for hot starting forward.
    #'
    #' @param learner [Learner]\cr
    #'   Target learner.
    #' @param task_hash [Task]\cr
    #'   Hash of the task on which the target learner is trained.
    #'
    # @return `numeric()`.
    start_cost = function(learner, task_hash) {
      if (!nrow(self$stack)) {
        return(numeric(0))
      }
      .learner_hash = learner_hotstart_hash(assert_learner(learner))
      .task_hash = assert_string(task_hash)
      hotstart_id = learner$param_set$ids(tags = "hotstart")

      set(self$stack, j = "cost", value = NA_real_)
      cost = self$stack[
        list(.task_hash, .learner_hash),
        "cost" := map_dbl(get("start_learner"), function(l) calculate_cost(l, learner, hotstart_id)),
        on = c("task_hash", "learner_hash")
      ][, get("cost")]
      self$stack[, "cost" := NULL]
      cost
    },

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s>", class(self)[1L])
    },

    #' @description
    #' Printer.
    #'
    #' @param ... (ignored).
    print = function(...) {
      cat_cli(cli_h1("{.cls {class(self)[1L]}}"))
      print(self$stack, digits = 2L)
    }
  ),

  private = list(
    # Queries the stack for the start learner with the lowest cost of hot
    # starting the target `learner`. The start learner's state is copied to
    # `learner` and `learner` is returned. This method is internally used by
    # `Learner$train()` `resample()` and `benchmark()` which call
    # `learner_train(learner, task, row_ids, mode = 'retrain')` with the
    # returned learner.
    .start_learner = function(learner, task_hash) {
      if (!nrow(self$stack)) {
        return(NULL)
      }
      .learner_hash = learner_hotstart_hash(assert_learner(learner))
      .task_hash = assert_character(task_hash, len = 1L)
      hotstart_id = learner$param_set$ids(tags = "hotstart")

      start_learner = self$stack[list(.task_hash, .learner_hash), on = c("task_hash", "learner_hash"), nomatch = NULL][,
        "cost" := map_dbl(start_learner, function(l) calculate_cost(l, learner, hotstart_id))
      ][which_min(get("cost"), na_rm = TRUE), start_learner]

      lg$debug("Found %i start learner(s) in hotstart stack of size %i.", length(start_learner), nrow(self$stack))

      if (!length(start_learner)) {
        return(NULL)
      }
      learner$state = start_learner[[1L]]$state
      learner
    }
  )
)

#' @description
#' Calculates cost to hotstart learner.
#' @param start_learner ([Learner]).
#' @param learner ([Learner]).
#' @param hotstart_id (`character(1)`).
#'
#' @return `numeric(1)`.
#' @noRd
calculate_cost = function(start_learner, learner, hotstart_id) {
  if (is.null(start_learner) || !length(hotstart_id)) {
    return(NA_real_)
  }

  cost = learner$param_set$values[[hotstart_id]] - start_learner$param_set$values[[hotstart_id]]
  if (cost == 0) {
    return(-1)
  }

  if ("hotstart_backward" %chin% learner$properties && "hotstart_forward" %chin% learner$properties) {
    if (cost < 0) 0 else cost
  } else if ("hotstart_backward" %chin% learner$properties) {
    if (cost < 0) 0 else NA_real_
  } else {
    if (cost > 0) cost else NA_real_
  }
}

#' @description
#' Hash (unique identifier) for learner object, excluding parameter values
#' tagged with `hotstart`.
#'
#' @param learner [Learner].
#'
#' @return `character(1)`.
#' @noRd
learner_hotstart_hash = function(learner) {
  param_vals = learner$param_set$values
  hotstart_id = learner$param_set$ids(tags = "hotstart")
  train_ids = setdiff(learner$param_set$ids(tags = "train"), hotstart_id)
  train_vals = param_vals[names(param_vals) %in% train_ids]

  calculate_hash(class(learner), learner$id, learner$predict_type, learner$fallback$hash, train_vals)
}

test_that("mlr_tasks", {
  expect_dictionary(mlr_tasks, min_items = 1L)
  keys = mlr_tasks$keys()

  for (key in keys) {
    t = tsk(key)
    expect_task_supervised(t)
  }
})

test_that("load_x", {
  ns = getNamespace("mlr3")
  nn = names(ns)
  nn = nn[startsWith(names(ns), "load_task")]

  for (fun in nn) {
    fun = get(fun, envir = ns, mode = "function")
    expect_task_supervised(fun())
  }
})

test_that("tasks are cloned", {
  if (packageVersion("mlr3misc") >= "0.9.2") {
    task = tsk("iris")
    mlr_tasks$add("foo", task)
    expect_different_address(task, tsk("foo"))
    mlr_tasks$remove("foo")
  }
})

test_that("as.data.table(..., objects = TRUE)", {
  tab = as.data.table(mlr_tasks, objects = TRUE)
  expect_data_table(tab)
  expect_list(tab$object, "Task", any.missing = FALSE)
})

test_that("Dictionary: clone works", {
  t1 = tsk("iris")
  expect_task(t1)
  t2 = tsk("iris")
  expect_task(t2)
  expect_different_address(t1, t2)
})

test_that("$keys(pattern) works", {
  expect_subset(mlr_learners$keys("classif"), mlr_learners$keys(), empty.ok = FALSE)
})

test_that("dictionary to data.table conversion works with prototype arguments", {
  LearnerRegrRpart2 = R6Class(
    "LearnerRegrRpart2",
    inherit = LearnerRegrRpart,
    public = list(
      x = NULL,
      initialize = function(x) {
        self$x = x
        super$initialize()
      }
    )
  )
  on.exit(mlr_learners$remove("regr.rpart2"))
  mlr_learners$add("regr.rpart2", LearnerRegrRpart2, .prototype_args = list(x = 123))

  expect_data_table(as.data.table(mlr_learners))
})

test_that("HotstartStack works with forward target learner and increased hotstart parameter", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 2)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 3)
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(2, 1))
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_2$state)
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with forward target learner when hotstart learner is different", {
  task = tsk("pima")
  learner_1 = lrn("classif.rpart")
  learner_1$train(task)

  learner = lrn("classif.debug", iter = 3)
  hot = HotstartStack$new(list(learner_1))

  expect_equal(hot$start_cost(learner, task$hash), NA_real_)
  expect_data_table(hot$stack, ncols = 3)
  expect_null(get_private(hot)$.start_learner(learner, task$hash))
})

test_that("HotstartStack works with forward target learner when cost of hotstarting is equal", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 2)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 2)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 3)
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(1, 1))
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works when hotstart values are lower and higher for forward target", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 3)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 2)
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(1, NA_real_))
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_1$state)
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works when forward hotstart and target learner are equal", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)

  learner = lrn("classif.debug", iter = 1)
  hot = HotstartStack$new(list(learner_1))

  expect_equal(hot$start_cost(learner, task$hash), -1)
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_1$state)
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with forward target learner when hotstart values are only higher", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 2)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 3)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 1)
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(NA_real_, NA_real_))
  expect_data_table(hot$stack, ncols = 3)
  expect_null(get_private(hot)$.start_learner(learner, task$hash))
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with forward target when learners differ and values vary", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1) # lower
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 2) # equal
  learner_2$train(task)
  learner_3 = lrn("classif.debug", iter = 3) # higher
  learner_3$train(task)
  learner_4 = lrn("classif.rpart") # different learner
  learner_4$train(task)

  learner = lrn("classif.debug", iter = 2)
  hot = HotstartStack$new(list(learner_1, learner_2, learner_3, learner_4))

  expect_equal(hot$start_cost(learner, task$hash), c(1, -1, NA_real_, NA_real_))
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works when forward target learner is unable to hotstart", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 2)
  learner_2$train(task)

  learner = lrn("classif.rpart")
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(NA_real_, NA_real_))
  expect_null(get_private(hot)$.start_learner(learner, task$hash))
})

test_that("HotstartStack works when stack is empty", {
  task = tsk("pima")
  learner = lrn("classif.debug", iter = 2)
  hot = HotstartStack$new()

  expect_equal(hot$start_cost(learner, task$hash), numeric(0))
  expect_null(get_private(hot)$.start_learner(learner, task$hash))
})

test_that("HotstartStack works with backward target learner and decreased hotstart parameter", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 2)
  learner_1$train(task)

  learner = lrn("classif.debug", iter = 1)
  learner$properties[learner$properties %chin% "hotstart_forward"] = "hotstart_backward"
  hot = HotstartStack$new(list(learner_1))

  expect_equal(hot$start_cost(learner, task$hash), 0)
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_1$state)
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with backward target learner when cost of hotstarting is equal", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 4)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 4)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 3)
  learner$properties[learner$properties %chin% "hotstart_forward"] = "hotstart_backward"
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(0, 0))
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works when hotstart values are lower and higher for backward target", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 3)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 1)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 2)
  learner$properties[learner$properties %chin% "hotstart_forward"] = "hotstart_backward"
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(0, NA_real_))
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_1$state)
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works when backward hotstart and target learner are equal", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)

  learner = lrn("classif.debug", iter = 1)
  learner$properties[learner$properties %chin% "hotstart_forward"] = "hotstart_backward"
  hot = HotstartStack$new(list(learner_1))

  expect_equal(hot$start_cost(learner, task$hash), -1)
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_1$state)
  expect_data_table(hot$stack, ncols = 3)
})


test_that("HotstartStack works with backward target learner when hotstart values are only lower", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)

  learner = lrn("classif.debug", iter = 2)
  learner$properties[learner$properties %chin% "hotstart_forward"] = "hotstart_backward"
  hot = HotstartStack$new(list(learner_1))

  expect_equal(hot$start_cost(learner, task$hash), NA_real_)
  expect_data_table(hot$stack, ncols = 3)
  expect_null(get_private(hot)$.start_learner(learner, task$hash))
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with backward target when learners differ and values vary", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1) # lower
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 2) # equal
  learner_2$train(task)
  learner_3 = lrn("classif.debug", iter = 3) # higher
  learner_3$train(task)
  learner_4 = lrn("classif.rpart") # different learner
  learner_4$train(task)

  learner = lrn("classif.debug", iter = 2)
  learner$properties[learner$properties %chin% "hotstart_forward"] = "hotstart_backward"
  hot = HotstartStack$new(list(learner_1, learner_2, learner_3, learner_4))

  expect_equal(hot$start_cost(learner, task$hash), c(NA_real_, -1, 0, NA_real_))
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with forward/backward target learner and increased hotstart parameter", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 3)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 2)
  learner$properties = c(learner$properties, "hotstart_backward")
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(1, 0))
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_2$state)
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with forward/backward target learner when cost of hotstarting is equal", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 2)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 2)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 3)
  learner$properties = c(learner$properties, "hotstart_backward")
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(1, 1))
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works when hotstart values are lower and higher for forward/backward target", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 3)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 2)
  learner$properties = c(learner$properties, "hotstart_backward")
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(1, 0))
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_2$state)
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works when forward/backward hotstart and target learner are equal", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)

  learner = lrn("classif.debug", iter = 1)
  learner$properties = c(learner$properties, "hotstart_backward")
  hot = HotstartStack$new(list(learner_1))

  expect_equal(hot$start_cost(learner, task$hash), -1)
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_1$state)
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with forward/backward target learner when hotstart values are only higher", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 2)
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 3)
  learner_2$train(task)

  learner = lrn("classif.debug", iter = 1)
  learner$properties = c(learner$properties, "hotstart_backward")
  hot = HotstartStack$new(list(learner_1, learner_2))

  expect_equal(hot$start_cost(learner, task$hash), c(0, 0))
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with forward/backward target learner when hotstart values are only lower", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)

  learner = lrn("classif.debug", iter = 2)
  learner$properties = c(learner$properties, "hotstart_backward")
  hot = HotstartStack$new(list(learner_1))

  expect_equal(hot$start_cost(learner, task$hash), 1)
  expect_data_table(hot$stack, ncols = 3)
  expect_equal(get_private(hot)$.start_learner(learner, task$hash)$state, learner_1$state)
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack works with forward/backward target when learners differ and values vary", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1) # lower
  learner_1$train(task)
  learner_2 = lrn("classif.debug", iter = 2) # equal
  learner_2$train(task)
  learner_3 = lrn("classif.debug", iter = 3) # higher
  learner_3$train(task)
  learner_4 = lrn("classif.rpart") # different learner
  learner_4$train(task)

  learner = lrn("classif.debug", iter = 2)
  learner$properties = c(learner$properties, "hotstart_backward")
  hot = HotstartStack$new(list(learner_1, learner_2, learner_3, learner_4))

  expect_equal(hot$start_cost(learner, task$hash), c(1, -1, 0, NA_real_))
  expect_data_table(hot$stack, ncols = 3)
})

test_that("HotstartStack add method works", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  hot = HotstartStack$new(list(learner_1))

  learner_2 = lrn("classif.debug", iter = 2)
  learner_2$train(task)
  hot$add(list(learner_2))

  expect_data_table(hot$stack, nrows = 2, ncols = 3)
  expect_equal(hot$stack$start_learner[[1]], learner_1)
  expect_equal(hot$stack$start_learner[[2]], learner_2)
})

test_that("copying state of learner works", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)

  learner_2 = lrn("classif.debug", iter = 2)
  hot = HotstartStack$new(learner_1)
  learner_2$hotstart_stack = hot

  start_learner = get_private(learner_2$hotstart_stack)$.start_learner(learner_2, task$hash)
  start_learner$param_set$values$iter = 2

  expect_equal(learner_1$param_set$values$iter, 1)
  expect_equal(start_learner$param_set$values$iter, 2)
})

test_that("HotstartStack works without hotstart parameter", {
  task = tsk("pima")
  start_learner = lrn("classif.rpart")
  start_learner$train(task)

  target_learner = lrn("classif.rpart")
  hot = HotstartStack$new(start_learner)
  target_learner$hotstart_stack = hot

  expect_equal(hot$start_cost(target_learner, task$hash), NA_real_)
  expect_null(get_private(target_learner$hotstart_stack)$.start_learner(target_learner, task$hash))
})


test_that("Learners without a model are not added to the hotstart stack", {
  task = tsk("pima")
  learner = lrn("classif.debug", iter = 1)
  learner$train(task)
  learner$state$model = NULL

  expect_error(HotstartStack$new(learner), "Learners must be trained before adding them to the hotstart stack.")
})

test_that("HotstartStack threshold works", {
  hot = HotstartStack$new(hotstart_threshold = c(iter = 2))

  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)

  hot$add(learner_1)
  expect_data_table(hot$stack, nrows = 0)

  task = tsk("pima")
  learner_2 = lrn("classif.debug", iter = 2)
  learner_2$train(task)

  hot$add(learner_2)
  expect_data_table(hot$stack, nrows = 1)
})

test_that("error when adding marshaled learner", {
  hot = HotstartStack$new()
  learner = lrn("classif.debug")
  learner$train(tsk("iris"))
  learner$marshal()
  expect_error(hot$add(learner), "unmarshaled")
})

#' @title Syntactic Sugar for Object Construction
#'
#' @name mlr_sugar
#' @description
#' Functions to retrieve objects, set hyperparameters and assign to fields in one go.
#' Relies on [mlr3misc::dictionary_sugar_get()] to extract objects from the respective [mlr3misc::Dictionary]:
#'
#' * `tsk()` for a [Task] from [mlr_tasks].
#' * `tsks()` for a list of [Task]s from [mlr_tasks].
#' * `tgen()` for a [TaskGenerator] from [mlr_task_generators].
#' * `tgens()` for a list of [TaskGenerator]s from [mlr_task_generators].
#' * `lrn()` for a [Learner] from [mlr_learners].
#' * `lrns()` for a list of [Learner]s from [mlr_learners].
#' * `rsmp()` for a [Resampling] from [mlr_resamplings].
#' * `rsmps()` for a list of [Resampling]s from [mlr_resamplings].
#' * `msr()` for a [Measure] from [mlr_measures].
#' * `msrs()` for a list of [Measure]s from [mlr_measures].
#'
#' @param .key (`character(1)`)\cr
#'   Key passed to the respective [dictionary][mlr3misc::Dictionary] to retrieve the object.
#' @param .keys (`character()`)\cr
#'   Keys passed to the respective [dictionary][mlr3misc::Dictionary] to retrieve multiple objects.
#' @param ... (named `list()`)\cr
#'   Named arguments passed to the constructor, to be set as parameters in the [paradox::ParamSet],
#'   or to be set as public field.
#'   See [mlr3misc::dictionary_sugar_get()] for more details.
#'
#' @return [R6::R6Class] object of the respective type,
#'   or a list of [R6::R6Class] objects for the plural versions.
#' @examples
#' # penguins task with new id
#' tsk("penguins", id = "penguins2")
#'
#' # classification tree with different hyperparameters
#' # and predict type set to predict probabilities
#' lrn("classif.rpart", cp = 0.1, predict_type = "prob")
#'
#' # multiple learners with predict type 'prob'
#' lrns(c("classif.featureless", "classif.rpart"), predict_type = "prob")
NULL

#' @rdname mlr_sugar
#' @export
tsk = function(.key, ...) {
  dictionary_sugar_get(dict = mlr_tasks, .key, ...)
}

#' @rdname mlr_sugar
#' @export
tsks = function(.keys, ...) {
  dictionary_sugar_mget(dict = mlr_tasks, .keys, ...)
}

#' @rdname mlr_sugar
#' @export
tgen = function(.key, ...) {
  dictionary_sugar_get(dict = mlr_task_generators, .key, ...)
}

#' @rdname mlr_sugar
#' @export
tgens = function(.keys, ...) {
  dictionary_sugar_mget(dict = mlr_task_generators, .keys, ...)
}

#' @rdname mlr_sugar
#' @export
lrn = function(.key, ...) {
  dictionary_sugar_get(dict = mlr_learners, .key, ...)
}

#' @rdname mlr_sugar
#' @export
lrns = function(.keys, ...) {
  dictionary_sugar_mget(dict = mlr_learners, .keys, ...)
}

#' @rdname mlr_sugar
#' @export
rsmp = function(.key, ...) {
  dictionary_sugar_get(dict = mlr_resamplings, .key, ...)
}

#' @rdname mlr_sugar
#' @export
rsmps = function(.keys, ...) {
  dictionary_sugar_mget(dict = mlr_resamplings, .keys, ...)
}

#' @rdname mlr_sugar
#' @export
msr = function(.key, ...) {
  dictionary_sugar_get(dict = mlr_measures, .key, ...)
}

#' @rdname mlr_sugar
#' @export
msrs = function(.keys, ...) {
  dictionary_sugar_mget(dict = mlr_measures, .keys, ...)
}

test_that("holdout has no duplicated ids", {
  r = rsmp("holdout")
  expect_identical(r$duplicated_ids, FALSE)
})

test_that("stratification", {
  data = data.table(
    y = factor(rep(letters[1:2], times = c(90, 10))),
    x1 = runif(100),
    x2 = rep(LETTERS[1:2], times = c(50, 50))
  )
  b = as_data_backend(data)
  task = TaskClassif$new("stratify_data", b, target = "y")
  task$col_roles$stratum = task$target_names

  r = rsmp("holdout", ratio = .5)
  r$instantiate(task)

  i = 1L
  expect_equal(task$data(r$train_set(i))[y == "a", .N], 45)
  expect_equal(task$data(r$train_set(i))[y == "b", .N], 5)
  expect_equal(task$data(r$test_set(i))[y == "a", .N], 45)
  expect_equal(task$data(r$test_set(i))[y == "b", .N], 5)
})

test_that("grouping", {
  r = rsmp("holdout")
  expect_grouping_works(r)
})

test_that("prediction does not drop dimension (#551)", {
  task = tsk("iris")
  learner = lrn("classif.rpart")
  resampling = rsmp("holdout")
  resampling$instantiate(task)

  design = data.table(
    learner = list(learner),
    task = list(task),
    resampling = list(resampling)
  )

  bmr = benchmark(design)
  expect_number(bmr$aggregate(msr("classif.ce"))[["classif.ce"]])
  expect_equal(map(get_private(bmr)$.data$data$fact$prediction, names), list("test"))
})

test_that("Construction", {
  task = tsk("california_housing")
  p = PredictionRegr$new(row_ids = task$row_ids, truth = task$truth(), response = task$truth())
  expect_prediction(p)
  expect_prediction_regr(p)
})

test_that("Internally constructed Prediction", {
  task = tsk("california_housing")
  lrn = lrn("regr.featureless")
  lrn$predict_type = "se"
  p = lrn$train(task)$predict(task)
  expect_prediction(p)
  expect_prediction_regr(p)
})


test_that("c", {
  task = tsk("california_housing")
  lrn = lrn("regr.featureless")
  lrn$predict_type = "se"
  rr = resample(task, lrn, rsmp("cv", folds = 3))

  pred = do.call(c, rr$predictions())
  expect_prediction(pred)
  expect_prediction_regr(pred)

  dt = as.data.table(pred)
  expect_data_table(dt, nrows = task$nrow, ncols = 4L, any.missing = FALSE)

  # duplicates are detected?
  p1 = get_private(rr)$.data$data$fact$prediction[[1L]]$test
  p2 = get_private(rr)$.data$data$fact$prediction[[1L]]$test
  p3 = c(p1, p2, keep_duplicates = FALSE)
  expect_equal(sort(p1$data$row_ids), sort(p2$data$row_ids))
  expect_equal(sort(p1$data$row_ids), sort(p3$data$row_ids))
  expect_numeric(p3$response, len = length(p1$response), any.missing = FALSE)
  expect_numeric(p3$se, len = length(p1$se), any.missing = FALSE)
})

test_that("c drops se (#250)", {
  task = tsk("california_housing")
  lrn = lrn("regr.featureless")
  rr = resample(task, lrn, rsmp("cv", folds = 3))

  pred = do.call(c, rr$predictions())
  expect_null(pred$data$se)
  expect_false("se" %chin% pred$predict_types)
  expect_true(allMissing(pred$se))
  expect_false("se" %chin% names(as.data.table(pred)))
})

test_that("distr", {
  skip_if_not_installed("distr6")

  task = tsk("mtcars")
  distr = distr6::VectorDistribution$new(
    distribution = "Binomial",
    params = replicate(task$nrow, list(prob = runif(1), size = 10), FALSE)
  )

  p = PredictionRegr$new(row_ids = task$row_ids, truth = task$truth(), distr = distr, check = TRUE)
  expect_output(print(p))
  expect_set_equal(p$predict_types, c("response", "se", "distr"))
  expect_numeric(p$response, len = task$nrow, any.missing = FALSE)
  expect_numeric(p$se, len = task$nrow, any.missing = FALSE, lower = 0)
  expect_integer(p$missing, len = 0)
  expect_prediction(p)

  expect_prediction(c(p, p))
  expect_output(print(p))
  expect_set_equal(p$predict_types, c("response", "se", "distr"))
  expect_numeric(p$response, len = task$nrow, any.missing = FALSE)
  expect_numeric(p$se, len = task$nrow, any.missing = FALSE, lower = 0)
  expect_integer(p$missing, len = 0)
})

test_that("as_prediction_regr", {
  task = tsk("mtcars")
  learner = lrn("regr.featureless")
  p = learner$train(task)$predict(task)

  tab = as.data.table(p)
  p2 = as_prediction_regr(tab)

  expect_equal(tab, as.data.table(p2))
})

test_that("filtering", {
  task = tsk("mtcars")
  p = PredictionRegr$new(row_ids = task$row_ids, truth = task$truth(), response = task$truth())

  p2 = p$clone()$filter(1:3)
  expect_set_equal(p$row_ids, 1:32)
  expect_set_equal(p2$row_ids, 1:3)
  expect_prediction(as_prediction_regr(as.data.table(p2)))
})

test_that("predictions with weights", {
  ll = lrn("regr.debug")$train(as_task_regr(cars, target = "dist"), row_ids = 1)

  pred_with_weights = ll$predict(cars_weights_measure)
  expect_equal(pred_with_weights$weights, rep(c(1, 10), each = 25))

  pred_without_weights = ll$predict(cars_weights_learner) # learner weights are ignored during predict
  expect_null(pred_without_weights$weights)

  expect_error(c(pred_with_weights, pred_without_weights), "Some predictions have weights, others do not")

  expect_equal(c(pred_with_weights, pred_with_weights)$weights, rep(c(1, 10, 1, 10), each = 25))

  expect_equal(
    c(pred_with_weights$clone(deep = TRUE)$filter(1:10), pred_with_weights)$weights,
    rep(c(1, 10, 1, 10), each = 25)[c(1:10, 51:100)]
  )
})

test_that("extra data is stored", {
  LearnerExtra = R6Class(
    "LearnerExtra",
    inherit = LearnerRegrDebug,
    private = list(
      .predict = function(task, ...) {
        pred = super$.predict(task, ...)
        pred$extra = list(extra_col = replicate(length(pred$response), "bar"))
        pred
      }
    )
  )

  learner = LearnerExtra$new()
  learner$train(tsk("mtcars"))
  pred = learner$predict(tsk("mtcars"))
  expect_list(pred$extra, len = 1)
  expect_equal(pred$extra[[1]], replicate(length(pred$response), "bar"))
  expect_prediction(pred)

  LearnerExtra = R6Class(
    "LearnerExtra",
    inherit = LearnerRegrDebug,
    private = list(
      .predict = function(task, ...) {
        pred = super$.predict(task, ...)
        pred$extra = list(extra_col = replicate(2, "bar"))
        pred
      }
    )
  )

  learner = LearnerExtra$new()
  learner$train(tsk("mtcars"))
  expect_error(learner$predict(tsk("mtcars")), "Extra data must have the same length as the number of predictions")
})

test_that("raw data is stored", {
  task = tsk("mtcars")
  raw_obj = list(a = 1, b = "hello")
  p = PredictionRegr$new(row_ids = task$row_ids, truth = task$truth(), response = task$truth(), raw = raw_obj)
  expect_equal(p$raw, raw_obj)

  # raw survives filtering
  p2 = p$clone()$filter(1:3)
  expect_equal(p2$raw, raw_obj)

  # raw is NULL when not provided
  p3 = PredictionRegr$new(row_ids = task$row_ids, truth = task$truth(), response = task$truth())
  expect_null(p3$raw)
})

test_that("raw data is combined into list", {
  task = tsk("mtcars")
  p1 = PredictionRegr$new(row_ids = 1:3, truth = task$truth(1:3), response = task$truth(1:3), raw = list(x = 1))
  p2 = PredictionRegr$new(row_ids = 4:6, truth = task$truth(4:6), response = task$truth(4:6), raw = list(x = 2))
  combined = c(p1, p2)
  expect_list(combined$raw, len = 2)
  expect_equal(combined$raw[[1]], list(x = 1))
  expect_equal(combined$raw[[2]], list(x = 2))

  # combining with and without raw
  p3 = PredictionRegr$new(row_ids = 7:9, truth = task$truth(7:9), response = task$truth(7:9))
  combined2 = c(p1, p3)
  expect_list(combined2$raw, len = 1)
  expect_equal(combined2$raw[[1]], list(x = 1))

  # combining without any raw
  p4 = PredictionRegr$new(row_ids = 10:12, truth = task$truth(10:12), response = task$truth(10:12))
  combined3 = c(p3, p4)
  expect_null(combined3$raw)
})

test_that("raw data via learner predict", {
  LearnerRaw = R6Class(
    "LearnerRaw",
    inherit = LearnerRegrDebug,
    private = list(
      .predict = function(task, ...) {
        pred = super$.predict(task, ...)
        pred$raw = list(upstream_output = "raw_value")
        pred
      }
    )
  )

  learner = LearnerRaw$new()
  learner$train(tsk("mtcars"))
  pred = learner$predict(tsk("mtcars"))
  expect_equal(pred$raw, list(upstream_output = "raw_value"))
})

test_that("predict_raw with regr.rpart", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart", predict_raw = TRUE)
  learner$train(task)
  pred = learner$predict(task)
  expect_numeric(pred$raw, len = task$nrow)

  learner = lrn("regr.rpart")
  learner$train(task)
  pred = learner$predict(task)
  expect_null(pred$raw)
})

test_that("obs_loss works", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  learner$train(task)
  prediction = learner$predict(task)
  obs_loss = prediction$obs_loss(msr("regr.mse"))
  expect_data_table(obs_loss, nrows = task$nrow)
  expect_numeric(obs_loss$regr.mse, len = task$nrow, any.missing = FALSE)

  obs_loss = prediction$obs_loss(msrs(c("regr.mse", "regr.rmse", "regr.ktau")))
  expect_data_table(obs_loss, nrows = task$nrow)
  expect_numeric(obs_loss$regr.mse, len = task$nrow, any.missing = FALSE)
  expect_numeric(obs_loss$regr.rmse, len = task$nrow, any.missing = FALSE)
  expect_true(all(is.na(obs_loss$regr.ktau)))
})

test_that("mlr_resamplings", {
  expect_dictionary(mlr_resamplings, min_items = 1L)
  keys = setdiff(mlr_resamplings$keys(), "ordered_holdout")

  task = tsk("iris")
  for (key in keys) {
    r = rsmp(key)
    expect_resampling(r) # construction works
    expect_false(r$is_instantiated)
    if (key == "custom") {
      ret = r$instantiate(task, list(1:3), list(5:9))
    } else if (key == "custom_cv") {
      ret = r$instantiate(task, iris$Species)
    } else {
      ret = r$instantiate(task)
    }
    expect_r6(ret, "Resampling")
    expect_true(r$is_instantiated)
    expect_resampling(r)
  }
})

test_that("mlr_resamplings: sugar", {
  r = rsmp("cv", id = "cv3", folds = 3L)
  expect_equal(r$id, "cv3")
  expect_equal(r$param_set$values$folds, 3L)
})

test_that("as.data.table(..., objects = TRUE)", {
  tab = as.data.table(mlr_resamplings, objects = TRUE)
  expect_data_table(tab)
  expect_list(tab$object, "Resampling", any.missing = FALSE)
})

skip_if_not_installed("mirai")

test_that("parallel resample", {
  with_mirai(
    {
      task = tsk("pima")
      learner = lrn("classif.rpart")
      rr = resample(task, learner, rsmp("cv", folds = 3))
      expect_resample_result(rr)
      expect_data_table(rr$errors, nrows = 0L)
    },
    compute = "mlr3_parallelization"
  )
})

test_that("parallel benchmark", {
  task = tsk("pima")
  learner = lrn("classif.rpart")

  with_mirai(
    {
      bmr = benchmark(benchmark_grid(task, learner, rsmp("cv", folds = 3)))
    },
    compute = "mlr3_parallelization"
  )
  expect_benchmark_result(bmr)
  expect_equal(bmr$aggregate(conditions = TRUE)$warnings, 0L)
  expect_equal(bmr$aggregate(conditions = TRUE)$errors, 0L)
})

test_that("mirai resample is reproducible", {
  with_mirai(
    {
      task = tsk("pima")
      learner = lrn("classif.debug")
      resampling = rsmp("cv", folds = 3)
      rr = resample(task, learner, resampling, store_models = TRUE)
    },
    compute = "mlr3_parallelization",
    seed = 123
  )
  random_number_1 = map_int(rr$learners, function(l) l$model$random_number)

  with_mirai(
    {
      task = tsk("pima")
      learner = lrn("classif.debug")
      resampling = rsmp("cv", folds = 3)
      rr = resample(task, learner, resampling, store_models = TRUE)
    },
    compute = "mlr3_parallelization",
    seed = 123
  )
  random_number_2 = map_int(rr$learners, function(l) l$model$random_number)

  expect_equal(random_number_1, random_number_2)
})

test_that("data table threads are not changed in main session", {
  skip_on_os("mac") # number of threads cannot be changed on mac
  skip_on_cran()

  old_dt_threads = getDTthreads()
  on.exit(
    {
      setDTthreads(old_dt_threads)
    },
    add = TRUE
  )
  setDTthreads(2L)

  task = tsk("sonar")
  learner = lrn("classif.debug", predict_type = "prob")
  resampling = rsmp("cv", folds = 3L)
  measure = msr("classif.auc")

  # nolint next
  with_seed(123L, with_mirai(resample(task, learner, resampling), compute = "mlr3_parallelization"))
  expect_equal(getDTthreads(), 2L)
})

test_that("parallel resample and encapsulation works", {
  skip_on_cran()

  with_mirai(
    {
      # start encapsulation daemons on all daemons
      mirai::everywhere(
        {
          mirai::daemons(1, .compute = "mlr3_encapsulation")
        },
        .compute = "mlr3_parallelization"
      )

      task = tsk("pima")
      learner = lrn("classif.debug")
      learner$encapsulate("mirai", lrn("classif.featureless"))
      resampling = rsmp("cv", folds = 3)
      rr = resample(task, learner, resampling, store_models = TRUE)

      expect_resample_result(rr)
      expect_data_table(rr$errors, nrows = 0L)
      # check that all resample iterations are evaluated on daemon
      mirai::status(.compute = "mlr3_parallelization")$mirai["completed"]
      # 3 resample iterations + everywhere call
      expect_equal(mirai::status(.compute = "mlr3_parallelization")$mirai["completed"], c("completed" = 4L))

      # check that all encapsulation calls are evaluated on daemon
      status_encapsulation = mirai::collect_mirai(mirai::mirai(
        {
          mirai::status(.compute = "mlr3_encapsulation")
        },
        .compute = "mlr3_parallelization"
      ))
      expect_equal(status_encapsulation$mirai["completed"], c("completed" = 6L)) # 3 x train + 3 x predict
    },
    compute = "mlr3_parallelization"
  )
})

test_that("mirai compute profile can be changed", {
  old_opts = getOption("mlr3.mirai_parallelization")
  on.exit(options(mlr3.mirai_parallelization = old_opts), add = TRUE)

  with_mirai(
    {
      options(mlr3.mirai_parallelization = "mlr3_parallelization2")
      task = tsk("pima")
      learner = lrn("classif.debug")
      resampling = rsmp("cv", folds = 3)
      rr = resample(task, learner, resampling, store_models = TRUE)
    },
    compute = "mlr3_parallelization2"
  )

  expect_resample_result(rr)
})

test_that("ContextResample works", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)
  iteration = 1

  ctx = ContextResample$new(task, learner, resampling, iteration)

  expect_task(ctx$task)
  expect_learner(ctx$learner)
  expect_resampling(ctx$resampling)
  expect_equal(ctx$iteration, iteration)

  expect_error(
    {
      ctx$task = tsk("spam")
    },
    "read-only"
  )
  expect_error(
    {
      ctx$resampling = rsmp("cv", folds = 5)
    },
    "read-only"
  )
  expect_error(
    {
      ctx$iteration = 2
    },
    "read-only"
  )
})

test_that("selected_features", {
  task = tsk("sonar")
  lrn = lrn("classif.rpart")
  measures = list(msr("classif.ce"), msr("selected_features"))

  rr = resample(task, lrn, rsmp("holdout"), store_models = TRUE)
  perf = rr$aggregate(measures)
  expect_count(perf[["selected_features"]])
})

skip_if(Sys.getenv("GITHUB_ACTIONS") != "true", "Not on GitHub Actions")
skip_if(!identical(Sys.getenv("BRANCH_NAME"), "release"), "Not in release branch")
github_path = Sys.getenv("GITHUB_WORKSPACE")

test_that("task classif backward compatibility", {
  task = readRDS(file.path(github_path, "tests", "testthat", "_object_snapshots", "task_classif.rds"))
  expect_task(task)
})

test_that("task regr backward compatibility", {
  task = readRDS(file.path(github_path, "tests", "testthat", "_object_snapshots", "task_regr.rds"))
  expect_task(task)
})

test_that("learner classif backward compatibility", {
  learner_classif = readRDS(file.path(github_path, "tests", "testthat", "_object_snapshots", "learner_classif.rds"))
  expect_learner(learner_classif)

  task = tsk("pima")
  suppressWarnings({
    pred = learner_classif$predict(task)
  })
  expect_prediction_classif(pred, task)

  learner_classif$train(task)
  expect_learner(learner_classif)
})

test_that("learner regr backward compatibility", {
  learner_regr = readRDS(file.path(github_path, "tests", "testthat", "_object_snapshots", "learner_regr.rds"))
  expect_learner(learner_regr)

  task = tsk("mtcars")
  suppressWarnings({
    pred = learner_regr$predict(task)
  })
  expect_prediction_regr(pred)

  learner_regr$train(task)
  expect_learner(learner_regr)
})

test_that("resampling backward compatibility", {
  resampling = readRDS(file.path(github_path, "tests", "testthat", "_object_snapshots", "resampling.rds"))
  expect_resampling(resampling)
})

test_that("resample result backward compatibility", {
  rr = readRDS(file.path(github_path, "tests", "testthat", "_object_snapshots", "rr.rds"))
  expect_resample_result(rr)

  score = rr$score(msr("classif.ce"))
  expect_data_table(score, nrows = 3L)
  expect_numeric(score$classif.ce, len = 3L)
})

test_that("benchmark result backward compatibility", {
  bmr = readRDS(file.path(github_path, "tests", "testthat", "_object_snapshots", "bmr.rds"))
  expect_benchmark_result(bmr)

  score = bmr$score(msr("classif.ce"))
  expect_data_table(score, nrows = 3L)
  expect_numeric(score$classif.ce, len = 3L)

  aggr = bmr$aggregate()
  expect_data_table(aggr, nrows = 1L)
  expect_numeric(aggr$classif.ce, len = 1L)
})

test_that("measure backward compatibility", {
  measure = readRDS(file.path(github_path, "tests", "testthat", "_object_snapshots", "measure.rds"))
  expect_measure(measure)

  task = tsk("pima")
  learner = lrn("classif.rpart")
  learner$train(task)
  pred = learner$predict(task)
  expect_numeric(measure$score(pred), len = 1L)
})

# stages -----------------------------------------------------------------------

test_that("on_resample_begin works", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)

  callback = callback_resample(
    "test",

    on_resample_begin = function(callback, context) {
      # expect_* does not work
      assert_task(context$task)
      assert_learner(context$learner)
      assert_resampling(context$resampling)
      assert_number(context$iteration)
      assert_null(context$pdatas)
    }
  )

  expect_resample_result(resample(task, learner, resampling, callbacks = callback))
})

test_that("on_resample_before_train works", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)

  callback = callback_resample(
    "test",

    on_resample_before_train = function(callback, context) {
      assert_task(context$task)
      assert_learner(context$learner)
      assert_resampling(context$resampling)
      assert_number(context$iteration)
      assert_null(context$pdatas)
    }
  )

  expect_resample_result(resample(task, learner, resampling, callbacks = callback))
})

test_that("on_resample_before_predict works", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)

  callback = callback_resample(
    "test",

    on_resample_before_predict = function(callback, context) {
      assert_task(context$task)
      assert_learner(context$learner)
      assert_resampling(context$resampling)
      assert_null(context$pdatas)
    }
  )

  expect_resample_result(resample(task, learner, resampling, callbacks = callback))
})

test_that("on_resample_end works", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)

  callback = callback_resample(
    "test",

    on_resample_end = function(callback, context) {
      assert_task(context$task)
      assert_learner(context$learner)
      assert_resampling(context$resampling)
      assert_number(context$iteration)
      assert_class(context$pdatas$test, "PredictionData")
    }
  )

  expect_resample_result(resample(task, learner, resampling, callbacks = callback))
})

# learner state ----------------------------------------------------------------

test_that("writing to learner$state works", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)

  callback = callback_resample("test", on_resample_end = function(callback, context) {
    context$learner$state$test = 1
  })

  # resample result
  rr = resample(task, learner, resampling, callbacks = callback)
  walk(rr$learners, function(learner) {
    expect_equal(learner$state$test, 1)
  })
  expect_data_table(rr$data_extra, nrows = 3)
  walk(rr$data_extra$data_extra, expect_null)
  tab = as.data.table(rr)
  expect_names(names(tab), disjunct.from = "data_extra")

  # benchmark result
  design = benchmark_grid(task, learner, resampling)
  bmr = benchmark(design, callbacks = callback)
  walk(bmr$score()$learner, function(learner) {
    expect_equal(learner$state$test, 1)
  })
  tab = as.data.table(bmr)
  expect_names(names(tab), disjunct.from = "data_extra")
})

# data_extra -------------------------------------------------------------------

test_that("writing to data_extra works", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)

  callback = callback_resample("test", on_resample_end = function(callback, context) {
    context$data_extra$test = 1
  })

  # resample result
  rr = resample(task, learner, resampling, callbacks = callback)
  walk(rr$data_extra$data_extra, function(x) {
    expect_equal(x$test, 1)
  })

  # resample result data.table
  tab = as.data.table(rr)
  expect_data_table(tab)
  expect_names(names(tab), must.include = "data_extra")
  walk(tab$data_extra, function(x) {
    expect_equal(x$test, 1)
  })

  # benchmark data.table
  design = benchmark_grid(task, learner, resampling)
  bmr = benchmark(design, callbacks = callback)
  tab = as.data.table(bmr)
  expect_names(names(tab), must.include = "data_extra")
  expect_list(tab$data_extra)
  walk(tab$data_extra, function(x) {
    expect_equal(x$test, 1)
  })
})

test_that("data_extra is a list column", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("holdout")

  callback = callback_resample("test", on_resample_end = function(callback, context) {
    context$data_extra$test = 1
  })

  rr = resample(task, learner, resampling, callbacks = callback)
  expect_list(as.data.table(rr)$data_extra, len = 1)
  expect_list(as.data.table(rr)$data_extra[[1]], len = 1)

  resampling = rsmp("cv", folds = 3)
  rr = resample(task, learner, resampling, callbacks = callback)
  expect_list(as.data.table(rr)$data_extra, len = 3)
  expect_list(as.data.table(rr)$data_extra[[1]], len = 1)
})

test_that("data_extra is null", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)

  callback = callback_resample("test", on_resample_end = function(callback, context) {
    context$learner$state$test = 1
  })

  # resample result
  rr = resample(task, learner, resampling, callbacks = callback)
  expect_data_table(rr$data_extra, nrows = 3)
  walk(rr$data_extra$data_extra, expect_null)

  # resample result data.table
  tab = as.data.table(rr)
  expect_data_table(tab)
  expect_names(names(tab), disjunct.from = "data_extra")
  walk(tab$data_extra, expect_null)

  # benchmark data.table
  design = benchmark_grid(task, learner, resampling)
  bmr = benchmark(design, callbacks = callback)
  tab = as.data.table(bmr)
  expect_data_table(tab)
  expect_names(names(tab), disjunct.from = "data_extra")
  walk(tab$data_extra, expect_null)
})

test_that("learner cloning in workhorse is passed to context", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("holdout")

  callback = callback_resample(
    "test",
    on_resample_begin = function(callback, context) {
      callback$state$address_1 = data.table::address(context$learner)
    },

    on_resample_before_train = function(callback, context) {
      callback$state$address_2 = data.table::address(context$learner)
    },

    on_resample_end = function(callback, context) {
      context$data_extra = list(
        address_1 = callback$state$address_1,
        address_2 = callback$state$address_2
      )
    }
  )

  rr = resample(task, learner, resampling, callbacks = callback)

  expect_true(rr$data_extra$data_extra[[1]]$address_1 != rr$data_extra$data_extra[[1]]$address_2)
})

test_that("returning data_extra sometimes works ", {
  learners = lrns(c("classif.rpart", "classif.featureless"))
  task = tsk("pima")
  resampling = rsmp("cv", folds = 3)

  callback = callback_resample("test", on_resample_end = function(callback, context) {
    if (context$learner$id == "classif.featureless") {
      context$data_extra$test = 1
    }
  })

  design = benchmark_grid(task, learners, resampling)
  bmr = benchmark(design, callbacks = callback)

  tab = as.data.table(bmr)
  expect_data_table(tab, nrows = 6)
  expect_names(names(tab), must.include = "data_extra")
  expect_list(tab$data_extra)
  expect_equal(sum(map_lgl(tab$data_extra, is.null)), 3)
  expect_data_table(get_private(bmr)$.data$data$data_extras, nrows = 3)
})

test_that("internal valid score", {
  task = tsk("iris")
  learner = lrn("classif.debug", validate = 0.2)$train(task)
  pred = learner$predict(task)
  rr = resample(task, learner, rsmp("holdout"))
  expect_equal(
    rr$score(msr("internal_valid_score", select = "acc"))$acc,
    rr$learners[[1]]$internal_valid_scores$acc
  )
  expect_equal(
    rr$score(msr("internal_valid_score", select = "wrong_name"))$wrong_name,
    NA_real_
  )
  rr = resample(task, lrn("classif.rpart"), rsmp("holdout"))
  expect_equal(
    rr$score(msr("internal_valid_score", select = "acc"))$acc,
    NA_real_
  )
  expect_measure(msr("internal_valid_score"))

  # learner that does not have it
  m = msr("internal_valid_score", select = "a")
  task = tsk("mtcars")
  learner = lrn("regr.debug")
  learner$train(task)
  pred = learner$predict(task)
  rr = resample(task, learner, rsmp("holdout"))
  expect_equal(rr$score(m)$a, NA_real_)

  task = tsk("iris")
  # the first validation score is taken by default
  rr = resample(task, lrn("classif.debug", predict_type = "prob", validate = 0.2), rsmp("holdout"))

  m = msr("internal_valid_score")
  expect_equal(
    rr$score(m)$internal_valid_score,
    rr$learners[[1]]$internal_valid_scores[[1L]]
  )

  measure = msr("internal_valid_score")
  expect_equal(measure$id, "internal_valid_score")
  expect_equal(measure$minimize, NA)
  measure = msr("internal_valid_score", select = "acc", minimize = TRUE)
  expect_equal(measure$id, "acc")
  expect_true(measure$minimize)
})

test_that("fallback = default_fallback() works", {
  learner = lrn("classif.rpart")
  fallback = default_fallback(learner)

  expect_class(fallback, "LearnerClassifFeatureless")
  expect_equal(fallback$predict_type, "response")

  learner = lrn("classif.rpart", predict_type = "prob")
  fallback = default_fallback(learner)

  expect_class(fallback, "LearnerClassifFeatureless")
  expect_equal(fallback$predict_type, "prob")

  learner = lrn("regr.rpart")
  fallback = default_fallback(learner)

  expect_class(fallback, "LearnerRegrFeatureless")
  expect_equal(fallback$predict_type, "response")

  learner = lrn("regr.debug", predict_type = "se")
  fallback = default_fallback(learner)

  expect_class(fallback, "LearnerRegrFeatureless")
  expect_equal(fallback$predict_type, "se")

  learner = lrn("regr.debug", predict_type = "quantiles", quantiles = c(0.1, 0.9), quantile_response = 0.1)
  fallback = default_fallback(learner)

  expect_class(fallback, "LearnerRegrFeatureless")
  expect_equal(fallback$predict_type, "quantiles")
  expect_equal(fallback$quantiles, c(0.1, 0.9))
  expect_equal(fallback$quantile_response, 0.1)
})

test_that("set_threads", {
  l1 = lrn("classif.featureless")
  expect_learner(set_threads(l1))

  l2 = lrn("classif.debug")
  expect_null(l2$param_set$values$threads)
  expect_learner(set_threads(l2, 1))
  expect_equal(l2$param_set$values$threads, 1)

  x = list(l1, l2)
  expect_list(set_threads(x, 2))
  expect_equal(l2$param_set$values$threads, 2)
})

test_that("autotest", {
  learner = lrn("classif.featureless")
  expect_learner(learner)
  result = run_autotest(learner, exclude = "sanity")
  expect_true(result, info = result$error)
})

test_that("Simple training/predict", {
  task = tsk("iris")
  learner = lrn("classif.featureless")
  expect_learner(learner, task)

  learner$train(task, row_ids = c(1:50, 51:70, 101:120))
  expect_class(learner$model, "classif.featureless_model")
  expect_numeric(learner$model$tab, len = 3L, any.missing = FALSE)
  prediction = learner$predict(task)
  expect_factor(prediction$response, any.missing = FALSE, levels = levels(iris$Species))
  perf = prediction$score(msr("classif.ce"))
  expect_number(perf, lower = 0.6, upper = 0.7)
})

test_that("Predict with prob", {
  task = tsk("penguins")
  learner = mlr_learners$get("classif.featureless")
  learner$predict_type = "prob"
  expect_learner(learner, task)

  p = learner$train(task)$predict(task)
  expect_matrix(p$prob, nrows = task$nrow, ncols = 3L)
  expect_names(colnames(p$prob), permutation.of = levels(task$truth()))

  p = lrn("classif.featureless", predict_type = "prob", method = "sample")$train(task)$predict(task)
  expect_number(unique(as.numeric(p$prob)), lower = 0.33, upper = 0.34)

  p = lrn("classif.featureless", predict_type = "prob", method = "weighted.sample")$train(task)$predict(task)
  expect_set_equal(unique(as.numeric(p$prob)), c(0, 1))
})

test_that("classif.featureless works on featureless task", {
  task = tsk("wine")$select(character())
  learner = lrn("classif.featureless")
  rr = resample(task, learner, rsmp("holdout"))
  expect_resample_result(rr)
  expect_number(rr$aggregate())
})

test_that("weights are respected", {
  # Create weighted task
  data = iris
  # Give 'setosa' a much higher weight
  data$weights = fifelse(data$Species == "setosa", 100, 1)
  task = TaskClassif$new(
    id = "iris_weights",
    backend = data,
    target = "Species"
  )
  task$set_col_roles("weights", roles = "weights_learner")

  # --- Response Prediction ---
  learner_resp = lrn("classif.featureless", predict_type = "response", method = "mode")
  learner_resp$train(task, row_ids = c(1, 2, 51:150))
  p_resp_mode = learner_resp$predict(task)
  # With 'mode', the response should always be the highest weighted class
  expect_true(all(p_resp_mode$response == "setosa"))

  learner_resp$param_set$values$method = "weighted.sample"
  learner_resp$train(task) # Re-train needed for model parameters, although method is predict param
  # With 'weighted.sample', the response should heavily favor the highest weighted class
  responses_ws = factor(levels = c("setosa", "versicolor", "virginica"))
  for (i in 1:100) {
    responses_ws[i] = learner_resp$predict(task)$response[1] # predict one sample at a time
  }
  response_table_ws = table(responses_ws)
  expect_true(response_table_ws["setosa"] > 90) # Expect > 90% setosa
  expect_true(all(names(response_table_ws) %in% task$class_names))

  # --- Probability Prediction ---
  learner_prob = lrn("classif.featureless", predict_type = "prob", method = "mode")
  learner_prob$train(task)
  p_prob_mode = learner_prob$predict(task)
  # Probabilities should reflect the weighted frequencies
  expected_probs_mode = c(setosa = 100 / (100 + 1 + 1), versicolor = 1 / (100 + 1 + 1), virginica = 1 / (100 + 1 + 1))
  # Rounding necessary due to potential floating point inaccuracies
  expect_equal(round(p_prob_mode$prob[1, ], 4), round(expected_probs_mode, 4))
})

test_that("mlr_task_generators", {
  expect_dictionary(mlr_task_generators, min_items = 1L, contains = "TaskGenerator")
  keys = mlr_task_generators$keys()
  n = 30L

  for (key in keys) {
    gen = mlr_task_generators$get(key)
    expect_task_generator(gen)

    task = gen$generate(n)
    expect_task(task)
    expect_equal(gen$task_type, task$task_type)
    if (!inherits(gen, "TaskGeneratorSimplex")) {
      expect_equal(task$nrow, n)
    }
  }
})

test_that("as.data.table(mlr_task_generators)", {
  tab = as.data.table(mlr_task_generators)
  expect_data_table(tab, min.cols = 2L)
  expect_character(tab$key, unique = TRUE, any.missing = FALSE)
})

test_that("sugar functions", {
  expect_task_generator(tgen("xor"))
  expect_list(tgens(c("xor", "smiley")), "TaskGenerator")
})

test_that("as.data.table(..., objects = TRUE)", {
  tab = as.data.table(mlr_task_generators, objects = TRUE)
  expect_data_table(tab)
  expect_list(tab$object, "TaskGenerator", any.missing = FALSE)
})

test_that("autotest", {
  learner = lrn("regr.featureless")
  expect_learner(learner)
  result = run_autotest(learner, exclude = "sanity")
  expect_true(result, info = result$error)
})

test_that("regr.featureless works on featureless task", {
  task = tsk("mtcars")$select(character())
  learner = lrn("regr.featureless")
  rr = resample(task, learner, rsmp("holdout"))
  expect_resample_result(rr)
  expect_number(rr$aggregate())
})

test_that("regr.featureless quantile prediction works", {
  task = tsk("mtcars")

  learner = lrn("regr.featureless", predict_type = "quantiles", quantiles = c(0.1, 0.5, 0.9), quantile_response = 0.5)

  learner$train(task)
  expect_numeric(learner$model$quantiles, len = 3L)

  pred = learner$predict(task)
  expect_prediction(pred)
  expect_subset("quantiles", pred$predict_types)
  expect_matrix(pred$quantiles, ncols = 3L, nrows = task$nrow, any.missing = FALSE)
  expect_names(colnames(pred$quantiles), identical.to = c("q0.1", "q0.5", "q0.9"))
  expect_equal(pred$response, pred$quantiles[, 2L])

  learner = lrn("regr.featureless", predict_type = "quantiles")
  expect_error(learner$train(task), "Quantiles must be set")

  learner$quantiles = c(0.25, 0.5, 0.75)
  expect_error(learner$train(task), "Quantile response must be set")
})

test_that("weights are respected", {
  # Task with weights_learner role defined in helper_misc.R
  task = cars_weights_learner
  learner = lrn("regr.featureless")

  # Check learner property
  expect_true("weights" %in% learner$properties)

  # --- Test robust = FALSE ---
  learner$param_set$values$robust = FALSE
  learner$predict_type = "response" # Reset to default for training
  learner$train(task)

  # Calculate expected weighted mean and sd manually
  dt = task$data()
  w = task$weights_learner$weight
  expected_mean = weighted.mean(dt$dist, w)

  # Check model components
  expect_equal(learner$model$location, expected_mean)

  # Test quantiles with robust = FALSE
  learner$predict_type = "quantiles"
  learner$quantiles = c(0.2, 0.8)
  learner$quantile_response = 0.5 # Needs to be set, even if not used here
  learner$train(task)
  expected_quantiles_nonrobust = quantile_weighted(dt$dist, probs = c(0.2, 0.5, 0.8), weights = w)
  expect_equal(learner$model$quantiles, expected_quantiles_nonrobust)

  # --- Test robust = TRUE ---
  learner$param_set$values$robust = TRUE
  learner$predict_type = "response" # Reset for training
  learner$train(task)

  # Calculate expected weighted median and mad manually
  expected_median = quantile_weighted(dt$dist, probs = 0.5, weights = w, continuous = FALSE)
  expected_mad = quantile_weighted(abs(dt$dist - expected_median), probs = 0.5, weights = w, continuous = FALSE) *
    1.4826

  # Check model components
  expect_equal(learner$model$location, expected_median)
  expect_equal(learner$model$dispersion, expected_mad)

  # Test quantiles with robust = TRUE (should be the same as non-robust)
  learner$predict_type = "quantiles"
  learner$quantiles = c(0.3, 0.7)
  learner$quantile_response = 0.5 # Needs to be set
  learner$train(task)
  expected_quantiles_robust = quantile_weighted(dt$dist, probs = c(0.3, 0.5, 0.7), weights = w)
  expect_equal(learner$model$quantiles, expected_quantiles_robust)
})


test_that("weighted quantile helper function", {
  wq = quantile_weighted(
    c(0, 1),
    weights = c(1, 2),
    probs = c(-.Machine$double.eps, 0, 1 / 6, 1 / 3, 0.5, 1, 1 + .Machine$double.eps)
  )

  expect_equal(unname(wq), c(0, 0, 1 / 3, 2 / 3, 3 / 4, 1, 1))

  wqdisc = quantile_weighted(
    c(0, 1),
    weights = c(1, 2),
    probs = c(-.Machine$double.eps, 0, 1 / 6, 1 / 3, 0.5, 1, 1 + .Machine$double.eps),
    continuous = FALSE
  )
  expect_equal(unname(wqdisc), c(0, 0, 0, 2 / 3, 1, 1, 1))

  expect_equal(unname(quantile_weighted(c(3, 4, 1, 2), weights = c(1, 1, 1, 1), probs = 0.5, continuous = FALSE)), 2.5)
  expect_equal(unname(quantile_weighted(c(3, 4, 1, 2), weights = c(1, 1, 1, 2), probs = 0.5, continuous = FALSE)), 2)
  expect_equal(unname(quantile_weighted(c(3, 4, 1, 2), weights = c(2, 1, 1, 2), probs = 0.5, continuous = FALSE)), 2.5)
  expect_equal(
    unname(quantile_weighted(c(3, 4, 1, 2), weights = c(1, 2, 1, 2), probs = 0.5, continuous = FALSE)),
    2 + 1 / 3
  )
})

test_that("as_learner conversion", {
  learner = lrn("classif.rpart")
  converted = as_learner(learner)
  cloned = as_learner(learner, clone = TRUE)

  expect_class(converted, "Learner")
  expect_same_address(learner, converted)
  expect_different_address(learner, cloned)

  expect_list(as_learners(learner), types = "Learner")
  expect_list(as_learners(list(learner)), types = "Learner")
})

test_that("discard_state", {
  learner = lrn("classif.rpart")$train(tsk("iris"))
  learner2 = as_learner(learner, clone = TRUE, discard_state = TRUE)
  expect_null(learner2$state)
  expect_false(is.null(learner$state))

  learner3 = lrn("classif.rpart")
  as_learner(learner3, clone = FALSE, discard_state = TRUE)
  expect_null(learner3$state)
})

test_that("error when arguments are misspelled", {
  expect_error(as_learner(lrn("classif.rpart"), clone2 = TRUE), "Received the following")
})

test_that("model extractor works", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)

  selected_features = function(learner) list(selected_features = learner$selected_features())
  callback = clbk("mlr3.model_extractor", fun = selected_features)

  rr = resample(task, learner, resampling = resampling, callbacks = callback)

  expect_list(rr$data_extra$data_extra)
  walk(rr$data_extra$data_extra, function(data) {
    expect_names(names(data), must.include = "selected_features")
    expect_subset(data[["selected_features"]], task$feature_names)
  })
})

test_that("holdout task works", {
  task = tsk("pima")
  task_holdout = task$clone()
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)
  splits = partition(task, 0.7)

  task$filter(splits$train)
  task_holdout$filter(splits$test)

  callback = clbk("mlr3.holdout_task", task = task_holdout)

  rr = resample(task, learner, resampling = resampling, callbacks = callback)

  expect_list(rr$data_extra$data_extra)
  walk(rr$data_extra$data_extra, function(data) {
    expect_names(names(data), must.include = "prediction_holdout")
    expect_prediction(data[["prediction_holdout"]])
  })
})

test_that("train fails gracefully", {
  task = tsk("iris")
  learner = lrn("classif.debug", error_train = 1)
  expect_error(learner$train(task), "classif.debug")
})

test_that("predict fails gracefully", {
  task = tsk("iris")
  learner = lrn("classif.debug", error_predict = 1)
  learner$train(task)
  expect_error(learner$predict(task), "classif.debug")
})

test_that("fail during train", {
  task = tsk("iris")
  learner = lrn("classif.debug", error_train = 1)
  learner$encapsulate("evaluate", lrn("classif.featureless"))
  expect_class(learner$fallback, "LearnerClassifFeatureless")
  learner$train(task)

  expect_s3_class(learner$state$fallback_state$model, "classif.featureless_model")
  expect_null(learner$state$model)
  expect_number(learner$state$train_time, lower = 0)

  expect_prediction(learner$predict(task))
})

test_that("fail during predict", {
  task = tsk("iris")
  learner = lrn("classif.debug", error_predict = 1)
  learner$encapsulate("evaluate", lrn("classif.featureless"))
  expect_class(learner$fallback, "LearnerClassifFeatureless")
  learner$train(task)

  expect_s3_class(learner$state$fallback_state$model, "classif.featureless_model")
  expect_s3_class(learner$state$model, "classif.debug_model")
  expect_number(learner$state$train_time, lower = 0)

  expect_prediction(learner$predict(task))
})

test_that("fail during resample", {
  task = tsk("iris")
  learner = lrn("classif.debug", error_predict = 1)
  learner$encapsulate("evaluate", lrn("classif.featureless"))
  expect_class(learner$fallback, "LearnerClassifFeatureless")

  rr = resample(tsk("iris"), learner, rsmp("cv", folds = 3))
  expect_data_table(rr$errors, nrows = 3)
  expect_number(rr$aggregate(msr("classif.ce")))
})

test_that("incomplete predictions", {
  task = tsk("iris")
  learner = lrn("classif.debug", predict_type = "prob", predict_missing = 0.5)
  learner$encapsulate("evaluate", lrn("classif.featureless", predict_type = "prob"))
  expect_class(learner$fallback, "LearnerClassifFeatureless")

  learner$train(task)
  p = learner$predict(task)
  expect_prediction(p)
  expect_factor(p$response, any.missing = FALSE)
  expect_matrix(p$prob, any.missing = FALSE)

  rr = resample(tsk("iris"), learner, rsmp("cv", folds = 3))
  expect_prediction(rr$prediction())
  expect_factor(rr$prediction()$response, any.missing = FALSE)
  expect_matrix(rr$prediction()$prob, any.missing = FALSE)
})

test_that("fallback properties are checked", {
  learner = lrn("classif.featureless")
  fallback = lrn("classif.debug")

  expect_warning(learner$encapsulate("evaluate", fallback), "The fallback learner")
})

test_that("fail during train, weights are used", {
  task = iris_weights_learner
  learner = lrn("classif.debug", error_train = 1, predict_type = "prob")
  learner$encapsulate("evaluate", lrn("classif.featureless", predict_type = "prob"))
  expect_class(learner$fallback, "LearnerClassifFeatureless")
  learner$train(task)

  expect_equal(
    unname(learner$predict(task)$prob),
    matrix(c(1 / 111, 10 / 111, 100 / 111), byrow = TRUE, ncol = 3, nrow = 150)
  )

  expect_equal(learner$predict(iris_weights_measure)$weights, rep(c(1, 10, 100), each = 50))
})

test_that("mlr_measures_regr.pinball", {
  task = tsk("california_housing")
  lrn = lrn("regr.featureless")

  expect_error(msr("regr.pinball", alpha = 2), "alpha: Element 1 is not <= 1")

  m = msr("regr.pinball")
  expect_equal(m$properties, "weights")
  preds_na = lrn$train(task)$predict(task)
  expect_warning(preds_na$score(m), "missing predict type 'quantiles'")
  score_na = suppressWarnings(unname(preds_na$score(m)))
  expect_equal(score_na, NaN)

  lrn$predict_type = "quantiles"
  lrn$quantiles = c(0.25, 0.5, 0.75)
  lrn$quantile_response = 0.5

  preds = lrn$train(task)$predict(task)
  expect_number(preds$score(m))

  # alpha must be in predicted quantiles
  expect_error(preds$score(msr("regr.pinball", alpha = 0.1)), "Assertion on 'alpha' failed: Must be element of set")
})

test_that("binary task", {
  task = tsk("sonar")
  costs = matrix(c(0, 5, 10, 0), nrow = 2)
  rownames(costs) = colnames(costs) = task$class_names

  m1 = mlr_measures$get("classif.costs")
  m1$costs = costs
  m1$id = "m1"
  m2 = msr("classif.costs", costs = -1 * costs)
  m2$id = "m2"
  costs[] = 0
  m3 = msr("classif.costs", costs = costs)
  m3$id = "m3"
  measures = list(m1, m2, m3)

  lrn = lrn("classif.featureless")
  perf = lrn$train(task)$predict(task)$score(measures)
  expect_equal(perf[[1]], -perf[[2]])
  expect_equal(perf[[3]], 0)
})

test_that("multiclass", {
  task = tsk("iris")
  costs = 1 - diag(3)
  rownames(costs) = colnames(costs) = task$class_names

  m = msr("classif.costs", costs = costs, normalize = FALSE)
  measures = list(m, mlr_measures$get("classif.ce"))

  lrn = lrn("classif.featureless")
  p = lrn$train(task)$predict(task)
  perf = p$score(measures, task = task)
  expect_equal(perf[["classif.costs"]], perf[["classif.ce"]] * task$nrow)
})

test_that("multiclass / level reordering", {
  task = tsk("iris")
  costs = matrix(runif(9), nrow = 3)
  diag(costs) = 0
  rownames(costs) = colnames(costs) = task$class_names

  m1 = msr("classif.costs", id = "c1", costs = costs, normalize = FALSE)
  costs = costs[c(2, 1, 3), c(3, 2, 1)]
  m2 = msr("classif.costs", id = "c2", costs = costs, normalize = FALSE)
  measures = list(m1, m2)

  lrn = lrn("classif.featureless")
  perf = lrn$train(task)$predict(task)$score(measures, task = task)
  expect_equal(unname(perf[1]), unname(perf[2]))
})

test_that("multiclass with weights_measure", {
  task = iris_weights_measure
  costs = matrix(c(0, 1, 2, 3, 0, 4, 5, 6, 0), nrow = 3, byrow = TRUE)
  dimnames(costs) = list(response = task$class_names, truth = task$class_names)

  m = msr("classif.costs", costs = costs, normalize = FALSE)

  # Use a deterministic learner for simplicity
  learner = lrn("classif.featureless")
  p = learner$train(task)$predict(task)

  # Score using the measure
  perf = p$score(m)

  # Calculate expected weighted cost manually
  truth = p$truth
  response = p$response
  weights = p$weights
  weighted_confusion = tapply(weights, list(response = response, truth = truth), sum, default = 0)

  # Ensure matrix dimensions and order match
  lvls = task$class_names
  weighted_confusion_full = matrix(
    0,
    nrow = length(lvls),
    ncol = length(lvls),
    dimnames = list(response = lvls, truth = lvls)
  )
  weighted_confusion_full[rownames(weighted_confusion), colnames(weighted_confusion)] = weighted_confusion

  expected_cost = sum(weighted_confusion_full * costs)

  expect_equal(perf, c(classif.costs = expected_cost))
})

test_that("custom has duplicated ids", {
  r = rsmp("custom")
  expect_identical(r$duplicated_ids, TRUE)
})

test_that("custom_cv accepts external factor", {
  task = tsk("penguins")
  task$filter(1:10)

  ccv = rsmp("custom_cv")
  split_f = factor(c(rep(letters[1:3], each = 3), NA))
  ccv$instantiate(task, f = split_f)

  expect_class(ccv$instance, "list")
  expect_length(ccv$instance, 3)
  expect_length(ccv$train_set(3), 6)

  expect_identical(ccv$duplicated_ids, FALSE)
})

test_that("custom_cv accepts task feature", {
  task = tsk("german_credit")
  ccv = rsmp("custom_cv")
  expect_identical(ccv$duplicated_ids, FALSE)

  ccv$instantiate(task, f = task$data(cols = "job")[[1L]])
  expect_class(ccv$instance, "list")
  expect_length(ccv$instance, 4)
  expect_length(ccv$train_set(3), 370)

  ccv$instantiate(task, col = "job")
  expect_class(ccv$instance, "list")
  expect_length(ccv$instance, 4)
  expect_length(ccv$train_set(3), 370)
})

test_that("mlr_measures_regr.rqr", {
  tsk = tsk("california_housing")
  lrn = lrn("regr.featureless")

  expect_error(msr("regr.rqr", alpha = 2), "alpha: Element 1 is not <= 1")

  # default
  m = msr("regr.rqr")
  expect_null(m$properties)
  expect_equal(m$param_set$values$alpha, 0.5)

  # missing predict type
  preds_na = lrn$train(tsk)$predict(tsk)
  expect_warning(preds_na$score(m), "missing predict type 'quantiles'")
  score_na = suppressWarnings(unname(preds_na$score(m)))
  expect_equal(score_na, NaN)

  # proper quantile prediction
  lrn$predict_type = "quantiles"
  lrn$quantiles = c(0.25, 0.5, 0.75)
  lrn$quantile_response = 0.5
  preds = lrn$train(tsk)$predict(tsk)

  expect_number(preds$score(m))
  expect_true(preds$score(m) == 0)

  # pred_set_mean
  m2 = msr("regr.rqr", pred_set_mean = FALSE)
  expect_equal(m2$properties, c("requires_task", "requires_train_set"))
  expect_number(preds$score(m2, task = tsk, train_set = tsk$nrow))

  m_25 = msr("regr.rqr", alpha = 0.25)
  expect_number(preds$score(m_25))

  # alpha must be in predicted quantiles
  expect_error(preds$score(msr("regr.pinball", alpha = 0.1)), "Must be element of set")
})

test_that("log to text file", {
  # See #566
  console_appender = if (packageVersion("lgr") >= "0.4.0") {
    lg$inherited_appenders$console
  } else {
    lg$inherited_appenders$appenders.console
  }
  f = tempfile("mlr3test_", fileext = "log")
  th1 = console_appender$threshold

  lg$set_threshold("debug")
  lg$add_appender(lgr::AppenderFile$new(f, threshold = "debug"), name = "testappender")
  console_appender$set_threshold("warn")

  on.exit({
    lg$remove_appender("testappender")
    lg$set_threshold(0)
    console_appender$set_threshold(th1)
  })

  task = tsk("iris")
  learner = lrn("classif.featureless")
  resampling = rsmp("cv", folds = 3L)
  rr = suppressWarnings(resample(task, learner, resampling))

  lines = readLines(f)
  expect_true(any(startsWith(lines, "INFO")))
  expect_true(any(startsWith(lines, "DEBUG")))
  expect_match(lines, "'iris'", fixed = TRUE, all = FALSE)
  expect_match(lines, "'classif.featureless'", fixed = TRUE, all = FALSE)
})

test_that("logger works", {
  on.exit({
    lgr::get_logger("mlr3")$set_threshold(0L)
    lgr::get_logger("mlr3/core")$set_threshold(0L)
  })

  lgr::get_logger("mlr3")$set_threshold(NULL)
  lgr::get_logger("mlr3/core")$set_threshold(NULL)

  res = capture_output(resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L)))
  expect_match(res, "\\[mlr3\\]")

  lgr::get_logger("mlr3")$set_threshold("error")
  res = capture_output(resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L)))
  expect_equal(res, "")

  lgr::get_logger("mlr3")$set_threshold("info")
  res = capture_output(resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L)))
  expect_match(res, "\\[mlr3\\]")

  lgr::get_logger("mlr3/core")$set_threshold("error")
  res = capture_output(resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L)))
  expect_equal(res, "")

  # if the child logger is configured, it overrides the parent logger
  lgr::get_logger("mlr3")$set_threshold("info")
  res = capture_output(resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L)))
  expect_equal(res, "")
})

test_that("thresholds are restored on workers", {
  on.exit({
    lgr::get_logger("mlr3")$set_threshold(0L)
    lgr::get_logger("mlr3/core")$set_threshold(0L)
  })

  lgr::get_logger("mlr3")$set_threshold(NULL)
  lgr::get_logger("mlr3/core")$set_threshold(NULL)

  res = capture_output(with_future(future::multisession, {
    resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L))
  }))
  expect_match(res, "\\[mlr3\\]")

  lgr::get_logger("mlr3")$set_threshold("error")

  res = capture_output(with_future(future::multisession, {
    resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L))
  }))
  expect_equal(res, "")

  lgr::get_logger("mlr3")$set_threshold("info")
  res = capture_output(with_future(future::multisession, {
    resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L))
  }))
  expect_match(res, "\\[mlr3\\]")

  lgr::get_logger("mlr3/core")$set_threshold("error")
  res = capture_output(with_future(future::multisession, {
    resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L))
  }))
  expect_equal(res, "")

  # if the child logger is configured, it overrides the parent logger
  lgr::get_logger("mlr3")$set_threshold("info")
  res = capture_output(with_future(future::multisession, {
    resample(tsk("pima"), lrn("classif.featureless"), rsmp("cv", folds = 3L))
  }))
  expect_equal(res, "")
})

test_that("no encapsulation", {
  task = tsk("iris")
  learner = lrn("classif.debug")

  learner$param_set$values = list(error_train = 1)
  expect_error(learner$train(task), "classif.debug->train")

  learner$param_set$values = list(error_predict = 1)
  learner$train(task)
  expect_error(learner$predict(task), "classif.debug->predict")
})

test_that("no encapsulation / resampling", {
  learner = lrn("classif.debug", error_train = 1)
  suppressWarnings(expect_error(resample(tsk("iris"), learner, rsmp("cv", folds = 3)), "classif.debug"))
})


test_that("encapsulation", {
  task = tsk("iris")
  learner = lrn("classif.debug")
  learner$encapsulate("evaluate", lrn("classif.featureless"))
  expect_class(learner$fallback, "LearnerClassifFeatureless")

  learner$param_set$values = list(warning_train = 1)
  learner$train(task)
  expect_data_table(learner$log, min.rows = 1L)
  expect_class(learner$warnings[[1]], "Mlr3Warning")
  expect_list(learner$errors, len = 0L)

  learner$param_set$values = list(error_train = 1)
  learner$train(task)
  expect_data_table(learner$log, min.rows = 1L)
  expect_list(learner$warnings, len = 0L, any.missing = FALSE)
  expect_class(learner$errors[[1]], "Mlr3Error")

  learner$param_set$values = list(warning_predict = 1)
  learner$train(task)
  expect_data_table(learner$log, nrows = 0L)
  expect_list(learner$warnings, len = 0L)
  expect_list(learner$errors, len = 0L)
  expect_prediction(learner$predict(task))
  expect_class(learner$warnings[[1]], "Mlr3Warning")
  expect_list(learner$errors, len = 0L)

  learner$param_set$values = list(error_predict = 1)
  learner$train(task)
  expect_data_table(learner$log, nrows = 0L)
  expect_list(learner$warnings, len = 0L)
  expect_list(learner$errors, len = 0L)
  learner$predict(task)
  expect_list(learner$warnings, len = 0L)
  expect_class(learner$errors[[1]], "Mlr3Error")
})

test_that("encapsulation / resample", {
  task = tsk("iris")
  learner = lrn("classif.debug")
  learner$param_set$values = list(warning_train = 1)
  learner$encapsulate("evaluate", lrn("classif.featureless"))
  expect_class(learner$fallback, "LearnerClassifFeatureless")

  rr = resample(task, learner, rsmp("cv", folds = 3))
  expect_data_table(rr$warnings, nrows = 3L)
  expect_data_table(rr$errors, nrows = 0L)

  learner$param_set$values = list(warning_train = 1, error_predict = 1)
  rr = resample(task, learner, rsmp("cv", folds = 3))
  expect_data_table(rr$warnings, nrows = 3L)
  expect_data_table(rr$errors, nrows = 3L)
})

test_that("encapsulation / benchmark", {
  task = tsk("iris")
  learner = lrn("classif.debug")
  learner$param_set$values = list(warning_train = 1)
  learner$encapsulate("evaluate", lrn("classif.featureless"))
  expect_class(learner$fallback, "LearnerClassifFeatureless")

  bmr = benchmark(benchmark_grid(task, learner, rsmp("cv", folds = 3)))
  aggr = bmr$aggregate(conditions = TRUE)
  expect_equal(aggr$warnings, 3L)
  expect_equal(aggr$errors, 0L)

  learner$param_set$values = list(warning_train = 1, error_predict = 1)
  bmr = benchmark(benchmark_grid(task, learner, rsmp("cv", folds = 3)))
  aggr = bmr$aggregate(conditions = TRUE)
  expect_equal(aggr$warnings, 3L)
  expect_equal(aggr$errors, 3L)
})

test_that("mlr_learners", {
  expect_dictionary(mlr_learners, min_items = 1L)
  keys = mlr_learners$keys()

  for (key in keys) {
    l = lrn(key)
    if (key == "classif.debug") {
      expect_learner(l, task = tsk("iris"))
    } else {
      expect_learner(l)
    }
    if (inherits(l, "TaskClassif")) {
      expect_true(startsWith(l$id, "classif."))
    }
    if (inherits(l, "TaskRegr")) {
      expect_true(startsWith(l$id, "regr."))
    }
  }
})

test_that("mlr_learners: sugar", {
  lrn = lrn("classif.rpart", id = "foo", cp = 0.001, predict_type = "prob")
  expect_equal(lrn$id, "foo")
  expect_equal(lrn$param_set$values$cp, 0.001)
  expect_equal(lrn$predict_type, "prob")
})

test_that("as.data.table(..., objects = TRUE)", {
  tab = as.data.table(mlr_learners, objects = TRUE)
  expect_data_table(tab)
  expect_list(tab$object, "Learner", any.missing = FALSE)
})

test_that("singular sugar functions", {
  expect_task(tsk("iris"))
  expect_learner(lrn("classif.featureless"))
  expect_resampling(rsmp("cv"))
  expect_measure(msr("classif.ce"))
})

test_that("plural sugar functions", {
  expect_list(tsks("iris"), "Task", len = 1L)
  expect_list(lrns("classif.featureless"), "Learner", len = 1L)
  expect_list(rsmps("cv"), "Resampling", len = 1L)
  expect_list(msrs("classif.ce"), "Measure", len = 1L)
})

disable_encapsulation = function(learner) {
  learner$encapsulate("none")
  learner
}

enable_encapsulation = function(learner) {
  learner$encapsulate("evaluate", default_fallback(learner))
  learner
}

task = tsk("iris")
learner = lrn("classif.debug")
learner$param_set$values = list(message_train = 1, warning_train = 1, message_predict = 1, warning_predict = 1)

test_that("evaluate / single step", {
  row_ids = 1:120
  expect_message(expect_warning(disable_encapsulation(learner)$train(task, row_ids)))
  log = learner$log
  expect_data_table(log)

  expect_silent(enable_encapsulation(learner)$train(task, row_ids))
  log = learner$log
  expect_data_table(log)
  expect_data_table(log, nrows = 2L, ncols = 3L, any.missing = FALSE)
  expect_factor(log$class)
  expect_set_equal(as.character(log$class), c("output", "warning"))
  msgs = map_chr(log$condition, conditionMessage)
  expect_match(msgs, "->train()", fixed = TRUE)
  expect_true("output" %in% log$class)
  expect_true("warning" %in% log$class)
  expect_false("error" %in% log$class)

  expect_message(expect_warning(disable_encapsulation(learner)$predict(task, row_ids = 101:150)))
  log = learner$log[stage == "predict"]
  expect_data_table(log)
  expect_shape(log, nrow = 0L)

  p = enable_encapsulation(learner)$predict(task, row_ids = 101:150)
  log = learner$log[stage == "predict"]
  expect_data_table(log)
  expect_data_table(log, nrows = 2L, ncols = 3L, any.missing = FALSE)
  expect_factor(log$class)
  expect_equal(as.character(log$class), c("output", "warning"))
  msgs = map_chr(log$condition, conditionMessage)
  expect_match(msgs, "->predict()", fixed = TRUE)
})

test_that("evaluate / resample", {
  resampling = rsmp("cv", folds = 3)

  rr = suppressMessages(suppressWarnings(resample(task, disable_encapsulation(learner), resampling)))
  expect_true(every(get_private(rr)$.data$data$fact$learner_state, function(x) nrow(x$log) == 0L))

  expect_silent(rr <- resample(task, enable_encapsulation(learner), resampling))
  expect_true(every(get_private(rr)$.data$data$fact$learner_state, function(x) all(table(x$log$stage) == 2)))
})

test_that("errors and warnings are printed with logger", {
  task = tsk("spam")

  old_threshold = lg$threshold
  lg$set_threshold("warn")
  on.exit({
    lg$set_threshold(old_threshold)
  })

  learner = enable_encapsulation(lrn("classif.debug", error_train = 1))
  expect_output(learner$train(task), "ERROR")

  learner = disable_encapsulation(lrn("classif.debug", error_train = 1))
  expect_error(learner$train(task))

  learner = enable_encapsulation(lrn("classif.debug", warning_train = 1))
  expect_output(learner$train(task), "WARN")

  learner = disable_encapsulation(lrn("classif.debug", warning_train = 1))
  expect_warning(learner$train(task))
})

test_that("encapsulate methods produce the same results", {
  rng_state = .GlobalEnv$.Random.seed
  on.exit({
    .GlobalEnv$.Random.seed = rng_state
  })

  set.seed(123)
  learner = lrn("classif.debug")
  learner$train(task)
  expect_equal(learner$model$random_number, 2986)
  expect_equal(sample(seq(1000), 1), 818)
  rr = resample(task, learner, rsmp("cv", folds = 3), store_models = TRUE)
  expect_equal(map_int(rr$learners, function(learner) learner$model$random_number), c(37151, 94567, 21057))

  set.seed(123)
  learner = lrn("classif.debug")
  learner$encapsulate("try", lrn("classif.featureless"))
  learner$train(task)
  expect_equal(learner$model$random_number, 2986)
  expect_equal(sample(seq(1000), 1), 818)
  rr = resample(task, learner, rsmp("cv", folds = 3), store_models = TRUE)
  expect_equal(map_int(rr$learners, function(learner) learner$model$random_number), c(37151, 94567, 21057))

  set.seed(123)
  learner = lrn("classif.debug")
  learner$encapsulate("evaluate", lrn("classif.featureless"))
  learner$train(task)
  expect_equal(learner$model$random_number, 2986)
  expect_equal(sample(seq(1000), 1), 818)
  rr = resample(task, learner, rsmp("cv", folds = 3), store_models = TRUE)
  expect_equal(map_int(rr$learners, function(learner) learner$model$random_number), c(37151, 94567, 21057))

  set.seed(123)
  learner = lrn("classif.debug")
  learner$encapsulate("callr", lrn("classif.featureless"))
  learner$train(task)
  expect_equal(learner$model$random_number, 2986)
  expect_equal(sample(seq(1000), 1), 818)
  rr = resample(task, learner, rsmp("cv", folds = 3), store_models = TRUE)
  expect_equal(map_int(rr$learners, function(learner) learner$model$random_number), c(37151, 94567, 21057))

  set.seed(123)
  learner = lrn("classif.debug")
  learner$encapsulate("mirai", lrn("classif.featureless"))
  with_mirai(
    {
      learner$train(task)
    },
    compute = "mlr3_encapsulation"
  )
  expect_equal(learner$model$random_number, 2986)
  expect_equal(sample(seq(1000), 1), 818)
  rr = resample(task, learner, rsmp("cv", folds = 3), store_models = TRUE)
  expect_equal(map_int(rr$learners, function(learner) learner$model$random_number), c(37151, 94567, 21057))
})

test_that("mirai compute profile can be changed", {
  old_opts = getOption("mlr3.mirai_encapsulation")
  on.exit(options(mlr3.mirai_encapsulation = old_opts), add = TRUE)

  with_mirai(
    {
      options(mlr3.mirai_encapsulation = "mlr3_encapsulation2")
      task = tsk("pima")
      learner = lrn("classif.debug")
      learner$encapsulate("mirai", lrn("classif.featureless"))
      learner$train(task)
    },
    compute = "mlr3_encapsulation2"
  )
})

test_that("extract_pkgs works", {
  skip_if_not_installed("remotes")

  # this is a noop, mlr3 is stripped from required packages
  expect_character(install_pkgs("mlr3"), len = 0)

  expect_set_equal(
    extract_pkgs("rpart"),
    "rpart"
  )

  expect_set_equal(
    extract_pkgs(lrn("classif.rpart")),
    c("mlr3", "rpart")
  )

  expect_set_equal(
    extract_pkgs(msr("classif.auc")),
    c("mlr3", "mlr3measures")
  )

  expect_set_equal(
    extract_pkgs(tgen("xor")),
    c("mlr3", "mlbench")
  )

  rr = resample(tsk("mtcars"), lrn("regr.featureless"), rsmp("holdout"))
  expect_set_equal(
    extract_pkgs(rr),
    c("mlr3", "stats")
  )

  bmr = as_benchmark_result(rr)
  expect_set_equal(
    extract_pkgs(rr),
    c("mlr3", "stats")
  )

  expect_set_equal(
    extract_pkgs(list(lrns(c("regr.rpart", "regr.featureless")))),
    c("mlr3", "rpart", "stats")
  )
})

test_that("learner is hotstarted when the hotstart parameter is increased", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  id = learner_1$model$id

  learner = lrn("classif.debug", iter = 2)
  hot = HotstartStack$new(list(learner_1))
  learner$hotstart_stack = hot
  learner$train(task)

  expect_equal(learner$model$id, id)
  expect_equal(learner$model$iter, 2)
  expect_equal(learner$param_set$values$iter, 2)
})

test_that("learner is directly returned when hotstart and target learner are equal", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  id = learner_1$model$id

  learner = lrn("classif.debug", iter = 1)
  hot = HotstartStack$new(list(learner_1))
  learner$hotstart_stack = hot
  learner$train(task)

  expect_equal(learner$model$id, id)
})

test_that("learner is trained when target learner has an additional non-hotstart parameter", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  id = learner_1$model$id

  learner = lrn("classif.debug", iter = 2, x = 1)
  hot = HotstartStack$new(list(learner_1))
  learner$hotstart_stack = hot
  learner$train(task)

  expect_true(learner$model$id != id)
  expect_equal(learner$param_set$values$x, 1)
  expect_equal(learner$param_set$values$iter, 2)
  expect_equal(learner$model$iter, 2)
})

test_that("learner is trained when target learner has an additional hotstart parameter", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", x = 0)
  learner_1$train(task)
  learner_1$state$param_vals$iter = NULL # iter set by default. Assume it is not.
  id = learner_1$model$id

  learner = lrn("classif.debug", iter = 5)
  hot = HotstartStack$new(list(learner_1))
  learner$hotstart_stack = hot
  learner$train(task)

  expect_true(learner$model$id != id)
})

test_that("learner is trained when target learner when both hotstart and non-hotstart parameters are increased", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  id = learner_1$model$id

  learner = lrn("classif.debug", iter = 2, x = 1)
  hot = HotstartStack$new(list(learner_1))
  learner$hotstart_stack = hot
  learner$train(task)

  expect_true(learner$model$id != id)
})

test_that("learner is trained when the hotstart parameter are equal but the non-hotstart parameter is altered", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1, x = 0)
  learner_1$train(task)
  id = learner_1$model$id

  learner = lrn("classif.debug", iter = 1, x = 1)
  hot = HotstartStack$new(list(learner_1))
  learner$hotstart_stack = hot
  learner$train(task)

  expect_true(learner$model$id != id)
})

test_that("learner is hotstarted when the non-hotstart parameter are equal but the hotstart parameter is increased", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1, x = 0)
  learner_1$train(task)
  id = learner_1$model$id

  learner = lrn("classif.debug", iter = 2, x = 0)
  hot = HotstartStack$new(list(learner_1))
  learner$hotstart_stack = hot
  learner$train(task)

  expect_equal(learner$model$id, id)
})

test_that("learner is trained when the task is different", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_1$train(task)
  id = learner_1$model$id

  learner = lrn("classif.debug", iter = 2)
  hot = HotstartStack$new(list(learner_1))
  learner$hotstart_stack = hot
  learner$train(tsk("iris"))

  expect_true(learner$model$id != id)
})

test_that("learner is trained when the stack is empty", {
  task = tsk("pima")
  learner = lrn("classif.debug", iter = 2)
  learner$hotstart_stack = HotstartStack$new()
  learner$train(tsk("iris"))

  expect_class(learner$model, "classif.debug_model")
})

test_that("learners are hotstarted when resample is used", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  resampling = rsmp("cv", folds = 3)
  resampling$instantiate(task)

  rr = resample(task, learner_1, resampling, store_models = TRUE)

  learner = lrn("classif.debug", iter = 2)
  hot = HotstartStack$new(rr$learners)
  learner$hotstart_stack = hot

  rr_2 = resample(task, learner, resampling, store_models = TRUE, allow_hotstart = TRUE)
  pmap(list(rr$learners, rr_2$learners), function(l1, l2) {
    expect_equal(l2$param_set$values$iter, 2)
    expect_class(l2$model, "classif.debug_model")
    expect_equal(l2$model$iter, 2)
    expect_equal(l1$model$id, l2$model$id)
    expect_null(l2$hotstart_stack)
  })
})

test_that("learners are hotstarted when benchmark is called", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_2 = lrn("classif.debug", iter = 2)
  resampling = rsmp("cv", folds = 3)
  resampling$instantiate(task)

  design = data.table(task = list(task), learner = list(learner_1, learner_2), resampling = list(resampling))
  bmr = benchmark(design, store_models = TRUE)

  learners = unlist(map(seq_len(bmr$n_resample_results), function(i) bmr$resample_result(i)$learners))
  hot = HotstartStack$new(learners)
  ids = map_chr(learners, function(l) l$model$id)

  learner = lrn("classif.debug", iter = 3)
  learner$hotstart_stack = hot

  design = benchmark_grid(task, learner, resampling)
  bmr_2 = benchmark(design, store_models = TRUE, allow_hotstart = TRUE)

  walk(bmr_2$resample_result(1)$learners, function(l1) {
    expect_equal(l1$param_set$values$iter, 3)
    expect_class(l1$model, "classif.debug_model")
    expect_equal(l1$model$iter, 3)
    expect_true(l1$model$id %in% ids)
    expect_null(l1$hotstart_stack)
  })
})

test_that("learners are trained and hotstarted when benchmark is called", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  learner_2 = lrn("classif.debug", iter = 2)
  resampling = rsmp("cv", folds = 3)
  resampling$instantiate(task)

  design = data.table(task = list(task), learner = list(learner_1, learner_2), resampling = list(resampling))
  bmr = benchmark(design, store_models = TRUE)

  learners = unlist(map(seq_len(bmr$n_resample_results), function(i) bmr$resample_result(i)$learners))
  hot = HotstartStack$new(learners)
  ids = map_chr(learners, function(l) l$model$id)

  learner_3 = lrn("classif.debug", iter = 4)
  learner_3$hotstart_stack = hot
  learner_4 = lrn("classif.rpart")
  learner_4$hotstart_stack = hot

  design = benchmark_grid(task, list(learner_3, learner_4), resampling)
  bmr_2 = benchmark(design, store_models = TRUE, allow_hotstart = TRUE)

  walk(bmr_2$resample_result(1)$learners, function(l1) {
    expect_equal(l1$param_set$values$iter, 4)
    expect_class(l1$model, "classif.debug_model")
    expect_equal(l1$model$iter, 4)
    expect_true(l1$model$id %in% ids[4:6])
    expect_null(l1$hotstart_stack)
  })

  walk(bmr_2$resample_result(2)$learners, function(l1) {
    expect_class(l1$model, "rpart")
    expect_null(l1$hotstart_stack)
  })
})

test_that("learners are cloned when hotstarting is applied", {
  task = tsk("pima")
  learner_1 = lrn("classif.debug", iter = 1)
  resampling = rsmp("holdout")
  resampling$instantiate(task)

  design = benchmark_grid(task, learner_1, resampling)
  bmr = benchmark(design, store_models = TRUE, allow_hotstart = TRUE)

  learner_2 = lrn("classif.debug", iter = 2)
  hot = HotstartStack$new(bmr$resample_result(1)$learners)
  learner_2$hotstart_stack = hot

  design = benchmark_grid(task, learner_2, resampling)
  bmr = benchmark(design, store_models = TRUE, allow_hotstart = TRUE)

  expect_equal(bmr$resample_result(1)$learners[[1]]$param_set$values$iter, 2)
  expect_equal(bmr$resample_result(1)$learners[[1]]$model$iter, 2)
  expect_equal(hot$stack$start_learner[[1]]$param_set$values$iter, 1)
  expect_equal(hot$stack$start_learner[[1]]$model$iter, 1)
  expect_equal(bmr$resample_result(1)$learners[[1]]$model$id, hot$stack$start_learner[[1]]$model$id)
})

test_that("hotstarting works when col role is set in task", {
  task = tsk("pima")
  task$col_roles$stratum = task$target_names
  learner_1 = lrn("classif.debug", iter = 1)
  resampling = rsmp("holdout")
  resampling$instantiate(task)

  rr = resample(task, learner_1, resampling, store_models = TRUE)

  learner = lrn("classif.debug", iter = 2)
  hot = HotstartStack$new(rr$learners)
  learner$hotstart_stack = hot

  rr_2 = resample(task, learner, resampling, store_models = TRUE, allow_hotstart = TRUE)
  pwalk(list(rr$learners, rr_2$learners), function(l1, l2) {
    expect_equal(l2$param_set$values$iter, 2)
    expect_class(l2$model, "classif.debug_model")
    expect_equal(l2$model$iter, 2)
    expect_equal(l1$model$id, l2$model$id)
    expect_null(l2$hotstart_stack)
  })
})

test_that("mlr_measures", {
  expect_dictionary(mlr_measures, min_items = 1L)
  keys = mlr_measures$keys()

  for (key in keys) {
    m = mlr_measures$get(key)
    expect_measure(m)
  }
})

test_that("as.data.table(mlr_measures)", {
  d = as.data.table(mlr_measures)
  expect_data_table(d)
  expect_character(d$key, unique = TRUE, any.missing = FALSE)
  expect_subset(d$task_type, c(mlr_reflections$task_types$type, NA))
  qexpectr(d$packages, "S")
  expect_subset(d$predict_type, c(NA, unlist(mlr_reflections$learner_predict_types)))
  qexpectr(d$task_properties, "S")
  expect_subset(unlist(d$task_properties), unlist(mlr_reflections$task_properties))
})

test_that("custom aggregation", {
  task = tsk("wine")
  lrn = lrn("classif.featureless")

  m = msr("classif.ce")
  m$id = "max_ce"
  m$aggregator = max
  measures = list(msr("classif.ce"), m)

  rr = resample(task, lrn, rsmp("cv", folds = 3))
  perf = rr$score(measures)
  aggr = rr$aggregate(measures)
  expect_equal(aggr[["max_ce"]], max(perf$classif.ce))
})

test_that("as.data.table(..., objects = TRUE)", {
  tab = as.data.table(mlr_measures, objects = TRUE)
  expect_data_table(tab)
  expect_list(tab$object, "Measure", any.missing = FALSE)
})

expect_hash_changes = function(x) {
  id_before = x$id
  hash_before = x$hash

  x$id = "foo"
  testthat::expect_false(identical(x$id, id_before))
  if (!is.na(hash_before)) {
    testthat::expect_false(identical(x$hash, hash_before))
  }
  x$id = id_before
  testthat::expect_identical(x$id, id_before)
  testthat::expect_identical(x$hash, hash_before)
  if ("use_weights" %in% names(x)) {
    use_weights_before = x$use_weights
    x$use_weights = "ignore"
    testthat::expect_false(identical(x$use_weights, use_weights_before))
    if (!is.na(hash_before)) {
      testthat::expect_false(identical(x$hash, hash_before))
    }
    x$use_weights = use_weights_before
    testthat::expect_identical(x$use_weights, use_weights_before)
    testthat::expect_identical(x$hash, hash_before)
  }
}

test_that("task$hash", {
  x = tsk("iris")
  expect_hash_changes(x)
})

test_that("learner$hash", {
  x = lrn("classif.rpart")
  expect_hash_changes(x)
})

test_that("measure$hash", {
  x = msr("classif.ce")
  expect_hash_changes(x)
})

test_that("resampling$hash", {
  x = rsmp("cv")
  expect_hash_changes(x)
})

test_that("score_roc_measure works", {
  learner = lrn("classif.rpart", predict_type = "prob")
  splits = partition(task = tsk("pima"), ratio = 0.7)
  task = tsk("pima")
  learner$train(task)
  pred = learner$predict(task)
  res = score_roc_measures(pred)

  expect_list(res, len = 2)
  expect_named(res, c("confusion_matrix", "measures"))
  expect_named(
    res$measures,
    c("tpr", "fpr", "fnr", "tnr", "ppv", "fdr", "npv", "fomr", "acc", "lr_plus", "lr_minus", "dor")
  )
})

test_that("partition two way split", {
  task = tsk("pima")
  li = partition(task, ratio = 0.66)
  expect_list(li, len = 3L)
  expect_names(names(li), identical.to = c("train", "test", "validation"))
  expect_equal(length(li$train), 507)
  expect_equal(length(li$test), 261)
  expect_equal(length(li$validation), 0)

  expect_disjunct(li$train, li$test)
  expect_disjunct(li$train, li$validation)
})

test_that("partition three way split", {
  task = tsk("pima")
  li = partition(task, ratio = c(0.66, 0.14))
  expect_list(li, len = 3L)
  expect_names(names(li), identical.to = c("train", "test", "validation"))
  expect_equal(length(li$train), 507)
  expect_equal(length(li$test), 107)
  expect_equal(length(li$validation), 154)

  expect_disjunct(li$train, li$test)
  expect_disjunct(li$train, li$validation)
})

skip_if_not_installed("future")

test_that("parallel resample", {
  skip_if_not_installed("future.callr")
  skip_if_not_installed("progressr")

  with_future(future.callr::callr, {
    task = tsk("iris")
    learner = lrn("classif.rpart")

    progressr::with_progress({
      rr = resample(task, learner, rsmp("cv", folds = 3))
    })
    expect_resample_result(rr)
    expect_data_table(rr$errors, nrows = 0L)
  })
})

test_that("seeds work identical during sequential and parallel execution", {
  task = tsk("sonar")
  learner = lrn("classif.debug", predict_type = "prob")
  resampling = rsmp("cv", folds = 3L)
  measure = msr("classif.auc")

  # nolint nexts
  rr1 = with_seed(123, with_future(future::sequential, resample(task, learner, resampling)))
  # nolint next
  rr2 = with_seed(123, with_future(future::multisession, resample(task, learner, resampling)))

  expect_equal(
    as.data.table(rr1$prediction())$prob.M,
    as.data.table(rr2$prediction())$prob.M
  )
})

test_that("parallel benchmark", {
  skip_if_not_installed("future.callr")

  task = tsk("iris")
  learner = lrn("classif.rpart")

  with_future(future.callr::callr, {
    progressr::with_progress({
      bmr = benchmark(benchmark_grid(task, learner, rsmp("cv", folds = 3)))
    })
  })
  expect_benchmark_result(bmr)
  expect_equal(bmr$aggregate(conditions = TRUE)$warnings, 0L)
  expect_equal(bmr$aggregate(conditions = TRUE)$errors, 0L)

  grid = benchmark_grid(
    list(tsk("wine"), tsk("sonar")),
    list(lrn("classif.debug", id = "learner_1"), lrn("classif.debug", id = "learner_2")),
    rsmp("cv", folds = 2)
  )
  njobs = 3L
  bmr = with_future(
    future::multisession,
    {
      benchmark(grid, store_models = TRUE)
    },
    workers = njobs
  )

  expect_benchmark_result(bmr)
  pids = map_int(as.data.table(bmr)$learner, function(x) x$model$pid)
  expect_equal(length(unique(pids)), njobs)
})

test_that("real parallel resample", {
  skip_if_not_installed("progressr")
  skip_on_os("windows") # currently buggy

  with_future(future::multisession, {
    task = tsk("iris")
    learner = lrn("classif.rpart")

    progressr::with_progress({
      rr = resample(task, learner, rsmp("cv", folds = 3))
    })
    expect_resample_result(rr)
    expect_data_table(rr$errors, nrows = 0L)
  })
})

test_that("parallel seed", {
  skip_if_not_installed("future.callr")

  task = tsk("wine")
  learner = lrn("classif.debug", predict_type = "prob")

  # nolint next
  rr1 = with_seed(123, resample(task, learner, rsmp("cv", folds = 3)))
  with_future(future.callr::callr, {
    # nolint next
    rr2 = with_seed(123, resample(task, learner, rsmp("cv", folds = 3)))
  })
  expect_equal(rr1$prediction()$prob, rr2$prediction()$prob)
})

test_that("data table threads are not changed in main session", {
  skip_on_os("mac") # number of threads cannot be changed on mac
  skip_on_cran()

  old_dt_threads = getDTthreads()
  on.exit(
    {
      setDTthreads(old_dt_threads)
    },
    add = TRUE
  )
  setDTthreads(2L)

  task = tsk("sonar")
  learner = lrn("classif.debug", predict_type = "prob")
  resampling = rsmp("cv", folds = 3L)
  measure = msr("classif.auc")

  # nolint next
  rr1 = with_seed(123, with_future(future::sequential, resample(task, learner, resampling)))
  expect_equal(getDTthreads(), 2L)

  # nolint next
  rr2 = with_seed(123, with_future(future::multisession, resample(task, learner, resampling)))
  expect_equal(getDTthreads(), 2L)
})

task = tsk("penguins")
learner = lrn("classif.rpart")
rr = resample(task, learner, rsmp("cv", folds = 3), store_models = TRUE)

test_that("similarity measures", {
  measures = mlr_measures$keys("^sim\\.")

  for (m in msrs(measures)) {
    expect_number(m$aggregate(rr))
    expect_true(is.na(m$score(rr$prediction(), learner = rr$learners[[1]])))
    expect_numeric(rr$aggregate(msrs(c("classif.acc", m$id))), len = 2, any.missing = FALSE)
    expect_true(allMissing(rr$score(m)[[m$id]]))
  }
})

test_that("similarity example", {
  task = tsk("penguins")
  learners = list(
    lrn("classif.rpart", maxdepth = 1, id = "r1"),
    lrn("classif.rpart", maxdepth = 2, id = "r2")
  )
  resampling = rsmp("cv", folds = 3)
  grid = benchmark_grid(task, learners, resampling)
  bmr = benchmark(grid, store_models = TRUE)
  x = bmr$aggregate(msr("sim.jaccard"))
  expect_numeric(x$sim.jaccard, any.missing = FALSE)
})

test_that("mlr_measures_auc", {
  task = tsk("sonar")
  lrn = lrn("classif.featureless", predict_type = "prob")
  m = msr("classif.auc")
  measures = list(m)

  p = lrn$train(task)$predict(task)
  perf = p$score(measures)
  expect_equal(unname(perf), 0.5)
})

old_mlr_reflections = as.environment(as.list(mlr_reflections, all.names = TRUE))

on.exit({
  mlr_reflections = as.environment(as.list(old_mlr_reflections, all.names = TRUE))
})

mlr_reflections$task_types = setkeyv(
  rbind(
    mlr_reflections$task_types,
    rowwise_table(
      ~type,
      ~package,
      ~task,
      ~learner,
      ~prediction,
      ~prediction_data,
      ~measure,
      "test",
      "mlr3",
      "TaskClassifTest",
      "LearnerClassif",
      "PredictionClassif",
      "PredictionDataClassif",
      "MeasureClassif"
    )
  ),
  "type"
)

mlr_reflections$task_col_roles$test = c(mlr_reflections$task_col_roles$classif, "test")
mlr_reflections$task_properties$test = mlr_reflections$task_properties$classif
mlr_reflections$default_measures$test = "classif.ce"
mlr_reflections$learner_predict_types$test = mlr_reflections$learner_predict_types$classif

TaskClassifTest = R6Class(
  "TaskClassifTest",
  inherit = TaskClassif,
  public = list(
    initialize = function(id, backend, target, positive = NULL, label = NA_character_, extra_args = list()) {
      super$initialize(id = id, backend = backend, target = target, label = label, extra_args = extra_args)

      self$task_type = "test"
      new_col_roles = named_list(
        setdiff(mlr_reflections$task_col_roles[["test"]], names(private$.col_roles)),
        character(0)
      )
      private$.col_roles = insert_named(private$.col_roles, new_col_roles)

      update_classif_property(self, private)
    }
  )
)

b = as_data_backend(load_dataset("PimaIndiansDiabetes2", "mlbench"))
task = TaskClassifTest$new("test", b, target = "diabetes", positive = "pos", label = "Pima Indian Diabetes")
learner = lrn("classif.rpart")
measure = msr("classif.ce")

test_that("assertions work", {
  expect_learner(assert_learner(learner, task))
  expect_error(assert_learner(learner, tsk("california_housing")), "must have task type")
  expect_null(assert_task_learner(task, learner))
  expect_error(assert_task_learner(tsk("california_housing"), learner), "not match type")
  expect_measure(assert_measure(measure, task, learner))
  expect_error(assert_measure(measure, tsk("california_housing"), learner), "is not compatible")
  expect_error(assert_measure(measure, task, lrn("regr.rpart")), "is not compatible")

  at = learner
  class(at) = c("AutoTuner", "Learner", "R6")
  expect_learner(assert_learner(at, task))
  expect_error(assert_learner(at, tsk("california_housing")), "must have task type")
  expect_null(assert_task_learner(task, at))
  expect_error(assert_task_learner(tsk("california_housing"), at))
})

test_that("train and predict works", {
  expect_learner(learner$train(task))
  expect_prediction(learner$predict(task))
})

test_that("resampling works", {
  rr = resample(task, learner, rsmp("cv", folds = 3))
  expect_equal(rr$task_type, "test")

  scores = rr$score(msr("classif.ce"), predictions = TRUE)
  expect_list(scores$prediction_test, "Prediction")
  expect_numeric(scores$classif.ce, any.missing = FALSE)
  expect_number(rr$aggregate(msr("classif.ce")))

  scores = rr$score(predictions = TRUE)
  expect_list(scores$prediction_test, "Prediction")
  expect_numeric(scores$classif.ce, any.missing = FALSE)
  expect_number(rr$aggregate(msr("classif.ce")))
})

test_that("benchmark works", {
  grid = benchmark_grid(task, learner, rsmp("cv", folds = 3))
  bmr = benchmark(grid)
  expect_benchmark_result(bmr)
  expect_equal(bmr$task_type, "test")
  tab = bmr$aggregate(measure)
  expect_data_table(tab, nrows = 1L)
  expect_names(
    names(tab),
    type = "unique",
    identical.to = c("nr", "resample_result", "task_id", "learner_id", "resampling_id", "iters", "classif.ce")
  )

  grid = benchmark_grid(list(task, tsk("mtcars"), tsk("california_housing")), learner, rsmp("cv", folds = 3))
  expect_error(benchmark(grid), "Multiple task types detected")
})

test_that("set column roles works", {
  expect_task(task$set_col_roles("age", "test"))
  expect_equal(task$col_roles$test, "age")
})

test_that("external packages can set column roles", {
  x = utils::getFromNamespace("mlr_reflections", ns = "mlr3")
  old_col_roles = x$task_col_roles$classif

  on.exit(
    {
      x$col_roles$classif = old_col_roles
    },
    add = TRUE
  )

  x$task_col_roles$classif = c(x$task_col_roles$classif, "extra_role")

  task = tsk("pima")

  with_future(future::multisession, {
    rr = resample(task, lrn("classif.rpart"), rsmp("cv", folds = 3))
  })
  expect_resample_result(rr)
})

test_that("PredictionDataRegr", {
  task = tsk("mtcars")
  learner = lrn("regr.featureless", predict_type = "se")
  p = learner$train(task)$predict(task)
  pdata = p$data

  expect_s3_class(pdata, "PredictionDataRegr")
  expect_integer(pdata$row_ids, any.missing = FALSE)
  expect_numeric(pdata$truth, any.missing = FALSE)
  expect_numeric(pdata$response, any.missing = FALSE)
  expect_numeric(pdata$se, any.missing = FALSE)

  expect_s3_class(c(pdata, pdata), "PredictionDataRegr")
  expect_prediction(as_prediction(pdata))
  expect_equal(as.data.table(p), as.data.table(as_prediction(pdata)))

  pdata = filter_prediction_data(pdata, row_ids = 1:3)
  expect_set_equal(pdata$row_ids, 1:3)
  expect_numeric(pdata$truth, len = 3)
  expect_numeric(pdata$response, len = 3)
})

test_that("construction of empty PredictionDataRegr", {
  task = tsk("mtcars")

  learner = lrn("regr.featureless")
  learner$train(task)
  pred = learner$predict(task, row_ids = integer())
  expect_prediction(pred)
  expect_set_equal(pred$predict_types, "response")
  expect_integer(pred$row_ids, len = 0L)
  expect_numeric(pred$truth, len = 0L)
  expect_null(pred$data$se)
  expect_null(pred$data$distr)
  expect_data_table(as.data.table(pred), nrows = 0L, ncols = 3L)

  learner = lrn("regr.featureless", predict_type = "se")
  learner$train(task)
  pred = learner$predict(task, row_ids = integer())
  expect_prediction(pred)
  expect_set_equal(pred$predict_types, c("response", "se"))
  expect_integer(pred$row_ids, len = 0L)
  expect_numeric(pred$truth, len = 0L)
  expect_numeric(pred$se, len = 0L)
  expect_data_table(as.data.table(pred), nrows = 0L, ncols = 4L)
})

test_that("PredictionDataRegr with quantiles", {
  n = 100
  probs = c(0.1, 0.5, 0.9)
  y = runif(n)
  task = as_task_regr(data.table(y = y), target = "y")

  quantiles = quantile(y, probs = probs)
  quantiles = matrix(rep(quantiles, n), nrow = n, byrow = TRUE)
  setattr(quantiles, "probs", probs)
  setattr(quantiles, "response", 0.5)

  data = list(quantiles = quantiles)
  pdata = as_prediction_data(data, task)

  pred = as_prediction(pdata)
  expect_prediction_regr(pred)
})

test_that("c.PredictionDataRegr preserves probs attribute on quantiles", {
  n = 10
  probs = c(0.1, 0.5, 0.9)
  y = runif(n)
  task = as_task_regr(data.table(y = y), target = "y")

  quantiles = quantile(y, probs = probs)
  quantiles = matrix(rep(quantiles, n), nrow = n, byrow = TRUE)
  setattr(quantiles, "probs", probs)
  setattr(quantiles, "response", 0.5)

  pdata = as_prediction_data(list(quantiles = quantiles), task)
  combined = c(pdata, pdata)

  expect_equal(attr(combined$quantiles, "probs"), probs)
  expect_equal(attr(combined$quantiles, "response"), "q0.5")
})

test_that("PredictionDataRegr with quantiles and response", {
  n = 100
  probs = c(0.1, 0.9)
  y = runif(n)
  task = as_task_regr(data.table(y = y), target = "y")

  quantiles = quantile(y, probs = probs)
  quantiles = matrix(rep(quantiles, n), nrow = n, byrow = TRUE)
  setattr(quantiles, "probs", probs)

  data = list(quantiles = quantiles, response = rep(0.5, n))
  pdata = as_prediction_data(data, task)

  pred = as_prediction(pdata)
  expect_prediction_regr(pred)
})

test_that("combine with extra data", {
  pdata = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = runif(2),
      response = runif(2),
      extra = list(extra_col = c("a", "b"), extra_col2 = c(1, 2))
    ),
    "regr"
  )

  pdata2 = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = runif(2),
      response = runif(2),
      extra = list(extra_col = c("c", "d"), extra_col2 = c(3, 4))
    ),
    "regr"
  )

  expect_equal(c(pdata, pdata2)$extra, list(extra_col = c("a", "b", "c", "d"), extra_col2 = c(1, 2, 3, 4)))
  expect_equal(c(pdata, pdata2, keep_duplicates = FALSE)$extra, list(extra_col = c("c", "d"), extra_col2 = c(3, 4)))

  pdata2 = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = runif(2),
      response = runif(2),
      extra = list(extra_col = c("c", "d"), extra_col3 = c(3, 4))
    ),
    "regr"
  )

  expect_equal(
    c(pdata, pdata2)$extra,
    list(extra_col = c("a", "b", "c", "d"), extra_col2 = c(1, 2, NA, NA), extra_col3 = c(NA, NA, 3, 4))
  )

  pdata2 = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = runif(2),
      response = runif(2),
      prob = matrix(c(0.5, 0.5, 0.5, 1), 2)
    ),
    "regr"
  )

  expect_error(c(pdata, pdata2), "Some predictions have extra data, others do not")
})

test_that("filtering with extra data", {
  pdata = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = runif(2),
      response = runif(2),
      extra = list(extra_col = c("a", "b"), extra_col2 = c(1, 2))
    ),
    "regr"
  )

  expect_equal(filter_prediction_data(pdata, row_ids = 1)$extra, list(extra_col = "a", extra_col2 = 1))
  expect_equal(filter_prediction_data(pdata, row_ids = 2)$extra, list(extra_col = "b", extra_col2 = 2))
})

test_that("Regression measures", {
  keys = mlr_measures$keys("^regr\\.")
  task = tsk("california_housing")
  learner = lrn("regr.rpart")
  learner$train(task)
  p = learner$predict(task)

  for (key in keys) {
    m = mlr_measures$get(key)

    if (is.na(m$task_type) || m$task_type == "regr") {
      if (m$predict_type == "quantiles") {
        learner_q = lrn("regr.featureless", predict_type = "quantiles", quantiles = 0.5)
        learner_q$train(task)
        p_q = learner_q$predict(task)
        perf = m$score(prediction = p_q, task = task, learner = learner_q)
        expect_number(perf, na.ok = FALSE, lower = m$range[1], upper = m$range[2])
      } else {
        perf = m$score(prediction = p, task = task, learner = learner)
        expect_number(perf, na.ok = FALSE, lower = m$range[1], upper = m$range[2])
      }
    }
  }
})

#' @include DataBackend.R
DataBackendRename = R6Class(
  "DataBackendRename",
  inherit = DataBackend,
  cloneable = FALSE,
  public = list(
    old = NULL,
    new = NULL,

    initialize = function(b, old, new) {
      super$initialize(data = b, b$primary_key)
      assert_character(old, any.missing = FALSE, unique = TRUE)
      assert_subset(old, b$colnames)
      assert_character(new, any.missing = FALSE, len = length(old))
      assert_names(new, "unique")

      ii = old != new
      old = old[ii]
      new = new[ii]

      if (self$primary_key %chin% old) {
        error_input("Renaming the primary key is not supported")
      }

      resulting_names = map_values(b$colnames, old, new)
      dup = anyDuplicated(resulting_names)
      if (dup > 0L) {
        error_input("Duplicated column name after rename: %s", resulting_names[dup])
      }

      self$old = old
      self$new = new
    },

    data = function(rows, cols) {
      assert_names(cols, type = "unique")
      b = private$.data
      cols = map_values(intersect(cols, self$colnames), self$new, self$old)
      data = b$data(rows, cols)
      set_col_names(data, map_values(names(data), self$old, self$new))
    },

    head = function(n = 6L) {
      data = private$.data$head(n)
      set_col_names(data, map_values(names(data), self$old, self$new))
    },

    distinct = function(rows, cols, na_rm = TRUE) {
      cols = map_values(intersect(cols, self$colnames), self$new, self$old)
      x = private$.data$distinct(rows, cols, na_rm = na_rm)
      set_names(x, map_values(names(x), self$old, self$new))
    },

    missings = function(rows, cols) {
      cols = map_values(intersect(cols, self$colnames), self$new, self$old)
      x = private$.data$missings(rows, cols)
      set_names(x, map_values(names(x), self$old, self$new))
    }
  ),

  active = list(
    rownames = function() {
      private$.data$rownames
    },

    colnames = function() {
      x = private$.data$colnames
      map_values(x, self$old, self$new)
    },

    nrow = function() {
      private$.data$nrow
    },

    ncol = function() {
      private$.data$ncol
    },

    col_hashes = function() {
      res = private$.data$col_hashes
      names(res) = map_values(names(res), self$old, self$new)
      res
    }
  ),

  private = list(
    .calculate_hash = function() {
      calculate_hash(self$old, self$new, private$.data$hash)
    }
  )
)

#' @title Convert to a Regression Prediction
#'
#' @description
#' Convert object to a [PredictionRegr].
#'
#' @inheritParams as_prediction
#'
#' @return [PredictionRegr].
#' @export
#' @examples
#' # create a prediction object
#' task = tsk("mtcars")
#' learner = lrn("regr.rpart")
#' learner$train(task)
#' p = learner$predict(task)
#'
#' # convert to a data.table
#' tab = as.data.table(p)
#'
#' # convert back to a Prediction
#' as_prediction_regr(tab)
#'
#' # split data.table into a list of data.tables
#' tabs = split(tab, cut(tab$truth, 3))
#'
#' # convert back to list of predictions
#' preds = lapply(tabs, as_prediction_regr)
#'
#' # calculate performance in each group
#' sapply(preds, function(p) p$score())
as_prediction_regr = function(x, ...) {
  UseMethod("as_prediction_regr")
}


#' @rdname as_prediction_regr
#' @export
# nolint next
as_prediction_regr.PredictionRegr = function(x, ...) {
  x
}


#' @rdname as_prediction_regr
#' @export
# nolint next
as_prediction_regr.data.frame = function(x, ...) {
  assert_names(names(x), must.include = c("row_ids", "truth", "response"))
  assert_names(names(x), subset.of = c("row_ids", "truth", "response", "se", "weights"))
  invoke(PredictionRegr$new, .args = x)
}

#' @import data.table
#' @import checkmate
#' @import cli
#' @import paradox
#' @import mlr3misc
#' @import palmerpenguins
#' @importFrom R6 R6Class is.R6
#' @importFrom utils data head tail getFromNamespace packageVersion
#' @importFrom graphics plot
#' @importFrom stats predict rnorm runif sd contr.treatment model.frame terms quantile weighted.mean
#' @importFrom uuid UUIDgenerate
#' @importFrom parallelly availableCores
#' @importFrom future nbrOfWorkers plan
#' @importFrom methods formalArgs
#'
#' @section Learn mlr3:
#' * Book on mlr3: \url{https://mlr3book.mlr-org.com}

#' * Use cases and examples gallery: \url{https://mlr3gallery.mlr-org.com}
#' * Cheat Sheets: \url{https://github.com/mlr-org/mlr3cheatsheets}
#'
#' @section mlr3 extensions:
#' * Preprocessing and machine learning pipelines: \CRANpkg{mlr3pipelines}
#' * Analysis of benchmark experiments: \CRANpkg{mlr3benchmark}
#' * More classification and regression tasks: \CRANpkg{mlr3data}
#' * Connector to [OpenML](https://www.openml.org): \CRANpkg{mlr3oml}
#' * Solid selection of good classification and regression learners: \CRANpkg{mlr3learners}
#' * Even more learners: \url{https://github.com/mlr-org/mlr3extralearners}
#' * Tuning of hyperparameters: \CRANpkg{mlr3tuning}
#' * Hyperband tuner: \CRANpkg{mlr3hyperband}
#' * Visualizations for many \pkg{mlr3} objects: \CRANpkg{mlr3viz}
#' * Survival analysis and probabilistic regression: \CRANpkg{mlr3proba}
#' * Cluster analysis: \CRANpkg{mlr3cluster}
#' * Feature selection filters: \CRANpkg{mlr3filters}
#' * Feature selection wrappers: \CRANpkg{mlr3fselect}
#' * Interface to real (out-of-memory) data bases: \CRANpkg{mlr3db}
#' * Performance measures as plain functions: \CRANpkg{mlr3measures}
#' * Resampling methods for spatiotemporal data: \CRANpkg{mlr3spatiotempcv}
#' * Data storage and prediction support for spatial objects: \CRANpkg{mlr3spatial}
#'
#' @section Suggested packages:
#' * Parallelization framework: \CRANpkg{future}
#' * Progress bars: \CRANpkg{progressr}
#' * Encapsulated evaluation: \CRANpkg{evaluate}, \CRANpkg{callr} (external process)
#'
#' @section Package Options:
#' * `"mlr3.exec_random"`: Randomize the order of execution in [resample()] and [benchmark()] during
#'   parallelization with \CRANpkg{future}. Defaults to `TRUE`.
#'   Note that this does not affect the order of results.
#' * `"mlr3.exec_chunk_size"`: Number of iterations to perform in a single [future::future()] during
#'   parallelization with \CRANpkg{future}. Defaults to 1.
#' * `"mlr3.exec_chunk_bins"`: Number of bins to split the iterations into. If set, `"mlr3.exec_chunk_size"` is ignored.
#' * `"mlr3.debug"`: If set to `TRUE`, parallelization via \CRANpkg{future} is disabled to simplify
#'   debugging and provide more concise tracebacks.
#'   Note that results computed in debug mode use a different seeding mechanism and are **not reproducible**.
#' * `"mlr3.warn_version_mismatch"`: Set to `FALSE` to silence warnings raised during predict if a learner has been
#'   trained with a different version version of mlr3.
#' * `"mlr3.prob_as_default"`: Set to `TRUE` to set the predict type of classification learners to
#'   `"prob"` by default (if they support it).
#' * `"mlr3.mirai_parallelization"`: Compute profile to use for parallelization with \CRANpkg{mirai}.
#'   Defaults to `"mlr3_parallelization"`.
#' * `"mlr3.mirai_encapsulation"`: Compute profile to use for encapsulation with \CRANpkg{mirai}.
#'   Defaults to `"mlr3_encapsulation"`.
#' * `"mlr3.print_class_ratio_threshold"`: Maximum number of rows for which class ratios are computed
#'   when printing a classification task. For tasks with more rows, only class names are shown.
#'   Defaults to `1000000`.
#'
#' @section Error Classes:
#' * `Mlr3Error`: The base mlr3 error class.
#' * `Mlr3ErrorConfig`: This error signals that the user has misconfigured something.
#'   By default, this error is not caught when the learner is encapsulated.
#' * `Mlr3ErrorInput`: This error signals that the input to the function is invalid.
#' * `Mlr3ErrorLearner`: The base error class for errors related to the learner.
#' * `Mlr3ErrorLearnerTrain`: This error signals that the learner failed to train the model.
#' * `Mlr3ErrorLearnerPredict`: This error signals that something went wrong during prediction.
#' * `Mlr3TimeoutError`: This error signals that the encapsulation during train or predict timed out.
#'
#' @section Warning Classes:
#' * `Mlr3Warning`: The base mlr3 warning class.
#' * `Mlr3WarningConfig`: This warning signals that the user has misconfigured something.
#' * `Mlr3WarningInput`: This warning signals that the input to the function is invalid.
#'
#' @references
#' `r tools::toRd(citation("mlr3"))`
"_PACKAGE"

dummy_import = function() {
  # nocov start
  # this function is required to silence R CMD check
  mlbench::mlbench.xor
  mlr3measures::mse
  evaluate::evaluate
} # nocov end


.onLoad = function(libname, pkgname) {
  # nocov start
  backports::import(pkgname)

  # callbacks
  x = utils::getFromNamespace("mlr_callbacks", ns = "mlr3misc")
  x$add("mlr3.model_extractor", load_callback_model_extractor)
  x$add("mlr3.holdout_task", load_callback_holdout_task)

  # setup logger
  lg = lgr::get_logger("mlr3/core")
  assign("lg", lg, envir = parent.env(environment()))
  f = function(event) {
    event$msg = paste0("[mlr3] ", event$msg)
    TRUE
  }
  lg$set_filters(list(f))
  if (Sys.getenv("IN_PKGDOWN") == "true") {
    lg$set_threshold("warn")
  }

  register_namespace_callback(pkgname, "mlr", function(...) {
    warning_mlr3("Packages 'mlr3' and 'mlr' are conflicting and should not be loaded in the same session")
  })
} # nocov end

leanify_package()

#' @title Dictionary of Learners
#'
#' @usage NULL
#' @format [R6::R6Class] object inheriting from [mlr3misc::Dictionary].
#'
#' @description
#' A simple [mlr3misc::Dictionary] storing objects of class [Learner].
#' Each learner has an associated help page, see `mlr_learners_[id]`.
#'
#' This dictionary can get populated with additional learners by add-on packages.
#' For an opinionated set of solid classification and regression learners,
#' install and load the \CRANpkg{mlr3learners} package.
#' More learners are connected via \url{https://github.com/mlr-org/mlr3extralearners}.
#'
#' For a more convenient way to retrieve and construct learners, see [lrn()]/[lrns()].
#'
#' @section Methods:
#' See [mlr3misc::Dictionary].
#'
#' @section S3 methods:
#' * `as.data.table(dict, ..., objects = FALSE)`\cr
#'   [mlr3misc::Dictionary] -> [data.table::data.table()]\cr
#'   Returns a [data.table::data.table()] with fields "key", "label", "task_type", "feature_types", "packages",
#'   "properties", and "predict_types" as columns.
#'   If `objects` is set to `TRUE`, the constructed objects are returned in the list column named `object`.
#'
#'
#' @family Dictionary
#' @family Learner
#' @seealso
#' Sugar functions: [lrn()], [lrns()]
#'
#' Extension Packages: \CRANpkg{mlr3learners}
#' @export
#' @examples
#' as.data.table(mlr_learners)
#' mlr_learners$get("classif.featureless")
#' lrn("classif.rpart")
mlr_learners = R6Class("DictionaryLearner", inherit = Dictionary, cloneable = FALSE, )$new()

#' @export
as.data.table.DictionaryLearner = function(x, ..., objects = FALSE) {
  assert_flag(objects)

  setkeyv(
    map_dtr(
      x$keys(),
      function(key) {
        l = withCallingHandlers(x$get(key, .prototype = TRUE), packageNotFoundWarning = function(w) {
          invokeRestart("muffleWarning")
        })
        insert_named(
          list(
            key = key,
            label = l$label,
            task_type = l$task_type,
            feature_types = list(l$feature_types),
            packages = list(l$packages),
            properties = list(l$properties),
            predict_types = list(l$predict_types)
          ),
          if (objects) list(object = list(l))
        )
      },
      .fill = TRUE
    ),
    "key"
  )[]
}

#' @title Palmer Penguins Data Set
#'
#' @name mlr_tasks_penguins
#' @format [R6::R6Class] inheriting from [TaskClassif].
#' @include mlr_tasks.R
#'
#' @description
#' Classification data to predict the species of penguins from the \CRANpkg{palmerpenguins} package,
#' see [palmerpenguins::penguins].
#' A better alternative to the [iris data set][iris].
#'
#' @templateVar id penguins
#' @template task
#'
#' @section Pre-processing:
#' * The unit of measurement have been removed from the column names.
#'   Lengths are given in millimeters (mm), weight in gram (g).
#'
#' @source \CRANpkg{palmerpenguins}
#'
#' @references
#' `r format_bib("gorman2014")`
#'
#' \url{https://github.com/allisonhorst/palmerpenguins}
#'
#' @template seealso_task
NULL

load_task_penguins = function() {
  penguins = as.data.table(palmerpenguins::penguins)
  setnames(
    penguins,
    old = c("bill_length_mm", "bill_depth_mm", "flipper_length_mm", "body_mass_g"),
    new = c("bill_length", "bill_depth", "flipper_length", "body_mass")
  )

  b = as_data_backend(penguins)
  task = TaskClassif$new("penguins", b, target = "species", label = "Palmer Penguins")
  b$hash = task$man = "mlr3::mlr_tasks_penguins"
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("penguins", load_task_penguins)

#' @title R-Squared for Quantile Regression
#'
#' @name mlr_measures_regr.rqr
#' @include Measure.R
#'
#' @description
#' Measure to compare true observed response with predicted quantiles in regression tasks.
#'
#' @details
#' \eqn{R^1(\alpha)} is defined as \deqn{
#'   1 - \frac{\sum_{i=1}^n \rho_\alpha \left( t_i - r_i(\alpha) \right)}{
#'     \sum_{i=1}^n \rho_\alpha \left( t_i - q_{\alpha} \right)},
#' }{
#'   1 - sum(pinball(t - r(alpha))) / sum(pinball(t - quantile(t, alpha))),
#' }
#' where for a quantile \eqn{\alpha}, \eqn{\rho_\alpha} is the pinball function, \eqn{r_i(\alpha)} are the predictions
#' for the quantile and \eqn{q_{\alpha}} is the empirical \eqn{\alpha}-quantile of the test or training data.
#'
#' \eqn{R^1(\alpha)} is analogous to \eqn{R^2} for regression tasks.
#' It compares the pinball function of the predictions relative to a naive model predicting the empirical quantile.
#'
#' This measure is undefined for constant \eqn{t}.
#'
#' @param pred_set_mean `logical(1)`\cr
#' If `TRUE`, the mean of the true values is calculated on the prediction set.
#' If `FALSE`, the mean of the true values is calculated on the training set.
#'
#' @param alpha `numeric(1)`\cr
#' The quantile for which to compute the measure.
#' Must be one of the quantiles that the Learner was trained on.
#
#' @references
#' `r format_bib("koenker_1999")`
#'
#' @templateVar id regr.rqr
#' @template measure
#'
#' @template seealso_measure
#' @export
MeasureRegrRQR = R6Class(
  "MeasureRegrRQR",
  inherit = MeasureRegr,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function(alpha = 0.5, pred_set_mean = TRUE) {
      private$.pred_set_mean = assert_flag(pred_set_mean)
      param_set = ps(alpha = p_dbl(lower = 0, upper = 1))
      param_set$set_values(alpha = alpha)

      super$initialize(
        id = "regr.rqr",
        param_set = param_set,
        properties = c(if (!private$.pred_set_mean) c("requires_task", "requires_train_set")),
        predict_type = "quantiles",
        minimize = FALSE,
        range = c(-Inf, 1),
        label = "R-Squared for Quantile Regression",
        man = "mlr3::mlr_measures_regr.rqr"
      )
    }
  ),

  active = list(
    #' @field pred_set_mean (`logical(1)`)\cr
    #' Whether the empirical quantile is calculated on the prediction set.
    #' Set during construction.
    pred_set_mean = function(rhs) {
      assert_ro_binding(rhs)
      private$.pred_set_mean
    }
  ),

  private = list(
    .pred_set_mean = NULL,

    .extra_hash = "pred_set_mean",

    .score = function(prediction, task = NULL, train_set = NULL, weights = NULL, ...) {
      alpha = self$param_set$values$alpha
      probs = attr(prediction$data$quantiles, "probs")
      assert_choice(alpha, probs)

      mu = if (private$.pred_set_mean) {
        quantile(prediction$truth, probs = alpha)
      } else {
        quantile(task$truth(train_set), probs = alpha)
      }

      numerator = sum(
        mlr3measures::pinball(
          truth = prediction$truth,
          response = prediction$data$quantiles[, which(probs == alpha)],
          alpha = alpha
        )
      )

      denominator = sum(
        mlr3measures::pinball(
          truth = prediction$truth,
          response = rep(mu, length(prediction$truth)),
          alpha = alpha
        )
      )

      1 - (numerator / denominator)
    }
  )
)

#' @include mlr_measures.R
mlr_measures$add("regr.rqr", MeasureRegrRQR)

#' @title Prediction Object for Classification
#'
#' @include Prediction.R
#'
#' @description
#' This object wraps the predictions returned by a learner of class [LearnerClassif], i.e.
#' the predicted response and class probabilities.
#'
#' If the response is not provided during construction, but class probabilities are,
#' the response is calculated from the probabilities: the class label with the highest
#' probability is chosen. In case of ties, a label is selected randomly.
#'
#' @note
#' If this object is constructed manually, make sure that the factor levels for `truth`
#' have the same levels as the task, in the same order.
#' In case of binary classification tasks, the positive class label must be the first level.
#'
#' @section Thresholding:
#' If probabilities are stored, it is possible to change the threshold which determines the predicted class label.
#' Usually, the label of the class with the highest predicted probability is selected.
#' For binary classification problems, such an threshold defaults to 0.5.
#' For cost-sensitive or imbalanced classification problems, manually adjusting the threshold can increase
#' the predictive performance.
#'
#' * For binary problems only a single threshold value can be set.
#'   If the probability exceeds the threshold, the positive class is predicted.
#'   If the probability equals the threshold, the label is selected randomly.
#' * For binary and multi-class problems, a named numeric vector of thresholds can be set.
#'   The length and names must correspond to the number of classes and class names, respectively.
#'   To determine the class label, the probabilities are divided by the threshold.
#'   This results in a ratio > 1 if the probability exceeds the threshold, and a ratio < 1 otherwise.
#'   Note that it is possible that either none or multiple ratios are greater than 1 at the same time.
#'   Anyway, the class label with maximum ratio is selected.
#'   In case of ties in the ratio, one of the tied class labels is selected randomly.
#'
#'   Note that there are the following edge cases for threshold equal to `0` which are handled specially:
#'   1. With threshold 0 the resulting ratio gets `Inf` and thus gets always selected.
#'      If there are multiple ratios with value `Inf`, one is selected according to `ties_method`
#'      (randomly per default).
#'   2. If additionally the predicted probability is also 0, the ratio `0/0` results in `NaN` values.
#'      These are simply replaced by `0` and thus will never get selected.
#'
#' @template seealso_prediction
#' @export
#' @examples
#' task = tsk("penguins")
#' learner = lrn("classif.rpart", predict_type = "prob")
#' learner$train(task)
#' p = learner$predict(task)
#' p$predict_types
#' head(as.data.table(p))
#'
#' # confusion matrix
#' p$confusion
#'
#' # change threshold
#' th = c(0.05, 0.9, 0.05)
#' names(th) = task$class_names
#'
#' # new predictions
#' p$set_threshold(th)$response
#' p$score(measures = msr("classif.ce"))
PredictionClassif = R6Class(
  "PredictionClassif",
  inherit = Prediction,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param task ([TaskClassif])\cr
    #'   Task, used to extract defaults for `row_ids` and `truth`.
    #'
    #' @param row_ids (`integer()`)\cr
    #'   Row ids of the predicted observations, i.e. the row ids of the test set.
    #'
    #' @param truth (`factor()`)\cr
    #'   True (observed) labels. See the note on manual construction.
    #'
    #' @param response (`character()` | `factor()`)\cr
    #'   Vector of predicted class labels.
    #'   One element for each observation in the test set.
    #'   Character vectors are automatically converted to factors.
    #'   See the note on manual construction.
    #'
    #' @param prob (`matrix()`)\cr
    #'   Numeric matrix of posterior class probabilities with one column for each class
    #'   and one row for each observation in the test set.
    #'   Columns must be named with class labels, row names are automatically removed.
    #'   If `prob` is provided, but `response` is not, the class labels are calculated from
    #'   the probabilities using [max.col()] with `ties.method` set to `"random"`.
    #'
    #' @param weights (`numeric()`)\cr
    #'   Vector of measure weights for each observation. Should be constructed from
    #'   the `Task`'s `weights_measure` column.
    #'
    #' @param check (`logical(1)`)\cr
    #'   If `TRUE`, performs some argument checks and predict type conversions.
    #'
    #' @param extra (`list()`)\cr
    #'   List of extra data to be stored in the prediction object.
    #'
    #' @param raw (any)\cr
    #'   Raw prediction object from the upstream model. Stored as-is without validation.
    initialize = function(
      task = NULL,
      row_ids = task$row_ids,
      truth = task$truth(),
      response = NULL,
      prob = NULL,
      weights = NULL,
      check = TRUE,
      extra = NULL,
      raw = NULL
    ) {
      pdata = new_prediction_data(
        list(
          row_ids = row_ids,
          truth = truth,
          response = response,
          prob = prob,
          weights = weights,
          extra = extra,
          raw = raw
        ),
        task_type = "classif"
      )

      if (check) {
        pdata = check_prediction_data(pdata)
      }
      self$task_type = "classif"
      self$man = "mlr3::PredictionClassif"
      self$data = pdata
      self$predict_types = intersect(c("response", "prob"), names(pdata))
    },

    #' @description
    #' Sets the prediction response based on the provided threshold.
    #' See the section on thresholding for more information.
    #'
    #' @param threshold (`numeric()`).
    #' @param ties_method (`character(1)`)\cr
    #'   One of `"random"`, `"first"` or `"last"` (c.f. [max.col()]) to determine how to deal with
    #'   tied probabilities.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    set_threshold = function(threshold, ties_method = "random") {
      self$data = set_threshold_pdata(self$data, threshold, ties_method)
      invisible(self)
    }
  ),

  active = list(
    #' @field response (`factor()`)\cr
    #' Access to the stored predicted class labels.
    response = function(rhs) {
      assert_ro_binding(rhs)
      self$data$response %??% factor(rep(NA_character_, length(self$data$row_ids)), levels(self$data$truth))
    },

    #' @field prob (`matrix()`)\cr
    #' Access to the stored probabilities.
    prob = function(rhs) {
      assert_ro_binding(rhs)
      self$data$prob
    },

    #' @field confusion (`matrix()`)\cr
    #' Confusion matrix, as resulting from the comparison of truth and response.
    #' Truth is in columns, predicted response is in rows.
    confusion = function(rhs) {
      assert_ro_binding(rhs)
      table(response = self$data$response, truth = self$data$truth, useNA = "ifany")
    }
  )
)

#' @export
# nolint next
as.data.table.PredictionClassif = function(x, ...) {
  tab = as.data.table(x$data[c("row_ids", "truth", "response")])

  if ("prob" %chin% x$predict_types) {
    prob = as.data.table(x$data$prob)
    setnames(prob, names(prob), paste0("prob.", names(prob)))
    tab = rcbind(tab, prob)
  }

  if (!is.null(x$data$extra)) {
    tab = cbind(tab, as.data.table(x$data$extra))
  }

  if (!is.null(x$data$weights)) {
    tab$weights = x$data$weights
  }

  tab[]
}

set_threshold_pdata = function(pdata, threshold, ties_method) {
  if (!is.matrix(pdata$prob)) {
    error_input("Cannot set threshold, no probabilities available")
  }
  lvls = levels(pdata$truth)

  if (length(threshold) == 1L) {
    assert_number(threshold, lower = 0, upper = 1)
    if (length(lvls) != 2L) {
      error_input("Setting a single threshold only supported for binary classification problems")
    }
    prob = cbind(pdata$prob[, 1L], threshold)
  } else {
    assert_numeric(threshold, any.missing = FALSE, lower = 0, upper = 1, len = length(lvls))
    assert_names(names(threshold), permutation.of = lvls)
    threshold = threshold[lvls] # reorder thresh so it is in the same order as levels

    # multiply all rows by threshold, then get index of max element per row
    prob = pdata$prob %*% diag(1 / threshold) # can generate Inf for threshold 0
    prob[is.na(prob)] = 0 # NaN results from 0 * Inf, replace with 0, c.f. #452
  }

  ind = max.col(prob, ties.method = ties_method)
  pdata$response = factor(lvls[ind], levels = lvls)
  pdata
}

#' @title Classification Task
#'
#' @include TaskSupervised.R
#'
#' @description
#' This task specializes [Task] and [TaskSupervised] for classification problems.
#' The target column is assumed to be a factor or ordered factor.
#' The `task_type` is set to `"classif"`.
#'
#' Additional task properties include:
#' * `"twoclass"`: The task is a binary classification problem.
#' * `"multiclass"`: The task is a multiclass classification problem.
#'
#' It is recommended to use [as_task_classif()] for construction.
#' Predefined tasks are stored in the [dictionary][mlr3misc::Dictionary] [mlr_tasks].
#'
#' @template param_id
#' @template param_backend
#' @template param_rows
#' @template param_cols
#'
#' @template seealso_task
#' @export
#' @examples
#' data("Sonar", package = "mlbench")
#' task = as_task_classif(Sonar, target = "Class", positive = "M")
#'
#' task$task_type
#' task$formula()
#' task$truth()
#' task$class_names
#' task$positive
#' task$data(rows = 1:3, cols = task$feature_names[1:2])
TaskClassif = R6Class(
  "TaskClassif",
  inherit = TaskSupervised,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #' The function [as_task_classif()] provides an alternative way to construct classification tasks.
    #'
    #' @template param_target
    #'
    #' @param positive (`character(1)`)\cr
    #'   Only for binary classification: Name of the positive class.
    #'   The levels of the target columns are reordered accordingly, so that the first element of `$class_names` is the
    #'   positive class, and the second element is the negative class.
    #' @template param_label
    #' @template param_extra_args
    initialize = function(id, backend, target, positive = NULL, label = NA_character_, extra_args = list()) {
      assert_string(target)
      super$initialize(
        id = id,
        task_type = "classif",
        backend = backend,
        target = target,
        label = label,
        extra_args = extra_args
      )

      update_classif_property(self, private)

      if (!is.null(positive)) {
        # NB: this also sets `extra_args$positive`
        self$positive = positive
      }
    },

    #' @description
    #' Printer.
    #' @param ... (ignored).
    print = function(...) {
      super$print(...)

      if (!is.null(private$.backend) && self$nrow <= getOption("mlr3.print_class_ratio_threshold", 1000000L)) {
        class_freqs = table(self$truth()) / self$nrow * 100
        class_freqs = class_freqs[order(-class_freqs, names(class_freqs))]
        classes = if ("twoclass" %in% self$properties) {
          sprintf(
            "%s (positive class, %.0f%%), %s (%.0f%%)",
            self$positive,
            class_freqs[[self$positive]],
            self$negative,
            class_freqs[[self$negative]]
          )
        } else {
          if (length(class_freqs) > 10L) {
            paste0(
              toString(sprintf("%s (%.0f%%)", names(class_freqs)[1:10], class_freqs[1:10])),
              " + ",
              length(class_freqs) - 10,
              " more"
            )
          } else {
            toString(sprintf("%s (%.0f%%)", names(class_freqs), class_freqs))
          }
        }
      } else {
        classes = toString(self$class_names)
      }
      cat_cli(cli_li("Target classes: {classes}"))
    },

    #' @description
    #' True response for specified `row_ids`. Format depends on the task type.
    #' Defaults to all rows with role `"use"`.
    #' @return `factor()`.
    truth = function(rows = NULL) {
      super$truth(rows)[[1L]]
    },

    #' @description
    #' Updates the cache of stored factor levels, removing all levels not present in the current set of active rows.
    #' `cols` defaults to all columns with storage type "factor" or "ordered".
    #' Also updates the task property `"twoclass"`/`"multiclass"`.
    #'
    #' @return Modified `self`.
    droplevels = function(cols = NULL) {
      super$droplevels(cols)
      update_classif_property(self, private)
      invisible(self)
    }
  ),

  active = list(
    #' @field class_names (`character()`)\cr
    #' Returns all class labels of the target column.
    class_names = function(rhs) {
      assert_ro_binding(rhs)
      fget_key(self$col_info, i = self$target_names, j = "levels", key = "id")[[1L]]
    },

    #' @field positive (`character(1)`)\cr
    #' Stores the positive class for binary classification tasks, and `NA` for multiclass tasks.
    #' To switch the positive class, assign a level to this field.
    positive = function(rhs) {
      lvls = self$class_names
      if (missing(rhs)) {
        if (length(lvls) != 2L) {
          return(NA_character_)
        }
        return(lvls[1L])
      }

      if (length(lvls) != 2L) {
        error_input("Setting the positive class is only feasible for binary classification")
      }
      positive = assert_choice(rhs, lvls)
      negative = setdiff(lvls, rhs)
      self$extra_args$positive = positive
      lvls = c(positive, negative)
      ii = self$col_info[list(self$target_names), on = "id", which = TRUE]
      set(self$col_info, i = ii, j = "levels", value = list(lvls))
      set(self$col_info, i = ii, j = "fix_factor_levels", value = TRUE)

      lvls
    },

    #' @field negative (`character(1)`)\cr
    #' Stores the negative class for binary classification tasks, and `NA` for multiclass tasks.
    negative = function(rhs) {
      assert_ro_binding(rhs)
      lvls = self$class_names
      if (length(lvls) != 2L) {
        return(NA_character_)
      }
      lvls[2L]
    }
  )
)

update_classif_property = function(self, private) {
  tn = self$target_names
  if (fget_key(self$col_info, tn, "type", key = "id") %nin% c("factor", "ordered")) {
    error_input("Target column '%s' must be a factor or ordered factor", tn)
  }

  nlvls = length(self$class_names)
  if (nlvls < 2L) {
    error_input("Target column '%s' must have at least two levels", tn)
  }

  private$.properties = setdiff(private$.properties, c("twoclass", "multiclass"))
  private$.properties = union(private$.properties, if (nlvls == 2L) "twoclass" else "multiclass")
}

#' @title Get the Default Measure
#'
#' @description
#' Gets the default measures using the information in [mlr_reflections$default_measures][mlr_reflections]:
#' * [`"classif.ce"`][mlr_measures_classif.ce] for classification (`"classif"`).
#' * [`"regr.mse"`][mlr_measures_regr.mse] for regression (`"regr"`).
#' * Add-on package may register additional default measures for their own task types.
#'
#' @param task_type (`character(1)`)\cr
#'   Get the default measure for the task type `task_type`, e.g., `"classif"` or `"regr"`.
#'   If `task_type` is `NULL`, an empty list is returned.
#'
#' @return list of [Measure].
#'
#' @export
#' @examples
#' default_measures("classif")
#' default_measures("regr")
default_measures = function(task_type) {
  if (is.null(task_type)) {
    return(list())
  }
  assert_choice(task_type, names(mlr_reflections$default_measures))
  keys = mlr_reflections$default_measures[[task_type]]
  mlr_measures$mget(keys)
}

#' @title Convert to a Task
#'
#' @description
#' Convert object to a [Task] or a list of [Task].
#'
#' The function supports:
#' * Converting existing [Task] objects (with optional cloning)
#' * Converting objects from other packages (e.g., `OMLTask` from \CRANpkg{mlr3oml})
#' * Converting lists of objects to lists of tasks
#'
#' For constructing tasks from data frames, use the dedicated converters:
#' * [as_task_classif()] for classification tasks
#' * [as_task_regr()] for regression tasks
#' * [as_task_unsupervised()] for unsupervised tasks
#'
#' @param x (any)\cr
#'   Object to convert.
#' @param ... (any)\cr
#'   Additional arguments.
#' @export
as_task = function(x, ...) {
  UseMethod("as_task")
}

#' @export
as_task.default = function(x, ...) {
  error_input(
    # nolint next
    "No method for class '%s'. To create a task from a `data.frame`, use dedicated converters such as `as_task_classif()` or `as_task_regr()`.",
    class(x)[1L]
  )
}

#' @rdname as_task
#' @param clone (`logical(1)`)\cr
#'   If `TRUE`, ensures that the returned object is not the same as the input `x`.
#' @export
# nolint next
as_task.Task = function(x, clone = FALSE, ...) {
  assert_empty_ellipsis(...)
  if (isTRUE(clone)) x$clone(deep = TRUE) else x
}

#' @rdname as_task
#' @export
as_tasks = function(x, ...) {
  UseMethod("as_tasks")
}

#' @rdname as_task
#' @export
# nolint next
as_tasks.default = function(x, ...) {
  list(as_task(x, ...))
}

#' @rdname as_task
#' @export
# nolint next
as_tasks.list = function(x, ...) {
  lapply(x, as_task, ...)
}

test_that("as_task conversion", {
  task = tsk("zoo")
  converted = as_task(task)
  cloned = as_task(task, clone = TRUE)

  expect_class(converted, "Task")
  expect_same_address(task, converted)
  expect_different_address(task, cloned)

  expect_list(as_tasks(task), types = "Task")
  expect_list(as_tasks(list(task)), types = "Task")
})

test_that("as_task_xx error messages (#944)", {
  expect_error(
    as_task_classif(data.frame(
      x = factor(c("a", "b", "a", "b"), levels = c("a", "b")),
      y = factor(c("a", "b", "a", "b"), levels = c("a", "b"))
    )),
    "argument \"target\" is missing, with no default"
  )

  expect_error(
    as_task_regr(data.frame(
      x = factor(c("a", "b", "a", "b"), levels = c("a", "b")),
      y = factor(c("a", "b", "a", "b"), levels = c("a", "b"))
    )),
    "argument \"target\" is missing, with no default"
  )
})

test_that("error when arguments are misspelled", {
  expect_error(as_task(tsk("iris"), clone2 = TRUE), "Received the following")
})

hashes = function(x) {
  map_chr(unname(x), "hash")
}

phashes = function(x) {
  map_chr(unname(x), "phash")
}

#' @description
#' Calculate task hashes of resampling iterations.
#'
#' @param task ([Task]).
#' @param resampling ([Resampling]).
#' @param learner ([Learner])\cr
#' Because we only set the validation task for learners that use it, a learner can be passed here to
#' calculate the hashes accordingly, i.e. ignore the test ids when the learner does not have the
#' 'validation' property but otherwise include them.
#'
#' @return (`character()`).
#' @noRd
resampling_task_hashes = function(task, resampling, learner = NULL) {
  # validation task is set on the worker
  learner_does_validation = !is.null(get0("validate", learner))
  map_chr(seq_len(resampling$iters), function(i) {
    train_set = resampling$train_set(i)
    test_set = if (learner_does_validation) resampling$test_set(i)
    task_hash(task, train_set, test_set)
  })
}

task_hash = function(task, use_ids, test_ids = NULL, ignore_internal_valid_task = FALSE) {
  # order matters: we first check for test_ids and then for the internal_valid_task
  internal_valid_task_hash = if (!is.null(test_ids)) {
    # this does the same as
    # task$internal_valid_task = test_ids
    # $internal_valid_task$hash
    # but avoids the deep clone
    task_hash(task, use_ids = test_ids, test_ids = NULL, ignore_internal_valid_task = TRUE)
  } else if (!ignore_internal_valid_task) {
    task$internal_valid_task$hash
  }

  calculate_hash(
    class(task),
    task$id,
    task$backend$hash,
    task$col_info,
    use_ids,
    task$col_roles,
    get_private(task)$.properties,
    internal_valid_task_hash,
    task$characteristics
  )
}

#' @title Debug Measure for Classification
#'
#' @name mlr_measures_debug_classif
#' @include Measure.R
#'
#' @description
#' This measure returns the number of observations in the [PredictionClassif] object.
#' Its main purpose is debugging.
#' The parameter `na_ratio` (`numeric(1)`) controls the ratio of scores which randomly
#' are set to `NA`, between 0 (default) and 1.
#'
#' @templateVar id debug_classif
#' @template measure
#'
#' @template seealso_measure
#' @export
#' @examples
#' task = tsk("wine")
#' learner = lrn("classif.featureless")
#' measure = msr("debug_classif", na_ratio = 0.5)
#' rr = resample(task, learner, rsmp("cv", folds = 5))
#' rr$score(measure)
MeasureDebugClassif = R6Class(
  "MeasureDebugClassif",
  inherit = Measure,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      param_set = ps(na_ratio = p_dbl(0, 1, tags = "required"))
      param_set$set_values(na_ratio = 0)
      super$initialize(
        id = "debug_classif",
        param_set = param_set,
        predict_type = "response",
        range = c(0, Inf),
        properties = "na_score",
        label = "Debug Classification Measure",
        man = "mlr3::mlr_measures_debug_classif"
      )
    }
  ),

  private = list(
    .score = function(prediction, ...) {
      na_ratio = self$param_set$get_values()$na_ratio
      if (na_ratio > runif(1L)) {
        return(NA_integer_)
      }
      length(prediction$row_ids)
    }
  )
)

#' @include mlr_measures.R
mlr_measures$add("debug_classif", function() MeasureDebugClassif$new())

#' @title Manually Partition into Training, Test and Validation Set
#'
#' @description
#' Creates a split of the row ids of a [Task] into a training and a test set, and optionally a validation set.
#'
#' @param task ([Task])\cr
#'   Task to operate on.
#' @param ratio (`numeric()`)\cr
#'   Ratio of observations to put into the training set.
#'   If a 2 element vector is provided, the first element is the ratio for the training set,
#'   the second element is the ratio for the test set.
#'   The validation set will contain the remaining observations.
#' @export
#' @examples
#' # regression task partitioned into training and test set
#' task = tsk("california_housing")
#' split = partition(task, ratio = 0.5)
#' data = data.frame(
#'   y = c(task$truth(split$train), task$truth(split$test)),
#'   split = rep(c("train", "predict"), lengths(split[c("train", "test")]))
#' )
#' boxplot(y ~ split, data = data)
#'
#' # classification task partitioned into training, test and validation set
#' task = tsk("pima")
#' split = partition(task, c(0.66, 0.14))
partition = function(task, ratio = 0.67) {
  assert_numeric(ratio, min.len = 1L, max.len = 2L)
  UseMethod("partition")
}

#' @rdname partition
#' @export
partition.Task = function(task, ratio = 0.67) {
  task = task$clone(deep = TRUE)
  if (sum(ratio) >= 1) {
    error_input("Sum of 'ratio' must be smaller than 1")
  }

  if (length(ratio) == 1L) {
    ratio[2L] = 1 - ratio
  } else {
    ratio[3L] = 1 - (ratio[1L] + ratio[2L])
  }

  r1 = rsmp("holdout", ratio = ratio[1L])$instantiate(task)
  task$row_roles$use = r1$test_set(1L)
  r2 = rsmp("holdout", ratio = ratio[2L] / (1 - ratio[1L]))$instantiate(task)

  list(
    train = r1$train_set(1L),
    test = r2$train_set(1L),
    validation = r2$test_set(1L)
  )
}

#' @rdname PredictionData
#' @export
# nolint next
check_prediction_data.PredictionDataRegr = function(pdata, ...) {
  pdata$row_ids = assert_row_ids(pdata$row_ids)
  n = length(pdata$row_ids)
  if (is.null(pdata$truth)) {
    pdata$truth = NA_real_
  }
  if (!length(pdata$row_ids)) {
    pdata$truth = numeric()
  }

  if (!is.null(pdata$response)) {
    pdata$response = assert_numeric(unname(pdata$response))
    assert_prediction_count(length(pdata$response), n, "response")
  }

  if (!is.null(pdata$se)) {
    pdata$se = assert_numeric(unname(pdata$se), lower = 0)
    assert_prediction_count(length(pdata$se), n, "se")
  }

  if (!is.null(pdata$quantiles)) {
    quantiles = pdata$quantiles
    assert_matrix(quantiles)
    assert_prediction_count(nrow(quantiles), n, "quantiles")

    if (is.null(attr(quantiles, "probs"))) {
      error_learner_predict("No probs attribute stored in 'quantile'")
    }

    if (is.null(attr(quantiles, "response")) && is.null(pdata$response)) {
      error_learner_predict("No response attribute stored in 'quantile' or response stored in 'pdata'")
    }

    colnames(pdata$quantiles) = sprintf("q%g", attr(quantiles, "probs"))
    setattr(pdata$quantiles, "response", sprintf("q%g", attr(quantiles, "response")))
  }

  if (!is.null(pdata$distr)) {
    assert_class(pdata$distr, "VectorDistribution")

    if (is.null(pdata$response)) {
      pdata$response = unname(pdata$distr$mean())
    }

    if (is.null(pdata$se)) {
      pdata$se = unname(pdata$distr$stdev())
    }
  }

  if (!is.null(pdata$weights)) {
    # weights may never be NA, even if no prediction was made.
    pdata$weights = assert_numeric(unname(pdata$weights), any.missing = FALSE)
    assert_prediction_count(length(pdata$weights), n, "weights")
  }

  if (!is.null(pdata$extra)) {
    assert_list(pdata$extra, names = "unique")
    len = lengths(pdata$extra)
    if (any(len != n)) {
      error_learner_predict("Extra data must have the same length as the number of predictions")
    }
  }

  pdata
}


#' @rdname PredictionData
#' @export
# nolint next
is_missing_prediction_data.PredictionDataRegr = function(pdata, ...) {
  miss = logical(length(pdata$row_ids))

  if (!is.null(pdata$response)) {
    miss = is.na(pdata$response)
  }

  if (!is.null(pdata$se)) {
    miss = miss | is.na(pdata$se)
  }

  if (!is.null(pdata$quantiles)) {
    miss = miss | row_any_na(pdata$quantiles)
  }

  # weights may never be NA, so we don't need to check for missingness

  pdata$row_ids[miss]
}


#' @rdname PredictionData
#' @export
# nolint next
c.PredictionDataRegr = function(..., keep_duplicates = TRUE) {
  dots = list(...)
  assert_list(dots, "PredictionDataRegr")
  assert_flag(keep_duplicates)
  if (length(dots) == 1L) {
    return(dots[[1L]])
  }

  predict_types = names(mlr_reflections$learner_predict_types$regr)
  predict_types = map(dots, function(x) intersect(names(x), predict_types))
  if (!every(predict_types[-1L], setequal, y = predict_types[[1L]])) {
    error_input("Cannot combine predictions: Different predict types")
  }

  if (length(unique(map_lgl(dots, function(x) is.null(x$weights)))) > 1L) {
    error_input("Cannot combine predictions: Some predictions have weights, others do not")
  }

  if (length(unique(map_lgl(dots, function(x) is.null(x$extra)))) > 1L) {
    error_input("Cannot rbind predictions: Some predictions have extra data, others do not")
  }

  nn = names(dots[[1L]])
  elems = c("row_ids", "truth", intersect(predict_types[[1L]], c("response", "se")), if ("weights" %chin% nn) "weights")
  tab = map_dtr(dots, function(x) x[elems], .fill = FALSE)

  quantiles = if ("quantiles" %chin% nn) {
    quantiles = map(dots, "quantiles")
    attrs = attributes(quantiles[[1L]])
    quantiles = do.call(rbind, quantiles)
    setattr(quantiles, "probs", attrs$probs)
    setattr(quantiles, "response", attrs$response)
  }
  extra = if ("extra" %chin% nn) {
    rbindlist(map(dots, "extra"), fill = TRUE, use.names = TRUE)
  }

  if (!keep_duplicates) {
    keep = !duplicated(tab, by = "row_ids", fromLast = TRUE)
    tab = tab[keep]
    quantiles = quantiles[keep, , drop = FALSE]
    extra = extra[keep]
  }

  result = as.list(tab)
  result$quantiles = quantiles
  if (!is.null(extra)) {
    result$extra = as.list(extra)
  }

  if ("distr" %chin% predict_types[[1L]]) {
    require_namespaces("distr6", msg = "To predict probability distributions, please install %s")
    result$distr = do.call(c, map(dots, "distr"))
  }

  raw = discard(map(dots, "raw"), is.null)
  if (length(raw)) {
    result$raw = raw
  }

  new_prediction_data(result, "regr")
}

#' @export
filter_prediction_data.PredictionDataRegr = function(pdata, row_ids, ...) {
  keep = pdata$row_ids %in% row_ids
  pdata$row_ids = pdata$row_ids[keep]
  pdata$truth = pdata$truth[keep]

  if (!is.null(pdata$response)) {
    pdata$response = pdata$response[keep]
  }

  if (!is.null(pdata$se)) {
    pdata$se = pdata$se[keep]
  }

  if (!is.null(pdata$quantiles)) {
    pdata$quantiles = pdata$quantiles[keep, , drop = FALSE]
  }

  if (!is.null(pdata$weights)) {
    pdata$weights = pdata$weights[keep]
  }

  if (!is.null(pdata$extra)) {
    pdata$extra = map(pdata$extra, function(x) x[keep])
  }

  pdata
}

#' @export
create_empty_prediction_data.TaskRegr = function(task, learner) {
  predict_types = mlr_reflections$learner_predict_types[["regr"]][[learner$predict_type]]

  pdata = list(
    row_ids = integer(),
    truth = numeric()
  )

  if ("response" %chin% predict_types) {
    pdata$response = pdata$truth
  }

  if ("se" %chin% predict_types) {
    pdata$se = pdata$truth
  }

  if ("distr" %chin% predict_types) {
    pdata$distr = list()
  }

  if ("weights_measure" %chin% task$properties) {
    pdata$weights = numeric()
  }

  new_prediction_data(pdata, "regr")
}

#' @title Task Class
#'
#' @include mlr_reflections.R
#' @include warn_deprecated.R
#'
#' @description
#' This is the abstract base class for [TaskSupervised] and [TaskUnsupervised].
#' [TaskClassif] and [TaskRegr] inherit from [TaskSupervised].
#' More supervised tasks are implemented in \CRANpkg{mlr3proba}, unsupervised cluster tasks
#' in package \CRANpkg{mlr3cluster}.
#'
#' Tasks serve two purposes:
#'
#' 1. Tasks wrap a [DataBackend], an object to transparently interface different data storage types.
#' 2. Tasks store meta-information, such as the role of the individual columns in the [DataBackend].
#'    For example, for a classification task a single column must be marked as target column, and others as features.
#'
#' Predefined (toy) tasks are stored in the [dictionary][mlr3misc::Dictionary] [mlr_tasks],
#' e.g. [`penguins`][mlr_tasks_penguins] or [`california_housing`][mlr_tasks_california_housing].
#' More toy tasks can be found in the dictionary after loading \CRANpkg{mlr3data}.
#'
#' @template param_id
#' @template param_backend
#' @template param_task_type
#' @template param_rows
#' @template param_cols
#' @template param_label
#' @template param_extra_args
#'
#' @section S3 methods:
#' * `as.data.table(t)`\cr
#'   [Task] -> [data.table::data.table()]\cr
#'   Returns the complete data as [data.table::data.table()].
#' * `head(t)`\cr
#'   Calls [head()] on the task's data.
#' * `summary(t)`\cr
#'   Calls [summary()] on the task's data.
#'
#'
#' @section Task mutators:
#' The following methods change the task in-place:
#' * Any modification of the lists `$col_roles` or `$row_roles`.
#'   This provides a different "view" on the data without altering the data itself.
#'   This may affect, e.g., `$data`, `$nrow`, `$ncol`, `n_features`, `row_ids`, and `$feature_names`.
#'   Altering `$col_roles` may affect, e.g., `$data`, `$ncol`, `$n_features`, and `$feature_names`.
#'   Altering `$row_roles` may affect, e.g., `$data`, `$nrow`, and `$row_ids`.
#' * Modification of column or row roles via `$set_col_roles()` or `$set_row_roles()`, respectively.
#'   They are an alternative to directly accessing `$col_roles` or `$row_roles`, with the same side effects.
#' * `$select()` and `$filter()` subset the set of active features or rows in `$col_roles` or `$row_roles`,
#'   respectively.
#' * `$cbind()` and `$rbind()` change the task in-place by binding new columns or rows to the data.
#' * `$rename()` changes column names.
#' * `$set_levels()` and `$droplevels()` update the field `$col_info()` to automatically repair
#'   factor levels while querying data with `$data()`.
#' * `$materialize_view()` creates a new [DataBackendDataTable] which keeps only the data in the currently active view
#'   possibly freeing some memory consumed by the [DataBackend] stored in the `Task`.
#'
#' @template seealso_task
#' @concept Task
#' @export
#' @examples
#' # We use the inherited class TaskClassif here,
#' # because the base class `Task` is not intended for direct use
#' task = TaskClassif$new("penguins", palmerpenguins::penguins, target = "species")
#'
#' task$nrow
#' task$ncol
#' task$feature_names
#' task$formula()
#'
#' # de-select "year"
#' task$select(setdiff(task$feature_names, "year"))
#'
#' task$feature_names
#'
#' # Add new column "foo"
#' task$cbind(data.frame(foo = 1:344))
#' head(task)
Task = R6Class(
  "Task",
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' Note that this object is typically constructed via a derived classes, e.g. [TaskClassif] or [TaskRegr].
    initialize = function(id, task_type, backend, label = NA_character_, extra_args = list()) {
      private$.id = assert_id(id)
      private$.label = assert_string(label, na.ok = TRUE)
      private$.task_type = assert_choice(task_type, mlr_reflections$task_types$type)
      if (!inherits(backend, "DataBackend")) {
        private$.backend = as_data_backend(backend)
      } else {
        private$.backend = assert_backend(backend)
      }

      cn = private$.backend$colnames
      rn = private$.backend$rownames

      assert_names(cn, "unique", .var.name = "column names")
      if (any(grepl("%", cn, fixed = TRUE))) {
        error_input("Column names may not contain special character '%%'")
      }

      private$.col_info = col_info(private$.backend)
      private$.col_info$label = NA_character_
      private$.col_info$fix_factor_levels = FALSE

      assert_subset(private$.col_info$type, mlr_reflections$task_feature_types, .var.name = "feature types")
      pmap(private$.col_info, function(id, levels, ...) {
        assert_character(
          levels,
          any.missing = FALSE,
          min.len = 1L,
          null.ok = TRUE,
          .var.name = sprintf("levels of '%s'", id)
        )
      })

      cn = private$.col_info$id # note: this sorts the columns!
      private$.row_roles = list(use = rn)
      private$.col_roles = named_list(mlr_reflections$task_col_roles[[task_type]], character())
      private$.col_roles$feature = setdiff(cn, private$.backend$primary_key)
      private$.extra_args = assert_list(extra_args, names = "unique")
      private$.man = NA_character_
      private$.mlr3_version = mlr_reflections$package_version
    },

    #' @description
    #' Opens the corresponding help page referenced by field `$man`.
    help = function() {
      open_help(self$man)
    },

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s:%s>", class(self)[1L], self$id)
    },

    #' @description
    #' Printer.
    #' @param ... (ignored).
    print = function(...) {
      msg_h = if (is.null(self$label) || is.na(self$label)) "" else paste0(": ", self$label)
      cat_cli(cli_h1("{.cls {class(self)[1L]}} ({self$nrow}x{self$ncol}){msg_h}"))

      roles = private$.col_roles
      roles = roles[lengths(roles) > 0L]

      # print additional columns as specified in reflections
      before = mlr_reflections$task_print_col_roles$before
      iwalk(before[before %chin% names(roles)], function(role, str) {
        cat_cli(cli_li("{str}: {roles[[role]]}"))
      })

      cat_cli(cli_li("Target: {self$target_names}"))

      properties = if (length(self$properties)) toString(self$properties) else "-"
      cat_cli(cli_li("Properties: {properties}"))

      types = self$feature_types

      if (nrow(types)) {
        cat_cli({
          id = type = NULL
          cli_li("Features ({nrow(types)}):")
          types = types[, list(N = .N, feats = str_collapse(id, n = 100L)), by = "type"][,
            "type" := translate_types(type)
          ]
          setorderv(types, "N", order = -1L)

          ulid = cli_ul()
          pmap(types, function(type, N, feats) {
            cli_li("{type} ({N}): {feats}")
          })
          cli_end(ulid)
        })
      }

      # print additional columns as specified in reflections
      after = mlr_reflections$task_print_col_roles$after
      iwalk(after[after %chin% names(roles)], function(role, str) {
        cat_cli(cli_li("{str}: {roles[[role]]}"))
      })

      if (!is.null(private$.internal_valid_task)) {
        cat_cli(cli_li("Validation Task: ({private$.internal_valid_task$nrow}x{private$.internal_valid_task$ncol})"))
      }

      if (!is.null(self$characteristics)) {
        cat_cli(cli_li("Characteristics: {as_short_string(self$characteristics)}"))
      }
    },

    #' @description
    #' Returns a slice of the data from the [DataBackend] as a `data.table`.
    #' Rows default to observations with role `"use"`,
    #' and columns default to features with roles `"target"` or `"feature"`.
    #' Rows must be a subset of `$row_ids`.
    #' If `rows` or `cols` are specified which do not exist in the [DataBackend], an exception is raised.
    #'
    #' Rows and columns are returned in the order specified via the arguments `rows` and `cols`.
    #' If `rows` is `NULL`, rows are returned in the order of `task$row_ids`.
    #' If `cols` is `NULL`, the column order defaults to `c(task$target_names, task$feature_names)`.
    #' Note that it is recommended to **not** rely on the order of columns,
    #' and instead always address columns with their respective column name.
    #'
    #' @param ordered (`logical(1)`)\cr
    #'   If `TRUE`, data is ordered according to the columns with column role `"order"`.
    #'
    #' @return Depending on the [DataBackend], but usually a [data.table::data.table()].
    #' @examples
    #' task = tsk("penguins")
    #' task$data(rows = 1:5, cols = c("species", "sex"))
    data = function(rows = NULL, cols = NULL, ordered = FALSE) {
      assert_has_backend(self)
      assert_flag(ordered)

      row_roles = private$.row_roles
      col_roles = private$.col_roles

      if (is.null(rows)) {
        rows = row_roles$use
      } else {
        assert_subset(rows, self$row_roles$use)
        if (is.double(rows)) {
          rows = as.integer(rows)
        }
      }

      if (is.null(cols)) {
        query_cols = cols = c(col_roles$target, col_roles$feature)
      } else {
        assert_subset(cols, private$.col_info$id)
        query_cols = cols
      }

      reorder_rows = length(col_roles$order) > 0L && ordered
      if (reorder_rows) {
        query_cols = union(query_cols, col_roles$order)
      }

      data = private$.backend$data(rows = rows, cols = query_cols)

      if (length(query_cols) && nrow(data) != length(rows)) {
        error_mlr3(
          "DataBackend did not return the queried rows correctly: %i requested, %i received.
        The resampling was probably instantiated on a different task.",
          length(rows),
          nrow(data)
        ) # TODO: more specific error necessary?
      }

      if (length(rows) && ncol(data) != length(query_cols)) {
        error_mlr3(
          "DataBackend did not return the queried cols correctly: %i requested, %i received",
          length(cols),
          ncol(data)
        ) # TODO: more specific error necessary?
      }

      # nolint next
      .__i__ = private$.col_info[["fix_factor_levels"]]
      if (any(.__i__)) {
        fix_factors = private$.col_info[.__i__, c("id", "levels"), with = FALSE]
        if (nrow(fix_factors)) {
          # ordering is slow
          if (nrow(fix_factors) > 1L) {
            fix_factors = fix_factors[list(names(data)), on = "id", nomatch = NULL]
          }
          data = fix_factor_levels(data, levels = set_names(fix_factors$levels, fix_factors$id))
        }
      }

      if (reorder_rows) {
        setorderv(data, col_roles$order)[]
        data = remove_named(data, setdiff(col_roles$order, cols))
      }

      return(data)
    },

    #' @description
    #' Constructs a [formula()], e.g. `[target] ~ [feature_1] + [feature_2] + ... + [feature_k]`,
    #' using the features provided in argument `rhs`
    #' (defaults to all columns with role `"feature"`, symbolized by `"."`).
    #'
    #' Note that it is currently not possible to change the formula.
    #' However, \CRANpkg{mlr3pipelines} provides a pipe operator interfacing [stats::model.matrix()] for this
    #' purpose: `"modelmatrix"`.
    #'
    #' @param rhs (`character(1)`)\cr
    #'   Right hand side of the formula. Defaults to `"."` (all features of the task).
    #' @return [formula()].
    #' @examples
    #' task = tsk("penguins")
    #' task$formula()
    formula = function(rhs = ".") {
      formulate(self$target_names, rhs)
    },

    #' @description
    #' Get the first `n` observations with role `"use"` of all columns with role `"target"` or `"feature"`.
    #'
    #' @param n (`integer(1)`).
    #' @return [data.table::data.table()] with `n` rows.
    #' @examples
    #' task = tsk("penguins")
    #' task$head(3)
    head = function(n = 6L) {
      assert_number(n, na.ok = FALSE)
      ids = head(private$.row_roles$use, n)
      self$data(rows = ids)
    },

    #' @description
    #' Returns the distinct values for columns referenced in `cols` with storage type "factor" or "ordered".
    #' Argument `cols` defaults to all such columns with role `"target"` or `"feature"`.
    #'
    #' Note that this function ignores the row roles, it returns all levels available in the [DataBackend].
    #' To update the stored level information, e.g. after subsetting a task with `$filter()`, call `$droplevels()`.
    #'
    #' @return named `list()`.
    #' @examples
    #' task = tsk("penguins")
    #' task$levels()
    levels = function(cols = NULL) {
      if (is.null(cols)) {
        cols = unlist(private$.col_roles[c("target", "feature")], use.names = FALSE)
        cols = private$.col_info[
          get("id") %chin% cols & get("type") %chin% c("factor", "ordered"),
          "id",
          with = FALSE
        ][[1L]]
      } else {
        assert_subset(cols, private$.col_info$id)
      }

      set_names(
        fget_keys(private$.col_info, cols, "levels", "id"),
        cols
      )
    },

    #' @description
    #' Returns the number of missing observations for columns referenced in `cols`.
    #' Considers only active rows with row role `"use"`.
    #' Argument `cols` defaults to all columns with role "target" or "feature".
    #'
    #' @return Named `integer()`.
    #' @examples
    #' task = tsk("penguins")
    #' task$missings()
    missings = function(cols = NULL) {
      assert_has_backend(self)

      if (is.null(cols)) {
        cols = unlist(private$.col_roles[c("target", "feature")], use.names = FALSE)
      } else {
        assert_subset(cols, private$.col_info$id)
      }

      private$.backend$missings(self$row_ids, cols = cols)
    },

    #' @description
    #' Subsets the task, keeping only the rows specified via row ids `rows`.
    #'
    #' This operation mutates the task in-place.
    #' See the section on task mutators for more information.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' task = tsk("penguins")
    #' task$filter(1:10)
    #' task$nrow
    filter = function(rows) {
      assert_has_backend(self)
      rows = assert_row_ids(rows)
      private$.row_roles$use = assert_subset(rows, self$row_ids_backend)
      private$.row_hash = NULL
      private$.hash = NULL
      invisible(self)
    },

    #' @description
    #' Subsets the task, keeping only the features specified via column names `cols`.
    #' Note that you cannot deselect the target column, for obvious reasons.
    #'
    #' This operation mutates the task in-place.
    #' See the section on task mutators for more information.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' task = tsk("penguins")
    #' task$select(c("bill_length", "bill_depth"))
    #' task$feature_names
    select = function(cols) {
      assert_has_backend(self)
      assert_character(cols)
      assert_subset(cols, private$.col_roles$feature)
      private$.hash = NULL
      private$.col_hashes = NULL
      private$.col_roles$feature = intersect(private$.col_roles$feature, cols)
      invisible(self)
    },

    #' @description
    #' Adds additional rows to the [DataBackend] stored in `$backend`.
    #' New row ids are automatically created, unless `data` has a column whose name matches
    #' the primary key of the [DataBackend] (`task$backend$primary_key`).
    #' In case of name clashes of row ids, rows in `data` have higher precedence
    #' and virtually overwrite the rows in the [DataBackend].
    #'
    #' All columns roles `"target"`, `"feature"`, `"weights_learner"`, `"weights_measure"`,
    #' `"group"`, `"stratum"`, and `"order"` must be present in `data`.
    #' Columns only present in `data` but not in the [DataBackend] of `task` will be discarded.
    #'
    #' This operation mutates the task in-place.
    #' See the section on task mutators for more information.
    #'
    #' @param data (`data.frame()`).
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' task = tsk("penguins")
    #' extra = task$data(rows = 1:2)
    #' task$rbind(extra)
    rbind = function(data) {
      assert_has_backend(self)

      pk = private$.backend$primary_key
      rn = private$.backend$rownames
      pk_in_backend = TRUE
      type_check = TRUE

      if (is.data.frame(data)) {
        pk_in_backend = pk %chin% names(data)
        type_check = FALSE # done by auto-converter

        keep_cols = intersect(names(data), private$.col_info$id)
        if (length(keep_cols) == pk_in_backend || nrow(data) == 0L) {
          return(invisible(self))
        }

        if (!pk_in_backend) {
          start = if (length(rn)) max(rn) + 1L else 1L
          pk = seq(from = start, to = start + nrow(data) - 1L)
        }

        ci = private$.col_info[list(keep_cols), on = "id"]
        data = do.call(
          data.table,
          Map(auto_convert, value = as.list(data)[ci$id], id = ci$id, type = ci$type, levels = ci$levels)
        )

        data = as_data_backend(data, primary_key = pk)
      } else {
        assert_backend(data)
        if (data$ncol <= 1L || data$nrow == 0L) {
          return(invisible(self))
        }
      }

      if (pk_in_backend && any(data$rownames %in% private$.backend$rownames)) {
        error_input("Cannot rbind data to task '%s', duplicated row ids", self$id)
      }

      # columns with these roles must be present in data
      mandatory_roles = c(
        "target",
        "feature",
        "group",
        "stratum",
        "order",
        "offset",
        "weights_learner",
        "weights_measure"
      )
      mandatory_cols = unlist(private$.col_roles[mandatory_roles], use.names = FALSE)
      missing_cols = setdiff(mandatory_cols, data$colnames)
      if (length(missing_cols)) {
        error_input(
          "Cannot rbind data to task '%s', missing the following mandatory columns: %s",
          self$id,
          str_collapse(missing_cols)
        )
      }

      # merge col infos
      tab = merge(
        private$.col_info,
        col_info(data),
        by = "id",
        all.x = TRUE,
        all.y = FALSE,
        suffixes = c("", "_y"),
        sort = TRUE
      )

      # type check
      if (type_check) {
        type = type_y = NULL
        ii = head(tab[type != type_y, which = TRUE], 1L)
        if (length(ii)) {
          error_input(
            "Cannot rbind to task: Types do not match for column: %s (%s != %s)",
            tab$id[ii],
            tab$type[ii],
            tab$type_y[ii]
          )
        }
      }

      # merge factor levels
      ii = tab[type %chin% c("factor", "ordered"), which = TRUE]
      for (i in ii) {
        x = tab[["levels"]][[i]]
        y = tab[["levels_y"]][[i]]
        if (any(y %nin% x)) {
          set(tab, i = i, j = "levels", value = list(union(x, y)))
          set(tab, i = i, j = "fix_factor_levels", value = TRUE)
        }
      }
      tab[, c("type_y", "levels_y") := list(NULL, NULL)]

      # everything looks good, modify task
      private$.hash = NULL
      private$.backend = DataBackendRbind$new(private$.backend, data)
      private$.col_info = tab[]
      private$.row_roles$use = c(private$.row_roles$use, data$rownames)

      invisible(self)
    },

    #' @description
    #'
    #' Adds additional columns to the [DataBackend] stored in `$backend`.
    #'
    #' The row ids must be provided as column in `data`
    #' (with column name matching the primary key name of the [DataBackend]).
    #' If this column is missing, it is assumed that the rows are exactly in the order of `$row_ids`.
    #' In case of name clashes of column names in `data` and [DataBackend], columns in `data` have higher precedence
    #' and virtually overwrite the columns in the [DataBackend].
    #'
    #' This operation mutates the task in-place.
    #' See the section on task mutators for more information.
    #' @param data (`data.frame()`).
    #' @examples
    #' task = tsk("penguins")
    #' task$cbind(data.table(extra_col = seq_len(task$nrow)))
    #' head(task$data(cols = "extra_col"))
    cbind = function(data) {
      assert_has_backend(self)
      pk = private$.backend$primary_key

      if (is.data.frame(data)) {
        # binding data with 0 rows is explicitly allowed
        if (ncol(data) == 0L) {
          return(invisible(self))
        }

        if (pk %nin% names(data)) {
          data[[pk]] = self$row_ids
        }

        data = as_data_backend(data, primary_key = pk)
      } else {
        assert_backend(data)
        if (data$ncol <= 1L) {
          return(invisible(self))
        }
        assert_set_equal(self$row_ids, data$rownames)
      }

      # update col_info for existing columns
      ci = col_info(data)
      private$.col_info = ujoin(private$.col_info, ci, key = "id")

      # add rows to col_info for new columns
      private$.col_info = rbindlist(
        list(
          private$.col_info,
          insert_named(ci[!list(private$.col_info), on = "id"], list(label = NA_character_, fix_factor_levels = FALSE))
        ),
        use.names = TRUE
      )
      setkeyv(private$.col_info, "id")

      # add new features
      private$.hash = NULL
      private$.col_hashes = NULL
      col_roles = private$.col_roles
      private$.col_roles$feature = union(col_roles$feature, setdiff(data$colnames, c(pk, col_roles$target)))

      # update backend
      private$.backend = DataBackendCbind$new(private$.backend, data)

      invisible(self)
    },

    #' @description
    #' Renames columns by mapping column names in `old` to new column names in `new` (element-wise).
    #'
    #' This operation mutates the task in-place.
    #' See the section on task mutators for more information.
    #'
    #' @param old (`character()`)\cr
    #'   Old names.
    #'
    #' @param new (`character()`)\cr
    #'   New names.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' task = tsk("penguins")
    #' task$rename("body_mass", "mass")
    #' task$feature_names
    rename = function(old, new) {
      assert_has_backend(self)
      private$.hash = NULL
      private$.col_hashes = NULL
      private$.backend = DataBackendRename$new(private$.backend, old, new)
      setkeyv(private$.col_info[old, ("id") := new, on = "id"], "id")
      private$.col_roles = map(private$.col_roles, map_values, old = old, new = new)
      invisible(self)
    },

    #' @description
    #' Modifies the roles in `$row_roles` **in-place**.
    #'
    #' @param rows (`integer()`)\cr
    #'   Row ids for which to change the roles for.
    #' @param roles (`character()`)\cr
    #'   Exclusively set rows to the specified `roles` (remove from other roles).
    #' @param add_to (`character()`)\cr
    #'   Add rows with row ids `rows` to roles specified in `add_to`.
    #'   Rows keep their previous roles.
    #' @param remove_from (`character()`)\cr
    #'   Remove rows with row ids `rows` from roles specified in `remove_from`.
    #'   Other row roles are preserved.
    #'
    #' @details
    #' Roles are first set exclusively (argument `roles`), then added (argument `add_to`) and finally
    #' removed (argument `remove_from`) from different roles.
    #' Duplicated row ids are explicitly allowed, so you can replicate an observation by repeating its
    #' `row_id`.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' task = tsk("penguins")
    #' task$set_row_roles(1:5, remove_from = "use")
    set_row_roles = function(rows, roles = NULL, add_to = NULL, remove_from = NULL) {
      assert_has_backend(self)
      assert_subset(rows, private$.backend$rownames)

      private$.row_hash = NULL
      private$.hash = NULL
      private$.row_roles = task_set_roles(private$.row_roles, rows, roles, add_to, remove_from, allow_duplicated = TRUE)

      invisible(self)
    },

    #' @description
    #' Modifies the roles in `$col_roles` **in-place**.
    #' See `$col_roles` for a list of possible roles.
    #'
    #' @param cols (`character()`)\cr
    #'   Column names for which to change the roles for.
    #' @param roles (`character()`)\cr
    #'   Exclusively set columns to the specified `roles` (remove from other roles).
    #' @param add_to (`character()`)\cr
    #'   Add columns with column names `cols` to roles specified in `add_to`.
    #'   Columns keep their previous roles.
    #' @param remove_from (`character()`)\cr
    #'   Remove columns with columns names `cols` from roles specified in `remove_from`.
    #'   Other column roles are preserved.
    #'
    #' @details
    #' Roles are first set exclusively (argument `roles`), then added (argument `add_to`) and finally removed
    #' (argument `remove_from`) from different roles.
    #' Duplicated columns are removed from the same role.
    #' For tasks that only allow one target, the target column cannot be set with `$set_col_roles()`.
    #' Use the `$col_roles` field to swap the target column.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' task = tsk("penguins")
    #' task$set_col_roles("sex", roles = "stratum")
    #' task$col_roles$stratum
    set_col_roles = function(cols, roles = NULL, add_to = NULL, remove_from = NULL) {
      assert_has_backend(self)
      assert_subset(cols, private$.col_info$id)

      private$.hash = NULL
      private$.col_hashes = NULL
      new_roles = task_set_roles(private$.col_roles, cols, roles, add_to, remove_from)
      private$.col_roles = task_check_col_roles(self, new_roles)

      invisible(self)
    },

    #' @description
    #' Set levels for columns of type `factor` and `ordered` in field `col_info`.
    #' You can add, remove or reorder the levels, affecting the data returned by
    #' `$data()` and `$levels()`.
    #' If you just want to remove unused levels, use `$droplevels()` instead.
    #'
    #' Note that factor levels which are present in the data but not listed in the task as
    #' valid levels are converted to missing values.
    #'
    #' @param levels (named `list()` of `character()`)\cr
    #'   List of character vectors of new levels, named by column names.
    #'
    #' @return Modified `self`.
    #' @examples
    #' task = tsk("penguins")
    #' task$set_levels(list(sex = c("male", "female", "unknown")))
    #' task$levels("sex")
    set_levels = function(levels) {
      assert_list(levels, types = "character", names = "unique", any.missing = FALSE)
      assert_subset(names(levels), private$.col_info$id)

      tab = enframe(lapply(levels, unname), name = "id", value = "levels")
      tab$fix_factor_levels = TRUE

      private$.hash = NULL
      private$.col_info = ujoin(private$.col_info, tab, key = "id")

      invisible(self)
    },

    #' @description
    #' Updates the cache of stored factor levels, removing all levels not present in the current set of active rows.
    #' `cols` defaults to all columns with storage type "factor" or "ordered".
    #' @return Modified `self`.
    #' @examples
    #' task = tsk("penguins")
    #' task$set_levels(list(sex = c("male", "female", "unknown")))
    #' task$levels("sex")
    droplevels = function(cols = NULL) {
      assert_has_backend(self)
      tab = private$.col_info[
        get("type") %chin% c("factor", "ordered"),
        c("id", "levels", "fix_factor_levels"),
        with = FALSE
      ]
      if (!is.null(cols)) {
        tab = tab[list(cols), on = "id", nomatch = NULL]
      }

      # update levels
      # note that we assume that new_levels is a subset of levels!
      new_levels = NULL
      tab$new_levels = private$.backend$distinct(rows = self$row_ids, cols = tab$id)
      tab = tab[lengths(levels) > lengths(new_levels)]
      tab[, c("levels", "fix_factor_levels") := list(Map(intersect, levels, new_levels), TRUE)]

      private$.hash = NULL
      private$.col_info = ujoin(private$.col_info, remove_named(tab, "new_levels"), key = "id")

      invisible(self)
    },

    #' @description
    #' Cuts numeric variables into new factors columns which are added to the task with role
    #' `"stratum"`.
    #' This ensures that all training and test splits contain observations from all bins.
    #' The columns are named `"..stratum_[col_name]"`.
    #'
    #' @param cols (`character()`)\cr
    #'   Names of columns to operate on.
    #' @param bins (`integer()`)\cr
    #'   Number of bins to cut into (passed to [cut()] as `breaks`).
    #'   Replicated to have the same length as `cols`.
    #' @return self (invisibly).
    #' @examples
    #' task = tsk("penguins")
    #' task$add_strata("flipper_length", bins = 4)
    add_strata = function(cols, bins = 3L) {
      assert_names(cols, "unique", subset.of = private$.backend$colnames)
      bins = assert_integerish(bins, any.missing = FALSE, coerce = TRUE)

      col_types = fget_keys(private$.col_info, i = cols, j = "type", key = "id")
      ii = wf(col_types %nin% c("integer", "numeric"))
      if (length(ii)) {
        error_input("For `add_strata`, all columns must be numeric, but '%s' is not", cols[ii])
      }

      strata = pmap_dtc(list(self$data(cols = cols), bins), cut, include.lowest = TRUE)
      setnames(strata, sprintf("..stratum_%s", cols))
      self$cbind(strata)
      self$set_col_roles(names(strata), roles = "stratum")
    },

    #' @description
    #' Certain operations change the view on the data, e.g., `$filter()` or `$select()`.
    #' This operation queries the [DataBackend] for all data required in the active view and
    #' replaces the internal [DataBackend] with the new one. In some scenarios this helps to
    #' free up memory or speeds up accesses to the data, especially after several `$rbind()`
    #' and `$cbind()` operations.
    #'
    #' @details
    #' For tasks containing the same observation more than once (duplicates in `$row_ids`),
    #' the resulting backend contains it only once.
    #'
    #' @param internal_valid_task (`logical(1)`)\cr
    #'   Also materialize the internal validation task. Default is `TRUE`.
    #'
    #' @return self (invisibly).
    #' @examples
    #' task = tsk("iris")
    #' task$backend$nrow
    #' task$filter(1:120)
    #' task$backend$nrow
    materialize_view = function(internal_valid_task = TRUE) {
      assert_flag(internal_valid_task)

      b = private$.backend
      # nolint next
      ..cns = union(b$primary_key, unlist(private$.col_roles, use.names = FALSE))
      dt = b$data(rows = unique(self$row_ids), cols = ..cns)
      private$.backend = as_data_backend(dt, primary_key = b$primary_key)
      private$.col_info = setkeyv(private$.col_info[list(..cns), on = "id"], "id")

      if (internal_valid_task && !is.null(private$.internal_valid_task)) {
        private$.internal_valid_task$materialize_view(FALSE)
      }
      invisible(self)
    }
  ),

  active = list(
    #' @template field_id
    id = function(rhs) {
      if (missing(rhs)) {
        return(private$.id)
      }

      private$.hash = NULL
      private$.id = assert_id(rhs)
    },

    #' @field internal_valid_task (`Task` | `integer()` | `NULL`)\cr
    #' Optional validation task that can, e.g., be used for early stopping with learners such as XGBoost.
    #' See also the `$validate` field of [`Learner`].
    #' If integers are assigned they are removed from the primary task and an internal validation task
    #' with those ids is created from the primary task using only those ids.
    #' When assigning a new task, it is always cloned.
    internal_valid_task = function(rhs) {
      if (missing(rhs)) {
        return(invisible(private$.internal_valid_task))
      }
      private$.hash = NULL
      if (is.null(rhs)) {
        private$.internal_valid_task = NULL
        return(invisible(private$.internal_valid_task))
      }

      if (test_integerish(rhs)) {
        train_ids = setdiff(self$row_ids, rhs)
        rhs = self$clone(deep = TRUE)$filter(rhs)
        rhs$internal_valid_task = NULL
        self$row_roles$use = train_ids
      } else {
        if (!is.null(rhs$internal_valid_task)) {
          # avoid recursive structures
          error_input("Trying to assign task '%s' as a validation task, remove its validation task first.", rhs$id)
        }
        assert_task(rhs, task_type = self$task_type)
        rhs = rhs$clone(deep = TRUE)
      }

      ci1 = private$.col_info
      ci2 = rhs$col_info
      # don't do this too strictly, some column roles might just be important during training (weights)
      cols = unlist(self$col_roles[c("target", "feature")], use.names = FALSE)
      walk(cols, function(.col) {
        if (.col %nin% ci2$id) {
          error_input("Primary task has column '%s' which is not present in the validation task.", .col)
        }
        if (ci1[get("id") == .col, "type"]$type != ci2[get("id") == .col, "type"]$type) {
          error_input(
            "The type of column '%s' from the validation task differs from the type in the primary task.",
            .col
          )
        }
      })

      private$.internal_valid_task = rhs
      if (private$.internal_valid_task$nrow == 0) {
        warning_input("Internal validation task has 0 observations.")
      }
      invisible(private$.internal_valid_task)
    },

    #' @field hash (`character(1)`)\cr
    #' Hash (unique identifier) for this object.
    #' The hash is calculated based on the complete task object and `$row_ids`.
    #' If an internal validation task is set, the hash is recalculated.
    hash = function(rhs) {
      assert_ro_binding(rhs)
      if (is.null(private$.hash)) {
        private$.hash = task_hash(self, self$row_ids, ignore_internal_valid_task = FALSE)
      }

      private$.hash
    },

    #' @field row_hash (`character(1)`)\cr
    #' Hash (unique identifier) calculated based on the row ids.
    row_hash = function(rhs) {
      assert_ro_binding(rhs)
      if (is.null(private$.row_hash)) {
        private$.row_hash = calculate_hash(self$row_ids)
      }
      private$.row_hash
    },

    #' @field row_ids (positive `integer()`)\cr
    #' Returns the row ids of the [DataBackend] for observations with role "use".
    row_ids = function(rhs) {
      assert_ro_binding(rhs)
      private$.row_roles$use
    },

    #' @field row_names ([data.table::data.table()])\cr
    #' Returns a table with two columns:
    #'
    #' * `"row_id"` (`integer()`), and
    #' * `"row_name"` (`character()`).
    row_names = function(rhs) {
      assert_ro_binding(rhs)
      nn = private$.col_roles$name
      if (length(nn) == 0L) {
        return(NULL)
      }
      setnames(
        private$.backend$data(rows = self$row_ids, cols = c(private$.backend$primary_key, nn)),
        c("row_id", "row_name")
      )
    },

    #' @field feature_names (`character()`)\cr
    #' Returns all column names with `role == "feature"`.
    #'
    #' Note that this vector determines the default order of columns for `task$data(cols = NULL, ...)`.
    #' However, it is recommended to **not** rely on the order of columns, but instead always
    #' address columns by their name. The default order is not well defined after some
    #' operations, e.g. after `task$cbind()` or after processing via \CRANpkg{mlr3pipelines}.
    feature_names = function(rhs) {
      assert_ro_binding(rhs)
      private$.col_roles$feature
    },

    #' @field target_names (`character()`)\cr
    #' Returns all column names with role "target".
    target_names = function(rhs) {
      assert_ro_binding(rhs)
      private$.col_roles$target
    },

    #' @field properties (`character()`)\cr
    #' Set of task properties.
    #' Possible properties are stored in [mlr_reflections$task_properties][mlr_reflections].
    #' The following properties are currently standardized and understood by tasks in \CRANpkg{mlr3}:
    #'
    #' * `"strata"`: The task is resampled using one or more stratification variables (role `"stratum"`).
    #' * `"groups"`: The task comes with grouping/blocking information (role `"group"`).
    #' * `"weights_learner"`: If the task has observation weights with this role,
    #'    they are passed to the [Learner] during train.
    #'    The use of weights can be disabled by setting the learner's hyperparameter `use_weights` to `FALSE`.
    #' * `"weights_measure"`: If the task has observation weights with this role,
    #'    they are passed to the [Measure] for weighted scoring.
    #'    The use of weights can be disabled by setting the measure's hyperparameter `use_weights` to `FALSE`.
    #' * `"offset"`: The task includes one or more offset columns specifying fixed adjustments for model training
    #'    and possibly for prediction (role `"offset"`).
    #' * `"ordered"`: The task has columns which define the row order (role `"order"`).
    #'
    #' Note that above listed properties are calculated from the `$col_roles`, and may not be set explicitly.
    properties = function(rhs) {
      assert_ro_binding(rhs)
      prop_roles = c(
        groups = "group",
        strata = "stratum",
        weights_learner = "weights_learner",
        weights_measure = "weights_measure",
        offset = "offset",
        ordered = "order"
      )
      c(private$.properties, names(prop_roles)[lengths(private$.col_roles[prop_roles]) > 0L])
    },

    #' @field row_roles (named `list()`)\cr
    #' Each row (observation) can have an arbitrary number of roles in the learning task:
    #'
    #' - `"use"`: Use in train / predict / resampling.
    #'
    #' `row_roles` is a named list whose elements are named by row role and each element is an `integer()` vector of
    #' row ids.
    #' To alter the roles, just modify the list,
    #' e.g. with \R's set functions ([intersect()], [setdiff()], [union()], \ldots).
    row_roles = function(rhs) {
      if (missing(rhs)) {
        return(private$.row_roles)
      }

      assert_has_backend(self)
      assert_list(rhs, .var.name = "row_roles")
      if ("test" %chin% names(rhs) || "holdout" %chin% names(rhs)) {
        error_input("Setting row roles 'test'/'holdout' is no longer possible.")
      }
      assert_names(
        names(rhs),
        "unique",
        permutation.of = mlr_reflections$task_row_roles,
        .var.name = "names of row_roles"
      )
      rhs = map(rhs, assert_row_ids, .var.name = "elements of row_roles")
      private$.row_hash = NULL
      private$.hash = NULL
      private$.row_roles = rhs
    },

    #' @field col_roles (named `list()`)\cr
    #' Each column can be in one or more of the following groups to fulfill different roles:
    #'
    #' * `"feature"`: Regular feature used in the model fitting process.
    #' * `"target"`: Target variable. Most tasks only accept a single target column.
    #' * `"name"`: Row names / observation labels. To be used in plots. Can be queried with `$row_names`.
    #'   Not more than a single column can be associated with this role.
    #' * `"order"`: Data returned by `$data()` is ordered by this column (or these columns).
    #'   Columns must be sortable with [order()].
    #' * `"group"`: During resampling, observations with the same value of the variable with role "group" are marked
    #'   as "belonging together".
    #'   For each resampling iteration, observations of the same group will be exclusively assigned to be either in
    #'   the training set or in the test set.
    #'   Not more than a single column can be associated with this role.
    #' * `"stratum"`: Stratification variables. Multiple discrete columns may have this role.
    #' * `"weights_learner"`: If the task has observation weights with this role,
    #'    they are passed to the [Learner] during train.
    #'    The use of weights can be disabled by setting the learner's hyperparameter `use_weights` to `FALSE`.
    #' * `"weights_measure"`: If the task has observation weights with this role,
    #'    they are passed to the [Measure] for weighted scoring.
    #'    The use of weights can be disabled by setting the measure's hyperparameter `use_weights` to `FALSE`.
    #' * `"offset"`: Numeric columns used to specify fixed adjustments for model training.
    #'   Some models use offsets to simply shift predictions,
    #'   while others incorporate them to boost predictions from a baseline model.
    #'   For learners supporting offsets in multiclass settings,
    #'   an offset column must be provided for each target class.
    #'   These columns must follow the naming convention `"offset_{target_class_name}"`.
    #'   For an example of a learner that supports offsets, see `LearnerClassifXgboost`  of \CRANpkg{mlr3learners}.
    #'
    #' `col_roles` is a named list whose elements are named by column role and each element is a `character()` vector
    #' of column names.
    #' To alter the roles, just modify the list,
    #' e.g. with \R's set functions ([intersect()], [setdiff()], [union()], \ldots).
    #' The method `$set_col_roles` provides a convenient alternative to assign columns to roles.
    #'
    #' The roles `weights_learner` and `weights_measure` may only point to a single numeric column, but they can
    #' all point to the same column or different columns.
    #' Weights must be non-negative numerics with at least one weight being > 0.
    #' They don't necessarily need to sum up to 1.
    col_roles = function(rhs) {
      if (missing(rhs)) {
        return(private$.col_roles)
      }

      assert_has_backend(self)
      qassertr(rhs, "S[1,]", .var.name = "col_roles")
      assert_names(names(rhs), "unique", permutation.of = mlr_reflections$task_col_roles[[self$task_type]])
      assert_subset(
        unlist(rhs, use.names = FALSE),
        setdiff(private$.col_info$id, private$.backend$primary_key),
        .var.name = "elements of col_roles"
      )

      private$.hash = NULL
      private$.col_hashes = NULL
      private$.col_roles = task_check_col_roles(self, rhs)
    },

    #' @field nrow (`integer(1)`)\cr
    #' Returns the total number of rows with role "use".
    nrow = function(rhs) {
      assert_ro_binding(rhs)
      length(private$.row_roles$use)
    },

    #' @field ncol (`integer(1)`)\cr
    #' Returns the total number of columns with role "target" or "feature".
    ncol = function(rhs) {
      assert_ro_binding(rhs)
      length(private$.col_roles$feature) + length(private$.col_roles$target)
    },

    #' @field n_features (`integer(1)`)\cr
    #' Returns the total number of columns with role "feature" (i.e. the number of "active" features in the task).
    n_features = function(rhs) {
      assert_ro_binding(rhs)
      length(private$.col_roles$feature)
    },

    #' @field feature_types ([data.table::data.table()])\cr
    #' Returns a table with columns `id` and `type` where `id` are the column names of "active"
    #' features of the task and `type` is the storage type.
    feature_types = function(rhs) {
      assert_ro_binding(rhs)
      setkeyv(private$.col_info[list(private$.col_roles$feature), c("id", "type"), on = "id"], "id")
    },

    #' @field strata ([data.table::data.table()])\cr
    #' If the task has columns designated with role `"stratum"`,
    #' returns a table with one subpopulation per row and two columns:
    #'
    #' * `N` (`integer()`) with the number of observations in the subpopulation, and
    #' * `row_id` (list of `integer()`) as list column with the row ids in the respective subpopulation.
    #' Returns `NULL` if there is no stratification variable.
    #' See [Resampling] for more information on stratification.
    strata = function(rhs) {
      assert_has_backend(self)
      assert_ro_binding(rhs)
      cols = private$.col_roles$stratum
      if (length(cols) == 0L) {
        return(NULL)
      }

      row_ids = self$row_ids
      tab = private$.backend$data(rows = row_ids, cols = cols)
      tab$..row_id = row_ids
      tab = tab[, list(..N = .N, ..row_id = list(.SD$..row_id)), by = cols, .SDcols = "..row_id"][, (cols) := NULL]
      setnames(tab, c("..N", "..row_id"), c("N", "row_id"))[]
    },

    #' @field groups ([data.table::data.table()])\cr
    #' If the task has a column with designated role `"group"`, a table with two columns:
    #'
    #' * `row_id` (`integer()`), and
    #' * grouping variable `group` (`vector()`).
    #'
    #' Returns `NULL` if there is no grouping column.
    #' See [Resampling] for more information on grouping.
    groups = function(rhs) {
      assert_has_backend(self)
      assert_ro_binding(rhs)
      group_cols = private$.col_roles$group
      if (length(group_cols) == 0L) {
        return(NULL)
      }
      data = private$.backend$data(private$.row_roles$use, c(private$.backend$primary_key, group_cols))
      setnames(data, c("row_id", "group"))[]
    },

    #' @field order ([data.table::data.table()])\cr
    #' If the task has at least one column with designated role `"order"`, a table with two columns:
    #'
    #' * `row_id` (`integer()`), and
    #' * ordering vector `order` (`integer()`).
    #'
    #' Returns `NULL` if there is no order column.
    order = function(rhs) {
      assert_has_backend(self)
      assert_ro_binding(rhs)

      order_cols = private$.col_roles$order
      if (length(order_cols) == 0L) {
        return(NULL)
      }

      data = private$.backend$data(private$.row_roles$use, order_cols)
      data.table(row_id = private$.row_roles$use, order = do.call(order, data))
    },

    #' @field weights ([data.table::data.table()])\cr
    #' Deprecated, use `$weights_learner` instead.
    weights = function(rhs) {
      assert_ro_binding(rhs)
      .Deprecated("Task$weights_learner", old = "Task$weights")
      self$weights_learner
    },

    #' @field weights_learner ([data.table::data.table()])\cr
    #' Returns the observation weights used for training a [Learner] (column role `weights_learner`)
    #' as a `data.table` with the following columns:
    #'
    #' * `row_id` (`integer()`), and
    #' * `weight` (`numeric()`).
    #'
    #' Returns `NULL` if there is no column with the designated role.
    weights_learner = function(rhs) {
      assert_has_backend(self)
      assert_ro_binding(rhs)
      weight_cols = private$.col_roles[["weights_learner"]]
      if (length(weight_cols) == 0L) {
        return(NULL)
      }
      data = private$.backend$data(private$.row_roles$use, c(private$.backend$primary_key, weight_cols))
      setnames(data, c("row_id", "weight"))[]
    },

    #' @field weights_measure ([data.table::data.table()])\cr
    #' Returns the observation weights used for scoring a prediction with a [Measure] (column role `weights_measure`)
    #' as a `data.table` with the following columns:
    #'
    #' * `row_id` (`integer()`), and
    #' * `weight` (`numeric()`).
    #'
    #' Returns `NULL` if there is no column with the designated role.
    weights_measure = function(rhs) {
      assert_has_backend(self)
      assert_ro_binding(rhs)
      weight_cols = private$.col_roles[["weights_measure"]]
      if (length(weight_cols) == 0L) {
        return(NULL)
      }
      data = private$.backend$data(private$.row_roles$use, c(private$.backend$primary_key, weight_cols))
      setnames(data, c("row_id", "weight"))[]
    },
    #' @field offset ([data.table::data.table()])\cr
    #' If the task has a column with designated role `"offset"`, a table with two or more columns:
    #'
    #' * `row_id` (`integer()`), and
    #' * offset variable(s) (`numeric()`).
    #'
    #' For regression or binary classification tasks, there will be only a single-column offset.
    #' For multiclass tasks, it may return multiple offset columns, one for each target class.
    #' If there is only one offset column, it will be named as `offset`.
    #'
    #' If there are no columns with the `"offset"` role, `NULL` is returned.
    offset = function(rhs) {
      assert_has_backend(self)
      assert_ro_binding(rhs)
      offset_cols = private$.col_roles$offset
      if (length(offset_cols) == 0L) {
        return(NULL)
      }

      data = private$.backend$data(private$.row_roles$use, c(private$.backend$primary_key, offset_cols))
      if (length(offset_cols) == 1L) {
        setnames(data, c("row_id", "offset"))[]
      } else {
        setnames(data, c("row_id", offset_cols))[]
      }
    },

    #' @field labels (named `character()`)\cr
    #'   Retrieve `labels` (prettier formatted names) from columns.
    #'   Internally queries the column `label` of the table in field `col_info`.
    #'   Columns ids referenced by the name of the vector, the labels are the actual string values.
    #'
    #'   Assigning to this column update the task by reference.
    #'   You have to provide a character vector of labels, named with column ids.
    #'   To remove a label, set it to `NA`.
    #'   Alternatively, you can provide a [data.frame()] with the two columns
    #'   `"id"` and `"label"`.
    labels = function(rhs) {
      active = union(self$target_names, self$feature_names)

      if (missing(rhs)) {
        tab = ijoin(private$.col_info, active, c("id", "label"), "id")
        return(set_names(tab[["label"]], tab[["id"]]))
      }

      if (is.data.frame(rhs)) {
        # convert to named character
        assert_data_frame(rhs, ncols = 2L)
        assert_names(names(rhs), permutation.of = c("id", "label"))
        rhs = set_names(rhs[["label"]], rhs[["id"]])
      }

      assert_names(names(rhs), type = "unique")
      assert_subset(names(rhs), active)
      private$.col_info[list(names(rhs)), "label" := rhs, on = "id"]

      invisible(self)
    },

    #' @template field_col_hashes
    col_hashes = function() {
      if (is.null(private$.col_hashes)) {
        private$.col_hashes = private$.backend$col_hashes[setdiff(
          unlist(private$.col_roles, use.names = FALSE),
          private$.backend$primary_key
        )]
      }
      private$.col_hashes
    },

    #' @field characteristics (`list()`)\cr
    #' List of characteristics of the task, e.g. `list(n = 5, p = 7)`.
    characteristics = function(rhs) {
      if (missing(rhs)) {
        return(private$.characteristics)
      }

      private$.characteristics = assert_list(rhs, null.ok = TRUE)
      private$.hash = NULL
    },

    #' @field row_ids_backend (`integer()`)\cr
    #' Returns all row ids from the backend, regardless of their roles.
    #' This is different from `$row_ids` which only returns rows with role "use".
    row_ids_backend = function(rhs) {
      assert_ro_binding(rhs)
      private$.backend$rownames
    },

    #' @template field_label
    label = function(rhs) {
      if (missing(rhs)) {
        return(private$.label)
      }
      private$.label = assert_string(rhs, na.ok = TRUE)
    },

    #' @template field_task_type
    task_type = function(rhs) {
      if (missing(rhs)) {
        return(private$.task_type)
      }
      private$.task_type = assert_choice(rhs, mlr_reflections$task_types$type)
    },

    #' @template field_man
    man = function(rhs) {
      if (missing(rhs)) {
        return(private$.man)
      }
      private$.man = assert_string(rhs, na.ok = TRUE)
    },

    #' @field extra_args (named `list()`)\cr
    #' Additional arguments set during construction.
    #' Required for [convert_task()].
    extra_args = function(rhs) {
      if (missing(rhs)) {
        return(private$.extra_args)
      }
      private$.extra_args = assert_list(rhs, names = "unique")
    },

    #' @field backend ([DataBackend])\cr
    #' Abstract interface to the data of the task.
    backend = function(rhs) {
      if (missing(rhs)) {
        return(private$.backend)
      }
      # Allow NULL for cleanup in deep_clone
      if (!is.null(rhs)) {
        assert_backend(rhs)
      }
      private$.backend = rhs
    },

    #' @field col_info ([data.table::data.table()])\cr
    #' Table with 4 columns, mainly for internal purposes:
    #' - `"id"` (`character()`) stores the name of the column.
    #' - `"type"` (`character()`) holds the storage type of the variable, e.g. `integer`, `numeric` or `character`.
    #'   See [mlr_reflections$task_feature_types][mlr_reflections] for a complete list of allowed types.
    #' - `"levels"` (`list()`) stores a vector of distinct values (levels) for ordered and unordered factor variables.
    #' - `"label"` (`character()`) stores a vector of prettier, formatted column names.
    #' - `"fix_factor_levels"` (`logical()`) stores flags which determine if the levels of the respective variable
    #'   need to be reordered after querying the data from the [DataBackend].
    #'
    #' Note that all columns of the [DataBackend], also columns which are not selected or have any role, are listed
    #' in this table.
    col_info = function(rhs) {
      if (missing(rhs)) {
        return(private$.col_info)
      }
      assert_data_table(rhs)
      private$.col_info = rhs
    },

    #' @field mlr3_version (`package_version`)\cr
    #' Package version of `mlr3` used to create the task.
    mlr3_version = function(rhs) {
      assert_ro_binding(rhs)
      private$.mlr3_version
    }
  ),

  private = list(
    .internal_valid_task = NULL,
    .id = NULL,
    .label = NULL,
    .task_type = NULL,
    .man = NULL,
    .extra_args = NULL,
    .backend = NULL,
    .col_info = NULL,
    .mlr3_version = NULL,
    .properties = NULL,
    .col_roles = NULL,
    .row_roles = NULL,
    .hash = NULL,
    .col_hashes = NULL,
    .characteristics = NULL,
    .row_hash = NULL,

    deep_clone = function(name, value) {
      # NB: DataBackends are never copied!
      if (name == ".col_info") {
        copy(value)
      } else if (name == ".internal_valid_task" && !is.null(value)) {
        value$clone(deep = TRUE)
      } else {
        value
      }
    }
  )
)

task_set_roles = function(li, elements, roles = NULL, add_to = NULL, remove_from = NULL, allow_duplicated = FALSE) {
  if (!is.null(roles)) {
    assert_subset(roles, names(li))
    if (allow_duplicated) {
      for (role in roles) {
        li[[role]] = c(li[[role]], elements)
      }

      for (role in setdiff(names(li), roles)) {
        li[[role]] = li[[role]][li[[role]] %nin% elements]
      }
    } else {
      for (role in roles) {
        li[[role]] = union(li[[role]], elements)
      }

      for (role in setdiff(names(li), roles)) {
        li[[role]] = setdiff(li[[role]], elements)
      }
    }
  }

  if (!is.null(add_to)) {
    assert_subset(add_to, names(li))
    if (allow_duplicated) {
      for (role in add_to) {
        li[[role]] = c(li[[role]], elements)
      }
    } else {
      for (role in add_to) {
        li[[role]] = union(li[[role]], elements)
      }
    }
  }

  if (!is.null(remove_from)) {
    assert_subset(remove_from, names(li))
    if (allow_duplicated) {
      for (role in remove_from) {
        li[[role]] = li[[role]][li[[role]] %nin% elements]
      }
    } else {
      for (role in remove_from) {
        li[[role]] = setdiff(li[[role]], elements)
      }
    }
  }

  li
}

#' @title Check Column Roles
#'
#' @description
#' Internal function to check column roles.
#'
#' @param task ([Task])\cr
#'  Task.
#' @param new_roles (`list()`)\cr
#'  Column roles.
#'
#' @keywords internal
#' @export
task_check_col_roles = function(task, new_roles, ...) {
  UseMethod("task_check_col_roles")
}

#' @rdname task_check_col_roles
#' @export
task_check_col_roles.Task = function(task, new_roles, ...) {
  if ("weight" %in% names(new_roles)) {
    error_input("Task role 'weight' is deprecated, use 'weights_learner' instead")
  }

  for (role in c("group", "name", "weights_learner", "weights_measure")) {
    if (length(new_roles[[role]]) > 1L) {
      error_input("There may only be up to one column with role '%s'", role)
    }
  }

  # check weights
  for (role in c("weights_learner", "weights_measure")) {
    if (length(new_roles[[role]]) > 0L) {
      col = task$backend$data(task$backend$rownames, cols = new_roles[[role]])
      assert_numeric(col[[1L]], lower = 0, any.missing = FALSE, .var.name = names(col))
    }
  }

  # check name
  if (length(new_roles[["name"]])) {
    row_names = task$backend$data(task$backend$rownames, cols = new_roles[["name"]])
    if (!is.character(row_names[[1L]]) && !is.factor(row_names[[1L]])) {
      error_input(
        "Assertion on '%s' failed: Must be of type 'character' or 'factor', not %s",
        names(row_names),
        class(row_names[[1]])
      )
    }
  }

  # check offset
  if (
    length(new_roles[["offset"]]) &&
      any(fget_keys(task$col_info, new_roles[["offset"]], "type", key = "id") %nin% c("numeric", "integer"))
  ) {
    error_input(
      "Offset column(s) %s must be a numeric or integer column",
      paste0("'", new_roles[["offset"]], "'", collapse = ", ")
    )
  }

  if (length(new_roles[["offset"]]) && any(task$missings(cols = new_roles[["offset"]]) > 0)) {
    missings = task$missings(cols = new_roles[["offset"]])
    missings = names(missings[missings > 0])
    error_input("Offset column(s) %s contain missing values", paste0("'", missings, "'", collapse = ", "))
  }

  new_roles
}

#' @rdname task_check_col_roles
#' @export
task_check_col_roles.TaskClassif = function(task, new_roles, ...) {
  # check target
  if (length(new_roles[["target"]]) > 1L) {
    error_input("There may only be up to one column with role 'target'")
  }

  if (
    length(new_roles[["target"]]) &&
      any(fget_keys(task$col_info, new_roles[["target"]], "type", key = "id") %nin% c("factor", "ordered"))
  ) {
    error_input("Target column(s) '%s' must be a factor or ordered factor", new_roles[["target"]])
  }

  if (length(new_roles[["offset"]]) > 1L && length(task$class_names) == 2L) {
    error_input("There may only be up to one column with role 'offset' for binary classification tasks")
  }

  if (length(new_roles[["offset"]]) > 1L) {
    expected_names = paste0("offset_", task$class_names)
    expect_subset(new_roles[["offset"]], expected_names, label = "col_roles")
  }

  NextMethod()
}

#' @rdname task_check_col_roles
#' @export
task_check_col_roles.TaskRegr = function(task, new_roles, ...) {
  for (role in c("target", "offset")) {
    if (length(new_roles[[role]]) > 1L) {
      error_input("There may only be up to one column with role '%s'", role)
    }
  }

  if (
    length(new_roles[["target"]]) &&
      any(fget_keys(task$col_info, new_roles[["target"]], "type", key = "id") %nin% c("numeric", "integer"))
  ) {
    error_input("Target column '%s' must be a numeric or integer column", new_roles[["target"]])
  }

  NextMethod()
}

#' @rdname task_check_col_roles
#' @export
task_check_col_roles.TaskSupervised = function(task, new_roles, ...) {
  # check target
  if (length(new_roles$target) == 0L) {
    error_input("Supervised tasks need at least one target column")
  }

  NextMethod()
}

#' @rdname task_check_col_roles
#' @export
task_check_col_roles.TaskUnsupervised = function(task, new_roles, ...) {
  # check target
  if (length(new_roles$target) != 0L) {
    error_input("Unsupervised tasks may not have a target column")
  }

  NextMethod()
}

#' @title Column Information for Backend
#'
#' @description
#' Collects column information for backend.
#'
#' Currently, this includes:
#' * storage type
#' * levels (factor / ordered), but not for the primary key column
#'
#' @param x (any)\cr
#'   A backend-like object for which to retrieve column information.
#' @param ... (any)\cr
#'   Additional arguments.
#'
#' @export
#' @keywords internal
col_info = function(x, ...) {
  UseMethod("col_info")
}

#' @rdname col_info
#' @param primary_key (`character()`)\cr
#'   The primary key of the backend.
#' @export
# nolint next
col_info.data.table = function(x, primary_key = character(), ...) {
  types = map_chr(x, function(x) class(x)[1L])
  discrete = setdiff(names(types)[types %chin% c("factor", "ordered")], primary_key)
  levels = insert_named(named_list(names(types)), lapply(x[, discrete, with = FALSE], distinct_values, drop = FALSE))
  data.table(id = names(types), type = unname(types), levels = levels, key = "id")
}

#' @rdname col_info
#' @export
# nolint next
col_info.DataBackend = function(x, ...) {
  types = map_chr(x$head(1L), function(x) class(x)[1L])
  discrete = setdiff(names(types)[types %chin% c("factor", "ordered")], x$primary_key)
  levels = insert_named(named_list(names(types)), x$distinct(rows = NULL, cols = discrete))
  data.table(id = names(types), type = unname(types), levels = levels, key = "id")
}

#' @export
# nolint next
as.data.table.Task = function(x, ...) {
  x$data()
}

#' @export
# nolint next
head.Task = function(x, n = 6L, ...) {
  assert_number(n, na.ok = FALSE)
  x$data(rows = head(x$row_ids, n))
}

#' @export
# nolint next
tail.Task = function(x, n = 6L, ...) {
  assert_number(n, na.ok = FALSE)
  x$data(rows = tail(x$row_ids, n))
}

task_rm_backend = function(task) {
  # fix task hash
  ee = get_private(task)
  ee$.hash = force(task$hash)
  ee$.col_hashes = force(task$col_hashes)
  ee$.internal_valid_task$backend = NULL

  # NULL backend
  task$backend = NULL

  task
}


#' @export
# nolint next
rd_info.Task = function(obj, section, ...) {
  x = c(
    "",
    sprintf("* Task type: %s", rd_format_string(obj$task_type)),
    sprintf("* Dimensions: %ix%i", obj$nrow, obj$ncol),
    sprintf("* Properties: %s", rd_format_string(obj$properties)),
    sprintf("* Has Missings: `%s`", any(obj$missings() > 0L)),
    sprintf("* Target: %s", rd_format_string(obj$target_names)),
    sprintf("* Features: %s", rd_format_string(obj$feature_names))
  )
  paste(x, collapse = "\n")
}

#' @export
summary.Task = function(object, limit = object$nrow, ...) {
  summary(head(object, limit))
}

with_seed = function(seed, expr) {
  old_seed = get0(".Random.seed", globalenv(), mode = "integer", inherits = FALSE)
  if (is.null(old_seed)) {
    runif(1L)
    old_seed = get0(".Random.seed", globalenv(), mode = "integer", inherits = FALSE)
  }

  on.exit(assign(".Random.seed", old_seed, globalenv()), add = TRUE)
  set.seed(seed)
  force(expr)
}

with_future = function(backend, expr, ...) {
  requireNamespace("future")
  oplan = force(future::plan(backend, ...))
  on.exit(future::plan(oplan), add = TRUE)
  force(expr)
}

with_mirai = function(expr, compute, seed = NULL) {
  requireNamespace("mirai")
  mirai::daemons(1, .compute = compute, seed = seed)
  on.exit(mirai::daemons(0, .compute = compute), add = TRUE)
  force(expr)
  testthat::expect_true(mirai::status(.compute = compute)$mirai["completed"] > 0)
}

private = function(x) {
  x[[".__enclos_env__"]][["private"]]
}


iris_weights_learner = TaskClassif$new(
  "iris_weights_learner",
  as_data_backend(cbind(datasets::iris, data.frame(w = rep(c(1, 10, 100), each = 50)))),
  target = "Species"
)
iris_weights_learner$set_col_roles("w", "weights_learner")

iris_weights_measure = TaskClassif$new(
  "iris_weights_measure",
  as_data_backend(cbind(datasets::iris, data.frame(w = rep(c(1, 10, 100), each = 50)))),
  target = "Species"
)
iris_weights_measure$set_col_roles("w", "weights_measure")

cars_weights_learner = TaskRegr$new(
  "cars_weights_learner",
  as_data_backend(cbind(datasets::cars, data.frame(w = rep(c(1, 10), each = 25)))),
  target = "dist"
)
cars_weights_learner$set_col_roles("w", "weights_learner")

cars_weights_measure = TaskRegr$new(
  "cars_weights_measure",
  as_data_backend(cbind(datasets::cars, data.frame(w = rep(c(1, 10), each = 25)))),
  target = "dist"
)
cars_weights_measure$set_col_roles("w", "weights_measure")

#' @title Learner Autotest Suite
#'
#' @description
#' The autotest suite is a collection of functions to test learners in a standardized way.
#' Extension packages need to specialize the S3 methods in the file.
#
#' @details
#' `run_autotest(learner)` generates multiple tasks, depending on the properties of the learner
#' and tests the learner on each task, with each predict type.
#' Calls `generate_tasks()` to generate tasks and `run_experiment()` to run the experiments.
#' See `generate_tasks()` for a list of tasks that are generated.
#' To debug, simply run `result = run_autotest(learner)`
#' and proceed with investigating he task, learner and prediction of the returned `result`.
#'
#' `run_experiment(task, learner)` runs a single experiment.
#' Calls `train()` and `predict()` on the learner and checks the prediction with `score()`.
#' The prediction is checked with `sanity_check()`.
#'
#' `generate_tasks(learner)` generates multiple tasks for a given learner.
#' Calls `generate_data()` and `generate_generic_tasks()` to generate tasks with different feature types.
#'
#' @noRd
NULL

#' @title Generate Tasks for a Learner
#'
#' @description
#' Generates multiple tasks for a given [Learner], based on its properties.
#'
#' @param learner [Learner]\cr
#'  Learner to generate tasks for.
#' @param proto [Task]\cr
#'  Prototype task to generate tasks from.
#'
#' @return (List of [Task]s).
#'
#' @noRd
generate_generic_tasks = function(learner, proto) {
  tasks = list()
  n = proto$nrow
  p = length(proto$feature_names)

  if (p > 0L) {
    # individual tasks with each supported feature type
    for (ftype in learner$feature_types) {
      sel = proto$feature_types[ftype, "id", on = "type", with = FALSE][[1L]]
      tasks[[sprintf("feat_single_%s", ftype)]] = proto$clone(deep = TRUE)$select(sel)
    }
  }

  # task with all supported features types
  sel = proto$feature_types[list(learner$feature_types), "id", on = "type", with = FALSE, nomatch = NULL][[1L]]
  tasks$feat_all = proto$clone(deep = TRUE)$select(sel)

  # task with missing values
  if ("missings" %in% learner$properties) {
    # one missing val in each feature
    features = proto$feature_names
    rows = sample(n, length(features))
    data = proto$data(cols = features)
    for (j in seq_along(features)) {
      data.table::set(data, rows[j], features[j], NA)
    }
    tasks$missings = proto$clone(deep = TRUE)$select(character())$cbind(data)

    if (length(features)) {
      # no row with no missing -> complete.cases() won't help
      features = sample(features, n, replace = TRUE)
      data = proto$data(cols = proto$feature_names)
      for (i in seq_along(features)) {
        data.table::set(data, i = i, j = features[i], NA)
      }
      tasks$missings_each_row = proto$clone(deep = TRUE)$select(character())$cbind(data)
    }
  }

  # task with weights
  if ("weights" %in% learner$properties) {
    tmp = proto$clone(deep = TRUE)$cbind(data.frame(weights = runif(n)))
    tmp$set_col_roles(cols = "weights", roles = "weights_learner")
    tasks$weights_learner = tmp
  }

  # task with offset
  if ("offset" %in% learner$properties) {
    if ("multiclass" %in% learner$properties && "multiclass" %in% proto$properties) {
      offset_cols = paste0("offset_", proto$class_names)
      # One offset column per class
      offset_data = as.data.frame(
        mlr3misc::set_names(
          lapply(offset_cols, function(col) runif(n)),
          offset_cols
        )
      )
      tmp = proto$clone(deep = TRUE)$cbind(offset_data)
      tmp$set_col_roles(cols = offset_cols, roles = "offset")
    } else {
      tmp = proto$clone(deep = TRUE)$cbind(data.frame(offset = runif(n)))
      tmp$set_col_roles(cols = "offset", roles = "offset")
    }
    tasks$offset = tmp
  }

  # task with non-ascii feature names
  if (p > 0L) {
    sel = proto$feature_types[list(learner$feature_types), "id", on = "type", with = FALSE, nomatch = NULL][[1L]]
    tasks$utf8_feature_names = proto$clone(deep = TRUE)$select(sel)
    old = sel[1L]
    new = "\u00e4 + \u1e9e"
    tasks$utf8_feature_names$rename(old, new)
  }

  # make sure that task ids match list names
  mlr3misc::imap(tasks, function(x, n) {
    x$id = n
    x
  })
}

#' @title Generate Data for a Learner
#'
#' @description
#' Generates data for a given [Learner], based on its supported feature types.
#' Data is created for logical, integer, numeric, character, factor, ordered, and POSIXct features.
#'
#' @param learner [Learner]\cr
#'  Learner to generate data for.
#' @param N `integer(1)`\cr
#'  Number of rows of generated data.
#'
#' @return [data.table::data.table()]
#'
#' @noRd
generate_data = function(learner, N) {
  generate_feature = function(type) {
    switch(
      type,
      logical = sample(rep_len(c(TRUE, FALSE), N)),
      integer = sample(rep_len(1:3, N)),
      numeric = runif(N),
      character = sample(rep_len(letters[1:2], N)),
      factor = sample(factor(rep_len(c("f1", "f2"), N), levels = c("f1", "f2"))),
      ordered = sample(ordered(rep_len(c("o1", "o2"), N), levels = c("o1", "o2"))),
      POSIXct = Sys.time() - runif(N, min = 0, max = 10 * 365 * 24 * 60 * 60),
      Date = Sys.Date() - runif(N, min = 0, max = 10 * 365)
    )
  }
  types = unique(learner$feature_types)
  do.call(data.table::data.table, mlr3misc::set_names(mlr3misc::map(types, generate_feature), types))
}

#' @title Generate Tasks for a Learner
#'
#' @description
#' Generates multiple tasks for a given [Learner], based on its properties.
#' This function is primarily used for unit tests, but can also assist while writing custom learners.
#' The following tasks are created:
#' * `feat_single_*`: Tasks with a single feature type.
#' * `feat_all_*`: Task with all supported feature types.
#' * `missings_*`: Task with missing values.
#' * `utf8_feature_names_*`: Task with non-ascii feature names.
#' * `sanity`: Task with a simple dataset to check if the learner is working.
#' * `sanity_reordered`: Task with the same dataset as `sanity`, but with reordered columns.
#' * `sanity_switched`: Task with the same dataset as `sanity`, but with the positive class switched.
#'
#' @param learner [Learner]\cr
#'  Learner to generate tasks for.
#' @param N `integer(1)`\cr
#'   Number of rows of generated tasks.
#'
#' @return `list` of [Task]s
#' @keywords internal
#' @export
#' @examples
#' tasks = generate_tasks(lrn("classif.rpart"))
#' tasks$missings_binary$data()
generate_tasks = function(learner, N = 30L) {
  N = checkmate::assert_int(N, lower = 10L, coerce = TRUE)
  UseMethod("generate_tasks")
}

#' @export
# nolint next
generate_tasks.LearnerClassif = function(learner, N = 30L) {
  tasks = list()

  # generate binary tasks
  if ("twoclass" %in% learner$properties) {
    target = factor(rep_len(head(LETTERS, 2L), N))
    data = cbind(data.table::data.table(target = target), generate_data(learner, N))
    task = mlr3::TaskClassif$new("proto", mlr3::as_data_backend(data), target = "target", positive = "A")
    gen_tasks = generate_generic_tasks(learner, task)
    # set names
    lapply(gen_tasks, function(x) x$id = paste0(x$id, "_binary"))
    gen_tasks = mlr3misc::set_names(gen_tasks, paste0(names(gen_tasks), "_binary"))
    tasks = c(tasks, gen_tasks)
  }

  # generate multiclass tasks
  if ("multiclass" %in% learner$properties) {
    target = factor(rep_len(head(LETTERS, 3L), N))
    data = cbind(data.table::data.table(target = target), generate_data(learner, N))
    task = mlr3::TaskClassif$new("proto", mlr3::as_data_backend(data), target = "target")
    gen_tasks = generate_generic_tasks(learner, task)
    # set names
    lapply(gen_tasks, function(x) x$id = paste0(x$id, "_multiclass"))
    gen_tasks = mlr3misc::set_names(gen_tasks, paste0(names(gen_tasks), "_multiclass"))
    tasks = c(tasks, gen_tasks)
  }

  # generate sanity task
  # nolint next
  data = with_seed(100, {
    data = data.table::data.table(
      x = c(rnorm(100, 0, 1), rnorm(100, 10, 1)),
      y = rep(as.factor(c("A", "B")), each = 100)
    )
    data$unimportant = runif(nrow(data), min = 0, max = 3)
    data
  })
  tasks$sanity = mlr3::TaskClassif$new("sanity", mlr3::as_data_backend(data), target = "y", positive = "A")

  # sanity task, but reorder columns between train and predict in run_experiment()
  tasks$sanity_reordered = mlr3::TaskClassif$new("sanity_reordered", mlr3::as_data_backend(data), target = "y")

  # sanity task, but with other label as positive class to detect label switches
  tasks$sanity_switched = mlr3::TaskClassif$new(
    "sanity_switched",
    mlr3::as_data_backend(data),
    target = "y",
    positive = "B"
  )

  tasks
}
registerS3method("generate_tasks", "LearnerClassif", generate_tasks.LearnerClassif)

#' @export
# nolint next
generate_tasks.LearnerRegr = function(learner, N = 30L) {
  target = rnorm(N)
  data = cbind(data.table::data.table(target = target), generate_data(learner, N))
  task = mlr3::TaskRegr$new("proto", mlr3::as_data_backend(data), target = "target")

  tasks = generate_generic_tasks(learner, task)

  # generate sanity task
  # nolint next
  data = with_seed(100, {
    y = seq(from = -10, to = 10, length.out = 100)
    data.table::data.table(
      y = y,
      x = y + rnorm(length(y), mean = 1),
      unimportant = runif(length(y), min = 0, max = 1)
    )
  })
  tasks$sanity = mlr3::TaskRegr$new("sanity", mlr3::as_data_backend(data), target = "y")
  tasks$sanity_reordered = mlr3::TaskRegr$new("sanity_reordered", mlr3::as_data_backend(data), target = "y")

  tasks
}
registerS3method("generate_tasks", "LearnerRegr", generate_tasks.LearnerRegr)

#' @title Sanity Check for Predictions
#'
#' @description
#' Checks the sanity of a prediction.
#'
#' @param prediction [Prediction]\cr
#'  Prediction to check.
#'
#' @return (`logical(1)`).
#'
#' @noRd
sanity_check = function(prediction, ...) {
  UseMethod("sanity_check")
}

# nolint next
sanity_check.PredictionClassif = function(prediction, ...) {
  prediction$score(mlr3::msr("classif.ce")) <= 0.3
}
registerS3method("sanity_check", "LearnerClassif", sanity_check.PredictionClassif)


# nolint next
sanity_check.PredictionRegr = function(prediction, ...) {
  prediction$score(mlr3::msr("regr.mse")) <= 2
}
registerS3method("sanity_check", "LearnerRegr", sanity_check.PredictionRegr)


#' @title Run a Single Learner Test
#'
#' @description
#' Runs a single experiment with a given task and learner.
#'
#' @param task [Task]\cr
#'  Task to run the experiment on.
#' @param learner [Learner]\cr
#'  Learner to run the experiment with.
#' @param seed `integer(1)`\cr
#'  Seed to use for the experiment.
#'  If `NULL`, a random seed is generated.
#' @param configure_learner `function(learner, task)`\cr
#'  Function to configure the learner before training.
#'  Useful when learner settings need to be adjusted for a specific task.
#'
#' @return `list` with the following elements:
#'  - `ok` (`logical(1)`): Success flag.
#'  - `learner` ([Learner]): Learner used for the experiment.
#'  - `prediction` ([Prediction]): Prediction object.
#'  - `error` (`character()`): Error message if `ok` is `FALSE`.
#'  - `seed` (`integer(1)`): Seed used for the experiment.
#'
#' @noRd
run_experiment = function(task, learner, seed = NULL, configure_learner = NULL) {
  # function to collect error message and objects
  err = function(info, ...) {
    info = sprintf(info, ...)
    list(
      ok = FALSE,
      seed = seed,
      task = task,
      learner = learner,
      prediction = prediction,
      score = score,
      prediction_marshaling = prediction_marshaling,
      learner_encapsulated = learner_encapsulated,
      error = sprintf("[%s] learner '%s' on task '%s' failed: %s", stage, learner$id, task$id, info)
    )
  }

  # some errors from upstream packages contain %
  error_as_character = function(error) {
    gsub("%", "%%", as.character(error), fixed = TRUE)
  }

  # seed handling
  if (is.null(seed)) {
    seed = sample.int(floor(.Machine$integer.max / 2L), 1L)
  }

  old_seed = get0(".Random.seed", globalenv(), mode = "integer", inherits = FALSE)
  if (is.null(old_seed)) {
    runif(1L)
    old_seed = get0(".Random.seed", globalenv(), mode = "integer", inherits = FALSE)
  }
  on.exit(assign(".Random.seed", old_seed, globalenv()), add = TRUE)

  set.seed(seed)

  task = mlr3::assert_task(mlr3::as_task(task))
  learner = mlr3::assert_learner(mlr3::as_learner(learner, clone = TRUE))
  if (!is.null(configure_learner)) {
    configure_learner(learner = learner, task = task)
  }
  prediction = NULL
  score = NULL
  prediction_marshaling = NULL
  learner_encapsulated = NULL

  # check train
  stage = "train()"

  # enable weights
  # the next lines are maybe not strictly necessary, but test that the defaults are
  # what they should be
  if ("weights" %in% learner$properties) {
    if (learner$use_weights != "use") {
      return(err("use_weights != 'use' for learner with property 'weights' on init!"))
    }
  } else {
    if (learner$use_weights != "error") {
      return(err("use_weights != 'error' for learner without property 'weights' on init!"))
    }
  }

  ok = suppressWarnings(try(learner$train(task), silent = TRUE))
  if (inherits(ok, "try-error")) {
    return(err(error_as_character(ok)))
  }
  if (is.null(learner$model)) {
    return(err("model is NULL"))
  }

  # check predict
  stage = "predict()"

  prediction = suppressWarnings(try(learner$predict(task), silent = TRUE))
  if (inherits(prediction, "try-error")) {
    ok = prediction
    prediction = NULL
    return(err(error_as_character(ok)))
  }
  msg = checkmate::check_class(prediction, "Prediction")
  if (!isTRUE(msg)) {
    return(err(msg))
  }
  if (prediction$task_type != learner$task_type) {
    return(err("learner and prediction have different task_type"))
  }

  # catch for mlr3proba tasks, which all return every possible predict type
  if (!(learner$task_type %in% c("dens", "surv"))) {
    expected = mlr3::mlr_reflections$learner_predict_types[[learner$task_type]][[learner$predict_type]]
    msg = checkmate::check_subset(expected, prediction$predict_types, empty.ok = FALSE)
    if (!isTRUE(msg)) {
      return(err(msg))
    }

    if (learner$predict_type == "response") {
      msg = checkmate::check_set_equal(learner$predict_type, prediction$predict_types)
      if (!isTRUE(msg)) {
        return(err(msg))
      }
    } else {
      msg = checkmate::check_subset(learner$predict_type, prediction$predict_types, empty.ok = FALSE)
      if (!isTRUE(msg)) {
        return(err(msg))
      }
    }
  }

  if (grepl("reordered", task$id)) {
    # compare prediction with reordered newdata
    newdata = task$data(cols = rev(task$feature_names))
    tmp = learner$predict_newdata(newdata)

    if (!isTRUE(all.equal(prediction$response, tmp$response))) {
      return(err("Task columns cannot be reordered"))
    }
  }

  # check score
  stage = "score()"

  score = try(
    prediction$score(
      mlr3::default_measures(learner$task_type),
      task = task,
      learner = learner,
      train_set = task$row_ids
    ),
    silent = TRUE
  )
  if (inherits(score, "try-error")) {
    ok = score
    score = NULL
    return(err(error_as_character(ok)))
  }
  msg = checkmate::check_numeric(score, any.missing = FALSE)
  if (!isTRUE(msg)) {
    return(err(msg))
  }

  # run sanity check on sanity task
  if (
    startsWith(task$id, "sanity") && !sanity_check(prediction, task = task, learner = learner, train_set = task$row_ids)
  ) {
    return(err("sanity check failed"))
  }

  # check importance, selected_features and oob_error methods
  if (startsWith(task$id, "feat_all")) {
    if ("importance" %in% learner$properties) {
      importance = learner$importance()
      msg = checkmate::check_numeric(rev(importance), any.missing = FALSE, min.len = 1L, sorted = TRUE)
      if (!isTRUE(msg)) {
        return(err(msg))
      }
      msg = checkmate::check_names(names(importance), subset.of = task$feature_names)
      if (!isTRUE(msg)) {
        return(err("Names of returned importance scores do not match task names: %s", str_collapse(names(importance))))
      }
      if ("unimportant" %in% head(names(importance), 1L)) {
        return(err("unimportant feature is important"))
      }
    }

    if ("selected_features" %in% learner$properties) {
      selected = learner$selected_features()
      msg = checkmate::check_subset(selected, task$feature_names)
      if (!isTRUE(msg)) {
        return(err(msg))
      }
    }

    if ("oob_error" %in% learner$properties) {
      oob = learner$oob_error()
      msg = checkmate::check_number(oob)
      if (!isTRUE(msg)) {
        return(err(msg))
      }
    }
  }

  stage = "weights"

  if ("weights" %in% learner$properties) {
    use_weights = learner$use_weights
    get_weights = learner$.__enclos_env__$private$.get_weights
    counter = R6::R6Class("counter", public = list(count = 0, count_up = function(x) self$count = self$count + 1))$new()
    get_weights_counter = function(...) {
      counter$count_up()
      get_weights(...)
    }
    on.exit({
      learner$.__enclos_env__$private$.get_weights = get_weights
      lockBinding(".get_weights", env = learner$.__enclos_env__$private)
      learner$use_weights = use_weights
    })

    unlockBinding(".get_weights", env = learner$.__enclos_env__$private)
    learner$.__enclos_env__$private$.get_weights = get_weights_counter

    learner$use_weights = "use"
    suppressWarnings(learner$train(task))
    if (counter$count == 0) {
      return(err("get_weights was not called"))
    }

    learner$use_weights = "ignore"
    suppressWarnings(learner$train(task))
    if (counter$count == 0) {
      return(err("get_weights was not called. It should be called even when use_weights = 'ignore'"))
    }

    learner$.__enclos_env__$private$.get_weights = get_weights
    lockBinding(".get_weights", env = learner$.__enclos_env__$private)
    learner$use_weights = use_weights
    on.exit()
  }

  # check predict_newdata_fast
  stage = "predict_newdata_fast()"

  if (startsWith(task$id, "feat_all") && exists("predict_newdata_fast", learner) && learner$predict_type == "prob") {
    prediction = suppressWarnings(try(learner$predict_newdata_fast(task$data()), silent = TRUE))

    if (inherits(prediction, "try-error")) {
      ok = prediction
      prediction = NULL
      return(err(error_as_character(ok)))
    }

    msg = checkmate::check_list(prediction)
    if (!isTRUE(msg)) {
      return(err("does not return a list"))
    }

    msg = checkmate::check_names(
      names(prediction),
      subset.of = c(
        mlr3::mlr_reflections$learner_predict_types[[learner$task_type]][[learner$predict_type]],
        "extra",
        "raw"
      )
    )
    if (!isTRUE(msg)) {
      return(err("Names of returned list do not match learner predict_types: %s", str_collapse(names(prediction))))
    }
  }

  # check marshalling
  stage = "marshal"

  if (startsWith(task$id, "feat_all") && "marshal" %in% learner$properties) {
    learner$marshal()
    # learner$marshaled checks if model is of class "marshaled"
    if (!learner$marshaled) {
      return(err("model not marshaled"))
    }

    learner$unmarshal()
    # checks if the "marshaled" class is removed
    if (learner$marshaled) {
      return(err("model not unmarshaled"))
    }

    prediction_marshaling = suppressWarnings(try(learner$predict(task), silent = TRUE))
    if (inherits(prediction_marshaling, "try-error")) {
      ok = prediction_marshaling
      prediction_marshaling = NULL
      return(err(error_as_character(ok)))
    }
  }

  # check encapsulation
  stage = "encapsulation"
  if (
    startsWith(task$id, "feat_all") && !(learner$task_type == "surv" && learner$predict_type %in% c("lp", "response"))
  ) {
    learner_encapsulated = learner$clone(deep = TRUE)
    learner_encapsulated$encapsulate("mirai", default_fallback(learner_encapsulated))

    rr = resample(task, learner_encapsulated, rsmp("holdout"), store_models = TRUE)
    log = rr$learners[[1]]$state$log
    if ("error" %in% log$class) {
      conditions = log[class == "error"]$condition
      return(err(
        "resample log has errors: %s",
        mlr3misc::str_collapse(mlr3misc::map_chr(conditions, conditionMessage))
      ))
    }
  }

  return(list(
    ok = TRUE,
    learner = learner,
    prediction = prediction,
    score = score,
    prediction_marshaling = prediction_marshaling,
    learner_encapsulated = learner_encapsulated,
    error = character(),
    seed = seed
  ))
}

#' @title Run Autotest for a Learner
#'
#' @description
#' Runs a series of experiments with a given learner on multiple tasks.
#'
#' @param learner ([Learner])\cr
#'  The learner to test.
#' @param N (`integer(1)`)\cr
#'  Number of rows of generated tasks.
#' @param exclude (`character()`)\cr
#'  Regular expression to exclude tasks from the test.
#'  Run `generate_tasks(learner)` to see all available tasks.
#' @param predict_types (`character()`)\cr
#'  Predict types to test.
#' @param check_replicable (`logical(1)`)\cr
#'  Check if the results are replicable.
#' @param configure_learner (`function(learner, task)`)\cr
#'  Function to configure the learner before training.
#'  Useful when learner settings need to be adjusted for a specific task.
#'
#' @return If the test was successful, `TRUE` is returned.
#' If the test failed, a `list` with the following elements is returned:
#'  - `ok` (`logical(1)`): Success flag.
#'  - `seed` (`integer(1)`): Seed used for the experiment.
#'  - `task` ([Task]): Task used for the experiment.
#'  - `learner` ([Learner]): Learner used for the experiment.
#'  - `prediction` ([Prediction]): Prediction object.
#'  - `score` (`numeric(1)`): Score of the prediction.
#'  - `error` (`character()`): Error message if `ok` is `FALSE`.
#
#' @noRd
run_autotest = function(
  learner,
  N = 30L,
  exclude = NULL,
  predict_types = learner$predict_types,
  check_replicable = TRUE,
  configure_learner = NULL
) {
  # nolint
  if (!is.null(configure_learner)) {
    checkmate::assert_function(configure_learner, args = c("learner", "task"))
  }
  learner = learner$clone(deep = TRUE)
  id = learner$id
  tasks = generate_tasks(learner, N = N)

  if (!is.null(exclude)) {
    tasks = tasks[!grepl(exclude, names(tasks))]
  }

  sanity_runs = list()
  make_err = function(msg, ...) {
    run$ok = FALSE
    run$error = sprintf(msg, ...)
    run
  }

  for (task in tasks) {
    for (predict_type in predict_types) {
      learner$id = sprintf("%s:%s", id, predict_type)
      learner$predict_type = predict_type

      if (predict_type == "quantiles") {
        learner$quantiles = 0.5
      }

      run = run_experiment(task, learner, NULL, configure_learner)
      if (!run$ok) {
        return(run)
      }

      # re-run task with same seed for feat_all
      if (startsWith(task$id, "feat_all")) {
        repeated_run = run_experiment(task, learner, seed = run$seed, configure_learner)

        if (!repeated_run$ok) {
          return(repeated_run)
        }

        if (
          check_replicable && !isTRUE(all.equal(as.data.table(run$prediction), as.data.table(repeated_run$prediction)))
        ) {
          return(make_err("Different results for replicated runs using fixed seed %i", run$seed))
        }
      }

      if (task$task_type == "classif" && task$id == "sanity") {
        sanity_runs[[predict_type]] = run
      }
    }
    if (task$task_type == "classif" && length(sanity_runs) > 1L) {
      responses = lapply(sanity_runs, function(r) r$prediction$response)
      if (!isTRUE(Reduce(all.equal, responses))) {
        return(make_err("Response is different for different predict types"))
      }
    }
  }

  return(TRUE)
}

#' @title Check Parameters of mlr3 Learners
#'
#' @description
#' Checks parameters of mlr3learners against parameters defined in the upstream functions of the respective learners.
#'
#' @details
#' Some learners do not have all of their parameters stored within the learner function
#' that is called within `.train()`.
#' Sometimes learners come with a "control" function, e.g. [glmnet::glmnet.control()].
#' Such need to be checked as well since they make up the full ParamSet of the respective learner.
#'
#' To work nicely with the defined ParamSet, certain parameters need to be excluded
#' because these are only present in either the "control" object or the actual top-level function call.
#' Such exclusions should go into argument `exclude` with a comment for the reason of the exclusion.
#' See examples for more information.
#'
#' @param learner ([mlr3::Learner])\cr
#'   The constructed learner.
#' @param fun (`function`)\cr
#'   The function of the upstream package for which parameters should
#'   be checked. E.g. `mboost::boost_control`.
#' @exclude (`character()`)\cr
#'   Parameters which should be excluded for this particular check. See details.
#' @examples
#' test_that("classif.gamboost", {
#'   learner = lrn("classif.gamboost")
#'   fun = mboost::gamboost
#'   exclude = c(
#'     "formula", # .train
#'     "data", # .train
#'     "na.action", # Only na.omit and na.fail available
#'     "weights", # .train
#'     "control" # mboost::boost_control
#'   )
#'
#'   run_paramtest(learner, fun, exclude)
#'   expect_true(result, info = result$error)
#' })
run_paramtest = function(learner, fun, exclude = character(), tag = NULL) {
  par_learner = learner$param_set$ids(tags = tag)
  if (checkmate::test_list(fun)) {
    # for xgboost we pass a character vector with info scraped from the web
    if (mlr3misc::some(fun, function(x) class(x) == "character")) {
      which = which(mlr3misc::map_lgl(fun, function(x) class(x) == "character"))
      par_package = fun[[which]]
      fun[[which]] = NULL
      other = unlist(lapply(fun, formalArgs))
      par_package = append(par_package, other)
    } else {
      par_package = unlist(lapply(fun, formalArgs))
    }
  } else {
    par_package = formalArgs(fun)
  }

  missing = setdiff(par_package, par_learner)
  missing = setdiff(missing, c(exclude, "..."))

  extra = setdiff(par_learner, par_package)
  extra = setdiff(extra, c(exclude, "..."))

  if (length(c(missing, extra)) == 0L) {
    return(TRUE)
  }

  merror = eerror = character(0)

  if (length(missing) > 0) {
    merror = sprintf("Missing parameters for learner '%s': %s", learner$id, paste0(missing, collapse = ", "))
  }

  if (length(extra) > 0) {
    eerror = sprintf("Extra parameters for learner '%s': %s", learner$id, paste0(extra, collapse = ", "))
  }

  error = paste(merror, eerror, sep = "\n")

  list(ok = FALSE, error = error, missing = missing, extra = extra)
}

# Helper function to convert a vector of probabilities to a matrix
#
# sometimes useful in tests, e.g., mlr3learners.partykit::LearnerClassifMob
# uses this in its tests to set up its custom prediction function for a mob
# version of a logit model
prob_vector_to_matrix = function(p, levs) {
  stopifnot(is.numeric(p))
  y = matrix(c(1 - p, p), ncol = 2L, nrow = length(p))
  colnames(y) = levs
  y
}

suppress_fallback_warnings = function(expr) {
  withCallingHandlers(
    {
      expr
    },
    warning = function(w) {
      if (inherits(w, "Mlr3WarningConfigFallbackPredictType") || inherits(w, "Mlr3WarningConfigFallbackProperties")) {
        invokeRestart("muffleWarning")
      }
    }
  )
}

expect_man_exists = function(man) {
  checkmate::expect_string(man, na.ok = TRUE, fixed = "::")
  if (!is.na(man)) {
    parts = strsplit(man, "::", fixed = TRUE)[[1L]]
    matches = help.search(parts[2L], package = parts[1L], ignore.case = FALSE)
    checkmate::expect_data_frame(matches$matches, min.rows = 1L, info = "man page lookup")
  }
}

expect_same_address = function(x, y) {
  requireNamespace("data.table")
  testthat::expect_identical(data.table::address(x), data.table::address(y))
}

expect_different_address = function(x, y) {
  requireNamespace("data.table")
  testthat::expect_false(identical(data.table::address(x), data.table::address(y)))
}

expect_id = function(x, len = NULL) {
  checkmate::expect_character(x, len = len, any.missing = FALSE, min.chars = 1L, unique = FALSE)
}

expect_hash = function(x, len = NULL) {
  pattern = "^[0-9a-z]{16}$"
  checkmate::expect_character(x, len = len, any.missing = FALSE, pattern = pattern, unique = TRUE)
}

expect_uhash = function(x, len = NULL) {
  pattern = "^[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}$"
  checkmate::expect_character(x, len = len, any.missing = FALSE, pattern = pattern, unique = TRUE)
}

expect_dictionary = function(d, contains = NA_character_, min_items = 0L) {
  checkmate::expect_r6(d, "Dictionary")
  testthat::expect_output(print(d), "Dictionary")
  keys = d$keys()

  checkmate::expect_environment(d$items)
  checkmate::expect_character(keys, any.missing = FALSE, min.len = min_items, min.chars = 1L)
  if (!is.na(contains)) {
    checkmate::expect_list(d$mget(keys), types = contains, names = "unique")
  }
  if (length(keys) >= 1L) {
    testthat::expect_error(d$get(keys[1], 1), "names")
  }
  checkmate::expect_data_table(data.table::as.data.table(d), key = "key", nrows = length(keys))
}

expect_backend = function(b) {
  checkmate::expect_r6(
    b,
    cloneable = FALSE,
    public = c("nrow", "ncol", "colnames", "rownames", "head", "data", "hash"),
    private = c(".data", ".hash", ".calculate_hash")
  )
  testthat::expect_output(print(b), "DataBackend")

  n = checkmate::expect_count(b$nrow)
  p = checkmate::expect_count(b$ncol)
  cn = checkmate::expect_character(b$colnames, any.missing = FALSE, len = p, unique = TRUE)
  rn = checkmate::expect_integerish(b$rownames, any.missing = FALSE, len = n, unique = TRUE)
  rn1 = head(rn, 1L)
  pk = b$primary_key

  x = b$data(rows = rn, cols = pk)
  checkmate::expect_data_table(x, ncols = 1L, nrows = n, col.names = "unique")
  x = x[[1L]]
  checkmate::expect_integerish(x, len = n, unique = TRUE)

  x = b$data(rows = rn, cols = setdiff(cn, pk)[1L])
  checkmate::expect_data_table(x, ncols = 1L, nrows = n, col.names = "unique")
  x = x[[1L]]
  checkmate::expect_atomic_vector(x, len = n)

  # extra cols are ignored
  x = b$data(rows = rn1, cols = c(cn[1L], "_not_existing_"))
  checkmate::expect_data_table(x, nrows = length(rn1), ncols = 1L)

  # zero cols matching
  x = b$data(rows = rn1, cols = "_not_existing_")
  checkmate::expect_data_table(x, nrows = 0L, ncols = 0L)

  # extra rows are ignored
  query_rows = c(rn1, if (is.numeric(rn)) b$nrow + 1L else "_not_existing_")
  x = b$data(query_rows, cols = cn[1L])
  checkmate::expect_data_table(x, nrows = length(rn1), ncols = 1L)

  # zero rows matching
  query_rows = if (is.numeric(rn)) b$nrow + 1L else "_not_existing_"
  x = b$data(rows = query_rows, cols = cn[1L])
  checkmate::expect_data_table(x, nrows = 0L, ncols = 1L)

  # rows are duplicated
  x = b$data(rows = rep(rn1, 2L), cols = b$colnames)
  checkmate::expect_data_table(x, nrows = 2L * length(rn1), ncols = p)

  # cols are returned in the right order
  j = rev(cn)
  x = b$data(rows = rn1, cols = j)
  testthat::expect_equal(j, colnames(x))

  # rows are returned in the right order
  i = sample(rn, min(n, 10L))
  x = b$data(rows = i, cols = b$primary_key)
  testthat::expect_equal(i, x[[1L]])

  # duplicated cols raise exception
  testthat::expect_error(b$data(rows = rn1, cols = rep(cn[1L], 2L)), "unique")

  # $head()
  checkmate::expect_data_table(b$head(.Machine$integer.max), nrows = n, ncols = p)
  checkmate::expect_data_table(b$head(0L), nrows = 0, ncols = p)
  checkmate::expect_data_table(b$head(2L), nrows = min(2L, b$nrow), ncols = p)

  # $distinct()
  d = b$distinct(rn, b$primary_key)[[1L]]
  checkmate::expect_integerish(d, any.missing = FALSE, len = n, unique = TRUE)
  checkmate::expect_list(b$distinct(rn, "_not_existing_"), len = 0L, names = "named")
  d = b$distinct(rn, c("_not_existing_", rev(cn), "_also_not_existing_"))
  checkmate::expect_list(d, names = "unique")
  checkmate::expect_names(names(d), identical.to = rev(cn))

  d = b$distinct(rn1, cn)
  checkmate::expect_list(d, len = length(cn), names = "unique", any.missing = FALSE)
  testthat::expect_true(all(lengths(d) <= 1L)) # NA -> 0 zero length

  ## missings are handled by distinct?
  d = b$distinct(rn, cn, na_rm = TRUE)
  checkmate::qexpectr(d, "V")

  d = b$distinct(rn, cn, na_rm = FALSE)
  m = b$missings(rn, cn)
  testthat::expect_equal(vapply(d, checkmate::anyMissing, FUN.VALUE = logical(1)), m > 0L)

  # $missings()
  x = b$missings(b$rownames, b$colnames)
  checkmate::expect_integerish(x, lower = 0L, upper = b$nrow, any.missing = FALSE)
  checkmate::expect_names(names(x), permutation.of = b$colnames)
  checkmate::expect_integerish(b$missings(b$rownames, "_not_existing_"), len = 0L, names = "named")
  checkmate::expect_integerish(b$missings(b$rownames[0L], b$colnames), len = b$ncol, names = "unique")
  checkmate::expect_integerish(b$missings(b$rownames[0L], "_not_existing_"), len = 0L, names = "unique")

  # $hash
  checkmate::expect_string(b$hash)

  # col_hashes
  col_hashes = b$col_hashes
  checkmate::expect_character(col_hashes, any.missing = FALSE)
  checkmate::expect_names(names(col_hashes), permutation.of = setdiff(b$colnames, b$primary_key))

  # if multiple cols have the same hash, check that they actually contain the same data.
  realhashes = sapply(b$data(rows = b$rownames, cols = setdiff(b$colnames, b$primary_key)), mlr3misc::calculate_hash)
  testthat::expect_equal(
    unname(sapply(split(realhashes, col_hashes), data.table::uniqueN)),
    rep(1, data.table::uniqueN(col_hashes))
  )
}

expect_iris_backend = function(b, n_missing = 0L) {
  testthat::expect_equal(b$nrow, 150L)
  testthat::expect_equal(b$ncol, 6L)
  checkmate::expect_set_equal(b$colnames, c(names(iris), b$primary_key))
  checkmate::expect_set_equal(b$rownames, seq_len(150L))

  x = b$head(2)
  checkmate::expect_data_table(x, nrows = 2L, ncols = 6L, any.missing = FALSE)
  checkmate::expect_names(names(x), permutation.of = c(names(iris), b$primary_key))

  x = b$distinct(b$rownames, "Species")
  checkmate::expect_list(x, "character", len = 1)
  checkmate::expect_set_equal(x$Species, levels(iris$Species))

  x = b$data(rows = 2:10, cols = c(b$primary_key, "Species", "Sepal.Width"))
  checkmate::expect_data_table(x, nrows = 9L, ncols = 3L)
  checkmate::expect_names(names(x), identical.to = c(b$primary_key, "Species", "Sepal.Width"))
  checkmate::expect_set_equal(x[[b$primary_key]], 2:10)

  if (is.factor(x$Species)) {
    checkmate::expect_factor(x$Species, levels = levels(iris$Species))
  } else {
    checkmate::expect_character(x$Species, any.missing = FALSE)
  }

  testthat::expect_true(all(x$Species == "setosa"))
  checkmate::expect_numeric(x$Sepal.Width, any.missing = FALSE)

  # invalid row ids
  x = b$data(rows = 150:151, cols = c(b$primary_key, "Species"))
  checkmate::expect_data_table(x, nrows = 1L, ncols = 2L)
  testthat::expect_equal(x[[b$primary_key]], 150L)
  testthat::expect_equal(as.character(x[["Species"]]), "virginica")

  # duplicated row ids
  x = b$data(rows = c(1L, 1L), cols = c(b$primary_key, "Species"))
  checkmate::expect_data_table(x, nrows = 2L, ncols = 2L)
  testthat::expect_equal(x[[b$primary_key]], c(1L, 1L))
  testthat::expect_equal(as.character(x[["Species"]]), c("setosa", "setosa"))

  # no missing values
  x = b$missings(b$rownames, b$colnames)
  checkmate::expect_integerish(x, names = "unique", lower = 0L, upper = b$nrow, any.missing = FALSE)
  checkmate::expect_names(names(x), permutation.of = b$colnames)
  testthat::expect_equal(sum(x), n_missing)
}

expect_task = function(task, null_backend_ok = TRUE, duplicated_ids = FALSE) {
  checkmate::expect_r6(
    task,
    "Task",
    cloneable = TRUE,
    public = c(
      "id",
      "backend",
      "task_type",
      "row_roles",
      "col_roles",
      "col_info",
      "head",
      "row_ids",
      "feature_names",
      "target_names",
      "formula",
      "nrow",
      "ncol",
      "feature_types"
    )
  )
  testthat::expect_output(print(task), "Task")
  expect_id(task$id)
  checkmate::expect_string(task$label, na.ok = TRUE)
  expect_man_exists(task$man)
  checkmate::expect_count(task$nrow)
  checkmate::expect_count(task$ncol)

  null_backend = is.null(task$backend)
  if (!null_backend_ok) {
    testthat::expect_false(is.null(task$backend))
  }

  if (task$nrow > 0L && !null_backend) {
    checkmate::expect_data_table(head(task, 1), nrows = 1L)
  }

  cols = c("id", "type", "levels", "label", "fix_factor_levels")
  checkmate::expect_data_table(task$col_info, key = "id", ncols = length(cols))
  checkmate::expect_names(names(task$col_info), permutation.of = cols)
  expect_id(task$col_info$id)
  checkmate::expect_subset(task$col_info$type, mlr3::mlr_reflections$task_feature_types)
  checkmate::expect_list(task$col_info$levels)
  checkmate::expect_character(task$col_info$label)
  checkmate::expect_logical(task$col_info$fix_factor_levels)

  checkmate::expect_list(task$col_roles, names = "unique", any.missing = FALSE)
  checkmate::expect_names(
    names(task$col_roles),
    permutation.of = mlr3::mlr_reflections$task_col_roles[[task$task_type]]
  )
  lapply(task$col_roles, checkmate::expect_character, any.missing = FALSE, unique = TRUE, min.chars = 1L)
  checkmate::expect_subset(unlist(task$col_roles, use.names = FALSE), task$col_info$id)

  checkmate::expect_list(
    task$row_roles,
    names = "unique",
    types = c("integer", "character", "numeric"),
    any.missing = FALSE
  )
  checkmate::expect_names(names(task$row_roles), permutation.of = mlr3::mlr_reflections$task_row_roles)
  lapply(task$row_roles, checkmate::expect_integerish, any.missing = FALSE, unique = !duplicated_ids)

  types = task$feature_types
  checkmate::expect_data_table(types, ncols = 2, nrows = length(task$feature_names), key = "id")
  checkmate::expect_set_equal(types$id, task$feature_names)
  checkmate::expect_subset(types$type, mlr3::mlr_reflections$task_feature_types)

  properties = task$properties
  checkmate::expect_subset(properties, mlr3::mlr_reflections$task_properties[[task$task_type]])

  levels = task$levels()
  checkmate::expect_list(levels, names = "unique")
  checkmate::qassertr(levels, c("0", "S+"))

  expect_hash(task$hash, 1L)

  if (!null_backend) {
    missings = task$missings()
    checkmate::expect_integer(missings, names = "unique", any.missing = FALSE, lower = 0L, upper = task$nrow)

    missings = task$missings(character())
    checkmate::expect_integer(missings, len = 0L)
    testthat::expect_named(missings)

    # query zero columns
    data = task$data(cols = character())
    checkmate::expect_data_table(data, ncols = 0L)

    # query zero rows
    data = task$data(rows = task$row_ids[0L])
    checkmate::expect_data_table(data, nrows = 0L)
  }

  # col_hashes
  col_hashes = task$col_hashes
  checkmate::expect_character(col_hashes, any.missing = FALSE)
  checkmate::expect_names(names(col_hashes), permutation.of = setdiff(unlist(task$col_roles), task$backend$primary_key))

  # if multiple cols have the same hash, check that they actually contain the same data.
  if (!null_backend) {
    realhashes = sapply(task$data(cols = names(col_hashes)), mlr3misc::calculate_hash)
    testthat::expect_equal(
      unname(sapply(split(realhashes, col_hashes), data.table::uniqueN)),
      rep(1, data.table::uniqueN(col_hashes))
    )
  }
}

expect_task_supervised = function(task) {
  checkmate::expect_r6(task, "TaskSupervised", cloneable = TRUE)
  checkmate::expect_subset(task$target_names, task$col_info$id, empty.ok = FALSE)

  f = task$formula()
  checkmate::expect_class(f, "formula")
  # tf = terms(f)
  # checkmate::expect_set_equal(labels(tf), task$feature_names) # rhs
  # checkmate::expect_set_equal(setdiff(all.vars(tf), labels(tf)), task$target_names) # lhs
  checkmate::expect_subset(task$feature_names, colnames(head(task)))
  expect_hash(task$hash, 1L)
}

expect_task_classif = function(task) {
  checkmate::expect_r6(task, "TaskClassif")
  y = task$truth()
  checkmate::expect_factor(y)

  testthat::expect_gte(length(task$class_names), 2L)
  checkmate::expect_character(task$class_names, any.missing = FALSE)
  checkmate::expect_subset(task$class_names, levels(y))
  if (length(task$class_names) > 2L) {
    testthat::expect_identical(task$positive, NA_character_)
    testthat::expect_identical(task$negative, NA_character_)
  } else {
    checkmate::expect_set_equal(c(task$positive, task$negative), task$class_names)
    testthat::expect_true(task$positive != task$negative)
  }
  expect_hash(task$hash, 1L)
}

is_special_learner = function(x) {
  inherits(x, "GraphLearner") || inherits(x, "AutoTuner") || inherits(x, "AutoFSelector")
}

expect_task_regr = function(task) {
  checkmate::expect_r6(task, "TaskRegr")
  y = task$truth()
  checkmate::expect_numeric(y, any.missing = FALSE)
  expect_hash(task$hash, 1L)
}

expect_task_unsupervised = function(task) {
  checkmate::expect_r6(task, "TaskUnsupervised")
  expect_hash(task$hash, 1L)
}

expect_task_generator = function(gen) {
  checkmate::expect_r6(gen, "TaskGenerator", private = ".generate")
  expect_id(gen$id)
  checkmate::expect_string(gen$label, na.ok = TRUE)
  expect_man_exists(gen$man)
  checkmate::expect_choice(gen$task_type, mlr3::mlr_reflections$task_types$type)
  checkmate::expect_function(gen$generate, args = "n")
  checkmate::expect_class(gen$param_set, "ParamSet")
  checkmate::expect_list(gen$param_set$values, names = "unique")
  testthat::expect_output(print(gen))
}


expect_learner = function(lrn, task = NULL, check_man = TRUE) {
  checkmate::expect_r6(lrn, "Learner", cloneable = TRUE)
  expect_id(lrn$id)
  checkmate::expect_string(lrn$label, na.ok = TRUE)
  if (check_man) {
    expect_man_exists(lrn$man)
  }
  testthat::expect_output(print(lrn))

  checkmate::expect_choice(lrn$task_type, mlr3::mlr_reflections$task_types$type)
  checkmate::expect_character(lrn$packages, any.missing = FALSE, min.chars = 1L, unique = TRUE)
  checkmate::expect_class(lrn$param_set, "ParamSet")
  testthat::expect_lte(length(lrn$param_set$ids(tags = "threads")), 1L)
  checkmate::expect_character(lrn$properties, any.missing = FALSE, min.chars = 1L, unique = TRUE)

  checkmate::expect_character(lrn$predict_types, any.missing = FALSE, min.chars = 1L, unique = TRUE)
  checkmate::expect_choice(lrn$predict_type, lrn$predict_types)
  checkmate::expect_subset(lrn$feature_types, mlr3::mlr_reflections$task_feature_types, empty.ok = FALSE)

  expect_hash(lrn$hash)
  expect_hash(lrn$phash)

  checkmate::expect_function(mlr3misc::get_private(lrn)$.train, args = "task", nargs = 1L)
  checkmate::expect_function(mlr3misc::get_private(lrn)$.predict, args = "task", nargs = 1L)
  expect_hash(lrn$hash, 1L)

  tags = lrn$param_set$tags
  tags = Filter(function(tags) !any(c("train", "predict") %in% tags), tags)
  testthat::expect_true(
    length(tags) == 0L,
    info = sprintf(
      "All hyperparameters of learner %s must be tagged with 'train' or 'predict'. Missing tags for: %s",
      lrn$id,
      paste0(names(tags), collapse = ", ")
    )
  )

  # FIXME: remove at and glrn when they have new releases supporting marshaling
  if (
    "marshal" %in%
      lrn$properties &&
      !inherits(lrn, "GraphLearner") &&
      !inherits(lrn, "AutoTuner") &&
      !inherits(lrn, "AutoFSelector")
  ) {
    checkmate::expect_function(lrn$marshal)
    checkmate::expect_function(lrn$unmarshal)
  }

  if (!is.null(task)) {
    checkmate::expect_class(task, "Task")
    checkmate::expect_subset(lrn$properties, mlr3::mlr_reflections$learner_properties[[task$task_type]])
    testthat::expect_identical(lrn$task_type, task$task_type)

    # FIXME: remove at and glrn when they have new releases supporting marshaling
    if (
      "marshal" %in%
        lrn$properties &&
        !inherits(lrn, "GraphLearner") &&
        !inherits(lrn, "AutoTuner") &&
        !inherits(lrn, "AutoFSelector")
    ) {
      expect_marshalable_learner(lrn, task)
    }
  }

  if (!inherits(lrn, "GraphLearner") && !inherits(lrn, "AutoTuner")) {
    # still not in pipelines, breaking check in mlr3tuning
    checkmate::expect_class(lrn$base_learner(), "Learner")
  }

  if ("validation" %in% lrn$properties && !checkmate::test_class(lrn, "GraphLearner")) {
    testthat::expect_true(exists("validate", lrn))
    testthat::expect_true(exists("internal_valid_scores", envir = lrn))
    checkmate::expect_function(mlr3misc::get_private(lrn)$.extract_internal_valid_scores)
  } else if (!is_special_learner(lrn)) {
    checkmate::assert_false(exists("validate", lrn))
  }
  if ("internal_tuning" %in% lrn$properties && !checkmate::test_class(lrn, "GraphLearner")) {
    any_internal_tuning = FALSE
    for (tags in lrn$param_set$tags) {
      if ("internal_tuning" %in% tags) {
        any_internal_tuning = TRUE
        break
      }
    }
    if (!any_internal_tuning) {
      stopf("at least one parameter must support internal tuning when the learner is tagged as such")
    }
    testthat::expect_true(exists("internal_tuned_values", envir = lrn))
    checkmate::expect_function(mlr3misc::get_private(lrn)$.extract_internal_tuned_values)
  }
}

expect_marshalable_learner = function(learner, task) {
  testthat::expect_true("marshal" %in% learner$properties)
  learner$state = NULL

  has_public = function(learner, x) {
    exists(x, learner, inherits = FALSE)
  }

  testthat::expect_true(has_public(learner, "marshal") && checkmate::test_function(learner$marshal, nargs = 0))
  testthat::expect_true(has_public(learner, "unmarshal") && checkmate::test_function(learner$unmarshal, nargs = 0))
  testthat::expect_true(has_public(learner, "marshaled"))

  testthat::expect_equal(learner$marshaled, FALSE)

  learner$train(task)
  model = learner$model
  class_prev = class(model)
  testthat::expect_false(learner$marshaled)
  testthat::expect_equal(mlr3::is_marshaled_model(learner$model), learner$marshaled)
  testthat::expect_invisible(learner$marshal())
  if (!inherits(learner, "GraphLearner")) {
    testthat::expect_true(learner$marshaled)
  }
  testthat::expect_equal(mlr3::is_marshaled_model(learner$model), learner$marshaled)

  # unmarshaling works
  testthat::expect_invisible(learner$unmarshal())
  # can predict after unmarshaling
  expect_prediction(learner$predict(task))
  # model is reset
  testthat::expect_equal(learner$model, model)
  # marshaled is set accordingly
  testthat::expect_false(learner$marshaled)
  testthat::expect_equal(class(learner$model), class_prev)
}

expect_resampling = function(r, task = NULL, strata = TRUE) {
  checkmate::expect_r6(r, "Resampling")
  testthat::expect_output(print(r), "Resampling")
  expect_id(r$id)
  expect_man_exists(r$man)

  instance = r$instance
  if (is.null(instance)) {
    testthat::expect_false(r$is_instantiated)
    testthat::expect_error(r$train_set(1L), "instantiated")
    testthat::expect_error(r$test_set(1L), "instantiated")
    # testthat::expect_identical(r$hash, NA_character_)
    if (r$id %in% c("custom", "custom_cv", "loo")) {
      checkmate::expect_count(r$iters, na.ok = TRUE)
      testthat::expect_true(is.na(r$iters))
    }
    testthat::expect_identical(r$task_hash, NA_character_)
  } else {
    testthat::expect_true(r$is_instantiated)
    expect_hash(r$hash, 1L)
    expect_hash(r$task_hash, 1L)
    expect_hash(r$task_row_hash, 1L)
    if (!is.null(task)) {
      ids = task$row_ids
    }
    checkmate::expect_count(r$iters, positive = TRUE)

    for (i in seq_len(r$iters)) {
      train = r$train_set(1L)
      test = r$test_set(1L)
      checkmate::expect_integerish(train, any.missing = FALSE)
      checkmate::expect_integerish(test, any.missing = FALSE)
      if (!inherits(r, "ResamplingCustom") && !inherits(r, "ResamplingInsample")) {
        testthat::expect_length(intersect(train, test), 0L)
      }
      if (!is.null(task)) {
        checkmate::expect_subset(train, ids)
        checkmate::expect_subset(test, ids)
      }
    }
  }

  checkmate::expect_list(r$param_set$values, names = "unique")
  testthat::expect_true(checkmate::qtestr(r$param_set$values, "V1"))

  # check re-instantiation with provided task
  if (
    !is.null(task) && !is.null(task$backend) && !inherits(r, "ResamplingCustom") && !inherits(r, "ResamplingCustomCV")
  ) {
    r = r$clone()$instantiate(task)
    checkmate::expect_subset(r$train_set(1), task$row_ids)
    checkmate::expect_subset(r$test_set(1), task$row_ids)

    # again with strata
    if (strata) {
      task = task$clone()
      task$col_roles$stratum = task$target_names
      r$instantiate(task)
      checkmate::expect_subset(r$train_set(1), task$row_ids)
      checkmate::expect_subset(r$test_set(1), task$row_ids)
    }
  }
}

expect_measure = function(m) {
  checkmate::expect_r6(
    m,
    "Measure",
    public = c("aggregate", "score", "id", "minimize", "packages", "range", "task_type", "task_properties")
  )
  expect_id(m$id)
  expect_man_exists(m$man)
  testthat::expect_output(print(m), "Measure")

  if ("requires_no_prediction" %in% m$properties) {
    testthat::expect_null(m$predict_sets)
  }

  expect_id(m$id)
  checkmate::expect_subset(m$task_type, c(NA_character_, mlr3::mlr_reflections$task_types$type), empty.ok = FALSE)
  checkmate::expect_numeric(m$range, len = 2, any.missing = FALSE)
  testthat::expect_lt(m$range[1], m$range[2])
  checkmate::expect_flag(m$minimize, na.ok = TRUE)
  checkmate::expect_character(m$packages, min.chars = 1L, any.missing = FALSE, unique = TRUE)
  if (is.null(mlr3misc::get_private(m)$.score)) {
    checkmate::expect_function(m$score_internal, args = c("prediction", "..."))
  } else {
    checkmate::expect_function(mlr3misc::get_private(m)$.score, args = c("prediction", "..."))
  }
  checkmate::expect_function(m$aggregate, args = "rr")
}

expect_prediction = function(p) {
  checkmate::expect_r6(p, "Prediction", public = c("row_ids", "truth", "predict_types"))
  testthat::expect_output(print(p), "Prediction")
  checkmate::expect_data_table(data.table::as.data.table(p), nrows = length(p$row_ids))
  checkmate::expect_integerish(p$missing)
}

expect_prediction_regr = function(p) {
  expect_prediction(p)
  checkmate::expect_r6(p, "PredictionRegr", public = c("row_ids", "response", "truth", "predict_types", "se", "distr"))
  checkmate::expect_numeric(p$truth, any.missing = TRUE, len = length(p$row_ids), null.ok = TRUE)
  checkmate::expect_numeric(p$response, any.missing = FALSE, len = length(p$row_ids), null.ok = TRUE)
  if ("se" %in% p$predict_types) {
    checkmate::expect_numeric(p$se, any.missing = FALSE, len = length(p$row_ids), lower = 0)
  }
  if ("distr" %in% p$predict_types) {
    checkmate::expect_class(p$distr, "VectorDistribution")
  }
}

expect_prediction_classif = function(p, task = NULL) {
  expect_prediction(p)
  checkmate::expect_r6(p, "PredictionClassif", public = c("row_ids", "response", "truth", "predict_types", "prob"))
  n = length(p$row_ids)
  lvls = if (is.null(task)) levels(p$truth) else task$class_names
  checkmate::expect_factor(p$truth, len = n, levels = lvls, null.ok = TRUE)
  checkmate::expect_factor(p$response, len = n, levels = lvls, null.ok = TRUE)
  testthat::expect_identical(levels(p$truth), lvls)
  testthat::expect_identical(levels(p$response), lvls)
  if ("prob" %in% p$predict_types) {
    checkmate::expect_matrix(p$prob, "numeric", any.missing = FALSE, ncols = nlevels(p$response), nrows = n)
    testthat::expect_identical(colnames(p$prob), lvls)
  }
  confusion = p$confusion
  checkmate::expect_matrix(confusion)
  checkmate::expect_integer(confusion, lower = 0L, any.missing = FALSE)
  if (!is.null(task)) {
    checkmate::expect_names(rownames(confusion), identical.to = task$class_names)
    checkmate::expect_names(colnames(confusion), identical.to = task$class_names)
  }
}

expect_resample_result = function(rr, allow_incomplete = FALSE) {
  checkmate::expect_r6(rr, "ResampleResult")
  expect_resultdata(mlr3misc::get_private(rr)$.data, FALSE)
  testthat::expect_output(print(rr), "ResampleResult")
  nr = mlr3misc::get_private(rr)$.data$iterations()

  if (nr > 0L) {
    expect_task(rr$task, null_backend_ok = is.null(rr$task$backend))
    expect_learner(rr$learner)
    lapply(rr$learners, expect_learner, task = rr$task)
    expect_resampling(rr$resampling, task = rr$task)
  }

  expected_iters = if (allow_incomplete || nr == 0L) nr else rr$resampling$iters

  checkmate::expect_data_table(
    rr$score(),
    nrows = expected_iters,
    min.cols = length(mlr3::mlr_reflections$rr_names),
    any.missing = FALSE
  )
  checkmate::expect_names(names(rr$score()), must.include = mlr3::mlr_reflections$rr_names)
  if (nr > 0L) {
    expect_uhash(rr$uhash, 1L)
  } else {
    testthat::expect_identical(rr$uhash, NA_character_)
  }

  if (nr > 0L) {
    m = mlr3::default_measures(rr$task$task_type)[[1L]]
    y = rr$score(m)
    aggr = rr$aggregate(m)
    checkmate::expect_numeric(
      y[[m$id]],
      lower = m$range[1],
      upper = m$range[2],
      any.missing = FALSE,
      label = sprintf("measure %s", m$id)
    )
    checkmate::expect_number(
      aggr[[m$id]],
      lower = m$range[1L],
      upper = m$range[2L],
      label = sprintf("measure %s", m$id)
    )
  }

  checkmate::expect_list(rr$predictions(), types = "Prediction", len = expected_iters)
  if (nr > 0L) {
    expect_prediction(rr$prediction())
  }
}

expect_benchmark_result = function(bmr) {
  checkmate::expect_r6(bmr, "BenchmarkResult", private = ".data")
  expect_resultdata(mlr3misc::get_private(bmr)$.data, TRUE)
  testthat::expect_output(print(bmr), "BenchmarkResult")

  checkmate::expect_names(
    names(as.data.table(bmr)),
    permutation.of = c(mlr3::mlr_reflections$rr_names, "prediction", "uhash", "task_id", "learner_id", "resampling_id")
  )

  tab = bmr$tasks
  checkmate::expect_data_table(tab, ncols = 3L)
  checkmate::expect_names(names(tab), identical.to = c("task_hash", "task_id", "task"))
  expect_hash(tab$task_hash)
  expect_id(tab$task_id)
  checkmate::expect_list(tab$task, "Task")
  checkmate::expect_set_equal(bmr$tasks$task_hash, mlr3misc::get_private(bmr)$.data$data$tasks$task_hash)

  tab = bmr$learners
  checkmate::expect_data_table(tab, ncols = 3L)
  checkmate::expect_names(names(tab), identical.to = c("learner_hash", "learner_id", "learner"))
  expect_hash(tab$learner_hash)
  expect_id(tab$learner_id)
  checkmate::expect_list(tab$learner, "Learner")
  checkmate::expect_set_equal(
    bmr$learners$learner_hash,
    mlr3misc::get_private(bmr)$.data$data$learner_components$learner_hash
  )

  tab = bmr$resamplings
  checkmate::expect_data_table(tab, ncols = 3L)
  checkmate::expect_names(names(tab), identical.to = c("resampling_hash", "resampling_id", "resampling"))
  expect_hash(tab$resampling_hash)
  expect_id(tab$resampling_id)
  checkmate::expect_list(tab$resampling, "Resampling")
  checkmate::expect_set_equal(
    bmr$resamplings$resampling_hash,
    mlr3misc::get_private(bmr)$.data$data$resamplings$resampling_hash
  )

  if (nrow(mlr3misc::get_private(bmr)$.data$data$fact) > 0L) {
    measures = mlr3::default_measures(bmr$task_type)
  } else {
    measures = mlr3::msrs("time_both")
  }
  tab = bmr$aggregate(measures, ids = TRUE)
  checkmate::expect_data_table(tab, min.cols = 5L + length(measures))
  checkmate::expect_names(
    names(tab),
    must.include = c(
      "nr",
      "resample_result",
      "task_id",
      "learner_id",
      "resampling_id",
      "iters",
      mlr3misc::map_chr(measures, "id")
    )
  )
  testthat::expect_equal(tab$nr, seq_len(nrow(tab)))
  checkmate::expect_list(tab$resample_result, "ResampleResult")
  expect_id(tab$task_id)
  expect_id(tab$learner_id)
  expect_id(tab$resampling_id)
  if (length(measures)) {
    measure = measures[[1L]]
    checkmate::expect_numeric(tab[[measure$id]], any.missing = FALSE)
  }

  tab = bmr$aggregate(measures = measures, params = TRUE)
  checkmate::expect_list(tab$params)

  uhashes = bmr$uhashes
  expect_uhash(uhashes, len = length(mlr3misc::get_private(bmr)$.data$uhashes()))
  checkmate::expect_set_equal(uhashes, unique(mlr3misc::get_private(bmr)$.data$uhashes()))
  testthat::expect_equal(bmr$n_resample_results, length(uhashes))

  tab = bmr$resample_results
  checkmate::expect_data_table(tab, ncols = 3L, nrows = bmr$n_resample_results, any.missing = FALSE)
  checkmate::expect_character(tab$uhash, any.missing = FALSE)
  checkmate::expect_integer(tab$nr, sorted = TRUE, any.missing = FALSE, lower = 1L)
  expect_integer(tab$nr, any.missing = FALSE, lower = 1L)
  checkmate::expect_list(tab$resample_result, types = "ResampleResult")

  ni = mlr3misc::get_private(bmr)$.data$iterations()
  if (ni) {
    checkmate::expect_choice(bmr$task_type, mlr3::mlr_reflections$task_types$type, null.ok = ni == 0L)
  } else {
    testthat::expect_null(bmr$task_type)
  }
}

expect_resultdata = function(rdata, consistency = TRUE) {
  checkmate::expect_class(rdata, "ResultData")
  data = rdata$data

  proto = mlr3:::star_init()
  checkmate::expect_names(names(data), identical.to = names(proto))

  for (nn in names(proto)) {
    checkmate::expect_data_table(data[[nn]], key = data.table::key(proto[[nn]]))
    checkmate::expect_names(names(data[[nn]]), identical.to = names(proto[[nn]]))
  }

  checkmate::expect_character(data$uhashes$uhash, unique = TRUE)

  expect_fsetequal = function(x, y, column) {
    testthat::expect_true(data.table::fsetequal(x[, column, with = FALSE], y[, column, with = FALSE], all = FALSE))
  }

  if (consistency) {
    expect_fsetequal(data$fact, data$uhashes, "uhash")
    expect_fsetequal(data$fact, data$tasks, "task_hash")
    expect_fsetequal(data$fact, data$learners, "learner_phash")
    expect_fsetequal(data$fact, data$learner_components, "learner_hash")
    expect_fsetequal(data$fact, data$resamplings, "resampling_hash")
  }
}

expect_no_extra_pkgs = function(expr, pkgs = character()) {
  req_pkgs = function(expr, pkgs) {
    library("mlr3")
    for (pkg in pkgs) {
      requireNamespace(pkg, quietly = TRUE)
    }
    snap = loadedNamespaces()
    eval(expr)
    setdiff(loadedNamespaces(), snap)
  }

  testthat::skip_if_not_installed("callr")
  extra = callr::r(req_pkgs, args = list(substitute(expr), pkgs = pkgs))
  testthat::expect_identical(extra, character(), info = "extra packages required for construction")
}

test_that("Basic ops on Ames Housing task", {
  task = tsk("california_housing")
  expect_task(task)
  expect_task_supervised(task)
  expect_task_regr(task)
  expect_equal(task$target_names, "median_house_value")

  f = task$formula()
  expect_class(f, "formula")
  # expect_set_equal(attr(terms(f), "term.labels"), task$feature_names)
})

test_that("Target is numeric", {
  b = as_data_backend(iris)
  expect_error(TaskRegr$new("iris", backend = b, target = "Species"), "Target column")
})

test_that("TaskRegr rejects NAs in target", {
  dt = data.table::data.table(x = 1:10, y = c(NA_real_, 2:10))
  expect_message(as_task_regr(dt, target = "y"), "contains missing")
})

test_that("TaskRegr: 0 feature task", {
  b = as_data_backend(data.table(y = runif(20)))
  task = TaskRegr$new(id = "zero_feat_task", b, target = "y")
  expect_output(print(task))
  b = task$backend
  expect_backend(b)
  expect_task(task)
  expect_task_supervised(task)
  expect_task_regr(task)
  expect_data_table(task$data(), ncols = 1L)

  lrn = lrn("regr.featureless")
  p = lrn$train(task)$predict(task)
  expect_prediction(p)
})

test_that("$add_strata", {
  tab = data.table(y = rep(c(1, 10), times = c(10, 10)), x = 1)
  task = TaskRegr$new("strata", tab, "y")
  expect_equal(task$col_roles$stratum, character())

  task$add_strata(task$target_names, bins = 2)
  expect_equal(task$col_roles$stratum, "..stratum_y")
  expect_equal(lengths(task$strata$row_id), c(10L, 10L))

  r = rsmp("holdout", ratio = 0.5)$instantiate(task)
  expect_equal(as.integer(table(r$train_set(1) <= 10L)), c(5L, 5L))
  expect_equal(as.integer(table(r$test_set(1) > 10L)), c(5L, 5L))

  tab = data.table(y = rep(c(1, 10), times = c(50, 10)), x = 1)
  task = TaskRegr$new("strata", tab, "y")
  task$add_strata(task$target_names, bins = 2)
  expect_identical(task$strata$N, c(50L, 10L))
})

test_that("offset column role works", {
  task = tsk("mtcars")
  expect_null(task$offset)
  task$set_col_roles("am", "offset")

  expect_subset("offset", task$properties)
  expect_data_table(task$offset, nrows = task$nrow, ncols = 2)
  expect_subset(c("row_id", "offset"), names(task$offset))

  expect_error(
    {
      task$col_roles$offset = c("am", "gear")
    },
    "up to one"
  )

  task$col_roles$offset = character()
  expect_true("offset" %nin% task$properties)
  expect_null(task$offset)
})

test_that("construction", {
  l = Learner$new("test-learner", task_type = "classif", predict_types = "prob")
  expect_class(l, "Learner")
})

test_that("clone", {
  l1 = lrn("classif.rpart")$train(tsk("iris"))
  l2 = l1$clone(deep = TRUE)

  expect_different_address(l1$state$log, l2$state$log)
  expect_different_address(l1$param_set, l2$param_set)

  l1$param_set$values = list(xval = 10L)
  expect_equal(l1$param_set$values$xval, 10L)
  expect_equal(l2$param_set$values$xval, 0L)
})

test_that("Learners are called with invoke / small footprint of call", {
  task = tsk("california_housing")
  learner = lrn("regr.rpart")
  learner$train(task)
  call = as.character(learner$model$call)
  expect_character(call, min.len = 1L, any.missing = FALSE)
  expect_match(call, "task$formula()", fixed = TRUE, all = FALSE)
  expect_match(call, "task$data", fixed = TRUE, all = FALSE)
  expect_lt(sum(nchar(call)), 1000)
})

test_that("Extra data slots of learners are kept / reset", {
  task = tsk("california_housing")
  learner = lrn("regr.rpart")
  learner$train(task)
  learner$state$foo = "bar"
  expect_equal(learner$state$foo, "bar")
  learner$predict(task)
  expect_equal(learner$state$foo, "bar")

  learner$train(task)
  expect_null(learner$state$foo)
})

test_that("task is checked in train() / predict()", {
  learner = lrn("regr.rpart")
  expect_error(learner$train(tsk("pima")), "type")
  expect_error(learner$predict(tsk("pima")), "type")
})

test_that("learner timings", {
  learner = lrn("regr.rpart")
  t = learner$timings
  expect_equal(unname(t), as.double(c(NA, NA)))
  expect_names(names(t), identical.to = c("train", "predict"))

  learner$train(tsk("mtcars"))
  t = learner$timings
  expect_number(t[["train"]])
  expect_equal(t[["predict"]], NA_real_)

  learner$predict(tsk("mtcars"))
  t = learner$timings
  expect_number(t[["train"]])
  expect_number(t[["predict"]])
})

test_that("predict on newdata works / classif", {
  task = tsk("iris")$filter(1:120)
  learner = lrn("classif.featureless")
  expect_error(learner$predict(task), "trained")
  learner$train(task)
  expect_task(learner$state$train_task, null_backend_ok = TRUE)
  newdata = tsk("iris")$filter(121:150)$data()

  # passing the task
  p = learner$predict_newdata(newdata = newdata, task = task)
  expect_data_table(as.data.table(p), nrows = 30)
  expect_set_equal(as.data.table(p)$row_ids, 1:30)
  expect_factor(p$truth, any.missing = FALSE, levels = task$class_names)

  # rely on internally stored task representation
  p = learner$predict_newdata(newdata = newdata, task = NULL)
  expect_data_table(as.data.table(p), nrows = 30)
  expect_set_equal(as.data.table(p)$row_ids, 1:30)
  expect_factor(p$truth, any.missing = FALSE, levels = task$class_names)

  # with missing target column
  newdata$Species = NULL
  p = learner$predict_newdata(newdata = newdata, task = task)
  expect_data_table(as.data.table(p), nrows = 30)
  expect_set_equal(as.data.table(p)$row_ids, 1:30)
  expect_factor(p$truth, levels = task$class_names)
  expect_true(allMissing(p$truth))
})

test_that("train task is properly cloned (#383)", {
  lr = lrn("classif.rpart")$train(tsk("iris"))
  p = lr$predict_newdata(iris[1:3, 1:4])
  expect_equal(length(p$row_ids), 3)
  p = lr$predict_newdata(iris[1:3, 1:4])
  expect_equal(length(p$row_ids), 3)
})

test_that("predict on newdata works / regr", {
  task = tsk("california_housing")
  train = which(seq_len(task$nrow) %% 2 == 0L)
  test = setdiff(seq_len(task$nrow), train)

  learner = lrn("regr.featureless")
  learner$train(task)

  newdata = task$clone()$filter(test)$data()
  p = learner$predict_newdata(newdata)

  expect_data_table(as.data.table(p), nrows = length(test))
  expect_set_equal(as.data.table(p)$row_ids, seq_along(test))
})


test_that("predict on newdata works / no target column", {
  task = tsk("california_housing")
  train = which(seq_len(task$nrow) %% 2 == 0L)
  test = setdiff(seq_len(task$nrow), train)

  learner = lrn("regr.featureless")
  learner$train(task$clone()$filter(train))

  newdata = remove_named(task$clone()$filter(test)$data(), task$target_names)
  p = learner$predict_newdata(newdata = newdata)

  expect_data_table(as.data.table(p), nrows = length(test))
  expect_set_equal(as.data.table(p)$row_ids, seq_along(test))

  # newdata with 1 col and learner that does not support missings
  xdt = data.table(x = 1, y = 1)
  task = as_task_regr(xdt, target = "y")
  learner = lrn("regr.featureless")
  learner$properties = setdiff(learner$properties, "missings")
  learner$train(task)
  learner$predict_newdata(xdt[, 1])
})

test_that("predict on newdata works / empty data frame", {
  task = tsk("mtcars")
  splits = partition(task)
  learner = lrn("regr.featureless")
  learner$train(task$clone()$filter(splits$train))

  newdata = as_data_backend(task$data(rows = integer()))
  p = learner$predict_newdata(newdata = newdata)

  expect_data_table(as.data.table(p), nrows = 0)
  expect_set_equal(as.data.table(p)$row_ids, integer())
})

test_that("predict on newdata works / data backend input", {
  task = tsk("mtcars")
  splits = partition(task)
  learner = lrn("regr.featureless")
  learner$train(task$clone()$filter(splits$train))

  newdata = task$data(rows = splits$test)
  newdata$new_row_id = sample(1e4, nrow(newdata))
  backend = as_data_backend(newdata, primary_key = "new_row_id")
  p = learner$predict_newdata(newdata = backend)

  expect_data_table(as.data.table(p), nrows = backend$nrow)
  expect_set_equal(as.data.table(p)$row_ids, backend$rownames)
})


test_that("predict on newdata works / titanic use case", {
  skip_if_not_installed("mlr3data")
  data("titanic", package = "mlr3data")
  drop = c("cabin", "name", "ticket", "passenger_id")
  data = setDT(remove_named(titanic, drop))
  task = TaskClassif$new(id = "titanic", data[!is.na(survived), ], target = "survived")

  learner = lrn("classif.rpart")$train(task)
  p = learner$predict_newdata(newdata = data[is.na(survived)])
  expect_prediction_classif(p)
  expect_factor(p$response, levels = task$class_names, any.missing = FALSE)
  expect_factor(p$truth, levels = task$class_names)
  expect_true(allMissing(p$truth))
})

test_that("predict train + test set", {
  task = tsk("iris")
  m1 = msr("debug_classif", id = "tr", predict_sets = "train")
  m2 = msr("debug_classif", id = "te", predict_sets = "test")
  m3 = msr("debug_classif", id = "trte", predict_sets = c("train", "test"))
  measures = list(m1, m2, m3)
  hout = rsmp("holdout")$instantiate(task)
  n_train = length(hout$train_set(1))
  n_test = length(hout$test_set(1))

  learner = lrn("classif.rpart")
  rr = resample(task, learner, hout)
  expect_warning(expect_warning(rr$aggregate(measures = measures), "predict sets", fixed = TRUE))

  learner = lrn("classif.rpart", predict_sets = "train")
  rr = resample(task, learner, hout)
  expect_warning(expect_warning(rr$aggregate(measures = measures), "predict sets", fixed = TRUE))

  learner = lrn("classif.rpart", predict_sets = c("train", "test"))
  rr = resample(task, learner, hout)
  aggr = rr$aggregate(measures = measures)
  expect_equal(unname(is.na(aggr)), c(FALSE, FALSE, FALSE))
  expect_equal(unname(aggr), c(n_train, n_test, n_train + n_test))
})

test_that("assertions work (#357)", {
  measures = lapply(c("classif.auc", "classif.acc"), msr)
  task = tsk("iris")
  lrn = lrn("classif.featureless")
  p = lrn$train(task)$predict(task)

  expect_error(p$score(msr("classif.auc"), "predict.type"))
  expect_error(p$score(msr("regr.mse"), "task.type"))
})

test_that("reset()", {
  task = tsk("iris")
  lrn = lrn("classif.rpart")

  lrn$train(task)
  expect_list(lrn$state, names = "unique")
  expect_learner(lrn$reset())
  expect_null(lrn$state)
})

test_that("empty predict set (#421)", {
  task = tsk("iris")
  learner = lrn("classif.featureless")
  resampling = rsmp("holdout", ratio = 1)
  hout = resampling$instantiate(task)
  learner$train(task, hout$train_set(1))
  pred = learner$predict(task, hout$test_set(1))
  expect_prediction(pred)
  expect_match(map_chr(learner$log$condition, conditionMessage), "No data to predict on", all = FALSE)
})

test_that("fallback learner is deep cloned (#511)", {
  l = lrn("classif.rpart")
  l$encapsulate("evaluate", lrn("classif.featureless"))
  expect_different_address(l$fallback, l$clone(deep = TRUE)$fallback)
})

test_that("learner cannot be trained with TuneToken present", {
  task = tsk("california_housing")
  learner = lrn("regr.rpart", cp = paradox::to_tune(0.1, 0.3))
  expect_error(
    learner$train(task),
    regexp = "<LearnerRegrRpart:regr.rpart> cannot be trained with TuneToken present",
    fixed = TRUE
  )
})

test_that("integer<->numeric conversion in newdata (#533)", {
  data = data.table(y = runif(10), x = 1:10)
  newdata1 = data.table(y = runif(10), x = as.double(1:10))
  newdata2 = data.table(y = runif(10), x = 1:10 + 0.1)

  task = TaskRegr$new("test", data, "y")
  learner = lrn("regr.featureless")
  learner$train(task)
  expect_prediction(learner$predict_newdata(data))
  expect_prediction(learner$predict_newdata(newdata1))
  expect_error(learner$predict_newdata(newdata2), "class 'numeric' to class 'integer'")
})

test_that("weights", {
  data = cbind(iris, w = rep(c(1, 100, 1), each = 50))
  task = TaskClassif$new("weighted_task", data, "Species")
  task$set_col_roles("w", "weights_learner")

  learner = lrn("classif.rpart", use_weights = "use")
  learner$train(task)

  conf = learner$predict(task)$confusion
  expect_equal(unname(conf[, 2]), c(0, 50, 0)) # no errors in class with weight 100

  expect_prediction(learner$predict_newdata(data[1:3, ]))
  expect_prediction(learner$predict_newdata(iris[1:3, ]))

  learner$use_weights = "error"
  expect_error(learner$train(task), "'use_weights' was set to 'error'")

  ll = R6Class(
    "dummy",
    inherit = LearnerClassif,
    public = list(
      initialize = function() {
        super$initialize(
          id = "dummy",
          param_set = ps(),
          feature_types = "numeric",
          predict_types = "response",
          properties = c("twoclass", "multiclass")
        )
      }
    ),
    private = list(
      .train = function(task, ...) {
        task
      }
    )
  )$new()
  expect_error(ll$train(task), "Learner does not support weights") # different error message from above!

  ll$use_weights = "ignore"
  expect_no_error(ll$train(task))
})

test_that("mandatory properties", {
  task = tsk("iris")
  learner = lrn("classif.rpart")
  learner$properties = setdiff(learner$properties, "multiclass")
  resample = rsmp("holdout")

  expect_error(learner$train(task), "multiclass")
  expect_error(resample(task, learner, resample), "multiclass")
  expect_error(benchmark(benchmark_grid(task, learner, resample)), "multiclass")
})

test_that("train task is cloned (#382)", {
  train_task = tsk("iris")
  lr = lrn("classif.rpart")$train(train_task)
  expect_different_address(lr$state$train_task, train_task)

  lrc = lr$clone(deep = TRUE)
  expect_different_address(lr$state$train_task, lrc$state$train_task)
})

test_that("Error on missing data (#413)", {
  task = tsk("pima")
  learner = lrn("classif.rpart")
  learner$properties = setdiff(learner$properties, "missings")
  expect_error(learner$train(task), "missing values")
})

test_that("Task prototype is stored in state", {
  task = tsk("california_housing")
  learner = lrn("regr.rpart")
  learner$train(task)

  prototype = learner$state$data_prototype
  expect_data_table(prototype, nrows = 0, ncols = 10)
  expect_names(names(prototype), permutation.of = c(task$feature_names, task$target_names))
})

test_that("Models can be replaced", {
  task = tsk("california_housing")
  learner = lrn("regr.featureless")
  learner$train(task)

  learner$model$location = 1
  expect_equal(learner$model$location, 1)
})

test_that("validation task's backend is removed", {
  learner = lrn("regr.rpart")
  task = tsk("mtcars")
  task$internal_valid_task = 1:10
  learner$train(task)
  expect_null(learner$state$train_task$internal_valid_task$backend)
})

test_that("manual $train() stores validation hash and validation ids", {
  task = tsk("iris")
  l = lrn("classif.debug", validate = 0.2)
  l$train(task)
  expect_character(l$state$internal_valid_task_hash)

  l = lrn("classif.debug", validate = "predefined")
  task = tsk("iris")
  task$internal_valid_task = 1:10
  l$train(task)
  expect_equal(l$state$internal_valid_task_hash, task$internal_valid_task$hash)

  # nothing is stored for learners that don't do it
  l2 = lrn("classif.featureless")
  l2$train(task)
  expect_null(l2$state$internal_valid_task_hash)
})

test_that("error when training a learner that sets valiadte to 'predefined' on a task without a validation task", {
  task = tsk("iris")
  learner = lrn("classif.debug", validate = "predefined")
  expect_error(learner$train(task), "is set to 'predefined'")
  task$internal_valid_task = 1:10
  expect_class(learner, "Learner")
})

test_that("properties are also checked on validation task", {
  task = tsk("iris")
  row = task$data(1)
  row[[1]][1] = NA
  row$..row_id = 151
  task$rbind(row)
  task$internal_valid_task = 151
  learner = lrn("classif.debug", validate = "predefined")
  learner$properties = setdiff(learner$properties, "missings")
  expect_error(learner$train(task), "missing values")
})

test_that("validate changes phash and hash", {
  l1 = lrn("classif.debug")
  l2 = lrn("classif.debug")
  l2$validate = 0.2
  expect_false(l1$hash == l2$hash)
})

test_that("marshaling and encapsulation", {
  task = tsk("iris")
  learner = lrn("classif.debug", count_marshaling = TRUE)

  # callr encapsulation causes marshaling
  learner$encapsulate("callr", lrn("classif.featureless"))
  learner$train(task)
  expect_equal(learner$model$marshal_count, 1)
  expect_false(learner$marshaled)
  expect_prediction(learner$predict(task))

  # no marshaling with no other encapsulation
  learner$encapsulate("none")
  learner$train(task)
  expect_equal(learner$model$marshal_count, 0)
})

test_that("marshal state", {
  state = lrn("classif.debug")$train(tsk("iris"))$state
  sm = marshal_model(state)
  expect_true(is_marshaled_model(sm))
  expect_equal(state, unmarshal_model(marshal_model(state)))
})


test_that("internal_valid_task is created correctly", {
  LearnerClassifTest = R6Class(
    "LearnerClassifTest",
    inherit = LearnerClassifDebug,
    public = list(
      task = NULL
    ),
    private = list(
      .train = function(task, ...) {
        self$task = task$clone(deep = TRUE)
        super$.train(task, ...)
      }
    )
  )
  # validate = NULL (but task has one)
  learner = LearnerClassifTest$new()
  task = tsk("iris")
  task$internal_valid_task = partition(task)$test
  learner$train(task)
  learner$validate = NULL
  expect_null(learner$internal_valid_scores)
  expect_null(learner$task$internal_valid_task)

  # validate = NULL (but task has none)
  learner1 = LearnerClassifTest$new()
  task1 = tsk("iris")
  learner1$train(task1)
  expect_null(learner1$internal_valid_scores)
  expect_null(learner1$task$internal_valid_task)

  # validate = "test"
  LearnerClassifTest2 = R6Class(
    "LearnerClassifTest2",
    inherit = LearnerClassifDebug,
    public = list(
      expected_valid_ids = NULL,
      expected_train_ids = NULL
    ),
    private = list(
      .train = function(task, ...) {
        if (!test_permutation(task$internal_valid_task$row_ids, self$expected_valid_ids)) {
          stopf("something went wrong")
        }
        if (!test_permutation(task$row_ids, self$expected_train_ids)) {
          stopf("something went wrong")
        }
        super$.train(task, ...)
      }
    )
  )
  task2 = tsk("iris")
  learner2 = LearnerClassifTest2$new()
  learner2$validate = "test"
  resampling = rsmp("holdout")$instantiate(task2)
  learner2$expected_valid_ids = resampling$test_set(1)
  learner2$expected_train_ids = resampling$train_set(1)
  expect_no_error(resample(task2, learner2, resampling))

  # ratio works
  LearnerClassifTest3 = R6Class(
    "LearnerClassifTest3",
    inherit = LearnerClassifDebug,
    private = list(
      .train = function(task, ...) {
        if (length(task$internal_valid_task$row_ids) != 20) {
          stopf("something went wrong")
        }
        super$.train(task, ...)
      }
    )
  )
  task = tsk("iris")$filter(1:100)
  learner3 = LearnerClassifTest3$new()
  learner3$validate = 0.2
  learner3$train(task)

  # check that validation task is reset discarded at the end
  learner4 = lrn("classif.debug", validate = 0.2)
  task = tsk("iris")
  learner4$train(task)
  expect_null(task$internal_valid_task)
})

test_that("compatibility check on validation task", {
  learner = lrn("classif.debug", validate = "predefined")
  task = tsk("german_credit")
  task$internal_valid_task = 1:10
  task$col_roles$feature = "age"
  expect_error(learner$train(task), "has different features")
  task$internal_valid_task$col_roles$feature = "age"
  task$internal_valid_task$col_roles$target = "credit_history"
  expect_error(learner$train(task), "has different target")
})

test_that("model is marshaled during parallel predict", {
  # by setting check_pid = TRUE, we ensure that unmarshal_model() sets the process id to the current
  # id. LearnerClassifDebug then checks during `.predict()`,
  # whether the marshal_id of the model is equal to the current process id and errs if this is not the case.
  task = tsk("iris")
  learner = lrn("classif.debug", check_pid = TRUE)
  learner$train(task)
  learner$parallel_predict = TRUE
  pred = with_future(future::multisession, {
    learner$predict(task)
  })
  expect_class(pred, "Prediction")
})

test_that("model is marshaled during callr prediction", {
  # by setting check_pid = TRUE, we ensure that unmarshal_model() sets the process id to the current
  # id. LearnerClassifDebug then checks during `.predict()`,
  # whether the marshal_id of the model is equal to the current process id and errs if this is not the case.
  task = tsk("iris")
  learner = lrn("classif.debug", check_pid = TRUE)
  learner$encapsulate("callr", lrn("classif.featureless"))
  learner$train(task)
  pred = learner$predict(task)
  expect_class(pred, "Prediction")
})

test_that("predict leaves marshaling status as-is", {
  task = tsk("iris")
  learner = lrn("classif.debug", check_pid = TRUE)
  learner$encapsulate("callr", lrn("classif.featureless"))
  learner$train(task)
  learner$marshal()
  expect_class(learner$predict(task), "Prediction")
  expect_true(learner$marshaled)
  learner$unmarshal()
  expect_class(learner$predict(task), "Prediction")
  expect_false(learner$marshaled)
})

test_that("learner state contains validate field", {
  learner = lrn("classif.debug", validate = 0.2)
  learner$train(tsk("iris"))
  expect_equal(learner$state$validate, 0.2)
})

test_that("learner state contains internal valid task information", {
  task = tsk("iris")
  # 1. resample
  learner = lrn("classif.debug", validate = 0.2)
  rr = resample(task, learner, rsmp("holdout"))
  expect_string(rr$learners[[1L]]$state$internal_valid_task_hash)

  # 1. manual
  learner$train(task)
  expect_string(learner$state$internal_valid_task_hash)
})

test_that("validation task with 0 observations", {
  learner = lrn("classif.debug", validate = "predefined")
  task = tsk("iris")
  expect_warning({
    task$internal_valid_task = integer(0)
  })
})

test_that("column info is compared during predict", {
  x = c("a", "b")
  d = data.table(
    x = factor(x, levels = c("a", "b")),
    y = 1:2
  )

  dflip = data.table(
    x = factor(x, labels = c("b", "a"), levels = c("b", "a")),
    y = 1:2
  )

  dother = data.table(
    x1 = factor(c("a", "b")),
    y = 1:2
  )

  task = as_task_classif(d, target = "y")
  task_flip = as_task_classif(dflip, target = "y")
  task_other = as_task_classif(dother, target = "y")
  l = lrn("classif.rpart")
  l$train(task)
  expect_error(l$predict(task_flip), "task with different column info")
  expect_error(l$predict(task_other), "with different columns")
})

test_that("quantiles in LearnerRegr", {
  task = tsk("mtcars")
  learner = lrn("regr.debug", predict_type = "quantiles")
  expect_learner(learner)
  quantiles = c(0.05, 0.5, 0.95)
  learner$quantiles = quantiles

  expect_numeric(learner$quantiles, any.missing = FALSE, len = 3)

  learner$quantiles = NULL
  expect_null(learner$quantiles)
  expect_null(learner$quantile_response)

  learner$quantiles = quantiles
  learner$quantile_response = 0.6
  expect_equal(learner$quantile_response, 0.6)
  expect_equal(learner$quantiles, c(0.05, 0.5, 0.6, 0.95))

  expect_error(
    {
      learner$quantiles = c(0.5, 0.1)
    },
    "sorted"
  )

  expect_error(
    {
      learner$quantiles = integer()
    },
    "length"
  )

  learner$train(task)

  expect_numeric(learner$model$quantiles, len = 4L)

  pred = learner$predict(task)
  expect_prediction(pred)
  expect_subset("quantiles", pred$predict_types)
  expect_matrix(pred$quantiles, ncols = 4L, nrows = task$nrow, any.missing = FALSE)
  expect_true(!any(apply(pred$quantiles, 1L, is.unsorted)))
  expect_equal(pred$response, pred$quantiles[, 3L])

  tab = as.data.table(pred)
  expect_data_table(tab, nrows = task$nrow)
  expect_subset("q0.5", names(tab))
})

test_that("predict time is cumulative", {
  # no runtime test on CRAN
  skip_on_cran()

  learner = lrn("classif.debug", sleep_predict = function() 0.05)
  task = tsk("iris")
  learner$train(task)$predict(task)
  t1 = learner$timings["predict"]
  learner$param_set$values$sleep_predict = function() 0.01
  learner$predict(task)
  t2 = learner$timings["predict"]
  expect_true(t1 > t2)
})

test_that("weights properties and defaults", {
  ll = lrn("classif.rpart")
  expect_true("weights" %in% ll$properties)
  expect_equal(ll$use_weights, "use")

  ll = LearnerClassif$new(
    id = "dummy",
    param_set = ps(),
    feature_types = "numeric",
    predict_types = "response",
    properties = "twoclass"
  )
  expect_true("weights" %nin% ll$properties)
  expect_equal(ll$use_weights, "error")
})

test_that("configure method works", {
  learner = lrn("classif.rpart")

  expect_learner(learner$configure())
  expect_learner(learner$configure(.values = list()))

  # set new hyperparameter value
  learner$configure(cp = 0.1)
  expect_equal(learner$param_set$values$cp, 0.1)

  # overwrite existing hyperparameter value
  learner$configure(xval = 10)
  expect_equal(learner$param_set$values$xval, 10)

  # set field
  learner$configure(predict_sets = "train")
  expect_equal(learner$predict_sets, "train")

  # hyperparameter and field
  learner$configure(minbucket = 2, parallel_predict = TRUE)
  expect_equal(learner$param_set$values$minbucket, 2)
  expect_true(learner$parallel_predict)

  # unknown hyperparameter and field
  expect_error(learner$configure(xvald = 1), "Cannot set argument")

  # use .values
  learner = lrn("classif.rpart")
  learner$configure(.values = list(cp = 0.1, xval = 10, predict_sets = "train"))
  expect_equal(learner$param_set$values$cp, 0.1)
  expect_equal(learner$param_set$values$xval, 10)
  expect_equal(learner$predict_sets, "train")
})

test_that("predict_raw flag works", {
  learner = lrn("classif.rpart")
  expect_false(learner$predict_raw)

  learner$predict_raw = TRUE
  expect_true(learner$predict_raw)

  learner$configure(predict_raw = FALSE)
  expect_false(learner$predict_raw)

  h1 = lrn("classif.rpart")$hash
  h2 = lrn("classif.rpart", predict_raw = TRUE)$hash
  expect_true(h1 != h2)
})

test_that("selected_features works", {
  task = tsk("spam")
  # alter rpart class to not support feature selection
  fun = LearnerClassifRpart$public_methods$selected_features
  on.exit({
    LearnerClassifRpart$public_methods$selected_features = fun
  })
  LearnerClassifRpart$public_methods$selected_features = NULL

  learner = lrn("classif.rpart")
  expect_error(learner$selected_features(), "No model stored")
  learner$train(task)
  expect_error(learner$selected_features(), "Learner does not support feature selection")

  learner$selected_features_impute = "all"
  expect_equal(learner$selected_features(), task$feature_names)
})

test_that("predict_newdata auto conversion (#685)", {
  l = lrn("classif.debug", save_tasks = TRUE)$train(tsk("iris")$select(c("Sepal.Length", "Sepal.Width")))
  expect_error(
    l$predict_newdata(data.table(Sepal.Length = 1, Sepal.Width = "abc")),
    "Incompatible types during auto-converting column 'Sepal.Width'",
    fixed = TRUE
  )
  expect_error(l$predict_newdata(data.table(Sepal.Length = 1L)), "but is missing elements")

  # New test for integerish value conversion to double
  p1 = l$predict_newdata(data.table(Sepal.Length = 1, Sepal.Width = 2))
  p2 = l$predict_newdata(data.table(Sepal.Length = 1L, Sepal.Width = 2))
  expect_equal(l$model$task_predict$col_info[list("Sepal.Length")]$type, "numeric")
  expect_double(l$model$task_predict$data(cols = "Sepal.Length")[[1]])

  expect_equal(p1, p2)
})

test_that("predict_newdata creates column info correctly", {
  learner = lrn("classif.debug", save_tasks = TRUE)
  task = tsk("iris")
  task$col_info$label = letters[1:6]
  task$col_info$fix_factor_levels = c(TRUE, TRUE, FALSE, TRUE, FALSE, TRUE)
  learner$train(task)

  ## data.frame is passed without task
  p1 = learner$predict_newdata(iris[10:11, ])
  expect_equal(learner$model$task_predict$col_info, task$col_info)
  expect_equal(p1$row_ids, 1:2)

  ## backend is passed without task
  d = iris[10:11, ]
  d$..row_id = 10:11
  b = as_data_backend(d, primary_key = "..row_id")
  p2 = learner$predict_newdata(b)
  expect_equal(p2$row_ids, 10:11)
  expect_equal(learner$model$task_predict$col_info, task$col_info)

  ## data.frame is passed with task
  task2 = tsk("iris")
  learner$predict_newdata(iris[10:11, ], task2)
  expect_equal(learner$model$task_predict$col_info, task2$col_info)

  ## backend is passed with task
  learner$predict_newdata(b, task2)
  expect_equal(learner$model$task_predict$col_info, task2$col_info)

  ## backend with different name for primary key
  d2 = iris[10:11, ]
  d2$row_id = 10:11
  b2 = as_data_backend(d2, primary_key = "row_id")
  p3 = learner$predict_newdata(b2, task2)
  expect_equal(p3$row_ids, 10:11)
  expect_true("row_id" %in% learner$model$task_predict$col_info$id)
})

test_that("predict_newdata preserves target level ordering (#1459)", {
  data = data.table(
    x = c(rnorm(50, 0), rnorm(50, 5)),
    y = factor(rep(c("0", "1"), each = 50), levels = c("0", "1"))
  )
  task = as_task_classif(data, target = "y", positive = "1")
  learner = lrn("classif.featureless", predict_type = "prob")
  learner$train(task)

  p1 = learner$predict(task)
  p2 = learner$predict_newdata(data)
  expect_equal(p1$prob, p2$prob)
})

test_that("marshaling and internal tuning", {
  l = lrn("classif.debug", validate = 0.3, early_stopping = TRUE, iter = 100)
  l$encapsulate("evaluate", lrn("classif.featureless"))
  task = tsk("iris")
  l$train(task)
  expect_list(l$internal_tuned_values, types = "integer")
  expect_list(l$internal_valid_scores, types = "numeric")
})

test_that("prob_as_default works", {
  on.exit(options(old_opts))
  old_opts = options(mlr3.prob_as_default = TRUE)
  l = lrn("classif.debug")
  expect_equal(l$predict_type, "prob")
  options(mlr3.prob_as_default = NULL)
  l = lrn("classif.debug")
  expect_equal(l$predict_type, "response")
  options(mlr3.prob_as_default = FALSE)
  expect_equal(l$predict_type, "response")
})

test_that("weights are used when appropriate", {
  learner = lrn("classif.featureless", predict_type = "prob")
  predict_task = tsk("iris")$filter(1)
  expect_equal(
    unname(learner$train(tsk("iris"))$predict(predict_task)$prob),
    matrix(c(1, 1, 1) / 3, nrow = 1, ncol = 3)
  )

  # weights_measure has no effect
  expect_equal(
    unname(learner$train(iris_weights_measure)$predict(predict_task)$prob),
    matrix(c(1, 1, 1) / 3, nrow = 1, ncol = 3)
  )

  expect_equal(
    unname(learner$train(iris_weights_learner)$predict(predict_task)$prob),
    matrix(c(1, 10, 100) / 111, nrow = 1, ncol = 3)
  )

  learner$use_weights = "ignore"

  # weights are ignored
  expect_equal(
    unname(learner$train(iris_weights_learner)$predict(predict_task)$prob),
    matrix(c(1, 1, 1) / 3, nrow = 1, ncol = 3)
  )

  learner$use_weights = "error"
  expect_error(learner$train(iris_weights_learner), "'use_weights' was set to\n  'error'")

  # behavior of learner that does not support weights
  llclass = R6Class(
    "dummy",
    inherit = LearnerClassif,
    public = list(
      initialize = function() {
        super$initialize(
          id = "dummy",
          param_set = ps(),
          feature_types = "numeric",
          predict_types = "response",
          properties = c("twoclass", "multiclass")
        )
      }
    ),
    private = list(
      .train = function(task) {
        list(response = as.character(sample(task$truth(), 1L)))
      }
    )
  )
  ll = llclass$new()

  # different error message
  expect_error(ll$train(iris_weights_learner), "Learner does not support\n  weights")
  ll$train(iris_weights_measure)

  ll$use_weights = "ignore"
  ll$train(iris_weights_learner)
})

test_that("Learner printer for use_weights", {
  expect_output(print(lrn("classif.featureless", use_weights = "ignore")), "use_weights = 'ignore'")
  expect_output(print(lrn("classif.featureless", use_weights = "use")), "use_weights = 'use'")
  expect_output(print(lrn("classif.featureless", use_weights = "error")), "use_weights = 'error'")
})

test_that("Learner printer for encapsulation", {
  expect_output(
    print(lrn("classif.featureless")$encapsulate("callr", lrn("classif.rpart"))),
    "Encapsulation: callr \\(fallback: LearnerClassifRpart\\)"
  )
  expect_output(
    print(lrn("classif.rpart")$encapsulate("evaluate", lrn("classif.featureless"))),
    "Encapsulation: evaluate \\(fallback: LearnerClassifFeatureless\\)"
  )
  expect_output(print(lrn("classif.rpart")$encapsulate("none")), "Encapsulation: none \\(fallback: -\\)")
})

test_that("error conditions are working: callr", {
  # no runtime test on CRAN
  skip_on_cran()

  l = lrn(
    "classif.debug",
    timeout = c(train = 0.01),
    # Sys.sleep does not get interrupted reliably
    sleep_train = function() {
      while (TRUE) {
        NULL
      }
    }
  )

  l$encapsulate(
    "callr",
    lrn("classif.featureless"),
    when = function(cond, ...) {
      !inherits(cond, "Mlr3ErrorTimeout")
    }
  )

  expect_error(l$train(tsk("iris")), regexp = "reached elapsed time limit")
  l$configure(error_train = 1, sleep_train = NULL, timeout = c(train = Inf, predict = Inf))
  expect_learner(l$train(tsk("iris")))
})

test_that("error conditions are working: evaluate", {
  # no runtime test on CRAN
  skip_on_cran()

  l = lrn(
    "classif.debug",
    timeout = c(train = 0.2),
    # Sys.sleep does not get interrupted reliably
    sleep_train = function() {
      while (TRUE) {
        NULL
      }
    }
  )

  l$encapsulate(
    "evaluate",
    lrn("classif.featureless"),
    function(cond, ...) {
      !inherits(cond, "Mlr3ErrorTimeout")
    }
  )

  expect_error(l$train(tsk("iris")), regexp = "reached elapsed time limit")
  l$configure(error_train = 1, sleep_train = NULL, timeout = c(train = Inf, predict = Inf))
  expect_learner(l$train(tsk("iris")))
})

test_that("error conditions are working: try", {
  # no runtime test on CRAN
  skip_on_cran()

  # try encapsulation prints to console
  null_con = file(nullfile(), open = "wt")
  sink(null_con, type = "message")
  sink(null_con, type = "output")

  on.exit({
    sink(type = "message")
    sink(type = "output")
    close(null_con)
  })

  l = lrn(
    "classif.debug",
    timeout = c(train = 0.01),
    # Sys.sleep does not get interrupted reliably
    sleep_train = function() {
      while (TRUE) {
        NULL
      }
    }
  )

  l$encapsulate(
    "try",
    lrn("classif.featureless"),
    function(cond, ...) {
      !inherits(cond, "Mlr3ErrorTimeout")
    }
  )

  expect_error(l$train(tsk("iris")), class = "Mlr3ErrorTimeout")
  l$configure(error_train = 1, sleep_train = NULL, timeout = c(train = Inf, predict = Inf))
  expect_learner(l$train(tsk("iris")))
})

test_that("error conditions are working: mirai", {
  # no runtime test on CRAN
  skip_on_cran()

  l = lrn(
    "classif.debug",
    timeout = c(train = 0.01),
    # Sys.sleep does not get interrupted reliably
    sleep_train = function() {
      while (TRUE) {
        NULL
      }
    }
  )

  l$encapsulate(
    "mirai",
    lrn("classif.featureless"),
    function(cond, ...) {
      !inherits(cond, "Mlr3ErrorTimeout")
    }
  )

  expect_error(l$train(tsk("iris")), regexp = "reached elapsed time limit")
  l$configure(error_train = 1, sleep_train = NULL, timeout = c(train = Inf, predict = Inf))
  expect_learner(l$train(tsk("iris")))
})

test_that("error conditions are working for predict", {
  # config errors are not caught
  task = tsk("iris")
  l = lrn("classif.debug", error_predict = 1, sleep_predict = function() {
    error_config("all working!")
  })
  l$encapsulate("evaluate", lrn("classif.featureless"))
  l$train(task)
  expect_error(l$predict(task), regexp = "all working!", fixed = TRUE)
})

test_that("when: stage parameter is working", {
  task = tsk("iris")
  l = lrn("classif.debug")

  # input checks on argument 'when':
  expect_error(
    l$encapsulate("evaluate", lrn("classif.featureless"), function(x, y) NULL),
    regexp = "subset of"
  )
  expect_error(
    l$encapsulate("evaluate", lrn("classif.featureless"), "a"),
    regexp = "function"
  )
  expect_error(
    l$encapsulate("evaluate", lrn("classif.featureless"), function(...) NULL),
    regexp = NA
  )

  # we catch error during train, but not during predict
  l$encapsulate("evaluate", lrn("classif.featureless"), function(cond, stage) {
    if (inherits(cond, "Mlr3TestError")) {
      return(stage == "train")
    }
    if (stage == "predict" && grepl("No model stored", cond$message)) {
      return(FALSE)
    }
    stop("test went wrong")
  })
  l$configure(
    sleep_train = function() error_mlr3("a", class = "Mlr3TestError"),
    sleep_predict = function() error_mlr3("a", class = "Mlr3TestError")
  )
  expect_error(l$train(task), regexp = NA)
  expect_null(l$model)
  expect_error(l$predict(task), regexp = "No model stored", class = "Mlr3ErrorLearnerNoModel")
  l$encapsulate("evaluate", lrn("classif.featureless"), function(cond, stage) {
    if (inherits(cond, "Mlr3TestError")) {
      return(stage == "train")
    }
    if (stage == "predict" && grepl("No model stored", cond$message)) {
      return(TRUE)
    }
    stop("test went wrong")
  })
  expect_class(l$predict(task), "PredictionClassif")

  l = lrn("classif.debug", error_train = 1)
  l$encapsulate("evaluate", lrn("classif.featureless"))
  l$train(task)
})

test_that("oob_error is available without storing models via $.extract_oob_error()", {
  LearnerDummyOOB = R6::R6Class(
    "LearnerDummyOOB",
    inherit = LearnerClassif,
    public = list(
      initialize = function() {
        super$initialize(
          id = "classif.dummy_oob",
          param_set = paradox::ps(),
          feature_types = c("logical", "integer", "numeric", "character", "factor", "ordered"),
          predict_types = c("response"),
          properties = c("twoclass", "multiclass", "oob_error"),
          man = NA_character_
        )
      }
    ),
    private = list(
      .train = function(task) {
        list(response = as.character(sample(task$truth(), 1L)))
      },
      .predict = function(task) {
        list(response = rep.int(self$model$response, task$nrow))
      },
      .extract_oob_error = function() {
        0.123
      }
    )
  )

  task = tsk("iris")
  learner = LearnerDummyOOB$new()
  rr = resample(task, learner, rsmp("holdout"), store_models = FALSE)

  expect_equal(rr$aggregate(msr("oob_error")), c(oob_error = 0.123))
})

test_that("config error does not trigger callback", {
  l = lrn("classif.debug", config_error = TRUE)
  l$encapsulate("evaluate", lrn("classif.featureless"), function(...) TRUE)
  expect_error(l$train(tsk("iris")), regexp = "You misconfigured the learner")
  l$encapsulate("evaluate", lrn("classif.featureless"))
  expect_error(l$train(tsk("iris")), regexp = "You misconfigured the learner")
})

test_that("new_levels property is working", {
  learner = lrn("classif.featureless")
  task = tsk("penguins")
  learner$train(task)
  data = task$data()
  set(data, i = 1L, j = "island", value = "NewIsland")

  expect_error(learner$predict_newdata(data), "received task with different column info")

  learner$properties = c(learner$properties, "new_levels")
  expect_prediction(learner$predict_newdata(data))
})

test_that("native_model returns model by default", {
  task = tsk("iris")
  learner = lrn("classif.rpart")
  learner$train(task)
  expect_identical(learner$native_model, learner$model)
  expect_class(learner$native_model, "rpart")
})

test_that("native_model can be overwritten by learner", {
  LearnerWithExtraInfo = R6Class(
    "LearnerWithExtraInfo",
    inherit = LearnerClassif,
    public = list(
      initialize = function() {
        super$initialize(
          id = "classif.extra_info",
          param_set = paradox::ps(),
          feature_types = c("logical", "integer", "numeric", "factor", "ordered"),
          predict_types = c("response"),
          properties = c("twoclass", "multiclass"),
          man = NA_character_
        )
      }
    ),
    active = list(
      native_model = function(rhs) {
        assert_ro_binding(rhs)
        self$model$model
      }
    ),
    private = list(
      .train = function(task) {
        list(
          model = list(response = as.character(sample(task$truth(), 1L))),
          extra_info = "some additional information"
        )
      },
      .predict = function(task) {
        list(response = rep.int(self$model$model$response, task$nrow))
      }
    )
  )

  task = tsk("iris")
  learner = LearnerWithExtraInfo$new()
  learner$train(task)
  expect_list(learner$model)
  expect_true("extra_info" %in% names(learner$model))
  expect_equal(learner$model$extra_info, "some additional information")
  expect_list(learner$native_model)
  expect_false("extra_info" %in% names(learner$native_model))
  expect_true("response" %in% names(learner$native_model))
})

test_that("deadline works as intended: callr", {
  # no runtime test on CRAN
  skip_on_cran()
  eternal_sleep = function() {
    while (TRUE) {
      Sys.sleep(0.01)
    }
  }

  l = lrn("classif.debug")
  l$encapsulate(
    "callr",
    lrn("classif.featureless"),
    function(cond, ...) {
      !inherits(cond, "Mlr3ErrorTimeout")
    }
  )

  # Train: Test that it interrupts, but not immediately
  allowed = 0.5
  l$configure(sleep_train = eternal_sleep, deadline = c(train = Sys.time() + allowed, predict = as.POSIXct(Inf)))
  expect_error(l$train(tsk("iris")), regexp = "reached elapsed time limit")
  expect_true(Sys.time() >= l$deadline[["train"]])

  # Train: Test that learner can finish before deadline
  allowed = 10
  l$configure(sleep_train = NULL, deadline = c(train = Sys.time() + allowed, predict = as.POSIXct(Inf)))
  expect_no_error(l$train(tsk("iris")))
  expect_true((l$timings["train"] < allowed))

  # Predict: Test that it interrupts, but not immediately
  allowed = 0.5
  l$configure(
    sleep_train = NULL,
    sleep_predict = eternal_sleep,
    deadline = c(train = as.POSIXct(Inf), predict = Sys.time() + allowed)
  )
  l$train(tsk("iris"))
  expect_error(l$predict(tsk("iris")), regexp = "reached elapsed time limit")
  expect_true(Sys.time() >= l$deadline[["predict"]])

  # Predict: Test that learner can finish before deadline
  allowed = 10
  l$configure(
    sleep_train = NULL,
    sleep_predict = NULL,
    deadline = c(train = as.POSIXct(Inf), predict = Sys.time() + allowed)
  )
  l$train(tsk("iris"))
  expect_no_error(l$predict(tsk("iris")))
  expect_true((l$timings["predict"] < allowed))
})

test_that("deadline works as intended: evaluate", {
  # no runtime test on CRAN
  skip_on_cran()
  eternal_sleep = function() {
    while (TRUE) {
      Sys.sleep(0.01)
    }
  }

  l = lrn("classif.debug")
  l$encapsulate(
    "evaluate",
    lrn("classif.featureless"),
    function(cond, ...) {
      !inherits(cond, "Mlr3ErrorTimeout")
    }
  )

  # Train: Test that it interrupts, but not immediately
  allowed = 0.5
  l$configure(sleep_train = eternal_sleep, deadline = c(train = Sys.time() + allowed, predict = as.POSIXct(Inf)))
  expect_error(l$train(tsk("iris")), regexp = "reached elapsed time limit")
  expect_true(Sys.time() >= l$deadline[["train"]])

  # Train: Test that learner can finish before deadline
  allowed = 10
  l$configure(sleep_train = NULL, deadline = c(train = Sys.time() + allowed, predict = as.POSIXct(Inf)))
  expect_no_error(l$train(tsk("iris")))
  expect_true((l$timings["train"] < allowed))

  # Predict: Test that it interrupts, but not immediately
  allowed = 0.5
  l$configure(
    sleep_train = NULL,
    sleep_predict = eternal_sleep,
    deadline = c(train = as.POSIXct(Inf), predict = Sys.time() + allowed)
  )
  l$train(tsk("iris"))
  expect_error(l$predict(tsk("iris")), regexp = "reached elapsed time limit")
  expect_true(Sys.time() >= l$deadline[["predict"]])

  # Predict: Test that learner can finish before deadline
  allowed = 10
  l$configure(
    sleep_train = NULL,
    sleep_predict = NULL,
    deadline = c(train = as.POSIXct(Inf), predict = Sys.time() + allowed)
  )
  l$train(tsk("iris"))
  expect_no_error(l$predict(tsk("iris")))
  expect_true((l$timings["predict"] < allowed))
})

test_that("deadline works as intended: mirai", {
  # no runtime test on CRAN
  skip_on_cran()
  eternal_sleep = function() {
    while (TRUE) {
      Sys.sleep(0.01)
    }
  }

  l = lrn("classif.debug")
  l$encapsulate(
    "mirai",
    lrn("classif.featureless"),
    function(cond, ...) {
      !inherits(cond, "Mlr3ErrorTimeout")
    }
  )

  # Train: Test that it interrupts, but not immediately
  allowed = 0.5
  l$configure(sleep_train = eternal_sleep, deadline = c(train = Sys.time() + allowed, predict = as.POSIXct(Inf)))
  expect_error(l$train(tsk("iris")), regexp = "reached elapsed time limit")
  expect_true(Sys.time() >= l$deadline[["train"]])

  # Train: Test that learner can finish before deadline
  allowed = 10
  l$configure(sleep_train = NULL, deadline = c(train = Sys.time() + allowed, predict = as.POSIXct(Inf)))
  expect_no_error(l$train(tsk("iris")))
  expect_true((l$timings["train"] < allowed))

  # Predict: Test that it interrupts, but not immediately
  allowed = 0.5
  l$configure(
    sleep_train = NULL,
    sleep_predict = eternal_sleep,
    deadline = c(train = as.POSIXct(Inf), predict = Sys.time() + allowed)
  )
  l$train(tsk("iris"))
  expect_error(l$predict(tsk("iris")), regexp = "reached elapsed time limit")
  expect_true(Sys.time() >= l$deadline[["predict"]])

  # Predict: Test that learner can finish before deadline
  allowed = 10
  l$configure(
    sleep_train = NULL,
    sleep_predict = NULL,
    deadline = c(train = as.POSIXct(Inf), predict = Sys.time() + allowed)
  )
  l$train(tsk("iris"))
  expect_no_error(l$predict(tsk("iris")))
  expect_true((l$timings["predict"] < allowed))
})

test_that("Feature columns can be reordered", {
  task = tsk("california_housing")
  new_order = shuffle(task$feature_names)

  task$col_roles$feature = new_order
  expect_equal(task$feature_names, new_order)
  expect_names(names(task$data(rows = 1)), identical.to = c("median_house_value", new_order))
})

test_that("Task duplicates rows", {
  # getting same row ids twice
  task = tsk("iris")
  data = task$data(c(1L, 1L))
  expect_data_table(data, nrows = 2L, any.missing = FALSE)

  # task with duplicated ids in row_roles$use
  # this happens in ResamplingBootstrap!
  task = tsk("iris")
  task$row_roles$use = c(1:5, 1:5, 146:150)
  expect_task(task, duplicated_ids = TRUE)

  expect_equal(task$nrow, 15L)
  expect_data_table(task$data(), nrows = 15)
  task$droplevels()
  expect_character(task$class_names, len = 2L)

  task$set_row_roles(1, remove_from = "use")
  expect_equal(task$nrow, 13L)
  task$set_row_roles(1L, add_to = "use")
  expect_equal(task$nrow, 14L)
  task$set_row_roles(1L, add_to = "use")
  expect_equal(task$nrow, 15L)
})

test_that("Rows return ordered", {
  x = load_dataset("nhtemp", "datasets", TRUE)
  data = as.data.frame(x)
  data$x = as.numeric(data$x)
  data$t = as.integer(time(x))
  data = data[rev(seq_row(data)), ]
  rownames(data) = NULL
  b = as_data_backend(data)
  task = TaskRegr$new(id = "nhtemp", b, target = "x")

  x = task$data()
  expect_true(is.unsorted(x$t))

  task$col_roles$order = "t"
  x = task$data(ordered = TRUE)
  expect_integer(x$t, sorted = TRUE, any.missing = FALSE)

  x = task$data(ordered = FALSE)
  expect_true(is.unsorted(x$t))

  x = task$data(rows = sample(nrow(data), 50), ordered = TRUE)
  expect_integer(x$t, sorted = TRUE, any.missing = FALSE)
})

test_that("Rows return ordered with multiple order cols", {
  task = tsk("iris")

  x = task$data()
  expect_true(is.unsorted(x$Petal.Length))

  task$col_roles$order = c("Petal.Length", "Petal.Width")
  expect_equal(task$col_roles$order, c("Petal.Length", "Petal.Width"))

  x = task$data(ordered = TRUE)
  expect_numeric(x$Petal.Length, sorted = TRUE, any.missing = FALSE)

  expect_true(x[, is.unsorted(Petal.Width)])
  expect_true(all(x[, is.unsorted(Petal.Width), by = Petal.Width]$V1 == FALSE))
})


test_that("Task rbind", {
  task = tsk("iris")
  # expect_error(task$rbind(task), "data.frame")
  data = iris[1:10, ]
  task$rbind(iris[1:10, ])
  expect_task(task)
  expect_equal(task$nrow, 160)

  task$rbind(iris[integer(), ])
  expect_equal(task$nrow, 160)

  # #185
  task = tsk("iris")
  task$select("Petal.Length")
  task$rbind(task$data())
  expect_set_equal(task$row_ids, 1:300)

  task$rbind(data.table())
  expect_equal(task$nrow, 300L)

  # #437
  task = tsk("zoo")
  data = task$data()
  data$foo = 101:1
  nt = task$clone()$rbind(data)
  expect_task(nt)
  expect_set_equal(nt$row_ids, 1:202)
  expect_equal(nt$row_names$row_name, c(task$row_names$row_name, rep(NA, 101)))
  expect_equal(nt$col_info[list("foo"), .N, nomatch = NULL], 0L)

  # 496
  data = iris
  data$blocks = sample(letters[1:2], nrow(iris), replace = TRUE)
  task = TaskClassif$new("iris", data, target = "Species")
  task$col_roles$feature = setdiff(task$col_roles$feature, "blocks")
  task$col_roles$group = "blocks"
  learner = lrn("classif.rpart")
  learner$train(task)
  expect_prediction(predict(learner, iris, predict_type = "<Prediction>"))

  # merge factor levels
  task = tsk("penguins")
  data = task$data(1)
  data$sex = factor("unsure", levels = c("male", "female", "unsure"))
  task$rbind(data)
  expect_equal(task$levels("sex")[[1]], c("female", "male", "unsure"))
  expect_equal(task$col_info[list("sex"), fix_factor_levels], TRUE)
})

test_that("Task cbind", {
  task = tsk("iris")

  iris_col_hashes = task$col_hashes

  # expect_error(task$cbind(task), "data.frame")
  data = cbind(data.frame(foo = 150:1), data.frame(..row_id = task$row_ids))
  task$cbind(data)
  expect_task(task)
  expect_equal(task$ncol, 6L)
  expect_names(task$feature_names, must.include = "foo")

  expect_equal(iris_col_hashes, task$col_hashes[names(iris_col_hashes)])

  data = data.frame(bar = runif(150))
  task$cbind(data)
  expect_task(task)
  expect_equal(task$ncol, 7L)
  expect_names(task$feature_names, must.include = "bar")

  data = data.frame(..row_id = task$row_ids)
  task$cbind(data)
  expect_equal(task$ncol, 7L)

  task$cbind(iris[, character()])
  expect_equal(task$ncol, 7L)

  task$cbind(data.table())
  expect_equal(task$ncol, 7L)

  y = task$data(cols = task$target_names, rows = shuffle(task$row_ids))
  task$cbind(y)
  expect_equal(task$ncol, 7L)
  expect_disjunct(task$feature_names, task$target_names)

  # cbind to subsetted task
  task = tsk("iris")$filter(1:120)
  backend = data.table(x = runif(120))
  task$cbind(backend)

  expect_equal(iris_col_hashes, task$col_hashes[names(iris_col_hashes)])

  # cbind 0-row data (#461)
  task = tsk("iris")$filter(integer())
  task$cbind(data.frame(x = integer()))
  expect_set_equal(c(task$target_names, task$feature_names), c(names(iris), "x"))
})

test_that("cbind/rbind works", {
  task = tsk("iris")
  data = data.table(..row_id = 1:150, foo = 150:1)

  task$cbind(data)
  expect_task(task)
  expect_equal(task$n_features, 5L) # "foo" was added as a feature
  expect_set_equal(c(task$feature_names, task$target_names), c(names(iris), "foo"))
  expect_data_table(task$data(), ncols = 6, any.missing = FALSE)

  task$rbind(cbind(data.table(..row_id = 201:210, foo = 99L), iris[1:10, ]))
  expect_task(task)
  expect_set_equal(task$row_ids, c(1:150, 201:210))
  expect_equal(task$n_features, 5L) # adding rows doesn't change #features
  expect_data_table(task$data(), ncols = 6, nrows = 160, any.missing = FALSE)

  # auto generated ids
  task = tsk("zoo")
  newdata = task$data(1)
  newdata$animal = "boy"
  task$rbind(newdata)
  expect_set_equal(task$row_ids, 1:102)
})

test_that("filter works", {
  task = tsk("iris")
  task$filter(1:100)
  expect_equal(task$nrow, 100L)
  expect_equal(task$row_ids, 1:100)

  task$filter(91:150)
  expect_equal(task$nrow, 60L)
  expect_equal(task$row_ids, 91:150)

  task$filter(91)
  expect_equal(task$nrow, 1L)
  expect_equal(task$row_ids, 91)
  expect_integer(task$row_ids) # #1285
  expect_data_table(task$data(), nrows = 1L, any.missing = FALSE)
})

test_that("select works", {
  task = tsk("iris")
  task$select(setdiff(task$feature_names, "Sepal.Length"))
  expect_equal(task$ncol, 4L)

  expect_error(task$select(c("Sepal.Width", "foobar")))

  task$select("Sepal.Width")
  expect_equal(task$feature_names, "Sepal.Width")

  expect_error(task$select(1:4), "character")
  expect_error(task$select("xxx", "subset"))
})

test_that("rename works", {
  task = tsk("iris")
  old = names(iris)
  new = paste0("xx_", old)
  task$rename(old, new)

  expect_set_equal(task$feature_names, setdiff(new, "xx_Species"))
  expect_equal(task$target_names, "xx_Species")
  expect_task_classif(task)
})

test_that("stratify works", {
  task = tsk("iris")
  expect_false("strata" %chin% task$properties)
  expect_null(task$strata)

  task$col_roles$stratum = task$target_names
  expect_true("strata" %chin% task$properties)
  tab = task$strata
  expect_data_table(tab, ncols = 2, nrows = 3)
  expect_list(tab$row_id, "integer")
})

test_that("groups/weights work", {
  b = as_data_backend(data.table(
    x = runif(20),
    y = runif(20),
    w = runif(20),
    o = runif(20),
    g = sample(letters[1:2], 20, replace = TRUE)
  ))
  task = TaskRegr$new("test", b, target = "y")
  task$set_row_roles(16:20, character())

  expect_false("groups" %in% task$properties)
  expect_false("weights" %in% task$properties)
  expect_false("weights_learner" %in% task$properties)
  expect_false("offset" %chin% task$properties)
  expect_null(task$groups)
  expect_null(task$weights_learner)

  # weight
  task$col_roles$weights_learner = "w"
  expect_subset("weights_learner", task$properties)
  expect_data_table(task$weights_learner, ncols = 2, nrows = 15)
  expect_numeric(task$weights_learner$weight, any.missing = FALSE)

  task$col_roles$weights_learner = character()
  expect_true("weights_learner" %nin% task$properties)

  # group
  task$col_roles$group = "g"
  expect_subset("groups", task$properties)
  expect_data_table(task$groups, ncols = 2, nrows = 15)
  expect_subset(task$groups$group, c("a", "b"))

  task$col_roles$group = character()
  expect_true("groups" %nin% task$properties)

  expect_error(
    {
      task$col_roles$weights_learner = c("w", "g")
    },
    "up to one"
  )
})

test_that("col roles are valid", {
  b = as_data_backend(data.table(
    y = runif(20),
    logical = sample(c(TRUE, FALSE), 20, replace = TRUE),
    numeric = runif(20),
    integer = sample(1:3, 20, replace = TRUE),
    factor = factor(sample(letters[1:3], 20, replace = TRUE))
  ))
  task = TaskRegr$new("test", b, target = "y")

  # weight
  expect_error(task$set_col_roles("logical", roles = "weights_learner"), "type")
  expect_error(task$set_col_roles("factor", roles = "weights_learner"), "type")
  expect_error(
    task$set_col_roles(c("integer", "numeric"), roles = "weights_learner"),
    "There may only be up to one column with role"
  )

  expect_error(task$set_col_roles("logical", roles = "weights_measure"), "type")
  expect_error(task$set_col_roles("factor", roles = "weights_measure"), "type")
  expect_error(
    task$set_col_roles(c("integer", "numeric"), roles = "weights_measure"),
    "There may only be up to one column with role"
  )

  # name
  expect_error(task$set_col_roles("logical", roles = "name"), "type")
  expect_error(task$set_col_roles("integer", roles = "name"), "type")
  expect_error(task$set_col_roles("numeric", roles = "name"), "type")
  expect_error(
    task$set_col_roles(c("integer", "numeric"), roles = "name"),
    "There may only be up to one column with role"
  )

  # group
  expect_error(
    task$set_col_roles(c("numeric", "factor"), roles = "group"),
    "There may only be up to one column with role"
  )

  # missing weights
  b = as_data_backend(data.table(y = runif(20), numeric = c(runif(19), NA_real_)))
  task = TaskRegr$new("test", b, target = "y")

  expect_error(task$set_col_roles("numeric", roles = "weights_learner"), "missing")
  expect_error(task$set_col_roles("numeric", roles = "weights_measure"), "missing")

  # negative weights
  b = as_data_backend(data.table(y = runif(20), numeric = c(runif(19), -10)))
  task = TaskRegr$new("test", b, target = "y")

  expect_error(task$set_col_roles("numeric", roles = "weights_learner"), "is not")
  expect_error(task$set_col_roles("numeric", roles = "weights_measure"), "is not")

  # target classif
  b = as_data_backend(data.table(
    y = factor(sample(letters[1:3], 20, replace = TRUE)),
    numeric = runif(20)
  ))
  task = as_task_classif(b, target = "y")

  expect_error(
    {
      task$col_roles = insert_named(task$col_roles, list(target = "numeric", feature = "y"))
    },
    "must be a factor or ordered factor"
  )

  expect_error(task$set_col_roles("numeric", roles = "target"), "up to one column with")

  # target regr
  b = as_data_backend(data.table(
    y = runif(20),
    factor = factor(sample(letters[1:3], 20, replace = TRUE))
  ))
  task = TaskRegr$new("test", b, target = "y")

  expect_error(
    {
      task$col_roles = insert_named(task$col_roles, list(target = "factor", feature = "y"))
    },
    "numeric or integer column"
  )

  expect_error(task$set_col_roles("factor", roles = "target"), "up to one column with")
})

test_that("ordered factors (#95)", {
  df = data.frame(
    x = c(1, 2, 3),
    y = factor(letters[1:3], levels = letters[1:3], ordered = TRUE),
    z = factor(c("M", "R", "R"), levels = c("M", "R"))
  )
  b = as_data_backend(df)
  task = TaskClassif$new(id = "id", backend = b, target = "z")
  expect_subset(c("numeric", "ordered", "factor"), task$col_info$type)
  expect_set_equal(task$col_info[id == "z", levels][[1L]], c("M", "R"))
  expect_set_equal(task$col_info[id == "y", levels][[1L]], letters[1:3])
})

test_that("as.data.table", {
  task = tsk("iris")
  expect_data_table(as.data.table(task), nrows = 150, ncols = 5)
})

test_that("extra factor levels are stored (#179)", {
  dt = data.table(
    x1 = factor(letters[1:5], levels = letters[5:1]),
    x2 = factor(letters[1:5], levels = letters),
    x3 = letters[1:5],
    target = 1:5
  )
  task = TaskRegr$new("extra_factor_levels", as_data_backend(dt), "target")
  expect_equal(task$levels("x2")$x2, letters)
})

test_that("task$droplevels works", {
  dt = data.table(
    x1 = factor(letters[1:3]),
    target = 1:3
  )
  task = TaskRegr$new("droplevels", as_data_backend(dt), "target")

  task$filter(1:2)
  expect_equal(task$nrow, 2L)
  expect_equal(task$levels("x1")$x1, letters[1:3])
  task$droplevels()
  expect_equal(task$levels("x1")$x1, letters[1:2])
})

test_that("task$missings() works", {
  task = tsk("pima")
  x = task$missings()
  y = map_int(task$data(), count_missing)
  expect_equal(x, y[match(names(x), names(y))])

  # issue #862
  task = tsk("iris")$cbind(data.frame(x = 1:150))$rename("x", "y")
  missings = task$missings(cols = character())
  expect_integer(missings, len = 0L)
  testthat::expect_named(missings)
})

test_that("task$feature_types preserves key (#193)", {
  task = tsk("iris")$select(character())$cbind(iris[1:4])
  expect_data_table(task$feature_types, ncols = 2L, nrows = 4L, key = "id")
})

test_that("switch columns on and off (#301)", {
  task = tsk("iris")
  expect_equal(task$n_features, 4L)
  task$col_roles$feature = setdiff(task$col_roles$feature, "Sepal.Length")
  expect_equal(task$n_features, 3L)
  task$cbind(data.table(x = 1:150))
  expect_equal(task$n_features, 4L)
  task$col_roles$feature = union(task$col_roles$feature, "Sepal.Length")
  expect_equal(task$n_features, 5L)
  expect_data_table(task$data(), ncols = 6, nrows = 150, any.missing = FALSE)
})

test_that("row roles setters", {
  task = tsk("iris")

  expect_error(
    {
      task$row_roles$use = "foo"
    },
    "integerish"
  )
  expect_error(
    {
      task$row_roles$foo = 1L
    },
    "extra elements"
  )

  task$row_roles$use = 1:20
  expect_equal(task$nrow, 20L)
})

test_that("row_ids_backend returns all backend rows", {
  task = tsk("iris")

  expect_set_equal(task$row_ids, task$row_ids_backend)

  task$filter(1:100)
  expect_set_equal(task$row_ids, 1:100)
  expect_set_equal(task$row_ids_backend, 1:150)

  task$set_row_roles(1:50, remove_from = "use")
  expect_set_equal(task$row_ids, 51:100)
  expect_set_equal(task$row_ids_backend, 1:150)

  expect_error(
    {
      task$row_ids_backend = 1:10
    },
    "read-only"
  )
})

test_that("col roles getters/setters", {
  task = tsk("iris")

  expect_error(
    {
      task$col_roles$feature = "foo"
    },
    "subset"
  )

  expect_error({
    task$col_roles$foo = "Species"
  })

  task$col_roles$feature = setdiff(task$col_roles$feature, "Sepal.Length")
  expect_false("Sepal.Length" %chin% task$feature_names)
})

test_that("Task$row_names", {
  task = tsk("mtcars")
  tab = task$row_names
  expect_data_table(tab, any.missing = FALSE, ncols = 2, nrows = task$nrow)
  expect_integer(tab$row_id, unique = TRUE)
  expect_character(tab$row_name)

  tab = task$filter(1:10)$row_names
  expect_data_table(tab, any.missing = FALSE, ncols = 2, nrows = task$nrow)
  expect_integer(tab$row_id, unique = TRUE)
  expect_character(tab$row_name)
})

test_that("Task$set_row_roles", {
  task = tsk("pima")

  task$set_row_roles(1:10, remove_from = "use")
  expect_true(all(1:10 %nin% task$row_ids))

  task$set_row_roles(1:10, add_to = "use")
  expect_true(all(1:10 %in% task$row_ids))
})


test_that("Task$set_col_roles", {
  task = tsk("pima")
  expect_equal(task$n_features, 8L)

  task$set_col_roles("mass", remove_from = "feature")
  expect_equal(task$n_features, 7L)
  expect_true("mass" %nin% task$feature_names)

  task$set_col_roles("mass", add_to = "feature")
  expect_equal(task$n_features, 8L)
  expect_true("mass" %chin% task$feature_names)

  task$set_col_roles("age", roles = "weights_learner")
  expect_equal(task$n_features, 7L)
  expect_true("age" %nin% task$feature_names)
  expect_data_table(task$weights_learner)

  task$set_col_roles("age", add_to = "feature", remove_from = "weights_learner")
  expect_equal(task$n_features, 8L)
  expect_true("age" %chin% task$feature_names)
  expect_null(task$weights_learner)
})

test_that("$add_strata", {
  task = tsk("mtcars")
  expect_equal(task$col_roles$stratum, character())

  task$add_strata("mpg", bins = 5)
  expect_set_equal(task$col_roles$stratum, "..stratum_mpg")
  expect_data_table(task$strata, nrows = 5)

  task$add_strata("am", bins = 3)
  expect_set_equal(task$col_roles$stratum, c("..stratum_mpg", "..stratum_am"))

  task = tsk("mtcars")
  task$add_strata(c("mpg", "am"), bins = c(2, 5))
  expect_set_equal(task$col_roles$stratum, c("..stratum_mpg", "..stratum_am"))
})

test_that("column labels", {
  task = tsk("iris")
  expect_character(task$col_info$label)
  expect_true(allMissing(task$col_info$label))
  expect_true(allMissing(task$labels))

  task$labels = c(Species = "sp")
  expect_equal(task$labels[["Species"]], "sp")
  expect_equal(count_missing(task$labels), 4L)

  fn = task$feature_names
  task$labels = set_names(toupper(fn), fn)
  expect_equal(unname(task$labels), c("sp", toupper(fn)))

  expect_error(
    {
      task$labels = c(foo = "as")
    },
    "names"
  )

  dt = data.table(id = c(task$target_names, task$feature_names))
  dt$label = tolower(dt$id)

  task$labels = dt
  expect_equal(
    unname(task$labels),
    tolower(c(task$target_names, task$feature_names))
  )
})

test_that("set_levels", {
  task = tsk("penguins")

  new_lvls = c("male", "female", "missing")
  task$set_levels(list(sex = new_lvls))

  tab = task$col_info[list("sex")]
  expect_equal(tab$levels[[1]], new_lvls)
  expect_equal(tab$fix_factor_levels[[1]], TRUE)
  expect_equal(levels(task$data(1)$sex), new_lvls)
  expect_equal(levels(head(task)$sex), new_lvls)

  new_lvls = c("female", "nothing")
  task$set_levels(list(sex = new_lvls))

  tab = task$col_info[list("sex")]
  expect_equal(tab$levels[[1]], new_lvls)
  expect_equal(tab$fix_factor_levels[[1]], TRUE)
  expect_equal(as.integer(task$data(1)$sex), NA_integer_)
  expect_equal(as.integer(head(task, 1)$sex), NA_integer_)
  expect_equal(levels(task$data(1)$sex), new_lvls)
  expect_equal(levels(head(task, 1)$sex), new_lvls)
})

test_that("special chars in feature names (#697)", {
  expect_error(
    TaskRegr$new("test", data.table(`%asd` = 1:3, t = 3:1), target = "t"),
    "special character"
  )
})

test_that("head/tail", {
  task = tsk("iris")
  expect_data_table(head(task, n = 3), nrows = 3)
  expect_data_table(head(task, n = -3), nrows = task$nrow - 3)

  expect_data_table(tail(task, n = 3), nrows = 3)
  expect_data_table(tail(task, n = -3), nrows = task$nrow - 3)

  expect_data_table(head(task, n = Inf), nrows = 150)
  expect_data_table(tail(task, n = Inf), nrows = 150)

  expect_data_table(head(task, n = -Inf), nrows = 0)
  expect_data_table(tail(task, n = -Inf), nrows = 0)
})

test_that("Roles get printed (#877)", {
  task = tsk("iris")
  task$col_roles$weights_learner = "Petal.Width"
  expect_output(print(task), "Weights/Learner: Petal.Width")
})

test_that("validation task is cloned", {
  task = tsk("iris")
  task$internal_valid_task = c(1:10, 51:60, 101:110)
  task2 = task$clone(deep = TRUE)
  expect_different_address(task$internal_valid_task, task2$internal_valid_task)
  # TODO: maybe re-enable after $weights has been removed?
  # expect_equal(task$internal_valid_task, task2$internal_valid_task)
})

test_that("task is cloned when assining internal validation task", {
  task = tsk("iris")
  task$internal_valid_task = task
  expect_false(identical(task, task$internal_valid_task))
})

test_that("validation task changes a task's hash", {
  task = tsk("iris")
  h1 = task$hash
  task$internal_valid_task = task$clone(deep = TRUE)$filter(1:10)
  h2 = task$hash
  expect_false(h1 == h2)
})

test_that("compatibility checks on internal_valid_task", {
  d1 = data.table(x = 1:10, y = 1:10)
  d2 = data.table(x = rnorm(10), y = 1:10)
  d3 = data.table(x1 = rnorm(10), y = 1:10)

  t1 = as_task_regr(d1, target = "y")
  t2 = as_task_regr(d2, target = "y")
  t3 = as_task_regr(d3, target = "y")
  expect_error(
    {
      t1$internal_valid_task = t2
    },
    "differs from the type"
  )
  expect_error(
    {
      t1$internal_valid_task = t3
    },
    "not present"
  )
})

test_that("can NULL validation task", {
  task = tsk("iris")
  task$internal_valid_task = 1
  task$internal_valid_task = NULL
  expect_equal(length(task$row_ids), 149)
})

test_that("internal_valid_task is printed", {
  task = tsk("iris")
  task$internal_valid_task = c(1:10, 51:60, 101:110)
  expect_output(print(task), "Validation Task: \\(30x5\\)")
})

test_that("task hashes during resample", {
  task = orig = tsk("iris")
  task = orig$clone(deep = TRUE)
  resampling = rsmp("holdout")
  resampling$instantiate(task)
  task$internal_valid_task = resampling$test_set(1)
  learner = lrn("classif.debug", validate = "test")
  expect_equal(resampling_task_hashes(task, resampling, learner), task$hash)
})

test_that("integer vector can be passed to internal_valid_task", {
  task = tsk("iris")$filter(1:5)
  task$internal_valid_task = 5
  expect_permutation(task$row_ids, 1:4)
  expect_equal(task$internal_valid_task$row_ids, 5)
})

test_that("cbind supports non-standard primary key (#961)", {
  tbl = data.table(x = runif(10), y = runif(10), myid = 1:10)
  b = as_data_backend(tbl, primary_key = "myid")
  task = as_task_regr(b, target = "y")
  task$cbind(data.table(x1 = 10:1))
  expect_true("x1" %chin% task$feature_names)
})

test_that("task weights", {
  # proper deprecation of rename weights -> weights_learner
  task = tsk("mtcars")
  task$cbind(data.table(w = runif(32)))
  expect_warning(task$weights)

  task$set_col_roles("w", "weights_learner")
  expect_data_table(task$weights_learner)
  expect_subset("weights_learner", task$properties)
  expect_task(task)
})

test_that("task$set_col_roles() with weights", {
  task = tsk("mtcars")
  task$cbind(data.table(w_lrn = runif(32), w_msr = runif(32)))

  # weights_learner
  task$set_col_roles("w_lrn", "weights_learner")
  expect_data_table(task$weights_learner)
  expect_subset("weights_learner", task$properties)
  expect_task(task)
  task$set_col_roles("w_lrn", remove_from = "weights_learner")
  expect_null(task$weights_learner)
  expect_false("weights_learner" %in% task$properties)

  # weights_measure
  task$set_col_roles("w_msr", "weights_measure")
  expect_data_table(task$weights_measure)
  expect_subset("weights_measure", task$properties)
  expect_task(task)
  task$set_col_roles("w_msr", remove_from = "weights_measure")
  expect_null(task$weights_measure)
  expect_false("weights_measure" %in% task$properties)

  # Test assigning the same column to both
  task = tsk("mtcars")$cbind(data.table(w = runif(32)))
  task$set_col_roles("w", add_to = c("weights_learner", "weights_measure"))
  expect_equal(task$weights_learner$weight, task$backend$data(task$row_ids, "w")$w)
  expect_equal(task$weights_measure$weight, task$backend$data(task$row_ids, "w")$w)
  expect_subset(c("weights_learner", "weights_measure"), task$properties)
})

test_that("task$set_col_roles errors with wrong weights", {
  dd = iris
  dd$ww_chr = sample(letters, 150, replace = TRUE)
  dd$ww_na = 1:150
  dd$ww_na[1] = NA
  dd$ww_neg = 1:150
  dd$ww_neg[1] = -99
  tt = as_task_classif(dd, target = "Species")

  expect_error(tt$set_col_roles("ww_chr", "weights_learner"), "Must be of type")
  expect_error(tt$set_col_roles("ww_chr", "weights_measure"), "Must be of type")

  expect_error(tt$set_col_roles("ww_na", "weights_learner"), "missing values")
  expect_error(tt$set_col_roles("ww_na", "weights_measure"), "missing values")

  expect_error(tt$set_col_roles("ww_neg", "weights_learner"), "is not >= 0")
  expect_error(tt$set_col_roles("ww_neg", "weights_measure"), "is not >= 0")
})

test_that("weights printing", {
  task = tsk("mtcars")
  task$cbind(data.table(w_lrn = runif(32), w_msr = runif(32)))
  task$set_col_roles("w_lrn", "weights_learner")
  task$set_col_roles("w_msr", "weights_measure")
  expect_output(print(task), "Weights/Learner: w_lrn")
  expect_output(print(task), "Weights/Measure: w_msr")
})

test_that("rbind with weights", {
  task = tsk("iris")
  task$cbind(data.table(w_lrn = 1:150, w_msr = 150:1))
  task$set_col_roles("w_lrn", "weights_learner")
  task$set_col_roles("w_msr", "weights_measure")

  original_rows = task$nrow
  original_row_ids = task$row_ids
  original_weights_lrn = task$weights_learner
  original_weights_msr = task$weights_measure

  new_data = task$data(1:10) # includes target, features, weights
  new_data$..row_id = 151:160
  new_data$w_lrn = 1001:1010
  new_data$w_msr = 2010:2001
  new_data$Petal.Length = new_data$Petal.Length + 100 # Change a feature to check backend update
  new_species = sample(c("setosa", "versicolor", "virginica", "new_level"), 10, replace = TRUE) # new factor level
  new_species[1] = "new_level" # include new level at least once
  new_data$Species = factor(new_species)

  task$rbind(new_data)

  expect_equal(task$nrow, original_rows + 10)
  expect_set_equal(task$row_ids, c(original_row_ids, 151:160))
  expect_true("weights_learner" %in% task$properties)
  expect_true("weights_measure" %in% task$properties)

  # Check combined weights
  combined_weights_lrn = task$weights_learner
  expect_shape(combined_weights_lrn, nrow = task$nrow)
  expect_equal(setkeyv(combined_weights_lrn[row_id %in% original_row_ids], "row_id"), original_weights_lrn)
  expect_equal(combined_weights_lrn[row_id %in% 151:160]$weight, 1001:1010)

  combined_weights_msr = task$weights_measure
  expect_shape(combined_weights_msr, nrow = task$nrow)
  # expect_equal(combined_weights_msr[row_id %in% original_row_ids], original_weights_msr) # ordering issue with join
  expect_equal(setkeyv(combined_weights_msr[list(original_row_ids), on = "row_id"], "row_id"), original_weights_msr)
  expect_equal(combined_weights_msr[row_id %in% 151:160]$weight, 2010:2001)

  # Check feature and factor level update
  expect_true(any(task$data(151:160)$Petal.Length > 100))
  expect_subset("new_level", task$levels("Species")$Species)
})

test_that("cbind with weights", {
  task = tsk("iris")
  original_n_features = task$n_features
  original_ncol = task$ncol

  new_cols_data = data.table(
    ..row_id = task$row_ids,
    w_lrn = 1:150,
    w_msr = 150:1,
    new_feature = rnorm(150)
  )

  task$cbind(new_cols_data)

  expect_equal(task$n_features, original_n_features + 3) # new_feature added automatically
  expect_equal(task$ncol, original_ncol + 3)
  expect_true("new_feature" %in% task$feature_names)
  expect_false("weights_learner" %in% task$properties) # Role not assigned yet
  expect_false("weights_measure" %in% task$properties)
  expect_null(task$weights_learner)
  expect_null(task$weights_measure)

  # Assign roles
  task$set_col_roles("w_lrn", "weights_learner")
  task$set_col_roles("w_msr", "weights_measure")

  expect_true("weights_learner" %in% task$properties)
  expect_true("weights_measure" %in% task$properties)

  weights_lrn = task$weights_learner
  expect_shape(weights_lrn, nrow = task$nrow)
  expect_equal(weights_lrn$weight, 1:150)

  weights_msr = task$weights_measure
  expect_shape(weights_msr, nrow = task$nrow)
  expect_equal(weights_msr$weight, 150:1)

  # Check that original features/target are still there
  expect_equal(task$target_names, "Species")
  expect_true(all(c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width") %in% task$feature_names))
})

test_that("$select changes hash", {
  task = tsk("iris")
  h1 = task$hash
  task$select("Petal.Length")
  h2 = task$hash
  expect_false(h1 == h2)
})

test_that("$characteristics works", {
  task = tsk("spam")
  characteristics = list(foo = 1, bar = "a")
  task$characteristics = characteristics

  expect_snapshot(task)
  expect_equal(task$characteristics, characteristics)

  tsk_1 = tsk("spam")
  tsk_1$characteristics = list(n = 300)
  tsk_2 = tsk("spam")
  tsk_2$characteristics = list(n = 200)

  expect_true(tsk_1$hash != tsk_2$hash)

  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 3)

  design = benchmark_grid(
    tasks = list(tsk_1, tsk_2),
    learners = learner,
    resamplings = resampling
  )

  bmr = benchmark(design)
  tab = as.data.table(bmr, task_characteristics = TRUE)
  expect_names(names(tab), must.include = "n")
  expect_subset(tab$n, c(300, 200))

  tsk_1$characteristics = list(n = 300, f = 3)
  tsk_2$characteristics = list(n = 200, f = 2)

  design = benchmark_grid(
    tasks = list(tsk_1, tsk_2),
    learners = learner,
    resamplings = resampling
  )

  bmr = benchmark(design)
  tab = as.data.table(bmr, task_characteristics = TRUE)
  expect_names(names(tab), must.include = c("n", "f"))
  expect_subset(tab$n, c(300, 200))
  expect_subset(tab$f, c(2, 3))

  tsk_1$characteristics = list(n = 300, f = 2)
  tsk_2$characteristics = list(n = 200)

  design = benchmark_grid(
    tasks = list(tsk_1, tsk_2),
    learners = learner,
    resamplings = resampling
  )

  bmr = benchmark(design)
  tab = as.data.table(bmr, task_characteristics = TRUE)

  expect_names(names(tab), must.include = c("n", "f"))
  expect_subset(tab$n, c(300, 200))
  expect_subset(tab$f, c(2, NA_real_))
})

test_that("warn when internal valid task has 0 obs", {
  task = tsk("iris")
  # expect_warning({task$internal_valid_task = 151}, "has 0 observations")
})


test_that("$data() is not called during task construction", {
  tbl = data.table(x = 1:10, y = 1:10, ..row_id = 1:10)
  backend = R6Class(
    "DataBackendTest",
    inherit = DataBackendDataTable,
    cloneable = FALSE,
    public = list(
      data = function(rows, cols) {
        stop("Bug")
      }
    )
  )$new(tbl, "..row_id")
  task = as_task_regr(backend, target = "y")
  expect_class(task, "Task")
})

test_that("row_hash works correctly", {
  task = tsk("pima")
  original_hash = task$row_hash

  # hash should change when row ids change with filter
  task$filter(1:100)
  expect_false(identical(task$row_hash, original_hash))
  new_hash = task$row_hash

  # hash should change when row roles change with set_row_roles
  task$set_row_roles(1:50, roles = "use")
  expect_false(identical(task$row_hash, new_hash))

  # hash should be read-only
  expect_error(
    {
      task$row_hash = "new_hash"
    },
    "is read-only"
  )
})

test_that("row_ids_backend works correctly", {
  task = tsk("pima")

  # should not change when filtering
  task$filter(1:100)
  expect_set_equal(task$row_ids_backend, 1:768)

  # should not change when modifying row roles
  task$set_row_roles(1:50, remove_from = "use")
  expect_set_equal(task$row_ids_backend, 1:768)

  # should be read-only
  expect_error(
    {
      task$row_ids_backend = 1:10
    },
    "read-only"
  )

  # should match backend$rownames
  expect_set_equal(task$row_ids_backend, task$backend$rownames)

  # should include all rows even after multiple filters
  task$filter(1:50)
  task$filter(1:25)
  expect_set_equal(task$row_ids_backend, 1:768)
})

test_that("$levels with cols works (#1323)", {
  task = TaskClassif$new(
    id = "example",
    target = "target",
    backend = data.table(
      target = factor(c("S", "T", "U")),
      x = ordered(c("A", "B", "C")),
      y = factor(c("L", "M", "N"))
    )
  )

  expect_equal(task$levels(), list(target = c("S", "T", "U"), x = c("A", "B", "C"), y = c("L", "M", "N")))
  expect_equal(task$levels(cols = c("y", "x")), list(y = c("L", "M", "N"), x = c("A", "B", "C")))
})

test_that("$materialize_view works", {
  task = tsk("iris")
  task$select(head(task$feature_names, 2))
  task$materialize_view()

  expect_backend(task$backend)
  expect_task(task)
  expect_set_equal(task$feature_names, c("Petal.Length", "Petal.Width"))
  expect_set_equal(task$target_names, "Species")
  expect_data_table(task$col_info, key = "id")
  expect_set_equal(task$col_info$id, c(task$backend$primary_key, task$feature_names, task$target_names))

  learner = lrn("classif.featureless")$train(task)
  expect_set_equal(learner$model$features, c("Petal.Length", "Petal.Width"))
})

test_that("$materialize_view works with internal valid task", {
  task = tsk("iris")
  task$select(head(task$feature_names, 2))
  task$filter(1:140)
  task$internal_valid_task = 1:10
  task$materialize_view(TRUE)

  iv = task$internal_valid_task
  expect_backend(iv$backend)
  expect_task(iv)
  expect_set_equal(iv$feature_names, c("Petal.Length", "Petal.Width"))
  expect_set_equal(iv$target_names, "Species")
  expect_data_table(iv$col_info, key = "id")
  expect_set_equal(iv$col_info$id, c(task$backend$primary_key, task$feature_names, task$target_names))
})

test_that("materialize_view works with duplicates", {
  task = tsk("iris")
  task2 = task$clone(deep = TRUE)
  task$filter(c(1, 1, 2))
  task2$filter(c(1, 1, 2))
  task2$materialize_view()
  expect_equal(task$data(), task2$data())
})

test_that("weights_measure + stratum works during resampling (#1405)", {
  data = cbind(datasets::iris, data.frame(w = rep(c(1, 10, 100), each = 50)))
  # 150 rows works, but 151 rows fails
  data = data[c(seq(150), 1), ]
  task = TaskClassif$new("iris_weights_measure", as_data_backend(data, target = "Species"), target = "Species")
  task$set_col_roles("w", "weights_measure")
  task$set_col_roles("Species", roles = c("target", "stratum"))
  expect_resample_result(resample(task, lrn("classif.featureless"), rsmp("cv", folds = 3)))
})

test_that("class ratios are not printed for large tasks (#1382)", {
  task = tsk("iris")
  old_threshold = getOption("mlr3.print_class_ratio_threshold")
  on.exit(options(mlr3.print_class_ratio_threshold = old_threshold))
  expect_output(print(task), "setosa \\(33%\\), versicolor \\(33%\\), virginica \\(33%\\)")

  options(mlr3.print_class_ratio_threshold = 0L)
  expect_output(print(task), "setosa, versicolor, virginica", fixed = TRUE)
})

test_that("id must not contain percent character (#1461)", {
  expect_error(TaskClassif$new("a %>% b", backend = iris, target = "Species"), "must not contain")
  task = tsk("iris")
  expect_error(task$id <- "a %>% b", "must not contain") # nolint
})

test_that("predict method works", {
  task = tsk("sonar")
  lrn = lrn("classif.featureless")$train(task)

  newdata = task$data(1:3)
  expect_factor(predict(lrn, newdata = newdata), len = 3)
  expect_factor(predict(lrn, newdata = newdata, predict_type = "response"), len = 3)
  expect_error(predict(lrn, newdata = newdata, predict_type = "prob"), "'prob'")

  lrn = lrn("classif.featureless", predict_type = "prob")$train(task)
  expect_factor(predict(lrn, newdata = newdata), len = 3)
  expect_factor(predict(lrn, newdata = newdata, predict_type = "response"), len = 3)
  expect_matrix(predict(lrn, newdata = newdata, predict_type = "prob"), nrows = 3, ncols = 2)

  expect_true(uniqueN(predict(lrn, newdata, method = "mode")) == 1L)
})

test_that("missing predictions are handled gracefully / classif", {
  task = tsk("sonar")
  learner = lrn("classif.debug", predict_missing = 1, predict_missing_type = "na", predict_type = "prob")
  learner$train(task)

  p = learner$predict(task)
  expect_factor(p$response, levels = task$class_names)
  expect_true(all(is.na(p$response)))
  expect_true(all(is.na(p$prob)))
  expect_error(p$score(), "missing")

  learner = lrn("classif.debug", predict_missing = 0.5, predict_missing_type = "omit", predict_type = "prob")
  learner$train(task)
  expect_error(learner$predict(task), "observations")
})

test_that("missing predictions are handled gracefully / regr", {
  task = tsk("mtcars")
  learner = lrn("regr.debug", predict_missing = 1, predict_missing_type = "na", predict_type = "se")
  learner$train(task)

  p = learner$predict(task)
  expect_numeric(p$response)
  expect_numeric(p$se)
  expect_true(all(is.na(p$response)))
  expect_true(all(is.na(p$se)))
  expect_error(p$score(), "missing")

  learner = lrn("regr.debug", predict_missing = 0.5, predict_missing_type = "omit", predict_type = "se")
  learner$train(task)
  expect_error(learner$predict(task), "observations")
})


test_that("predict_newdata with weights (#519)", {
  # we had a problem where predict did not work if weights were present in the task
  # especially with the "predict_newdata" function

  task = tsk("california_housing")
  task$set_col_roles("households", "weights_learner")
  task$set_col_roles("total_rooms", "weights_measure")
  learner = lrn("regr.featureless")
  learner$train(task)

  # predict with different API calls
  # normal predict on the task
  pred = learner$predict(task)
  expect_prediction(pred)
  expect_equal(pred$weights, tsk("california_housing")$data()$total_rooms)
  # w/o weights in the new-df
  pred = learner$predict_newdata(task$data())
  expect_prediction(pred)
  expect_null(pred$weights)
  # w weights in the new-df
  pred = learner$predict_newdata(task$data(cols = c(task$target_names, task$feature_names, "households")))
  expect_prediction(pred)
  expect_null(pred$weights)

  # now measure-weights are present
  pred = learner$predict_newdata(task$data(cols = c(task$target_names, task$feature_names, "total_rooms")))
  expect_prediction(pred)
  expect_equal(pred$weights, tsk("california_housing")$data()$total_rooms)
  # now both are present
  pred = learner$predict_newdata(task$data(
    cols = c(task$target_names, task$feature_names, "households", "total_rooms")
  ))
  expect_prediction(pred)
  expect_equal(pred$weights, tsk("california_housing")$data()$total_rooms)
})

test_that("parallel predict works", {
  skip_if_not_installed("future")
  task = tsk("sonar")
  lrn = lrn("classif.featureless")$train(task)

  lrn$parallel_predict = FALSE
  p1 = lrn$predict(task, row_ids = 20:1)

  lrn$parallel_predict = TRUE
  p2 = with_future(future::multisession, lrn$predict(task, row_ids = 20:1))

  expect_equal(as.data.table(p1), as.data.table(p2))
})

test_that("re-instantiating", {
  t1 = tsk("iris")
  t2 = tsk("california_housing")
  r = rsmp("cv", folds = 2)

  expect_resampling(r$instantiate(t1), task = t1)
  expect_resampling(r$instantiate(t2), task = t2)

  r = rsmp("custom")
  expect_error(r$instantiate(t1), "missing")

  expect_resampling(r$instantiate(t1, train_sets = list(1), test_sets = list(1)), task = t1)
  expect_resampling(r$instantiate(t2, train_sets = list(1), test_sets = list(2)), task = t2)
})

test_that("param_vals", {
  task = tsk("iris")
  r = rsmp("bootstrap", repeats = 100L, ratio = 1)

  expect_identical(r$param_set$values$ratio, 1)
  expect_identical(r$param_set$values$repeats, 100L)

  r$instantiate(task)
  expect_true(r$is_instantiated)
  expect_identical(r$iters, 100L)
  expect_integerish(r$train_set(100), len = task$nrow)

  expect_resampling(r)

  expect_error(
    {
      r$param_set$values = list(repeats = 10L)
      r$param_set$get_values()
    },
    "ratio"
  )

  expect_error(
    {
      r$param_set$values = list(ratio = 0.5, repeats = 10L, foobar = 12)
    },
    "foobar"
  )
})

test_that("hashing", {
  task = tsk("iris")
  keys = setdiff(mlr_resamplings$keys(), c("custom", "custom_cv", "ordered_holdout"))

  for (key in keys) {
    r = rsmp(key)

    # nolint next
    with_seed(123L, r$instantiate(task))
    hash = r$hash
    expect_string(hash, pattern = "^[a-z0-9]+$")

    # nolint next
    with_seed(123L, r$instantiate(task))
    expect_identical(r$hash, hash)

    if (key != "insample") {
      # nolint next
      with_seed(124L, r$instantiate(task))
      expect_false(identical(r$hash, hash))
    }
  }
})

test_that("cloning", {
  task = tsk("iris")
  keys = setdiff(mlr_resamplings$keys(), c("custom", "custom_cv", "ordered_holdout"))

  for (key in keys) {
    r = rsmp(key)$instantiate(task)

    clone = r$clone(deep = TRUE)
    expect_different_address(r$param_set, clone$param_set)
    if (is.data.table(r$instance)) {
      expect_different_address(r$instance, clone$instance)
    }
  }
})

test_that("integer grouping col (#396)", {
  df = data.frame(
    id = rep(1L:10L, each = 2),
    x = rnorm(20)
  )

  tsk = TaskRegr$new(id = "task", backend = df, target = "x")
  tsk$set_col_roles("id", "group")

  bs = rsmp("bootstrap", repeats = 10L, ratio = 1)
  bs$instantiate(tsk)

  set = bs$train_set(1)
  expect_integer(set)
  expect_true(every(split(seq_row(df), f = df$id), function(x) all(x %in% set) || all(x %nin% set)))

  set = bs$test_set(1)
  expect_integer(set)
  expect_true(every(split(seq_row(df), f = df$id), function(x) all(x %in% set) || all(x %nin% set)))
})

test_that("as.data.table.Resampling", {
  r = rsmp("bootstrap")
  r$instantiate(tsk("mtcars"))

  tab = as.data.table(r)
  expect_data_table(tab, ncols = 3)
  expect_names(names(tab), permutation.of = c("set", "iteration", "row_id"))
  expect_integer(tab$iteration, any.missing = FALSE)
  expect_factor(tab$set, levels = c("train", "test"), any.missing = FALSE)
  expect_integer(tab$row_id, any.missing = FALSE)
})

test_that("custom_cv", {
  task = tsk("penguins")
  ccv = rsmp("custom_cv")

  f = task$data(cols = "island")[[1L]]
  ccv$instantiate(task, f = f)
  expect_resampling(ccv, task = task)
  expect_equal(ccv$iters, 3L)
  expect_list(ccv$instance, "integer", len = 3)
  expect_names(names(ccv$instance), permutation.of = levels(f))

  ccv$instantiate(task, col = "island")
  expect_resampling(ccv, task = task)
  expect_equal(ccv$iters, 3L)
  expect_list(ccv$instance, "integer", len = 3)
  names(ccv$instance)
  expect_names(names(ccv$instance), permutation.of = levels(f))

  task = task$clone(TRUE)$filter(1:10)
  f = factor(rep(letters[1:3], each = 3))
  expect_error(ccv$instantiate(task, f), "length")
  f[10] = NA
  ccv$instantiate(task, f)
  expect_data_table(as.data.table(ccv), nrows = 3L * 9L)

  f[] = NA
  expect_error(ccv$instantiate(task, f), "only missing")
})

test_that("loo with groups", {
  task = tsk("penguins")
  task$set_col_roles("island", add_to = "group")
  loo = rsmp("loo")
  loo$instantiate(task)
  expect_equal(loo$iters, 3L)

  islands = cbind(row_id = task$row_ids, task$data(cols = "island"))
  tab = merge(as.data.table(loo), islands, by = "row_id")
  expect_true(all(tab[, .(n_islands = uniqueN(island)), by = row_id]$n_islands == 1L))
})

test_that("task_row_hash in Resampling works correctly", {
  task = tsk("penguins")
  resampling = rsmp("cv", folds = 2)

  # task_row_hash should be NA before instantiation
  expect_identical(resampling$task_row_hash, NA_character_)

  # task_row_hash should be set after instantiation
  resampling$instantiate(task)
  expect_identical(resampling$task_row_hash, task$row_hash)
})

test_that("DataBackendCbind", {
  data = as.data.table(iris)
  data$Petal.Length[91:120] = NA
  data$id = 1:150

  b1 = as_data_backend(data[, -"Sepal.Length"], primary_key = "id")
  b2 = as_data_backend(data[, c("id", "Sepal.Length")], primary_key = "id")
  b = DataBackendCbind$new(b1, b2)
  expect_backend(b)
  expect_iris_backend(b, n_missing = 30L)

  expect_set_equal(b$rownames, 1:150)
  expect_set_equal(b$colnames, names(data))
  expect_data_table(b$data(b$rownames, b$colnames), nrows = 150, ncols = 6)
  expect_set_equal(b$distinct(b$rownames, "Species")$Species, distinct_values(iris$Species, drop = FALSE))

  x = b$missings(b$rownames, c("Petal.Width", "Petal.Length"))
  expect_equal(x, set_names(c(0L, 30L), c("Petal.Width", "Petal.Length")))

  expect_equal(b$col_hashes, c(b1$col_hashes, b2$col_hashes)[setdiff(b$colnames, b$primary_key)])
})

test_that("issue #124", {
  task = tsk("iris")
  newcols = cbind(data.table(col = 1:150))
  task$select(character())$cbind(newcols)
  expect_data_table(task$data(cols = "col"), ncols = 1L, nrows = 150L)
})

test_that("cbind backends with same columns", {
  data = as.data.table(iris)
  data$id = 1:150
  data1 = copy(data)
  data2 = copy(data)

  data1$Petal.Width = NA
  b1 = as_data_backend(data1[, -"Sepal.Length"], primary_key = "id")

  data2$Sepal.Width = NA
  b2 = as_data_backend(data2[, c("id", "Sepal.Length", "Sepal.Width", "Petal.Width")], primary_key = "id")

  rows = 1:10
  cols = b2$colnames

  b = DataBackendCbind$new(b1, b2)
  expect_backend(b)
  data = b$head(Inf)
  expect_false(anyMissing(data$Petal.Width))
  expect_true(allMissing(data$Sepal.Width))

  expect_equal(b$col_hashes, insert_named(b1$col_hashes, b2$col_hashes)[setdiff(b$colnames, b$primary_key)])
})

test_that("Backends with different rows", {
  data = as.data.table(iris)
  data$id = 1:150

  b1 = as_data_backend(data[1:20, -"Sepal.Length"], primary_key = "id")
  b2 = as_data_backend(data[1:10, c("id", "Sepal.Length")], primary_key = "id")

  b = DataBackendCbind$new(b1, b2)

  expect_set_equal(b$colnames, c(names(iris), "id"))
  expect_set_equal(b$rownames, 1:20)

  expect_data_table(b$head(Inf), nrows = 20, ncols = 6)

  expect_equal(b$col_hashes, c(b1$col_hashes, b2$col_hashes)[setdiff(b$colnames, b$primary_key)])
})

test_that("Basic ops on iris task", {
  task = tsk("iris")
  expect_task(task)
  expect_task_supervised(task)
  expect_task_classif(task)
  expect_equal(task$target_names, "Species")
  expect_set_equal(task$class_names, levels(iris$Species))

  f = task$formula()
  expect_class(f, "formula")
  # expect_set_equal(attr(terms(f), "term.labels"), setdiff(names(iris), "Species"))
})

test_that("$class_names consider also inactive rows", {
  task = tsk("iris")
  task$set_row_roles(1:100, remove_from = "use")

  expect_set_equal(task$class_names, levels(iris$Species))
})

test_that("Factor levels are preserved in prediction", {
  task = tsk("iris")
  learner = lrn("classif.featureless")
  learner$predict_type = "prob"
  learner$train(task, 1:100)

  pred = as.data.table(learner$predict(task, 1:10))
  expect_factor(pred$truth, levels = levels(iris$Species), any.missing = FALSE)
  expect_factor(pred$response, levels = levels(iris$Species), any.missing = FALSE)
  expect_equal(levels(pred$truth), levels(pred$response))
  expect_numeric(pred$prob.virginica, lower = 0, upper = 0, any.missing = FALSE)

  pred = as.data.table(learner$predict(task, 101:150))
  expect_factor(pred$truth, levels = levels(iris$Species), any.missing = FALSE)
  expect_factor(pred$response, levels = levels(iris$Species), any.missing = FALSE)
  expect_equal(levels(pred$truth), levels(pred$response))
  expect_numeric(pred$prob.virginica, lower = 0, upper = 0, any.missing = FALSE)
})

test_that("Target is character/factor", {
  b = as_data_backend(iris)
  expect_error(TaskClassif$new("iris", backend = b, target = "Sepal.Length"), "Target column")
})

test_that("TaskClassif rejects NAs in target", {
  dt = data.table::data.table(x = 1:10, y = factor(c(NA, letters[1:9])))
  expect_message(as_task_classif(dt, target = "y"), "contains missing")
})

test_that("0 feature task", {
  b = as_data_backend(iris[, 5L, drop = FALSE])
  task = TaskClassif$new(id = "zero_feat_task", b, target = "Species")
  expect_output(print(task))
  b = task$backend
  expect_backend(b)
  expect_task(task)
  expect_task_supervised(task)
  expect_task_classif(task)
  expect_data_table(task$data(), ncols = 1L)

  lrn = lrn("classif.featureless")
  p = lrn$train(task)$predict(task)
  expect_prediction(p)
})

test_that("Positive class always comes first", {
  sonar = load_dataset("Sonar", package = "mlbench")
  tmp = list(c("M", "R"), c("R", "M"))
  lrn = lrn("classif.featureless", predict_type = "prob", method = "sample")

  for (lvls in tmp) {
    task = TaskClassif$new("sonar", backend = sonar, target = "Class", positive = lvls[[1]])
    expect_equal(task$positive, lvls[1])
    expect_equal(task$negative, lvls[2])
    expect_equal(task$class_names, lvls)
    expect_equal(levels(task$truth()), lvls)
    expect_equal(levels(task$data(cols = task$target_names)[[1L]]), lvls)
    preds = lrn$train(task)$predict(task)
    expect_equal(levels(preds$truth), lvls)
    expect_equal(levels(preds$response), lvls)
    expect_set_equal(colnames(preds$prob), lvls)
  }
})

test_that("class property is updated", {
  task = tsk("iris")

  task$filter(1:100)
  expect_subset("multiclass", task$properties)
  task$droplevels()
  expect_subset("twoclass", task$properties)
  task$filter(1:50)
  expect_subset("twoclass", task$properties)
  expect_error(task$droplevels())

  task$rbind(iris)
  task$droplevels()
  expect_subset("multiclass", task$properties)
})


test_that("droplevels keeps level order", {
  task = tsk("iris")
  my_order = c("virginica", "setosa", "versicolor")
  task$col_info[id == "Species", levels := list(my_order)]
  lvls = task$class_names
  expect_equal(lvls, my_order)
  task$filter(c(101:150, 1:50)) # remove versicolor
  task$droplevels()
  expect_equal(task$class_names, c("virginica", "setosa"))
})

test_that("target is encoded as factor (#629)", {
  dt = data.table(a = c(1, 2, 3, 4), target = c(1, 1, 0, 1))
  expect_error(TaskClassif$new(id = "XX", backend = dt, target = "target"), "factor")

  dt$target = ordered(dt$target)
  TaskClassif$new(id = "XX", backend = dt, target = "target")
})

test_that("offset column role works with binary tasks", {
  task = tsk("pima")
  expect_null(task$offset)

  task$set_col_roles("age", "offset")
  expect_subset("offset", task$properties)
  expect_data_table(task$offset, nrows = task$nrow, ncols = 2)
  expect_subset(c("row_id", "offset"), names(task$offset))

  expect_error(
    {
      task$col_roles$offset = c("glucose", "diabetes")
    },
    "There may only be up to one column with role"
  )

  expect_error(
    {
      task$col_roles$offset = c("glucose")
    },
    "contain missing values"
  )

  expect_warning(lrn("classif.rpart")$train(task), "has offset")
})

test_that("offset column role works with multiclass tasks", {
  task = tsk("penguins")
  expect_null(task$offset)
  task$set_col_roles("year", "offset")
  expect_subset("offset", task$properties)
  expect_data_table(task$offset, nrows = task$nrow, ncols = 2)
  expect_subset(c("row_id", "offset"), names(task$offset))

  expect_error(
    {
      task$col_roles$offset = "bill_length"
    },
    "contain missing values"
  )

  task = tsk("wine")

  expect_error(
    {
      task$col_roles$offset = c("alcohol", "ash")
    },
    "Must be a subset of"
  )

  task = tsk("wine")
  data = task$data()
  set(data, j = "offset_1", value = runif(nrow(data)))
  set(data, j = "offset_2", value = runif(nrow(data)))
  task = as_task_classif(data, target = "type")
  task$set_col_roles(c("offset_1", "offset_2"), "offset")

  expect_subset("offset", task$properties)
  expect_data_table(task$offset, nrows = task$nrow, ncols = 3)
  expect_subset(c("row_id", "offset_1", "offset_2"), names(task$offset))
})

#' @title PredictionData
#'
#' @description
#' Convert object to a [PredictionData] or a list of [PredictionData].
#'
#' @inheritParams as_task
#' @param task ([Task]).
#' @template param_row_ids
#' @param check (`logical(1)`)\cr
#'   Perform argument checks and type conversions?
#' @param ... (any)\cr
#'   Additional arguments.
#'
#' @return [PredictionData].
#' @export
as_prediction_data = function(x, task, row_ids = task$row_ids, check = TRUE, ...) {
  UseMethod("as_prediction_data")
}

#' @rdname as_prediction_data
#' @export
# nolint next
as_prediction_data.Prediction = function(x, task, row_ids = task$row_ids, check = TRUE, ...) {
  x$data
}

#' @rdname as_prediction_data
#' @export
# nolint next
as_prediction_data.PredictionData = function(x, task, row_ids = task$row_ids, check = TRUE, ...) {
  x
}

#' @rdname as_prediction_data
#' @param train_task ([Task])\cr
#'   Task used for training the learner.
#' @export
# nolint next
as_prediction_data.list = function(x, task, row_ids = task$row_ids, check = TRUE, ..., train_task) {
  assert_list(x, names = "unique")
  predict_types = names(mlr_reflections$learner_predict_types[[task$task_type]])
  if (!is.null(predict_types)) {
    predict_types = c(predict_types, "extra", "raw")
  }
  assert_names(names(x), subset.of = predict_types)

  x$row_ids = row_ids
  if (inherits(task, "TaskSupervised")) {
    x$truth = task$truth(row_ids)
  }

  if ("weights_measure" %chin% task$properties) {
    x$weights = task$weights_measure[list(row_id = row_ids), on = "row_id", "weight"][[1L]]
  }

  task = if (task$task_type == "unsupervised") train_task else task
  pdata = new_prediction_data(x, task$task_type)
  if (check) {
    pdata = check_prediction_data(pdata, train_task = task)
  }

  pdata
}

#' @title Friedman1 Regression Task Generator
#'
#' @name mlr_task_generators_friedman1
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] for the friedman1 task in [mlbench::mlbench.friedman1()].
#'
#' @templateVar id friedman1
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("friedman1")
#' task = generator$generate(200)
#' str(task$data())
TaskGeneratorFriedman1 = R6Class(
  "TaskGeneratorFriedman1",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        sd = p_dbl(0L, default = 1)
      )

      super$initialize(
        id = "friedman1",
        "regr",
        "mlbench",
        ps,
        label = "Friedman Regression",
        man = "mlr3::mlr_task_generators_friedman1"
      )
    }
  ),

  private = list(
    .generate = function(n) {
      obj = invoke(mlbench::mlbench.friedman1, n = n, .args = self$param_set$values)
      colnames(obj$x) = c(sprintf("important%i", 1:5), sprintf("unimportant%i", 1:5))
      data = insert_named(as.data.table(obj$x), list(y = obj$y))
      TaskRegr$new(sprintf("%s_%i", self$id, n), data, target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("friedman1", function() TaskGeneratorFriedman1$new())

#' @title Convert to ResampleResult
#'
#' @description
#' Convert object to a [ResampleResult].
#'
#' The S3 method for `list` expects argument `x` to be a list of [Prediction] objects and
#' all other relevant objects ([Task], [Learner]s, and instantiated [Resampling]) must
#' be provided, too.
#' A more flexible way to manually create a [ResampleResult] is implemented in [as_result_data()].
#'
#' @param x (any)\cr
#'  Object to convert.
#' @param ... (any)\cr
#'  Currently not used.
#'
#' @return ([ResampleResult]).
#' @export
as_resample_result = function(x, ...) {
  UseMethod("as_resample_result")
}


#' @rdname as_resample_result
#' @export
# nolint next
as_resample_result.ResampleResult = function(x, ...) {
  x
}

#' @rdname as_resample_result
#' @param view (`character()`)\cr
#'   See construction argument `view` of [`ResampleResult`].
#' @export
# nolint next
as_resample_result.ResultData = function(x, view = NULL, ...) {
  ResampleResult$new(x, view = view)
}

#' @rdname as_resample_result
#' @inheritParams as_result_data
#' @export
# nolint next
as_resample_result.list = function(x, task, learners, resampling, store_backends = TRUE, ...) {
  rdata = as_result_data(
    task = task,
    learners = learners,
    resampling = resampling,
    iterations = seq_len(resampling$iters),
    predictions = x,
    store_backends = store_backends
  )

  ResampleResult$new(rdata)
}

#' @title Assertion for mlr3 Objects
#'
#' @description
#' Functions intended to be used in packages extending \pkg{mlr3}.
#' Most assertion functions ensure the right class attribute, and optionally additional properties.
#' Additionally, the following compound assertions are implemented:
#'
#' * `assert_learnable(task, learner)`\cr
#'   ([Task], [Learner]) -> `NULL`\cr
#'   Checks if the learner is applicable to the task.
#'   This includes type checks on the type, the feature types, and properties.
#'
#' If an assertion fails, an exception is raised.
#' Otherwise, the input object is returned invisibly.
#'
#' @name mlr_assertions
#' @keywords internal
NULL


#' @export
#' @param b ([DataBackend]).
#' @rdname mlr_assertions
# nolint next
assert_backend = function(b, .var.name = vname(b)) {
  assert_class(b, "DataBackend", .var.name = .var.name)
}


#' @param task ([Task]).
#' @template param_feature_types
#' @param task_properties (`character()`)\cr
#'   Set of required task properties.
#' @rdname mlr_assertions
#' @export
# nolint next
assert_task = function(task, task_type = NULL, feature_types = NULL, task_properties = NULL, .var.name = vname(task)) {
  assert_class(task, "Task", .var.name = .var.name)

  if (!is.null(task_type) && task$task_type != task_type) {
    error_input("Task '%s' must have type '%s'", task$id, task_type)
  }

  if (!is.null(feature_types)) {
    tmp = setdiff(task$feature_types$type, feature_types)
    if (length(tmp)) {
      error_input("Task '%s' has the following unsupported feature types: %s", task$id, str_collapse(tmp))
    }
  }

  if (!is.null(task_properties)) {
    tmp = setdiff(task_properties, task$properties)
    if (length(tmp)) {
      error_input("Task '%s' is missing the following properties: %s", task$id, str_collapse(tmp))
    }
  }

  invisible(task)
}


#' @export
#' @param tasks (list of [Task]).
#' @rdname mlr_assertions
assert_tasks = function(
  tasks,
  task_type = NULL,
  feature_types = NULL,
  task_properties = NULL,
  # nolint next
  .var.name = vname(tasks)
) {
  assert_list(tasks, types = "Task")
  invisible(lapply(
    tasks,
    assert_task,
    task_type = task_type,
    feature_types = feature_types,
    task_properties = task_properties,
    .var.name = .var.name
  ))
}


#' @export
#' @param learner ([Learner]).
#' @param task_type (`character(1)`).
#' @rdname mlr_assertions
assert_learner = function(
  learner,
  task = NULL,
  task_type = NULL,
  properties = character(),
  # nolint next
  .var.name = vname(learner)
) {
  assert_class(learner, "Learner", .var.name = .var.name)

  task_type = task_type %??% task$task_type
  # check on class(learner) does not work with GraphLearner and AutoTuner
  # check on learner$task_type does not work with TaskUnsupervised
  if (!test_matching_task_type(task_type, learner, "learner")) {
    error_input("Learner '%s' must have task type '%s'", learner$id, task_type)
  }

  if (length(properties)) {
    miss = setdiff(properties, learner$properties)
    if (length(miss)) {
      error_input("Learner '%s' must have the properties: %s", learner$id, str_collapse(miss))
    }
  }

  invisible(learner)
}

test_matching_task_type = function(task_type, object, class) {
  if (is.null(task_type) || object$task_type == task_type) {
    return(TRUE)
  }

  cl_task_type = fget_key(mlr_reflections$task_types, task_type, class, "type")
  if (inherits(object, cl_task_type)) {
    return(TRUE)
  }

  cl_object = fget_key(mlr_reflections$task_types, object$task_type, class, "type")
  cl_task_type == cl_object
}


#' @export
#' @param learners (list of [Learner]).
#' @rdname mlr_assertions
assert_learners = function(
  learners,
  task = NULL,
  task_type = NULL,
  properties = character(),
  unique_ids = FALSE,
  # nolint next
  .var.name = vname(learners)
) {
  assert_list(learners, types = "Learner")
  if (unique_ids) {
    ids = map_chr(learners, "id")
    if (!test_character(ids, unique = TRUE)) {
      error_input("Learners need to have unique IDs: %s", str_collapse(ids))
    }
  }
  invisible(lapply(
    learners,
    assert_learner,
    task = task,
    task_type = NULL,
    properties = properties,
    .var.name = .var.name
  ))
}

# this does not check the validation task, as this is only possible once the validation set is known,
# which happens during worker(), so it cannot be checked before that
assert_task_learner = function(task, learner, param_values = NULL, cols = NULL) {
  pars = learner$param_set$get_values(type = "only_token", check_required = FALSE)
  # remove pars that are covered by param_values
  pars = pars[names(pars) %nin% names(param_values)]
  if (length(pars) > 0L) {
    error_config(
      "%s cannot be trained with TuneToken present in hyperparameter: %s",
      format_angle_brackets(learner),
      str_collapse(names(pars))
    )
  }
  # check on class(learner) does not work with GraphLearner and AutoTuner
  # check on learner$task_type does not work with TaskUnsupervised

  if (!test_matching_task_type(task$task_type, learner, "learner")) {
    error_input(
      "Type '%s' of %s does not match type '%s' of %s",
      task$task_type,
      format_angle_brackets(task),
      learner$task_type,
      format_angle_brackets(learner)
    )
  }

  tmp = setdiff(task$feature_types$type, learner$feature_types)
  if (length(tmp) > 0L) {
    error_input("%s has the following unsupported feature types: %s", format_angle_brackets(task), str_collapse(tmp))
  }

  if ("missings" %nin% learner$properties) {
    miss = task$missings(cols = cols) > 0L
    if (any(miss)) {
      error_input(
        "Task '%s' has missing values in column(s) %s, but learner '%s' does not support this",
        task$id,
        str_collapse(names(miss)[miss], quote = "'"),
        learner$id
      )
    }
  }

  if ("offset" %in% task$properties && "offset" %nin% learner$properties) {
    warning_input(
      "Task '%s' has offset, but learner '%s' does not support this, so it will be ignored",
      task$id,
      learner$id
    )
  }

  tmp = mlr_reflections$task_mandatory_properties[[task$task_type]]
  if (length(tmp)) {
    tmp = setdiff(intersect(task$properties, tmp), learner$properties)
    if (length(tmp)) {
      error_input("Task '%s' has property '%s', but learner '%s' does not support that", task$id, tmp[1L], learner$id)
    }
  }

  validate = get0("validate", learner)
  if (!is.null(task$internal_valid_task) && (is.numeric(validate) || identical(validate, "test"))) {
    error_config(
      # nolint next
      "Parameter 'validate' of Learner '%s' cannot be set to 'test' or a ratio when internal_valid_task is present, remove it first",
      learner$id
    )
  }
}

#' @export
#' @param param_values (`list()`)\cr
#'  TuneToken are not allowed in the parameter set of the learner.
#'  If the `param_values` overwrite the TuneToken, the assertion will pass.
#' @rdname mlr_assertions
assert_learnable = function(task, learner, param_values = NULL) {
  if (task$task_type == "unsupervised") {
    error_input("%s cannot be trained with %s", format_angle_brackets(learner), format_angle_brackets(task))
  }
  # we only need to check whether the learner wants to error on weights in training,
  # since weights_learner are always ignored during prediction.
  if (learner$use_weights == "error" && "weights_learner" %in% task$properties) {
    error_config(
      "%s cannot be trained with weights in %s%s",
      format_angle_brackets(learner),
      format_angle_brackets(task),
      if ("weights" %in% learner$properties) {
        " since 'use_weights' was set to 'error'."
      } else {
        # nolint next
        " since the Learner does not support weights.\nYou may set 'use_weights' to 'ignore' if you want the Learner to ignore weights."
      }
    )
  }
  assert_task_learner(task, learner, param_values)
}

#' @export
#' @rdname mlr_assertions
assert_predictable = function(task, learner) {
  if (!is.null(learner$state$train_task)) {
    train_task = learner$state$train_task
    cols_train = train_task$feature_names
    cols_predict = task$feature_names

    if (!test_permutation(cols_train, cols_predict)) {
      error_input("Learner '%s' has received tasks with different columns in train and predict.", learner$id)
    }

    ids = fget_keys(train_task$col_info, i = cols_train, j = "id", key = "id")
    train_type = fget_keys(train_task$col_info, i = ids, j = "type", key = "id")
    train_levels = fget_keys(train_task$col_info, i = ids, j = "levels", key = "id")
    predict_type = fget_keys(task$col_info, i = ids, j = "type", key = "id")
    predict_levels = fget_keys(task$col_info, i = ids, j = "levels", key = "id")

    ok = all(train_type == predict_type)

    if ("new_levels" %nin% learner$properties) {
      ok = ok && all(pmap_lgl(list(x = train_levels, y = predict_levels), identical))
    }

    if (!ok) {
      error_input(
        # nolint next
        "Learner '%s' received task with different column info (feature type or factor level ordering) during train and predict.",
        learner$id
      )
    }
  }

  assert_task_learner(task, learner, cols = task$feature_names)
}


#' @export
#' @param measure ([Measure]).
#' @param prediction ([Prediction]).
#' @rdname mlr_assertions
# nolint next
assert_measure = function(measure, task = NULL, learner = NULL, prediction = NULL, .var.name = vname(measure)) {
  assert_class(measure, "Measure", .var.name = .var.name)

  if (measure$use_weights == "error" && (!is.null(prediction$weights) || "weights_measure" %chin% task$properties)) {
    error_input(
      "%s cannot be evaluated with weights%s%s",
      format_angle_brackets(measure),
      if (!is.null(task)) paste0(" in ", format_angle_brackets(task)) else "",
      if ("weights" %in% measure$properties) {
        " since 'use_weights' was set to 'error'."
      } else {
        # nolint next
        " since the Measure does not support weights.\nYou may set 'use_weights' to 'ignore' if you want the Measure to ignore weights."
      }
    )
  }

  if (!is.null(task)) {
    if (!is_scalar_na(measure$task_type) && !test_matching_task_type(task$task_type, measure, "measure")) {
      error_input("Measure '%s' is not compatible with type '%s' of task '%s'", measure$id, task$task_type, task$id)
    }

    if (measure$check_prerequisites != "ignore") {
      miss = setdiff(measure$task_properties, task$properties)
      if (length(miss) > 0L) {
        warning_config(
          "Measure '%s' is missing properties %s of task '%s'",
          measure$id,
          str_collapse(miss, quote = "'"),
          task$id
        )
      }
    }
  }

  if (!is.null(learner)) {
    if (!is_scalar_na(measure$task_type) && !test_matching_task_type(measure$task_type, learner, "learner")) {
      error_input(
        "Measure '%s' is not compatible with type '%s' of learner '%s'",
        measure$id,
        learner$task_type,
        learner$id
      )
    }

    if (!is_scalar_na(measure$predict_type) && measure$check_prerequisites != "ignore") {
      predict_types = mlr_reflections$learner_predict_types[[learner$task_type]][[learner$predict_type]]
      if (measure$predict_type %nin% predict_types) {
        warning_config(
          "Measure '%s' is missing predict type '%s' of learner '%s'",
          measure$id,
          measure$predict_type,
          learner$id
        )
      }
    }

    if (measure$check_prerequisites != "ignore") {
      miss = setdiff(measure$predict_sets, learner$predict_sets)
      if (length(miss) > 0L) {
        warning_config(
          "Measure '%s' needs predict sets %s, but learner '%s' only predicted on sets %s",
          measure$id,
          str_collapse(miss, quote = "'"),
          learner$id,
          str_collapse(learner$predict_sets, quote = "'")
        )
      }
    }
  }

  if (!is.null(prediction) && is.null(learner)) {
    # same as above but works without learner e.g. measure$score(prediction)
    if (measure$check_prerequisites != "ignore" && measure$predict_type %nin% prediction$predict_types) {
      warning_config("Measure '%s' is missing predict type '%s' of prediction", measure$id, measure$predict_type)
    }
  }

  invisible(measure)
}

#' @export
#' @param measure ([Measure]).
#' @param prediction ([Prediction]).
#' @rdname mlr_assertions
# nolint next
assert_scorable = function(measure, task, learner, prediction = NULL, .var.name = vname(measure)) {
  if ("requires_model" %chin% measure$properties && is.null(learner$model)) {
    error_input("Measure '%s' requires the trained model", measure$id)
  }

  if ("requires_model" %chin% measure$properties && is_marshaled_model(learner$model)) {
    error_input("Measure '%s' requires the trained model, but model is in marshaled form", measure$id)
  }

  assert_measure(measure, task = task, learner = learner, prediction = prediction, .var.name = .var.name)
}

#' @export
#' @param measures (list of [Measure]).
#' @rdname mlr_assertions
# nolint next
assert_measures = function(measures, task = NULL, learner = NULL, .var.name = vname(measures)) {
  assert_list(measures, types = "Measure")
  lapply(measures, assert_measure, task = task, learner = learner, .var.name = .var.name)
  if (anyDuplicated(ids(measures))) {
    error_input("Measures need to have unique IDs")
  }
  invisible(measures)
}

#' @export
#' @param resampling ([Resampling]).
#' @rdname mlr_assertions
# nolint next
assert_resampling = function(resampling, instantiated = NULL, .var.name = vname(resampling)) {
  assert_class(resampling, "Resampling", .var.name = .var.name)

  if (!is.null(instantiated)) {
    if (instantiated && !resampling$is_instantiated) {
      error_input("Resampling '%s' must be instantiated", resampling$id)
    }
    if (!instantiated && resampling$is_instantiated) {
      error_input("Resampling '%s' may not be instantiated", resampling$id)
    }
  }

  invisible(resampling)
}


#' @export
#' @param resamplings (list of [Resampling]).
#' @rdname mlr_assertions
# nolint next
assert_resamplings = function(resamplings, instantiated = NULL, .var.name = vname(resamplings)) {
  assert_list(resamplings, types = "Resampling")
  invisible(lapply(resamplings, assert_resampling, instantiated = instantiated, .var.name = .var.name))
}


#' @export
#' @param prediction ([Prediction]).
#' @rdname mlr_assertions
# nolint next
assert_prediction = function(prediction, .var.name = vname(prediction), null.ok = FALSE) {
  if (null.ok && is.null(prediction)) {
    return(prediction)
  }
  assert_class(prediction, "Prediction", .var.name = .var.name)
}


#' @export
#' @param rr ([ResampleResult]).
#' @rdname mlr_assertions
# nolint next
assert_resample_result = function(rr, .var.name = vname(rr)) {
  assert_class(rr, "ResampleResult", .var.name = .var.name)
}


#' @export
#' @param bmr ([BenchmarkResult]).
#' @rdname mlr_assertions
# nolint next
assert_benchmark_result = function(bmr, .var.name = vname(bmr)) {
  assert_class(bmr, "BenchmarkResult", .var.name = .var.name)
}


# nolint next
assert_set = function(x, empty = TRUE, .var.name = vname(x)) {
  assert_character(
    x,
    min.len = as.integer(!empty),
    any.missing = FALSE,
    min.chars = 1L,
    unique = TRUE,
    .var.name = .var.name
  )
}


# nolint next
assert_range = function(range, .var.name = vname(range)) {
  assert_numeric(range, len = 2L, any.missing = FALSE, .var.name = .var.name)

  if (diff(range) < 0) {
    error_input(
      "Invalid range specified. First value (%f) must be less than or equal to second value (%f)",
      range[1L],
      range[2L]
    )
  }

  invisible(range)
}


#' @export
#' @template param_row_ids
#' @param task ([Task])\cr
#'   Task to check if row ids exist in it.
#' @rdname mlr_assertions
# nolint next
assert_row_ids = function(row_ids, task = NULL, null.ok = FALSE, .var.name = vname(row_ids)) {
  row_ids = assert_integerish(row_ids, coerce = TRUE, null.ok = null.ok)
  if (!is.null(task)) {
    if (any(row_ids %nin% task$row_ids)) {
      error_input("The provided row ids do not exist in task '%s'", task$id)
    }
  }
  invisible(row_ids)
}

#' @export
#' @param task ([Task]).
#' @rdname mlr_assertions
assert_has_backend = function(task) {
  if (is.null(task$backend)) {
    error_input(
      "The backend of Task '%s' has been removed. Set `store_backends` to `TRUE` during model fitting to conserve it.",
      task$id
    )
  }
}

assert_id = function(id, .var.name = "id") {
  assert_string(id, min.chars = 1L, .var.name = .var.name)
  if (grepl("%", id, fixed = TRUE)) {
    error_config("ID '%s' must not contain the special character '%%'", id)
  }
  invisible(id)
}

# assertion to ensure a helpful error message
assert_prediction_count = function(actual, expected, type) {
  if (actual != expected) {
    if (actual < expected) {
      error_learner_predict(
        "Predicted %s not complete, %s for %i observations is missing",
        type,
        type,
        expected - actual
      )
    } else {
      error_learner_predict(
        "Predicted %s contains %i additional predictions without matching rows",
        type,
        actual - expected
      )
    }
  }
}

assert_row_sums = function(prob) {
  for (i in seq_row(prob)) {
    x = prob[i, , drop = TRUE]
    n_missing = count_missing(x)
    if (n_missing > 0L) {
      if (n_missing < length(x)) {
        error_input("Probabilities for observation %i are partly missing", i)
      }
    } else {
      s = sum(x)
      if (abs(s - 1) > 0.001) {
        error_input("Probabilities for observation %i do sum up to %f != 1", i, s)
      }
    }
  }
}


#' @export
#' @param learner ([Learner]).
#' @param quantile_response (`logical(1)`)\cr
#'   Whether to check if the quantile response is set.
#'   If `TRUE`, the learner must have the `$quantile_response` field set.
#' @rdname mlr_assertions
assert_quantiles = function(learner, quantile_response = FALSE) {
  if (is.null(learner$quantiles)) {
    error_config("Quantiles must be set via `$quantiles`")
  }

  if (quantile_response && is.null(learner$quantile_response)) {
    error_config("Quantile response must be set via `$quantile_response`")
  }

  invisible(learner)
}

# nolint next
assert_param_values = function(x, n_learners = NULL, .var.name = vname(x)) {
  assert_list(x, len = n_learners, .var.name = .var.name)

  ok = every(x, function(x) {
    test_list(x) && every(x, test_list, names = "unique", null.ok = TRUE)
  })

  if (!ok) {
    error_input("'%s' must be a three-time nested list and the most inner list must be named", .var.name)
  }
  invisible(x)
}

#' @title Assert Empty Ellipsis
#' @description
#' Assert that `...` arguments are empty.
#' Use this function in S3-methods to ensure that misspelling of arguments does not go unnoticed.
#' @param ... (any)\cr
#'    Ellipsis arguments to check.
#' @keywords internal
#' @return `NULL`
#' @export
assert_empty_ellipsis = function(...) {
  nx = ...length()
  if (nx == 0L) {
    return(NULL)
  }
  names = ...names()
  if (is.null(names)) {
    error_input("Received %i unnamed argument that was not used.", nx)
  }
  names2 = names[nzchar(names)]
  if (length(names2) == length(names)) {
    error_input(
      "Received the following named arguments that were unused: %s.",
      toString(names2)
    )
  }
  error_input(
    "Received unused arguments: %i unnamed, as well as named arguments %s.",
    length(names) - length(names2),
    toString(names2)
  )
}

test_that("assert_empty_ellipsis works", {
  expect_error(assert_empty_ellipsis(1), "Received 1 unnamed argument")
  expect_error(assert_empty_ellipsis(1, 2), "Received 2 unnamed argument")
  expect_error(assert_empty_ellipsis(a = 1), "that were unused: a")
  expect_error(assert_empty_ellipsis(a = 1, b = 2), "that were unused: a, b")
  expect_error(assert_empty_ellipsis(a = 1, b = 1, 2), "1 unnamed, as well as named arguments a, b")
  expect_null(assert_empty_ellipsis())
})

#' @title Predict Method for Learners
#'
#' @description
#' Extends the generic [stats::predict()] with a method for [Learner].
#' Note that this function is intended as glue code to be used in third party packages.
#' We recommend to work with the [Learner] directly, i.e. calling `learner$predict()`
#' or `learner$predict_newdata()` directly.
#'
#' Performs the following steps:
#' * Sets additional hyperparameters passed to this function.
#' * Creates a [Prediction] object by calling `learner$predict_newdata()`.
#' * Returns (subset of) [Prediction].
#'
#' @param object ([Learner])\cr
#'   Any [Learner].
#' @param newdata ([data.frame()])\cr
#'   New data to predict on.
#' @param predict_type (`character(1)`)\cr
#'   The predict type to return.
#'   Set to `<Prediction>` to retrieve the complete [Prediction] object.
#'   If set to `NULL` (default), the first predict type for the respective class of the [Learner]
#'   as stored in [mlr_reflections] is used.
#' @param ... (any)\cr
#'   Hyperparameters to pass down to the [Learner].
#'
#' @export
#' @examples
#' task = tsk("spam")
#'
#' learner = lrn("classif.rpart", predict_type = "prob")
#' learner$train(task)
#' predict(learner, task$data(1:3), predict_type = "response")
#' predict(learner, task$data(1:3), predict_type = "prob")
#' predict(learner, task$data(1:3), predict_type = "<Prediction>")
predict.Learner = function(object, newdata, predict_type = NULL, ...) {
  assert_data_frame(newdata)
  assert_string(predict_type, null.ok = TRUE)

  if (...length()) {
    pars = assert_list(list(...), names = "unique")
    tags = object$param_set$tags
    predict_pars = names(tags)[map_lgl(tags, is.element, el = "predict")]
    i = which(names(pars) %nin% predict_pars)
    if (length(i)) {
      error_input("Unknown parameters: %s", str_collapse(names(pars)[i]))
    }

    object = object$clone()
    object$param_set$values = insert_named(object$param_set$values, list(...))
  }

  prediction = object$predict_newdata(newdata)

  if (identical(predict_type, "<Prediction>")) {
    return(prediction)
  }

  predict_type = predict_type %??% head(names(mlr_reflections$learner_predict_types[[object$task_type]]), 1L)
  if (predict_type %nin% prediction$predict_types) {
    error_input("Predict type '%s' not available", predict_type)
  }

  prediction[[predict_type]]
}

expect_grouping_works = function(r) {
  data = insert_named(as.data.table(iris), list(grp = rep_len(letters[1:10], 150)))
  task = TaskClassif$new("iris-grp", as_data_backend(data), target = "Species")
  task$col_roles$group = "grp"

  r$instantiate(task)
  for (i in seq_len(r$iters)) {
    expect_integer(r$train_set(i), lower = 1L, upper = 150L, any.missing = FALSE)
    expect_integer(r$test_set(i), lower = 1L, upper = 150L, any.missing = FALSE)
    if (!inherits(r, "ResamplingInsample")) {
      testthat::expect_length(intersect(data[r$train_set(i), get("grp")], data[r$test_set(i), get("grp")]), 0L)
    }
  }
}

#' @title Calculate ROC Measures
#'
#' @description
#' Calculate a set of roc performance measures based on the confusion matrix.
#'
#' * `tpr` True positive rate (Sensitivity, Recall)
#' * `fpr` False positive rate (Fall-out)
#' * `fnr` False negative rate (Miss rate)
#' * `tnr` True negative rate (Specificity)
#' * `ppv` Positive predictive value (Precision)
#' * `fomr` False omission rate
#' * `lrp` Positive likelihood ratio (LR+)
#' * `fdr` False discovery rate
#' * `npv` Negative predictive value
#' * `acc` Accuracy
#' * `lrm` Negative likelihood ratio (LR-)
#' * `dor` Diagnostic odds ratio
#'
#' @param pred ([PredictionClassif])\cr
#' The prediction object.
#'
#' @return `list()`\cr
#'  A list containing two elements `confusion_matrix` which is the 2 times 2 confusion matrix
#'  of absolute frequencies and `measures`, a list of the above mentioned measures.
#' @export
#' @examples
#' learner = lrn("classif.rpart", predict_type = "prob")
#' splits = partition(task = tsk("pima"), ratio = 0.7)
#' task = tsk("pima")
#' learner$train(task)
#' pred = learner$predict(task)
#' score_roc_measures(pred)
score_roc_measures = function(pred) {
  assert_prediction(pred)

  score_tpr = pred$score(msr("classif.tpr"))
  score_fnr = pred$score(msr("classif.fnr"))
  score_fpr = pred$score(msr("classif.fpr"))
  score_tnr = pred$score(msr("classif.tnr"))
  score_ppv = pred$score(msr("classif.ppv"))
  score_fdr = 1 - score_ppv
  score_npv = pred$score(msr("classif.npv"))
  score_fomr = 1 - score_npv
  score_acc = pred$score(msr("classif.acc"))
  score_fpr = pred$score(msr("classif.fpr"))
  score_lr_plus = score_tpr / score_fpr
  score_lr_minus = score_fnr / score_tnr
  score_dor = score_lr_plus / score_lr_minus

  set_class(
    list(
      confusion_matrix = pred$confusion,
      measures = list(
        tpr = score_tpr,
        fpr = score_fpr,
        fnr = score_fnr,
        tnr = score_tnr,
        ppv = score_ppv,
        fdr = score_fdr,
        npv = score_npv,
        fomr = score_fomr,
        acc = score_acc,
        lr_plus = score_lr_plus,
        lr_minus = score_lr_minus,
        dor = score_dor
      )
    ),
    "roc_measures"
  )
}

#' @title Print ROC Measures
#'
#' @description
#' Print the confusion matrix and a set of roc performance measures.
#'
#' @param x (`roc_measures`)\cr
#'   The object returned by `score_roc_measures`.
#' @param abbreviations (`logical(1)`)\cr
#'   If `TRUE`, print a list of abbreviations for the measures.
#' @param digits (`integer(1)`)\cr
#'   Number of digits to round the measures to.
#' @param ... (`any`)\cr
#'   Additional parameters, currently unused.
#' @export
print.roc_measures = function(x, abbreviations = TRUE, digits = 2L, ...) {
  assert_flag(abbreviations)
  assert_int(digits, lower = 1L)

  # format measures
  x$measures = mapply(function(m, v) paste0(m, ": ", round(v, digits)), names(x$measures), x$measures)

  res = cbind(
    round(x$confusion_matrix, digits = digits),
    c(x$measures[["tpr"]], x$measures[["fpr"]]),
    c(x$measures[["fnr"]], x$measures[["tnr"]])
  )
  res = rbind(
    res,
    c(x$measures[["ppv"]], x$measures[["fomr"]], x$measures[["lr_plus"]], x$measures[["acc"]]),
    c(x$measures[["fdr"]], x$measures[["npv"]], x$measures[["lr_minus"]], x$measures[["dor"]])
  )

  names(dimnames(res)) = c("true", "predicted")
  cat_cli({
    cli_h2("ROC Measures")
  })
  print(noquote(res))
  if (abbreviations) {
    cat_cli({
      cli_h3("Abbreviations:")
      cli_li("tpr - True positive rate (Sensitivity, Recall)")
      cli_li("fpr - False positive rate (Fall-out)")
      cli_li("fnr - False negative rate (Miss rate)")
      cli_li("tnr - True negative rate (Specificity)")
      cli_li("ppv - Positive predictive value (Precision)")
      cli_li("fomr - False omission rate")
      cli_li("lrp - Positive likelihood ratio (LR+)")
      cli_li("fdr - False discovery rate")
      cli_li("npv - Negative predictive value")
      cli_li("acc - Accuracy")
      cli_li("lrm - Negative likelihood ratio (LR-)")
      cli_li("dor - Diagnostic odds ratio")
    })
  }
}

#' @title German Credit Classification Task
#'
#' @name mlr_tasks_german_credit
#' @format [R6::R6Class] inheriting from [TaskClassif].
#' @include mlr_tasks.R
#'
#' @description
#' A classification task for the German credit data set.
#' The aim is to predict creditworthiness, labeled as "good" and "bad".
#' Positive class is set to label "good".
#'
#' See the example for the creation of a [MeasureClassifCosts] with the described misclassification costs.
#'
#' @templateVar id german_credit
#' @template task
#'
#' @source
#' Data set originally published on [UCI](http://archive.ics.uci.edu/dataset/144/statlog+german+credit+data).
#' This is the preprocessed version taken from package \CRANpkg{rchallenge} with
#' factors instead of dummy variables, and corrected as proposed by Ulrike
#' Grömping.
#'
#' Donor:
#' Professor Dr. Hans Hofmann\cr
#' Institut für Statistik und Ökonometrie\cr
#' Universität Hamburg\cr
#' FB Wirtschaftswissenschaften\cr
#' Von-Melle-Park 5\cr
#' 2000 Hamburg 13
#'
#' @references
#' `r format_bib("gromping_2019")`
#'
#' @template seealso_task
#' @examples
#' task = tsk("german_credit")
#' costs = matrix(c(0, 1, 5, 0), nrow = 2)
#' dimnames(costs) = list(predicted = task$class_names, truth = task$class_names)
#' measure = msr("classif.costs", id = "german_credit_costs", costs = costs)
#' print(measure)
NULL

load_task_german_credit = function(id = "german_credit") {
  b = as_data_backend(readRDS(system.file("extdata", "german_credit.rds", package = "mlr3")))
  task = TaskClassif$new(id, b, target = "credit_risk", positive = "good", label = "German Credit")
  b$hash = task$man = "mlr3::mlr_tasks_german_credit"
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("german_credit", load_task_german_credit)

#' @importFrom utils bibentry
# fmt: skip
bibentries = c(
  anderson_1936 = bibentry("article",
    doi       = "10.2307/2394164",
    year      = "1936",
    month     = "9",
    publisher = "JSTOR",
    volume    = "23",
    number    = "3",
    pages     = "457",
    author    = "Edgar Anderson",
    title     = "The Species Problem in Iris",
    journal   = "Annals of the Missouri Botanical Garden"
  ),

  bischl_2012 = bibentry("article",
    title = "Resampling Methods for Meta-Model Validation with Recommendations for Evolutionary Computation",
    author = "Bernd Bischl and Olaf Mersmann and Heike Trautmann and Claus Weihs",
    year = "2012",
    month = "jun",
    journal = "Evolutionary Computation",
    publisher = "MIT Press - Journals",
    volume = "20",
    number = "2",
    pages = "249--275",
    doi = "10.1162/evco_a_00069"
  ),

  breiman_1984 = bibentry("book",
    title        = "Classification And Regression Trees",
    author       = "Leo Breiman and Jerome H. Friedman and Richard A. Olshen and Charles J. Stone",
    year         = "1984",
    publisher    = "Routledge",
    doi          = "10.1201/9781315139470"
  ),

  dua_2017 = bibentry("misc",
    title        = "{UCI} Machine Learning Repository",
    author       = "Dua, Dheeru and Graff, Casey",
    year         = "2017",
    url          = "http://archive.ics.uci.edu/datasets",
    institution  = "University of California, Irvine, School of Information and Computer Sciences"
  ),

  gorman2014 = bibentry("article",
    doi       = "10.1371/journal.pone.0090081",
    year      = "2014",
    month     = "mar",
    publisher = "Public Library of Science ({PLoS})",
    volume    = "9",
    number    = "3",
    pages     = "e90081",
    author    = "Kristen B. Gorman and Tony D. Williams and William R. Fraser",
    editor    = "Andr\u00e8 Chiaradia",
    # nolint next
    title     = "Ecological Sexual Dimorphism and Environmental Variability within a Community of Antarctic Penguins (Genus Pygoscelis)",
    journal   = "{PLoS} {ONE}"
  ),

  gromping_2019 = bibentry("techreport",
    title        = "South German Credit Data: Correcting a Widely Used Data Set",
    author       = "Ulrike Gr\u00f6mping",
    year         = "2019",
    number       = "4",
    type         = "Reports in Mathematics, Physics and Chemistry",
    institution  = "Department II, Beuth University of Applied Sciences Berlin"
  ),

  sasaki_2007 = bibentry("article",
    title        = "The truth of the F-measure",
    author       = "Sasaki, Yutaka",
    year         = "2007",
    journal      = "Teach Tutor mater",
    volume       = "1",
    number       = "5",
    pages        = "1--5",
    url          = "https://www.cs.odu.edu/~mukka/cs795sum10dm/Lecturenotes/Day3/F-measure-YS-26Oct07.pdf"
  ),

  koenker_1999 = bibentry("article",
    title = "Goodness of Fit and Related Inference Processes for Quantile Regression",
    author = "Koenker, Roger and Machado, Jos\u00e9 A. F.",
    year = "1999",
    doi = "10.1080/01621459.1999.10473882",
    journal = "Journal of the American Statistical Association",
    volume = "94",
    number = "448",
    pages = "1296--1310"
  )

)

#' @title Dictionary of Task Generators
#'
#' @usage NULL
#' @format [R6::R6Class] object inheriting from [mlr3misc::Dictionary].
#'
#' @description
#' A simple [mlr3misc::Dictionary] storing objects of class [TaskGenerator].
#' Each task generator has an associated help page, see `mlr_task_generators_[id]`.
#'
#' This dictionary can get populated with additional task generators by add-on packages.
#'
#' For a more convenient way to retrieve and construct task generators, see [tgen()]/[tgens()].
#'
#' @section Methods:
#' See [mlr3misc::Dictionary].
#'
#' @section S3 methods:
#' * `as.data.table(dict, ..., objects = FALSE)`\cr
#'   [mlr3misc::Dictionary] -> [data.table::data.table()]\cr
#'   Returns a [data.table::data.table()] with fields "key", "label", "task_type", "params", and "packages" as columns.
#'   If `objects` is set to `TRUE`, the constructed objects are returned in the list column named `object`.
#'
#' @family Dictionary
#' @family TaskGenerator
#' @seealso
#' Sugar functions: [tgen()], [tgens()]
#' @export
#' @examples
#' mlr_task_generators$get("smiley")
#' tgen("2dnormals")
mlr_task_generators = R6Class("DictionaryTaskGenerator", inherit = Dictionary, cloneable = FALSE)$new()

#' @export
as.data.table.DictionaryTaskGenerator = function(x, ..., objects = FALSE) {
  assert_flag(objects)

  setkeyv(
    map_dtr(
      x$keys(),
      function(key) {
        g = withCallingHandlers(x$get(key), packageNotFoundWarning = function(w) invokeRestart("muffleWarning"))
        insert_named(
          list(
            key = key,
            label = g$label,
            task_type = g$task_type,
            params = list(g$param_set$ids()),
            packages = list(g$packages)
          ),
          if (objects) list(object = list(g))
        )
      },
      .fill = TRUE
    ),
    "key"
  )[]
}

#' @title Custom Cross-Validation
#'
#' @name mlr_resamplings_custom_cv
#' @include Resampling.R
#'
#' @description
#' Splits data into training and test sets in a cross-validation fashion based
#' on a user-provided categorical vector.
#' This vector can be passed during instantiation either via an arbitrary factor `f`
#' with the same length as `task$nrow`, or via a single string `col` referring to a
#' column in the task.
#'
#' An alternative but equivalent approach using leave-one-out resampling is
#' showcased in the examples of [mlr_resamplings_loo].
#'
#' @templateVar id custom_cv
#' @template resampling
#'
#' @template seealso_resampling
#' @export
#' @examples
#' # Create a task with 10 observations
#' task = tsk("penguins")
#' task$filter(1:10)
#'
#' # Instantiate Resampling:
#' custom_cv = rsmp("custom_cv")
#' f = factor(c(rep(letters[1:3], each = 3), NA))
#' custom_cv$instantiate(task, f = f)
#' custom_cv$iters # 3 folds
#'
#' # Individual sets:
#' custom_cv$train_set(1)
#' custom_cv$test_set(1)
#'
#' # Disjunct sets:
#' intersect(custom_cv$train_set(1), custom_cv$test_set(1))
ResamplingCustomCV = R6Class(
  "ResamplingCustomCV",
  inherit = Resampling,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      super$initialize(
        id = "custom_cv",
        duplicated_ids = FALSE,
        label = "Custom Split Cross-Validation",
        man = "mlr3::mlr_resamplings_custom_cv"
      )
    },

    #' @description
    #' Instantiate this [Resampling] as cross-validation with custom splits.
    #'
    #' @param task [Task]\cr
    #'   Used to extract row ids.
    #'
    #' @param f (`factor()` | `character()`)\cr
    #'   Vector of type factor or character with the same length as `task$nrow`.
    #'   Row ids are split on this vector, each distinct value results in a fold.
    #'   Empty factor levels are dropped and row ids corresponding to missing values are removed,
    #'   c.f. [split()].
    #' @param col (`character(1)`)\cr
    #'   Name of the task column to use for splitting.
    #'   Alternative and mutually exclusive to providing the factor levels as a vector via
    #'   parameter `f`.
    instantiate = function(task, f = NULL, col = NULL) {
      task = assert_task(as_task(task))
      private$.hash = NULL
      if (!xor(is.null(f), is.null(col))) {
        error_input("Exactly one of `f` or `col` must be provided")
      }

      if (!is.null(col)) {
        assert_choice(col, task$col_info$id)
        f = task$data(cols = col)[[1L]]
        assert_atomic_vector(f, all.missing = FALSE)
      } else {
        assert_factor(f, len = task$nrow, all.missing = FALSE)
      }

      self$instance = split(task$row_ids, f, drop = TRUE)
      self$task_hash = task$hash
      self$task_nrow = task$nrow
      self$task_row_hash = task$row_hash
      invisible(self)
    }
  ),

  active = list(
    #' @template field_iters
    iters = function(rhs) {
      assert_ro_binding(rhs)
      if (self$is_instantiated) length(self$instance) else NA_integer_
    }
  ),

  private = list(
    .get_train = function(i) {
      unlist(self$instance[-i], use.names = FALSE) %??% integer()
    },

    .get_test = function(i) {
      self$instance[[i]]
    }
  )
)

#' @include mlr_resamplings.R
mlr_resamplings$add("custom_cv", function() ResamplingCustomCV$new())

#' @title Peak Regression Task Generator
#'
#' @name mlr_task_generators_peak
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] for the peak task in [mlbench::mlbench.peak()].
#'
#' @templateVar id peak
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("peak", d = 5)
#' task = generator$generate(200)
#' str(task$data())
TaskGeneratorPeak = R6Class(
  "TaskGeneratorPeak",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        d = p_int(lower = 1, default = 20L)
      )

      super$initialize(
        id = "peak",
        "regr",
        "mlbench",
        ps,
        label = "Peak Regression",
        man = "mlr3::mlr_task_generators_peak"
      )
    }
  ),

  private = list(
    .generate_obj = function(n) {
      invoke(
        mlbench::mlbench.peak,
        n = n,
        .args = self$param_set$values,
        .opts = allow_partial_matching
      )
    },

    .generate = function(n) {
      obj = invoke(
        mlbench::mlbench.peak,
        n = n,
        .args = self$param_set$values
      )
      colnames(obj$x) = c(sprintf("x%0i", seq_col(obj$x)))
      data = insert_named(as.data.table(obj$x), list(y = obj$y))
      TaskRegr$new(sprintf("%s_%i", self$id, n), data, target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("peak", function() TaskGeneratorPeak$new())

#' @title DataBackend
#'
#' @include mlr_reflections.R
#' @include warn_deprecated.R
#'
#' @description
#' This is the abstract base class for data backends.
#'
#' Data backends provide a layer of abstraction for various data storage systems.
#' It is not recommended to work directly with the DataBackend.
#' Instead, all data access is handled transparently via the [Task].
#'
#' This package currently ships with one implementation for backends:
#'
#' * [DataBackendDataTable] which stores the data as [data.table::data.table()].
#'
#' To connect to out-of-memory database management systems such as SQL servers,
#' see the extension package \CRANpkg{mlr3db}.
#'
#' @details
#' The required set of fields and methods to implement a custom `DataBackend` is
#' listed in the respective sections (see [DataBackendDataTable]).
#'
#' @template seealso_databackend
#'
#' @export
#' @examples
#' data = data.table::data.table(id = 1:5, x = runif(5),
#'   y = sample(letters[1:3], 5, replace = TRUE))
#'
#' b = DataBackendDataTable$new(data, primary_key = "id")
#' print(b)
#' b$head(2)
#' b$data(rows = 1:2, cols = "x")
#' b$distinct(rows = b$rownames, "y")
#' b$missings(rows = b$rownames, cols = names(data))
DataBackend = R6Class(
  "DataBackend",
  cloneable = FALSE,
  public = list(
    #' @field primary_key (`character(1)`)\cr
    #' Column name of the primary key column of positive and unique integer row ids.
    primary_key = NULL,

    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' Note: This object is typically constructed via a derived classes, e.g.
    #' [DataBackendDataTable], or via the S3 method
    #' [as_data_backend()].
    #'
    #' @param data (any)\cr
    #'   The format of the input data depends on the specialization. E.g.,
    #'   [DataBackendDataTable] expects a [data.table::data.table()].
    #'
    #' @param primary_key (`character(1)`)\cr
    #'   Each DataBackend needs a way to address rows, which is done via a
    #'   column of unique integer values, referenced here by `primary_key`. The
    #'   use of this variable may differ between backends.
    initialize = function(data, primary_key) {
      private$.data = data
      self$primary_key = assert_string(primary_key)
    },

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s>", class(self)[1L])
    },

    #' @description
    #' Printer.
    print = function() {
      nr = self$nrow
      cat_cli(cli_h1("{.cls {class(self)[1L]}} ({.val {nr}}x{.val {self$ncol}})"))
      print(self$head(6L), row.names = FALSE, print.keys = FALSE)
      if (nr > 6L) {
        catf("[...] (%i rows omitted)", nr - 6L)
      }
    }
  ),

  active = list(
    #' @field hash (`character(1)`)\cr
    #' Hash (unique identifier) for this object.
    hash = function(rhs) {
      if (missing(rhs)) {
        if (is.na(private$.hash)) {
          private$.hash = private$.calculate_hash()
        }
        return(private$.hash)
      }
      private$.hash = assert_string(rhs)
    },

    #' @template field_col_hashes
    col_hashes = function() {
      cn = setdiff(self$colnames, self$primary_key)
      set_names(sprintf("%s.%s", self$hash, cn), cn)
    }
  ),

  private = list(
    .data = NULL,
    .hash = NA_character_
  )
)

#' @title Zoo Classification Task
#'
#' @name mlr_tasks_zoo
#' @format [R6::R6Class] inheriting from [TaskClassif].
#' @include mlr_tasks.R
#'
#' @description
#' A classification task for the [mlbench::Zoo] data set.
#' Rownames are stored as variable `"..rownames"` with column role `"name"`.
#'
#' @templateVar id zoo
#' @template task
#'
#' @template seealso_task
NULL

load_task_zoo = function(id = "zoo") {
  b = as_data_backend(load_dataset("Zoo", "mlbench", keep_rownames = TRUE), keep_rownames = "animal")
  task = TaskClassif$new(id, b, target = "type", label = "Zoo Animals")
  b$hash = task$man = "mlr3::mlr_tasks_zoo"
  task$col_roles$name = "animal"
  task$col_roles$feature = setdiff(task$col_roles$feature, "animal")
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("zoo", load_task_zoo)

#' @title Similarity Measure
#'
#' @include Measure.R
#'
#' @description
#' This measure specializes [Measure] for measures quantifying the similarity of
#' sets of selected features.
#' To calculate similarity measures, the [Learner] must have the property
#' `"selected_features"`.
#'
#' * `task_type` is set to `NA_character_`.
#' * `average` is set to `"custom"`.
#'
#' Predefined measures can be found in the [dictionary][mlr3misc::Dictionary]
#' [mlr_measures], prefixed with `"sim."`.
#'
#' @template param_id
#' @template param_param_set
#' @template param_range
#' @template param_minimize
#' @template param_aggregator
#' @template param_predict_type
#' @template param_measure_properties
#' @template param_predict_sets
#' @template param_task_properties
#' @template param_packages
#' @template param_label
#' @template param_man
#'
#' @template seealso_measure
#' @export
#' @examples
#' task = tsk("penguins")
#' learners = list(
#'   lrn("classif.rpart", maxdepth = 1, id = "r1"),
#'   lrn("classif.rpart", maxdepth = 2, id = "r2")
#' )
#' resampling = rsmp("cv", folds = 3)
#' grid = benchmark_grid(task, learners, resampling)
#' bmr = benchmark(grid, store_models = TRUE)
#' bmr$aggregate(msrs(c("classif.ce", "sim.jaccard")))
MeasureSimilarity = R6Class(
  "MeasureSimilarity",
  inherit = Measure,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function(
      id,
      param_set = ps(),
      range,
      minimize = NA,
      aggregator = NULL,
      properties = character(),
      predict_type = NA_character_,
      task_properties = character(),
      packages = character(),
      label = NA_character_,
      man = NA_character_
    ) {
      super$initialize(
        id,
        task_type = NA_character_,
        param_set = param_set,
        range = range,
        minimize = minimize,
        average = "custom",
        aggregator = aggregator,
        properties = c("requires_model", "requires_no_prediction", properties),
        predict_type = predict_type,
        predict_sets = NULL,
        task_properties = task_properties,
        packages = packages,
        label = label,
        man = man
      )
    }
  ),

  private = list(
    .score = function(prediction, ...) {
      NA_real_
    }
  )
)

#' @title Convert to BenchmarkResult
#'
#' @description
#' Convert object to a [BenchmarkResult].
#'
#' @inheritParams as_task
#'
#' @return ([BenchmarkResult]).
#' @export
as_benchmark_result = function(x, ...) {
  UseMethod("as_benchmark_result")
}


#' @rdname as_benchmark_result
#' @export
# nolint next
as_benchmark_result.BenchmarkResult = function(x, ...) {
  x
}

#' @rdname as_benchmark_result
#' @export
# nolint next
as_benchmark_result.ResampleResult = function(x, ...) {
  private = get_private(x)
  rdata = private$.data$clone(deep = TRUE)
  if (!is.null(private$.view)) {
    rdata$data$fact = rdata$data$fact[list(private$.view), on = "uhash", nomatch = NULL]
    rdata$sweep()
  }
  BenchmarkResult$new(rdata)
}

# extract values from a single column `j` of a data table where the `key` column matches `i`
# tries to avoid the overhead of data.table for small tables
# i must be a single value, use fget_keys for multiple values
# returns all matches
# if `i` has no match in `key`, `NULL` is returned
fget_key = function(tab, i, j, key) {
  if (nrow(tab) > 1000L) {
    ijoin(tab, i, j, key, mult = "all", nomatch = NULL)[[1L]]
  } else {
    x = tab[[key]]
    if (is.character(x) && is.character(i)) {
      tab[[j]][x %chin% i]
    } else {
      tab[[j]][x %in% i]
    }
  }
}

# extract values from a single column `j` of a data table where the `key` column matches `i`
# tries to avoid the overhead of data.table for small tables
# i can be a vector of values
# returns a vector sorted by the order of i
# if the key column is not unique, the first match is returned for each i
# if no-matching elements are found, NA is returned
fget_keys = function(tab, i, j, key) {
  if (nrow(tab) > 1000L) {
    ijoin(tab, i, j, key, mult = "first", nomatch = NA)[[1L]]
  } else {
    x = tab[[key]]
    tab[[j]][match(i, x)]
  }
}

# nolint next
ijoin = function(tab, .__i__, .__j__, .__key__, nomatch = NULL, mult = "all") {
  if (!is.list(.__i__)) {
    # nolint next
    .__i__ = list(.__i__)
  }
  tab[.__i__, .__j__, with = FALSE, nomatch = nomatch, on = .__key__, mult = mult]
}

# updating join:
# replaces values in x with values in y
ujoin = function(x, y, key) {
  cn = setdiff(intersect(names(x), names(y)), key)
  expr = parse(text = paste0("`:=`(", paste0(sprintf("%1$s=i.%1$s", cn), collapse = ","), ")"))
  x[y, eval(expr), on = key][]
}

# fget_key ---------------------------------------------------------------------

test_that("fget_key works with character keys", {
  # small table
  tab = data.table(key_col = c("k0001", "k0002", "k0003"), val = c(1L, 2L, 3L))
  res = fget_key(tab, i = "k0002", j = "val", key = "key_col")
  expect_equal(res, 2L)

  # large table
  n = 1200L
  tab = data.table(key_col = sprintf("k%04d", 1:n), val = 1:n)
  res = fget_key(tab, i = "k1000", j = "val", key = "key_col")
  expect_equal(res, 1000L)
})

test_that("fget_key returns all matches on non-unique key column", {
  # small table
  tab = data.table(key_col = c("k0001", "k0002", "k0002"), val = c(1L, 2L, 3L))
  res = fget_key(tab, i = "k0002", j = "val", key = "key_col")
  expect_equal(res, c(2L, 3L))

  # large table
  n = 1200L
  tab = data.table(key_col = sprintf("k%04d", 1:n), val = 1:n)
  tab[1001, key_col := "k1000"]
  res = fget_key(tab, i = "k1000", j = "val", key = "key_col")
  expect_equal(res, c(1000L, 1001L))
})

test_that("fget_key works with non-character keys", {
  # small table
  tab = data.table(key_col = c(1L, 2L, 3L), val = c(1L, 2L, 3L))
  res = fget_key(tab, i = 2L, j = "val", key = "key_col")
  expect_equal(res, 2L)

  # large table
  n = 1200L
  tab = data.table(key_col = 1:n, val = 1:n)
  res = fget_key(tab, i = 1000L, j = "val", key = "key_col")
  expect_equal(res, 1000L)
})

test_that("fget returns empty vector if no match", {
  # small table
  tab = data.table(key_col = c("k0001", "k0002", "k0003"), val = c(1L, 2L, 3L))
  res = fget_key(tab, i = "k0004", j = "val", key = "key_col")
  expect_equal(res, integer())

  # large table
  n = 1200L
  tab = data.table(key_col = 1:n, val = 1:n)
  res = fget_key(tab, i = 1201L, j = "val", key = "key_col")
  expect_equal(res, integer())
})

# fget_keys ---------------------------------------------------------------------

test_that("fget_keys returns values in order of i", {
  # small table
  tab = data.table(key_col = c("k0001", "k0002", "k0003"), val = c(1L, 2L, 3L))
  res = fget_keys(tab, i = c("k0002", "k0001"), j = "val", key = "key_col")
  expect_equal(res, c(2L, 1L))

  # large table
  n = 1200L
  tab = data.table(key_col = sprintf("k%04d", 1:n), val = 1:n)
  res = fget_keys(tab, i = c("k1000", "k0201"), j = "val", key = "key_col")
  expect_equal(res, c(1000L, 201L))
})

test_that("non-matching elements return NA", {
  # small table
  tab = data.table(key_col = c("k0001", "k0002", "k0003"), val = c(1L, 2L, 3L))
  res = fget_keys(tab, i = c("k0002", "k0004"), j = "val", key = "key_col")
  expect_equal(res, c(2L, NA))

  # large table
  n = 1200L
  tab = data.table(key_col = sprintf("k%04d", 1:n), val = 1:n)
  res = fget_keys(tab, i = c("k1000", "k1201"), j = "val", key = "key_col")
  expect_equal(res, c(1000L, NA))
})

test_that("fget_keys repeats matches", {
  # small table
  tab = data.table(key_col = c("k0001", "k0002", "k0003"), val = c(1L, 2L, 3L))
  res = fget_keys(tab, i = c("k0002", "k0002", "k0003"), j = "val", key = "key_col")
  expect_equal(res, c(2L, 2L, 3L))

  # large table
  n = 1200L
  tab = data.table(key_col = sprintf("k%04d", 1:n), val = 1:n)
  res = fget_keys(tab, i = tab$key_col[c(1000, 200, 700, 1000)], j = "val", key = "key_col")
  expect_equal(res, c(1000L, 200L, 700L, 1000L))
})

test_that("fget_keys on non-unique key column", {
  tab = data.table(key_col = c("k0001", "k0002", "k0002"), val = c(1L, 2L, 3L))
  res = fget_keys(tab, i = c("k0001", "k0002"), j = "val", key = "key_col")
  expect_equal(res, c(1L, 2L))

  n = 1200L
  tab = data.table(key_col = sprintf("k%04d", 1:n), val = 1:n)
  tab[3, key_col := "k0002"]
  res = fget_keys(tab, i = c("k0001", "k0002"), j = "val", key = "key_col")
  expect_equal(res, c(1L, 2L))
})

#' @title Convert to a Regression Task
#'
#' @description
#' Convert object to a [TaskRegr].
#' This is a S3 generic. mlr3 ships with methods for the following objects:
#'
#' 1. [TaskRegr]: returns the object as-is, possibly cloned.
#' 2. [`formula`], [data.frame()], [matrix()], and [DataBackend]: provides an alternative to the
#'    constructor of [TaskRegr].
#' 3. [TaskClassif]: Calls [convert_task()].
#'
#' @inheritParams as_task
#'
#' @return [TaskRegr].
#' @export
#' @examples
#' as_task_regr(datasets::mtcars, target = "mpg")
as_task_regr = function(x, ...) {
  UseMethod("as_task_regr")
}


#' @rdname as_task_regr
#' @export
# nolint next
as_task_regr.TaskRegr = function(x, clone = FALSE, ...) {
  if (clone) x$clone() else x
}


#' @rdname as_task_regr
#' @template param_target
#' @param id (`character(1)`)\cr
#'   Id for the new task.
#'   Defaults to the (deparsed and substituted) name of the data argument.
#' @template param_label
#' @export
# nolint next
as_task_regr.data.frame = function(x, target, id = deparse1(substitute(x)), label = NA_character_, ...) {
  force(id)

  assert_data_frame(x, min.rows = 1L, min.cols = 1L, col.names = "unique")
  assert_choice(target, names(x))

  ii = which(map_lgl(keep(x, is.double), anyInfinite))
  if (length(ii)) {
    warning_input("Detected columns with unsupported Inf values in data: %s", str_collapse(names(ii)))
  }

  if (anyMissing(x[[target]])) {
    messagef("Target column '%s' contains missing values. This will be an error in the future.", target)
    # error_input("Target column '%s' must not contain missing values", target)
  }

  TaskRegr$new(id = id, backend = x, target = target, label = label)
}

#' @rdname as_task_regr
#' @export
# nolint next
as_task_regr.matrix = function(x, target, id = deparse1(substitute(x)), label = NA_character_, ...) {
  force(id)

  assert_matrix(x, mode = "numeric")
  assert_choice(target, colnames(x))

  as_task_regr(as.data.table(x), target = target, id = id, label = label, ...)
}

#' @rdname as_task_regr
#' @export
# nolint next
as_task_regr.DataBackend = function(x, target, id = deparse1(substitute(x)), label = NA_character_, ...) {
  force(id)

  assert_choice(target, x$colnames)

  TaskRegr$new(id = id, backend = x, target = target, label = label, ...)
}

#' @rdname as_task_regr
#' @inheritParams convert_task
#' @export
# nolint next
as_task_regr.TaskClassif = function(x, target, drop_original_target = FALSE, drop_levels = TRUE, ...) {
  convert_task(intask = x, target = target, new_type = "regr", drop_original_target = drop_original_target, drop_levels = drop_levels)
}

#' @rdname as_task_regr
#' @param data (`data.frame()`)\cr
#'   Data frame containing all columns referenced in formula `x`.
#' @export
# nolint next
as_task_regr.formula = function(x, data, id = deparse1(substitute(data)), label = NA_character_, ...) {
  force(id)

  assert_data_frame(data)
  assert_subset(all.vars(x), c(names(data), "."), .var.name = "formula")

  if (!attr(terms(x, data = data), "response")) {
    error_input("Formula %s is missing a response", format(x))
  }
  tab = model.frame(x, data, na.action = "na.pass")
  setattr(tab, "terms", NULL)
  setattr(tab, "na.action", NULL)
  target = all.vars(x)[1L]

  as_task_regr(tab, target = target, id = id, label = label, ...)
}

#' @title Convert to ResultData
#'
#' @description
#' This function allows to construct or convert to a [ResultData] object, the
#' result container used by [ResampleResult] and [BenchmarkResult].
#' A [ResampleResult] or [BenchmarkResult] can be initialized with the returned object.
#' Note that [ResampleResult]s can be converted to a [BenchmarkResult] with [as_benchmark_result()]
#' and multiple [BenchmarkResult]s can be combined to a larger [BenchmarkResult] with the
#' `$combine()` method of [BenchmarkResult].
#'
#' @param task ([Task]).
#' @param learners (list of trained [Learner]s).
#' @param resampling ([Resampling]).
#' @param iterations (`integer()`).
#' @param predictions (list of list of [Prediction]s).
#' @param learner_states (`list()`)\cr
#'   Learner states. If not provided, the states of `learners` are automatically extracted.
#' @param data_extra (`list()`)\cr
#'  Additional data for each iteration.
#' @param store_backends (`logical(1)`)\cr
#'   If set to `FALSE`, the backends of the [Task]s provided in `data` are
#'   removed.
#'
#' @return `ResultData` object which can be passed to the constructor of [ResampleResult].
#' @export
#' @examples
#' task = tsk("penguins")
#' learner = lrn("classif.rpart")
#' resampling = rsmp("cv", folds = 2)$instantiate(task)
#' iterations = seq_len(resampling$iters)
#'
#' # manually train two learners.
#' # store learners and predictions
#' learners = list()
#' predictions = list()
#' for (i in iterations) {
#'   l = learner$clone(deep = TRUE)
#'   learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
#'   predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
#' }
#'
#' rdata = as_result_data(task, learners, resampling, iterations, predictions)
#' ResampleResult$new(rdata)
as_result_data = function(
  task,
  learners,
  resampling,
  iterations,
  predictions,
  learner_states = NULL,
  data_extra = NULL,
  store_backends = TRUE
) {
  assert_task(task)
  assert_learners(learners, task = task)
  assert_resampling(resampling, instantiated = TRUE)
  assert_integer(iterations, any.missing = FALSE, lower = 1L, upper = resampling$iters, unique = TRUE)
  assert_list(predictions, types = "list")
  assert_list(learner_states, null.ok = TRUE)
  assert_list(data_extra, null.ok = TRUE)
  predictions = map(predictions, function(x) map(x, as_prediction_data))

  N = length(iterations)

  if (length(learners) != N) {
    error_input("Number of learners (%i) must match the number of resampling iterations (%i)", length(learners), N)
  }

  if (length(predictions) != N) {
    error_input(
      "Number of predictions (%i) must match the number of resampling iterations (%i)",
      length(predictions),
      N
    )
  }

  if (is.null(learner_states)) {
    learner_states = map(learners, "state")
  } else if (length(learner_states) != N) {
    error_input(
      "Number of learner_states (%i) must match the number of resampling iterations (%i)",
      length(learner_states),
      N
    )
  }

  if (resampling$task_hash != task$hash) {
    error_input("Resampling '%s' has not been trained on task '%s', hashes do not match", resampling$id, task$id)
  }

  if (!is.null(data_extra) && length(data_extra) != N) {
    error_input("Length of data_extra (%i) must match the number of resampling iterations (%i)", length(data_extra), N)
  }

  ResultData$new(
    data.table(
      task = list(task),
      learner = learners,
      learner_hash = hashes(learners),
      learner_state = learner_states,
      param_values = map(learners, function(x) x$param_set$values),
      resampling = list(resampling),
      iteration = iterations,
      prediction = predictions,
      uhash = UUIDgenerate()
    ),
    data_extra = data_extra,
    store_backends = store_backends
  )
}

#' @title Convert to PredictionData
#'
#' @name PredictionData
#' @rdname PredictionData
#'
#' @description
#' Objects of type `PredictionData` serve as a intermediate representation for objects of type [Prediction].
#' It is an internal data structure, implemented to optimize runtime and solve some issues
#' emerging while serializing R6 objects.
#' End-users typically do not need to worry about the details,
#' package developers are advised to continue reading for some technical information.
#'
#' Unlike most other \CRANpkg{mlr3} objects, `PredictionData` relies on the S3 class system.
#' The following operations must be supported to extend mlr3 for new task types:
#'
#' * [as_prediction_data()] converts objects to class `PredictionData`, e.g. objects of type [Prediction].
#' * [as_prediction()] converts objects to class [Prediction], e.g. objects of type `PredictionData`.
#' * `check_prediction_data()` is called on the return value of the predict method of a [Learner]
#'   to perform assertions and type conversions.
#'   Returns an update object of class `PredictionData`.
#' * `is_missing_prediction_data()` is used for the fallback learner (see [Learner]) to impute
#'   missing predictions.
#'   Returns vector with row ids which need imputation.
#'
#'
NULL

new_prediction_data = function(li, task_type) {
  extra_attributes = attributes(li)
  extra_attributes = extra_attributes[names(extra_attributes) %nin% c("names", "class")]

  li = discard(li, is.null)
  class(li) = c(fget_key(mlr_reflections$task_types, task_type, "prediction_data", "type"), "PredictionData")

  for (i in seq_along(extra_attributes)) {
    setattr(li, names(extra_attributes)[i], extra_attributes[[i]])
  }

  li
}

#' @rdname PredictionData
#'
#' @param task ([Task]).
#' @param learner ([Learner]).
#'
#' @export
create_empty_prediction_data = function(task, learner) {
  UseMethod("create_empty_prediction_data")
}

#' @export
print.PredictionData = function(x, ...) {
  catf("<%s:%i>", class(x)[1L], length(x$row_ids))
}

#' @rdname PredictionData
#' @param pdata ([PredictionData])\cr
#'   Named list inheriting from `"PredictionData"`.
#' @export
check_prediction_data = function(pdata, ...) {
  UseMethod("check_prediction_data")
}

#' @rdname PredictionData
#' @export
is_missing_prediction_data = function(pdata, ...) {
  UseMethod("is_missing_prediction_data")
}

#' @rdname PredictionData
#' @template param_row_ids
#' @export
filter_prediction_data = function(pdata, row_ids, ...) {
  UseMethod("filter_prediction_data")
}

test_that("PredictionDataClassif", {
  task = tsk("iris")
  learner = lrn("classif.featureless", predict_type = "prob")
  p = learner$train(task)$predict(task)
  pdata = p$data

  expect_s3_class(pdata, "PredictionDataClassif")
  expect_integer(pdata$row_ids, any.missing = FALSE)
  expect_factor(pdata$truth, levels = task$class_names, any.missing = FALSE)
  expect_factor(pdata$response, levels = task$class_names, any.missing = FALSE)
  expect_matrix(pdata$prob, nrows = task$nrow, ncols = length(task$class_names), any.missing = FALSE)

  expect_s3_class(c(pdata, pdata), "PredictionDataClassif")
  expect_prediction(as_prediction(pdata))
  expect_equal(as.data.table(p), as.data.table(as_prediction(pdata)))

  pdata = filter_prediction_data(pdata, row_ids = 1:3)
  expect_set_equal(pdata$row_ids, 1:3)
  expect_factor(pdata$truth, len = 3)
  expect_factor(pdata$response, len = 3)
  expect_matrix(pdata$prob, nrows = 3)
})

test_that("row sums of prob sums up to 1 ", {
  pdata = new_prediction_data(
    list(row_ids = 1:2, truth = factor(c("a", "b")), response = c("a", "b"), prob = matrix(c(0.5, 0.5, 0.5, 1), 2)),
    "classif"
  )
  expect_error(check_prediction_data(pdata), "sum up")
})

test_that("construction of empty PredictionDataClassif", {
  task = tsk("iris")

  learner = lrn("classif.featureless")
  learner$train(task)
  pred = learner$predict(task, row_ids = integer())
  expect_prediction(pred)
  expect_set_equal(pred$predict_types, "response")
  expect_integer(pred$row_ids, len = 0L)
  expect_factor(pred$truth, len = 0L, levels = task$class_names)
  expect_null(pred$prob)
  expect_data_table(as.data.table(pred), nrows = 0L, ncols = 3L)

  learner = lrn("classif.featureless", predict_type = "prob")
  learner$train(task)
  pred = learner$predict(task, row_ids = integer())
  expect_prediction(pred)
  expect_set_equal(pred$predict_types, c("response", "prob"))
  expect_integer(pred$row_ids, len = 0L)
  expect_factor(pred$truth, len = 0L, levels = task$class_names)
  expect_matrix(pred$prob, nrows = 0L, ncols = length(task$class_names))
  expect_data_table(as.data.table(pred), nrows = 0L, ncols = 6L)
})

test_that("combine with extra data", {
  pdata = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = factor(c("a", "b")),
      response = c("a", "b"),
      prob = matrix(c(0.5, 0.5, 0.5, 1), 2),
      extra = list(extra_col = c("a", "b"), extra_col2 = c(1, 2))
    ),
    "classif"
  )

  pdata2 = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = factor(c("a", "b")),
      response = c("a", "b"),
      prob = matrix(c(0.5, 0.5, 0.5, 1), 2),
      extra = list(extra_col = c("c", "d"), extra_col2 = c(3, 4))
    ),
    "classif"
  )

  expect_equal(c(pdata, pdata2)$extra, list(extra_col = c("a", "b", "c", "d"), extra_col2 = c(1, 2, 3, 4)))
  expect_equal(c(pdata, pdata2, keep_duplicates = FALSE)$extra, list(extra_col = c("c", "d"), extra_col2 = c(3, 4)))

  pdata2 = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = factor(c("a", "b")),
      response = c("a", "b"),
      prob = matrix(c(0.5, 0.5, 0.5, 1), 2),
      extra = list(extra_col = c("c", "d"), extra_col3 = c(3, 4))
    ),
    "classif"
  )

  expect_equal(
    c(pdata, pdata2)$extra,
    list(extra_col = c("a", "b", "c", "d"), extra_col2 = c(1, 2, NA, NA), extra_col3 = c(NA, NA, 3, 4))
  )

  pdata2 = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = factor(c("a", "b")),
      response = c("a", "b"),
      prob = matrix(c(0.5, 0.5, 0.5, 1), 2)
    ),
    "classif"
  )

  expect_error(c(pdata, pdata2), "Some predictions have extra data, others do not")
})

test_that("filtering with extra data", {
  pdata = new_prediction_data(
    list(
      row_ids = 1:2,
      truth = factor(c("a", "b")),
      response = c("a", "b"),
      prob = matrix(c(0.5, 0.5, 0.5, 1), 2),
      extra = list(extra_col = c("a", "b"), extra_col2 = c(1, 2))
    ),
    "classif"
  )

  expect_equal(filter_prediction_data(pdata, row_ids = 1)$extra, list(extra_col = "a", extra_col2 = 1))
  expect_equal(filter_prediction_data(pdata, row_ids = 2)$extra, list(extra_col = "b", extra_col2 = 2))
})

#' @title Learner Class
#'
#' @include mlr_reflections.R
#' @include warn_deprecated.R
#'
#' @description
#' This is the abstract base class for learner objects like [LearnerClassif] and [LearnerRegr].
#'
#' Learners are built around the three following key parts:
#'
#' * Methods `$train()` and `$predict()` which call internal methods or private methods `$.train()`/`$.predict()`).
#' * A [paradox::ParamSet] which stores meta-information about available hyperparameters,
#'   and also stores hyperparameter settings.
#' * Meta-information about the requirements and capabilities of the learner.
#' * The fitted model stored in field `$model`, available after calling `$train()`.
#'
#' Predefined learners are stored in the [dictionary][mlr3misc::Dictionary] [mlr_learners],
#' e.g. [`classif.rpart`][mlr_learners_classif.rpart] or [`regr.rpart`][mlr_learners_regr.rpart].
#'
#' More classification and regression learners are implemented in the add-on package \CRANpkg{mlr3learners}.
#' Learners for survival analysis (or more general, for probabilistic regression) can be found in \CRANpkg{mlr3proba}.
#' Unsupervised cluster algorithms are implemented in \CRANpkg{mlr3cluster}.
#' The dictionary [mlr_learners] gets automatically populated with the new learners
#' as soon as the respective packages are loaded.
#'
#' More (experimental) learners can be found in the GitHub repository:
#' \url{https://github.com/mlr-org/mlr3extralearners}.
#' A guide on how to extend \CRANpkg{mlr3} with custom learners can be found in the
#' [mlr3book](https://mlr3book.mlr-org.com).
#'
#' To combine the learner with preprocessing operations like factor encoding, \CRANpkg{mlr3pipelines} is recommended.
#' Hyperparameters stored in the `param_set` can be tuned with \CRANpkg{mlr3tuning}.
#'
#' @template param_id
#' @template param_task_type
#' @template param_param_set
#' @template param_predict_types
#' @template param_feature_types
#' @template param_learner_properties
#' @template param_packages
#' @template param_label
#' @template param_man
#'
#'
#' @section Optional Extractors:
#'
#' Specific learner implementations are free to implement additional getters to ease the access of certain parts
#' of the model in the inherited subclasses.
#'
#' For the following operations, extractors are standardized:
#'
#' * `importance(...)`: Returns the feature importance score as numeric vector.
#'   The higher the score, the more important the variable.
#'   The returned vector is named with feature names and sorted in decreasing order.
#'   Note that the model might omit features it has not used at all.
#'   The learner must be tagged with property `"importance"`.
#'   To filter variables using the importance scores, see package \CRANpkg{mlr3filters}.
#'
#' * `selected_features(...)`: Returns a subset of selected features as `character()`.
#'   The learner must be tagged with property `"selected_features"`.
#'
#' * `oob_error(...)`: Returns the out-of-bag error of the model as `numeric(1)`.
#'   The learner must be tagged with property `"oob_error"`.
#'
#' * `internal_valid_scores`: Returns the internal validation score(s) of the model as a named `list()`.
#'   Only available for [`Learner`]s with the `"validation"` property.
#'   If the learner is not trained yet, this returns `NULL`.
#'
#' * `internal_tuned_values`: Returns the internally tuned hyperparameters of the model as a named `list()`.
#'   Only available for [`Learner`]s with the `"internal_tuning"` property.
#'   If the learner is not trained yet, this returns `NULL`.
#'
#' @section Weights:
#'
#' Many learners support observation weights, indicated by their property `"weights"`.
#' The weights are stored in the [Task] where the column role `weights_learner` needs to be assigned
#' to a single numeric column.
#' If a task has weights and the learner supports them, they are used automatically.
#' If a task has weights but the learner does not support them, an error is thrown by default.
#' Both of these behaviors can be disabled by setting the `use_weights` field to `"ignore"`.
#' See the description of `use_weights` for more information.
#'
#' If the learner is set-up to use weights but the task does not have a designated weight column,
#' samples are considered to have equal weight.
#' When weights are being used, they are passed down to the learner directly;
#' the effect of weights depends on the specific learner.
#' Generally, weights do not need to sum up to 1.
#'
#' When implementing a Learner that uses weights, the `"weights"` property should be set.
#' The `$.train()`-method should then call the `$.get_weights()`-method to retrieve the weights from the task.
#' `$.get_weights()` will automatically discard weights when `use_weights` is set to `"ignore"`;
#'
#' @section Setting Hyperparameters:
#'
#' All information about hyperparameters is stored in the slot `param_set` which is a [paradox::ParamSet].
#' The printer gives an overview about the ids of available hyperparameters, their storage type, lower and upper bounds,
#' possible levels (for factors), default values and assigned values.
#' To set hyperparameters, call the `set_values()` method on the `param_set`:
#' ```
#' lrn = lrn("classif.rpart")
#' lrn$param_set$set_values(minsplit = 3, cp = 0.01)
#' ```
#' Note that this operation replaces all previously set hyperparameter values.
#' If you only intend to change one specific hyperparameter value and leave the others as-is,
#' you can use the helper function [mlr3misc::insert_named()]:
#' ```
#' lrn$param_set$values = mlr3misc::insert_named(lrn$param_set$values, list(cp = 0.001))
#' ```
#' If the learner has additional hyperparameters which are not encoded in the
#' [ParamSet][paradox::ParamSet], you can easily extend the learner.
#' Here, we add a factor hyperparameter with id `"foo"` and possible levels `"a"` and `"b"`:
#' ```
#' lrn$param_set$add(paradox::ParamFct$new("foo", levels = c("a", "b")))
#' ```
#'
#' @section Implementing Validation:
#' Some Learners, such as `XGBoost`, other boosting algorithms, or deep learning models (`mlr3torch`),
#' utilize validation data during the training to prevent overfitting or to log the validation performance.
#' It is possible to configure learners to be able to receive such an independent validation set during training.
#' To do so, one must:
#' * annotate the learner with the `"validation"` property
#' * implement the active binding `$internal_valid_scores` (see section *Optional Extractors*), as well as the
#'   private method `$.extract_internal_valid_scores()` which returns the (final) internal validation scores from the
#'   model of the [`Learner`] and returns them as a named `list()` of `numeric(1)`.
#'   If the model is not trained yet, this method should return `NULL`.
#' * Add the `validate` parameter, which can be either `NULL`, a ratio in $(0, 1)$, `"test"`, or `"predefined"`:
#'   * `NULL`: no validation
#'   * `ratio`: only proportion `1 - ratio` of the task is used for training and `ratio` is used for validation.
#'   * `"test"` means that the `"test"` task is used.
#'     **Warning**: This can lead to biased performance estimation.
#'     This option is only available if the learner is being trained via [resample()], [benchmark()] or functions that
#'     internally use them, e.g. `tune()` of \CRANpkg{mlr3tuning} or `batchmark()` of \CRANpkg{mlr3batchmark}.
#'     This is especially useful for hyperparameter tuning, where one might e.g. want to use the same validation data
#'     for early stopping and model evaluation.
#'   * `"predefined"` means that the task's (manually set) `$internal_valid_task` is used.
#'     See the [`Task`] documentation for more information.
#'
#' For an example how to do this, see [`LearnerClassifDebug`].
#' Note that in `.train()`, the `$internal_valid_task` will only be present if the `$validate` field of the `Learner`
#' is set to a non-`NULL` value.
#'
#' @section Implementing Internal Tuning:
#' Some learners such as `XGBoost` or `cv.glmnet` can internally tune hyperparameters.
#' XGBoost, for example, can tune the number of boosting rounds based on the validation performance.
#' CV Glmnet, on the other hand, can tune the regularization parameter based on an internal cross-validation.
#' Internal tuning *can* therefore rely on the internal validation data, but does not necessarily do so.
#'
#' In order to be able to combine this internal hyperparameter tuning with the standard hyperparameter optimization
#' implemented via \CRANpkg{mlr3tuning}, one must:
#' * annotate the learner with the `"internal_tuning"` property
#' * implement the active binding `$internal_tuned_values` (see section *Optional Extractors*) as well as the
#'   private method `$.extract_internal_tuned_values()` which extracts the internally tuned values
#'   from the [`Learner`]'s
#'   model and returns them as a named `list()`.
#'   If the model is not trained yet, this method should return `NULL`.
#' * Have at least one parameter tagged with `"internal_tuning"`, which requires to also provide a `in_tune_fn` and
#'   `disable_tune_fn`, and *should* also include a default aggregation function (`aggr`).
#'
#' For an example how to do this, see [`LearnerClassifDebug`].
#'
#' @section Implementing Marshaling:
#' Some [`Learner`]s have models that cannot be serialized as they e.g. contain external pointers.
#' In order to still be able to save them, use them with parallelization or callr encapsulation it is necessary
#' to implement how they should be (un)-marshaled. See [`marshaling`] for how to do this.
#'
#' @section Implementing Out-of-Bag Error:
#' Some [`Learner`]s can compute the out-of-bag error during training.
#' In order to do this, the learner must:
#' * annotate the learner with the `"oob_error"` property
#' * implement the private method `$.extract_oob_error()` which extracts the out-of-bag error
#'   from the [`Learner`]'s model and returns it as a `numeric(1)`.
#'
#' @template seealso_learner
#' @export
Learner = R6Class(
  "Learner",
  public = list(
    #' @field state (named `list()` | `NULL`)\cr
    #' Current (internal) state of the learner.
    #' Contains all information gathered during `train()` and `predict()`.
    #' It is not recommended to access elements from `state` directly.
    #' This is an internal data structure which may change in the future.
    state = NULL,

    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' Note that this object is typically constructed via a derived classes, e.g. [LearnerClassif] or [LearnerRegr].
    initialize = function(
      id,
      task_type,
      param_set = ps(),
      predict_types = character(),
      feature_types = character(),
      properties = character(),
      packages = character(),
      label = NA_character_,
      man = NA_character_
    ) {
      private$.id = assert_id(id)
      private$.label = assert_string(label, na.ok = TRUE)
      private$.task_type = assert_choice(task_type, mlr_reflections$task_types$type)
      private$.feature_types = assert_ordered_set(
        feature_types,
        mlr_reflections$task_feature_types,
        .var.name = "feature_types"
      )
      private$.predict_types = assert_ordered_set(
        predict_types,
        names(mlr_reflections$learner_predict_types[[task_type]]),
        empty.ok = FALSE,
        .var.name = "predict_types"
      )
      private$.predict_type = predict_types[1L]
      private$.properties = sort(assert_subset(properties, mlr_reflections$learner_properties[[task_type]]))
      private$.packages = union("mlr3", assert_character(packages, any.missing = FALSE, min.chars = 1L))
      private$.man = assert_string(man, na.ok = TRUE)
      private$.predict_sets = "test"
      private$.parallel_predict = FALSE
      private$.predict_raw = FALSE
      private$.timeout = c(train = Inf, predict = Inf)
      private$.deadline = as.POSIXct(c(train = Inf, predict = Inf))

      if ("weights" %in% private$.properties) {
        private$.use_weights = "use"
      } else {
        private$.use_weights = "error"
      }
      private$.param_set = param_set

      check_packages_installed(packages, msg = sprintf("Package '%%s' required but not installed for Learner '%s'", id))
    },

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s:%s>", class(self)[1L], self$id)
    },

    #' @description
    #' Printer.
    #' @param ... (ignored).
    print = function(...) {
      msg_h = if (is.null(self$label) || is.na(self$label)) "" else paste0(": ", self$label)
      model = if (is.null(self$model)) {
        "-"
      } else if (is_marshaled_model(self$model)) {
        "<marshaled>"
      } else {
        paste0(class(self$model)[1L])
      }

      cat_cli({
        cli_h1("{.cls {class(self)[1L]}} ({self$id}){msg_h}")
        cli_li("Model: {model}")
        cli_li("Parameters: {as_short_string(self$param_set$values, 1000L)}")
      })

      if (exists("validate", self)) {
        validate = format(self$validate)
        cat_cli(cli_li("Validate: {validate}"))
      }
      cat_cli(cli_li("Packages: {.pkg {self$packages}}"))

      pred_typs = replace(
        self$predict_types,
        self$predict_types == self$predict_type,
        paste0("[", self$predict_type, "]")
      )
      encapsulation = self$encapsulation[[1L]]
      fallback = if (encapsulation != "none") class(self$fallback)[[1L]] else "-"

      cat_cli({
        cli_li("Predict Types: {pred_typs}")
        cli_li("Feature Types: {self$feature_types}")
        cli_li("Encapsulation: {encapsulation} (fallback: {fallback})")
        cli_li("Properties: {self$properties}")
        cli_li("Other settings: use_weights = '{self$use_weights}', predict_raw = '{self$predict_raw}'")
      })

      w = self$warnings
      e = self$errors
      if (length(w)) {
        cat_cli(cli_alert_warning("Warnings: {w}"))
      }
      if (length(e)) {
        cat_cli(cli_alert_danger("Errors: {e}"))
      }
    },

    #' @description
    #' Opens the corresponding help page referenced by field `$man`.
    help = function() {
      open_help(self$man)
    },

    #' @description
    #' Train the learner on a set of observations of the provided `task`.
    #' Mutates the learner by reference, i.e. stores the model alongside other information in field `$state`.
    #'
    #' @param task ([Task]).
    #'
    #' @param row_ids (`integer()`)\cr
    #'   Vector of training indices as subset of `task$row_ids`.
    #'   For a simple split into training and test set, see [partition()].
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' task   = tsk("penguins")
    #' learner = lrn("classif.rpart")
    #' learner$train(task)
    train = function(task, row_ids = NULL) {
      task = assert_task(as_task(task))
      assert_learnable(task, self)
      row_ids = assert_row_ids(row_ids, task = task, null.ok = TRUE)

      if (!is.null(self$hotstart_stack)) {
        # search for hotstart learner
        start_learner = get_private(self$hotstart_stack)$.start_learner(self, task$hash)
      }
      if (is.null(self$hotstart_stack) || is.null(start_learner)) {
        # no hotstart learners stored or no adaptable model found
        learner = self
        mode = "train"
      } else {
        self$state = start_learner$clone()$state
        learner = self
        mode = "hotstart"
      }

      train_row_ids = if (!is.null(row_ids)) row_ids else task$row_roles$use

      train_result = learner_train(learner, task, train_row_ids = train_row_ids, mode = mode)
      self$model = unmarshal_model(model = self$state$model, inplace = TRUE)

      # store data prototype
      proto = task$data(rows = integer())
      self$state$data_prototype = proto

      # store the task w/o the data
      self$state$train_task = task_rm_backend(task$clone(deep = TRUE))

      invisible(self)
    },
    #' @description
    #' Uses the fitted model stored in `$state` to generate predictions for a set of observations
    #' from the provided `task`.
    #' This method requires that the learner has been previously trained using `$train()`.
    #'
    #' @param task ([Task])\cr
    #'   The task containing the observations to predict on.
    #'   Must be compatible with the learner's task type and feature types.
    #'   Unlike `$predict_newdata()`, no type conversion is done.
    #'
    #' @param row_ids (`integer()`)\cr
    #'   Vector of row indices from `task$row_ids` to predict on.
    #'   If `NULL` (default), predictions are made for all rows in the task.
    #'   For a simple train-test split, see [partition()].
    #'
    #' @return [Prediction] object containing the predictions for the specified observations.
    #' @examples
    #' task = tsk("penguins")
    #' learner = lrn("classif.rpart")$train(task)
    #' learner$predict(task)
    predict = function(task, row_ids = NULL) {
      # improve error message for the common mistake of passing a data.frame here
      if (is.data.frame(task)) {
        error_input("To predict on data.frames, use the method `$predict_newdata()` instead of `$predict()`")
      }
      task = assert_task(as_task(task))
      assert_predictable(task, self)
      row_ids = assert_row_ids(row_ids, task = task, null.ok = TRUE)

      if (is.null(self$state$model) && is.null(self$state$fallback_state$model)) {
        error_input("Cannot predict, Learner '%s' has not been trained yet", self$id)
      }

      # we need to marshal for call-r prediction and parallel prediction, but afterwards we reset the model
      # to it original state
      model_was_marshaled = is_marshaled_model(self$model)
      on.exit(
        {
          if (model_was_marshaled) {
            self$model = marshal_model(self$model, inplace = TRUE)
          } else {
            self$model = unmarshal_model(self$model, inplace = TRUE)
          }
        },
        add = TRUE
      )

      # reset learner predict time; this is only cumulative for multiple predict sets,
      # not for multiple calls to predict / predict_newdata
      self$state$predict_time = 0

      # we only have to marshal here for the parallel prediction case, because learner_predict() handles the
      # marshaling for call-r encapsulation itself
      if (isTRUE(self$parallel_predict) && nbrOfWorkers() > 1L) {
        row_ids = row_ids %??% task$row_ids
        chunked = chunk_vector(row_ids, n_chunks = nbrOfWorkers(), shuffle = FALSE)
        self$model = marshal_model(self$model, inplace = TRUE)
        pdata = future.apply::future_lapply(
          chunked,
          learner_predict,
          learner = self,
          task = task,
          future.globals = FALSE,
          future.seed = TRUE
        )
        pdata = do.call(c, pdata)
      } else {
        pdata = learner_predict(self, task, row_ids)
      }

      if (is.null(pdata)) {
        return(NULL)
      } else {
        as_prediction(pdata)
      }
    },

    #' @description
    #' Uses the model fitted during `$train()` to create a new [Prediction] based on the new data in `newdata`.
    #' Object `task` is the task used during `$train()` and required for conversion of `newdata`.
    #' If the learner's `$train()` method has been called, there is a (size reduced) version
    #' of the training task stored in the learner.
    #' If the learner has been fitted via [resample()] or [benchmark()],
    #' you need to pass the corresponding task stored in the [ResampleResult] or
    #' [BenchmarkResult], respectively.
    #' Further, [`auto_convert`] is used for type-conversions to ensure compatibility
    #' of features between `$train()` and `$predict()`.
    #'
    #' If the stored training task has a `weights_measure` column,
    #' *and* if `newdata` contains a column with the same name,
    #' that column must be numeric with no missing values and is used as measure weights column.
    #' Otherwise, no measure weights are used.
    #'
    #' @param newdata (any object supported by [as_data_backend()])\cr
    #'   New data to predict on.
    #'   All data formats convertible by [as_data_backend()] are supported, e.g.
    #'   `data.frame()` or [DataBackend].
    #'   If a [DataBackend] is provided as `newdata`, the row ids are preserved,
    #'   otherwise they are set to the sequence `1:nrow(newdata)`.
    #'
    #' @param task ([Task]).
    #'
    #' @return [Prediction].
    #' @examples
    #' task = tsk("penguins")
    #' learner = lrn("classif.rpart")$train(task)
    #' learner$predict_newdata(task$data(rows = 1:5))
    predict_newdata = function(newdata, task = NULL) {
      if (is.null(task)) {
        if (is.null(self$state$train_task)) {
          error_input("No task stored, and no task provided")
        }
        task = self$state$train_task$clone()
      } else {
        task = assert_task(as_task(task, clone = TRUE))
        assert_learnable(task, self)
        task = task_rm_backend(task)
      }

      newdata = as_data_backend(newdata)
      assert_names(newdata$colnames, must.include = task$feature_names)

      # set the target columns to NA if missing
      impute = setdiff(task$col_roles[["target"]], newdata$colnames)
      tab1 = if (length(impute)) {
        # create list with correct NA types and cbind it to the backend
        ci = insert_named(
          task$col_info[list(impute), c("id", "type", "levels"), on = "id", with = FALSE],
          list(value = NA)
        )
        na_cols = set_names(pmap(ci, function(..., nrow) rep(auto_convert(...), nrow), nrow = newdata$nrow), ci$id)
        invoke(data.table, .args = insert_named(na_cols, set_names(list(newdata$rownames), newdata$primary_key)))
      }

      # Perform type conversion where necessary
      keep_cols = intersect(newdata$colnames, task$col_info$id)
      ci = task$col_info[list(keep_cols), ][
        get("type") != col_info(newdata)[list(keep_cols), on = "id"]$type
      ]
      tab2 = do.call(
        data.table,
        Map(
          auto_convert,
          value = as.list(newdata$data(rows = newdata$rownames, cols = ci$id)),
          id = ci$id,
          type = ci$type,
          levels = ci$levels
        )
      )

      tab = cbind(tab1, tab2)
      if (ncol(tab)) {
        tab[[newdata$primary_key]] = newdata$rownames
        newdata = DataBackendCbind$new(newdata, DataBackendDataTable$new(tab, primary_key = newdata$primary_key))
      }

      prevci = task$col_info
      task$backend = newdata
      task$col_info = col_info(task$backend)
      task$col_info[, c("label", "fix_factor_levels")] = prevci[
        list(task$col_info$id),
        on = "id",
        c("label", "fix_factor_levels")
      ]
      task$col_info$fix_factor_levels[is.na(task$col_info$fix_factor_levels)] = FALSE

      # restore target levels from training task to preserve level ordering (e.g. positive class first)
      target_ids = task$target_names
      task$col_info[list(target_ids), "levels" := prevci[list(target_ids), "levels", on = "id"], on = "id"]
      task$row_roles$use = task$backend$rownames

      # reset column roles that are not in the newdata if they are optional
      optional_col_roles = mlr_reflections$task_col_roles_optional_newdata[[task$task_type]] %??%
        c("weights_learner", "weights_measure", "name", "order", "stratum", "group")
      task_col_roles = task$col_roles
      for (role in optional_col_roles) {
        if (any(task_col_roles[[role]] %nin% newdata$colnames)) {
          task_col_roles[[role]] = character(0)
        }
      }
      task$col_roles = task_col_roles

      self$predict(task)
    },

    #' @description
    #' Reset the learner, i.e. un-train by resetting the `state`.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' task = tsk("penguins")
    #' learner = lrn("classif.rpart")$train(task)
    #' learner$reset()
    reset = function() {
      self$state = NULL
      invisible(self)
    },

    #' @description
    #' Extracts the base learner from nested learner objects like
    #' `GraphLearner` in \CRANpkg{mlr3pipelines} or `AutoTuner` in
    #' \CRANpkg{mlr3tuning}.
    #' Returns the [Learner] itself for regular learners.
    #'
    #' @param recursive (`integer(1)`)\cr
    #'   Depth of recursion for multiple nested objects.
    #'
    #' @return [Learner]
    base_learner = function(recursive = Inf) {
      if (exists(".base_learner", envir = private, inherits = FALSE)) {
        private$.base_learner(recursive)
      } else {
        self
      }
    },

    #' @description
    #' Sets the encapsulation method and fallback learner for the train and predict steps.
    #' There are currently four different methods implemented:
    #'
    #' * `"none"`: Just runs the learner in the current session and measures the elapsed time.
    #'   Does not keep a log, output is printed directly to the console.
    #'   Works well together with [traceback()].
    #' * `"try"`: Similar to `"none"`, but catches error.
    #'   Output is printed to the console and not logged.
    #' * `"evaluate"`: Uses the package \CRANpkg{evaluate} to call the learner, measure time and do the logging.
    #' * `"callr"`: Uses the package \CRANpkg{callr} to call the learner, measure time and do the logging.
    #'   This encapsulation spawns a separate R session in which the learner is called.
    #'   While this comes with a considerable overhead,
    #'   it also guards your session from being torn down by segfaults.
    #' * `"mirai"`: Uses the package \CRANpkg{mirai} to call the learner,
    #'   measure time and do the logging.
    #'   This encapsulation calls the function in a `mirai` on a `daemon`.
    #'   The `daemon` can be pre-started via `daemons(1, .compute = "mlr3_encapsulation")`,
    #'   otherwise a new R session will be created for each encapsulated call.
    #'   If a `daemon` is already running with compute profile `"mlr3_encapsulation"`,
    #'   it will be used to execute all calls.
    #'   Using `mirai` is similarly safe as `callr` but much faster if several learners
    #'   are encapsulated one after the other on the same daemon.
    #'
    #' The fallback learner is fitted to create valid predictions in case that either the
    #' model fitting or the prediction of the original learner fails.
    #' If the training step or the predict step of the original learner fails,
    #' the fallback is used to make the predictions.
    #' If the original learner only partially fails during predict step
    #' (usually in the form of missing to predict some observations or producing
    #' some `NA` predictions),
    #' these missing predictions are imputed by the fallback.
    #' Note that the fallback is always trained, as we do not know in advance whether prediction will fail.
    #' If the training step fails, the `$model` field of the original learner is `NULL`.
    #' The results are reproducible across the different encapsulation methods.
    #'
    #' Note that for errors of class `Mlr3ErrorConfig`, the function always errs and no fallback learner
    #' is trained.
    #'
    #' Also see the section on error handling in the mlr3book:
    #' \url{https://mlr3book.mlr-org.com/chapters/chapter10/advanced_technical_aspects_of_mlr3.html#sec-error-handling}
    #'
    #' @param method `character(1)`\cr
    #'  One of `"none"`, `"try"`, `"evaluate"`, `"callr"` or `"mirai"`.
    #'  See the description for details.
    #' @param fallback [Learner]\cr
    #'  The fallback learner for failed predictions.
    #' @param when (`function(cond, stage)`)\cr
    #'  Function that takes in the condition (`cond`) and the stage (`"train"` or `"predict"`) and
    #'  returns `logical(1)` indicating whether to run the fallback learner.
    #'
    #'  If `NULL` (default), the fallback is always used, except for errors of class `Mlr3ErrorConfig`.
    #'
    #' @return `self` (invisibly).
    #' @examples
    #' learner = lrn("classif.rpart")
    #' fallback = lrn("classif.featureless")
    #' learner$encapsulate("try", fallback = fallback)
    encapsulate = function(method, fallback = NULL, when = NULL) {
      assert_choice(method, c("none", "try", "evaluate", "callr", "mirai"))

      private$.when = assert_function(when, null.ok = TRUE)

      if (!is.null(when)) {
        fargs = formalArgs(when)
        if ("..." %nin% fargs) {
          assert_subset(fargs, c("cond", "stage"))
        }
      }

      if (method != "none") {
        assert_learner(fallback, task_type = self$task_type)

        if (!identical(self$predict_type, fallback$predict_type)) {
          warning_config(
            "The fallback learner '%s' and the base learner '%s' have different predict types: '%s' != '%s'.",
            fallback$id,
            self$id,
            fallback$predict_type,
            self$predict_type,
            class = "Mlr3WarningConfigFallbackPredictType"
          )
        }

        # check properties
        properties = intersect(
          self$properties,
          c("twoclass", "multiclass", "missings", "importance", "selected_features")
        )
        missing_properties = setdiff(properties, fallback$properties)

        if (length(missing_properties)) {
          warning_config(
            "The fallback learner '%s' does not have the following properties of the learner '%s': %s.",
            fallback$id,
            self$id,
            str_collapse(missing_properties),
            class = "Mlr3WarningConfigFallbackProperties"
          )
        }
      } else if (method == "none" && !is.null(fallback)) {
        error_input("Fallback learner must be `NULL` if encapsulation is set to `none`.")
      }

      private$.encapsulation = c(train = method, predict = method)
      private$.fallback = fallback

      return(invisible(self))
    },

    #' @description
    #' Sets parameter values and fields of the learner.
    #' All arguments whose names match the name of a parameter of the [paradox::ParamSet] are set as parameters.
    #' All remaining arguments are assumed to be regular fields.
    #'
    #' @param ... (named `any`)\cr
    #'   Named arguments to set parameter values and fields.
    #' @param .values (named `any`)\cr
    #'   Named list of parameter values and fields.
    #' @examples
    #' learner = lrn("classif.rpart")
    #' learner$configure(minsplit = 3, parallel_predict = FALSE)
    #' learner$configure(.values = list(cp = 0.005))
    configure = function(..., .values = list()) {
      dots = list(...)
      assert_list(dots, names = "unique")
      assert_list(.values, names = "unique")
      assert_disjunct(names(dots), names(.values))
      new_values = insert_named(dots, .values)

      # set params in ParamSet
      if (length(new_values)) {
        param_ids = self$param_set$ids()
        ii = names(new_values) %in% param_ids
        if (any(ii)) {
          self$param_set$values = insert_named(self$param_set$values, new_values[ii])
          new_values = new_values[!ii]
        }
      } else {
        param_ids = character()
      }

      # remaining args go into fields
      if (length(new_values)) {
        ndots = names(new_values)
        for (i in seq_along(new_values)) {
          nn = ndots[[i]]
          if (!exists(nn, envir = self, inherits = FALSE)) {
            error_config(
              "Cannot set argument '%s' for '%s' (not a parameter, not a field).%s",
              nn,
              class(self)[1L],
              did_you_mean(nn, c(param_ids, setdiff(names(self), ".__enclos_env__")))
            )
          }
          self[[nn]] = new_values[[i]]
        }
      }

      return(invisible(self))
    },

    #' @description
    #' Returns the features selected by the model.
    #' The field `selected_features_impute` controls the behavior if the learner does not support feature selection.
    #' If set to `"error"`, an error is thrown, otherwise all features are returned.
    selected_features = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      if (private$.selected_features_impute == "error") {
        error_input("Learner does not support feature selection")
      } else {
        self$state$feature_names
      }
    }
  ),

  active = list(
    #' @field use_weights (`character(1)`)\cr
    #' How weights should be handled.
    #' Settings are `"use"` `"ignore"`, and `"error"`.
    #'
    #' * `"use"`: use weights, as supported by the underlying `Learner`.
    #'   Only available for `Learner`s with the property `"weights"`.
    #' * `"ignore"`: do not use weights.
    #' * `"error"`: throw an error if weights are present in the training `Task`.
    #'
    #' For `Learner`s with the property `"weights"`, this is initialized as `"use"`.
    #' For `Learner`s that do not support weights, i.e. without the `"weights"` property,
    #' this is initialized as `"error"`.
    #' The latter behavior is to avoid cases where a user erroneously assumes that a `Learner`
    #' supports weights when it does not.
    #' For `Learner`s that do not support weights, `use_weights` needs to be set to `"ignore"`
    #' if tasks with weights should be handled (by dropping the weights).
    #' See Section 'weights' for more details.
    use_weights = function(rhs) {
      if (!missing(rhs)) {
        assert_choice(rhs, c(if ("weights" %in% self$properties) "use", "ignore", "error"))
        private$.use_weights = rhs
      }
      private$.use_weights
    },

    #' @field model (any)\cr
    #' The fitted model. Only available after `$train()` has been called.
    model = function(rhs) {
      if (!missing(rhs)) {
        self$state$model = rhs
      }
      self$state$model
    },

    #' @field native_model (any)\cr
    #' The native model object from the upstream package.
    #' For most learners, this is identical to `$model`.
    #' However, some learners store additional information beyond the model from the upstream package.
    #' In such cases, `$model` contains a named list with the native model stored in element `model`
    #' along with additional information.
    #' The `$native_model` field can be overwritten by the learner to return the actual model object
    #' from the upstream package.
    #' The default returns `$model`.
    native_model = function(rhs) {
      assert_ro_binding(rhs)
      self$model
    },

    #' @field timings (named `numeric(2)`)\cr
    #' Elapsed time in seconds for the steps `"train"` and `"predict"`.
    #'
    #' When predictions for multiple predict sets were made during [resample()] or [benchmark()],
    #' the predict time shows the cumulative duration of all predictions.
    #' If `learner$predict()` is called manually, the last predict time gets overwritten.
    #'
    #' Measured via [mlr3misc::encapsulate()].
    timings = function(rhs) {
      assert_ro_binding(rhs)
      set_names(c(self$state$train_time %??% NA_real_, self$state$predict_time %??% NA_real_), c("train", "predict"))
    },

    #' @field log ([data.table::data.table()])\cr
    #' Returns the output (including warning and errors) as table with columns
    #'
    #' * `"stage"` ("train" or "predict"),
    #' * `"class"` ("output", "warning", or "error"), and
    #' * `"condition"` (`condition`).
    log = function(rhs) {
      assert_ro_binding(rhs)
      self$state$log
    },

    #' @field warnings (`list()`)\cr
    #' Logged warnings condition objects.
    warnings = function(rhs) {
      assert_ro_binding(rhs)
      get_log_condition(self$state, "warning")
    },

    #' @field errors (`list()`)\cr
    #' Logged error condition objects.
    errors = function(rhs) {
      assert_ro_binding(rhs)
      get_log_condition(self$state, "error")
    },

    #' @field hash (`character(1)`)\cr
    #' Hash (unique identifier) for this object.
    #' The hash is calculated based on the learner id, the parameter settings,
    #' the predict type, the fallback hash, the parallel predict setting,
    #' the validate setting, the predict sets, and the predict raw setting.
    hash = function(rhs) {
      assert_ro_binding(rhs)
      calculate_hash(
        class(self),
        self$id,
        self$param_set$values,
        private$.predict_type,
        self$fallback$hash,
        self$parallel_predict,
        get0("validate", self),
        self$predict_sets,
        private$.use_weights,
        private$.predict_raw
      )
    },

    #' @field phash (`character(1)`)\cr
    #' Hash (unique identifier) for this partial object, excluding some components
    #' which are varied systematically during tuning (parameter values).
    phash = function(rhs) {
      assert_ro_binding(rhs)
      calculate_hash(
        class(self),
        self$id,
        private$.predict_type,
        self$fallback$hash,
        self$parallel_predict,
        get0("validate", self),
        private$.use_weights,
        private$.predict_raw
      )
    },

    #' @field predict_type (`character(1)`)\cr
    #' Stores the currently active predict type, e.g. `"response"`.
    #' Must be an element of `$predict_types`.
    #' A few learners already use the predict type during training.
    #' So there is no guarantee that changing the predict type after training will have any effect
    #' or does not lead to errors.
    predict_type = function(rhs) {
      if (missing(rhs)) {
        return(private$.predict_type)
      }

      assert_string(rhs, .var.name = "predict_type")
      if (rhs %nin% self$predict_types) {
        error_input("Learner '%s' does not support predict type '%s'", self$id, rhs) # TODO error_learner?
      }
      private$.predict_type = rhs
    },

    #' @template field_param_set
    param_set = function(rhs) {
      if (!missing(rhs) && !identical(rhs, private$.param_set)) {
        error_input("param_set is read-only.")
      }
      private$.param_set
    },

    #' @field fallback ([Learner])\cr
    #' Returns the fallback learner set with `$encapsulate()`.
    fallback = function(rhs) {
      assert_ro_binding(rhs)
      private$.fallback
    },

    #' @field encapsulation (`character(2)`)\cr
    #' Returns the encapsulation settings set with `$encapsulate()`.
    encapsulation = function(rhs) {
      assert_ro_binding(rhs)
      private$.encapsulation
    },

    #' @field hotstart_stack ([HotstartStack])\cr.
    #' Stores `HotstartStack`.
    hotstart_stack = function(rhs) {
      if (missing(rhs)) {
        return(private$.hotstart_stack)
      }
      assert_r6(rhs, "HotstartStack", null.ok = TRUE)
      private$.hotstart_stack = rhs
    },

    #' @field selected_features_impute (`character(1)`)\cr
    #' Controls the behavior if the learner does not support feature selection.
    #' If set to `"error"`, an error is thrown.
    #' If set to `"all"` the complete feature set is returned.
    selected_features_impute = function(rhs) {
      if (missing(rhs)) {
        return(private$.selected_features_impute)
      }
      private$.selected_features_impute = assert_choice(rhs, c("error", "all"))
    },

    #' @field predict_types (`character()`)\cr
    #' Stores the possible predict types the learner is capable of.
    #' A complete list of candidate predict types, grouped by task type,
    #' is stored in [`mlr_reflections$learner_predict_types`][mlr_reflections].
    #' This field is read-only.
    predict_types = function(rhs) {
      assert_ro_binding(rhs)
      private$.predict_types
    },

    #' @template field_id
    id = function(rhs) {
      if (missing(rhs)) {
        return(private$.id)
      }
      private$.id = assert_id(rhs)
    },

    #' @template field_label
    label = function(rhs) {
      if (missing(rhs)) {
        return(private$.label)
      }
      private$.label = assert_string(rhs, na.ok = TRUE)
    },

    #' @template field_task_type
    task_type = function(rhs) {
      if (missing(rhs)) {
        return(private$.task_type)
      }
      private$.task_type = assert_choice(rhs, mlr_reflections$task_types$type)
    },

    #' @field feature_types (`character()`)\cr
    #' Stores the feature types the learner can handle, e.g. `"logical"`, `"numeric"`, or `"factor"`.
    #' A complete list of candidate feature types, grouped by task type,
    #' is stored in [`mlr_reflections$task_feature_types`][mlr_reflections].
    feature_types = function(rhs) {
      if (missing(rhs)) {
        return(private$.feature_types)
      }
      private$.feature_types = assert_subset(rhs, mlr_reflections$task_feature_types)
    },

    #' @field properties (`character()`)\cr
    #' Stores a set of properties/capabilities the learner has.
    #' A complete list of candidate properties, grouped by task type,
    #' is stored in [`mlr_reflections$learner_properties`][mlr_reflections].
    properties = function(rhs) {
      if (missing(rhs)) {
        return(private$.properties)
      }
      # Allow all valid properties from any task type to support dynamic property setting
      all_properties = unique(unlist(mlr_reflections$learner_properties, use.names = FALSE))
      private$.properties = sort(assert_subset(rhs, all_properties))
    },

    #' @template field_packages
    packages = function(rhs) {
      if (missing(rhs)) {
        return(private$.packages)
      }
      # workaround for mlr3torch
      if (is.null(rhs)) {
        rhs = character(0)
      }
      private$.packages = union("mlr3", assert_character(rhs, any.missing = FALSE, min.chars = 1L))
    },

    #' @template field_predict_sets
    predict_sets = function(rhs) {
      if (missing(rhs)) {
        return(private$.predict_sets)
      }
      if (is.null(rhs)) {
        private$.predict_sets = NULL
      } else {
        private$.predict_sets = assert_subset(rhs, mlr_reflections$predict_sets, empty.ok = FALSE)
      }
    },

    #' @field parallel_predict (`logical(1)`)\cr
    #' If set to `TRUE`, use \CRANpkg{future} to calculate predictions in parallel (default: `FALSE`).
    #' The row ids of the `task` will be split into [future::nbrOfWorkers()] chunks,
    #' and predictions are evaluated according to the active [future::plan()].
    #' This currently only works for methods `Learner$predict()` and `Learner$predict_newdata()`,
    #' and has no effect during [resample()] or [benchmark()] where you have other means
    #' to parallelize.
    #'
    #' Note that the recorded time required for prediction is not properly defined and
    #' depends on the parallelization backend.
    parallel_predict = function(rhs) {
      if (missing(rhs)) {
        return(private$.parallel_predict)
      }
      private$.parallel_predict = assert_flag(rhs)
    },

    #' @field predict_raw (`logical(1)`)\cr
    #' If set to `TRUE`, the raw prediction object from the upstream model is stored in the
    #' [Prediction] object (default: `FALSE`).
    #' The raw prediction is stored as-is, without validation or subsetting during filtering.
    #' When multiple predictions are combined via `c()`, the individual raw objects are collected into a `list()`.
    #' Individual learner implementations must support this flag by including a `raw` element in the list returned
    #' by `$.predict()`.
    predict_raw = function(rhs) {
      if (missing(rhs)) {
        return(private$.predict_raw)
      }
      private$.predict_raw = assert_flag(rhs)
    },

    #' @field timeout (named `numeric(2)`)\cr
    #' Timeout for the learner's train and predict steps, in seconds.
    #' This works differently for different encapsulation methods, see
    #' [mlr3misc::encapsulate()].
    #' Default is `c(train = Inf, predict = Inf)`.
    #' Also see the section on error handling in the mlr3book:
    #' \url{https://mlr3book.mlr-org.com/chapters/chapter10/advanced_technical_aspects_of_mlr3.html#sec-error-handling}
    timeout = function(rhs) {
      if (missing(rhs)) {
        return(private$.timeout)
      }
      assert_numeric(rhs, lower = 0, any.missing = FALSE, .var.name = "timeout")
      assert_names(names(rhs), type = "unique", subset.of = c("train", "predict"), .var.name = "names of timeout")
      # Merge with current timeout values
      private$.timeout[names(rhs)] = rhs
      private$.timeout
    },

    #' @field deadline (named `POSIXct(2)`)\cr
    #' Deadline for the learner's train and predict steps.
    #' This uses the same mechanism as `timeout`, but with absolute time (as
    #' date-time) instead of a relative time in seconds.
    #' Default is `as.POSIXct(c(train = Inf, predict = Inf))`.
    deadline = function(rhs) {
      if (missing(rhs)) {
        return(private$.deadline)
      }
      if (is.numeric(rhs)) {
        rhs = as.POSIXct(rhs)
      }
      assert_names(names(rhs), type = "unique", subset.of = c("train", "predict"), .var.name = "names of deadline")
      assert_names(names(rhs), subset.of = c("train", "predict"), .var.name = "names of deadline")
      private$.deadline[names(rhs)] = rhs
      private$.deadline = rhs
    },

    #' @template field_man
    man = function(rhs) {
      if (missing(rhs)) {
        return(private$.man)
      }
      private$.man = assert_string(rhs, na.ok = TRUE)
    }
  ),

  private = list(
    .id = NULL,
    .label = NULL,
    .task_type = NULL,
    .feature_types = NULL,
    .properties = NULL,
    .packages = NULL,
    .predict_sets = NULL,
    .parallel_predict = NULL,
    .predict_raw = NULL,
    .timeout = NULL,
    .deadline = NULL,
    .man = NULL,
    .when = NULL,
    .use_weights = NULL,
    .encapsulation = c(train = "none", predict = "none"),
    .fallback = NULL,
    .predict_type = NULL,
    .predict_types = NULL,
    .param_set = NULL,
    .hotstart_stack = NULL,
    .selected_features_impute = "error",

    # retrieve weights from a task, if it has weights and if the user did not
    # deactivate weight usage through `self$use_weights`.
    # - `task`: Task to retrieve weights from
    # - `no_weights_val`: Value to return if no weights are found (default NULL)
    # return: Numeric vector of weights or `no_weights_val` (default NULL)
    .get_weights = function(task, no_weights_val = NULL) {
      if ("weights" %nin% self$properties) {
        error_mlr3("private$.get_weights should not be used in Learners that do not have the 'weights' property.")
      }
      if (self$use_weights == "use" && "weights_learner" %in% task$properties) {
        task$weights_learner$weight
      } else {
        no_weights_val
      }
    },

    deep_clone = function(name, value) {
      switch(
        name,
        .param_set = value$clone(deep = TRUE),
        .fallback = if (is.null(value)) NULL else value$clone(deep = TRUE),
        state = {
          if (!is.null(value$train_task)) {
            value$train_task = value$train_task$clone(deep = TRUE)
          }
          value$log = copy(value$log)
          value
        },
        value
      )
    }
  )
)

#' @export
rd_info.Learner = function(obj, ...) {
  x = c(
    "",
    sprintf("* Task type: %s", rd_format_string(obj$task_type)),
    sprintf("* Predict Types: %s", rd_format_string(obj$predict_types)),
    sprintf("* Feature Types: %s", rd_format_string(obj$feature_types)),
    sprintf("* Required Packages: %s", rd_format_packages(obj$packages))
  )
  paste(x, collapse = "\n")
}

get_log_condition = function(state, condition) {
  if (is.null(state$log)) {
    list()
  } else {
    fget_key(state$log, condition, "condition", "class")
  }
}

#' @export
default_values.Learner = function(x, search_space, task, ...) {
  # nolint next
  values = default_values(x$param_set)

  if (any(search_space$ids() %nin% names(values))) {
    error_input(
      "Could not find default values for the following parameters: %s",
      str_collapse(setdiff(search_space$ids(), names(values)))
    )
  }

  values[search_space$ids()]
}

#' @export
marshal_model.learner_state = function(model, inplace = FALSE, ...) {
  if (is.null(model$model)) {
    return(model)
  }
  mm = marshal_model(model$model, inplace = inplace, ...)
  if (!is_marshaled_model(mm)) {
    return(model)
  }
  model$model = mm
  structure(
    list(
      marshaled = model,
      packages = "mlr3"
    ),
    class = c("learner_state_marshaled", "list_marshaled", "marshaled")
  )
}

#' @export
unmarshal_model.learner_state_marshaled = function(model, inplace = FALSE, ...) {
  mm = model$marshaled
  mm$model = unmarshal_model(mm$model, inplace = inplace, ...)
  return(mm)
}

test_that("autotest", {
  learner = lrn("regr.rpart")
  expect_learner(learner)
  result = run_autotest(learner)
  expect_true(result, info = result$error)

  exclude = c(
    "formula",
    "data",
    "weights",
    "subset",
    "na.action",
    "method",
    "model",
    "x",
    "y",
    "parms",
    "control",
    "cost",
    "keep_model",
    "use_weights"
  )
  result = run_paramtest(learner, list(rpart::rpart, rpart::rpart.control), exclude, tag = "train")
  expect_true(result, info = result$error)

  exclude = c(
    "object", # handled internally
    "newdata", # handled internally
    "type", # handled internally
    "na.action" # handled internally
  )
  result = run_paramtest(learner, rpart:::predict.rpart, exclude, tag = "predict")
  expect_true(result, info = result$error)
})

test_that("variable importance", {
  task = TaskRegr$new(
    "foo",
    as_data_backend(cbind(iris, data.frame(unimportant = runif(150)))),
    target = "Sepal.Length"
  )
  learner = lrn("regr.rpart")
  learner$train(task)
  imp = learner$importance()
  expect_numeric(imp, min.len = 1L, any.missing = FALSE)
  expect_names(names(imp), subset.of = task$feature_names)
  expect_false(is.unsorted(rev(imp)))
})

test_that("selected_features", {
  task = TaskRegr$new(
    "foo",
    as_data_backend(cbind(iris, data.frame(unimportant = runif(150)))),
    target = "Sepal.Length"
  )
  learner = lrn("regr.rpart", maxdepth = 2)
  sf = learner$train(task)$selected_features()
  expect_subset(sf, task$feature_names, empty.ok = FALSE)
})

test_that("default_values on rpart", {
  learner = lrn("regr.rpart")
  search_space = ps(
    minsplit = p_int(2, 128, logscale = TRUE),
    minbucket = p_int(1, 64, logscale = TRUE),
    cp = p_dbl(1e-04, 1e-1, logscale = TRUE)
  )
  task = tsk("pima")

  values = default_values(learner, search_space, task)
  expect_names(names(values), permutation.of = c("minsplit", "minbucket", "cp"))
})

test_that("use_weights actually influences the model", {
  # Task with weights_learner role defined in helper_misc.R
  task = cars_weights_learner
  learner = lrn("regr.rpart", use_weights = "use")
  learner$train(task)
  p1 = learner$predict(task)

  learner = lrn("regr.rpart", use_weights = "ignore")
  learner$train(task)
  p2 = learner$predict(task)

  # Predictions should differ if weights were used
  expect_false(all(p1$response == p2$response))
})

test_that("regression task", {
  data("mtcars", package = "datasets")
  task_train = as_task_regr(mtcars, "mpg")
  learner = lrn("regr.rpart")
  learner$train(task_train)

  mtcars$mpg = NULL
  task_predict = as_task_unsupervised(mtcars, id = "Test")
  expect_task_unsupervised(task_predict)
  expect_prediction(learner$predict(task_predict))
})


test_that("classification task", {
  data("iris", package = "datasets")
  task_train = as_task_classif(iris, "Species")
  learner = lrn("classif.rpart")
  learner$train(task_train)

  iris$Species = NULL
  task_predict = as_task_unsupervised(iris, id = "Test")
  expect_task_unsupervised(task_predict)
  expect_prediction(learner$predict(task_predict))
})

test_that("empty classif task", {
  data("iris", package = "datasets")
  task_train = as_task_classif(iris, "Species")
  learner = lrn("classif.rpart")
  learner$train(task_train)

  iris$Species = NULL
  iris = iris[NULL, ]
  task_predict = as_task_unsupervised(iris, id = "Test")
  expect_task_unsupervised(task_predict)
})

test_that("empty regr task", {
  data("mtcars", package = "datasets")
  task_train = as_task_regr(mtcars, "mpg")
  learner = lrn("regr.rpart")
  learner$train(task_train)

  mtcars$mpg = NULL
  mtcars = mtcars[NULL, ]
  task_predict = as_task_unsupervised(mtcars, id = "Test")
  expect_task_unsupervised(task_predict)
})

test_that("fallback", {
  data("iris", package = "datasets")
  task_train = as_task_classif(iris, "Species")
  learner = lrn("classif.debug", error_predict = 1)
  learner$encapsulate("evaluate", lrn("classif.featureless"))
  learner$train(task_train)

  iris$Species = NULL
  task_predict = as_task_unsupervised(iris, id = "Test")
  expect_task_unsupervised(task_predict)
  expect_prediction(learner$predict(task_predict))
})

test_that("assertions work", {
  data("iris", package = "datasets")
  iris$Species = NULL
  task = as_task_unsupervised(iris, id = "Test")
  learner = lrn("classif.rpart")

  expect_error(assert_learnable(task, learner), "cannot be trained")
  expect_null(assert_predictable(task, learner))

  at = learner
  class(at) = c("AutoTuner", "Learner", "R6")
  expect_error(assert_learnable(task, at), "cannot be trained")
  expect_null(assert_predictable(task, at))
})

test_that("train with unsupervised task fails", {
  data("iris", package = "datasets")
  iris$Species = NULL
  task_train = as_task_unsupervised(iris, id = "Test")
  learner = lrn("classif.rpart")
  expect_error(learner$train(task_train), "cannot be trained")
})

test_that("autotest", {
  learner = lrn("classif.rpart")
  expect_learner(learner)
  result = run_autotest(learner)
  expect_true(result, info = result$error)

  exclude = c(
    "formula",
    "data",
    "weights",
    "subset",
    "na.action",
    "method",
    "model",
    "x",
    "y",
    "parms",
    "control",
    "cost",
    "keep_model",
    "use_weights"
  )
  result = run_paramtest(learner, list(rpart::rpart, rpart::rpart.control), exclude, tag = "train")
  expect_true(result, info = result$error)

  exclude = c(
    "object", # handled internally
    "newdata", # handled internally
    "type", # handled internally
    "na.action" # handled internally
  )
  result = run_paramtest(learner, rpart:::predict.rpart, exclude, tag = "predict")
  expect_true(result, info = result$error)
})

test_that("variable importance", {
  task = TaskClassif$new("foo", as_data_backend(cbind(iris, data.frame(unimportant = runif(150)))), target = "Species")
  learner = lrn("classif.rpart")
  learner$train(task)
  imp = learner$importance()
  expect_numeric(imp, min.len = 1L, any.missing = FALSE)
  expect_names(names(imp), subset.of = task$feature_names)
  expect_false(is.unsorted(rev(imp)))
})

test_that("selected_features", {
  task = TaskClassif$new("foo", as_data_backend(cbind(iris, data.frame(unimportant = runif(150)))), target = "Species")
  learner = lrn("classif.rpart", maxdepth = 2)
  sf = learner$train(task)$selected_features()
  expect_subset(sf, task$feature_names, empty.ok = FALSE)
})

test_that("use_weights actually influences the model", {
  task = TaskClassif$new(
    "foo",
    as_data_backend(cbind(iris, data.frame(w = rep(c(1, 10, 100), each = 50)))),
    target = "Species"
  )
  task$set_col_roles("w", "weights_learner")
  learner = lrn("classif.rpart", use_weights = "use")
  learner$train(task)
  c1 = learner$predict(task)$confusion
  learner = lrn("classif.rpart", use_weights = "ignore")
  learner$train(task)
  c2 = learner$predict(task)$confusion
  expect_false(all(c1 == c2))
})

test_that("default_values on rpart", {
  learner = lrn("classif.rpart")
  search_space = ps(
    minsplit = p_int(2, 128, logscale = TRUE),
    minbucket = p_int(1, 64, logscale = TRUE),
    cp = p_dbl(1e-04, 1e-1, logscale = TRUE)
  )
  task = tsk("pima")

  values = default_values(learner, search_space, task)
  expect_names(names(values), permutation.of = c("minsplit", "minbucket", "cp"))
})

# this test / files was missing, only classif.debug was unit-tested
# I added at least a few basic tests when i added methods "importance" and "selected_features"

test_that("Simple training/predict", {
  task = tsk("mtcars")
  learner = lrn("regr.debug")
  expect_learner(learner, task)

  prediction = learner$train(task)$predict(task)
  expect_class(learner$model, "regr.debug_model")
  expect_numeric(learner$model$response, len = 1L, any.missing = FALSE)
  expect_numeric(prediction$response, any.missing = FALSE)
})


test_that("importance and selected features", {
  l = lrn("regr.debug")
  task = tsk("mtcars")
  l$train(task)
  expect_equal(l$selected_features(), character(0))
  expect_equal(l$importance(), set_names(rep(0, task$n_features), task$feature_names))
})

test_that("weights are respected", {
  # Task with weights_learner role defined in helper_misc.R
  task = cars_weights_learner
  learner = lrn("regr.debug")

  # Check learner property
  expect_true("weights" %in% learner$properties)

  # Train with response prediction
  learner$train(task)

  # Calculate expected weighted mean and sd manually
  dt = task$data()
  expected_mean = weighted.mean(dt$dist, task$weights_learner$weight)

  # Check model components
  expect_equal(learner$model$response, expected_mean)

  # Train with quantile prediction
  learner$predict_type = "quantiles"
  learner$quantiles = c(0.25, 0.5, 0.75)
  learner$quantile_response = 0.5
  learner$train(task)

  # Calculate expected weighted quantiles manually
  expected_quantiles = quantile_weighted(dt$dist, probs = c(0.25, 0.5, 0.75), weights = task$weights_learner$weight)

  # Check model components for quantiles
  expect_equal(learner$model$quantiles, unname(expected_quantiles))
  expect_equal(learner$model$quantile_probs, c(0.25, 0.5, 0.75))
})

test_that("quantiles work", {
  task = tsk("mtcars")
  learner = lrn("regr.debug", predict_type = "quantiles")
  learner$quantiles = c(0.25, 0.5, 0.75)
  learner$quantile_response = 0.5
  learner$train(task)
  expect_equal(learner$quantiles, c(0.25, 0.5, 0.75))
  expect_equal(learner$quantile_response, 0.5)

  prediction = learner$predict(tsk("mtcars"))
  expect_matrix(prediction$quantiles, nrows = nrow(task$data()), ncols = length(learner$quantiles), any.missing = FALSE)
  expect_numeric(prediction$response, len = nrow(task$data()), any.missing = FALSE)

  learner = lrn("regr.debug", predict_type = "quantiles")
  expect_error(learner$train(task), "Quantiles must be set")

  learner$quantiles = c(0.25, 0.5, 0.75)
  expect_error(learner$train(task), "Quantile response must be set")
})

test_that("Simple training/predict", {
  task = tsk("iris")
  learner = lrn("classif.debug")
  expect_learner(learner, task)

  prediction = learner$train(task)$predict(task)
  expect_class(learner$model, "classif.debug_model")
  expect_character(learner$model$response, len = 1L, any.missing = FALSE)
  expect_factor(prediction$response, any.missing = FALSE, levels = levels(learner$model))
})

test_that("updating model works", {
  task = tsk("iris")
  learner = lrn("classif.debug", save_tasks = TRUE)
  learner$train(task, 1:10)
  expect_task(learner$model$task_train)
  prediction = learner$predict(task, row_ids = 11:20)
  expect_task(learner$model$task_predict)

  itrain = task$clone(TRUE)$filter(1:10)
  ipredict = task$clone(TRUE)$filter(11:20)

  expect_equal(hashes(learner$model[c("task_train", "task_predict")]), hashes(list(itrain, ipredict)))
})

test_that("updating model works / resample", {
  learner = lrn("classif.debug", save_tasks = TRUE)
  rr = resample(tsk("iris"), learner, rsmp("holdout"), store_models = TRUE)
  new_learner = rr$learners[[1]]
  expect_list(new_learner$model, len = 7)
})

test_that("NA predictions", {
  task = tsk("iris")
  learner = lrn("classif.debug", predict_missing = 0.5, predict_type = "response")
  learner$train(task)
  p = learner$predict(task)
  expect_equal(count_missing(p$response), 75L)

  learner = lrn("classif.debug", predict_missing = 0.5, predict_type = "prob")
  learner$train(task)
  p = learner$predict(task)
  expect_equal(count_missing(p$response), 75L)
  expect_equal(is.na(p$response), apply(p$prob, 1, anyMissing))
})

test_that("default_values", {
  learner = lrn("classif.debug")
  search_space = ps(iter = p_int(1, 10))
  task = tsk("pima")

  values = default_values(learner, search_space, task)
  expect_names(names(values), identical.to = "iter")
})

test_that("default_values works with empty search space", {
  learner = lrn("classif.debug")
  expect_list(default_values(learner, ps(), task), len = 0)
})

test_that("validation and internal tuning", {
  task = tsk("iris")
  learner = lrn("classif.debug", iter = 100, early_stopping = TRUE, validate = 0.3, predict_type = "prob")
  learner$train(task)
  expect_list(learner$internal_valid_scores, len = 2L, types = "numeric")
  expect_permutation(names(learner$internal_valid_scores), c("acc", "mbrier"))
})

test_that("disable internal tuning", {
  learner = lrn("classif.debug", early_stopping = TRUE, iter = 100, validate = 0.2)
  learner$param_set$disable_internal_tuning("iter")
  expect_false(learner$param_set$values$early_stopping)
})

test_that("marshaling", {
  l = lrn("classif.debug")
  expect_learner(l, tsk("iris"))
  task = tsk("iris")
  l$train(task)
  p1 = l$predict(task)
  p2 = l$marshal()$unmarshal()$predict(task)
  expect_equal(p1, p2)
})

test_that("importance and selected features", {
  l = lrn("classif.debug")
  task = tsk("iris")
  l$train(task)
  expect_equal(l$selected_features(), character(0))
  expect_equal(l$importance(), set_names(rep(0, task$n_features), task$feature_names))
})

test_that("weights are respected", {
  # Create weighted task
  data = iris
  data$weights = fifelse(data$Species == "setosa", 1000, 1)
  task = TaskClassif$new(
    id = "iris_weights",
    backend = data,
    target = "Species"
  )
  task$set_col_roles("weights", roles = "weights_learner")

  # Create learner
  learner = lrn("classif.debug")

  # Train repeatedly and collect responses
  responses = character(100)
  for (i in 1:100) {
    # learner$reset() # reset not needed as train overwrites model
    learner$train(task)
    responses[i] = learner$model$response
  }

  # Check results
  # Expect 'setosa' to be the overwhelmingly most frequent response
  response_table = table(responses)
  # check if setosa is the most frequent sampled class, probability > 0.9
  expect_true(response_table["setosa"] > 90)
  expect_true(all(names(response_table) %in% task$class_names))
})

#' @title Measure Internal Validation Score
#'
#' @name mlr_measures_internal_valid_score
#' @include Measure.R
#'
#' @description
#' Returns the selected internal validation score of the [Learner] for learners with property `"validation"`.
#' Returns `NA` for unsupported learners, when no validation was done, or when the selected id was not found.
#' The `id` of this measure is set to the value of `select` if provided.
#'
#' @templateVar id internal_valid_score
#' @template measure
#'
#' @template seealso_measure
#' @export
#' @examples
#' rr = resample(tsk("iris"), lrn("classif.debug", validate = 0.3), rsmp("holdout"))
#' rr$score(msr("internal_valid_score", select = "acc"))
MeasureInternalValidScore = R6Class(
  "MeasureInternalValidScore",
  inherit = Measure,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #' @param select  (`character(1)`)\cr
    #'   Which of the internal validation scores to select.
    #'   Which scores are available depends on the learner and its configuration.
    #'   By default, the first score is chosen.
    #' @param minimize (`logical(1)`)\cr
    #'   Whether smaller values are better.
    #'   Must be set to use for tuning.
    initialize = function(select = NULL, minimize = NA) {
      private$.select = assert_string(select, null.ok = TRUE)
      super$initialize(
        id = select %??% "internal_valid_score",
        task_type = NA_character_,
        properties = c("na_score", "requires_learner", "requires_no_prediction"),
        predict_sets = NULL,
        predict_type = NA_character_,
        range = c(-Inf, Inf),
        minimize = assert_flag(minimize, na.ok = TRUE),
        label = "Internal Validation Score",
        man = "mlr3::mlr_measures_internal_valid_score"
      )
    }
  ),

  private = list(
    .select = NULL,
    .score = function(prediction, learner, ...) {
      x = get0("internal_valid_scores", learner)
      x[[private$.select %??% 1]] %??% NA_real_
    }
  )
)

#' @include mlr_measures.R
mlr_measures$add("internal_valid_score", MeasureInternalValidScore)

#' @title Regression Learner
#'
#' @include Learner.R
#'
#' @description
#' This Learner specializes [Learner] for regression problems:
#'
#' * `task_type` is set to `"regr"`.
#' * Creates [Prediction]s of class [PredictionRegr].
#' * Possible values for `predict_types` are:
#'   - `"response"`: Predicts a numeric response for each observation in the test set.
#'   - `"se"`: Predicts the standard error for each value of response for each observation in the test set.
#'   - `"distr"`: Probability distribution as `VectorDistribution` object (requires package `distr6`, available via
#'     repository \url{https://raphaels1.r-universe.dev}).
#'  - `"quantiles"`: Predicts quantile estimates for each observation in the test set.
#'    Set `$quantiles` to specify the quantiles to predict and `$quantile_response` to specify the response quantile.
#'    See the [mlr3book](https://mlr3book.mlr-org.com/chapters/chapter13/beyond_regression_and_classification.html)
#'    on quantile regression for more details.
#'
#' Predefined learners can be found in the [dictionary][mlr3misc::Dictionary] [mlr_learners].
#' Essential regression learners can be found in this dictionary after loading \CRANpkg{mlr3learners}.
#' Additional learners are implemented in the Github package \url{https://github.com/mlr-org/mlr3extralearners}.
#'
#' @template param_id
#' @template param_param_set
#' @template param_predict_types
#' @template param_feature_types
#' @template param_learner_properties
#' @template param_packages
#' @template param_label
#' @template param_man
#' @template param_task_type
#'
#' @template seealso_learner
#' @export
#' @examples
#' # get all regression learners from mlr_learners:
#' lrns = mlr_learners$mget(mlr_learners$keys("^regr"))
#' names(lrns)
#'
#' # get a specific learner from mlr_learners:
#' mlr_learners$get("regr.rpart")
#' lrn("regr.featureless")
LearnerRegr = R6Class(
  "LearnerRegr",
  inherit = Learner,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function(
      id,
      task_type = "regr",
      param_set = ps(),
      predict_types = "response",
      feature_types = character(),
      properties = character(),
      packages = character(),
      label = NA_character_,
      man = NA_character_
    ) {
      super$initialize(
        id = id,
        task_type = task_type,
        param_set = param_set,
        feature_types = feature_types,
        predict_types = predict_types,
        properties = properties,
        packages = packages,
        label = label,
        man = man
      )
    },

    #' @description
    #' Predicts outcomes for new data in `newdata` using the model fitted during `$train()`.
    #' This method is faster than `$predict_newdata()` as it skips assertions,
    #' type conversions, encapsulation, and logging.
    #' The learner must be trained before calling this method.
    #'
    #' Unlike `$predict_newdata()`, this method does not return a [Prediction] object.
    #' Instead, it returns a list with either a `"response"`, `"se"`, or `"quantiles"` element,
    #' depending on the prediction type.
    #'
    #' Note that `state$predict_time` and `state$log` will remain empty after using this method.
    #' Some learners may not support this method and may fail when it is called.
    #' Prefer `$predict_newdata()` unless performance is critical.
    #'
    #' If the model was trained via [resample()] or [benchmark()],
    #' you must pass the associated task object stored in the corresponding [ResampleResult] or [BenchmarkResult].
    #'
    #' @param newdata [`data.table::data.table()`]\cr
    #'   New data to predict on.
    #' @param task ([Task]).
    #'
    #' @return `list()` with elements `"response"`, `"se"` or `"quantiles"` depending on the predict type.
    predict_newdata_fast = function(newdata, task = NULL) {
      if (is.null(task) && is.null(self$state$train_task)) {
        error_input("No task stored, and no task provided")
      }
      feature_names = self$state$train_task$feature_names %??% task$feature_names

      # add data and most common used meta data
      fake_task = list(
        # some learners require that newdata contains the features names in the same order as the training data
        data = function(...) newdata[, feature_names, with = FALSE],
        feature_names = feature_names,
        nrow = nrow(newdata)
      )

      # train failed, use fallback
      if (is.null(self$model) && !is.null(self$state$fallback_state$model)) {
        return(self$fallback$predict_newdata_fast(newdata))
      }
      pred = get_private(self)$.predict(fake_task)

      # predict missing predictions with fallback
      miss = logical(fake_task$nrow)
      if (!is.null(pred$response)) {
        miss = is.na(pred$response)
      }

      if (!is.null(pred$se)) {
        miss = miss | is.na(pred$se)
      }

      if (!is.null(pred$quantiles)) {
        miss = miss | row_any_na(pred$quantiles)
      }

      miss_ids = which(miss)
      if (length(miss_ids) && !is.null(self$state$fallback_state$model)) {
        pred_miss = self$fallback$predict_newdata_fast(newdata[miss_ids, ])

        if (!is.null(pred$response)) {
          pred$response[miss_ids] = pred_miss$response
        }

        if (!is.null(pred$se)) {
          pred$se[miss_ids] = pred_miss$se
        }

        if (!is.null(pred$quantiles)) {
          pred$quantiles[miss_ids, ] = pred_miss$quantiles
        }
      }

      return(pred)
    }
  ),

  active = list(
    #' @field quantiles (`numeric()`)\cr
    #' Numeric vector of probabilities to be used while predicting quantiles.
    #' Elements must be between 0 and 1, not missing and provided in ascending order.
    #' If only one quantile is provided, it is used as response.
    #' Otherwise, set `$quantile_response` to specify the response quantile.
    #' Set to `NULL` to reset both `$quantiles` and `$quantile_response`.
    quantiles = function(rhs) {
      if (missing(rhs)) {
        return(private$.quantiles)
      }

      if ("quantiles" %nin% self$predict_types) {
        error_config("Learner does not support predicting quantiles")
      }
      if (is.null(rhs)) {
        private$.quantiles = NULL
        private$.quantile_response = NULL
        return(invisible(NULL))
      }
      private$.quantiles = assert_numeric(
        rhs,
        lower = 0,
        upper = 1,
        any.missing = FALSE,
        min.len = 1L,
        sorted = TRUE,
        .var.name = "quantiles"
      )

      if (length(private$.quantiles) == 1L) {
        private$.quantile_response = private$.quantiles
      }
    },

    #' @field quantile_response (`numeric(1)`)\cr
    #' The quantile to be used as response.
    quantile_response = function(rhs) {
      if (missing(rhs)) {
        return(private$.quantile_response)
      }

      if ("quantiles" %nin% self$predict_types) {
        error_config("Learner does not support predicting quantiles")
      }

      private$.quantile_response = assert_number(rhs, lower = 0, upper = 1, .var.name = "response")
      private$.quantiles = sort(union(private$.quantiles, private$.quantile_response))
    }
  ),

  private = list(
    .quantiles = NULL,
    .quantile_response = NULL
  )
)

#' @title Classification Learner for Debugging
#'
#' @name mlr_learners_classif.debug
#' @include LearnerClassif.R
#'
#' @description
#' A simple [LearnerClassif] used primarily in the unit tests and for debugging purposes.
#' If no hyperparameter is set, it simply constantly predicts a randomly selected label.
#' The following hyperparameters trigger the following actions:
#' \describe{
#'    \item{error_predict:}{Probability to raise an exception during predict.}
#'    \item{error_train:}{Probability to raises an exception during train.}
#'    \item{message_predict:}{Probability to output a message during predict.}
#'    \item{message_train:}{Probability to output a message during train.}
#'    \item{predict_missing:}{Ratio of predictions which will be NA.}
#'    \item{predict_missing_type:}{How to encode missingness.
#'      \dQuote{na} will insert NA values,
#'      \dQuote{omit} will just return fewer predictions than requested.}
#'    \item{save_tasks:}{Saves input task in `model` slot during training and prediction.}
#'    \item{segfault_predict:}{Probability to provokes a segfault during predict.}
#'    \item{segfault_train:}{Probability to provokes a segfault during train.}
#'    \item{sleep_train:}{Function returning a single number determining how many seconds
#'      to sleep during `$train()`.}
#'    \item{sleep_predict:}{Function returning a single number determining how many seconds
#'      to sleep during `$predict()`.}
#'    \item{threads:}{Number of threads to use. Has no effect.}
#'    \item{warning_predict:}{Probability to signal a warning during predict.}
#'    \item{warning_train:}{Probability to signal a warning during train.}
#'    \item{x:}{Numeric tuning parameter. Has no effect.}
#'    \item{iter:}{Integer parameter for testing hotstarting.}
#'    \item{count_marshaling:}{If `TRUE`, `marshal_model` will increase the `marshal_count`
#'      by 1 each time it is called. The default is `FALSE`.}
#'    \item{check_pid:}{If `TRUE`, the `$predict()` function will throw an error if the model
#'      was not unmarshaled in the same session that is used for prediction.)}
#' }
#' Note that segfaults may not be triggered reliably on your operating system.
#' Also note that if they work as intended, they will tear down your R session immediately!
#'
#' @templateVar id classif.debug
#' @template learner
#'
#' @template seealso_learner
#' @export
#' @examples
#' learner = lrn("classif.debug")
#' learner$param_set$set_values(message_train = 1, save_tasks = TRUE)
#'
#' # this should signal a message
#' task = tsk("penguins")
#' learner$train(task)
#' learner$predict(task)
#'
#' # task_train and task_predict are the input tasks for train() and predict()
#' names(learner$model)
LearnerClassifDebug = R6Class(
  "LearnerClassifDebug",
  inherit = LearnerClassif,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      iter_aggr = crate(function(x) as.integer(ceiling(mean(unlist(x, use.names = FALSE)))), .parent = topenv())
      iter_tune_fn = crate(
        function(domain, param_vals) {
          assert_true(isTRUE(param_vals$early_stopping))
          assert_true(domain$lower <= 1)
          domain$upper
        },
        .parent = topenv()
      )

      p_iter = p_int(
        1,
        default = 1,
        tags = c("train", "hotstart", "internal_tuning"),
        aggr = iter_aggr,
        in_tune_fn = iter_tune_fn,
        disable_in_tune = list(early_stopping = FALSE)
      )

      # fmt: skip
      param_set = ps(
        error_predict        = p_dbl(0, 1, default = 0, tags = "predict"),
        error_train          = p_dbl(0, 1, default = 0, tags = "train"),
        message_predict      = p_dbl(0, 1, default = 0, tags = "predict"),
        message_train        = p_dbl(0, 1, default = 0, tags = "train"),
        predict_missing      = p_dbl(0, 1, default = 0, tags = "predict"),
        predict_missing_type = p_fct(c("na", "omit"), default = "na", tags = "predict"),
        save_tasks           = p_lgl(default = FALSE, tags = c("train", "predict")),
        segfault_predict     = p_dbl(0, 1, default = 0, tags = "predict"),
        segfault_train       = p_dbl(0, 1, default = 0, tags = "train"),
        sleep_train          = p_uty(tags = "train"),
        sleep_predict        = p_uty(tags = "predict"),
        threads              = p_int(1L, tags = c("train", "threads")),
        warning_predict      = p_dbl(0, 1, default = 0, tags = "predict"),
        warning_train        = p_dbl(0, 1, default = 0, tags = "train"),
        x                    = p_dbl(0, 1, tags = "train"),
        iter                 = p_iter,
        early_stopping       = p_lgl(default = FALSE, tags = "train"),
        count_marshaling     = p_lgl(default = FALSE, tags = "train"),
        check_pid            = p_lgl(default = TRUE, tags = "train"),
        config_error         = p_lgl(default = FALSE, tags = "train")
      )
      super$initialize(
        id = "classif.debug",
        param_set = param_set,
        feature_types = c("logical", "integer", "numeric", "character", "factor", "ordered"),
        predict_types = c("response", "prob"),
        properties = c(
          "twoclass",
          "multiclass",
          "missings",
          "hotstart_forward",
          "validation",
          "internal_tuning",
          "marshal",
          "weights"
        ),
        man = "mlr3::mlr_learners_classif.debug",
        label = "Debug Learner for Classification"
      )
    },
    #' @description
    #' Marshal the learner's model.
    #' @param ... (any)\cr
    #'   Additional arguments passed to [`marshal_model()`].
    marshal = function(...) {
      learner_marshal(.learner = self, ...)
    },
    #' @description
    #' Unmarshal the learner's model.
    #' @param ... (any)\cr
    #'   Additional arguments passed to [`unmarshal_model()`].
    unmarshal = function(...) {
      learner_unmarshal(.learner = self, ...)
    },

    #' @description
    #' Returns 0 for each feature seen in training.
    #' @return Named `numeric()`.
    importance = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      fns = self$state$feature_names
      set_names(rep(0, length(fns)), fns)
    },

    #' @description
    #' Always returns character(0).
    #' @return `character()`.
    selected_features = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      character(0)
    }
  ),
  active = list(
    #' @field marshaled (`logical(1)`)\cr
    #' Whether the learner has been marshaled.
    marshaled = function() {
      learner_marshaled(self)
    },
    #' @field internal_valid_scores
    #' Retrieves the internal validation scores as a named `list()`.
    #' Returns `NULL` if learner is not trained yet.
    internal_valid_scores = function() {
      self$state$internal_valid_scores
    },
    #' @field internal_tuned_values
    #' Retrieves the internally tuned values as a named `list()`.
    #' Returns `NULL` if learner is not trained yet.
    internal_tuned_values = function() {
      self$state$internal_tuned_values
    },

    #' @field validate
    #' How to construct the internal validation data. This parameter can be either `NULL`,
    #' a ratio in $(0, 1)$, `"test"`, or `"predefined"`.
    validate = function(rhs) {
      if (!missing(rhs)) {
        private$.validate = assert_validate(rhs)
      }
      private$.validate
    }
  ),
  private = list(
    .validate = NULL,
    .train = function(task) {
      pv = self$param_set$get_values(tags = "train")
      if (isTRUE(pv$config_error)) {
        error_config("You misconfigured the learner")
      }

      pv$count_marshaling = pv$count_marshaling %??% FALSE
      roll = function(name) {
        name %chin% names(pv) && pv[[name]] > runif(1L)
      }

      if (!is.null(pv$sleep_train)) {
        secs = assert_number(pv$sleep_train())
        Sys.sleep(max(0, secs))
      }

      if (roll("message_train")) {
        message("Message from classif.debug->train()")
      }
      if (roll("warning_train")) {
        warning_mlr3("Warning from classif.debug->train()")
      }
      if (roll("error_train")) {
        error_learner_train("Error from classif.debug->train()")
      }
      if (roll("segfault_train")) {
        get("attach")(structure(list(), class = "UserDefinedDatabase"))
      }

      valid_truth = if (!is.null(task$internal_valid_task)) task$internal_valid_task$truth()

      if (isTRUE(pv$early_stopping) && is.null(valid_truth)) {
        error_config("Early stopping is only possible when a validation task is present.")
      }

      model = list(
        response = as.character(sample(task$truth(), 1L, prob = private$.get_weights(task))),
        pid = Sys.getpid(),
        id = UUIDgenerate(),
        random_number = sample(100000, 1),
        iter = if (isTRUE(pv$early_stopping)) sample(pv$iter %??% 1L, 1L) else pv$iter %??% 1L
      )

      if (!is.null(valid_truth)) {
        valid_pred = private$.make_prediction(
          task$internal_valid_task,
          model,
          self$param_set$get_values(tags = "predict")
        )

        valid_pred = as_prediction(as_prediction_data(
          valid_pred,
          task = task$internal_valid_task,
          check = TRUE,
          train_task = task
        ))

        model$internal_valid_scores = list(acc = mlr3measures::acc(valid_truth, valid_pred$response))
        if (self$predict_type == "prob") {
          model$internal_valid_scores$mbrier = mlr3measures::mbrier(valid_truth, valid_pred$prob)
        }
      }

      if (isTRUE(pv$save_tasks)) {
        model$task_train = task$clone(deep = TRUE)
      }

      if (pv$check_pid %??% FALSE) {
        model$marshal_pid = Sys.getpid()
      }

      if (isTRUE(pv$count_marshaling)) {
        model$marshal_count = 0L
      }

      set_class(model, "classif.debug_model")
    },

    .extract_internal_tuned_values = function() {
      if (!isTRUE(self$state$param_vals$early_stopping)) {
        named_list()
      } else {
        self$model["iter"]
      }
    },
    .extract_internal_valid_scores = function() {
      if (is.null(self$model$internal_valid_scores)) {
        named_list()
      } else {
        self$model$internal_valid_scores
      }
    },
    .predict = function(task) {
      if (!is.null(self$model$marshal_pid) && self$model$marshal_pid != Sys.getpid()) {
        error_mlr3("Model was not unmarshaled correctly")
      }
      n = task$nrow
      pv = self$param_set$get_values(tags = "predict")
      roll = function(name) {
        name %chin% names(pv) && pv[[name]] > runif(1L)
      }

      if (!is.null(pv$sleep_predict)) {
        secs = assert_number(pv$sleep_predict())
        Sys.sleep(max(0, secs))
      }

      if (roll("message_predict")) {
        message("Message from classif.debug->predict()")
      }
      if (roll("warning_predict")) {
        warning_mlr3("Warning from classif.debug->predict()")
      }
      if (roll("error_predict")) {
        error_learner_predict("Error from classif.debug->predict()")
      }
      if (roll("segfault_predict")) {
        get("attach")(structure(list(), class = "UserDefinedDatabase"))
      }

      if (isTRUE(pv$save_tasks)) {
        self$state$model$task_predict = task$clone(deep = TRUE)
      }

      private$.make_prediction(task, self$model, pv)
    },

    .make_prediction = function(task, model, pv) {
      n = task$nrow
      response = prob = NULL
      missing_type = pv$predict_missing_type %??% "na"

      if ("response" %chin% self$predict_type) {
        response = rep.int(unclass(model$response), n)
        if (!is.null(pv$predict_missing)) {
          ii = sample.int(n, n * pv$predict_missing)
          response = switch(missing_type, "na" = replace(response, ii, NA), "omit" = response[ii])
        }
      }

      if ("prob" %chin% self$predict_type) {
        cl = task$class_names
        prob = matrix(runif(n * length(cl)), nrow = n)
        prob = prob / rowSums(prob)
        colnames(prob) = cl

        if (!is.null(pv$predict_missing)) {
          ii = sample.int(n, n * pv$predict_missing)
          prob = switch(
            missing_type,
            "na" = {
              prob[ii, ] = NA_real_
              prob
            },
            "omit" = {
              prob[ii, , drop = FALSE]
            }
          )
        }
      }

      list(response = response, prob = prob)
    },

    .hotstart = function(task) {
      model = self$model
      pars = self$param_set$get_values(tags = "train")
      id = self$model$id

      model = list(response = as.character(sample(task$truth(), 1L)), pid = Sys.getpid(), iter = pars$iter, id = id)
      set_class(model, "classif.debug_model")
    }
  )
)

#' @include mlr_learners.R
mlr_learners$add("classif.debug", function() LearnerClassifDebug$new())

#' @export
#' @method marshal_model classif.debug_model
marshal_model.classif.debug_model = function(model, inplace = FALSE, ...) {
  if (!is.null(model$marshal_count)) {
    model$marshal_count = model$marshal_count + 1L
  }
  structure(
    list(
      marshaled = model,
      packages = "mlr3"
    ),
    class = c("classif.debug_model_marshaled", "marshaled")
  )
}

#' @export
#' @method unmarshal_model classif.debug_model_marshaled
unmarshal_model.classif.debug_model_marshaled = function(model, inplace = FALSE, ...) {
  unmarshaled = model$marshaled
  if (!is.null(unmarshaled$marshal_pid)) {
    unmarshaled$marshal_pid = Sys.getpid()
  }
  unmarshaled
}

#' @title DataBackend for data.table
#'
#' @description
#' [DataBackend] for \CRANpkg{data.table} which serves as an efficient in-memory data base.
#'
#' @template param_rows
#' @template param_cols
#' @template param_primary_key
#' @template param_na_rm
#'
#' @template seealso_databackend
#' @export
#' @examples
#' data = as.data.table(palmerpenguins::penguins)
#' data$id = seq_len(nrow(palmerpenguins::penguins))
#' b = DataBackendDataTable$new(data = data, primary_key = "id")
#' print(b)
#' b$head()
#' b$data(rows = 100:101, cols = "species")
#'
#' b$nrow
#' head(b$rownames)
#'
#' b$ncol
#' b$colnames
#'
#' # alternative construction
#' as_data_backend(palmerpenguins::penguins)
DataBackendDataTable = R6Class(
  "DataBackendDataTable",
  inherit = DataBackend,
  cloneable = FALSE,
  public = list(
    #' @field compact_seq `logical(1)`\cr
    #' If `TRUE`, row ids are a natural sequence from 1 to `nrow(data)` (determined internally).
    #' In this case, row lookup uses faster positional indices instead of equi joins.
    compact_seq = FALSE,

    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' Note that `DataBackendDataTable` does not copy the input data,
    #' while `as_data_backend()` calls [data.table::copy()].
    #' `as_data_backend()` also takes care about casting to a `data.table()` and adds a primary key column if necessary.
    #'
    #' @param data ([data.table::data.table()])\cr
    #'   The input [data.table()].
    initialize = function(data, primary_key) {
      assert_data_table(data, col.names = "unique")
      super$initialize(setkeyv(data, primary_key), primary_key)
      ii = match(primary_key, names(data))
      if (is.na(ii)) {
        error_input("Primary key '%s' not in 'data'", primary_key)
      }
      private$.cache = set_names(replace(rep(NA, ncol(data)), ii, FALSE), names(data))
    },

    #' @description
    #' Returns a slice of the data.
    #' The rows must be addressed as vector of primary key values, columns must be referred to via column names.
    #' Queries for rows with no matching row id and queries for columns with no matching column name are
    #' silently ignored.
    #' Rows are guaranteed to be returned in the same order as `rows`, columns may be returned in an arbitrary order.
    #' Duplicated row ids result in duplicated rows, duplicated column names lead to an exception.
    data = function(rows, cols) {
      rows = assert_integerish(rows, coerce = TRUE)
      assert_names(cols, type = "unique")
      cols = intersect(cols, colnames(private$.data))

      if (self$compact_seq) {
        private$.data[rows, cols, with = FALSE, nomatch = NULL]
      } else {
        ijoin(private$.data, rows, cols, self$primary_key)
      }
    },

    #' @description
    #' Retrieve the first `n` rows.
    #'
    #' @param n (`integer(1)`)\cr
    #'   Number of rows.
    #'
    #' @return [data.table::data.table()] of the first `n` rows.
    head = function(n = 6L) {
      head(private$.data, n)
    },

    #' @description
    #' Returns a named list of vectors of distinct values for each column
    #' specified. If `na_rm` is `TRUE`, missing values are removed from the
    #' returned vectors of distinct values. Non-existing rows and columns are
    #' silently ignored.
    #'
    #' @return Named `list()` of distinct values.
    distinct = function(rows, cols, na_rm = TRUE) {
      cols = intersect(cols, colnames(private$.data))
      if (is.null(rows)) {
        set_names(lapply(cols, function(x) distinct_values(private$.data[[x]], drop = FALSE, na_rm = na_rm)), cols)
      } else {
        lapply(self$data(rows, cols), distinct_values, drop = TRUE, na_rm = na_rm)
      }
    },

    #' @description
    #' Returns the number of missing values per column in the specified slice
    #' of data. Non-existing rows and columns are silently ignored.
    #'
    #' @return Total of missing values per column (named `numeric()`).
    missings = function(rows, cols) {
      missind = private$.cache
      missind = missind[reorder_vector(names(missind), cols)]

      # update cache
      ii = which(is.na(missind))
      if (length(ii)) {
        missind[ii] = map_lgl(private$.data[, names(missind[ii]), with = FALSE], anyMissing)
        private$.cache = insert_named(private$.cache, missind[ii])
      }

      # query required columns
      query_cols = which(missind)
      insert_named(
        named_vector(names(missind), 0L),
        map_int(self$data(rows, names(query_cols)), count_missing)
      )
    }
  ),

  active = list(
    #' @field rownames (`integer()`)\cr
    #' Returns vector of all distinct row identifiers, i.e. the contents of the primary key column.
    rownames = function() {
      private$.data[[self$primary_key]]
    },

    #' @field colnames (`character()`)\cr
    #' Returns vector of all column names, including the primary key column.
    colnames = function() {
      colnames(private$.data)
    },

    #' @field nrow (`integer(1)`)\cr
    #' Number of rows (observations).
    nrow = function() {
      nrow(private$.data)
    },

    #' @field ncol (`integer(1)`)\cr
    #' Number of columns (variables), including the primary key column.
    ncol = function() {
      ncol(private$.data)
    }
  ),

  private = list(
    .calculate_hash = function() {
      calculate_hash(self$compact_seq, private$.data)
    },

    .cache = NULL
  )
)

#' @title Elapsed Time Measure
#'
#' @name mlr_measures_elapsed_time
#' @aliases mlr_measures_time_train mlr_measures_time_predict mlr_measures_time_both
#' @include Measure.R
#'
#' @description
#' Measures the elapsed time during train ("time_train"), predict ("time_predict"), or both ("time_both").
#' Aggregation of elapsed time defaults to mean but can be configured via the field `aggregator` of the [Measure].
#'
#' When predictions for multiple predict sets were made during [resample()] or [benchmark()],
#' the predict time shows the cumulative duration of all predictions.
#' If `learner$predict()` is called manually, the last predict time gets overwritten.
#' The elapsed time accounts only for the training duration of the primary learner,
#' excluding the time required for training the fallback learner.
#'
#' @template param_id
#' @templateVar id time_train
#' @template measure
#'
#' @template seealso_measure
#' @export
MeasureElapsedTime = R6Class(
  "MeasureElapsedTime",
  inherit = Measure,
  public = list(
    #' @field stages (`character()`)\cr
    #' Which stages of the learner to measure?
    #' Usually set during construction.
    stages = NULL,

    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param stages (`character()`)\cr
    #'   Subset of `("train", "predict")`.
    #'   The runtime of provided stages will be summed.
    initialize = function(id = "elapsed_time", stages) {
      super$initialize(
        id = id,
        task_type = NA_character_,
        predict_sets = NULL,
        predict_type = NA_character_,
        range = c(0, Inf),
        minimize = TRUE,
        properties = c("requires_learner", "requires_no_prediction"),
        label = "Elapsed Time",
        man = "mlr3::mlr_measures_elapsed_time"
      )
      self$stages = assert_subset(stages, c("train", "predict"), empty.ok = FALSE)
    }
  ),

  private = list(
    .score = function(prediction, learner, ...) {
      sum(unlist(learner$state[sprintf("%s_time", self$stages)], use.names = FALSE))
    },

    .extra_hash = "stages"
  )
)

#' @include mlr_measures.R
mlr_measures$add("time_train", function() MeasureElapsedTime$new(id = "time_train", stages = "train"))
mlr_measures$add("time_predict", function() MeasureElapsedTime$new(id = "time_predict", stages = "predict"))
mlr_measures$add("time_both", function() MeasureElapsedTime$new(id = "time_both", stages = c("train", "predict")))

#' @title XOR Classification Task Generator
#'
#' @name mlr_task_generators_xor
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] for the xor task in [mlbench::mlbench.xor()].
#'
#' @templateVar id xor
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("xor")
#' plot(generator, n = 200)
#'
#' task = generator$generate(200)
#' str(task$data())
TaskGeneratorXor = R6Class(
  "TaskGeneratorXor",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        d = p_int(1L, default = 1L)
      )

      super$initialize(
        id = "xor",
        "classif",
        "mlbench",
        ps,
        label = "XOR Classification",
        man = "mlr3::mlr_task_generators_xor"
      )
    },

    #' @description
    #' Creates a simple plot of generated data.
    #' @param n (`integer(1)`)\cr
    #'   Number of samples to draw for the plot. Default is 200.
    #' @param pch (`integer(1)`)\cr
    #'   Point char. Passed to [plot()].
    #' @param ... (any)\cr
    #'   Additional arguments passed to [plot()].
    plot = function(n = 200L, pch = 19L, ...) {
      plot(private$.generate_obj(n), pch = pch, ...)
    }
  ),

  private = list(
    .generate_obj = function(n) {
      invoke(mlbench::mlbench.xor, n = n, .args = self$param_set$values, .opts = allow_partial_matching)
    },

    .generate = function(n) {
      obj = private$.generate_obj(n)
      TaskClassif$new(sprintf("%s_%i", self$id, n), convert_mlbench(obj), target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("xor", function() TaskGeneratorXor$new())

#' @title Spam Classification Task
#'
#' @name mlr_tasks_spam
#' @format [R6::R6Class] inheriting from [TaskClassif].
#' @include mlr_tasks.R
#'
#' @description
#' Spam data set from the UCI machine learning repository (\url{http://archive.ics.uci.edu/dataset/94/spambase}).
#' Data set collected at Hewlett-Packard Labs to classify emails as spam or non-spam.
#' 57 variables indicate the frequency of certain words and characters in the e-mail.
#' The positive class is set to "spam".
#'
#' @templateVar id spam
#' @template task
#'
#' @source
#' Creators:
#' Mark Hopkins, Erik Reeber, George Forman, Jaap Suermondt.
#' Hewlett-Packard Labs, 1501 Page Mill Rd., Palo Alto, CA 94304
#'
#' Donor:
#' George Forman (gforman at nospam hpl.hp.com) 650-857-7835
#'
#' Preprocessing:
#' Columns have been renamed. Preprocessed data taken from the \CRANpkg{kernlab} package.
#'
#' @references
#' `r format_bib("dua_2017")`
#'
#' @template seealso_task
NULL

load_task_spam = function(id = "spam") {
  b = as_data_backend(readRDS(system.file("extdata", "spam.rds", package = "mlr3")))
  task = TaskClassif$new(id, b, target = "type", positive = "spam", label = "HP Spam Detection")
  b$hash = task$man = "mlr3::mlr_tasks_spam"
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("spam", load_task_spam)

#' @title Smiley Classification Task Generator
#'
#' @name mlr_task_generators_smiley
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] for the smiley task in [mlbench::mlbench.smiley()].
#'
#' @templateVar id smiley
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("smiley")
#' plot(generator, n = 200)
#'
#' task = generator$generate(200)
#' str(task$data())
TaskGeneratorSmiley = R6Class(
  "TaskGeneratorSmiley",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        sd1 = p_dbl(0L),
        sd2 = p_dbl(0L)
      )

      super$initialize(
        id = "smiley",
        "classif",
        "mlbench",
        ps,
        label = "Smiley Classification",
        man = "mlr3::mlr_task_generators_smiley"
      )
    },

    #' @description
    #' Creates a simple plot of generated data.
    #' @param n (`integer(1)`)\cr
    #'   Number of samples to draw for the plot. Default is 200.
    #' @param pch (`integer(1)`)\cr
    #'   Point char. Passed to [plot()].
    #' @param ... (any)\cr
    #'   Additional arguments passed to [plot()].
    plot = function(n = 200L, pch = 19L, ...) {
      plot(private$.generate_obj(n), pch = pch, ...)
    }
  ),

  private = list(
    .generate_obj = function(n) {
      obj = invoke(mlbench::mlbench.smiley, n = n, .args = self$param_set$values)
      colnames(obj$x) = sprintf("x.%i", seq_col(obj$x))
      obj
    },

    .generate = function(n) {
      obj = private$.generate_obj(n)
      TaskClassif$new(sprintf("%s_%i", self$id, n), convert_mlbench(obj), target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("smiley", function() TaskGeneratorSmiley$new())

allow_partial_matching = list(
  warnPartialMatchArgs = FALSE,
  warnPartialMatchAttr = FALSE,
  warnPartialMatchDollar = FALSE
)

set_encapsulation = function(learners, encapsulate) {
  assert_choice(encapsulate, c(NA_character_, "none", "evaluate", "callr", "try"))

  if (!is.na(encapsulate)) {
    lapply(learners, function(learner) {
      fallback = if (encapsulate != "none") default_fallback(learner)
      if (is.null(fallback)) {
        error_input("Could not find default fallback learner for learner '%s'", learner$id)
      }
      learner$encapsulate(encapsulate, fallback)
    })
  }
  learners
}

future_map = function(n, FUN, ..., MoreArgs = list()) {
  if (getOption("mlr3.debug", FALSE)) {
    lg$info("Running experiments sequentially in debug mode with %i iterations", n)
    mapply(FUN, ..., MoreArgs = MoreArgs, SIMPLIFY = FALSE, USE.NAMES = FALSE)
  } else if (
    isNamespaceLoaded("mirai") &&
      mirai::daemons_set(.compute = getOption("mlr3.mirai_parallelization", "mlr3_parallelization"))
  ) {
    lg$debug("Running resample() via mirai with %i iterations", n)
    mirai::collect_mirai(mirai::mirai_map(
      data.table(...),
      FUN,
      .args = c(MoreArgs, list(is_sequential = FALSE)),
      .compute = getOption("mlr3.mirai_parallelization", "mlr3_parallelization")
    ))
  } else {
    is_sequential = inherits(plan(), "sequential")
    scheduling = if (!is_sequential && isTRUE(getOption("mlr3.exec_random", TRUE))) {
      structure(TRUE, ordering = "random")
    } else {
      TRUE
    }
    chunk_size = getOption("mlr3.exec_chunk_size", 1)
    chunk_bins = getOption("mlr3.exec_chunk_bins")
    if (!is.null(chunk_bins)) {
      chunk_size = ceiling(n / chunk_bins)
    }
    stdout = if (is_sequential) NA else TRUE

    # workaround for sequential plan checking the size of the globals
    # see https://github.com/futureverse/future/issues/197
    if (is_sequential) {
      old_opts = options(future.globals.maxSize = Inf)
      on.exit(options(old_opts), add = TRUE)
    }

    MoreArgs = c(MoreArgs, list(is_sequential = is_sequential))

    lg$debug("Running resample() via future with %i iterations", n)
    future.apply::future_mapply(
      FUN,
      ...,
      MoreArgs = MoreArgs,
      SIMPLIFY = FALSE,
      USE.NAMES = FALSE,
      future.globals = FALSE,
      future.packages = mlr_reflections$loaded_packages,
      future.seed = TRUE,
      future.scheduling = scheduling,
      future.chunk.size = chunk_size,
      future.stdout = stdout
    )
  }
}

#' @title Abstract Prediction Object
#'
#' @description
#' This is the abstract base class for task objects like [PredictionClassif] or [PredictionRegr].
#'
#' Prediction objects store the following information:
#' 1. The row ids of the test set
#' 2. The corresponding true (observed) response.
#' 3. The corresponding predicted response.
#' 4. Additional predictions based on the class and `predict_type`.
#'    E.g., the class probabilities for classification or the estimated standard error for regression.
#'
#' Note that this object is usually constructed via a derived classes, e.g. [PredictionClassif] or [PredictionRegr].
#'
#' @template param_measures
#'
#' @section S3 Methods:
#' * `as.data.table(rr)`\cr
#'   [Prediction] -> [data.table::data.table()]\cr
#'   Converts the data to a [data.table::data.table()].
#'
#' * `c(..., keep_duplicates = TRUE)`\cr
#'   ([Prediction], [Prediction], ...) -> [Prediction]\cr
#'   Combines multiple `Prediction`s to a single `Prediction`.
#'   If `keep_duplicates` is `FALSE` and there are duplicated row ids,
#'   the data of the former passed objects get overwritten by the data of the later passed objects.
#'
#' @template seealso_prediction
#' @export
Prediction = R6Class(
  "Prediction",
  public = list(
    #' @field data (named `list()`)\cr
    #' Internal data structure.
    data = list(),

    #' @field task_type (`character(1)`)\cr
    #' Required type of the [Task].
    task_type = NULL,

    #' @field task_properties (`character()`)\cr
    #' Required properties of the [Task].
    task_properties = NULL,

    #' @field predict_types (`character()`)\cr
    #' Set of predict types this object stores.
    predict_types = character(),

    #' @template field_man
    man = "mlr3::Prediction",

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s>", class(self)[1L])
    },

    #' @description
    #' Printer.
    #' @param ... (ignored).
    print = function(...) {
      n = length(self$data$row_ids)
      if (n == 0L) {
        cat_cli(cli_h1("{.cls {class(self)[1L]}} for {.val 0} observations"))
      } else {
        data = as.data.table(self)
        cat_cli(cli_h1("{.cls {class(self)[1L]}} for {.val {n}} observations:"))
        print(data, nrows = 10L, topn = 3L, class = FALSE, row.names = FALSE, print.keys = FALSE)
      }
    },

    #' @description
    #' Opens the corresponding help page referenced by field `$man`.
    help = function() {
      open_help(self$man)
    },

    #' @description
    #' Calculates the performance for all provided measures
    #' [Task] and [Learner] may be `NULL` for most measures, but some measures need to extract information
    #' from these objects.
    #' Note that the `predict_sets` of the `measures` are ignored by this method,
    #' instead all predictions are used.
    #'
    #' @param task ([Task]).
    #'
    #' @param learner ([Learner]).
    #'
    #' @param train_set (`integer()`).
    #'
    #' @return [Prediction].
    score = function(measures = NULL, task = NULL, learner = NULL, train_set = NULL) {
      measures = assert_measures(as_measures(measures, task_type = self$task_type))
      scores = map_dbl(measures, function(m) {
        m$score(prediction = self, task = task, learner = learner, train_set = train_set)
      })
      set_names(scores, ids(measures))
    },

    #' @description
    #' Calculates the observation-wise loss via the [Measure]'s `obs_loss` method.
    #' Returns a `data.table()` with the columns of the matching [Prediction] object plus one additional numeric column
    #' for each measure, named with the respective measure id.
    #' If there is no observation-wise loss function for the measure, the column is filled with `NA_real_` values.
    #' Note that some measures such as RMSE, do have an `$obs_loss`,
    #' but they require an additional transformation after aggregation,
    #' in this example taking the square-root.
    obs_loss = function(measures = NULL) {
      measures = assert_measures(as_measures(measures, task_type = self$task_type))
      tab = as.data.table(self)
      walk(measures, function(m) set(tab, j = m$id, value = m$obs_loss(prediction = self)))
      tab[]
    },

    #' @description
    #' Filters the [Prediction], keeping only predictions for the provided row_ids.
    #' This changes the object in-place, you need to create a clone to preserve
    #' the original [Prediction].
    #'
    #' @template param_row_ids
    #' @return `self`, modified.
    filter = function(row_ids) {
      row_ids = assert_row_ids(row_ids)
      self$data = filter_prediction_data(self$data, row_ids)
      invisible(self)
    }
  ),

  active = list(
    #' @field row_ids (`integer()`)\cr
    #'   Vector of row ids for which predictions are stored.
    row_ids = function(rhs) {
      assert_ro_binding(rhs)
      self$data$row_ids
    },

    #' @field truth (any)\cr
    #'   True (observed) outcome.
    truth = function(rhs) {
      assert_ro_binding(rhs)
      self$data$truth
    },

    #' @field missing (`integer()`)\cr
    #'   Returns `row_ids` for which the predictions are missing or incomplete.
    missing = function(rhs) {
      assert_ro_binding(rhs)
      is_missing_prediction_data(self$data)
    },

    #' @field weights (`numeric()`)\cr
    #'   Vector of measure weights, obtained from the `weights_measure` column of the [Task] if present.
    #'   This is `NULL` if no weights are present.
    weights = function(rhs) {
      assert_ro_binding(rhs)
      self$data$weights
    },

    #' @field extra (`list()`)\cr
    #'   List of extra data stored in the prediction object.
    extra = function(rhs) {
      assert_ro_binding(rhs)
      self$data$extra
    },

    #' @field raw (any)\cr
    #'   Raw prediction object from the upstream model.
    #'   Only available if the learner's `predict_raw` flag was set to `TRUE` and the learner implementation
    #'   supports it.
    #'   The object is stored as-is, without validation or subsetting during filtering.
    #'   When multiple predictions are combined via `c()`, the individual raw objects are collected into a `list()`.
    raw = function(rhs) {
      assert_ro_binding(rhs)
      self$data$raw
    }
  )
)

#' @export
# nolint next
c.Prediction = function(..., keep_duplicates = TRUE) {
  dots = list(...)
  if (length(dots) == 1L) {
    return(dots[[1L]])
  }

  classes = unique(map_chr(dots, function(x) class(x)[1L]))
  if (length(classes) > 1L) {
    error_input("Cannot combine objects of different type: %s", str_collapse(classes))
  }
  assert_flag(keep_duplicates)

  pdata = invoke(c, .args = c(map(dots, "data"), list(keep_duplicates = keep_duplicates)))
  as_prediction(pdata, check = FALSE)
}

#' @title ResultData
#'
#' @description
#' Internal object to store results in list of data.tables, arranged in a star schema.
#' It is advised to not directly work on this data structure as it may be changed in the future
#' without further warnings.
#'
#' The main motivation of this data structure is the necessity to avoid storing duplicated [R6][R6::R6Class] objects.
#' While this is usually no problem in a single R session, serialization via [serialize()] (which is
#' used in [save()]/[saveRDS()] or during parallelization) leads to objects with unreasonable memory
#' requirements.
#'
#' @template param_view
#' @keywords internal
#' @export
#' @examples
#' # table overview
#' print(ResultData$new()$data)
ResultData = R6Class(
  "ResultData",
  public = list(
    #' @field data (`list()`)\cr
    #' List of [data.table::data.table()], arranged in a star schema.
    #' Do not operate directly on this list.
    data = NULL,

    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #' An alternative construction method is provided by [as_result_data()].
    #'
    #' @param data ([data.table::data.table()]) | `NULL`)\cr
    #'   Do not initialize this object yourself, use [as_result_data()] instead.
    #' @param data_extra (`list()`)\cr
    #'   Additional data to store.
    #'   This can be used to store additional information for each iteration.
    #' @param store_backends (`logical(1)`)\cr
    #'   If set to `FALSE`, the backends of the [Task]s provided in `data` are removed.
    initialize = function(data = NULL, data_extra = NULL, store_backends = TRUE) {
      assert_flag(store_backends)

      if (is.null(data)) {
        self$data = star_init()
      } else {
        assert_names(
          names(data),
          permutation.of = c(
            "task",
            "learner",
            "learner_state",
            "resampling",
            "iteration",
            "param_values",
            "prediction",
            "uhash",
            "learner_hash"
          )
        )

        if (nrow(data) == 0L) {
          self$data = star_init()
        } else {
          setcolorder(
            data,
            c(
              "uhash",
              "iteration",
              "learner_state",
              "prediction",
              "task",
              "learner",
              "resampling",
              "param_values",
              "learner_hash"
            )
          )
          uhashes = data.table(uhash = unique(data$uhash))
          setkeyv(data, c("uhash", "iteration"))

          task = learner = resampling = NULL
          data[, "task_hash" := task[[1L]]$hash, by = "uhash"]
          data[, "learner_phash" := learner[[1L]]$phash, by = "uhash"]
          data[, "resampling_hash" := resampling[[1L]]$hash, by = "uhash"]

          tasks = data[, list(task = .SD$task[1L]), keyby = "task_hash"]
          learners = data[, list(learner = list(.SD$learner[[1L]]$reset())), keyby = "learner_phash"]
          resamplings = data[, list(resampling = .SD$resampling[1L]), keyby = "resampling_hash"]
          learner_components = data[, list(learner_param_vals = list(.SD$param_values[[1L]])), keyby = "learner_hash"]

          set(data, j = "task", value = NULL)
          set(data, j = "learner", value = NULL)
          set(data, j = "resampling", value = NULL)
          set(data, j = "param_values", value = NULL)

          data_extras = if (!is.null(data_extra)) {
            assert_list(data_extra, len = nrow(data))
            data.table(uhash = data$uhash, iteration = data$iteration, data_extra = data_extra)
          } else {
            data.table(uhash = character(), iteration = integer(), data_extra = list())
          }
          # remove null rows in data_extra
          data_extras = data_extras[!map_lgl(data_extra, is.null)]
          setkeyv(data_extras, c("uhash", "iteration"))

          if (!store_backends) {
            set(tasks, j = "task", value = lapply(tasks$task, task_rm_backend))
          }

          self$data = list(
            fact = data,
            uhashes = uhashes,
            tasks = tasks,
            learners = learners,
            resamplings = resamplings,
            learner_components = learner_components,
            data_extras = data_extras
          )
        }
      }
    },

    #' @description
    #' Returns all unique hashes (`uhash` values) of all included [ResampleResult]s.
    #'
    #' @return `character()`.
    uhashes = function(view = NULL) {
      if (is.null(view)) {
        self$data$uhashes$uhash
      } else {
        intersect(self$data$uhashes$uhash, view)
      }
    },

    #' @description
    #' Returns a [`data.table`] with columns `uhash`, `learner_id`, `task_id` and `resampling_id`
    #' for the given view.
    #' The `uhash` uniquely identifies an individual [`ResampleResult`].
    #'
    #' @return `data.table()`
    uhash_table = function(view = NULL) {
      newtbl = list()
      tbl = self$data$fact[
        private$get_view_index(view),
        c("uhash", "learner_phash", "task_hash", "resampling_hash"),
        with = FALSE
      ]
      tbl = tbl[!duplicated(get("uhash")), ]
      # retrieve learner id, task id and resampling id from self$data$learners, self$data$tasks
      # and self$data$resamplings
      newtbl$learner_id = self$data$learners[list(tbl$learner_phash), ids(get("learner")), on = "learner_phash"]
      newtbl$task_id = self$data$tasks[list(tbl$task_hash), ids(get("task")), on = "task_hash"]
      newtbl$resampling_id = self$data$resamplings[
        list(tbl$resampling_hash),
        ids(get("resampling")),
        on = "resampling_hash"
      ]
      newtbl$uhash = tbl$uhash
      setDT(newtbl)
      merge(self$data$uhashes, newtbl, by = "uhash", sort = FALSE)
    },

    #' @description
    #' Returns the number of recorded iterations / experiments.
    #'
    #' @return `integer(1)`.
    iterations = function(view = NULL) {
      if (is.null(view)) {
        nrow(self$data$fact)
      } else {
        length(private$get_view_index(view))
      }
    },

    #' @description
    #' Returns a table of included [Task]s.
    #'
    #' @return `data.table()` with columns `"task_hash"` (`character()`) and `"task"` ([Task]).
    tasks = function(view = NULL) {
      # nolint next
      .__ii__ = private$get_view_index(view)
      tab = unique(self$data$fact[.__ii__, "task_hash", with = FALSE], by = "task_hash")
      merge(tab, self$data$tasks, by = "task_hash", sort = TRUE)
    },

    #' @description
    #' Returns a table of included [Learner]s.
    #'
    #' @param states (`logical(1)`)\cr
    #'   If `TRUE`, returns a learner for each iteration/experiment in the [ResultData] object.
    #'   If `FALSE`, returns an exemplary learner (without state) for each [ResampleResult].
    #' @param reassemble (`logical(1)`)\cr
    #'   Reassemble the learners, i.e. re-set the `state` and the hyperparameters which are stored separately before
    #'   returning the learners.
    #'
    #' @return `data.table()` with columns `"learner_hash"` (`character()`) and `"learner"` ([Learner]).
    learners = function(view = NULL, states = TRUE, reassemble = TRUE) {
      # nolint next
      .__ii__ = private$get_view_index(view)

      if (states) {
        tab = self$data$fact[.__ii__, c("learner_hash", "learner_phash", "learner_state"), with = FALSE]
        tab = merge(tab, self$data$learners, by = "learner_phash", sort = FALSE)
        tab = merge(tab, self$data$learner_components, by = "learner_hash", sort = TRUE)
        set(
          tab,
          j = "learner",
          value = reassemble_learners(tab$learner, states = tab$learner_state, param_vals = tab$learner_param_vals)
        )
      } else {
        tab = unique(self$data$fact[.__ii__, c("learner_hash", "learner_phash"), with = FALSE], by = "learner_hash")
        tab = merge(tab, self$data$learners, by = "learner_phash", sort = FALSE)

        if (reassemble) {
          tab = merge(tab, self$data$learner_components, by = "learner_hash", sort = TRUE)
          set(tab, j = "learner", value = reassemble_learners(tab$learner, param_vals = tab$learner_param_vals))
        } else {
          setkeyv(tab, "learner_hash")
        }
      }

      tab[, c("learner_hash", "learner"), with = FALSE]
    },

    #' @description
    #' Returns a list of states of included [Learner]s without reassembling the learners.
    #'
    #'  @return list of `list()`
    learner_states = function(view = NULL) {
      # nolint next
      .__ii__ = private$get_view_index(view)
      learner_hash = learner_state = NULL
      self$data$fact[.__ii__][order(learner_hash), learner_state]
    },

    #' @description
    #' Returns a table of included [Resampling]s.
    #'
    #' @return `data.table()` with columns `"resampling_hash"` (`character()`) and `"resampling"` ([Resampling]).
    resamplings = function(view = NULL) {
      # nolint next
      .__ii__ = private$get_view_index(view)
      tab = unique(self$data$fact[.__ii__, "resampling_hash", with = FALSE], by = "resampling_hash")
      merge(tab, self$data$resamplings, by = "resampling_hash", sort = TRUE)
    },

    #' @description
    #' Returns a list of [Prediction] objects.
    #'
    #' @template param_predict_sets
    #'
    #' @return `list()` of [Prediction].
    predictions = function(view = NULL, predict_sets = "test") {
      # nolint next
      .__ii__ = private$get_view_index(view)
      as_predictions(self$data$fact[.__ii__, "prediction", with = FALSE][[1L]], predict_sets = predict_sets)
    },

    #' @description
    #' Returns a combined [Prediction] objects.
    #'
    #' @template param_predict_sets
    #'
    #' @return [Prediction].
    prediction = function(view = NULL, predict_sets = "test") {
      do.call(c, self$predictions(view = view, predict_sets = predict_sets))
    },

    #' @description
    #' Returns additional data stored.
    #'
    #' @return [data.table::data.table()].
    data_extra = function(view = NULL) {
      # nolint next
      .__ii__ = private$get_view_index(view)
      tab = self$data$fact[.__ii__, c("uhash", "iteration"), with = FALSE]
      merge(tab, self$data$data_extras, by = c("uhash", "iteration"), all.x = TRUE, sort = TRUE)
    },

    #' @description
    #' Combines multiple [ResultData] objects, modifying `self` in-place.
    #'
    #' @param rdata ([ResultData]).
    #'
    #' @return `self` (invisibly).
    combine = function(rdata) {
      assert_class(rdata, "ResultData")

      rbind_if_new = function(x, y, on = key(x)) {
        if (nrow(x) == 0L) {
          return(y[, names(x), with = FALSE])
        }

        new_rows = y[!x, on = on]
        if (nrow(new_rows)) {
          setkeyv(rbindlist(list(x, new_rows), use.names = TRUE), on)
        } else {
          x
        }
      }

      self$data$uhashes = funion(self$data$uhashes, rdata$data$uhashes, all = FALSE)
      for (nn in setdiff(names(self$data), "uhashes")) {
        self$data[[nn]] = rbind_if_new(self$data[[nn]], rdata$data[[nn]])
      }

      invisible(self)
    },

    #' @description
    #' Updates the [ResultData] object, removing rows from all tables which are not referenced
    #' by the fact table anymore.
    #' E.g., can be called after filtering/subsetting the fact table.
    #'
    #' @return Modified `self` (invisibly).
    sweep = function() {
      fact = self$data$fact
      uhashes = unique(self$data$fact[, "uhash", with = FALSE])
      self$data$uhashes = uhashes[self$data$uhashes, on = "uhash", nomatch = NULL]

      for (nn in c("tasks", "learners", "learner_components", "resamplings")) {
        tab = self$data[[nn]]
        if (nrow(tab)) {
          keycol = key(tab)
          y = unique(fact[, keycol, with = FALSE], by = keycol)
          self$data[[nn]] = merge(tab, y, by = keycol, sort = TRUE)
        }
      }

      invisible(self)
    },

    #' @description
    #' Marshals all stored learner models.
    #' This will do nothing to models that are already marshaled.
    #' @param ... (any)\cr
    #'   Additional arguments passed to [`marshal_model()`].
    marshal = function(...) {
      learner_state = NULL
      self$data$fact[,
        learner_state := lapply(learner_state, function(x) marshal_state_if_model(.state = x, inplace = TRUE, ...))
      ]
      invisible(self)
    },
    #' @description
    #' Unmarshals all stored learner models.
    #' This will do nothing to models which are not marshaled.
    #' @param ... (any)\cr
    #'   Additional arguments passed to [`unmarshal_model()`].
    unmarshal = function(...) {
      learner_state = NULL
      self$data$fact[,
        learner_state := lapply(learner_state, function(x) unmarshal_state_if_model(.state = x, inplace = TRUE, ...))
      ]
      invisible(self)
    },

    #' @description
    #' Shrinks the object by discarding parts of the stored data.
    #'
    #' @param backends (`logical(1)`)\cr
    #'   If `TRUE`, the [DataBackend] is removed from all stored [Task]s.
    #' @param models (`logical(1)`)\cr
    #'   If `TRUE`, the stored model is removed from all [Learner]s.
    #'
    #' @return Modified `self` (invisibly).
    discard = function(backends = FALSE, models = FALSE) {
      if (assert_flag(backends)) {
        tab = self$data$tasks
        set(tab, j = "task", value = lapply(tab$task, task_rm_backend))
      }

      if (assert_flag(models)) {
        tab = self$data$fact
        set(tab, j = "learner_state", value = lapply(tab$learner_state, remove_named, "model"))
      }

      invisible(self)
    },

    #' @description
    #' Combines internal tables into a single flat [data.table()].
    #'
    #' @param reassemble_learners (`logical(1)`)\cr
    #'   Reassemble the tasks?
    #' @param convert_predictions (`logical(1)`)\cr
    #'   Convert [PredictionData] to [Prediction]?
    #' @template param_predict_sets
    as_data_table = function(
      view = NULL,
      reassemble_learners = TRUE,
      convert_predictions = TRUE,
      predict_sets = "test"
    ) {
      # nolint next
      .__ii__ = private$get_view_index(view)

      tab = self$data$fact[.__ii__]
      tab = merge(tab, self$data$tasks, by = "task_hash", sort = FALSE)
      tab = merge(tab, self$data$learners, by = "learner_phash", sort = FALSE)
      tab = merge(tab, self$data$resamplings, by = "resampling_hash", sort = FALSE)
      tab = merge(tab, self$data$learner_components, by = "learner_hash", sort = FALSE)
      if (nrow(self$data$data_extras)) {
        tab = merge(tab, self$data$data_extras, by = c("uhash", "iteration"), all.x = TRUE, sort = FALSE)
      }

      if (nrow(tab)) {
        if (reassemble_learners) {
          set(
            tab,
            j = "learner",
            value = reassemble_learners(tab$learner, states = tab$learner_state, param_vals = tab$learner_param_vals)
          )
        }

        if (convert_predictions) {
          set(tab, j = "prediction", value = as_predictions(tab$prediction, predict_sets = predict_sets))
        }
      }

      cns = c(
        "uhash",
        "task",
        "task_hash",
        "learner",
        "learner_hash",
        "learner_param_vals",
        "resampling",
        "resampling_hash",
        "iteration",
        "prediction",
        if (nrow(self$data$data_extras)) "data_extra"
      )
      merge(self$data$uhashes, tab[, cns, with = FALSE], by = "uhash", sort = FALSE)
    },

    #' @description
    #' Get a table of recorded learner logs.
    #'
    #' @param condition (`character(1)`)
    #'   The condition to extract. One of `"message"`, `"warning"` or `"error"`.
    #'
    #' @return [data.table::data.table()].
    logs = function(view = NULL, condition) {
      # nolint next
      .__ii__ = private$get_view_index(view)
      learner_state = NULL
      logs = map(self$data$fact[.__ii__, learner_state], function(s) list(condition = get_log_condition(s, condition)))
      rbindlist(logs, idcol = "iteration", use.names = TRUE)
    },

    #' @description
    #' Sets the threshold for the response prediction of classification learners,
    #' given they have output a probability prediction.
    #' @template param_view
    #' @param threshold (`numeric(1)`)\cr
    #'   Threshold value.
    #' @template param_ties_method
    set_threshold = function(view = NULL, threshold, ties_method = "random") {
      assert_numeric(threshold, len = 1L, lower = 0, upper = 1)
      assert_choice(ties_method, c("random", "first", "last"))
      # nolint next
      .__ii__ = private$get_view_index(view)
      self$data$fact[
        .__ii__,
        "prediction" := lapply(get("prediction"), function(ps) {
          list(lapply(ps, set_threshold_pdata, threshold, ties_method))
        })
      ]
    }
  ),

  active = list(
    #' @field task_type (`character(1)`)\cr
    #'   Returns the task type of stored objects, e.g. `"classif"` or `"regr"`.
    #'   Returns `NULL` if the [ResultData] is empty.
    task_type = function() {
      tab = self$data$tasks
      if (nrow(tab)) {
        tab$task[[1L]]$task_type
      } else {
        NULL
      }
    }
  ),

  private = list(
    get_view_index = function(view) {
      if (is.null(view)) {
        return(TRUE)
      }
      self$data$fact[list(view), on = "uhash", nomatch = NULL, which = TRUE]
    },

    deep_clone = function(name, value) {
      if (name == "data") {
        lapply(self$data, copy)
      } else {
        value
      }
    }
  )
)

#######################################################################################################################
### constructor
#######################################################################################################################
star_init = function() {
  fact = data.table(
    uhash = character(),
    iteration = integer(),
    learner_state = list(),
    prediction = list(),

    learner_hash = character(),
    task_hash = character(),
    learner_phash = character(),
    resampling_hash = character(),

    key = c("uhash", "iteration")
  )

  uhashes = data.table(
    uhash = character()
  )

  tasks = data.table(
    task_hash = character(),
    task = list(),
    key = "task_hash"
  )

  learners = data.table(
    learner_phash = character(),
    learner = list(),
    key = "learner_phash"
  )

  resamplings = data.table(
    resampling_hash = character(),
    resampling = list(),
    key = "resampling_hash"
  )

  learner_components = data.table(
    learner_hash = character(),
    learner_param_vals = list(),
    key = "learner_hash"
  )

  data_extras = data.table(
    uhash = character(),
    iteration = integer(),
    data_extra = list(),
    key = c("uhash", "iteration")
  )

  list(
    fact = fact,
    uhashes = uhashes,
    tasks = tasks,
    learners = learners,
    resamplings = resamplings,
    learner_components = learner_components,
    data_extras = data_extras
  )
}


#' @title Sets the State and/or ParamSet values in a Learner
#'
#' @param learner ([Learner]).
#' @param states (`list()`).
#' @param param_vals (`list()`).
#'
#' @return list of ([Learner]) with updated state and param values.
#' @noRd
reassemble_learners = function(learners, states = NULL, param_vals = NULL) {
  learners = lapply(learners, function(l) l$clone(deep = TRUE))

  if (!is.null(states)) {
    Map(
      function(l, s) {
        l$state = s
      },
      l = learners,
      s = states
    )
  }

  if (!is.null(param_vals)) {
    Map(
      function(l, pv) {
        l$param_set$values = pv
      },
      l = learners,
      pv = param_vals
    )
  }
  learners
}

test_that("results are ordered", {
  grid = data.table(
    task = tsks(c("iris", "sonar")),
    learner = lrns(c("classif.featureless", "classif.debug")),
    resampling = rsmps("cv", folds = 3)
  )
  grid$resampling = pmap(grid, function(task, resampling, ...) resampling$clone(deep = TRUE)$instantiate(task))

  bmr = benchmark(grid, store_models = TRUE)
  rdata = get_private(bmr)$.data

  tab = rdata$as_data_table()
  expect_equal(rdata$uhashes(), rdata$data$uhashes$uhash)
  expect_equal(unique(hashes(tab$task)), hashes(grid$task))
  expect_equal(unique(hashes(tab$learner)), hashes(grid$learner))
  expect_equal(unique(hashes(tab$resampling)), hashes(grid$resampling))

  rdata$data$uhashes$uhash = rev(rdata$data$uhashes$uhash)
  tab = rdata$as_data_table()
  expect_equal(rdata$uhashes(), rdata$data$uhashes$uhash)
  expect_equal(unique(hashes(tab$task)), rev(hashes(grid$task)))
  expect_equal(unique(hashes(tab$learner)), rev(hashes(grid$learner)))
  expect_equal(unique(hashes(tab$resampling)), rev(hashes(grid$resampling)))

  rr = resample(tsk("pima"), lrn("classif.rpart"), rsmp("holdout"))
  rdata$combine(get_private(rr)$.data)
  expect_resultdata(rdata)
  expect_equal(rdata$uhashes()[3], rr$uhash)

  # remove rr in the middle
  uhashes = rdata$uhashes()
  rdata$data$fact = rdata$data$fact[!list(uhashes[2])]
  rdata$sweep()
  expect_resultdata(rdata, TRUE)

  expect_equal(rdata$uhashes(), uhashes[c(1, 3)])

  # test discard
  expect_true(!every(map(rdata$data$fact$learner_state, "model"), is.null))
  expect_true(!some(map(rdata$data$tasks$task, "backend"), is.null))

  rdata$discard(models = TRUE)
  expect_true(every(map(rdata$data$fact$learner_state, "model"), is.null))

  rdata$discard(backends = TRUE)
  expect_true(every(map(rdata$data$tasks$task, "backend"), is.null))
})

test_that("mlr3tuning use case", {
  task = tsk("iris")
  learners = replicate(3, lrn("classif.rpart"), simplify = FALSE)
  learners[[1]]$param_set$values = list(xval = 0, cp = 0.1)
  learners[[2]]$param_set$values = list(xval = 0, cp = 0.2)
  learners[[3]]$param_set$values = list(xval = 0, cp = 0.3)
  resampling = rsmp("holdout")
  resampling$instantiate(task)

  bmr = benchmark(data.table(task = list(task), learner = learners, resampling = list(resampling)))

  rdata = get_private(bmr)$.data

  expect_resultdata(rdata)
  expect_data_table(rdata$data$fact, nrows = 3L)
  expect_data_table(rdata$data$tasks, nrows = 1L)
  expect_data_table(rdata$data$learners, nrows = 1L)
  expect_data_table(rdata$data$learner_components, nrows = 3L)
  expect_data_table(rdata$data$resamplings, nrows = 1L)

  expect_set_equal(map_dbl(bmr$learners$learner, function(l) l$param_set$values$cp), 1:3 / 10)

  get_params = function(l) l$param_set$values$cp
  has_state = function(l) length(l$state) > 0L
  expect_set_equal(map_dbl(bmr$learners$learner, get_params), 1:3 / 10)
  expect_true(all(!map_lgl(bmr$learners$learner, has_state)))

  aggr = bmr$aggregate()
  expect_set_equal(map_dbl(map(aggr$resample_result, "learner"), get_params), 1:3 / 10)
  expect_true(all(!map_lgl(map(aggr$resample_result, "learner"), has_state)))

  scores = bmr$score()
  expect_set_equal(map_dbl(scores$learner, get_params), 1:3 / 10)
  expect_true(every(scores$learner, has_state))

  learner_states = rdata$learner_states()
  expect_list(learner_states, any.missing = FALSE, len = 3)
  expect_set_equal(map_dbl(learner_states, function(l) l$param_vals$cp), 1:3 / 10)
})

test_that("predict set selection", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart", predict_sets = c("train", "test"))
  resampling = rsmp("holdout")
  rr = resample(task, learner, resampling)

  p1 = rr$predictions("train")[[1]]
  p2 = rr$predictions("test")[[1]]
  expect_prediction(p1)
  expect_prediction(p2)
  expect_disjunct(p1$row_ids, p2$row_ids)

  p1 = rr$prediction("train")
  p2 = rr$prediction("test")
  expect_prediction(p1)
  expect_prediction(p2)
  expect_disjunct(p1$row_ids, p2$row_ids)
})

# data_extra -------------------------------------------------------------------

test_that("data_extra works", {
  # empty result data
  rdata = ResultData$new()

  expect_data_table(rdata$data$data_extras, nrows = 0L)
  expect_names(names(rdata$data$data_extras), identical.to = c("uhash", "iteration", "data_extra"))
  expect_data_table(rdata$data_extra(), nrows = 0L)

  # result data without data_extra
  task = tsk("penguins")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 2)$instantiate(task)
  iterations = seq_len(resampling$iters)
  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }

  rdata = as_result_data(task, learners, resampling, iterations, predictions)
  expect_data_table(rdata$data$data_extras, nrows = 0L)
  expect_names(names(rdata$data$data_extras), identical.to = c("uhash", "iteration", "data_extra"))
  expect_data_table(rdata$data_extra(), nrows = 2L)
  walk(rdata$data_extra()$data_extra, expect_null)
  tab = rdata$as_data_table()
  expect_names(names(tab), disjunct.from = "data_extra")

  # result data with data_extra
  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }
  data_extra = list(list(a = 1), list(b = 2))
  rdata = as_result_data(task, learners, resampling, iterations, predictions, data_extra = data_extra)

  expect_data_table(rdata$data$data_extras, nrows = 2L)
  expect_data_table(rdata$data_extra(), nrows = 2L)
  expect_equal(rdata$data_extra()$data_extra, data_extra)
  tab = rdata$as_data_table()
  expect_names(names(tab), must.include = "data_extra")

  # data_extra is partially missing
  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }
  data_extra = list(list(a = 1), NULL)

  rdata = as_result_data(task, learners, resampling, iterations, predictions, data_extra = data_extra)
  expect_data_table(rdata$data$data_extras, nrows = 1L)
  expect_data_table(rdata$data_extra(), nrows = 2L)
  expect_equal(rdata$data_extra()$data_extra, data_extra)
  tab = rdata$as_data_table()
  expect_names(names(tab), must.include = "data_extra")
})

test_that("combine with data_extra works", {
  # empty result data + result_data with data_extra
  rdata_1 = ResultData$new()

  task = tsk("penguins")
  learner = lrn("classif.rpart")
  resampling = rsmp("cv", folds = 2)$instantiate(task)
  iterations = seq_len(resampling$iters)
  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }
  data_extra = list(list(a = 1), list(b = 2))
  rdata_2 = as_result_data(task, learners, resampling, iterations, predictions, data_extra = data_extra)

  rdata_1$combine(rdata_2)
  expect_data_table(rdata_1$data$data_extras, nrows = 2L)
  expect_data_table(rdata_1$data_extra(), nrows = 2L)

  # result data without data_extra + result_data with data_extra
  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }
  rdata_1 = as_result_data(task, learners, resampling, iterations, predictions)

  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }
  data_extra = list(list(a = 1), list(b = 2))
  rdata_2 = as_result_data(task, learners, resampling, iterations, predictions, data_extra = data_extra)
  rdata_2_uhash = rdata_2$data$uhashes$uhash

  rdata_1$combine(rdata_2)

  expect_data_table(rdata_1$data$data_extras, nrows = 2L)
  expect_data_table(rdata_1$data_extra(), nrows = 4L)
  expect_set_equal(rdata_1$data$data_extras$uhash, rdata_2_uhash)
  expect_equal(rdata_1$data$data_extras$iteration, seq(2))
  expect_equal(rdata_1$data$data_extras$data_extra, list(list(a = 1), list(b = 2)))

  # result data without data_extra
  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }
  rdata_1 = as_result_data(task, learners, resampling, iterations, predictions)

  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }
  rdata_2 = as_result_data(task, learners, resampling, iterations, predictions)

  rdata_1$combine(rdata_2)
  expect_data_table(rdata_1$data$data_extras, nrows = 0L)
  expect_data_table(rdata_1$data_extra(), nrows = 4L)
  expect_names(names(rdata_1$data$data_extras), identical.to = c("uhash", "iteration", "data_extra"))

  # result data with data_extra
  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }
  data_extra = list(list(a = 1), list(b = 2))

  rdata_1 = as_result_data(task, learners, resampling, iterations, predictions, data_extra = data_extra)
  rdata_1_uhash = rdata_1$data$uhashes$uhash

  learners = list()
  predictions = list()
  for (i in iterations) {
    l = learner$clone(deep = TRUE)
    learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
    predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
  }
  data_extra = list(list(c = 3), list(d = 4))

  rdata_2 = as_result_data(task, learners, resampling, iterations, predictions, data_extra = data_extra)
  rdata_2_uhash = rdata_2$data$uhashes$uhash

  rdata_1$combine(rdata_2)

  expect_data_table(rdata_1$data$data_extras, nrows = 4L)
  expect_data_table(rdata_1$data_extra(), nrows = 4L)
  expect_set_equal(rdata_1$data$data_extras$uhash, c(rdata_1_uhash, rdata_2_uhash))
  expect_equal(rdata_1$data$data_extras$iteration, rep(seq(2), 2))
  expect_set_equal(unlist(rdata_1$data_extra(c(rdata_1_uhash, rdata_2_uhash))$data_extra), seq(4))
})

#' Convert a Task from One Type to Another
#'
#' @description
#' The task's target is replaced by a different column from the data.
#'
#' @param intask ([Task])\cr
#'   A [Task] to be converted.
#' @param target (`character(1)`)\cr
#'   New target to be set, must be a column in the `intask` data.
#'   If `NULL`, no new target is set, and task is converted as-is.
#' @param new_type (`character(1)`)\cr
#'   The new task type. Must be in [`mlr_reflections$task_types`][mlr_reflections]].
#'   If `NULL` (default), a new task with the same task_type is created.
#' @param drop_original_target (`logical(1)`)\cr
#'   If `FALSE` (default), the original target is added as a feature.
#'   Otherwise the original target is dropped.
#' @param drop_levels (`logical(1)`)\cr
#'   If `TRUE` (default), unused levels of the new target variable are dropped.
#'
#' @return [Task] of requested type.
#' @export
convert_task = function(intask, target = NULL, new_type = NULL, drop_original_target = FALSE, drop_levels = TRUE) {
  assert_task(intask)
  target = assert_subset(target, choices = intask$col_info$id) %??% intask$target_names
  new_type = assert_choice(new_type, choices = mlr_reflections$task_types$type, null.ok = TRUE) %??% intask$task_type
  assert_logical(drop_original_target, any.missing = FALSE, len = 1L)

  # get task_type from mlr_reflections and call constructor
  constructor = get(fget_key(mlr_reflections$task_types, new_type, "task", key = "type")[[1L]])
  common_args = intersect(names(intask$extra_args), names(formals(constructor$public_methods$initialize)))
  newtask = invoke(
    constructor$new,
    id = intask$id,
    backend = intask$backend,
    target = target,
    label = intask$label,
    .args = intask$extra_args[common_args]
  )
  newtask$extra_args = intask$extra_args

  # copy row_roles / col_roles / properties
  newtask$row_roles = intask$row_roles
  props = intersect(mlr_reflections$task_col_roles[[intask$task_type]], mlr_reflections$task_col_roles[[new_type]])
  col_roles = intask$col_roles[props]
  # add the original target(s) as features, only keeping 'new_target'
  col_roles$feature = c(col_roles$feature, setdiff(intask$col_roles$target, target))
  col_roles$target = target
  # remove new target from features
  col_roles$feature = setdiff(col_roles$feature, target)
  newtask$col_roles[props] = col_roles

  # during prediction, when target is NA, we do not call droplevels
  if (assert_flag(drop_levels)) {
    newtask$droplevels()
  }

  # if drop_original_target, remove the original target from the features
  if (drop_original_target) {
    newtask$col_roles$feature = setdiff(newtask$col_roles$feature, intask$col_roles$target)
  }

  # preserve internal_valid_task
  if (!is.null(intask$internal_valid_task)) {
    newtask$internal_valid_task = convert_task(
      intask$internal_valid_task,
      target = target,
      new_type = new_type,
      drop_original_target = drop_original_target,
      drop_levels = drop_levels
    )
  }

  newtask
}

translate_types = function(x) {
  r_types = mlr_reflections$task_feature_types
  p_types = names(mlr_reflections$task_feature_types)
  factor(map_values(x, r_types, p_types), levels = p_types)
}

get_featureless_learner = function(task_type) {
  if (!is.na(task_type)) {
    id = paste0(task_type, ".featureless")
    if (mlr_learners$has(id)) {
      return(mlr_learners$get(id))
    }
  }

  NULL
}

assert_ordered_set = function(x, y, ...) {
  assert_subset(x, y, ...)
  x[reorder_vector(x, y)]
}

set_data_table_class = function(x, class = NULL) {
  setattr(x, "class", c(class, "data.table", "data.frame"))
}

print_data_table = function(x, hidden_columns) {
  hidden_columns = intersect(names(x), hidden_columns)
  extra_class = class(x)[1L]
  set_data_table_class(x)
  print(x[, .SD, .SDcols = !hidden_columns])
  if (length(hidden_columns)) {
    catf(str_indent("Hidden columns:", hidden_columns))
  }
  set_data_table_class(x, extra_class)
}

clone_without = function(x, y) {
  y_prev = x[[y]]
  x[[y]] = NULL
  x2 = x$clone(deep = TRUE)
  x[[y]] = y_prev
  x2
}

clone_rep = function(x, n) {
  xc = x$clone(deep = TRUE)
  lapply(seq_len(n), function(i) xc)
}

row_any_na = function(x) {
  if (!anyNA(x)) {
    return(logical(nrow(x)))
  }
  rowSums(is.na(x)) > 0L
}

#' @title Assert Validate
#' @description
#' Asserts whether the input is a valid value for the `$validate` field of a [`Learner`].
#' @param x (any)\cr
#'   The input to check.
#' @export
#' @rdname mlr_assertions
assert_validate = function(x) {
  if (test_numeric(x, lower = 0, upper = 1, len = 1L, any.missing = FALSE)) {
    return(x)
  }
  assert_choice(x, c("predefined", "test"), null.ok = TRUE)
}

# Generalization of quantile(type = 7) for weighted data.

# nolint next
quantile_weighted = function(x, probs, na.rm = FALSE, weights = NULL, digits = 7L, continuous = TRUE) {
  assert_flag(na.rm)
  assert_flag(continuous)
  assert_numeric(x, any.missing = na.rm)
  assert_numeric(probs, lower = -100 * .Machine$double.eps, upper = 1 + 100 * .Machine$double.eps)
  assert_numeric(weights, lower = 0, any.missing = FALSE, len = length(x), null.ok = TRUE)
  weights = weights[!is.na(x)]
  x = x[!is.na(x)] # if na.rm is FALSE and there are NAs, the assert stops us from getting here

  # we default to the unweighted quantile if there are no weights,
  # no (non-NA) probs, or no (non-NA) x, or all weights are the same
  if (
    is.null(weights) || length(x) == 0L || length(probs) == 0L || all(is.na(probs)) || length(unique(weights)) == 1L
  ) {
    return(quantile(x, probs, na.rm = na.rm, digits = digits, type = if (continuous) 7 else 2))
  }

  # We create a piecewise linear function with breaks
  # both at the midpoints of the weights, as well as at the boundaries of the weights
  # So if we have weights, ordered by corresponding x-value, w_(1), w_(2), ..., then we have breaks at
  # 0 [weight midpoint], w_(1)/2 [weight boundary], w_(1)/2 + w_(2)/2 [weight midpoint],
  # w_(1)/2 + w_(2) [weight boundary], w_(2)/2 + w_(3)/2 [weight midpoint]
  # The function values at the weight midpoints are the x-values corresponding to the weights,
  # and the function values at the weight boundaries are the
  # weighted average of the x-values of the previous and next weight midpoint.
  x_order = order(x)
  x = x[x_order]
  weights = weights[x_order]
  if (continuous) {
    # avoid 0 weights so we don't have to handle division by 0
    double_weights = rep(weights / 2, each = 2)[c(-1, -2 * length(weights))] + .Machine$double.xmin
    double_x = rep(x, each = 2)[c(-1, -2 * length(x))] * double_weights

    double_x_weighted = (c(double_x, double_x[[length(double_x)]]) + c(double_x[[1]], double_x)) /
      (c(double_weights, double_weights[[length(double_weights)]]) + c(double_weights[[1]], double_weights))

    pivots = c(0, cumsum(double_weights))

    weights_total = pivots[[length(pivots)]]
    weight_targets = weights_total * probs
    weight_targets[weight_targets < 0] = 0 # since we allow -100 * .Machine$double.eps probs

    lo_indices = findInterval(weight_targets, pivots)
    hi_indices = lo_indices + 1L
    lo_values = double_x_weighted[lo_indices]
    hi_values = double_x_weighted[hi_indices]
    result = lo_values +
      (weight_targets - pivots[lo_indices]) * (hi_values - lo_values) / (pivots[hi_indices] - pivots[lo_indices])
    result[hi_indices > length(double_x_weighted)] = double_x_weighted[length(double_x_weighted)]
  } else {
    pivots = c(0, cumsum(weights))
    weight_targets = pivots[[length(pivots)]] * probs
    weight_targets[weight_targets < 0] = 0 # since we allow -100 * .Machine$double.eps probs
    pivots_used = pivots[-length(pivots)]
    hi_indices = findInterval(weight_targets, pivots_used)
    lo_indices = hi_indices - (weight_targets %in% pivots_used[-1]) # use weighted average for ties
    lo_values = x[lo_indices]
    hi_values = x[hi_indices]
    weights = weights + .Machine$double.xmin
    result = (lo_values * weights[lo_indices] + hi_values * (weights[hi_indices])) /
      (weights[lo_indices] + weights[hi_indices])
  }

  pnames = paste0(formatC(probs * 100, format = "fg", width = 1, digits = digits), "%")
  pnames[is.na(pnames)] = ""
  names(result) = pnames
  result
}

weighted_mean_sd = function(x, weights) {
  if (is.null(weights)) {
    return(list(mean = mean(x), sd = sd(x)))
  }
  weights_sum = sum(weights)
  mean = sum(x * weights) / weights_sum
  sd = sqrt(sum(weights * (x - mean)^2) / (weights_sum - sum(weights^2) / weights_sum))
  list(mean = mean, sd = sd)
}

# Alternative formatting function for Task / Learner / Measure because mlr3misc::error_* and warning_* fail with input
# that is formatted with angle brackets, e.g. "<TaskClassif:iris>",
# see https://github.com/r-lib/cli/issues/789
# Replace this with x$format() when the issue is solved.
format_angle_brackets = function(x) {
  sprintf("<<%s:%s>>", class(x)[1L], x$id)
}

#' @exportS3Method data.table::format_col
format_col.condition = function(x, ...) {
  msg = cli::ansi_strip(conditionMessage(x))
  trimws(gsub("\n", " ", msg, fixed = TRUE))
}

#' @exportS3Method data.table::format_list_item
format_list_item.condition = function(x, ...) {
  format_col(x, ...)
}

#' @title Container for Results of `resample()`
#'
#' @include mlr_reflections.R
#'
#' @description
#' This is the result container object returned by [resample()].
#'
#' Note that all stored objects are accessed by reference.
#' Do not modify any object without cloning it first.
#'
#' [ResampleResult]s can be visualized via \CRANpkg{mlr3viz}'s `autoplot()` function.
#'
#' @template param_measures
#'
#' @section S3 Methods:
#' * `as.data.table(rr, reassemble_learners = TRUE, convert_predictions = TRUE, predict_sets = "test")`\cr
#'   [ResampleResult] -> [data.table::data.table()]\cr
#'   Returns a tabular view of the internal data.
#' * `c(...)`\cr
#'   ([ResampleResult], ...) -> [BenchmarkResult]\cr
#'   Combines multiple objects convertible to [BenchmarkResult] into a new [BenchmarkResult].
#'
#' @template seealso_resample
#' @export
#' @examples
#' task = tsk("penguins")
#' learner = lrn("classif.rpart")
#' resampling = rsmp("cv", folds = 3)
#' rr = resample(task, learner, resampling)
#' print(rr)
#'
#' # combined predictions and predictions for each fold separately
#' rr$prediction()
#' rr$predictions()
#'
#' # folds scored separately, then aggregated (macro)
#' rr$aggregate(msr("classif.acc"))
#'
#' # predictions first combined, then scored (micro)
#' rr$prediction()$score(msr("classif.acc"))
#'
#' # check for warnings and errors
#' rr$warnings
#' rr$errors
ResampleResult = R6Class(
  "ResampleResult",
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #' An alternative construction method is provided by [as_resample_result()].
    #'
    #' @param data ([ResultData] | [data.table()])\cr
    #'   An object of type [ResultData], either extracted from another [ResampleResult], another
    #'   [BenchmarkResult], or manually constructed with [as_result_data()].
    #' @param view (`character()`)\cr
    #'   Single `uhash` of the [ResultData] to operate on.
    #'   Used internally for optimizations.
    initialize = function(data = ResultData$new(), view = NULL) {
      private$.data = assert_class(data, "ResultData")
      private$.view = assert_string(view, null.ok = TRUE)
    },

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s>", class(self)[1L])
    },

    #' @description
    #' Printer.
    #' @param ... (ignored).
    print = function(...) {
      tab = self$score(measures = list(), conditions = TRUE)
      set_data_table_class(tab)
      tab[, "warnings" := map(get("warnings"), length)]
      tab[, "errors" := map(get("errors"), length)]
      cat_cli(cli_h1("{.cls {class(self)[1L]}} with {.val {self$iters}} resampling iterations"))
      if (nrow(tab)) {
        tab = remove_named(tab, c("task", "learner", "resampling", "prediction"))
        print(tab, class = FALSE, row.names = FALSE, print.keys = FALSE, digits = 3L)
      }
    },

    #' @description
    #' Opens the corresponding help page referenced by field `$man`.
    help = function() {
      open_help("mlr3::ResampleResult")
    },

    #' @description
    #' Combined [Prediction] of all individual resampling iterations, and all provided predict sets.
    #' Note that, per default, most performance measures do not operate on this object directly,
    #' but instead on the prediction objects from the resampling iterations separately, and then combine
    #' the performance scores with the aggregate function of the respective [Measure] (macro averaging).
    #'
    #' If you calculate the performance on this prediction object directly, this is called micro averaging.
    #'
    #' @param predict_sets (`character()`)\cr
    #'   Subset of `{"train", "test"}`.
    #' @return [Prediction] or empty `list()` if no predictions are available.
    #' @examples
    #' rr$prediction()
    prediction = function(predict_sets = "test") {
      private$.data$prediction(private$.view, predict_sets)
    },

    #' @description
    #' List of prediction objects, sorted by resampling iteration.
    #' If multiple sets are given, these are combined to a single one for each iteration.
    #'
    #' If you evaluate the performance on all of the returned prediction objects and then average them, this
    #' is called macro averaging. For micro averaging, operate on the combined prediction object as returned by
    #' `$prediction()`.
    #'
    #' @param predict_sets (`character()`)\cr
    #'   Subset of `{"train", "test", "internal_valid"}`.
    #' @return List of [Prediction] objects, one per element in `predict_sets`.
    #' Or list of empty `list()`s if no predictions are available.
    #' @examples
    #' rr$predictions()
    predictions = function(predict_sets = "test") {
      assert_subset(predict_sets, mlr_reflections$predict_sets, empty.ok = FALSE)
      private$.data$predictions(private$.view, predict_sets)
    },

    #' @description
    #' Returns a table with one row for each resampling iteration, including all involved objects:
    #' [Task], [Learner], [Resampling], iteration number (`integer(1)`), and (if enabled)
    #' one [Prediction] for each predict set of the [Learner].
    #' Additionally, a column with the individual (per resampling iteration) performance is added
    #' for each [Measure] in `measures`, named with the id of the respective measure id.
    #' If `measures` is `NULL`, `measures` defaults to the return value of [default_measures()].
    #'
    #' @param ids (`logical(1)`)\cr
    #'   If `ids` is `TRUE`, extra columns with the ids of objects (`"task_id"`, `"learner_id"`, `"resampling_id"`)
    #'   are added to the returned table.
    #'   These allow to subset more conveniently.
    #'
    #' @param conditions (`logical(1)`)\cr
    #'   Adds condition messages (`"warnings"`, `"errors"`) as extra
    #'   list columns of character vectors to the returned table
    #'
    #' @param predictions (`logical(1)`)\cr
    #'   Additionally return prediction objects, one column for each `predict_set` of the learner.
    #'   Columns are named `"prediction_train"`, `"prediction_test"` and `"prediction_internal_valid"`,
    #'   if present.
    #'
    #' @return [data.table::data.table()].
    #' @examples
    #' rr$score(msr("classif.acc"))
    score = function(measures = NULL, ids = TRUE, conditions = FALSE, predictions = TRUE) {
      measures = assert_measures(as_measures(measures, task_type = self$task_type))
      assert_flag(ids)
      assert_flag(conditions)
      assert_flag(predictions)

      tab = score_measures(self, measures, view = private$.view)

      if (ids) {
        set(tab, j = "task_id", value = ids(tab[["task"]]))
        set(tab, j = "learner_id", value = ids(tab[["learner"]]))
        set(tab, j = "resampling_id", value = ids(tab[["resampling"]]))
        setcolorder(
          tab,
          c("task", "task_id", "learner", "learner_id", "resampling", "resampling_id", "iteration", "prediction")
        )
      }

      if (conditions) {
        set(tab, j = "warnings", value = map(tab$learner, "warnings"))
        set(tab, j = "errors", value = map(tab$learner, "errors"))
      }

      if (predictions && nrow(tab)) {
        predict_sets = intersect(mlr_reflections$predict_sets, tab$learner[[1L]]$predict_sets)
        predict_cols = sprintf("prediction_%s", predict_sets)
        for (i in seq_along(predict_sets)) {
          set(
            tab,
            j = predict_cols[i],
            value = map(tab$prediction, function(p) as_prediction(p[[predict_sets[i]]], check = FALSE))
          )
        }
      } else {
        predict_cols = character()
      }

      set_data_table_class(tab, "rr_score")

      cns = c(
        "task",
        "task_id",
        "learner",
        "learner_id",
        "resampling",
        "resampling_id",
        "iteration",
        predict_cols,
        "warnings",
        "errors",
        ids(measures)
      )
      cns = intersect(cns, names(tab))
      tab[, cns, with = FALSE]
    },

    #' @description
    #' Calculates the observation-wise loss via the [Measure]'s `obs_loss` method.
    #' Returns a `data.table()` with an `iteration` column plus one numeric column for each measure,
    #' named with the respective measure id.
    #' If there is no observation-wise loss function for the measure, the column is filled with `NA_real_` values.
    #' Note that some measures such as RMSE, do have an `$obs_loss`,
    #' but they require an additional transformation after aggregation,
    #' in this example taking the square-root.
    #'
    #' @param predict_sets (`character()`)\cr
    #'   The predict sets.
    #' @examples
    #' rr$obs_loss(msr("classif.acc"))
    obs_loss = function(measures = NULL, predict_sets = "test") {
      measures = assert_measures(as_measures(measures, task_type = self$task_type))
      map_dtr(
        self$predictions(predict_sets),
        function(pred) {
          pred$obs_loss(measures)
        },
        .idcol = "iteration"
      )
    },

    #' @description
    #' Calculates and aggregates performance values for all provided measures, according to the
    #' respective aggregation function in [Measure].
    #' If `measures` is `NULL`, `measures` defaults to the return value of [default_measures()].
    #'
    #' @return Named `numeric()`.
    #' @examples
    #' rr$aggregate(msr("classif.acc"))
    aggregate = function(measures = NULL) {
      measures = assert_measures(as_measures(measures, task_type = self$task_type))
      resample_result_aggregate(self, measures)
    },

    #' @description
    #' Subsets the [ResampleResult], reducing it to only keep the iterations specified in `iters`.
    #'
    #' @param iters (`integer()`)\cr
    #'   Resampling iterations to keep.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' rr$filter(1L)
    filter = function(iters) {
      iters = assert_integerish(
        iters,
        lower = 1L,
        upper = self$resampling$iters,
        any.missing = FALSE,
        unique = TRUE,
        coerce = TRUE
      )

      private$.data = private$.data$clone(deep = TRUE)
      fact = private$.data$data$fact
      if (!is.null(private$.view)) {
        fact = fact[list(private$.view), on = "uhash", nomatch = NULL]
      }

      private$.data$data$fact = fact[list(iters), on = "iteration", nomatch = NULL]

      invisible(self)
    },

    #' @description
    #' Shrinks the [ResampleResult] by discarding parts of the internally stored data.
    #' Note that certain operations might stop work, e.g. extracting
    #' importance values from learners or calculating measures requiring the task's data.
    #'
    #' @param backends (`logical(1)`)\cr
    #'   If `TRUE`, the [DataBackend] is removed from all stored [Task]s.
    #' @param models (`logical(1)`)\cr
    #'   If `TRUE`, the stored model is removed from all [Learner]s.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    discard = function(backends = FALSE, models = FALSE) {
      private$.data$discard(backends = backends, models = models)
    },

    #' @description
    #' Marshals all stored models.
    #' @param ... (any)\cr
    #'   Additional arguments passed to [`marshal_model()`].
    #' @examples
    #' rr$marshal()
    marshal = function(...) {
      private$.data$marshal(...)
    },
    #' @description
    #' Unmarshals all stored models.
    #' @param ... (any)\cr
    #'   Additional arguments passed to [`unmarshal_model()`].
    #' @examples
    #' rr$unmarshal()
    unmarshal = function(...) {
      private$.data$unmarshal(...)
    },

    #' @description
    #' Sets the threshold for the response prediction of classification learners, given they have
    #' output a probability prediction for a binary classification task.
    #' This modifies the object in-place.
    #' @param threshold (`numeric(1)`)\cr
    #'   Threshold value.
    #' @template param_ties_method
    #' @examples
    #' learner = lrn("classif.rpart", predict_type = "prob")
    #' rr = resample(tsk("sonar"), learner, rsmp("cv", folds = 3))
    #' rr$set_threshold(0.6)
    set_threshold = function(threshold, ties_method = "random") {
      if (!self$task_type == "classif") {
        error_input("Can only change the threshold for classification problems, but task type is '%s'.", self$task_type)
      }
      private$.data$set_threshold(self$uhash, threshold, ties_method)
    }
  ),

  active = list(
    #' @field task_type (`character(1)`)\cr
    #' Task type of objects in the `ResampleResult`, e.g. `"classif"` or `"regr"`.
    #' This is `NA` for empty [ResampleResult]s.
    task_type = function(rhs) {
      assert_ro_binding(rhs)
      private$.data$task_type
    },

    #' @field uhash (`character(1)`)\cr
    #' Unique hash for this object.
    uhash = function(rhs) {
      assert_ro_binding(rhs)
      uhash = private$.data$uhashes(private$.view)
      if (length(uhash) == 0L) NA_character_ else uhash
    },

    #' @field iters (`integer(1)`)\cr
    #' Number of resampling iterations stored in the `ResampleResult`.
    iters = function(rhs) {
      private$.data$iterations(private$.view)
    },

    #' @field task ([Task])\cr
    #' The task [resample()] operated on.
    task = function(rhs) {
      assert_ro_binding(rhs)
      tab = private$.data$tasks(private$.view)
      if (nrow(tab) == 0L) {
        return(NULL)
      }
      tab$task[[1L]]
    },

    #' @field learner ([Learner])\cr
    #' Learner prototype [resample()] operated on.
    #' For a list of **trained** learners, see methods `$learners()`.
    learner = function(rhs) {
      assert_ro_binding(rhs)
      tab = private$.data$learners(private$.view, states = FALSE)
      if (nrow(tab) == 0L) {
        return(NULL)
      }
      tab$learner[[1L]]
    },

    #' @field resampling ([Resampling])\cr
    #' Instantiated [Resampling] object which stores the splits into training and test.
    resampling = function(rhs) {
      assert_ro_binding(rhs)
      tab = private$.data$resamplings(private$.view)
      if (nrow(tab) == 0L) {
        return(NULL)
      }
      tab$resampling[[1L]]
    },

    #' @field learners (list of [Learner])\cr
    #' List of trained learners, sorted by resampling iteration.
    learners = function(rhs) {
      assert_ro_binding(rhs)
      private$.data$learners(private$.view)$learner
    },

    #' @field data_extra (list())\cr
    #' Additional data stored in the [ResampleResult].
    data_extra = function(rhs) {
      assert_ro_binding(rhs)
      private$.data$data_extra(private$.view)
    },

    #' @field warnings ([data.table::data.table()])\cr
    #' A table with all warning messages.
    #' Column names are `"iteration"` and `"condition"`.
    #' Note that there can be multiple rows per resampling iteration if multiple warnings have been recorded.
    warnings = function(rhs) {
      assert_ro_binding(rhs)
      private$.data$logs(private$.view, "warning")
    },

    #' @field errors ([data.table::data.table()])\cr
    #' A table with all error messages.
    #' Column names are `"iteration"` and `"condition"`.
    #' Note that there can be multiple rows per resampling iteration if multiple errors have been recorded.
    errors = function(rhs) {
      assert_ro_binding(rhs)
      private$.data$logs(private$.view, "error")
    }
  ),

  private = list(
    # @field data (`ResultData`)\cr
    # Internal data storage object of type `ResultData`.
    .data = NULL,

    # @field view (`character(1)`)\cr
    # Subset of uhashes in the [ResultData] object to operate on.
    .view = NULL,

    deep_clone = function(name, value) {
      if (name == ".data") value$clone(deep = TRUE) else value
    }
  )
)

#' @export
# nolint next
as.data.table.ResampleResult = function(x, ..., predict_sets = "test") {
  private = get_private(x)
  tab = private$.data$as_data_table(view = private$.view, predict_sets = predict_sets)
  cns = c("task", "learner", "resampling", "iteration", "prediction", if ("data_extra" %in% names(tab)) "data_extra")
  tab[, cns, with = FALSE]
}


#' @export
c.ResampleResult = function(...) {
  do.call(c, lapply(list(...), as_benchmark_result))
}


resample_result_aggregate = function(rr, measures) {
  unlist(map(unname(measures), function(m) {
    val = m$aggregate(rr)
    # CIs in mlr3inferr return more than 1 value and are already named
    if (length(val) == 1L) {
      return(set_names(val, m$id))
    }
    val
  })) %??%
    set_names(numeric(), character())
}

#' @export
print.rr_score = function(x, ...) {
  predict_cols = sprintf("prediction_%s", mlr_reflections$predict_sets)
  print_data_table(x, c("task", "learner", "resampling", predict_cols))
}

tmp = tsk("iris", id = "iris_small")$select("Sepal.Length")
tasks = c(mlr_tasks$mget(c("iris", "sonar")), list(tmp))
learners = mlr_learners$mget(c("classif.featureless", "classif.rpart"))
resamplings = rsmp("cv", folds = 3)
design = benchmark_grid(tasks, learners, resamplings)
bmr = benchmark(design)

test_that("Basic benchmarking", {
  expect_benchmark_result(bmr)
  expect_names(
    names(as.data.table(bmr)),
    permutation.of = c(mlr_reflections$rr_names, "uhash", "prediction", "task_id", "learner_id", "resampling_id")
  )

  tab = as.data.table(bmr)
  expect_data_table(tab, nrows = 18L, ncols = 9L)
  expect_names(
    names(tab),
    permutation.of = c("uhash", "prediction", mlr_reflections$rr_names, "task_id", "learner_id", "resampling_id")
  )
  measures = list(msr("classif.acc"))

  tab = bmr$score(measures, ids = FALSE, predictions = TRUE)
  expect_data_table(tab, nrows = 18L, ncols = 7L + length(measures))
  expect_names(names(tab), must.include = c("nr", "uhash", "prediction_test", mlr_reflections$rr_names, ids(measures)))
  expect_list(tab$prediction_test, "Prediction")

  tab = bmr$tasks
  expect_data_table(tab, nrows = 3L, any.missing = FALSE)
  expect_names(names(tab), identical.to = c("task_hash", "task_id", "task"))
  expect_hash(tab$task_hash, len = 3L)

  tab = bmr$learners
  expect_data_table(tab, nrows = 2L, any.missing = FALSE)
  expect_names(names(tab), identical.to = c("learner_hash", "learner_id", "learner"))
  expect_hash(tab$learner_hash, len = 2L)
  qexpectr(map(tab$learner, "state"), "0")

  tab = bmr$resamplings
  expect_data_table(tab, nrows = 3L, any.missing = FALSE)
  expect_names(names(tab), permutation.of = c("resampling_hash", "resampling", "resampling_id"))
  expect_hash(tab$resampling_hash, len = 3L)

  tab = bmr$aggregate(measures)
  expect_data_table(tab, nrows = 6L)
  expect_names(
    names(tab),
    type = "unique",
    identical.to = c("nr", "resample_result", "task_id", "learner_id", "resampling_id", "iters", ids(measures))
  )
  m = measures[[1L]]
  expect_numeric(tab[[m$id]], any.missing = FALSE, lower = m$range[1], upper = m$range[2])
})

test_that("ResampleResult / hash", {
  m = msr("classif.ce")
  aggr = bmr$aggregate(m, uhashes = TRUE)
  nr = aggr$nr
  expect_integer(nr, len = 6L, any.missing = FALSE, unique = TRUE)

  for (i in nr) {
    rr = aggr$resample_result[[i]]
    expect_resample_result(rr)
    expect_equal(unname(rr$aggregate(m)), aggr[["classif.ce"]][i])
    expect_equal(bmr$uhashes[i], rr$uhash)
  }
})


test_that("discarding model", {
  bmr2 = benchmark(benchmark_grid(tasks[1L], learners[1L], resamplings), store_models = FALSE)
  expect_benchmark_result(bmr2)
  expect_true(every(map(as.data.table(bmr2)$learner, "model"), is.null))

  bmr2 = benchmark(benchmark_grid(tasks[1L], learners[1L], resamplings), store_models = TRUE)
  expect_benchmark_result(bmr2)
  expect_false(every(map(as.data.table(bmr2)$learner, "model"), is.null))
})

test_that("bmr$combine()", {
  bmr_new = benchmark(benchmark_grid(mlr_tasks$mget("pima"), learners, resamplings))

  combined = list(
    bmr$clone(deep = TRUE)$combine(bmr_new),
    c(bmr, bmr_new)
  )

  for (bmr_combined in combined) {
    expect_benchmark_result(bmr)
    expect_benchmark_result(bmr_new)
    expect_benchmark_result(bmr_combined)

    expect_data_table(get_private(bmr)$.data$data$fact, nrows = 18L)
    expect_data_table(get_private(bmr_new)$.data$data$fact, nrows = 6L)
    expect_data_table(get_private(bmr_combined)$.data$data$fact, nrows = 24L)

    expect_false("pima" %chin% bmr$tasks$task_id)
    expect_true("pima" %chin% bmr_new$tasks$task_id)
    expect_true("pima" %chin% bmr_combined$tasks$task_id)
  }

  rr = resample(tsk("zoo"), lrn("classif.rpart"), rsmp("holdout"))
  bmr2 = c(combined[[1]], rr)
  expect_benchmark_result(bmr2)
  expect_data_table(get_private(bmr2)$.data$data$fact, nrows = 25L)
})

test_that("empty bmr", {
  bmr_new = BenchmarkResult$new()
  expect_benchmark_result(bmr_new)

  bmr_new$combine(NULL)
  expect_benchmark_result(bmr_new)

  bmr_new$combine(bmr)
  expect_benchmark_result(bmr_new)
  expect_data_table(get_private(bmr_new)$.data$data$fact, nrows = nrow(get_private(bmr)$.data))
})

test_that("bmr$resample_result()", {
  uhashes = bmr$uhashes
  expect_resample_result(bmr$resample_result(1L))
  expect_resample_result(bmr$resample_result(uhash = uhashes[1]))
  expect_resample_result(bmr$resample_result(learner_id = "classif.featureless", task_id = "iris"))
  expect_error(bmr$resample_result(learner_id = "classif.featureless"), "requires selecting exactly one")
  expect_error(bmr$resample_result(0))
  expect_error(bmr$resample_result(100))
  expect_error(bmr$resample_result(uhash = "a"))
  expect_error(bmr$resample_result(i = 1, uhash = uhashes[1]))
})

test_that("inputs are cloned", {
  task = tsk("iris")
  learner = lrn("classif.featureless")
  resampling = rsmp("holdout")

  expect_error(
    benchmark(data.table(task = list(task), learner = list(learner), resampling = list(resampling))),
    "instantiated"
  )
  bmr = benchmark(
    design = data.table(task = list(task), learner = list(learner), resampling = list(resampling$instantiate(task)))
  )
  rr = bmr$resample_result(1)

  expect_different_address(task, rr$task)
  expect_different_address(learner, get_private(rr)$.data$data$learner_components$learner[[1L]])
  expect_different_address(resampling, rr$resampling)
})

test_that("memory footprint", {
  expect_equal(uniqueN(map_chr(design$task, address)), 3L)
  expect_equal(uniqueN(map_chr(design$learner, address)), 2L)
  expect_equal(uniqueN(map_chr(design$resampling, address)), 3L)

  x = as.data.table(bmr)
  expect_equal(uniqueN(map_chr(x$task, address)), 3L)
  expect_equal(uniqueN(map_chr(x$learner, address)), 18L)
  expect_equal(uniqueN(map_chr(x$resampling, address)), 3L)
})

test_that("resampling validation in benchmark_grid", {
  task1 = tsk("iris")
  task2 = tsk("pima")
  resampling_1 = rsmp("holdout")
  resampling_2 = rsmp("holdout")

  # should work when resamplings are instantiated on their corresponding tasks
  resampling_1$instantiate(task1)
  resampling_2$instantiate(task2)
  expect_data_table(benchmark_grid(
    list(task1, task2),
    lrn("classif.rpart"),
    list(resampling_1, resampling_2),
    paired = TRUE
  ))

  # should fail when resamplings are not instantiated
  resampling_1 = rsmp("holdout")
  resampling_2 = rsmp("holdout")
  expect_error(
    benchmark_grid(list(task1, task2), lrn("classif.rpart"), list(resampling_1, resampling_2), paired = TRUE),
    "is not instantiated"
  )

  # should fail when resampling is instantiated on wrong task
  resampling_1$instantiate(task1)
  resampling_2$instantiate(task1)
  expect_error(
    benchmark_grid(list(task1, task2), lrn("classif.rpart"), list(resampling_1, resampling_2), paired = TRUE),
    "not instantiated"
  )

  # should fail when task row hashes don't match
  task1 = tsk("iris")
  task2 = tsk("iris")$filter(1:100)
  resampling_1 = rsmp("holdout")
  resampling_1$instantiate(task1)

  expect_error(benchmark_grid(list(task1, task2), lrn("classif.rpart"), list(resampling_1)), "not instantiated")

  # should fail when the tasks have the same number of rows but different row hashes
  task1 = tsk("iris")$filter(1:75)
  task2 = tsk("iris")$filter(76:150)
  resampling_1 = rsmp("holdout")
  resampling_1$instantiate(task1)
  expect_error(benchmark_grid(list(task1, task2), lrn("classif.rpart"), list(resampling_1)), "not instantiated")

  # should work when all resamplings are uninstantiated
  res1 = rsmp("holdout")
  res2 = rsmp("holdout")
  expect_data_table(benchmark_grid(list(task1, task2), lrn("classif.rpart"), list(res1, res2)))

  # should fail when some resamplings are instantiated and others are not
  res1$instantiate(task1)
  expect_error(
    benchmark_grid(list(task1, task2), lrn("classif.rpart"), list(res1, res2)),
    "All resamplings must be instantiated, or none at all"
  )
})

test_that("multiple measures", {
  tasks = list(tsk("iris"), tsk("sonar"))
  learner = lrn("classif.featureless")
  measures = list(msr("classif.ce"), msr("classif.acc"))
  bmr = benchmark(design = benchmark_grid(tasks, learner, rsmp("cv", folds = 3)))
  expect_subset(c("classif.ce", "classif.acc"), names(bmr$aggregate(measures)))
})

test_that("predict_type is checked", {
  task = tsk("sonar")
  learner = lrn("classif.featureless")
  resampling = rsmp("cv", folds = 3L)
  design = benchmark_grid(task, learner, resampling)
  bmr = benchmark(design)
  expect_error(bmr$aggregate("classif.auc", "predict_type"))
})

test_that("custom resampling (#245)", {
  task_boston = tsk("california_housing")
  lrn = lrn("regr.featureless")

  rdesc = rsmp("custom")
  train_sets = list((1:200), (1:300), (1:400))
  test_sets = list((201:301), (301:401), (401:501))
  rdesc$instantiate(task_boston, train_sets, test_sets)

  expect_resample_result(mlr3::resample(task_boston, lrn, rdesc))

  design = data.table(task = list(task_boston), learner = list(lrn), resampling = list(rdesc))
  bmr = benchmark(design)
  expect_benchmark_result(bmr)

  # Issue #451
  design = benchmark_grid(tasks = task_boston, learners = lrn, resamplings = rdesc)
  expect_data_table(design, nrows = 1)
})

test_that("extract params in aggregate and score", {
  # set params differently in a few learners
  lrns = list(
    lrn("classif.rpart", id = "rp1", xval = 0),
    lrn("classif.rpart", id = "rp2", xval = 0, cp = 0.2, minsplit = 2),
    lrn("classif.rpart", id = "rp3", xval = 0, cp = 0.1)
  )
  bmr = benchmark(benchmark_grid(tsk("wine"), lrns, rsmp("cv", folds = 3)))
  aggr = bmr$aggregate(params = TRUE)
  setorder(aggr, "learner_id")
  expect_list(aggr$params[[1]], names = "unique", len = 1L)
  expect_list(aggr$params[[2]], names = "unique", len = 3L)
  expect_list(aggr$params[[3]], names = "unique", len = 2L)

  scores = bmr$score()
  pvs = map(scores$learner, function(l) l$param_set$values)
  expect_true(all(sapply(split(lengths(pvs), scores$nr), uniqueN) == 1))
  expect_set_equal(lengths(pvs), 1:3)

  # only one params
  lrns = mlr_learners$mget("classif.featureless")
  bmr = benchmark(benchmark_grid(tsk("wine"), lrns, rsmp("cv", folds = 3)))
  aggr = bmr$aggregate(params = TRUE)
  expect_list(aggr$params[[1]], names = "unique", len = 1L)

  # no params
  lrns = mlr_learners$mget("classif.debug")
  lrns$classif.debug$param_set$values = list()
  bmr = benchmark(benchmark_grid(tsk("wine"), lrns, rsmp("cv", folds = 3)))
  aggr = bmr$aggregate(params = TRUE)
  expect_list(aggr$params[[1]], names = "unique", len = 0L)

  expect_true(all(c("warnings", "errors") %chin% names(bmr$score(conditions = TRUE))))
})

test_that("benchmark_grid", {
  learner = lrn("classif.rpart")

  tasks = tsks(c("iris", "iris"))
  resamp = rsmp("cv")$instantiate(tasks[[1]])
  expect_data_table(benchmark_grid(tasks, learner, resamp))

  tasks = tsks(c("iris", "sonar"))
  resamp = rsmp("cv")$instantiate(tasks[[1]])
  expect_error(benchmark_grid(tasks, learner, resamp), "not instantiated")
})

test_that("filter", {
  tasks = lapply(c("iris", "sonar"), tsk)
  learners = lapply(c("classif.featureless", "classif.rpart"), lrn)
  resamplings = list(rsmp("cv", folds = 3), rsmp("holdout"))

  design = benchmark_grid(tasks, learners, resamplings)
  bmr = benchmark(design)

  expect_data_table(get_private(bmr)$.data$data$fact, nrows = 16)

  bmr$filter(task_ids = "sonar")
  expect_data_table(get_private(bmr)$.data$data$fact, nrows = 8)
  expect_resultdata(get_private(bmr)$.data, TRUE)

  bmr$filter(learner_ids = "classif.rpart")
  expect_data_table(get_private(bmr)$.data$data$fact, nrows = 4)
  expect_resultdata(get_private(bmr)$.data, TRUE)

  bmr2 = bmr$clone(deep = TRUE)$filter(resampling_ids = "cv")
  expect_data_table(get_private(bmr2)$.data$data$fact, nrows = 3)
  expect_resultdata(get_private(bmr2)$.data, TRUE)

  bmr$filter(i = 2)
  expect_data_table(get_private(bmr)$.data$data$fact, nrows = 1)
  expect_resultdata(get_private(bmr)$.data, TRUE)

  expect_benchmark_result(bmr)
  expect_benchmark_result(bmr2)
})

test_that("aggregated performance values are calculated correctly (#555)", {
  task = tsk("spam")
  learner1 = lrn("classif.featureless")
  learner2 = lrn("classif.rpart")
  resampling = rsmp("subsampling", repeats = 2)

  design = benchmark_grid(task, list(learner1, learner2), resampling)
  bmr = benchmark(design = design, store_models = TRUE)

  y = bmr$aggregate()$classif.ce
  expect_gt(y[1], y[2])

  y = c(
    bmr$resample_result(1)$aggregate(msr("classif.ce")),
    bmr$resample_result(2)$aggregate(msr("classif.ce"))
  )
  expect_gt(y[1], y[2])
})

test_that("save/load roundtrip", {
  path = tempfile()
  saveRDS(bmr, file = path)

  bmr2 = readRDS(path)
  expect_benchmark_result(bmr2)
})

test_that("debug branch", {
  tmp = tsk("iris", id = "iris_small")$select("Sepal.Length")
  tasks = c(mlr_tasks$mget(c("iris", "sonar")), list(tmp))
  learners = mlr_learners$mget(c("classif.featureless", "classif.rpart"))
  resamplings = rsmp("cv", folds = 2)
  design = benchmark_grid(tasks, learners, resamplings)
  bmr = invoke(benchmark, design, .opts = list(mlr3.debug = TRUE))
  expect_benchmark_result(bmr)
})

# uncomment when evaluate 1.0.4 is released

# test_that("encapsulation", {
#   learners = list(lrn("classif.debug", error_train = 1), lrn("classif.rpart"))
#   grid = benchmark_grid(tasks, learners, resamplings)

#   expect_error(benchmark(grid), "classif.debug->train()")
#   bmr = benchmark(grid, encapsulate = "evaluate")
#   aggr = bmr$aggregate(conditions = TRUE)
#   expect_true(all(aggr[learner_id == "classif.debug", errors] == 3L))
#   expect_true(all(aggr[learner_id != "classif.debug", errors] == 0L))

#   for (learner in bmr$learners$learner) {
#     expect_class(learner$fallback, "LearnerClassifFeatureless")
#     expect_equal(learner$encapsulation[["train"]], "evaluate")
#     expect_equal(learner$encapsulation[["predict"]], "evaluate")
#   }
# })

test_that("disable cloning", {
  grid = benchmark_grid(
    tasks = tsk("iris"),
    learners = lrn("classif.featureless"),
    resamplings = rsmp("holdout")
  )
  task = grid$task[[1L]]
  learner = grid$learner[[1L]]
  resampling = grid$resampling[[1L]]

  bmr = benchmark(grid, clone = c())

  expect_same_address(task, bmr$tasks$task[[1]])
  expect_same_address(learner, get_private(bmr)$.data$data$learners$learner[[1]])
  expect_same_address(resampling, bmr$resamplings$resampling[[1]])

  expect_identical(task$hash, bmr$tasks$task[[1]]$hash)
  expect_identical(learner$hash, bmr$learners$learner[[1]]$hash)
  expect_identical(resampling$hash, bmr$resamplings$resampling[[1]]$hash)
})

test_that("task and learner assertions", {
  grid = benchmark_grid(
    tasks = tsks(c("iris", "california_housing")),
    learners = lrn("classif.rpart"),
    resamplings = rsmp("holdout")
  )

  expect_error(benchmark(grid), "task types")

  grid = benchmark_grid(
    tasks = tsk("iris"),
    learners = lrns(c("classif.rpart", "regr.rpart")),
    resamplings = rsmp("holdout")
  )

  expect_error(benchmark(grid), "learner types")

  grid = benchmark_grid(
    tasks = tsk("iris"),
    learners = lrn("regr.rpart"),
    resamplings = rsmp("holdout")
  )

  expect_error(benchmark(grid), "not match type")
})


test_that("benchmark_grid works if paired = TRUE", {
  tasks = mlr3::tsks(c("pima", "iris"))
  learners = lrns(c("classif.featureless", "classif.rpart"))
  resampling = rsmp("cv")
  resamplings = pmap(
    list(tasks, rsmps(c("cv", "holdout"))),
    function(task, resampling) resampling$instantiate(task)
  )
  design = benchmark_grid(tasks, learners, resamplings, paired = TRUE)
  expect_class(design, "benchmark_grid")
  # design[, identical(task), by = task]]
  # expect(identical(design$resampling[class(learner)[[1]] ==)]))
  expect_true(nrow(design) == 4L) #
  expect_identical(design$task[[1]], design$task[[2]])
  expect_identical(design$task[[3]], design$task[[4]])
  expect_false(identical(design$task[[1]], design$task[[3]]))

  expect_identical(design$resampling[[1]], design$resampling[[2]])
  expect_identical(design$resampling[[3]], design$resampling[[4]])
  expect_false(identical(design$resampling[[1]], design$resampling[[3]]))

  expect_identical(design$learner[[1]], design$learner[[3]])
  expect_identical(design$learner[[2]], design$learner[[4]])
  expect_false(identical(design$learner[[2]], design$learner[[3]]))

  # Resamplings must be instantiated
  tasks = tsks(c("pima", "iris"))
  learners = lrns(c("classif.featureless", "classif.rpart"))
  resamplings = mlr3::rsmps(c("cv", "holdout"))
  expect_error(benchmark_grid(tasks, learners, resamplings, paired = TRUE))

  # Resamplings and tasks must have the same length
  tasks = tsks(c("pima", "iris"))
  learners = lrns(c("classif.featureless", "classif.rpart"))
  resamplings = pmap(
    list(tasks, mlr3::rsmps(c("cv", "holdout"))),
    function(task, resampling) resampling$instantiate(task)
  )
  resamplings = c(resamplings, resamplings)
  expect_error(benchmark_grid(tasks, learners, resamplings, paired = TRUE))

  # Resamplings and tasks must have corresponding hashes

  tasks = tsks(c("pima", "iris"))
  learners = lrns(c("classif.featureless", "classif.rpart"))
  resamplings = pmap(
    list(tasks, mlr3::rsmps(c("cv", "holdout"))),
    function(task, resampling) resampling$instantiate(task)
  )
  resamplings = rev(resamplings)
  expect_error(benchmark_grid(tasks, learners, resamplings, paired = TRUE))
})

test_that("param_values in benchmark", {
  # setup
  tasks = tsks("iris")
  resamplings = list(rsmp("cv", folds = 3)$instantiate(tasks[[1]]))
  learners = lrns("classif.debug")

  # single parameter set via manual design
  design = data.table(
    task = tasks,
    learner = learners,
    resampling = resamplings,
    param_values = list(list(list(x = 1)))
  )
  bmr = benchmark(design)
  expect_benchmark_result(bmr)
  expect_equal(bmr$n_resample_results, 1)
  expect_shape(as.data.table(bmr), nrow = 3L)
  learner = bmr$resample_result(1)$learner
  expect_equal(learner$param_set$values$x, 1)
  expect_shape(as.data.table(bmr), nrow = 3L)

  # multiple parameters set via manual design
  design = data.table(
    task = tasks,
    learner = learners,
    resampling = resamplings,
    param_values = list(list(list(x = 1), list(x = 0.5)))
  )
  bmr = benchmark(design)
  expect_benchmark_result(bmr)
  expect_equal(bmr$n_resample_results, 2)
  expect_shape(as.data.table(bmr), nrow = 6L)
  learner = bmr$resample_result(1)$learner
  expect_equal(learner$param_set$values$x, 1)
  learner = bmr$resample_result(2)$learner
  expect_equal(learner$param_set$values$x, 0.5)

  # benchmark grid does not attach param_values if empty
  design = benchmark_grid(tasks, learners, resamplings)
  expect_names(names(design), permutation.of = c("task", "learner", "resampling"))

  # benchmark grid with param_values
  design = benchmark_grid(tasks, learners, resamplings, param_values = list(list(list(x = 1))))
  expect_data_table(design, nrows = 1)
  expect_names(names(design), permutation.of = c("task", "learner", "resampling", "param_values"))
  bmr = benchmark(design)
  expect_benchmark_result(bmr)

  # benchmark grid with param_values and paired = TRUE
  design = benchmark_grid(tasks, learners, resamplings, param_values = list(list(list(x = 1))), paired = TRUE)
  expect_data_table(design, nrows = 1)
  bmr = benchmark(design)
  expect_benchmark_result(bmr)
  expect_equal(bmr$n_resample_results, 1)

  # benchmark grid with multiple params
  design = benchmark_grid(tasks, learners, resamplings, param_values = list(list(list(x = 1), list(x = 0.5))))
  expect_data_table(design, nrows = 1)
  bmr = benchmark(design)
  expect_benchmark_result(bmr)
  expect_equal(bmr$n_resample_results, 2)

  # benchmark grid with multiple params and multiple learners
  design = benchmark_grid(
    tasks,
    lrns(c("classif.debug", "classif.rpart")),
    rsmp("holdout"),
    param_values = list(list(list(x = 1), list(x = 0.5)), list())
  )
  bmr = benchmark(design)
  expect_benchmark_result(bmr)
  expect_equal(bmr$n_resample_results, 3)

  # constant values are inserted
  learners = lrns("classif.rpart", minsplit = 12)
  design = data.table(
    task = tasks,
    learner = learners,
    resampling = resamplings,
    param_values = list(list(list(cp = 0.1), list(minbucket = 2)))
  )
  bmr = benchmark(design)

  sortnames = function(x) {
    if (!is.null(names(x))) {
      x = x[order(names(x))]
    }
    x
  }
  trained = bmr$learners$learner
  ii = which(map_lgl(trained, function(x) "cp" %chin% names(x$param_set$values))) # find learner with cp
  expect_count(ii)

  expect_equal(sortnames(bmr$learners$learner[-ii][[1]]$param_set$values), list(minbucket = 2, minsplit = 12, xval = 0))
  expect_equal(sortnames(bmr$learners$learner[[ii]]$param_set$values), list(cp = 0.1, minsplit = 12, xval = 0))
})


test_that("learner's validate cannot be 'test' if internal_valid_set is present", {
  # otherwise, predict_set = "internal_valid" would be ambiguous
  learner = lrn("classif.debug", validate = "test", predict_sets = c("train", "internal_valid"))
  task = tsk("iris")
  task$internal_valid_task = 1
  expect_error(benchmark(benchmark_grid(task, learner, rsmp("holdout"))), "cannot be set to ")
})

test_that("learner's validate cannot be a ratio if internal_valid_set is present", {
  # otherwise, predict_set = "internal_valid" would be ambiguous
  learner = lrn("classif.debug", validate = 0.5, predict_sets = c("train", "internal_valid"))
  task = tsk("iris")
  task$internal_valid_task = 1
  expect_error(benchmark(benchmark_grid(task, learner, rsmp("holdout"))), "cannot be set to ")
})

test_that("properties are also checked on validation task", {
  task = tsk("iris")
  row = task$data(1)
  row[[1]][1] = NA
  row$..row_id = 151
  task$rbind(row)
  task$internal_valid_task = 151
  learner = lrn("classif.debug", validate = "predefined")
  learner$properties = setdiff(learner$properties, "missings")

  suppressWarnings(expect_error(benchmark(benchmark_grid(task, learner, rsmp("holdout"))), "missing values"))
})

test_that("unmarshal parameter is respected", {
  learner = lrn("classif.debug", count_marshaling = TRUE)
  learner$encapsulate("callr", lrn("classif.featureless"))
  task = tsk("iris")
  resampling = rsmp("holdout")
  design = benchmark_grid(task, learner, resampling)
  bmr = with_future(future::multisession, {
    list(
      marshaled = benchmark(design, store_models = TRUE, unmarshal = FALSE),
      unmarshaled = benchmark(design, store_models = TRUE, unmarshal = TRUE)
    )
  })
  expect_false(bmr$unmarshaled$resample_result(1)$learners[[1]]$marshaled)
  expect_true(bmr$marshaled$resample_result(1)$learners[[1]]$marshaled)
})

test_that("BenchmarkResult can be (un)marshaled", {
  bmr = benchmark(benchmark_grid(tsk("iris"), lrn("classif.debug"), rsmp("holdout")), store_models = TRUE)
  expect_false(bmr$resample_result(1)$learners[[1]]$marshaled)
  bmr$marshal()
  expect_true(bmr$resample_result(1)$learners[[1]]$marshaled)
  bmr$unmarshal()
  expect_false(bmr$resample_result(1)$learners[[1]]$marshaled)

  # also works with non-marshalable learner
  bmr1 = benchmark(benchmark_grid(tsk("iris"), lrn("classif.featureless"), rsmp("holdout")), store_models = TRUE)
  model = bmr1$resample_result(1)$learners[[1]]$model
  bmr1$unmarshal()
  expect_equal(bmr1$resample_result(1)$learners[[1]]$model, model)
})

test_that("predictions retrieved with as.data.table and predictions method are equal", {
  tab = as.data.table(bmr)
  predictions = unlist(map(bmr$resample_results$resample_result, function(rr) rr$predictions()), recursive = FALSE)
  expect_equal(tab$prediction, predictions)

  tab = as.data.table(bmr, predict_sets = "train")
  predictions = unlist(
    map(bmr$resample_results$resample_result, function(rr) rr$predictions(predict_sets = "train")),
    recursive = FALSE
  )
  expect_equal(tab$prediction, predictions)
})

test_that("score works with predictions and empty predictions", {
  learner_1 = lrn("classif.rpart", predict_sets = "train", id = "learner_1")
  learner_2 = lrn("classif.rpart", predict_sets = "test", id = "learner_2")
  task = tsk("pima")

  design = benchmark_grid(task, list(learner_1, learner_2), rsmp("holdout"))

  bmr = benchmark(design)

  expect_warning(
    {
      tab = bmr$score(msr("classif.ce", predict_sets = "test"))
    },
    "Measure"
  )
  expect_equal(tab$classif.ce[1], NaN)
})

test_that("benchmark_grid only allows unique learner ids", {
  task = tsk("iris")
  learner = lrn("classif.rpart")
  resampling = rsmp("holdout")

  expect_error(benchmark_grid(task, list(learner, learner), resampling), "unique")
})

test_that("benchmark allows that param_values overwrites tune token", {
  learner = lrn("classif.rpart", cp = to_tune(0.01, 0.1))
  design = benchmark_grid(tsk("pima"), learner, rsmp("cv", folds = 3), param_values = list(list(list(cp = 0.01))))
  expect_benchmark_result(benchmark(design))

  learner = lrn("classif.rpart", cp = to_tune(0.01, 0.1))
  design = benchmark_grid(tsk("pima"), learner, rsmp("cv", folds = 3))
  expect_error(benchmark(design), "cannot be trained with TuneToken present")
})

test_that("uhash_table works", {
  design = benchmark_grid(
    tsks(c("iris", "sonar")),
    lrns(c("classif.debug", "classif.featureless")),
    rsmps(c("holdout", "insample"))
  )
  bmr = benchmark(design)
  u = bmr$uhash_table

  # results agree with uhash_table from resample result, which is also tested for correctness
  for (i in seq_len(nrow(u))) {
    rr = bmr$resample_result(i)
    learner_id = rr$learner$id
    task_id = rr$task$id
    resampling_id = rr$resampling$id

    expect_equal(u$learner_id[i], learner_id)
    expect_equal(u$task_id[i], task_id)
    expect_equal(u$resampling_id[i], resampling_id)
    expect_equal(u$uhash[i], rr$uhash)
  }

  # uhash is in correct order
  expect_equal(u$uhash, bmr$uhashes)
  expect_equal(u$uhash, as.data.table(bmr)$uhash)
})

test_that("can change the threshold", {
  task = tsk("iris")$filter(1:80)$droplevels("Species")
  design = benchmark_grid(task, lrn("classif.featureless", predict_type = "prob"), rsmp("insample"))
  bmr = benchmark(design)

  # we can set the threshold and pass ties_method correctly
  expect_true(all(bmr$resample_result(1)$prediction()$response == "setosa"))
  bmr$set_threshold(0.9)
  expect_true(all(bmr$resample_result(1)$prediction()$response == "versicolor"))
  bmr$set_threshold(0.1)
  expect_true(all(bmr$resample_result(1)$prediction()$response == "setosa"))
  bmr$set_threshold(0.625, ties_method = "first")
  expect_true(all(bmr$resample_result(1)$prediction()$response == "setosa"))
  bmr$set_threshold(0.625, ties_method = "last")
  expect_true(all(bmr$resample_result(1)$prediction()$response == "versicolor"))
  # nolint next
  with_seed(1, {
    bmr$set_threshold(0.625, ties_method = "random")
    expect_true(
      "setosa" %in%
        bmr$resample_result(1)$prediction()$response &&
        "versicolor" %in% bmr$resample_result(1)$prediction()$response
    )
  })

  # Don't modify any threshold when at least one operation is invalid
  design = suppressWarnings(benchmark_grid(
    task,
    c(lrn("classif.featureless", predict_type = "prob"), lrn("classif.debug")),
    rsmp("insample")
  ))
  bmr = benchmark(design)
  response = bmr$resample_result(1)$prediction()$response
  expect_error(bmr$set_threshold(0.9), "Cannot set threshold, no probabilities available")
  # the other prediction was also not affected, we want to avoid partial updates
  expect_equal(bmr$resample_result(1)$prediction()$response, response)

  bmr$set_threshold(0.9, uhashes = uhashes(bmr, learner_ids = "classif.featureless"))

  expect_true(all(bmr$resample_result(1)$prediction()$response == "versicolor"))

  # can also use the iters argument
  design = benchmark_grid(
    task,
    c(lrn("classif.featureless", predict_type = "prob"), lrn("classif.debug", predict_type = "prob")),
    rsmp("insample")
  )
  bmr = benchmark(design)
  bmr$set_threshold(0.9, i = 1)
  expect_true(all(bmr$resample_result(1)$prediction()$response == "versicolor"))
  expect_false(all(bmr$resample_result(2)$prediction()$response == "versicolor"))
})

test_that("uhashe(s) work", {
  design = benchmark_grid(
    tsks(c("iris", "sonar")),
    lrns(c("classif.featureless", "classif.rpart")),
    rsmp("holdout")
  )
  bmr = benchmark(design)

  tbl = bmr$uhash_table
  expect_equal(bmr$uhashes, uhashes(bmr))
  expect_equal(tbl[get("learner_id") == "classif.debug", "uhash"]$uhash, uhashes(bmr, learner_ids = "classif.debug"))
  expect_equal(tbl[get("task_id") == "sonar", "uhash"]$uhash, uhashes(bmr, task_ids = "sonar"))
  expect_equal(tbl[get("resampling_id") == "holdout", "uhash"]$uhash, uhashes(bmr, resampling_ids = "holdout"))
  all_uhashes = bmr$uhashes
  expect_equal(length(all_uhashes), 4) # 2 tasks * 2 learners

  # Test filtering by single ID
  featureless_uhashes = uhashes(bmr, learner_ids = "classif.featureless")
  expect_equal(length(featureless_uhashes), 2)
  expect_true(all(featureless_uhashes %in% all_uhashes))

  iris_uhashes = uhashes(bmr, task_ids = "iris")
  expect_equal(length(iris_uhashes), 2)
  expect_true(all(iris_uhashes %in% all_uhashes))

  holdout_uhashes = uhashes(bmr, resampling_ids = "holdout")
  expect_equal(length(holdout_uhashes), 4)
  expect_true(all(holdout_uhashes %in% all_uhashes))

  # Test filtering by multiple IDs
  learner_subset = uhashes(bmr, learner_ids = c("classif.featureless", "classif.rpart"))
  expect_equal(length(learner_subset), 4)
  expect_setequal(learner_subset, all_uhashes)

  task_subset = uhashes(bmr, task_ids = c("iris", "sonar"))
  expect_equal(length(task_subset), 4)
  expect_setequal(task_subset, all_uhashes)

  # Test combined filtering
  iris_featureless = uhashes(bmr, learner_ids = "classif.featureless", task_ids = "iris")
  expect_equal(length(iris_featureless), 1)

  # Test uhash function with single valid combination
  single_uhash = uhash(bmr, learner_id = "classif.featureless", task_id = "iris", resampling_id = "holdout")
  expect_string(single_uhash)
  expect_true(single_uhash %in% all_uhashes)
  expect_error(uhash(bmr), "got 4")

  # no match
  expect_equal(uhashes(bmr, "not-existing"), character(0))
  expect_error(uhash(bmr, "not-existing"), "Expected exactly one uhash")
  expect_equal(bmr$uhashes, uhashes(bmr))
  expect_equal(bmr$filter(1)$uhashes, uhash(bmr))
})

test_that("resampling validation", {
  # test with uninstantiated resampling
  task = tsk("iris")
  learner = lrn("classif.rpart")
  resampling = rsmp("holdout")
  design = data.table(task = list(task), learner = list(learner), resampling = list(resampling))
  expect_error(benchmark(design), "instantiated")

  # test with resampling instantiated on wrong task
  task1 = tsk("iris")
  task2 = tsk("pima")
  resampling = rsmp("holdout")
  resampling$instantiate(task1)
  design = data.table(task = list(task2), learner = list(learner), resampling = list(resampling))
  expect_error(benchmark(design), "not instantiated")

  # test with resampling instantiated on filtered task
  task = tsk("iris")
  resampling = rsmp("holdout")
  resampling$instantiate(task)
  task$filter(1:100)
  design = data.table(task = list(task), learner = list(learner), resampling = list(resampling))
  expect_error(benchmark(design), "not instantiated")

  # test with resampling instantiated on correct task
  task = tsk("iris")
  resampling = rsmp("holdout")
  resampling$instantiate(task)
  design = data.table(task = list(task), learner = list(learner), resampling = list(resampling))
  expect_benchmark_result(benchmark(design))
})

test_that("warning when mixing predict types", {
  expect_warning(
    benchmark_grid(
      tsk("iris"),
      list(
        lrn("classif.debug", predict_type = "prob"),
        lrn("classif.featureless", predict_type = "response")
      ),
      rsmp("cv", folds = 3)
    ),
    regexp = "Multiple predict types detected"
  )
})

test_that("benchmark with tasks with weights", {
  learners = list(
    lrn("classif.featureless", use_weights = "ignore", predict_type = "prob", id = "ignores_weights"),
    lrn("classif.featureless", use_weights = "use", predict_type = "prob", id = "uses_weights")
  )
  measures = list(
    msr("classif.acc", use_weights = "ignore", id = "acc_ignore"),
    msr("classif.acc", use_weights = "use", id = "acc_use")
  )
  tasks = list(
    iris_weights_learner,
    iris_weights_measure,
    tsk("iris")
  )

  resamplings = list(
    rsmp("custom")$instantiate(tsk("iris"), train_sets = list(c(1:50, 140:150)), test_sets = list(c(1, 150)))
  )

  design = benchmark_grid(tasks, learners, resamplings)
  bmr = benchmark(design)

  predictions = map(1:6, function(i) bmr$resample_result(i)$prediction())

  # learner ignores weights, sees 50 'setosa' and 10 'virginica' -> predicts setosa
  expect_equal(predictions[[1]]$response[1], factor("setosa", levels = levels(iris$Species)))

  # learner uses weights, sees 50 'setosa' (weight 1) and 10 'virginica' (weight 100) -> predicts virginica
  expect_equal(predictions[[2]]$response[1], factor("virginica", levels = levels(iris$Species)))

  # task has no weights_learner -> unweighted 'setosa' predictions
  expect_equal(predictions[[3]]$response[1], factor("setosa", levels = levels(iris$Species)))
  expect_equal(predictions[[4]]$response[1], factor("setosa", levels = levels(iris$Species)))
  expect_equal(predictions[[5]]$response[1], factor("setosa", levels = levels(iris$Species)))
  expect_equal(predictions[[6]]$response[1], factor("setosa", levels = levels(iris$Species)))

  # 'weights' is NULL for all predictions not made for iris_weights_measure
  expect_null(predictions[[1]]$weights)
  expect_null(predictions[[2]]$weights)
  expect_null(predictions[[5]]$weights)
  expect_null(predictions[[6]]$weights)

  agpred = bmr$aggregate(measures)

  expect_equal(agpred$acc_ignore, rep(0.5, 6)) # made one correct, one incorrect prediction, unweighted

  # made one correct, one incorrect prediction, but 2nd task weighs the incorrect prediction x100
  expect_equal(agpred$acc_use, c(0.5, 0.5, 1 / 101, 1 / 101, 0.5, 0.5))

  expect_error(bmr$aggregate(msr("classif.acc", use_weights = "error")), "'use_weights' was set to 'error'")

  # no error when task has no weights_measure
  bmr = benchmark(design[c(1:2, 5:6)])
  expect_equal(bmr$aggregate(msr("classif.acc", use_weights = "error"))$classif.acc, c(0.5, 0.5, 0.5, 0.5))

  learners[[1]]$use_weights = "error"

  design = benchmark_grid(tasks, learners, resamplings)
  expect_error(benchmark(design), "'use_weights' was set to 'error'")

  # no error when task has no weights_learner
  design = benchmark_grid(tasks[2:3], learners, resamplings)
  bmr = benchmark(design)
})


test_that("obs_loss works", {
  bmr = benchmark(benchmark_grid(
    tsk("iris"),
    lrn("classif.rpart"),
    rsmp("cv", folds = 3)
  ))

  obs_loss = bmr$obs_loss(msr("classif.acc"))
  expect_data_table(obs_loss, nrows = 150L)
  expect_numeric(obs_loss$classif.acc, len = 150L)
  expect_set_equal(obs_loss$iteration, seq(3L))
  expect_set_equal(obs_loss$resample_result, 1L)

  obs_loss = bmr$obs_loss(msrs(c("classif.acc", "classif.auc")))
  expect_data_table(obs_loss, nrows = 150L)
  expect_numeric(obs_loss$classif.acc, len = 150L)
  expect_true(all(is.na(obs_loss$classif.auc)))
  expect_set_equal(obs_loss$iteration, seq(3L))
  expect_set_equal(obs_loss$resample_result, 1L)

  bmr = benchmark(benchmark_grid(
    tsk("iris"),
    lrn("classif.rpart"),
    c(rsmp("cv", folds = 3), rsmp("holdout"))
  ))

  obs_loss = bmr$obs_loss(msr("classif.acc"))
  expect_data_table(obs_loss, nrows = 200L)
  expect_numeric(obs_loss$classif.acc, len = 200L)
  expect_set_equal(obs_loss$iteration, seq(3L))
  expect_set_equal(obs_loss$resample_result, seq(2L))
})

test_that("MeasureRegrRSQ works", {
  measure = msr("regr.rsq")
  expect_equal(measure$properties, "weights")

  pred = PredictionRegr$new(truth = 0:10, response = 2:12, row_ids = seq(11))
  expect_equal(pred$score(measure), c(regr.rsq = 0.6))

  pred = PredictionRegr$new(truth = seq(0, 2, 0.5), response = seq(0, 2, 0.5), row_ids = seq(5))
  expect_equal(pred$score(measure), c(regr.rsq = 1.0))

  pred = PredictionRegr$new(truth = seq(1, 4), response = c(1, 2, 3, 5), row_ids = seq(4))
  expect_equal(pred$score(measure), c(regr.rsq = 0.8))

  measure = msr("regr.rsq", pred_set_mean = FALSE)
  expect_subset(measure$properties, c("requires_task", "requires_train_set", "weights"))

  pred = PredictionRegr$new(truth = 0:10, response = 2:12, row_ids = seq(11))
  task = as_task_regr(data.table(y = 0:10), target = "y")
  expect_equal(pred$score(measure, task = task, train_set = seq(11)), c(regr.rsq = 0.6))

  pred = PredictionRegr$new(truth = seq(0, 2, 0.5), response = seq(0, 2, 0.5), row_ids = seq(5))
  task = as_task_regr(data.table(y = seq(0, 2, 0.5)), target = "y")
  expect_equal(pred$score(measure, task = task, train_set = seq(5)), c(regr.rsq = 1.0))

  pred = PredictionRegr$new(truth = seq(1, 4), response = c(1, 2, 3, 5), row_ids = seq(4))
  task = as_task_regr(data.table(y = seq(1, 4)), target = "y")
  expect_equal(pred$score(measure, task = task, train_set = seq(4)), c(regr.rsq = 0.8))
})

test_that("MeasureRegrRSQ works with weights", {
  # Setup: truth, response, weights
  truth = 1:5
  response = c(1.1, 1.9, 3.2, 4.1, 4.8)
  weights = c(1, 2, 1, 2, 1)
  row_ids = 1:5

  pred = PredictionRegr$new(row_ids = row_ids, truth = truth, response = response, weights = weights)
  measure_weighted = msr("regr.rsq")
  measure_weighted$use_weights = "use" # explicitly set, although TRUE is default if 'weights' property exists
  expect_true("weights" %in% measure_weighted$properties)

  # Case 1: pred_set_mean = TRUE (default)
  # Manual calculation - mu now uses unweighted mean
  mu_pred = mean(truth) # No longer using weighted.mean()
  ss_res_pred = sum(weights * (truth - response)^2)
  ss_tot_pred = sum(weights * (truth - mu_pred)^2)
  rsq_expected_pred = 1 - ss_res_pred / ss_tot_pred
  expect_equal(pred$score(measure_weighted), c(regr.rsq = rsq_expected_pred))
  expect_equal(measure_weighted$score(pred), rsq_expected_pred)

  # Case 2: pred_set_mean = FALSE
  measure_train_mean = msr("regr.rsq", pred_set_mean = FALSE)
  measure_train_mean$use_weights = "use"
  expect_subset(c("requires_task", "requires_train_set", "weights"), measure_train_mean$properties)

  # Setup Task
  train_data = data.table(y = c(0, 8)) # mean = 4
  task = as_task_regr(train_data, target = "y")
  train_set_indices = 1:2 # Indices within the task data

  # Manual calculation - still uses unweighted mean for train set
  mu_train = mean(task$truth(train_set_indices)) # mean(c(0, 8)) = 4
  ss_res_train = ss_res_pred # Residuals don't depend on the mean used for SS_tot
  ss_tot_train = sum(weights * (truth - mu_train)^2)
  rsq_expected_train = 1 - ss_res_train / ss_tot_train
  expect_equal(
    pred$score(measure_train_mean, task = task, train_set = train_set_indices),
    c(regr.rsq = rsq_expected_train)
  )
  expect_equal(measure_train_mean$score(pred, task = task, train_set = train_set_indices), rsq_expected_train)

  # Case 3: Perfect fit
  pred_perfect = PredictionRegr$new(row_ids = row_ids, truth = truth, response = truth, weights = weights)
  expect_equal(pred_perfect$score(measure_weighted), c(regr.rsq = 1))
  expect_equal(measure_weighted$score(pred_perfect), 1)
  expect_equal(pred_perfect$score(measure_train_mean, task = task, train_set = train_set_indices), c(regr.rsq = 1))
  expect_equal(measure_train_mean$score(pred_perfect, task = task, train_set = train_set_indices), 1)

  # # Case 4: Constant truth (denominator = 0 -> NaN)
  ### this seems to be an edge case that does not need to be checked...
  pred_const = PredictionRegr$new(row_ids = row_ids, truth = rep(3, 5), response = response, weights = weights)
  # mu_const_pred = weighted.mean(pred_const$truth, weights) # 3
  # ss_tot_const_pred = sum(weights * (pred_const$truth - mu_const_pred)^2) # sum(weights * (3 - 3)^2) = 0
  # expect_equal(pred_const$score(measure_weighted), c(regr.rsq = -Inf))
  # # 1 - ss_res / 0 -> 1 - Inf -> -Inf? No, should be NaN
  # # R's 0/0 is NaN. If ss_res > 0, then pos/0 is Inf. 1 - Inf is -Inf.
  # # Let's check mlr3measures code (uses base::mean and base::weighted.mean)
  # # R: 1 - 5/0 = -Inf. The code uses sum((truth - response)^2) / sum((truth - mu)^2). So Inf. 1-Inf=-Inf.
  # # Let's test it:
  # # nolint next
  # expect_equal(1 - sum(weights * (pred_const$truth - pred_const$response)^2) / sum(weights * (pred_const$truth - mu_const_pred)^2), -Inf)
  # # The measure code directly calculates this, so it should be -Inf, not NaN
  # # Let's re-evaluate based on the definition provided in the help page:
  # # "This measure is undefined for constant t." -> NaN seems more appropriate user-facing value?
  # # Test what the code *actually* does:
  # expect_equal(measure_weighted$score(pred_const), -Inf)
  # expect_equal(pred_const$score(measure_weighted)[["regr.rsq"]], -Inf)

  # Use train mean = 4
  mu_const_train = 4
  # sum(weights * (3 - 4)^2) = sum(weights * (-1)^2) = sum(weights) = 7
  ss_tot_const_train = sum(weights * (pred_const$truth - mu_const_train)^2)
  rsq_expected_const_train = 1 - sum(weights * (pred_const$truth - pred_const$response)^2) / ss_tot_const_train
  expect_equal(
    pred_const$score(measure_train_mean, task = task, train_set = train_set_indices),
    c(regr.rsq = rsq_expected_const_train)
  )
  expect_equal(
    measure_train_mean$score(pred_const, task = task, train_set = train_set_indices),
    rsq_expected_const_train
  )
})

test_that("MeasureRegrRSQ works with weights during resampling", {
  task = cars_weights_measure$clone() # Use predefined task with 'weights_measure' role
  learner = lrn("regr.rpart")
  resampling = rsmp("cv", folds = 3)

  # Ensure the measure uses weights
  measure = msr("regr.rsq")
  measure$use_weights = "use"

  rr = resample(task, learner, resampling)

  # Check individual scores (pred_set_mean = TRUE implicitly)
  scores_manual = map_dbl(rr$predictions("test"), function(p) {
    w = p$weights
    t = p$truth
    r = p$response
    mu = mean(t)
    1 - sum(w * (t - r)^2) / sum(w * (t - mu)^2)
  })
  expect_equal(rr$score(measure)$regr.rsq, scores_manual)

  # Check aggregation - macro
  measure$average = "macro"
  aggr_macro = rr$aggregate(measure)
  expect_equal(aggr_macro[["regr.rsq"]], mean(scores_manual))

  # Check aggregation - micro
  measure$average = "micro"
  aggr_micro = rr$aggregate(measure)
  pred_combined = rr$prediction("test")
  w_comb = pred_combined$weights
  t_comb = pred_combined$truth
  r_comb = pred_combined$response
  mu_comb = mean(t_comb)
  rsq_micro_manual = 1 - sum(w_comb * (t_comb - r_comb)^2) / sum(w_comb * (t_comb - mu_comb)^2)
  expect_equal(aggr_micro[["regr.rsq"]], rsq_micro_manual)

  # Check aggregation - macro_weighted
  measure$average = "macro_weighted"
  aggr_macro_w = rr$aggregate(measure)
  weights_per_fold = map_dbl(rr$predictions("test"), function(p) sum(p$weights))
  rsq_macro_w_manual = weighted.mean(scores_manual, weights_per_fold)
  expect_equal(aggr_macro_w[["regr.rsq"]], rsq_macro_w_manual)

  # Check pred_set_mean = FALSE during resampling
  measure_train = msr("regr.rsq", pred_set_mean = FALSE)
  measure_train$use_weights = "use"
  rr_train_mean = resample(task, learner, resampling) # Need task for train mean

  scores_manual_train = map_dbl(seq_len(resampling$iters), function(i) {
    p = rr_train_mean$predictions()[[i]]
    train_set = rr_train_mean$resampling$train_set(i)
    w = p$weights
    t = p$truth
    r = p$response
    mu = mean(task$truth(train_set)) # Unweighted mean of train set truth
    1 - sum(w * (t - r)^2) / sum(w * (t - mu)^2)
  })
  expect_equal(rr_train_mean$score(measure_train)$regr.rsq, scores_manual_train)

  # Aggregation with pred_set_mean = FALSE
  measure_train$average = "macro"
  expect_equal(rr_train_mean$aggregate(measure_train)[["regr.rsq"]], mean(scores_manual_train))

  measure_train$average = "micro"
  expect_error(rr_train_mean$aggregate(measure_train), "requires the train_set")

  measure_train$average = "macro_weighted"
  weights_per_fold_train = map_dbl(rr_train_mean$predictions("test"), function(p) sum(p$weights))
  rsq_macro_w_manual_train = weighted.mean(scores_manual_train, weights_per_fold_train)
  expect_equal(rr_train_mean$aggregate(measure_train)[["regr.rsq"]], rsq_macro_w_manual_train)
})

test_that("MeasureRegrRSQ use_weights works", {
  task_plain = as_task_regr(cars, target = "dist")
  task_weighted = cars_weights_measure$clone()
  learner = lrn("regr.rpart")
  resampling = rsmp("holdout")$instantiate(task_plain)

  rr_plain = resample(task_plain, learner, resampling)
  rr_weighted = resample(task_weighted, learner, resampling)

  measure = msr("regr.rsq")

  # Default for measure with "weights" property is "use"
  expect_equal(measure$use_weights, "use")
  score_plain_use = rr_plain$aggregate(measure)
  score_weighted_use = rr_weighted$aggregate(measure)
  expect_false(isTRUE(all.equal(score_plain_use, score_weighted_use))) # Scores should differ

  # Ignore weights
  measure$use_weights = "ignore"
  score_plain_ignore = rr_plain$aggregate(measure)
  score_weighted_ignore = rr_weighted$aggregate(measure)
  expect_equal(score_plain_ignore, score_plain_use) # Ignoring weights on unweighted task = no change
  expect_equal(score_weighted_ignore, score_plain_use) # Ignoring weights should yield same as unweighted task

  # Error on weights
  measure$use_weights = "error"
  score_plain_error = rr_plain$aggregate(measure)
  expect_equal(score_plain_error, score_plain_use) # No error on unweighted task
  expect_error(rr_weighted$aggregate(measure), "since 'use_weights' was set to 'error'")
})

test_that("print.Measure", {
  m = mlr_measures$get("classif.ce")
  expect_output(print(Measure))
})

test_that("assert_measure", {
  task = tsk("iris")
  learner = lrn("classif.featureless")
  m = mlr_measures$get("time_train")
  expect_class(assert_measure(m), "Measure")
  expect_class(assert_measure(m, task = task), "Measure")
})

test_that("average with micro/macro", {
  task = tsk("german_credit")
  learner = lrn("classif.featureless")
  macro = msr("classif.fbeta", id = "macro")
  micro = msr("classif.fbeta", id = "micro", average = "micro")
  # nolint next
  with_seed(123, {
    rs = rsmp("cv", folds = 5)
    rr = resample(task, learner, rs)
  })
  aggr = rr$aggregate(list(macro, micro))

  expect_true(diff(aggr) > 0)
  expect_equal(aggr[[1]], mean(map_dbl(rr$predictions(), micro$score)))
  expect_equal(aggr[[1]], mean(map_dbl(rr$predictions(), macro$score)))
  expect_equal(aggr[[2]], micro$score(rr$prediction("test")))
  expect_equal(aggr[[2]], macro$score(rr$prediction("test")))
})

test_that("k is hashed in AIC", {
  m1 = msr("aic", k = 2)
  m2 = msr("aic", k = 3)

  expect_true(m1$hash != m2$hash)
})

test_that("can set beta for fbeta during construction", {
  # nolint next
  # https://stackoverflow.com/questions/66883242/how-to-change-beta-value-when-using-classif-fbeta-as-a-performance-measure-in

  task = tsk("sonar")
  l = lrn("classif.rpart")
  l$train(task)
  p = l$predict(task)

  m1 = msr("classif.fbeta", beta = 1)
  m2 = msr("classif.fbeta", beta = 10)
  expect_equal(m1$param_set$values$beta, 1)
  expect_equal(m2$param_set$values$beta, 10)

  expect_gt(abs(m1$score(p) - m2$score(p)), 0.0001)
})

test_that("check_prerequisites / task_properties", {
  task = tsk("penguins")
  learner = lrn("classif.featureless", predict_type = "prob")
  p = learner$train(task)$predict(task)
  m = msr("classif.auc")

  expect_error(unname(m$score(p)), "exactly")
  expect_error(unname(p$score(m)), "exactly")
  expect_warning(m$score(p, task = task), "properties", fixed = TRUE)
  expect_warning(p$score(m, task = task), "properties", fixed = TRUE)

  rr = resample(task, learner, rsmp("holdout"))
  expect_warning(res <- rr$score(m), "properties", fixed = TRUE)
  expect_identical(res$classif.auc, NaN)
  expect_warning(res <- rr$aggregate(m), "properties", fixed = TRUE)
  expect_identical(unname(res), NaN)

  bmr = as_benchmark_result(rr)
  expect_warning(res <- bmr$score(m), "properties", fixed = TRUE)
  expect_identical(res$classif.auc, NaN)
  expect_warning(res <- bmr$aggregate(m), "properties", fixed = TRUE)
  expect_identical(res$classif.auc, NaN)
})

test_that("check_prerequisites / predict_type", {
  task = tsk("sonar")
  learner = lrn("classif.featureless")
  p = learner$train(task)$predict(task)
  m = msr("classif.auc")

  expect_identical(unname(suppressWarnings(m$score(p))), NaN)
  expect_identical(unname(suppressWarnings(p$score(m))), NaN)
  expect_warning(m$score(p, learner = learner), "predict type", fixed = TRUE)
  expect_warning(p$score(m, learner = learner), "predict type", fixed = TRUE)

  rr = resample(task, learner, rsmp("holdout"))
  expect_warning(res <- rr$score(m), "predict type", fixed = TRUE)
  expect_identical(res$classif.auc, NaN)
  expect_warning(res <- rr$aggregate(m), "predict type", fixed = TRUE)
  expect_identical(unname(res), NaN)

  bmr = as_benchmark_result(rr)
  expect_warning(res <- bmr$score(m), "predict type", fixed = TRUE)
  expect_identical(res$classif.auc, NaN)
  expect_warning(res <- bmr$aggregate(m), "predict type", fixed = TRUE)
  expect_identical(res$classif.auc, NaN)
})

test_that("check_prerequisites / predict_sets", {
  task = tsk("sonar")
  learner = lrn("classif.featureless", predict_sets = "train")
  rr = resample(task, learner, rsmp("holdout"))
  m = msr("classif.ce")

  expect_warning(res <- rr$score(m), "predict sets", fixed = TRUE)
  expect_identical(res$classif.ce, NaN)
  expect_warning(res <- rr$aggregate(m), "predict sets", fixed = TRUE)
  expect_identical(unname(res), NaN)

  bmr = as_benchmark_result(rr)
  expect_warning(res <- bmr$score(m), "predict sets", fixed = TRUE)
  expect_identical(res$classif.ce, NaN)
  expect_warning(res <- bmr$aggregate(m), "predict set", fixed = TRUE)
  expect_identical(res$classif.ce, NaN)
})

test_that("time_train works with different predict type (#832)", {
  rr = resample(tsk("iris"), lrn("classif.debug", predict_type = "prob"), rsmp("holdout"))
  res = rr$score(msr("time_train"))
  expect_number(res$time_train)
})

test_that("time_train is > 0", {
  # no runtime test on CRAN
  skip_on_cran()

  rr = resample(tsk("iris"), lrn("classif.debug"), rsmp("holdout"))
  res = rr$score(msr("time_train"))
  expect_gte(res$time_train, 0)
})

test_that("scoring fails when measure requires_model, but model is in marshaled state", {
  measure = msr("classif.acc")
  measure$properties = c(measure$properties, "requires_model")

  task = tsk("iris")
  learner = lrn("classif.debug")
  pred = learner$train(task)$predict(task)
  learner$marshal()
  expect_error(measure$score(pred, learner = learner, task = task), regexp = "model is in marshaled")
})

test_that("measure weights", {
  learner = lrn("classif.featureless", predict_type = "prob")
  learner$train(tsk("iris"), row_ids = 1)
  prediction_no_weights = learner$predict(tsk("iris"), row_ids = c(1, 150))
  # should behave the same as no weights
  prediction_learner_weights = learner$predict(iris_weights_learner, row_ids = c(1, 150))
  prediction_measure_weights = learner$predict(iris_weights_measure, row_ids = c(1, 150))

  m = msr("classif.acc", use_weights = "use")
  expect_true("weights" %in% m$properties)

  # evaluating prediction with weights
  expect_equal(prediction_no_weights$score(m), c(classif.acc = 0.5))
  expect_equal(m$score(prediction_no_weights), 0.5)
  expect_equal(prediction_learner_weights$score(m), c(classif.acc = 0.5))
  expect_equal(m$score(prediction_learner_weights), 0.5)
  expect_equal(prediction_measure_weights$score(m), c(classif.acc = 1 / 101))
  expect_equal(m$score(prediction_measure_weights), 1 / 101)

  m$use_weights = "ignore"
  expect_equal(prediction_no_weights$score(m), c(classif.acc = 0.5))
  expect_equal(prediction_learner_weights$score(m), c(classif.acc = 0.5))
  expect_equal(prediction_measure_weights$score(m), c(classif.acc = 0.5))
  expect_equal(m$score(prediction_measure_weights), 0.5)

  m$use_weights = "error"
  expect_equal(prediction_no_weights$score(m), c(classif.acc = 0.5))
  expect_equal(prediction_learner_weights$score(m), c(classif.acc = 0.5))
  expect_error(prediction_measure_weights$score(m), "'use_weights' was set to 'error'")
  expect_error(m$score(prediction_measure_weights), "'use_weights' was set to 'error'")

  mauc = msr("classif.mauc_au1p")
  prediction_no_weights = learner$predict(tsk("iris"), row_ids = c(1, 2, 51, 52, 101, 102))
  # should behave the same as no weights
  prediction_learner_weights = learner$predict(iris_weights_learner, row_ids = c(1, 2, 51, 52, 101, 102))
  prediction_measure_weights = learner$predict(iris_weights_measure, row_ids = c(1, 2, 51, 52, 101, 102))

  expect_equal(prediction_no_weights$score(mauc), c(classif.mauc_au1p = 0.5))
  expect_equal(prediction_learner_weights$score(mauc), c(classif.mauc_au1p = 0.5))
  expect_error(prediction_measure_weights$score(mauc), "cannot be evaluated with weights")
  expect_error(mauc$score(prediction_measure_weights), "cannot be evaluated with weights")

  mauc$use_weights = "ignore"
  expect_equal(prediction_measure_weights$score(mauc), c(classif.mauc_au1p = 0.5))
  expect_equal(mauc$score(prediction_measure_weights), 0.5)
  expect_error(
    {
      mauc$use_weights = "use"
    },
    "Must be element of set"
  )

  # evaluating resampling with weights
  resampling = rsmp("custom")$instantiate(
    tsk("iris"),
    train_sets = list(1, 1),
    test_sets = list(c(1, 2, 51, 52, 101, 102), c(1:3, 51:53, 101:102))
  )
  rr_no_weights = resample(tsk("iris"), learner, resampling)
  rr_learner_weights = resample(iris_weights_learner, learner, resampling)
  rr_measure_weights = resample(iris_weights_measure, learner, resampling)

  m$use_weights = "use"
  expect_equal(rr_no_weights$score(m)$classif.acc, c(1 / 3, 3 / 8))
  expect_equal(rr_learner_weights$score(m)$classif.acc, c(1 / 3, 3 / 8))
  expect_equal(rr_measure_weights$score(m)$classif.acc, c(1 / 111, 3 / 233))

  m$use_weights = "ignore"
  expect_equal(rr_no_weights$score(m)$classif.acc, c(1 / 3, 3 / 8))
  expect_equal(rr_learner_weights$score(m)$classif.acc, c(1 / 3, 3 / 8))
  expect_equal(rr_measure_weights$score(m)$classif.acc, c(1 / 3, 3 / 8))

  m$use_weights = "error"
  expect_equal(rr_no_weights$score(m)$classif.acc, c(1 / 3, 3 / 8))
  expect_equal(rr_learner_weights$score(m)$classif.acc, c(1 / 3, 3 / 8))
  expect_error(rr_measure_weights$score(m), "since 'use_weights' was set to 'error'")

  mauc$use_weights = "ignore"
  expect_equal(rr_no_weights$score(mauc)$classif.mauc_au1p, c(1, 1) / 2)
  expect_equal(rr_learner_weights$score(mauc)$classif.mauc_au1p, c(1, 1) / 2)
  expect_equal(rr_measure_weights$score(mauc)$classif.mauc_au1p, c(1, 1) / 2)

  mauc$use_weights = "error"
  expect_equal(rr_no_weights$score(mauc)$classif.mauc_au1p, c(1, 1) / 2)
  expect_equal(rr_learner_weights$score(mauc)$classif.mauc_au1p, c(1, 1) / 2)
  expect_error(
    rr_measure_weights$score(mauc),
    "cannot be evaluated with weights in\n  .*Task.*since the Measure does not support\n  weights"
  )

  # aggregating resampling with weights
  m$use_weights = "use"
  expect_equal(rr_no_weights$aggregate(m), c(classif.acc = 1 / 6 + 3 / 16))
  expect_equal(m$aggregate(rr_no_weights), c(classif.acc = 1 / 6 + 3 / 16))
  expect_equal(rr_learner_weights$aggregate(m), c(classif.acc = 1 / 6 + 3 / 16))
  expect_equal(m$aggregate(rr_learner_weights), c(classif.acc = 1 / 6 + 3 / 16))
  expect_equal(rr_measure_weights$aggregate(m), c(classif.acc = 1 / 222 + 3 / 466))
  expect_equal(m$aggregate(rr_measure_weights), c(classif.acc = 1 / 222 + 3 / 466))

  m$use_weights = "ignore"
  expect_equal(rr_no_weights$aggregate(m), c(classif.acc = 1 / 6 + 3 / 16))
  expect_equal(rr_learner_weights$aggregate(m), c(classif.acc = 1 / 6 + 3 / 16))
  expect_equal(rr_measure_weights$aggregate(m), c(classif.acc = 1 / 6 + 3 / 16))
  expect_equal(m$aggregate(rr_measure_weights), c(classif.acc = 1 / 6 + 3 / 16))

  m$use_weights = "error"
  expect_equal(rr_no_weights$aggregate(m), c(classif.acc = 1 / 6 + 3 / 16))
  expect_equal(rr_learner_weights$aggregate(m), c(classif.acc = 1 / 6 + 3 / 16))
  expect_error(rr_measure_weights$aggregate(m), "since 'use_weights' was set to 'error'")
  expect_error(m$aggregate(rr_measure_weights), "since 'use_weights' was set to 'error'")

  mauc$use_weights = "ignore"
  expect_equal(rr_no_weights$aggregate(mauc), c(classif.mauc_au1p = 0.5))
  expect_equal(rr_learner_weights$aggregate(mauc), c(classif.mauc_au1p = 0.5))
  expect_equal(rr_measure_weights$aggregate(mauc), c(classif.mauc_au1p = 0.5))
  expect_equal(mauc$aggregate(rr_measure_weights), c(classif.mauc_au1p = 0.5))

  mauc$use_weights = "error"
  expect_equal(rr_no_weights$aggregate(mauc), c(classif.mauc_au1p = 0.5))
  expect_equal(rr_learner_weights$aggregate(mauc), c(classif.mauc_au1p = 0.5))
  expect_error(
    rr_measure_weights$aggregate(mauc),
    "cannot be evaluated with weights in\n  .*Task.*since the Measure does not support\n  weights"
  )
  expect_error(
    mauc$aggregate(rr_measure_weights),
    "cannot be evaluated with weights in\n  .*Task.*since the Measure does not support\n  weights"
  )

  m$use_weights = "use"
  m$average = "macro_weighted"
  expect_equal(rr_no_weights$aggregate(m), c(classif.acc = 5 / 14))
  expect_equal(m$aggregate(rr_no_weights), c(classif.acc = 5 / 14))
  expect_equal(rr_learner_weights$aggregate(m), c(classif.acc = 5 / 14))
  expect_equal(m$aggregate(rr_learner_weights), c(classif.acc = 5 / 14))
  expect_equal(rr_measure_weights$aggregate(m), c(classif.acc = 5 / 455))
  expect_equal(m$aggregate(rr_measure_weights), c(classif.acc = 5 / 455))

  m$use_weights = "ignore"
  expect_equal(rr_no_weights$aggregate(m), c(classif.acc = 5 / 14))
  expect_equal(rr_learner_weights$aggregate(m), c(classif.acc = 5 / 14))
  expect_equal(rr_measure_weights$aggregate(m), c(classif.acc = 5 / 14)) # weighs by number of samples
  expect_equal(m$aggregate(rr_measure_weights), c(classif.acc = 5 / 14))
})

test_that("primary iters are respected", {
  task = tsk("sonar")
  learner = lrn("classif.rpart", predict_type = "prob")
  resampling = rsmp("cv", folds = 10)
  resampling$instantiate(task)
  get_private(resampling, ".primary_iters") = 1:2

  rr = resample(task, learner, resampling)
  m = msr("classif.acc")
  m$average = "macro"
  scores = rr$score(m)$classif.acc

  # macro aggregation
  expect_equal(unname(rr$aggregate(m)), mean(scores[1:2]))
  expect_true(unname(rr$aggregate(m)) != mean(scores))

  # micro aggregation
  pred_micro = do.call(c, rr$predictions()[1:2])
  scores = pred_micro$score(m)
  expect_equal(unname(m$score(pred_micro)), unname(scores))

  jaccard = msr("sim.jaccard")
  expect_error(rr$aggregate(jaccard), "primary_iters")
})

test_that("no predict_sets required (#1094)", {
  m = msr("internal_valid_score")
  expect_null(m$predict_sets)
  rr = resample(tsk("iris"), lrn("classif.debug", validate = 0.3, predict_sets = NULL), rsmp("holdout"))
  expect_double(rr$aggregate(m))
  expect_warning(rr$aggregate(msr("classif.ce")), "needs predict sets")
})

test_that("checks on predict_sets", {
  m = msr("classif.ce")
  expect_error(
    {
      m$predict_sets = NULL
    },
    "Must be a subset"
  )
  expect_error(
    {
      m$predict_sets = "imaginary"
    },
    "Must be a subset"
  )
})

test_that("measure and prediction type is checked", {
  learner = lrn("classif.rpart")
  task = tsk("pima")
  learner$train(task)
  pred = learner$predict(task)

  measure = msr("classif.logloss")
  expect_warning(measure$score(pred), "is missing predict type")
})

test_that("obs_loss works", {
  MeasureTest = R6Class(
    "MeasureTest",
    inherit = Measure,
    public = list(
      initialize = function() {
        super$initialize("test", task_type = "classif", param_set = ps(), properties = "obs_loss")
      }
    ),
    private = list(
      .score = function(prediction, task, weights = NULL, ...) {
        2
      },

      .obs_loss = function(prediction, task, ...) {
        x = rnorm(length(prediction$row_ids))
        x - mean(x) + 2
      }
    )
  )

  measure = MeasureTest$new()
  learner = lrn("classif.rpart", predict_type = "prob")
  task = tsk("pima")
  learner$train(task)
  pred = learner$predict(task)
  obs_loss = measure$obs_loss(pred)
  expect_numeric(obs_loss, len = task$nrow, any.missing = FALSE)
})

test_that("Construction", {
  task = tsk("iris")
  p = PredictionClassif$new(row_ids = task$row_ids, truth = task$truth(), response = task$truth())
  expect_prediction(p)
  expect_prediction_classif(p)
})

test_that("Internally constructed Prediction", {
  task = tsk("iris")
  lrn = lrn("classif.featureless")
  lrn$predict_type = "prob"
  p = lrn$train(task)$predict(task)
  expect_prediction(p)
  expect_prediction_classif(p, task = task)

  p = PredictionClassif$new(row_ids = task$row_ids, truth = task$truth(), prob = p$prob, check = TRUE)
  expect_set_equal(p$predict_types, c("response", "prob"))
})

test_that("setting threshold binaryclass", {
  task = tsk("sonar")
  lrn = lrn("classif.rpart", predict_type = "prob")
  p = lrn$train(task)$predict(task)
  expect_factor(p$response, levels = task$class_names)
  expect_equal(as.character(p$response), colnames(p$prob)[max.col(p$prob)])

  response_before = p$response
  expect_false(withVisible(p$set_threshold(0.5))$visible) # 356
  expect_factor(p$response, levels = task$class_names, any.missing = FALSE)
  expect_equal(p$response, response_before)
  expect_lt(p$score(msr("classif.ce")), 0.25)

  p$set_threshold(0)
  expect_factor(p$response, levels = task$class_names, any.missing = FALSE)
  expect_true(all(as.character(p$response) == task$positive | p$prob[, task$positive] == 0))
  expect_gt(p$score(msr("classif.ce")), 0.25)

  p$set_threshold(1)
  expect_factor(p$response, levels = task$class_names, any.missing = FALSE)
  expect_true(all(as.character(p$response) == task$negative | p$prob[, task$negative] == 0))
  expect_gt(p$score(msr("classif.ce")), 0.25)
})

test_that("setting threshold multiclass", {
  task = tsk("zoo")
  lrn = lrn("classif.rpart", predict_type = "prob")
  p = lrn$train(task)$predict(task)

  # a small fix for our tests ... Add a small number to all probabilities so that
  # we can knock off single labels
  p$data$prob = t(apply(p$data$prob, 1, function(x) {
    x = x + 0.01
    x / sum(x)
  }))

  expect_factor(p$response, levels = task$class_names)
  expect_equal(as.character(p$response), colnames(p$prob)[max.col(p$prob)])

  prob_before = p$prob
  response_before = p$response

  expect_error(
    {
      p$set_threshold(c(0.5, 0.5))
    },
    "length"
  ) # check for correct length(threshold) = nclass

  x = p$set_threshold(set_names(c(1, 1, 1, 1, 1, 1, 1), task$class_names))
  expect_factor(x$response, levels = task$class_names, any.missing = FALSE)
  expect_equal(x$response, response_before)

  threshold = set_names(c(0, 1, 1, 1, 1, 1, 1), task$class_names)
  x = p$set_threshold(threshold)
  expect_factor(x$response, levels = task$class_names, any.missing = FALSE)
  expect_equal(as.character(unique(x$response)), task$class_names[1L])
})

test_that("setting threshold edge cases (#452)", {
  learner = lrn("classif.rpart", predict_type = "prob")
  t = tsk("iris")
  prd = learner$train(t)$predict(t$clone()$filter(c(1, 51, 101)))

  prd$set_threshold(c(setosa = 0, versicolor = 0, virginica = 0), ties_method = "first")
  expect_equal(as.character(prd$response), c("setosa", "versicolor", "versicolor"))
  prd$set_threshold(c(setosa = 0, versicolor = 0, virginica = 0), ties_method = "last")
  expect_equal(as.character(prd$response), c("setosa", "virginica", "virginica"))

  prd$set_threshold(c(setosa = 1, versicolor = 0, virginica = 0), ties_method = "first")
  expect_equal(as.character(prd$response), c("setosa", "versicolor", "versicolor"))
  prd$set_threshold(c(setosa = 1, versicolor = 0, virginica = 0), ties_method = "last")
  expect_equal(as.character(prd$response), c("setosa", "virginica", "virginica"))

  prd$set_threshold(c(setosa = 0, versicolor = 1, virginica = 0), ties_method = "first")
  expect_equal(as.character(prd$response), c("setosa", "virginica", "virginica"))
  prd$set_threshold(c(setosa = 0, versicolor = 1, virginica = 0), ties_method = "last")
  expect_equal(as.character(prd$response), c("setosa", "virginica", "virginica"))

  prd$set_threshold(c(setosa = 0, versicolor = 0, virginica = 1), ties_method = "first")
  expect_equal(as.character(prd$response), c("setosa", "versicolor", "versicolor"))
  prd$set_threshold(c(setosa = 0, versicolor = 0, virginica = 1), ties_method = "last")
  expect_equal(as.character(prd$response), c("setosa", "versicolor", "versicolor"))
})

test_that("confusion", {
  task = tsk("iris")
  lrn = lrn("classif.featureless")
  lrn$predict_type = "prob"
  p = lrn$train(task)$predict(task)
  cm = p$confusion

  expect_matrix(cm, nrows = 3, ncols = 3, any.missing = FALSE)
  expect_equal(colnames(p$confusion), task$class_names)
  expect_equal(rownames(p$confusion), task$class_names)
  expect_names(names(dimnames(cm)), identical.to = c("response", "truth"))
})

test_that("c", {
  task = tsk("iris")
  lrn = lrn("classif.featureless")
  lrn$predict_type = "prob"
  rr = resample(task, lrn, rsmp("cv", folds = 3))

  pred = do.call(c, rr$predictions())
  expect_prediction_classif(pred)

  dt = as.data.table(pred)
  expect_data_table(dt, nrows = task$nrow, ncols = 6L, any.missing = FALSE)

  conf = pred$confusion
  expect_equal(sum(conf), 150L)
  expect_equal(rownames(conf), task$class_names)
  expect_equal(colnames(conf), task$class_names)
  expect_equal(conf, Reduce("+", map(rr$predictions(), "confusion")))

  # duplicates are detected?
  p1 = get_private(rr)$.data$data$fact$prediction[[1]]$test
  p2 = get_private(rr)$.data$data$fact$prediction[[1]]$test
  p3 = c(p1, p2, keep_duplicates = FALSE)
  expect_equal(sort(p1$data$row_ids), sort(p2$data$row_ids))
  expect_equal(sort(p1$data$row_ids), sort(p3$data$row_ids))
  expect_factor(p3$response, len = length(p1$response), any.missing = FALSE)
  expect_matrix(p3$prob, nrows = nrow(p1$prob), ncols = ncol(p1$prob))
})

test_that("as_prediction_classif", {
  task = tsk("penguins")
  learner = lrn("classif.featureless", method = "weighted.sample")
  p = learner$train(task)$predict(task)

  tab = as.data.table(p)
  p2 = as_prediction_classif(tab)

  expect_equal(tab, as.data.table(p2))

  # issue #870
  tab = data.frame(row_ids = 1:3, truth = factor(c("a", "b", "a")), response = factor(c("a", "b", "b")))
  expect_class(as_prediction_classif(tab), "PredictionClassif")
})

test_that("#615", {
  task = tsk("iris")
  training = task$clone()$filter(1:100)
  testing = task$clone()$filter(101:150)

  l = lrn("classif.rpart")
  l$train(training)
  expect_equal(unname(l$predict(testing)$score()), 1)
})

test_that("filtering", {
  task = tsk("iris")
  p = PredictionClassif$new(row_ids = task$row_ids, truth = task$truth(), response = task$truth())

  p2 = p$clone()$filter(1:3)
  expect_set_equal(p$row_ids, 1:150)
  expect_set_equal(p2$row_ids, 1:3)
  expect_prediction(as_prediction_classif(as.data.table(p2)))
})

test_that("predictions with weights", {
  ll = lrn("classif.debug")$train(tsk("iris"), row_ids = 1)

  pred_with_weights = ll$predict(iris_weights_measure)
  expect_equal(pred_with_weights$weights, rep(c(1, 10, 100), each = 50))

  pred_without_weights = ll$predict(tsk("iris"))
  expect_null(pred_without_weights$weights)

  expect_error(c(pred_with_weights, pred_without_weights), "Some predictions have weights, others do not")

  expect_equal(c(pred_with_weights, pred_with_weights)$weights, rep(c(1, 10, 100, 1, 10, 100), each = 50))

  expect_equal(
    c(pred_with_weights$clone(deep = TRUE)$filter(1:10), pred_with_weights)$weights,
    rep(c(1, 10, 100, 1, 10, 100), each = 50)[c(1:10, 151:300)]
  )
})

test_that("extra data is stored", {
  LearnerExtra = R6Class(
    "LearnerExtra",
    inherit = LearnerClassifDebug,
    private = list(
      .predict = function(task, ...) {
        pred = super$.predict(task, ...)
        pred$extra = list(extra_col = replicate(length(pred$response), "bar"))
        pred
      }
    )
  )

  learner = LearnerExtra$new()
  learner$train(tsk("iris"))
  pred = learner$predict(tsk("iris"))
  expect_list(pred$extra, len = 1)
  expect_equal(pred$extra[[1]], replicate(length(pred$response), "bar"))
  expect_prediction(pred)

  LearnerExtra = R6Class(
    "LearnerExtra",
    inherit = LearnerClassifDebug,
    private = list(
      .predict = function(task, ...) {
        pred = super$.predict(task, ...)
        pred$extra = list(extra_col = replicate(2, "bar"))
        pred
      }
    )
  )

  learner = LearnerExtra$new()
  learner$train(tsk("iris"))
  expect_error(learner$predict(tsk("iris")), "Extra data must have the same length as the number of predictions")
})

test_that("raw data is stored", {
  task = tsk("iris")
  raw_obj = list(a = 1, b = "hello")
  p = PredictionClassif$new(row_ids = task$row_ids, truth = task$truth(), response = task$truth(), raw = raw_obj)
  expect_equal(p$raw, raw_obj)

  # raw survives filtering
  p2 = p$clone()$filter(1:3)
  expect_equal(p2$raw, raw_obj)

  # raw is NULL when not provided
  p3 = PredictionClassif$new(row_ids = task$row_ids, truth = task$truth(), response = task$truth())
  expect_null(p3$raw)
})

test_that("raw data is combined into list", {
  task = tsk("iris")
  p1 = PredictionClassif$new(row_ids = 1:3, truth = task$truth(1:3), response = task$truth(1:3), raw = list(x = 1))
  p2 = PredictionClassif$new(row_ids = 4:6, truth = task$truth(4:6), response = task$truth(4:6), raw = list(x = 2))
  combined = c(p1, p2)
  expect_list(combined$raw, len = 2)
  expect_equal(combined$raw[[1]], list(x = 1))
  expect_equal(combined$raw[[2]], list(x = 2))

  # combining with and without raw
  p3 = PredictionClassif$new(row_ids = 7:9, truth = task$truth(7:9), response = task$truth(7:9))
  combined2 = c(p1, p3)
  expect_list(combined2$raw, len = 1)
  expect_equal(combined2$raw[[1]], list(x = 1))

  # combining without any raw
  p4 = PredictionClassif$new(row_ids = 10:12, truth = task$truth(10:12), response = task$truth(10:12))
  combined3 = c(p3, p4)
  expect_null(combined3$raw)
})

test_that("raw data via learner predict", {
  LearnerRaw = R6Class(
    "LearnerRaw",
    inherit = LearnerClassifDebug,
    private = list(
      .predict = function(task, ...) {
        pred = super$.predict(task, ...)
        pred$raw = list(upstream_output = "raw_value")
        pred
      }
    )
  )

  learner = LearnerRaw$new()
  learner$train(tsk("iris"))
  pred = learner$predict(tsk("iris"))
  expect_equal(pred$raw, list(upstream_output = "raw_value"))
})

test_that("predict_raw with classif.rpart", {
  task = tsk("iris")
  learner = lrn("classif.rpart", predict_raw = TRUE)
  learner$train(task)
  pred = learner$predict(task)
  expect_class(pred$raw, "factor")
  expect_equal(length(pred$raw), task$nrow)

  learner = lrn("classif.rpart", predict_type = "prob", predict_raw = TRUE)
  learner$train(task)
  pred = learner$predict(task)
  expect_matrix(pred$raw, nrows = task$nrow)

  learner = lrn("classif.rpart")
  learner$train(task)
  pred = learner$predict(task)
  expect_null(pred$raw)
})

test_that("predict_raw with resample", {
  task = tsk("iris")
  learner = lrn("classif.rpart", predict_raw = TRUE)
  rr = resample(task, learner, rsmp("cv", folds = 3))
  pred = rr$prediction()
  expect_list(pred$raw, len = 3)
})

test_that("obs_loss works", {
  learner = lrn("classif.rpart", predict_type = "prob")
  task = tsk("pima")
  learner$train(task)
  prediction = learner$predict(task)
  # binary simple measure
  obs_loss = prediction$obs_loss(msr("classif.bbrier"))
  expect_data_table(obs_loss, nrows = task$nrow, any.missing = FALSE)
  expect_numeric(obs_loss$classif.bbrier, len = task$nrow, any.missing = FALSE)

  # classif simple measure
  obs_loss = prediction$obs_loss(msr("classif.logloss"))
  expect_data_table(obs_loss, nrows = task$nrow, any.missing = FALSE)
  expect_numeric(obs_loss$classif.logloss, len = task$nrow, any.missing = FALSE)

  obs_loss = prediction$obs_loss(msrs(c("classif.acc", "classif.logloss", "classif.auc")))
  expect_data_table(obs_loss, nrows = task$nrow)
  expect_numeric(obs_loss$classif.acc, len = task$nrow, any.missing = FALSE)
  expect_numeric(obs_loss$classif.logloss, len = task$nrow, any.missing = FALSE)
  expect_true(all(is.na(obs_loss$classif.auc)))
})

task = tsk("iris")
learner = lrn("classif.featureless")
resampling = rsmp("cv", folds = 3)
rr = resample(task, learner, resampling)

test_that("resample", {
  expect_resample_result(rr)

  scores = rr$score(msr("classif.ce"), predictions = TRUE)
  expect_list(scores$prediction_test, "Prediction")
  expect_numeric(scores$classif.ce, any.missing = FALSE)
  expect_number(rr$aggregate(msr("classif.ce")))
  learners = rr$learners
  expect_different_address(learners[[1L]], learners[[2L]])
  expect_equal(uniqueN(hashes(learners)), 1L)

  rr = rr$clone(TRUE)$filter(2:3)
  tab = as.data.table(rr)
  expect_data_table(tab, nrows = 2L)
  expect_data_table(tab, nrows = 2L)
  expect_equal(tab$iteration, 2:3)
  expect_resample_result(rr, allow_incomplete = TRUE)
})

test_that("empty RR", {
  rr = ResampleResult$new()
  expect_resample_result(rr)
})

test_that("resample with no or multiple measures", {
  for (measures in list(mlr_measures$mget(c("classif.ce", "classif.acc")), list())) {
    tab = rr$score(measures, ids = FALSE, predictions = TRUE)
    expect_data_table(
      tab,
      ncols = length(mlr_reflections$rr_names) + length(learner$predict_sets) + length(measures),
      nrows = 3L
    )
    expect_names(
      names(tab),
      permutation.of = c(mlr_reflections$rr_names, ids(measures), paste0("prediction_", learner$predict_sets))
    )
    perf = rr$aggregate(measures)
    expect_numeric(perf, any.missing = FALSE, len = length(measures), names = "unique")
    expect_names(names(perf), identical.to = unname(ids(measures)))
  }
})

test_that("as_benchmark_result.ResampleResult", {
  measures = list(msr("classif.ce"), msr("classif.acc"))
  bmr = as_benchmark_result(rr)
  expect_benchmark_result(bmr)
  expect_identical(nrow(get_private(bmr)$.data), nrow(get_private(rr)$.data))
  expect_set_equal(bmr$uhashes, rr$uhash)
  aggr = bmr$aggregate()
  expect_data_table(aggr, nrows = 1)
  expect_set_equal(bmr$uhashes, rr$uhash)
})


test_that("discarding model", {
  expect_equal(map(as.data.table(rr)$learner, "model"), vector("list", 3L))
})

test_that("inputs are cloned", {
  expect_different_address(task, get_private(rr)$.data$data$tasks$task[[1]])
  expect_different_address(learner, get_private(rr)$.data$data$learners$learner[[1]])
  expect_different_address(resampling, get_private(rr)$.data$data$resamplings$resampling[[1]])
})

test_that("memory footprint", {
  expect_shape(get_private(rr)$.data$data$learners, nrow = 1L)
  expect_shape(get_private(rr)$.data$data$tasks, nrow = 1L)
  expect_shape(get_private(rr)$.data$data$resamplings, nrow = 1L)
})

test_that("predict_type is checked", {
  task = tsk("sonar")
  learner = lrn("classif.featureless")
  resampling = rsmp("cv", folds = 3L)
  measure = msr("classif.auc")
  rr = resample(task, learner, resampling)

  expect_warning(rr$score(measure), "predict type", fixed = TRUE)
  expect_warning(rr$aggregate(measure), "predict type", fixed = TRUE)
})

test_that("empty train/predict sets", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")

  expect_error(learner$train(task, integer()))

  learner$train(task)
  expect_prediction(learner$predict(task, integer()))

  resampling = rsmp("holdout", ratio = 0)
  suppressWarnings(expect_error(resample(task, learner, resampling)))

  resampling = rsmp("holdout", ratio = 1)
  expect_prediction(resample(task, learner, resampling)$predictions()[[1]])
})

test_that("conditions are returned", {
  expect_true(all(c("warnings", "errors") %chin% names(rr$score(conditions = TRUE))))
})

test_that("save/load roundtrip", {
  path = tempfile()
  saveRDS(rr, file = path)

  rr2 = readRDS(path)
  expect_resample_result(rr2)
})

test_that("debug branch", {
  task = tsk("iris")
  learner = lrn("classif.featureless")
  resampling = rsmp("cv", folds = 2)
  rr = invoke(resample, task = task, learner = learner, resampling = resampling, .opts = list(mlr3.debug = TRUE))
  expect_resample_result(rr)
})

test_that("encapsulation", {
  task = tsk("iris")
  learner = lrn("classif.debug", error_train = 1)
  resampling = rsmp("holdout")

  suppressWarnings(expect_error(resample(task, learner, resampling), "classif.debug->train()"))

  rr = resample(task, learner, resampling, encapsulate = "evaluate")
  expect_data_table(rr$errors, nrows = 1L)
  expect_class(rr$learner$fallback, "LearnerClassifFeatureless")
  expect_equal(rr$learner$encapsulation[["train"]], "evaluate")
  expect_equal(rr$learner$encapsulation[["predict"]], "evaluate")
})

test_that("disable cloning", {
  task = tsk("iris")
  learner = lrn("classif.featureless")
  resampling = rsmp("holdout")

  rr = resample(task, learner, resampling, clone = c())

  expect_same_address(task, rr$task)
  expect_same_address(learner, get_private(rr)$.data$data$learners$learner[[1]])
  expect_same_address(resampling, rr$resampling)

  expect_identical(task$hash, rr$task$hash)
  expect_identical(learner$hash, rr$learner$hash)
  expect_true(resampling$is_instantiated)
  expect_identical(resampling$hash, rr$resampling$hash)
})

test_that("as_resample_result works for result data", {
  task = tsk("iris")
  learner = lrn("classif.featureless")
  resampling = rsmp("holdout")

  rr = resample(task, learner, resampling, clone = c())

  result_data = get_private(rr)$.data

  rr2 = as_resample_result(result_data)
  expect_class(rr2, "ResampleResult")
})

test_that("encapsulation triggers marshaling correctly", {
  learner1 = lrn("classif.debug", count_marshaling = TRUE)
  learner1$encapsulate("callr", fallback = lrn("classif.featureless"))
  learner2 = lrn("classif.debug", count_marshaling = TRUE)
  learner2$encapsulate("none")
  task = tsk("iris")
  resampling = rsmp("holdout")
  rr1 = resample(task, learner1, resampling, store_models = TRUE, unmarshal = FALSE)
  expect_true(rr1$learners[[1]]$marshaled)
  rr2 = resample(task, learner2, resampling, store_models = TRUE, unmarshal = FALSE)
  expect_true(rr2$learners[[1]]$marshaled)
})

test_that("parallel execution automatically triggers marshaling", {
  learner = lrn("classif.debug", count_marshaling = TRUE)
  task = tsk("iris")
  resampling = rsmp("holdout")
  rr = with_future(future::multisession, {
    resample(task, learner, resampling, store_models = TRUE, unmarshal = TRUE)
  })
  expect_equal(rr$learners[[1]]$model$marshal_count, 1)
  expect_false(rr$learners[[1]]$marshaled)
})

test_that("sequential execution does not trigger marshaling", {
  learner = lrn("classif.debug", count_marshaling = TRUE)
  task = tsk("iris")
  resampling = rsmp("holdout")
  rr = with_future(future::sequential, {
    resample(task, learner, resampling, store_models = TRUE, unmarshal = TRUE)
  })
  expect_equal(rr$learners[[1]]$model$marshal_count, 0)
})

test_that("parallel execution and callr marshal once", {
  learner = lrn("classif.debug", count_marshaling = TRUE)
  learner$encapsulate("callr", lrn("classif.featureless"))
  task = tsk("iris")
  resampling = rsmp("holdout")
  rr = with_future(future::multisession, {
    resample(task, learner, resampling, store_models = TRUE, unmarshal = TRUE)
  })
  expect_equal(rr$learners[[1]]$model$marshal_count, 1)
  expect_false(rr$learners[[1]]$marshaled)
})

test_that("marshaling works when store_models is FALSE", {
  learner = lrn("classif.debug", count_marshaling = FALSE)
  learner$encapsulate("callr", lrn("classif.featureless"))
  task = tsk("iris")
  resampling = rsmp("holdout")
  rr = with_future(future::multisession, {
    resample(task, learner, resampling, store_models = FALSE, unmarshal = TRUE)
  })
  expect_resample_result(rr)
  expect_null(rr$learners[[1]]$model)

  rr1 = with_future(future::sequential, {
    resample(task, learner, resampling, store_models = FALSE, unmarshal = TRUE)
  })
  expect_resample_result(rr1)
  expect_null(rr1$learners[[1]]$model)
})


test_that("unmarshal parameter is respected", {
  learner = lrn("classif.debug", count_marshaling = TRUE)
  learner$encapsulate("callr", lrn("classif.featureless"))
  task = tsk("iris")
  resampling = rsmp("holdout")
  rr = with_future(future::multisession, {
    list(
      marshaled = resample(task, learner, resampling, store_models = TRUE, unmarshal = FALSE),
      unmarshaled = resample(task, learner, resampling, store_models = TRUE, unmarshal = TRUE)
    )
  })
  expect_false(rr$unmarshaled$learners[[1]]$marshaled)
  expect_true(rr$marshaled$learners[[1]]$marshaled)
})

test_that("ResampleResult can be (un)marshaled", {
  rr = resample(tsk("iris"), lrn("classif.debug"), rsmp("holdout"), store_models = TRUE)
  expect_false(rr$learners[[1]]$marshaled)
  rr$marshal()
  expect_true(rr$learners[[1]]$marshaled)
  rr$unmarshal()
  expect_false(rr$learners[[1]]$marshaled)

  # also works with non-marshalable learner
  rr1 = resample(tsk("iris"), lrn("classif.featureless"), rsmp("holdout"), store_models = TRUE)
  model = rr1$learners[[1]]$model
  rr1$unmarshal()
  expect_equal(rr1$learners[[1]]$model, model)
})

test_that("does not unnecessarily clone state", {
  task = tsk("iris")
  learner = R6Class(
    "LearnerTest",
    inherit = LearnerClassifDebug,
    private = list(
      deep_clone = function(name, value) {
        if (name == "state" && !is.null(value)) {
          stop("Buggy bug bug")
        } else {
          super$deep_clone(name, value)
        }
      }
    )
  )$new()
  learner$train(task)
  expect_resample_result(resample(task, learner, rsmp("holdout")))
})

test_that("task hashes differ depending on whether test set is used", {
  task = tsk("iris")
  resampling = rsmp("holdout")
  learner1 = lrn("classif.debug")
  learner2 = lrn("classif.debug", validate = "test")
  rr1 = resample(task, learner1, resampling)
  rr2 = resample(task, learner2, resampling)
  expect_false(rr1$learners[[1]]$state$task_hash == rr2$learners[[1]]$state$task_hash)
})

test_that("can make predictions for internal_valid_task", {
  task = tsk("iris")
  learner = lrn("classif.debug", validate = 0.5, predict_sets = c("train", "internal_valid"))
  rr = resample(task, learner, rsmp("insample"))
  expect_equal(length(rr$predictions("internal_valid")[[1L]]$row_ids), task$nrow / 2)
})

test_that("learner's validate cannot be 'test' if internal_valid_set is present", {
  # otherwise, predict_set = "internal_valid" would be ambiguous
  learner = lrn("classif.debug", validate = "test", predict_sets = c("train", "internal_valid"))
  task = tsk("iris")
  task$internal_valid_task = 1
  expect_error(resample(task, learner, rsmp("holdout")), "cannot be set to ")
})

test_that("learner's validate cannot be a ratio if internal_valid_set is present", {
  # otherwise, predict_set = "internal_valid" would be ambiguous
  learner = lrn("classif.debug", validate = 0.5, predict_sets = c("train", "internal_valid"))
  task$internal_valid_task = 1
  expect_error(resample(task, learner, rsmp("holdout")), "cannot be set to")
})

test_that("internal_valid and train predictions", {
  task = tsk("iris")
  task$internal_valid_task = 1:2
  learner = lrn("classif.debug", validate = "predefined", predict_sets = c("train", "internal_valid", "test"))
  rr = resample(task, learner, rsmp("insample"))
  measure_valid = msr("classif.acc")
  measure_valid$predict_sets = "internal_valid"
  expect_equal(
    rr$score(measure_valid)$classif.acc,
    rr$learners[[1L]]$internal_valid_scores$acc
  )

  # if valid = "test", internal_valid and test predictions are the same
  task = tsk("iris")
  learner = lrn("classif.debug", validate = "test", predict_sets = c("train", "internal_valid", "test"))
  rr2 = resample(task, learner, rsmp("holdout"))

  expect_equal(
    rr2$score(measure_valid)$classif.acc,
    rr2$score(msr("classif.acc"))$classif.acc
  )
  expect_equal(
    rr2$predictions("internal_valid")[[1L]]$response,
    rr2$predictions("test")[[1L]]$response
  )

  # train predictions include the validation data
  learner = lrn("classif.debug", validate = 0.5, predict_sets = c("train", "internal_valid", "test"))
  task = tsk("iris")$filter(1:10)
  rr2 = resample(task, learner, rsmp("insample"))
  expect_equal(length(rr2$predictions("train")[[1L]]$row_ids), 10L)

  rr3 = resample(task, learner, rsmp("holdout", ratio = 0.8))

  expect_equal(length(rr3$predictions("train")[[1L]]$row_ids), 8L)
  expect_subset(
    rr3$predictions("internal_valid")$row_ids,
    rr3$predictions("internal_valid")$row_ids
  )
  expect_equal(length(rr3$predictions("internal_valid")[[1L]]$row_ids), 4L)
  expect_equal(length(rr3$predictions("test")[[1L]]$row_ids), 2L)
})

test_that("properties are also checked on validation task", {
  task = tsk("iris")
  row = task$data(1)
  row[[1]][1] = NA
  row$..row_id = 151
  task$rbind(row)
  task$internal_valid_task = 151
  learner = lrn("classif.debug", validate = "predefined")
  learner$properties = setdiff(learner$properties, "missings")

  suppressWarnings(expect_error(resample(task, learner, rsmp("holdout")), "missing values"))
})

test_that("marshaling does not change class of learner state when reassembling", {
  rr = resample(tsk("iris"), lrn("classif.debug"), rsmp("holdout"), encapsulate = "callr", store_models = TRUE)
  expect_class(rr$learners[[1]]$state, "learner_state")
})

test_that("marshaled model is sent back, when unmarshal is FALSE, sequential exec and callr", {
  learner = lrn("classif.debug", count_marshaling = TRUE)
  rr = resample(tsk("iris"), learner, rsmp("holdout"), store_models = TRUE, encapsulate = "callr", unmarshal = FALSE)
  expect_true(rr$learners[[1L]]$marshaled)
})

test_that("predict_set internal_valid throws error when none is available", {
  suppressWarnings(expect_error(
    resample(tsk("iris"), lrn("classif.debug", predict_sets = "internal_valid"), rsmp("holdout")),
    "Cannot set the predict_type"
  ))
})

test_that("can even use internal_valid predict set on learners that don't support validation", {
  task = tsk("mtcars")
  task$internal_valid_task = 1:10
  rr = resample(task, lrn("regr.debug", predict_sets = "internal_valid"), rsmp("holdout"))
  expect_warning(rr$score(), "predicted on sets 'internal_valid'")
})

test_that("callr during prediction triggers marshaling", {
  learner1 = lrn("classif.debug", count_marshaling = TRUE)
  learner1$encapsulate("callr", lrn("classif.featureless"))
  learner2 = lrn("classif.debug", count_marshaling = TRUE)
  learner2$encapsulate("callr", lrn("classif.featureless"))

  rr1 = with_future(future::multisession, {
    resample(tsk("iris"), learner1, rsmp("holdout"), unmarshal = FALSE, store_models = TRUE)
  })
  l1 = rr1$learners[[1L]]
  expect_true(l1$marshaled)
  expect_equal(l1$model$marshaled$marshal_count, 1L)

  rr2 = with_future(future::sequential, {
    resample(tsk("iris"), learner1, rsmp("holdout"), unmarshal = FALSE, store_models = TRUE)
  })
  l2 = rr2$learners[[1L]]
  expect_true(l2$marshaled)
  expect_equal(l2$model$marshaled$marshal_count, 1L)

  rr3 = with_future(future::multisession, {
    resample(tsk("iris"), learner2, rsmp("holdout"), unmarshal = FALSE, store_models = TRUE)
  })
  l3 = rr3$learners[[1L]]
  expect_true(l3$marshaled)
  expect_equal(l3$model$marshaled$marshal_count, 1L)

  rr4 = with_future(future::sequential, {
    resample(tsk("iris"), learner2, rsmp("holdout"), unmarshal = FALSE, store_models = TRUE)
  })
  l4 = rr4$learners[[1L]]
  expect_true(l4$marshaled)
  expect_equal(l4$model$marshaled$marshal_count, 1L)

  rr5 = with_future(future::multisession, {
    resample(tsk("iris"), learner1, rsmp("holdout"), unmarshal = TRUE, store_models = TRUE)
  })
  l5 = rr5$learners[[1L]]
  expect_false(l5$marshaled)
  expect_equal(l5$model$marshal_count, 1L)

  rr6 = with_future(future::sequential, {
    resample(tsk("iris"), learner1, rsmp("holdout"), unmarshal = TRUE, store_models = TRUE)
  })
  l6 = rr6$learners[[1L]]
  expect_false(l6$marshaled)
  expect_equal(l6$model$marshal_count, 1L)

  rr7 = with_future(future::multisession, {
    resample(tsk("iris"), learner2, rsmp("holdout"), unmarshal = TRUE, store_models = TRUE)
  })
  l7 = rr7$learners[[1L]]
  expect_false(l7$marshaled)
  expect_equal(l7$model$marshal_count, 1L)

  rr8 = with_future(future::sequential, {
    resample(tsk("iris"), learner2, rsmp("holdout"), unmarshal = TRUE, store_models = TRUE)
  })
  l8 = rr8$learners[[1L]]
  expect_false(l8$marshaled)
  expect_equal(l8$model$marshal_count, 1L)
})

test_that("multiple named measures", {
  rr = resample(tsk("iris"), lrn("classif.featureless"), rsmp("holdout"))
  res = rr$aggregate(c(acc = msr("classif.acc"), ce = msr("classif.ce")))
  expect_numeric(res[["classif.acc"]])
  expect_numeric(res[["classif.ce"]])
})

test_that("empty predictions", {
  rr = resample(tsk("iris"), lrn("classif.debug", validate = 0.3, predict_sets = NULL), rsmp("holdout"))
  preds = rr$predictions()
  expect_equal(preds, list(list()))
  pred = rr$prediction()
  expect_equal(pred, list())
})

test_that("resample result works with not predicted predict set", {
  learner = lrn("classif.debug", predict_sets = "train")
  task = tsk("iris")
  resampling = rsmp("holdout")

  rr = resample(task, learner, resampling)

  expect_list(rr$prediction(predict_sets = "test"), len = 0)
  expect_list(rr$predictions(predict_sets = "test"), len = 1)
  expect_list(rr$predictions(predict_sets = "test")[[1L]], len = 0)

  tab = as.data.table(rr)
  expect_list(tab$prediction, len = 1)
  expect_list(tab$prediction[[1]], len = 0)

  expect_warning(
    {
      tab = rr$score(msr("classif.ce", predict_sets = "test"))
    },
    "Measure"
  )
  expect_equal(tab$classif.ce, NaN)
})

test_that("resample results works with no predicted predict set", {
  learner = lrn("classif.debug", predict_sets = NULL)
  task = tsk("iris")
  resampling = rsmp("holdout")

  rr = resample(task, learner, resampling)

  expect_list(rr$prediction(predict_sets = "test"), len = 0)
  expect_list(rr$predictions(predict_sets = "test"), len = 1)
  expect_list(rr$predictions(predict_sets = "test")[[1L]], len = 0)

  tab = as.data.table(rr)
  expect_list(tab$prediction, len = 1)
  expect_list(tab$prediction[[1]], len = 0)

  expect_warning(
    {
      tab = rr$score(msr("classif.ce", predict_sets = "test"))
    },
    "Measure"
  )
  expect_equal(tab$classif.ce, NaN)
})

test_that("predict_time is 0 if no predict_set is specified", {
  learner = lrn("classif.featureless", predict_sets = NULL)
  rr = resample(task, learner, resampling)
  times = rr$score(msr("time_predict"))$time_predict
  expect_true(all(times == 0))
})

test_that("resampling instantiated on a different task throws an error", {
  task = tsk("spam")
  resampling = rsmp("cv", folds = 3)
  resampling$instantiate(task)

  expect_error(resample(tsk("pima"), lrn("classif.rpart"), resampling), "not instantiated")
})

test_that("resampling task row hash validation", {
  task = tsk("iris")
  resampling = rsmp("cv", folds = 3)
  resampling$instantiate(task)

  # Should work with same task
  expect_resample_result(resample(task, lrn("classif.rpart"), resampling))

  # Should fail if task is filtered
  task$filter(1:100)
  expect_error(resample(task, lrn("classif.rpart"), resampling), "not instantiated on task")
})

test_that("$score() checks for models", {
  rr = resample(tsk("mtcars"), lrn("regr.debug"), rsmp("holdout"))
  expect_error(rr$score(msr("aic")), "requires the trained model")
})

test_that("can change the threshold", {
  task = tsk("iris")$filter(1:80)$droplevels("Species")
  rr = resample(task, lrn("classif.featureless", predict_type = "prob"), rsmp("insample"))
  expect_true(all(rr$prediction()$response == "setosa"))
  rr$set_threshold(0.9)
  expect_true(all(rr$prediction()$response == "versicolor"))
  rr$set_threshold(0.1)
  expect_true(all(rr$prediction()$response == "setosa"))
  rr$set_threshold(0.625, ties_method = "first")
  expect_true(all(rr$prediction()$response == "setosa"))
  rr$set_threshold(0.625, ties_method = "last")
  expect_true(all(rr$prediction()$response == "versicolor"))
  # nolint next
  with_seed(1, {
    rr$set_threshold(0.625, ties_method = "random")
    expect_true("setosa" %in% rr$prediction()$response && "versicolor" %in% rr$prediction()$response)
  })
})

test_that("hashes work", {
  task = tsk("spam")
  learner = lrn("classif.debug")
  resampling = rsmp("cv", folds = 3)

  rr = resample(task, learner, resampling)
  task_hashes = map(rr$learners, function(learner) learner$state$task_hash)
  expect_length(unique(task_hashes), 3)
})

test_that("obs_loss works", {
  task = tsk("iris")
  learner = lrn("classif.rpart", predict_type = "prob")
  resampling = rsmp("cv", folds = 3)
  rr = resample(task, learner, resampling)

  obs_loss = rr$obs_loss(msr("classif.logloss"))
  expect_data_table(obs_loss, nrows = task$nrow)
  expect_numeric(obs_loss$classif.logloss, len = task$nrow, any.missing = FALSE)
  expect_set_equal(obs_loss$iteration, seq(3))

  obs_loss = rr$obs_loss(msrs(c("classif.acc", "classif.logloss", "classif.auc")))
  expect_data_table(obs_loss, nrows = task$nrow)
  expect_numeric(obs_loss$classif.acc, len = task$nrow, any.missing = FALSE)
  expect_numeric(obs_loss$classif.logloss, len = task$nrow, any.missing = FALSE)
  expect_true(all(is.na(obs_loss$classif.auc)))
})

test_that("learner methods", {
  learner = lrn("classif.debug")
  task = tsk("iris")
  expect_false(learner_marshaled(learner))

  learner$train(task)
  expect_false(learner_marshaled(learner))
  learner$marshal()
  expect_true(learner_marshaled(learner))
  learner$unmarshal()
  expect_false(learner_marshaled(learner))
})

test_that("default method does nothing", {
  x = 1
  expect_equal(x, marshal_model(x))
})

test_that("marshaling a marshaled object does nothing", {
  x = 1
  xm = marshal_model(x)
  expect_equal(marshal_model(xm), xm)
})

test_that("unmarshaling an unmarshaled object does nothing", {
  x = 1
  xm = marshal_model(x)
  expect_equal(unmarshal_model(xm), x)
  expect_equal(unmarshal_model(unmarshal_model(xm)), x)
})


test_that("marshaling for LearnerClassif", {
  task = tsk("iris")
  learner = lrn("classif.debug")
  learner$train(task)
  expect_false(learner$marshaled)
  learner$marshal()
  expect_true(learner$marshaled)
  learner$unmarshal()
  expect_learner(learner, task)
})

test_that("marshal count works for LearnerClassifDebug", {
  # to mock that marshaling behaves as expected, we need to be know how often it happened
  # note that this means that marshal_model modifies the model in a permanent way, i.e. this is not reversed by
  # unmarshal_model.
  learner = lrn("classif.debug", count_marshaling = TRUE)
  task = tsk("iris")
  learner$train(task)
  expect_equal(learner$model$marshal_count, 0)
  learner$marshal()$unmarshal()
  expect_equal(learner$model$marshal_count, 1)
  learner$marshal()$unmarshal()
  expect_equal(learner$model$marshal_count, 2)

  learner2 = lrn("classif.debug", count_marshaling = FALSE)
  learner2$train(task)
  expect_null(learner2$model$marshal_count)
  model1 = learner2$model
  model2 = learner2$marshal()$unmarshal()$model
  expect_equal(model1, model2)
})

test_that("printer", {
  expect_output(print(lrn("classif.debug")$train(tsk("iris"))$marshal()$model), "<classif.debug_model_marshaled>")
})

#' @title Convert to a Learner
#'
#' @description
#' Convert object to a [Learner] or a list of [Learner].
#'
#' @inheritParams as_task
#'
#' @return [Learner].
#' @export
# nolint next
as_learner = function(x, ...) {
  UseMethod("as_learner")
}

#' @export
#' @param discard_state (`logical(1)`)
#'   Whether to discard the state.
#' @rdname as_learner
# nolint next
as_learner.Learner = function(x, clone = FALSE, discard_state = FALSE, ...) {
  assert_empty_ellipsis(...)
  if (isTRUE(clone) && isTRUE(discard_state)) {
    clone_without(x, "state")
  } else if (isTRUE(clone)) {
    x$clone(deep = TRUE)
  } else if (isTRUE(discard_state)) {
    x$state = NULL
    x
  } else {
    x
  }
}

#' @export
#' @rdname as_learner
# nolint next
as_learners = function(x, ...) {
  UseMethod("as_learners")
}

#' @export
#' @rdname as_learner
# nolint next
as_learners.default = function(x, ...) {
  list(as_learner(x, ...))
}

#' @export
#' @rdname as_learner
# nolint next
as_learners.list = function(x, ...) {
  lapply(x, as_learner, ...)
}

#' @title Container for Benchmarking Results
#'
#' @include mlr_reflections.R
#'
#' @description
#' This is the result container object returned by [benchmark()].
#' A [BenchmarkResult] consists of the data of multiple [ResampleResult]s.
#' The contents of a `BenchmarkResult` and [ResampleResult] are almost identical
#' and the stored [ResampleResult]s can be extracted via the `$resample_result(i)` method,
#' where i is the index of the performed resample experiment.
#' This allows us to investigate the extracted [ResampleResult]
#' and individual resampling iterations, as well as the predictions and models from each fold.
#'
#' [BenchmarkResult]s can be visualized via \CRANpkg{mlr3viz}'s `autoplot()` function.
#'
#' For statistical analysis of benchmark results and more advanced plots, see \CRANpkg{mlr3benchmark}.
#'
#' @note
#' All stored objects are accessed by reference.
#' Do not modify any extracted object without cloning it first.
#'
#' @template param_measures
#'
#' @section S3 Methods:
#' * `as.data.table(rr, ..., reassemble_learners = TRUE, convert_predictions = TRUE,
#'   predict_sets = "test", task_characteristics = FALSE)`\cr
#'   [BenchmarkResult] -> [data.table::data.table()]\cr
#'   Returns a tabular view of the internal data.
#' * `c(...)`\cr
#'   ([BenchmarkResult], ...) -> [BenchmarkResult]\cr
#'   Combines multiple objects convertible to [BenchmarkResult] into a new [BenchmarkResult].
#'
#' @template seealso_benchmark
#' @export
#' @examples
#' set.seed(123)
#' learners = list(
#'   lrn("classif.featureless", predict_type = "prob"),
#'   lrn("classif.rpart", predict_type = "prob")
#' )
#'
#' design = benchmark_grid(
#'   tasks = list(tsk("sonar"), tsk("penguins")),
#'   learners = learners,
#'   resamplings = rsmp("cv", folds = 3)
#' )
#' print(design)
#'
#' bmr = benchmark(design)
#' print(bmr)
#'
#' bmr$tasks
#' bmr$learners
#'
#' # first 5 resampling iterations
#' head(as.data.table(bmr, measures = c("classif.acc", "classif.auc")), 5)
#'
#' # aggregate results
#' bmr$aggregate()
#'
#' # aggregate results with hyperparameters as separate columns
#' mlr3misc::unnest(bmr$aggregate(params = TRUE), "params")
#'
#' # extract resample result for classif.rpart
#' rr = bmr$aggregate()[learner_id == "classif.rpart", resample_result][[1]]
#' print(rr)
#'
#' # access the confusion matrix of the first resampling iteration
#' rr$predictions()[[1]]$confusion
#'
#' # reduce to subset with task id "sonar"
#' bmr$filter(task_ids = "sonar")
#' print(bmr)
BenchmarkResult = R6Class(
  "BenchmarkResult",
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param data (`ResultData`)\cr
    #'   An object of type `ResultData`, either extracted from another [ResampleResult], another
    #'   [BenchmarkResult], or manually constructed with [as_result_data()].
    initialize = function(data = NULL) {
      if (is.null(data)) {
        private$.data = ResultData$new()
      } else {
        private$.data = assert_class(data, "ResultData")
      }
    },

    #' @description
    #' Opens the help page for this object.
    help = function() {
      open_help("mlr3::BenchmarkResult")
    },

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s>", class(self)[1L])
    },

    #' @description
    #' Printer.
    print = function() {
      tab = self$aggregate(measures = list(), conditions = TRUE)
      set_data_table_class(tab)
      cat_cli(cli_h1(
        "{.cls {class(self)[1L]}} of {.val {private$.data$iterations()}} rows with {.val {nrow(tab)}} resampling run"
      ))
      if (nrow(tab)) {
        tab = remove_named(tab, c("uhash", "resample_result"))
        print(tab, class = FALSE, row.names = FALSE, print.keys = FALSE, digits = 3L)
      }
    },

    #' @description
    #' Fuses a second [BenchmarkResult] into itself, mutating the [BenchmarkResult] in-place.
    #' If the second [BenchmarkResult] `bmr` is `NULL`, simply returns `self`.
    #' Note that you can alternatively use the combine function [c()] which calls this method internally.
    #'
    #' @param bmr ([BenchmarkResult])\cr
    #'   A second [BenchmarkResult] object.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    combine = function(bmr) {
      if (!is.null(bmr)) {
        assert_benchmark_result(bmr)
        if (private$.data$iterations() && self$task_type != bmr$task_type) {
          error_input("BenchmarkResult is of task type '%s', but must be '%s'", bmr$task_type, self$task_type)
        }

        private$.data$combine(get_private(bmr)$.data)
      }

      invisible(self)
    },

    #' @description
    #' Marshals all stored models.
    #' @param ... (any)\cr
    #'   Additional arguments passed to [`marshal_model()`].
    #' @examples
    #' bmr$marshal()
    marshal = function(...) {
      private$.data$marshal(...)
    },
    #' @description
    #' Unmarshals all stored models.
    #' @param ... (any)\cr
    #'   Additional arguments passed to [`unmarshal_model()`].
    #' @examples
    #' bmr$unmarshal()
    unmarshal = function(...) {
      private$.data$unmarshal(...)
    },

    #' @description
    #' Returns a table with one row for each resampling iteration, including
    #' all involved objects: [Task], [Learner], [Resampling], iteration number
    #' (`integer(1)`), and [Prediction]. If `ids` is set to `TRUE`, character
    #' column of extracted ids are added to the table for convenient
    #' filtering: `"task_id"`, `"learner_id"`, and `"resampling_id"`.
    #'
    #' Additionally calculates the provided performance measures and binds the
    #' performance scores as extra columns. These columns are named using the id of
    #' the respective [Measure].
    #'
    #' @param ids (`logical(1)`)\cr
    #'   Adds object ids (`"task_id"`, `"learner_id"`, `"resampling_id"`) as
    #'   extra character columns to the returned table.
    #'
    #' @param conditions (`logical(1)`)\cr
    #'   Adds condition messages (`"warnings"`, `"errors"`) as extra
    #'   list columns of character vectors to the returned table
    #'
    #' @param predictions (`logical(1)`)\cr
    #'   Additionally return prediction objects, one column for each `predict_set` of all learners combined.
    #'   Columns are named `"prediction_train"`, `"prediction_test"` and `"prediction_internal_valid"`,
    #'   if present.
    #'
    #' @return [data.table::data.table()].
    #' @examples
    #' bmr$score(msr("classif.acc"))
    score = function(measures = NULL, ids = TRUE, conditions = FALSE, predictions = TRUE) {
      measures = assert_measures(as_measures(measures, task_type = self$task_type))
      assert_flag(ids)
      assert_flag(conditions)
      assert_flag(predictions)

      tab = score_measures(self, measures, view = NULL)
      tab = merge(private$.data$data$uhashes, tab, by = "uhash", sort = FALSE)
      tab[, "nr" := .GRP, by = "uhash"]

      if (ids) {
        set(tab, j = "task_id", value = ids(tab$task))
        set(tab, j = "learner_id", value = ids(tab$learner))
        set(tab, j = "resampling_id", value = ids(tab$resampling))
      }

      if (conditions) {
        set(tab, j = "warnings", value = map(tab$learner, "warnings"))
        set(tab, j = "errors", value = map(tab$learner, "errors"))
      }

      if (predictions && nrow(tab)) {
        predict_sets = intersect(
          mlr_reflections$predict_sets,
          unlist(map(self$learners$learner, "predict_sets"), use.names = FALSE)
        )
        predict_cols = sprintf("prediction_%s", predict_sets)
        for (i in seq_along(predict_sets)) {
          set(
            tab,
            j = predict_cols[i],
            value = map(tab$prediction, function(p) as_prediction(p[[predict_sets[i]]], check = FALSE))
          )
        }
      } else {
        predict_cols = character()
      }

      set_data_table_class(tab, "bmr_score")

      cns = c(
        "uhash",
        "nr",
        "task",
        "task_id",
        "learner",
        "learner_id",
        "resampling",
        "resampling_id",
        "iteration",
        predict_cols,
        "warnings",
        "errors",
        ids(measures)
      )
      cns = intersect(cns, names(tab))
      tab[, cns, with = FALSE]
    },

    #' @description
    #' Calculates the observation-wise loss via the [Measure]'s `obs_loss` method.
    #' Returns a `data.table()` with columns from the predictions (e.g., `row_ids`, `truth`, `response`, etc.),
    #' plus one numeric column for each measure, named with the respective measure id, and a `resample_result` column.
    #' If there is no observation-wise loss function for the measure, the column is filled with `NA_real_` values.
    #' Note that some measures such as RMSE, do have an `$obs_loss`,
    #' but they require an additional transformation after aggregation,
    #' in this example taking the square-root.
    #'
    #' @param predict_sets (`character()`)\cr
    #'   The predict sets.
    #' @examples
    #' bmr$obs_loss(msr("classif.acc"))
    obs_loss = function(measures = NULL, predict_sets = "test") {
      measures = assert_measures(as_measures(measures, task_type = self$task_type))
      map_dtr(
        self$resample_results$resample_result,
        function(rr) {
          rr$obs_loss(measures, predict_sets)
        },
        .idcol = "resample_result"
      )
    },

    #' @description
    #' Returns a result table where resampling iterations are combined into
    #' [ResampleResult]s. A column with the aggregated performance score is
    #' added for each [Measure], named with the id of the respective measure.
    #'
    #' The method for aggregation is controlled by the [Measure], e.g. micro
    #' aggregation, macro aggregation or custom aggregation. Most measures
    #' default to macro aggregation.
    #'
    #' Note that the aggregated performances just give a quick impression which
    #' approaches work well and which approaches are probably underperforming.
    #' However, the aggregates do not account for variance and cannot replace
    #' a statistical test.
    #' See \CRANpkg{mlr3viz} to get a better impression via boxplots or
    #' \CRANpkg{mlr3benchmark} for critical difference plots and
    #' significance tests.
    #'
    #' For convenience, different flags can be set to extract more
    #' information from the returned [ResampleResult].
    #'
    #' @param uhashes (`logical(1)`)\cr
    #'   Adds the uhash values of the [ResampleResult] as extra character
    #'   column `"uhash"`.
    #'
    #' @param ids (`logical(1)`)\cr
    #'   Adds object ids (`"task_id"`, `"learner_id"`, `"resampling_id"`) as
    #'   extra character columns for convenient subsetting.
    #'
    #' @param params (`logical(1)`)\cr
    #'   Adds the hyperparameter values as extra list column `"params"`. You
    #'   can unnest them with [mlr3misc::unnest()].
    #'
    #' @param conditions (`logical(1)`)\cr
    #'   Adds the number of resampling iterations with at least one warning as
    #'   extra integer column `"warnings"`, and the number of resampling
    #'   iterations with errors as extra integer column `"errors"`.
    #'
    #' @return [data.table::data.table()].
    #' @examples
    #' bmr$aggregate()
    aggregate = function(measures = NULL, ids = TRUE, uhashes = FALSE, params = FALSE, conditions = FALSE) {
      measures = if (is.null(measures)) {
        default_measures(self$task_type)
      } else {
        assert_measures(as_measures(measures))
      }
      assert_flag(ids)
      assert_flag(uhashes)
      assert_flag(params)
      assert_flag(conditions)

      create_rr = function(view) {
        if (length(view)) ResampleResult$new(private$.data, view = copy(view)) else list()
      }

      rdata = private$.data$data
      tab = rdata$fact[
        rdata$uhashes,
        list(
          nr = .GRP,
          iters = .N,
          task_hash = .SD$task_hash[1L],
          learner_hash = .SD$learner_hash[1L],
          learner_phash = .SD$learner_phash[1L],
          resampling_hash = .SD$resampling_hash[1L],
          resample_result = list(create_rr(.BY[[1L]])),
          warnings = if (conditions) {
            sum(map_int(.SD$learner_state, function(s) sum(s$log$class == "warning")))
          } else {
            NA_integer_
          },
          errors = if (conditions) {
            sum(map_int(.SD$learner_state, function(s) sum(s$log$class == "error")))
          } else {
            NA_integer_
          }
        ),
        by = "uhash",
        on = "uhash",
        nomatch = NULL
      ]

      if (ids) {
        tab = merge(
          tab,
          rdata$tasks[, list(task_hash = .SD$task_hash, task_id = ids(.SD$task))],
          by = "task_hash",
          sort = FALSE
        )
        tab = merge(
          tab,
          rdata$learners[, list(learner_phash = .SD$learner_phash, learner_id = ids(.SD$learner))],
          by = "learner_phash",
          sort = FALSE
        )
        tab = merge(
          tab,
          rdata$resamplings[, list(resampling_hash = .SD$resampling_hash, resampling_id = ids(.SD$resampling))],
          by = "resampling_hash",
          sort = FALSE
        )
      }

      if (!uhashes) {
        set(tab, j = "uhash", value = NULL)
      }

      if (params) {
        tab = merge(tab, rdata$learner_components, by = "learner_hash", sort = FALSE)
        setnames(tab, "learner_param_vals", "params")
      }

      if (!conditions) {
        tab = remove_named(tab, c("warnings", "errors"))
      }

      if (nrow(tab) > 0L) {
        scores = map_dtr(tab$resample_result, function(rr) as.list(resample_result_aggregate(rr, measures)))
      } else {
        scores = setDT(named_list(ids(measures), double()))
      }
      tab = insert_named(tab, scores)

      set_data_table_class(tab, "bmr_aggregate")

      cns = c(
        "uhash",
        "nr",
        "resample_result",
        "task_id",
        "learner_id",
        "resampling_id",
        "iters",
        "warnings",
        "errors",
        "params",
        names(scores)
      )
      cns = intersect(cns, names(tab))
      tab[, cns, with = FALSE]
    },

    #' @description
    #' Subsets the benchmark result.
    #' You can either directly provide the row IDs or the uhashes of the resample results to keep,
    #' or use the `learner_ids`, `task_ids` and `resampling_ids` arguments to filter for learner,
    #' task and resampling IDs.
    #' The three options are mutually exclusive.
    #'
    #' @param i (`integer()` | `NULL`)\cr
    #'   The iteration values to filter for.
    #' @param uhashes (`character()` | `NULL`)\cr
    #'   The uhashes of the resample results to filter for.
    #' @param learner_ids (`character()` | `NULL`)\cr
    #'   The learner IDs to filter for.
    #' @param task_ids (`character()` | `NULL`)\cr
    #'   The task IDs to filter for.
    #' @param resampling_ids (`character()` | `NULL`)\cr
    #'   The resampling IDs to filter for.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' design = benchmark_grid(
    #'   tsks(c("iris", "sonar")),
    #'   lrns(c("classif.debug", "classif.featureless")),
    #'   rsmp("holdout")
    #' )
    #' bmr = benchmark(design)
    #' bmr
    #' bmr2 = bmr$clone(deep = TRUE)
    #' bmr2$filter(learner_ids = "classif.featureless")
    #' bmr2
    filter = function(i = NULL, uhashes = NULL, learner_ids = NULL, task_ids = NULL, resampling_ids = NULL) {
      uhashes = private$.get_uhashes(i, uhashes, learner_ids, task_ids, resampling_ids)

      fact = private$.data$data$fact
      fact = if (is.null(uhashes)) {
        fact
      } else {
        fact[unique(uhashes), on = "uhash", nomatch = NULL]
      }

      setkeyv(fact, c("uhash", "iteration"))
      private$.data$data$fact = fact
      private$.data$sweep()

      invisible(self)
    },

    #' @description
    #' Retrieve the i-th [ResampleResult], by position, by unique hash `uhash` or by learner,
    #' task and resampling IDs.
    #' All three options are mutually exclusive.
    #'
    #' @param i (`integer(1)` | `NULL`)\cr
    #'   The iteration value to filter for.
    #' @param uhash (`character(1)` | `NULL`)\cr
    #'   The unique identifier to filter for.
    #' @param learner_id (`character(1)` | `NULL`)\cr
    #'   The learner ID to filter for.
    #' @param task_id (`character(1)` | `NULL`)\cr
    #'   The task ID to filter for.
    #' @param resampling_id (`character(1)` | `NULL`)\cr
    #'   The resampling ID to filter for.
    #'
    #' @examples
    #' design = benchmark_grid(
    #'   tsk("iris"),
    #'   lrns(c("classif.debug", "classif.featureless")),
    #'   rsmp("holdout")
    #' )
    #' bmr = benchmark(design)
    #' bmr$resample_result(learner_id = "classif.featureless")
    #' bmr$resample_result(i = 1)
    #' bmr$resample_result(uhash = uhashes(bmr, learner_id = "classif.debug"))
    #' @return [ResampleResult].
    resample_result = function(i = NULL, uhash = NULL, task_id = NULL, learner_id = NULL, resampling_id = NULL) {
      uhash = private$.get_uhashes(i, uhash, learner_id, task_id, resampling_id)
      if (length(uhash) != 1L) {
        error_input("Method requires selecting exactly one ResampleResult, but got %s", length(uhash))
      }
      ResampleResult$new(private$.data, view = uhash)
    },

    #' @description
    #' Shrinks the [BenchmarkResult] by discarding parts of the internally stored data.
    #' Note that certain operations might stop work, e.g. extracting
    #' importance values from learners or calculating measures requiring the task's data.
    #'
    #' @param backends (`logical(1)`)\cr
    #'   If `TRUE`, the [DataBackend] is removed from all stored [Task]s.
    #' @param models (`logical(1)`)\cr
    #'   If `TRUE`, the stored model is removed from all [Learner]s.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' bmr$discard(models = TRUE)
    discard = function(backends = FALSE, models = FALSE) {
      private$.data$discard(backends = backends, models = models)
    },

    #' @description
    #' Sets the threshold for the response prediction of classification learners, given they have
    #' output a probability prediction for a binary classification task.
    #'
    #' The resample results for which to change the threshold can either be specified directly
    #' via `uhashes`, by selecting the specific iterations (`i`) or by filtering according to
    #' learner, task and resampling IDs.
    #'
    #' If none of the three options is specified, the threshold is set for all resample results.
    #'
    #' @param threshold (`numeric(1)`)\cr
    #'   Threshold value.
    #' @param i (`integer()` | `NULL`)\cr
    #'   The iteration values to filter for.
    #' @param uhashes (`character()` | `NULL`)\cr
    #'   The unique identifiers of the [ResampleResult]s for which the threshold should be set.
    #' @param learner_ids (`character()` | `NULL`)\cr
    #'   The learner IDs for which the threshold should be set.
    #' @param task_ids (`character()` | `NULL`)\cr
    #'   The task IDs for which the threshold should be set.
    #' @param resampling_ids (`character()` | `NULL`)\cr
    #'   The resampling IDs for which the threshold should be set.
    #' @template param_ties_method
    #' @examples
    #' design = benchmark_grid(
    #'   tsk("sonar"),
    #'   lrns(c("classif.debug", "classif.featureless"), predict_type = "prob"),
    #'   rsmp("holdout")
    #' )
    #' bmr = benchmark(design)
    #' bmr$set_threshold(0.8, learner_ids = "classif.featureless")
    #' bmr$set_threshold(0.3, i = 2)
    #' bmr$set_threshold(0.7, uhashes = uhashes(bmr, learner_ids = "classif.featureless"))
    set_threshold = function(
      threshold,
      i = NULL,
      uhashes = NULL,
      learner_ids = NULL,
      task_ids = NULL,
      resampling_ids = NULL,
      ties_method = "random"
    ) {
      uhashes = private$.get_uhashes(i, uhashes, learner_ids, task_ids, resampling_ids)
      private$.data$set_threshold(uhashes, threshold, ties_method)
    }
  ),

  active = list(
    #' @field task_type (`character(1)`)\cr
    #' Task type of objects in the `BenchmarkResult`.
    #' All stored objects ([Task], [Learner], [Prediction]) in a single `BenchmarkResult` are
    #' required to have the same task type, e.g., `"classif"` or `"regr"`.
    #' This is `NA` for empty [BenchmarkResult]s.
    task_type = function(rhs) {
      assert_ro_binding(rhs)
      private$.data$task_type
    },

    #' @field tasks ([data.table::data.table()])\cr
    #' Table of included [Task]s with three columns:
    #'
    #' * `"task_hash"` (`character(1)`),
    #' * `"task_id"` (`character(1)`), and
    #' * `"task"` ([Task]).
    tasks = function(rhs) {
      assert_ro_binding(rhs)

      tab = private$.data$tasks()
      set(tab, j = "task_id", value = ids(tab$task))
      setcolorder(tab, c("task_hash", "task_id", "task"))[]
    },

    #' @field learners ([data.table::data.table()])\cr
    #' Table of included [Learner]s with three columns:
    #'
    #' * `"learner_hash"` (`character(1)`),
    #' * `"learner_id"` (`character(1)`), and
    #' * `"learner"` ([Learner]).
    #'
    #' Note that it is not feasible to access learned models via this field, as the training task would be ambiguous.
    #' For this reason the returned learners are reset before they are returned.
    #' Instead, select a row from the table returned by `$score()`.
    learners = function(rhs) {
      assert_ro_binding(rhs)

      tab = private$.data$learners(states = FALSE)
      set(tab, j = "learner_id", value = ids(tab$learner))
      setcolorder(tab, c("learner_hash", "learner_id", "learner"))[]
    },

    #' @field resamplings ([data.table::data.table()])\cr
    #' Table of included [Resampling]s with three columns:
    #'
    #' * `"resampling_hash"` (`character(1)`),
    #' * `"resampling_id"` (`character(1)`), and
    #' * `"resampling"` ([Resampling]).
    resamplings = function(rhs) {
      assert_ro_binding(rhs)

      tab = private$.data$resamplings()
      set(tab, j = "resampling_id", value = ids(tab$resampling))
      setcolorder(tab, c("resampling_hash", "resampling_id", "resampling"))[]
    },

    #' @field resample_results ([data.table::data.table()])\cr
    #' Returns a table with three columns:
    #' * `uhash` (`character()`).
    #' * `resample_result` ([ResampleResult]).
    resample_results = function(rhs) {
      assert_ro_binding(rhs)
      rdata = private$.data$data

      create_rr = function(view) {
        if (length(view)) ResampleResult$new(private$.data, view = copy(view)) else list()
      }
      rdata$fact[
        rdata$uhashes,
        list(
          nr = .GRP,
          resample_result = list(create_rr(.BY[[1L]]))
        ),
        by = "uhash"
      ]
    },

    #' @field n_resample_results (`integer(1)`)\cr
    #' Returns the total number of stored [ResampleResult]s.
    n_resample_results = function(rhs) {
      assert_ro_binding(rhs)
      length(private$.data$uhashes())
    },

    #' @field uhashes (`character()`)\cr
    #' Set of (unique) hashes of all included [ResampleResult]s.
    uhashes = function(rhs) {
      assert_ro_binding(rhs)
      private$.data$uhashes()
    },

    #' @field uhash_table ([data.table::data.table])\cr
    #' Table with columns `uhash`, `learner_id`, `task_id` and `resampling_id`.
    uhash_table = function(rhs) {
      assert_ro_binding(rhs)
      private$.data$uhash_table()
    }
  ),

  private = list(
    # @field data (`ResultData`)\cr
    # Internal data storage object of type `ResultData`.
    .data = NULL,

    .get_uhashes = function(i, uhashes, learner_ids, task_ids, resampling_ids) {
      args = discard(list(learner_ids = learner_ids, task_ids = task_ids, resampling_ids = resampling_ids), is.null)

      if (sum(!is.null(i), !is.null(uhashes), length(args) > 0L) > 1) {
        error_input("At most one of `i`, `uhash`, or IDs can be provided.")
      }
      if (!is.null(i)) {
        uhashes = self$uhashes
        uhashes = uhashes[assert_integerish(i, lower = 1L, upper = length(uhashes))]
      } else if (!is.null(uhashes)) {
        assert_subset(uhashes, self$uhashes)
      } else {
        uhashes = invoke(match.fun("uhashes"), bmr = self, .args = args)
      }
      if (length(uhashes) == 0L) {
        error_input("No resample results found for the given arguments.")
      }
      uhashes
    },

    deep_clone = function(name, value) {
      if (name == ".data") value$clone(deep = TRUE) else value
    }
  )
)

#' @export
# nolint next
as.data.table.BenchmarkResult = function(x, ..., hashes = FALSE, predict_sets = "test", task_characteristics = FALSE) {
  assert_flag(task_characteristics)
  tab = get_private(x)$.data$as_data_table(view = NULL, predict_sets = predict_sets)
  cns = c(
    "uhash",
    "task",
    "learner",
    "resampling",
    "iteration",
    "prediction",
    if ("data_extra" %in% names(tab)) "data_extra"
  )

  tab = tab[, cns, with = FALSE]

  set(tab, j = "task_id", value = ids(tab$task))
  set(tab, j = "learner_id", value = ids(tab$learner))
  set(tab, j = "resampling_id", value = ids(tab$resampling))

  if (task_characteristics) {
    set(tab, j = "characteristics", value = map(tab$task, "characteristics"))
    tab = unnest(tab, "characteristics")
  }

  tab[]
}

#' @export
# nolint next
c.BenchmarkResult = function(...) {
  bmrs = lapply(list(...), as_benchmark_result)
  init = BenchmarkResult$new()
  Reduce(function(lhs, rhs) lhs$combine(rhs), bmrs, init = init)
}

#' @export
print.bmr_score = function(x, ...) {
  print_data_table(x, c("uhash", "task", "learner", "resampling", "prediction"))
}

#' @export
print.bmr_aggregate = function(x, ...) {
  print_data_table(x, "resample_result")
}

#' @param learner_ids (`character()` | `NULL`)\cr
#'   Learner IDs.
#' @param task_ids (`character()` | `NULL`)\cr
#'   Task IDs.
#' @param resampling_ids (`character()` | `NULL`)\cr
#'   Resampling IDs.
#' @rdname uhash
#' @export
uhashes = function(bmr, learner_ids = NULL, task_ids = NULL, resampling_ids = NULL) {
  assert_class(bmr, "BenchmarkResult")
  assert_character(learner_ids, null.ok = TRUE)
  assert_character(task_ids, null.ok = TRUE)
  assert_character(resampling_ids, null.ok = TRUE)
  bmr$uhash_table[
    (is.null(learner_ids) | (get("learner_id") %in% learner_ids)) &
      (is.null(task_ids) | (get("task_id") %in% task_ids)) &
      (is.null(resampling_ids) | (get("resampling_id") %in% resampling_ids)),
    "uhash"
  ]$uhash
}

#' @title Obtain specific uhashes from a [BenchmarkResult]
#' @description
#' In a [`BenchmarkResult`], each [ResampleResult] is uniquely identified by a *hash* (*uhash*).
#' Operations that select specific [ResampleResult]s from a [BenchmarkResult] operate using
#' these hashes.
#' This function allows to obtain uhashes for specific learners, tasks, and resamplings.
#'
#' If you want more control, you can also directly obtain the uhash table from the [`BenchmarkResult`]
#' via the field `$uhash_table`.
#'
#' @param bmr (`BenchmarkResult`)\cr
#'   Benchmark result.
#' @param learner_id (`character(1)` | `NULL`)\cr
#'   Learner ID.
#' @param task_id (`character(1)` | `NULL`)\cr
#'   Task ID.
#' @param resampling_id (`character(1)` | `NULL`)\cr
#'   Resampling ID.
#' @export
#' @examples
#' design = benchmark_grid(
#'   tsks(c("sonar", "iris")),
#'   lrns(c("classif.debug", "classif.featureless", "classif.rpart")),
#'   rsmp("holdout")
#' )
#' bmr = benchmark(design)
#' bmr
#' bmr$uhashes
#' uhash(bmr, learner_id = "classif.debug", task_id = "sonar", resampling_id = "holdout")
#' uhashes(bmr, learner_ids = c("classif.debug", "classif.featureless"))
uhash = function(bmr, learner_id = NULL, task_id = NULL, resampling_id = NULL) {
  assert_string(learner_id, null.ok = TRUE)
  assert_string(task_id, null.ok = TRUE)
  assert_string(resampling_id, null.ok = TRUE)
  uhash = uhashes(bmr, learner_id, task_id, resampling_id)
  if (length(uhash) != 1L) {
    error_input("Expected exactly one uhash, got %s", length(uhash))
  }
  uhash
}

#' @title Pima Indian Diabetes Classification Task
#'
#' @name mlr_tasks_pima
#' @format [R6::R6Class] inheriting from [TaskClassif].
#' @include mlr_tasks.R
#'
#' @description
#' A classification task for the [mlbench::PimaIndiansDiabetes2] data set.
#' Positive class is set to `"pos"`.
#'
#' @templateVar id pima
#' @template task
#'
#' @template seealso_task
NULL

load_task_pima = function(id = "pima") {
  b = as_data_backend(load_dataset("PimaIndiansDiabetes2", "mlbench"))
  task = TaskClassif$new(id, b, target = "diabetes", positive = "pos", label = "Pima Indian Diabetes")
  b$hash = task$man = "mlr3::mlr_tasks_pima"
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("pima", load_task_pima)

#' @rdname PredictionData
#' @param train_task ([Task])\cr
#'   Task used for training the learner.
#' @export
check_prediction_data.PredictionDataClassif = function(pdata, train_task, ...) {
  # nolint next
  pdata$row_ids = assert_row_ids(pdata$row_ids)
  n = length(pdata$row_ids)
  assert_factor(pdata$truth, len = n, null.ok = TRUE)
  # unsupervised task
  if (is.null(pdata$truth)) {
    lvls = fget_key(train_task$col_info, train_task$target_names, "levels", "id")
    pdata$truth = if (length(pdata$row_ids)) factor(NA, lvls) else factor(levels = lvls)
  }
  lvls = levels(pdata$truth)

  if (!is.null(pdata$response)) {
    pdata$response = assert_factor(as_factor(unname(pdata$response), levels = lvls))
    assert_prediction_count(length(pdata$response), n, "response")
  }

  if (!is.null(pdata$prob)) {
    prob = assert_matrix(pdata$prob)
    assert_prediction_count(nrow(pdata$prob), n, "prob")
    assert_numeric(prob, lower = 0, upper = 1)
    assert_row_sums(prob)

    if (!identical(colnames(prob), lvls)) {
      assert_subset(colnames(prob), lvls)

      # add missing columns with prob == 0
      miss = setdiff(lvls, colnames(prob))
      if (length(miss)) {
        prob = cbind(prob, matrix(0, nrow = n, ncol = length(miss), dimnames = list(NULL, miss)))
      }

      # reorder columns to match the level order of `truth`
      prob = prob[, reorder_vector(colnames(prob), lvls), drop = FALSE]
    }

    if (!is.null(rownames(prob))) {
      rownames(prob) = NULL
    }

    pdata$prob = prob

    if (is.null(pdata$response)) {
      # calculate response from prob
      i = max.col(prob, ties.method = "random")
      pdata$response = factor(colnames(prob)[i], levels = lvls)
    }
  }

  if (!is.null(pdata$weights)) {
    # weights may never be NA, even if no prediction was made.
    pdata$weights = assert_numeric(unname(pdata$weights), any.missing = FALSE)
    assert_prediction_count(length(pdata$weights), n, "weights")
  }

  if (!is.null(pdata$extra)) {
    assert_list(pdata$extra, names = "unique")
    len = lengths(pdata$extra)
    if (any(len != n)) {
      error_learner_predict("Extra data must have the same length as the number of predictions")
    }
  }

  pdata
}


#' @rdname PredictionData
#' @export
# nolint next
is_missing_prediction_data.PredictionDataClassif = function(pdata, ...) {
  miss = logical(length(pdata$row_ids))
  if (!is.null(pdata$response)) {
    miss = is.na(pdata$response)
  }
  if (!is.null(pdata$prob)) {
    miss = miss | row_any_na(pdata$prob)
  }

  # weights may never be NA, so we don't need to check for missingness

  pdata$row_ids[miss]
}

#' @rdname PredictionData
#'
#' @param keep_duplicates (`logical(1)`)
#'   If `TRUE`, the combined [PredictionData] object is filtered for duplicated row ids (starting from last).
#' @param ... (one or more [PredictionData] objects).
#' @export
c.PredictionDataClassif = function(..., keep_duplicates = TRUE) {
  dots = list(...)
  assert_list(dots, "PredictionDataClassif")
  assert_flag(keep_duplicates)
  if (length(dots) == 1L) {
    return(dots[[1L]])
  }

  predict_types = names(mlr_reflections$learner_predict_types$classif)
  predict_types = map(dots, function(x) intersect(names(x), predict_types))
  if (!every(predict_types[-1L], setequal, y = predict_types[[1L]])) {
    error_input("Cannot rbind predictions: Different predict types")
  }

  if (length(unique(map_lgl(dots, function(x) is.null(x$weights)))) > 1L) {
    error_input("Cannot rbind predictions: Some predictions have weights, others do not")
  }

  if (length(unique(map_lgl(dots, function(x) is.null(x$extra)))) > 1L) {
    error_input("Cannot rbind predictions: Some predictions have extra data, others do not")
  }

  elems = c(
    "row_ids",
    "truth",
    intersect(predict_types[[1L]], "response"),
    if ("weights" %chin% names(dots[[1L]])) "weights"
  )
  tab = map_dtr(dots, function(x) x[elems], .fill = FALSE)
  prob = do.call(rbind, map(dots, "prob"))

  extra = if ("extra" %chin% names(dots[[1L]])) {
    rbindlist(map(dots, "extra"), fill = TRUE, use.names = TRUE)
  }

  if (!keep_duplicates) {
    keep = !duplicated(tab, by = "row_ids", fromLast = TRUE)
    tab = tab[keep]
    prob = prob[keep, , drop = FALSE]
    extra = extra[keep]
  }

  result = as.list(tab)
  if (!is.null(extra)) {
    result$extra = as.list(extra)
  }
  result$prob = prob

  raw = discard(map(dots, "raw"), is.null)
  if (length(raw)) {
    result$raw = raw
  }

  new_prediction_data(result, "classif")
}

#' @export
filter_prediction_data.PredictionDataClassif = function(pdata, row_ids, ...) {
  keep = pdata$row_ids %in% row_ids
  pdata$row_ids = pdata$row_ids[keep]
  pdata$truth = pdata$truth[keep]

  if (!is.null(pdata$response)) {
    pdata$response = pdata$response[keep]
  }

  if (!is.null(pdata$prob)) {
    pdata$prob = pdata$prob[keep, , drop = FALSE]
  }

  if (!is.null(pdata$weights)) {
    pdata$weights = pdata$weights[keep]
  }

  if (!is.null(pdata$extra)) {
    pdata$extra = map(pdata$extra, function(x) x[keep])
  }

  pdata
}

#' @export
create_empty_prediction_data.TaskClassif = function(task, learner) {
  predict_types = mlr_reflections$learner_predict_types[["classif"]][[learner$predict_type]]
  cn = task$class_names

  pdata = list(
    row_ids = integer(),
    truth = factor(character(), levels = cn)
  )

  if ("response" %chin% predict_types) {
    pdata$response = pdata$truth
  }

  if ("prob" %chin% predict_types) {
    pdata$prob = matrix(numeric(), nrow = 0L, ncol = length(cn), dimnames = list(NULL, cn))
  }

  if ("weights_measure" %chin% task$properties) {
    pdata$weights = numeric()
  }

  new_prediction_data(pdata, "classif")
}

#' @title Create a Data Backend
#'
#' @description
#' Wraps a [DataBackend] around data.
#' \CRANpkg{mlr3} ships with methods for `data.frame` (converted to a [DataBackendDataTable].
#'
#' Additional methods are implemented in the package \CRANpkg{mlr3db}, e.g. to connect
#' to real DBMS like PostgreSQL (via \CRANpkg{dbplyr}) or DuckDB (via \CRANpkg{DBI}/\CRANpkg{duckdb}).
#'
#' @param data (any)\cr
#'   Data to create a [DataBackend] from.
#'   For a `data.frame()` (this includes `tibble()` from \CRANpkg{tibble} and [data.table::data.table()]),
#'   a [DataBackendDataTable] is created.
#'
#' @template param_primary_key
#'
#' @param ... (any)\cr
#'   Additional arguments passed to the respective [DataBackend] method.
#'
#' @return [DataBackend].
#'
#' @template seealso_databackend
#' @export
#' @examples
#' # create a new backend using the penguins data:
#' as_data_backend(palmerpenguins::penguins)
as_data_backend = function(data, primary_key = NULL, ...) {
  UseMethod("as_data_backend")
}


#' @export
# nolint next
as_data_backend.DataBackend = function(data, primary_key = NULL, ...) {
  data
}


#' @param data ([data.frame()])\cr
#'   The input [data.frame()].
#'   Automatically converted to a [data.table::data.table()].
#'
#' @param keep_rownames (`logical(1)` | `character(1)`)\cr
#'   If `TRUE` or a single string, keeps the row names of `data` as a new column.
#'   The column is named like the provided string, defaulting to `"..rownames"` for `keep_rownames == TRUE`.
#'   Note that the created column will be used as a regular feature by the task
#'   unless you manually change the column role.
#'   Also see [data.table::as.data.table()].
#'
#' @rdname as_data_backend
#' @export
# nolint next
as_data_backend.data.frame = function(data, primary_key = NULL, keep_rownames = FALSE, ...) {
  assert_data_frame(data, min.cols = 1L, col.names = "unique")

  if (!isFALSE(keep_rownames)) {
    if (isTRUE(keep_rownames)) {
      keep_rownames = "..rownames"
    } else {
      assert_string(keep_rownames)
    }
  }

  data = as.data.table(data, keep.rownames = keep_rownames)
  compact_seq = FALSE

  if (is.character(primary_key)) {
    assert_string(primary_key)
    assert_choice(primary_key, colnames(data))
    assert_integer(data[[primary_key]], any.missing = FALSE, unique = TRUE, lower = 0L)
  } else {
    if (is.null(primary_key)) {
      row_ids = seq_row(data)
      compact_seq = TRUE
    } else if (is.integer(primary_key)) {
      row_ids = assert_integer(primary_key, len = nrow(data), any.missing = FALSE, unique = TRUE)
    } else {
      error_input("Argument 'primary_key' must be NULL, a column name or a vector of ids")
    }

    primary_key = "..row_id"
    data = insert_named(data, list("..row_id" = row_ids))
  }

  b = DataBackendDataTable$new(data, primary_key)
  b$compact_seq = compact_seq

  return(b)
}

test_that("TaskSupervised Construction", {
  b = as_data_backend(iris)
  task = TaskSupervised$new(id = "foo", "classif", b, target = "Species")
  b = task$backend
  expect_backend(b)
  expect_task(task)
  expect_task_supervised(task)
})

#' @title Regression Learner for Debugging
#'
#' @name mlr_learners_regr.debug
#' @include LearnerRegr.R
#'
#' @description
#' A simple [LearnerRegr] used primarily in the unit tests and for debugging purposes.
#' If no hyperparameter is set, it simply constantly predicts the mean value of the training data.
#' The following hyperparameters trigger the following actions:
#' \describe{
#'    \item{predict_missing:}{Ratio of predictions which will be NA.}
#'    \item{predict_missing_type:}{How to encode missingness.
#'      \dQuote{na} will insert NA values,
#'      \dQuote{omit} will just return fewer predictions than requested.}
#'    \item{save_tasks:}{Saves input task in `model` slot during training and prediction.}
#'    \item{threads:}{Number of threads to use. Has no effect.}
#'    \item{x:}{Numeric tuning parameter. Has no effect.}
#' }
#'
#' @templateVar id regr.debug
#' @template learner
#'
#' @template seealso_learner
#' @export
#' @examples
#' task = tsk("mtcars")
#' learner = lrn("regr.debug", save_tasks = TRUE)
#' learner$train(task, row_ids = 1:20)
#' prediction = learner$predict(task, row_ids = 21:32)
#'
#' learner$model$task_train
#' learner$model$task_predict
LearnerRegrDebug = R6Class(
  "LearnerRegrDebug",
  inherit = LearnerRegr,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      super$initialize(
        id = "regr.debug",
        feature_types = c("logical", "integer", "numeric", "character", "factor", "ordered"),
        predict_types = c("response", "se", "quantiles"),
        # fmt: skip
        param_set = ps(
          error_predict        = p_dbl(0, 1, default = 0, tags = "predict"),
          error_train          = p_dbl(0, 1, default = 0, tags = "train"),
          message_predict      = p_dbl(0, 1, default = 0, tags = "predict"),
          message_train        = p_dbl(0, 1, default = 0, tags = "train"),
          predict_missing      = p_dbl(0, 1, default = 0, tags = "predict"),
          predict_missing_type = p_fct(c("na", "omit"), default = "na", tags = "predict"),
          save_tasks           = p_lgl(default = FALSE, tags = c("train", "predict")),
          segfault_predict     = p_dbl(0, 1, default = 0, tags = "predict"),
          segfault_train       = p_dbl(0, 1, default = 0, tags = "train"),
          threads              = p_int(1L, tags = c("train", "threads")),
          warning_predict      = p_dbl(0, 1, default = 0, tags = "predict"),
          warning_train        = p_dbl(0, 1, default = 0, tags = "train"),
          x                    = p_dbl(0, 1, tags = "train")
        ),
        properties = c("missings", "weights"),
        packages = "stats",
        man = "mlr3::mlr_learners_regr.debug",
        label = "Debug Learner for Regression"
      )
    },

    #' @description
    #' Returns 0 for each feature seen in training.
    #' @return Named `numeric()`.
    importance = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      fns = self$state$feature_names
      set_names(rep(0, length(fns)), fns)
    },

    #' @description
    #' Always returns character(0).
    #' @return `character()`.
    selected_features = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      character(0)
    }
  ),
  private = list(
    .train = function(task) {
      pv = self$param_set$get_values(tags = "train")
      roll = function(name) {
        name %chin% names(pv) && pv[[name]] > runif(1L)
      }

      if (roll("message_train")) {
        message("Message from regr.debug->train()")
      }
      if (roll("warning_train")) {
        warning_mlr3("Warning from regr.debug->train()")
      }
      if (roll("error_train")) {
        error_learner_train("Error from regr.debug->train()")
      }
      if (roll("segfault_train")) {
        get("attach")(structure(list(), class = "UserDefinedDatabase"))
      }

      truth = task$truth()
      weights = private$.get_weights(task)
      wmd = weighted_mean_sd(truth, weights)
      model = list(
        response = wmd$mean,
        se = wmd$sd,
        pid = Sys.getpid()
      )

      if (self$predict_type == "quantiles") {
        assert_quantiles(self, quantile_response = TRUE)
        probs = self$quantiles
        model$quantiles = unname(quantile_weighted(truth, probs, weights = weights))
        model$quantile_probs = probs
      }

      if (isTRUE(pv$save_tasks)) {
        model$task_train = task$clone(deep = TRUE)
      }
      set_class(model, "regr.debug_model")
    },

    .predict = function(task) {
      n = task$nrow
      pv = self$param_set$get_values(tags = "predict")

      roll = function(name) {
        name %chin% names(pv) && pv[[name]] > runif(1L)
      }

      if (roll("message_predict")) {
        message("Message from regr.debug->predict()")
      }
      if (roll("warning_predict")) {
        warning_mlr3("Warning from regr.debug->predict()")
      }
      if (roll("error_predict")) {
        error_learner_predict("Error from regr.debug->predict()")
      }
      if (roll("segfault_predict")) {
        get("attach")(structure(list(), class = "UserDefinedDatabase"))
      }

      if (isTRUE(pv$save_tasks)) {
        self$state$model$task_predict = task$clone(deep = TRUE)
      }

      if (self$predict_type == "quantiles") {
        assert_quantiles(self, quantile_response = TRUE)
        prediction = list(
          quantiles = matrix(self$model$quantiles, nrow = n, ncol = length(self$model$quantiles), byrow = TRUE)
        )
        setattr(prediction$quantiles, "probs", self$model$quantile_probs)
        setattr(prediction$quantiles, "response", self$quantile_response)
        return(prediction)
      }

      predict_types = setdiff(self$predict_type, "quantiles")
      prediction = named_list(mlr_reflections$learner_predict_types[["regr"]][[predict_types]])
      missing_type = pv$predict_missing_type %??% "na"

      for (pt in names(prediction)) {
        value = rep.int(self$model[[pt]], n)
        if (!is.null(pv$predict_missing)) {
          ii = sample.int(n, n * pv$predict_missing)
          # fmt: skip
          value = switch(missing_type,
            "na" = replace(value, ii, NA),
            "omit" = value[ii]
          )
        }

        prediction[[pt]] = value
      }

      prediction
    }
  )
)

#' @include mlr_learners.R
mlr_learners$add("regr.debug", function() LearnerRegrDebug$new())

#' @title Dictionary of Performance Measures
#'
#' @usage NULL
#' @format [R6::R6Class] object inheriting from [mlr3misc::Dictionary].
#'
#' @description
#' A simple [mlr3misc::Dictionary] storing objects of class [Measure].
#' Each measure has an associated help page, see `mlr_measures_[id]`.
#'
#' This dictionary can get populated with additional measures by add-on packages.
#' E.g., \CRANpkg{mlr3proba} adds survival measures and \CRANpkg{mlr3cluster} adds
#' cluster analysis measures.
#'
#' For a more convenient way to retrieve and construct measures, see [msr()]/[msrs()].
#'
#' @section Methods:
#' See [mlr3misc::Dictionary].
#'
#' @section S3 methods:
#' * `as.data.table(dict, ..., objects = FALSE)`\cr
#'   [mlr3misc::Dictionary] -> [data.table::data.table()]\cr
#'   Returns a [data.table::data.table()] with fields "key", "label", "task_type", "packages",
#'   "predict_type", and "task_properties" as columns.
#'   If `objects` is set to `TRUE`, the constructed objects are returned in the list column named `object`.
#'
#' @family Dictionary
#' @family Measure
#' @seealso
#' Sugar functions: [msr()], [msrs()]
#'
#' Implementation of most measures: \CRANpkg{mlr3measures}
#' @export
#' @examples
#' as.data.table(mlr_measures)
#' mlr_measures$get("classif.ce")
#' msr("regr.mse")
mlr_measures = R6Class("DictionaryMeasure", inherit = Dictionary, cloneable = FALSE)$new()

#' @export
as.data.table.DictionaryMeasure = function(x, ..., objects = FALSE) {
  assert_flag(objects)

  setkeyv(
    map_dtr(
      x$keys(),
      function(key) {
        m = withCallingHandlers(x$get(key, .prototype = TRUE), packageNotFoundWarning = function(w) {
          invokeRestart("muffleWarning")
        })
        insert_named(
          list(
            key = key,
            label = m$label,
            task_type = m$task_type,
            packages = list(m$packages),
            predict_type = m$predict_type,
            properties = list(m$properties),
            task_properties = list(m$task_properties)
          ),
          if (objects) list(object = list(m))
        )
      },
      .fill = TRUE
    ),
    "key"
  )[]
}

#' @title Regression Task
#'
#' @include TaskSupervised.R
#'
#' @description
#' This task specializes [Task] and [TaskSupervised] for regression problems.
#' The target column is assumed to be numeric.
#' The `task_type` is set to `"regr"`.
#'
#' It is recommended to use [as_task_regr()] for construction.
#' Predefined tasks are stored in the [dictionary][mlr3misc::Dictionary] [mlr_tasks].
#'
#' @template param_rows
#' @template param_id
#' @template param_backend
#'
#' @template seealso_task
#' @export
#' @examples
#' task = as_task_regr(mtcars, target = "mpg")
#' task$task_type
#' task$formula()
#' task$truth()
#' task$data(rows = 1:3, cols = task$feature_names[1:2])
TaskRegr = R6Class(
  "TaskRegr",
  inherit = TaskSupervised,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #' The function [as_task_regr()] provides an alternative way to construct regression tasks.
    #'
    #' @template param_target
    #' @template param_label
    #' @template param_extra_args
    initialize = function(id, backend, target, label = NA_character_, extra_args = list()) {
      assert_string(target)
      super$initialize(
        id = id,
        task_type = "regr",
        backend = backend,
        target = target,
        label = label,
        extra_args = extra_args
      )

      type = fget_key(self$col_info, i = target, j = "type", key = "id")
      if (type %nin% c("integer", "numeric")) {
        error_input("Target column '%s' must be numeric", target)
      }
    },

    #' @description
    #' True response for specified `row_ids`. Format depends on the task type.
    #' Defaults to all rows with role "use".
    #' @return `numeric()`.
    truth = function(rows = NULL) {
      super$truth(rows)[[1L]]
    }
  )
)

#' @title Enable Validation for Learner
#'
#' @description
#' Helper function to configure the `$validate` field(s) of a [`Learner`].
#'
#' This is especially useful for learners such as `AutoTuner` of \CRANpkg{mlr3tuning}
#' or `GraphLearner` of \CRANpkg{mlr3pipelines}
#' which have multiple levels of `$validate` fields,
#' where the `$validate` fields need to be configured on multiple levels.
#'
#' @param learner (any)\cr
#'   The learner.
#' @param validate (`numeric(1)` | `"predefined"` | `"test"` | `NULL`)\cr
#'   Which validation set to use.
#' @param ... (any)\cr
#'   Additional arguments.
#'
#' @export
#' @return Modified [`Learner`]
#' @rdname mlr_sugar
#' @examples
#' learner = lrn("classif.debug")
#' set_validate(learner, 0.2)
#' learner$validate
set_validate = function(learner, validate, ...) {
  UseMethod("set_validate")
}

#' @export
set_validate.Learner = function(learner, validate, ...) {
  if (!"validation" %chin% learner$properties) {
    error_input("Learner '%s' does not support validation.", learner$id)
  }
  learner$validate = validate
  invisible(learner)
}

#' @title Subsampling Resampling
#'
#' @name mlr_resamplings_subsampling
#' @include Resampling.R
#'
#' @description
#' Splits data `repeats` (default: 30) times into training and test set
#' with a ratio of `ratio` (default: 2/3) observations going into the training set.
#'
#' @templateVar id subsampling
#' @template resampling
#'
#' @section Parameters:
#' * `repeats` (`integer(1)`)\cr
#'   Number of repetitions.
#' * `ratio` (`numeric(1)`)\cr
#'   Ratio of observations to put into the training set.
#'
#' @references
#' `r format_bib("bischl_2012")`
#'
#' @template seealso_resampling
#' @export
#' @examples
#' # Create a task with 10 observations
#' task = tsk("penguins")
#' task$filter(1:10)
#'
#' # Instantiate Resampling
#' subsampling = rsmp("subsampling", repeats = 2, ratio = 0.5)
#' subsampling$instantiate(task)
#'
#' # Individual sets:
#' subsampling$train_set(1)
#' subsampling$test_set(1)
#'
#' # Disjunct sets:
#' intersect(subsampling$train_set(1), subsampling$test_set(1))
#'
#' # Internal storage:
#' subsampling$instance$train # list of index vectors
ResamplingSubsampling = R6Class(
  "ResamplingSubsampling",
  inherit = Resampling,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        ratio = p_dbl(0, 1, tags = "required"),
        repeats = p_int(1L, tags = "required")
      )
      ps$set_values(repeats = 30L, ratio = 2 / 3)

      super$initialize(
        id = "subsampling",
        param_set = ps,
        label = "Subsampling",
        man = "mlr3::mlr_resamplings_subsampling"
      )
    }
  ),

  active = list(
    #' @template field_iters
    iters = function(rhs) {
      assert_ro_binding(rhs)
      as.integer(self$param_set$values$repeats)
    }
  ),

  private = list(
    .sample = function(ids, ...) {
      pv = self$param_set$values
      n = length(ids)
      nr = round(n * pv$ratio)

      train = replicate(pv$repeats, sample.int(n, nr), simplify = FALSE)
      list(train = train, row_ids = ids)
    },

    .get_train = function(i) {
      self$instance$row_ids[self$instance$train[[i]]]
    },

    .get_test = function(i) {
      self$instance$row_ids[-self$instance$train[[i]]]
    },

    .combine = function(instances) {
      Reduce(
        function(lhs, rhs) {
          n = length(lhs$row_ids)
          list(train = Map(function(x, y) c(x, y + n), lhs$train, rhs$train), row_ids = c(lhs$row_ids, rhs$row_ids))
        },
        instances
      )
    }
  )
)

#' @include mlr_resamplings.R
mlr_resamplings$add("subsampling", function() ResamplingSubsampling$new())

#' @title Resample Context
#'
#' @description
#' A [CallbackResample] accesses and modifies data during [resample()] and [benchmark()] via the `ContextResample`.
#' See the section on fields for a list of modifiable objects.
#' See [callback_resample()] for a list of stages that access `ContextResample`.
#'
#' @export
ContextResample = R6Class(
  "ContextResample",
  inherit = Context,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param task ([Task])\cr
    #'   The task to be evaluated.
    #' @param learner ([Learner])\cr
    #'   The learner to be evaluated.
    #' @param resampling ([Resampling])\cr
    #'   The resampling strategy to be used.
    #' @param iteration (`integer()`)\cr
    #'   The current iteration.
    initialize = function(task, learner, resampling, iteration) {
      # no assertions to avoid overhead
      private$.task = task
      private$.learner = learner
      private$.resampling = resampling
      private$.iteration = iteration

      super$initialize(id = "evaluate", label = "Evaluation")
    }
  ),

  active = list(
    #' @field task ([Task])\cr
    #' The task to be evaluated.
    #' The task is unchanged during the evaluation.
    #' The task is read-only.
    task = function(rhs) {
      assert_ro_binding(rhs)
      private$.task
    },

    #' @field learner ([Learner])\cr
    #' The learner to be evaluated.
    #' The learner contains the models after stage `on_resample_before_train`.
    learner = function(rhs) {
      if (missing(rhs)) {
        return(private$.learner)
      }
      private$.learner = assert_learner(rhs)
    },

    #' @field resampling [Resampling]\cr
    #' The resampling strategy to be used.
    #' The resampling is unchanged during the evaluation.
    #' The resampling is read-only.
    resampling = function(rhs) {
      assert_ro_binding(rhs)
      private$.resampling
    },

    #' @field iteration (`integer()`)\cr
    #' The current iteration.
    #' The iteration is read-only.
    iteration = function(rhs) {
      assert_ro_binding(rhs)
      private$.iteration
    },

    #' @field pdatas (List of [PredictionData])\cr
    #' The prediction data.
    #' The data is available on stage `on_resample_end`.
    pdatas = function(rhs) {
      if (missing(rhs)) {
        return(private$.pdatas)
      }
      private$.pdatas = assert_list(rhs, "PredictionData")
    },

    #' @field data_extra (list())\cr
    #' Data saved in the [ResampleResult] or [BenchmarkResult].
    #' Use this field to save results.
    #' Must be a `list()`.
    data_extra = function(rhs) {
      if (missing(rhs)) {
        return(private$.data_extra)
      }
      private$.data_extra = assert_list(rhs)
    }
  ),

  private = list(
    .task = NULL,
    .learner = NULL,
    .resampling = NULL,
    .iteration = NULL,
    .pdatas = NULL,
    .data_extra = NULL
  )
)

#' @title Wine Classification Task
#'
#' @name mlr_tasks_wine
#' @format [R6::R6Class] inheriting from [TaskClassif].
#' @include mlr_tasks.R
#'
#' @description
#' Wine data set from the UCI machine learning repository (\url{http://archive.ics.uci.edu/dataset/109/wine}).
#' Results of a chemical analysis of three types of wines grown in the same region in Italy
#' but derived from three different cultivars.
#'
#' @templateVar id wine
#' @template task
#'
#' @source
#' Original owners:
#' Forina, M. et al, PARVUS - An Extendible Package for Data Exploration, Classification and Correlation.
#' Institute of Pharmaceutical and Food Analysis and Technologies, Via Brigata Salerno, 16147 Genoa, Italy.
#'
#' Donor:
#' Stefan Aeberhard, email: stefan@coral.cs.jcu.edu.au
#'
#' @references
#' `r format_bib("dua_2017")`
#'
#' @template seealso_task
NULL

load_task_wine = function(id = "wine") {
  b = as_data_backend(readRDS(system.file("extdata", "wine.rds", package = "mlr3")))
  task = TaskClassif$new(id, b, target = "type", label = "Wine Regions")
  b$hash = task$man = "mlr3::mlr_tasks_wine"
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("wine", load_task_wine)

#' @title Supervised Task
#'
#' @include Task.R
#'
#' @description
#' This is the abstract base class for task objects like [TaskClassif] and [TaskRegr].
#' It extends [Task] with methods to handle a target column.
#' Supervised tasks for probabilistic regression (including survival analysis) can be
#' found in \CRANpkg{mlr3proba}.
#'
#' @template param_id
#' @template param_task_type
#' @template param_backend
#' @template param_rows
#' @template param_label
#' @template param_extra_args
#'
#' @template seealso_task
#' @keywords internal
#' @export
#' @examples
#' TaskSupervised$new("penguins", task_type = "classif", backend = palmerpenguins::penguins,
#'   target = "species")
TaskSupervised = R6Class(
  "TaskSupervised",
  inherit = Task,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param target (`character(1)`)\cr
    #'   Name of the target column.
    initialize = function(id, task_type, backend, target, label = NA_character_, extra_args = list()) {
      super$initialize(id = id, task_type = task_type, backend = backend, label = label, extra_args = extra_args)
      assert_subset(target, self$col_roles$feature)
      self$col_roles$target = target
      self$col_roles$feature = setdiff(self$col_roles$feature, target)
    },

    #' @description
    #' True response for specified `row_ids`. Format depends on the task type.
    #' Defaults to all rows with role "use".
    truth = function(rows = NULL) {
      self$data(rows, cols = self$target_names)
    }
  )
)

#' @title Akaike Information Criterion Measure
#'
#' @name mlr_measures_aic
#' @include Measure.R
#'
#' @description
#' Calculates the Akaike Information Criterion (AIC) which is a
#' trade-off between goodness of fit (measured in terms of
#' log-likelihood) and model complexity (measured in terms of number
#' of included features).
#' Internally, [stats::AIC()] is called with parameter `k` (defaulting to 2).
#' Requires the learner property `"loglik"`, `NA` is returned for unsupported learners.
#'
#' @templateVar id aic
#' @template measure
#'
#' @template seealso_measure
#' @export
MeasureAIC = R6Class(
  "MeasureAIC",
  inherit = Measure,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      param_set = ps(k = p_dbl(lower = 0))
      super$initialize(
        id = "aic",
        task_type = NA_character_,
        param_set = param_set,
        predict_sets = NULL,
        properties = c("na_score", "requires_learner", "requires_model", "requires_no_prediction"),
        predict_type = NA_character_,
        minimize = TRUE,
        label = "Akaike Information Criterion",
        man = "mlr3::mlr_measures_aic"
      )
    }
  ),

  private = list(
    .score = function(prediction, learner, ...) {
      learner = learner$base_learner()
      k = self$param_set$values$k %??% 2

      tryCatch(
        {
          stats::AIC(stats::logLik(learner$model), k = k)
        },
        error = function(e) {
          warning_config("Learner '%s' does not support AIC calculation", learner$id)
          NA_real_
        }
      )
    }
  )
)

#' @include mlr_measures.R
mlr_measures$add("aic", function() MeasureAIC$new())

#' @title Create a Fallback Learner
#'
#' @description
#' Create a fallback learner for a given learner.
#' The function searches for a suitable fallback learner based on the task type.
#' Additional checks are performed to ensure that the fallback learner supports the predict type.
#'
#' @param learner [Learner]\cr
#'  The learner for which a fallback learner should be created.
#' @param ... `any`\cr
#' ignored.
#'
#' @return [Learner]
#' @export
default_fallback = function(learner, ...) {
  UseMethod("default_fallback")
}

#' @rdname default_fallback
#' @export
default_fallback.Learner = function(learner, ...) {
  # FIXME: remove when new encapsulate/fallback system is in place
  NULL
}

#' @rdname default_fallback
#' @export
default_fallback.LearnerClassif = function(learner, ...) {
  fallback = lrn("classif.featureless")

  # set predict type
  if (learner$predict_type %nin% fallback$predict_types) {
    error_input("Fallback learner '%s' does not support predict type '%s'.", fallback$id, learner$predict_type)
  }

  fallback$predict_type = learner$predict_type

  fallback
}

#' @rdname default_fallback
#' @export
default_fallback.LearnerRegr = function(learner, ...) {
  fallback = lrn("regr.featureless")

  # set predict type
  if (learner$predict_type %nin% fallback$predict_types) {
    error_config("Fallback learner '%s' does not support predict type '%s'.", fallback$id, learner$predict_type)
  }

  fallback$predict_type = learner$predict_type

  # set quantiles
  if (learner$predict_type == "quantiles") {
    if (is.null(learner$quantiles) || is.null(learner$quantile_response)) {
      error_config(
        "Cannot set quantiles for fallback learner. Set `$quantiles` and `$quantile_response` in %s.",
        learner$id
      )
    }

    fallback$quantiles = learner$quantiles
    fallback$quantile_response = learner$quantile_response
  }

  fallback
}

learner_train = function(learner, task, train_row_ids = NULL, test_row_ids = NULL, mode = "train") {
  # This wrapper calls learner$.train, and additionally performs some basic
  # checks that the training was successful.
  # Exceptions here are possibly encapsulated, so that they get captured
  # and turned into log messages.
  train_wrapper = function(learner, task) {
    if (task$nrow == 0L) {
      error_input("Cannot %s Learner '%s' on task '%s': No observations", mode, learner$id, task$id)
    }

    model = if (mode == "train") {
      get_private(learner)$.train(task)
    } else if (mode == "hotstart") {
      get_private(learner)$.hotstart(task)
    }

    if (is.null(model)) {
      error_learner_train("Learner '%s' on task '%s' returned NULL during internal %s()", learner$id, task$id, mode)
    }

    # In order to avoid unnecessary (un-)marshaling steps,
    # we already extract the internal tuned values and validation scores here.
    # They should only operate on the model and the param_vals so the
    # information above should be enough.
    # In the future, we might want to refactor this, so the extractors get directly
    # called with the model and param_vals
    learner$state$model = model
    learner$state$param_vals = learner$param_set$values

    # Extract internal valid scores and tuned values if applicable.
    internal_valid_scores = if (
      !is.null(get0("validate", learner)) &&
        exists(".extract_internal_valid_scores", get_private(learner))
    ) {
      get_private(learner)$.extract_internal_valid_scores()
    }

    internal_tuned_values = if (exists(".extract_internal_tuned_values", get_private(learner))) {
      get_private(learner)$.extract_internal_tuned_values()
    }

    # extract out-of-bag error if supported
    oob_error = if (exists(".extract_oob_error", envir = get_private(learner))) {
      get_private(learner)$.extract_oob_error()
    }

    if (learner$encapsulation[["train"]] %in% c("callr", "mirai")) {
      model = marshal_model(model, inplace = TRUE)
    }

    list(
      model = model,
      internal_valid_scores = internal_valid_scores,
      internal_tuned_values = internal_tuned_values,
      oob_error = oob_error
    )
  }

  assert_choice(mode, c("train", "hotstart"))
  assert_task(task)
  assert_learner(learner)

  # ensure that required packages are installed
  require_namespaces(learner$packages)

  # subset to train set w/o cloning
  if (!is.null(train_row_ids)) {
    lg$debug(
      "Subsetting task '%s' to %i rows",
      task$id,
      length(train_row_ids),
      task = task$clone(),
      row_ids = train_row_ids
    )

    task_private = get_private(task)
    prev_use = task_private$.row_roles$use
    on.exit(
      {
        task_private$.row_roles$use = prev_use
      },
      add = TRUE
    )
    task_private$.row_roles$use = train_row_ids
    task_private$.row_hash = NULL
    task_private$.hash = NULL
  } else {
    lg$debug("Skip subsetting of task '%s'", task$id)
  }

  # handle the internal validation task
  validate = get0("validate", learner)
  prev_valid = task$internal_valid_task
  on.exit(
    {
      task$internal_valid_task = prev_valid
    },
    add = TRUE
  )

  # depending on the validate parameter, create the internal validation task (if needed)
  # modifies the task in place
  create_internal_valid_task(validate, task, test_row_ids, prev_valid, learner)
  if (!is.null(task$internal_valid_task) && !task$internal_valid_task$nrow) {
    error_config("Internal validation task for task '%s' has 0 observations", task$id)
  }

  if (mode == "train") {
    learner$state = list()
  }

  lg$debug(
    "Calling %s method of Learner '%s' on task '%s' with %i observations",
    mode,
    learner$id,
    task$id,
    task$nrow,
    learner = learner$clone()
  )

  deadline_time_left = max(as.numeric(difftime(learner$deadline[["train"]], Sys.time(), units = "secs")), 0)

  # call train_wrapper with encapsulation
  result = encapsulate(
    learner$encapsulation["train"],
    .f = train_wrapper,
    .args = list(learner = learner, task = task),
    .pkgs = c(learner$packages, mlr_reflections$loaded_packages),
    .seed = NA_integer_,
    .timeout = min(learner$timeout["train"], deadline_time_left),
    .compute = getOption("mlr3.mirai_encapsulation", "mlr3_encapsulation")
  )
  cond = cond_from_log(result$log)

  # We still log the warnings and messages in the case of an error,
  # so we don't throw immediately
  err = learner_will_err(cond, learner, stage = "train")

  # we don't log an existing (uncaught) error, because it's signalled
  log = append_log(NULL, stage = "train", class = result$log$class, condition = result$log$condition, log_error = !err)

  if (err) {
    stop(cond)
  }

  train_time = result$elapsed

  learner$state = set_class(
    insert_named(
      learner$state,
      list(
        model = result$result$model,
        log = log,
        train_time = train_time,
        param_vals = learner$param_set$values,
        task_hash = task$hash,
        feature_names = task$feature_names,
        validate = get0("validate", learner),
        mlr3_version = mlr_reflections$package_version
      )
    ),
    c("learner_state", "list")
  )

  # store the results of the internal tuning / internal validation in the learner's state
  # otherwise this information is only available with store_models = TRUE
  if (!is.null(result$result$internal_valid_scores)) {
    learner$state$internal_valid_scores = result$result$internal_valid_scores
    learner$state$internal_valid_task_hash = task$internal_valid_task$hash
  }

  learner$state$internal_tuned_values = result$result$internal_tuned_values
  learner$state$oob_error = result$result$oob_error

  if (is.null(result$result$model)) {
    lg$info(
      "Learner '%s' on task '%s' failed to %s a model",
      learner$id,
      task$id,
      mode,
      learner = learner$clone(),
      messages = result$log$msg
    )
  } else {
    lg$debug(
      "Learner '%s' on task '%s' succeeded to %s a model",
      learner$id,
      task$id,
      mode,
      learner = learner$clone(),
      result = result$result$model,
      messages = result$log$msg
    )
  }

  # fit fallback learner
  fb = learner$fallback
  if (!is.null(fb)) {
    lg$info(
      "Calling train method of fallback '%s' on task '%s' with %i observations",
      fb$id,
      task$id,
      task$nrow,
      learner = fb$clone()
    )

    fb = assert_learner(as_learner(fb))
    require_namespaces(fb$packages)
    fb$train(task)
    learner$state$fallback_state = fb$state

    lg$debug("Fitted fallback learner '%s'", fb$id, learner = fb$clone())
  }

  list(
    learner = learner,
    internal_valid_task_ids = if (!is.null(validate)) task$internal_valid_task$row_ids
  )
}

learner_predict = function(learner, task, row_ids = NULL) {
  # This wrapper calls learner$predict, and additionally performs some basic
  # checks that the prediction was successful.
  # Exceptions here are possibly encapsulated, so that they get captured and turned into log messages.

  predict_wrapper = function(task, learner) {
    # default method does nothing
    learner$model = unmarshal_model(learner$model, inplace = TRUE)
    if (is.null(learner$state$model)) {
      error_learner_predict("No trained model available for learner '%s' on task '%s'", learner$id, task$id)
    }

    result = get_private(learner)$.predict(task)
    as_prediction_data(result, task = task, check = TRUE, train_task = learner$state$train_task)
  }

  assert_task(task)
  assert_learner(learner)

  # ensure that required packages are installed
  require_namespaces(learner$packages)

  if (getOption("mlr3.warn_version_mismatch", TRUE)) {
    v_train = learner$state$mlr3_version
    v_predict = mlr_reflections$package_version

    if (!is.null(v_train) && v_train != v_predict) {
      warning_mlr3(
        # nolint next
        "Detected version mismatch: Learner '%s' has been trained with mlr3 version '%s', not matching currently installed version '%s'",
        learner$id,
        v_train,
        v_predict
      )
    }
  }

  # subset to test set w/o cloning
  if (!is.null(row_ids)) {
    lg$debug("Subsetting task '%s' to %i rows", task$id, length(row_ids), task = task$clone(), row_ids = row_ids)

    task_private = get_private(task)
    prev_use = task_private$.row_roles$use
    on.exit(
      {
        task_private$.row_roles$use = prev_use
      },
      add = TRUE
    )
    task_private$.row_roles$use = row_ids
  } else {
    lg$debug("Skip subsetting of task '%s'", task$id)
  }

  if (task$nrow == 0L) {
    # return an empty prediction object, #421
    lg$debug("No observations in task, returning empty prediction data", task = task)
    learner$state$log = append_log(
      learner$state$log,
      stage = "predict",
      class = "output",
      condition = list(simpleMessage("No data to predict on, create empty prediction"))
    )
    return(create_empty_prediction_data(task, learner))
  }

  if (is.null(learner$state$model)) {
    lg$debug("Learner '%s' has no model stored", learner$id, learner = learner$clone())

    pdata = NULL
    learner$state$predict_time = NA_real_
    cond = error_learner_predict("No model stored", signal = FALSE, class = "Mlr3ErrorLearnerNoModel")
    if (learner_will_err(cond, learner, stage = "predict")) {
      stop(cond)
    }
  } else {
    # call predict with encapsulation
    lg$debug(
      "Calling predict method of Learner '%s' on task '%s' with %i observations",
      learner$id,
      task$id,
      task$nrow,
      learner = learner$clone()
    )

    if (learner$encapsulation[["predict"]] %in% c("callr", "mirai")) {
      learner$model = marshal_model(learner$model, inplace = TRUE)
    }

    deadline_time_left = max(as.numeric(difftime(learner$deadline[["predict"]], Sys.time(), units = "secs")), 0)

    result = encapsulate(
      learner$encapsulation["predict"],
      .f = predict_wrapper,
      .args = list(task = task, learner = learner),
      .pkgs = learner$packages,
      .seed = NA_integer_,
      .timeout = min(learner$timeout["predict"], deadline_time_left),
      .compute = getOption("mlr3.mirai_encapsulation", "mlr3_encapsulation")
    )

    cond = cond_from_log(result$log)
    # Still log messages and warnings in the case of an error
    err = learner_will_err(cond, learner, stage = "predict")

    pdata = result$result
    # don't log an existing (uncaught) error, because it's signalled
    learner$state$log = append_log(
      learner$state$log,
      stage = "predict",
      class = result$log$class,
      condition = result$log$condition,
      log_error = !err
    )

    if (err) {
      stop(cond)
    }
    learner$state$predict_time = sum(learner$state$predict_time, result$elapsed)

    lg$debug(
      "Learner '%s' returned an object of class '%s'",
      learner$id,
      class(pdata)[1L],
      learner = learner$clone(),
      prediction_data = pdata,
      messages = result$log$msg
    )
  }

  fb = learner$fallback
  if (!is.null(fb)) {
    predict_fb = function(row_ids) {
      fb = assert_learner(as_learner(fb))
      fb$predict_type = learner$predict_type
      fb$state = learner$state$fallback_state
      as_prediction_data(fb$predict(task, row_ids), task, row_ids, check = TRUE, train_task = learner$state$train_task)
    }

    if (is.null(pdata)) {
      lg$debug("Creating new Prediction using fallback '%s'", fb$id, learner = fb$clone())

      learner$state$log = append_log(
        learner$state$log,
        stage = "predict",
        class = "output",
        condition = list(simpleMessage("Using fallback learner for predictions"))
      )
      pdata = predict_fb(task$row_ids)
    } else {
      miss_ids = is_missing_prediction_data(pdata)

      lg$debug(
        "Imputing %i/%i predictions using fallback '%s'",
        length(miss_ids),
        length(pdata$row_ids),
        fb$id,
        learner = fb$clone()
      )

      if (length(miss_ids)) {
        learner$state$log = append_log(
          learner$state$log,
          stage = "predict",
          class = "output",
          condition = list(simpleMessage("Using fallback learner to impute predictions"))
        )

        pdata = c(pdata, predict_fb(miss_ids), keep_duplicates = FALSE)
      }
    }
  }

  return(pdata)
}


workhorse = function(
  iteration,
  task,
  learner,
  resampling,
  param_values = NULL,
  lgr_index,
  store_models = FALSE,
  pb = NULL,
  mode = "train",
  is_sequential = TRUE,
  unmarshal = TRUE,
  callbacks = NULL
) {
  ctx = ContextResample$new(task, learner, resampling, iteration)

  call_back("on_resample_begin", callbacks, ctx)

  if (!is.null(pb)) {
    pb(sprintf("%s|%s|i:%i", task$id, learner$id, iteration))
  }
  # fmt: skip
  if (
    "internal_valid" %chin% learner$predict_sets &&
      is.null(task$internal_valid_task) &&
      is.null(get0("validate", learner))
  ) {
    error_config(
      # nolint next
      "Cannot set the predict_type field of learner '%s' to 'internal_valid' if there is no internal validation task configured",
      learner$id
    )
  }

  # restore settings on the workers
  if (!is_sequential) {
    # reduce data.table threads to 1
    setDTthreads(1, restore_after_fork = TRUE)

    # reduce blas threads to 1
    # RhpcBLASctl is licensed under AGPL and therefore should be in suggest #1023
    if (require_namespaces("RhpcBLASctl", quietly = TRUE)) {
      old_blas_threads = RhpcBLASctl::blas_get_num_procs()
      on.exit(RhpcBLASctl::blas_set_num_threads(old_blas_threads), add = TRUE)
      RhpcBLASctl::blas_set_num_threads(1)
    } else {
      # try the bare minimum to disable threading of the most popular blas implementations
      old_blas = Sys.getenv("OPENBLAS_NUM_THREADS")
      old_mkl = Sys.getenv("MKL_NUM_THREADS")
      Sys.setenv(OPENBLAS_NUM_THREADS = 1)
      Sys.setenv(MKL_NUM_THREADS = 1)

      on.exit(
        {
          Sys.setenv(OPENBLAS_NUM_THREADS = old_blas)
          Sys.setenv(MKL_NUM_THREADS = old_mkl)
        },
        add = TRUE
      )
    }

    # restore logger thresholds
    # skip inherited thresholds
    lgr_index = lgr_index[!lgr_index$threshold_inherited, ]
    mapply(
      function(name, threshold) {
        logger = lgr::get_logger(name)
        logger$set_threshold(threshold)
      },
      lgr_index$name,
      lgr_index$threshold
    )
  }

  lg$info(
    "%s learner '%s' on task '%s' (iter %i/%i)",
    if (mode == "train") "Applying" else "Hotstarting",
    learner$id,
    task$id,
    iteration,
    resampling$iters
  )

  sets = list(
    train = resampling$train_set(iteration),
    test = resampling$test_set(iteration)
  )

  # train model
  # use `learner` reference instead of `ctx$learner` to avoid going through the active binding
  ctx$learner = learner = ctx$learner$clone()
  if (length(param_values)) {
    learner$param_set$values = list()
    learner$param_set$set_values(.values = param_values)
  }
  learner_hash = learner$hash

  validate = get0("validate", learner)

  test_set = if (identical(validate, "test")) sets$test

  call_back("on_resample_before_train", callbacks, ctx)

  train_result = learner_train(learner, task, sets[["train"]], test_set, mode = mode)
  ctx$learner = learner = train_result$learner

  # process the model so it can be used for prediction (e.g. marshal for callr prediction), but also
  # keep a copy of the model in current form in case this is the format that we want to send back to the main process
  # and not the format that we need for prediction
  model_copy_or_null = process_model_before_predict(
    learner = learner,
    store_models = store_models,
    is_sequential = is_sequential,
    unmarshal = unmarshal
  )

  # predict for each set
  predict_sets = learner$predict_sets

  # creates the tasks and row_ids for all selected predict sets
  pred_data = prediction_tasks_and_sets(task, train_result, validate, sets, predict_sets)
  call_back("on_resample_before_predict", callbacks, ctx)

  pdatas = Map(
    function(set, row_ids, task) {
      lg$debug("Creating Prediction for predict set '%s'", set)

      learner_predict(learner, task, row_ids)
    },
    set = predict_sets,
    row_ids = pred_data$sets,
    task = pred_data$tasks
  )

  if (!length(predict_sets)) {
    learner$state$predict_time = 0L
  }
  ctx$pdatas = discard(pdatas, is.null)

  # set the model slot after prediction so it can be sent back to the main process
  process_model_after_predict(
    learner = learner,
    store_models = store_models,
    is_sequential = is_sequential,
    model_copy = model_copy_or_null,
    unmarshal = unmarshal
  )

  call_back("on_resample_end", callbacks, ctx)

  if (!store_models) {
    lg$debug("Erasing stored model for learner '%s'", learner$id)
    learner$state$model = NULL
  }

  learner_state = set_class(learner$state, c("learner_state", "list"))

  list(
    learner_state = learner_state,
    prediction = ctx$pdatas,
    param_values = learner$param_set$values,
    learner_hash = learner_hash,
    data_extra = ctx$data_extra
  )
}

# creates the tasks and row ids for the selected predict sets
prediction_tasks_and_sets = function(task, train_result, validate, sets, predict_sets) {
  predict_sets = predict_sets[predict_sets %in% mlr_reflections$predict_sets]
  tasks = list(train = task, test = task)
  if ("internal_valid" %nin% predict_sets) {
    return(list(tasks = tasks[predict_sets], sets = sets[predict_sets]))
  }

  if ("internal_valid" %chin% predict_sets) {
    if (is.numeric(validate) || identical(validate, "test")) {
      # in this scenario, the internal_valid_task was created during learner_train, which means that it used the
      # primary task. The selected ids are returned via the train result
      tasks$internal_valid = task
      sets$internal_valid = train_result$internal_valid_task_ids
    } else {
      # the predefined internal_valid_task was used
      tasks$internal_valid = task$internal_valid_task
      sets$internal_valid = task$internal_valid_task$row_ids
    }
  }

  list(tasks = tasks[predict_sets], sets = sets[predict_sets])
}

process_model_before_predict = function(learner, store_models, is_sequential, unmarshal) {
  # there are three states of the model that have to be considered to minimize how often we marshal a model:
  # 1. the current form: is it marshaled or not?
  # 2. the form for prediction: do we need to marshal it?
  # 3. the final form that is returned: does it have to be marshaled?
  #
  # and also, do we even need to send it back at all?

  currently_marshaled = is_marshaled_model(learner$model)
  predict_needs_marshaling = learner$encapsulation[["predict"]] %in% c("callr", "mirai")
  final_needs_marshaling = !is_sequential || !unmarshal

  # the only scenario in which we keep a copy is when we now have the model in the correct form but need to transform
  # it for prediction
  keep_copy = store_models &
    (currently_marshaled == final_needs_marshaling) &&
    (currently_marshaled != predict_needs_marshaling)

  if (!keep_copy) {
    # here we either
    # * don't return the model at all --> no copy
    # * the predict form is equal to the final form --> no copy
    # * we do store models but the current form is not the final form --> no copy
    if (predict_needs_marshaling) {
      learner$model = marshal_model(learner$model, inplace = TRUE)
    } else {
      learner$model = unmarshal_model(learner$model, inplace = TRUE)
    }
    return(NULL)
  }

  # here, we do store models, the current form is the final form and the current form is not the predict form
  # in order to avoid a marshaling cycle, we therefore keep a copy of the current model and then continue to
  # process the model for prediction

  # note that even though learner_predict takes care of the marshaling itself, it does do it in-place
  # Because we here have a copy of the model, we transform it NOT in-place. This is important because otherwise
  # we will mess up our copy

  model_copy = learner$model
  if (predict_needs_marshaling) {
    learner$model = marshal_model(learner$model, inplace = FALSE)
  } else {
    learner$model = unmarshal_model(learner$model, inplace = FALSE)
  }
  model_copy
}

process_model_after_predict = function(learner, store_models, is_sequential, unmarshal, model_copy) {
  if (store_models && !is.null(model_copy)) {
    # we created a copy of the model to avoid additional marshaling cycles
    learner$model = model_copy
  } else if (store_models && !is_sequential || !unmarshal) {
    # no copy was created, here we make sure that we return the model the way the user wants it
    learner$model = marshal_model(learner$model, inplace = TRUE)
  }
}

append_log = function(log = NULL, stage = NA_character_, class = NA_character_, condition = NULL, log_error = TRUE) {
  if (is.null(log)) {
    log = data.table(
      stage = factor(levels = c("train", "predict")),
      class = factor(levels = c("output", "warning", "error"), ordered = TRUE),
      condition = list()
    )
  }

  if (length(condition)) {
    log = rbindlist(list(log, data.table(stage = stage, class = class, condition = condition)), use.names = TRUE)

    # log messages with lgr / print to console
    pwalk(list(stage, class, condition), function(s, cls, cond) {
      m = conditionMessage(cond)
      if (cls == "error" && log_error) {
        lg$error("%s: %s", s, m)
      }
      if (cls == "warning") lg$warn("%s: %s", s, m)
    })
  }

  log
}

create_internal_valid_task = function(validate, task, test_row_ids, prev_valid, learner) {
  if (is.null(validate)) {
    task$internal_valid_task = NULL
    return(task)
  }

  # Otherwise, predict_set = "internal_valid" is ambiguous
  if (!is.null(prev_valid) && (is.numeric(validate) || identical(validate, "test"))) {
    error_config(
      "Parameter 'validate' of Learner '%s' cannot be set to 'test' or a ratio when internal_valid_task is present",
      learner$id
    )
  }

  if (is.character(validate)) {
    if (validate == "predefined") {
      if (is.null(task$internal_valid_task)) {
        error_config(paste0(
          "Parameter 'validate' is set to 'predefined' but no internal validation task is present. ",
          "This commonly happens in GraphLearners and can be avoided by configuring the validation data for the ",
          "GraphLearner via `set_validate(glrn, validate = values)`. ",
          "See https://mlr3book.mlr-org.com/chapters/chapter15/predsets_valid_inttune.html for more information."
        ))
      }
      if (!identical(task$target_names, task$internal_valid_task$target_names)) {
        error_config(
          # nolint next
          "Internal validation task '%s' has different target names than primary task '%s', did you modify the task after creating the internal validation task?",
          task$internal_valid_task$id,
          task$id
        )
      }
      if (!test_permutation(task$feature_names, task$internal_valid_task$feature_names)) {
        error_config(
          # nolint next
          "Internal validation task '%s' has different features than primary task '%s', did you modify the task after creating the internal validation task?",
          task$internal_valid_task$id,
          task$id
        )
      }
      return(task)
    } else {
      # validate is "test"
      if (is.null(test_row_ids)) {
        error_config("Parameter 'validate' cannot be set to 'test' when calling train manually.")
      }
      # at this point, the train rows are already set to the train set, i.e. we don't have to remove the test ids
      # from the primary task (this would cause bugs for resamplings with overlapping train and test set)
      valid_task = task$clone(deep = TRUE)
      valid_task$row_roles$use = test_row_ids
      task$internal_valid_task = valid_task
      return(task)
    }
  }

  # validate is numeric
  task$internal_valid_task = partition(task, ratio = 1 - validate)$test
  task
}

learner_will_err = function(cond, learner, stage) {
  when = get_private(learner)$.when
  if (is.null(cond)) {
    return(FALSE)
  }
  if (inherits(cond, "Mlr3ErrorConfig")) {
    return(TRUE)
  }
  if (is.null(when)) {
    return(FALSE)
  }
  !when(cond = cond, stage = stage)
}

cond_from_log = function(log) {
  x = log[class == "error", "condition"][[1L]]
  if (length(x)) {
    x[[1L]]
  } else {
    NULL
  }
}

#' @title Motor Trend Regression Task
#'
#' @name mlr_tasks_mtcars
#' @format [R6::R6Class] inheriting from [TaskRegr].
#' @include mlr_tasks.R
#'
#' @description
#' A regression task for the [datasets::mtcars] data set.
#' Target variable is `mpg` (Miles/(US) gallon).
#' Rownames are stored as variable `"model"` with column role `"name"`.
#'
#' @section Construction:
#' ```
#' mlr_tasks$get("mtcars")
#' tsk("mtcars")
#' ```
#'
#' @section Meta Information:
#' `r rd_info(tsk("mtcars"))`
#'
#' @template seealso_task
NULL

load_task_mtcars = function(id = "mtcars") {
  b = as_data_backend(load_dataset("mtcars", "datasets", keep_rownames = TRUE), keep_rownames = "model")
  task = TaskRegr$new(id, b, target = "mpg", label = "Motor Trends")
  b$hash = task$man = "mlr3::mlr_tasks_mtcars"
  task$col_roles$name = "model"
  task$col_roles$feature = setdiff(task$col_roles$feature, "model")
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("mtcars", load_task_mtcars)

#' @title Spiral Classification Task Generator
#'
#' @name mlr_task_generators_spirals
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] for the spirals task in [mlbench::mlbench.spirals()].
#'
#' @templateVar id spirals
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("spirals")
#' plot(generator, n = 200)
#'
#' task = generator$generate(200)
#' str(task$data())
TaskGeneratorSpirals = R6Class(
  "TaskGeneratorSpirals",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        cycles = p_int(1L, default = 1L),
        sd = p_dbl(0, default = 0)
      )

      super$initialize(
        id = "spirals",
        "classif",
        "mlbench",
        ps,
        label = "Spiral Classification",
        man = "mlr3::mlr_task_generators_spirals"
      )
    },

    #' @description
    #' Creates a simple plot of generated data.
    #' @param n (`integer(1)`)\cr
    #'   Number of samples to draw for the plot. Default is 200.
    #' @param pch (`integer(1)`)\cr
    #'   Point char. Passed to [plot()].
    #' @param ... (any)\cr
    #'   Additional arguments passed to [plot()].
    plot = function(n = 200L, pch = 19L, ...) {
      plot(private$.generate_obj(n), pch = pch, ...)
    }
  ),

  private = list(
    .generate_obj = function(n) {
      invoke(mlbench::mlbench.spirals, n = n, .args = self$param_set$values, .opts = allow_partial_matching)
    },

    .generate = function(n) {
      obj = private$.generate_obj(n)
      TaskClassif$new(sprintf("%s_%i", self$id, n), convert_mlbench(obj), target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("spirals", function() TaskGeneratorSpirals$new())

#' @title Benchmark Multiple Learners on Multiple Tasks
#'
#' @description
#' Runs a benchmark on arbitrary combinations of tasks ([Task]), learners ([Learner]),
#' and resampling strategies ([Resampling]), possibly in parallel.
#'
#' For large-scale benchmarking we recommend to use the \CRANpkg{mlr3batchmark} package.
#' This package runs benchmark experiments on high-performance computing clusters and handles failed experiments.
#'
#' @section Stochasticity:
#' Note that uninstantiated [`Resampling`]s are instantiated on the task, making
#' the function stochastic even in case of deterministic learners.
#'
#' @param design ([data.frame()])\cr
#'   Data frame (or [data.table::data.table()]) with three columns: "task", "learner", and "resampling".
#'   Each row defines a resampling by providing a [Task], [Learner] and an instantiated [Resampling] strategy.
#'   The helper function [benchmark_grid()] can assist in generating an exhaustive design (see examples) and
#'   instantiate the [Resampling]s per [Task].
#'   Additionally, you can set the additional column 'param_values', see [benchmark_grid()].
#' @template param_store_models
#' @template param_store_backends
#' @template param_encapsulate
#' @template param_allow_hotstart
#' @template param_clone
#' @template param_unmarshal
#' @template param_callbacks
#'
#' @return [BenchmarkResult].
#'
#' @note
#' The fitted models are discarded after the predictions have been scored in order to reduce memory consumption.
#' If you need access to the models for later analysis, set `store_models` to `TRUE`.
#'
#' @template section_predict_sets
#' @template section_parallelization
#' @template section_progress_bars
#' @template section_logging
#'
#' @template seealso_benchmark
#' @export
#' @examples
#' # benchmarking with benchmark_grid()
#' tasks = lapply(c("penguins", "sonar"), tsk)
#' learners = lapply(c("classif.featureless", "classif.rpart"), lrn)
#' resamplings = rsmp("cv", folds = 3)
#'
#' design = benchmark_grid(tasks, learners, resamplings)
#' print(design)
#'
#' set.seed(123)
#' bmr = benchmark(design)
#'
#' ## Data of all resamplings
#' head(as.data.table(bmr))
#'
#' ## Aggregated performance values
#' aggr = bmr$aggregate()
#' print(aggr)
#'
#' ## Extract predictions of first resampling result
#' rr = aggr$resample_result[[1]]
#' as.data.table(rr$prediction())
#'
#' # Benchmarking with a custom design:
#' # - fit classif.featureless on penguins with a 3-fold CV
#' # - fit classif.rpart on sonar using a holdout
#' tasks = list(tsk("penguins"), tsk("sonar"))
#' learners = list(lrn("classif.featureless"), lrn("classif.rpart"))
#' resamplings = list(rsmp("cv", folds = 3), rsmp("holdout"))
#'
#' design = data.table::data.table(
#'   task = tasks,
#'   learner = learners,
#'   resampling = resamplings
#' )
#'
#' ## Instantiate resamplings
#' design$resampling = Map(
#'   function(task, resampling) resampling$clone()$instantiate(task),
#'   task = design$task, resampling = design$resampling
#' )
#'
#' ## Run benchmark
#' bmr = benchmark(design)
#' print(bmr)
#'
#' ## Get the training set of the 2nd iteration of the featureless learner on penguins
#' rr = bmr$aggregate()[learner_id == "classif.featureless"]$resample_result[[1]]
#' rr$resampling$train_set(2)
benchmark = function(
  design,
  store_models = FALSE,
  store_backends = TRUE,
  encapsulate = NA_character_,
  allow_hotstart = FALSE,
  clone = c("task", "learner", "resampling"),
  unmarshal = TRUE,
  callbacks = NULL
) {
  assert_subset(clone, c("task", "learner", "resampling"))
  assert_data_frame(design, min.rows = 1L)
  assert_names(names(design), must.include = c("task", "learner", "resampling"))
  assert_flag(unmarshal)
  design$task = list(assert_tasks(as_tasks(design$task)))
  design$learner = list(assert_learners(as_learners(design$learner)))
  design$resampling = list(assert_resamplings(as_resamplings(design$resampling), instantiated = TRUE))
  if (is.null(design$param_values)) {
    design$param_values = list()
  } else {
    design$param_values = list(assert_param_values(design$param_values, n_learners = length(design$learner)))
  }
  assert_flag(store_models)
  assert_flag(store_backends)
  callbacks = assert_callbacks(as_callbacks(callbacks))

  # check for multiple task types
  task_types = unique(map_chr(design$task, "task_type"))
  if (length(task_types) > 1L) {
    error_input("Multiple task types detected, but mixing types is not supported: %s", str_collapse(task_types))
  }
  learner_types = unique(map_chr(design$learner, "task_type"))
  if (length(learner_types) > 1L) {
    error_input("Multiple learner types detected, but mixing types is not supported: %s", str_collapse(learner_types))
  }

  setDT(design)
  task = learner = resampling = NULL
  if ("task" %chin% clone) {
    design[, "task" := list(list(task[[1L]]$clone())), by = list(hashes(task))]
  }
  if ("learner" %chin% clone) {
    design[, "learner" := list(list(learner[[1L]]$clone())), by = list(hashes(learner))]
  }
  if ("resampling" %chin% clone) {
    design[, "resampling" := list(list(resampling[[1L]]$clone())), by = list(hashes(resampling))]
  }

  # set encapsulation + fallback
  set_encapsulation(design$learner, encapsulate)

  # expand the design: add rows for each resampling iteration and param_values
  grid = pmap_dtr(design, function(task, learner, resampling, param_values) {
    iters = resampling$iters
    n_params = max(1L, length(param_values))
    # insert constant values
    param_values = map(param_values, function(values) insert_named(learner$param_set$values, values))
    assert_learnable(task, learner, unlist(param_values, recursive = FALSE))

    # check that all row ids of the resampling are present in the task
    if (resampling$task_row_hash != task$row_hash) {
      error_input("Resampling '%s' is not instantiated on task '%s'", resampling$id, task$id)
    }

    data.table(
      task = list(task),
      learner = list(learner),
      resampling = list(resampling),
      iteration = rep(seq_len(iters), times = n_params),
      param_values = if (is.null(param_values)) list() else rep(param_values, each = iters),
      uhash = rep(UUIDgenerate(n = n_params), each = iters)
    )
  })

  n = nrow(grid)

  # set default mode
  set(grid, j = "mode", value = "train")

  lg$info("Running benchmark with %i resampling iterations", n)
  pb = if (isNamespaceLoaded("progressr")) {
    # NB: the progress bar needs to be created in this env
    pb = progressr::progressor(steps = n)
  } else {
    NULL
  }

  # add hot start learners
  if (allow_hotstart) {
    hotstart_grid = pmap_dtr(grid, function(task, learner, resampling, iteration, ...) {
      if (!is.null(learner$hotstart_stack)) {
        # search for hotstart learner
        learner = learner$clone()
        task_hashes = resampling_task_hashes(task, resampling, learner)
        start_learner = get_private(learner$hotstart_stack)$.start_learner(learner, task_hashes[iteration])
      }
      if (is.null(learner$hotstart_stack) || is.null(start_learner)) {
        # no hotstart learners stored or no adaptable model found
        mode = "train"
      } else {
        # hotstart learner found
        start_learner$param_set$values = insert_named(start_learner$param_set$values, learner$param_set$values)
        learner = start_learner
        mode = "hotstart"
      }
      data.table(learner = list(learner), mode = mode)
    })

    # null hotstart stack to reduce overhead in parallelization
    walk(hotstart_grid$learner, function(learner) {
      learner$hotstart_stack = NULL
      learner
    })
    set(grid, j = "learner", value = hotstart_grid$learner)
    set(grid, j = "mode", value = hotstart_grid$mode)
  }

  res = future_map(
    n,
    workhorse,
    task = grid$task,
    learner = grid$learner,
    resampling = grid$resampling,
    iteration = grid$iteration,
    param_values = grid$param_values,
    mode = grid$mode,
    MoreArgs = list(
      store_models = store_models,
      lgr_index = lgr::logger_index(),
      pb = pb,
      unmarshal = unmarshal,
      callbacks = callbacks
    )
  )

  grid = insert_named(
    grid,
    list(
      learner_state = map(res, "learner_state"),
      prediction = map(res, "prediction"),
      param_values = map(res, "param_values"),
      learner_hash = map_chr(res, "learner_hash")
    )
  )

  lg$info("Finished benchmark")

  set(grid, j = "mode", value = NULL)

  data_extra = if (length(callbacks) && any(map_lgl(res, function(x) !is.null(x$data_extra)))) map(res, "data_extra")

  result_data = ResultData$new(grid, data_extra, store_backends = store_backends)

  if (unmarshal && store_models) {
    result_data$unmarshal()
  }

  BenchmarkResult$new(result_data)
}

#' @title Convert to a Resampling
#'
#' @description
#' Convert object to a [Resampling] or a list of [Resampling].
#' This method e.g. allows to convert an `OMLTask` of \CRANpkg{mlr3oml} to a [`Resampling`].
#' @inheritParams as_task
#' @export
# nolint next
as_resampling = function(x, ...) {
  UseMethod("as_resampling")
}

#' @export
#' @rdname as_resampling
# nolint next
as_resampling.Resampling = function(x, clone = FALSE, ...) {
  assert_empty_ellipsis(...)
  if (isTRUE(clone)) x$clone() else x
}

#' @export
#' @rdname as_resampling
# nolint next
as_resamplings = function(x, ...) {
  UseMethod("as_resamplings")
}

#' @export
#' @rdname as_resampling
# nolint next
as_resamplings.default = function(x, ...) {
  list(as_resampling(x, ...))
}

#' @export
#' @rdname as_resampling
# nolint next
as_resamplings.list = function(x, ...) {
  lapply(x, as_resampling, ...)
}

test_that("as_resampling conversion", {
  resampling = rsmp("subsampling")
  converted = as_resampling(resampling)
  cloned = as_resampling(resampling, clone = TRUE)

  expect_class(converted, "Resampling")
  expect_same_address(resampling, converted)
  expect_different_address(resampling, cloned)

  expect_list(as_resamplings(resampling), types = "Resampling")
  expect_list(as_resamplings(list(resampling)), types = "Resampling")
})

test_that("error when arguments are misspelled", {
  expect_error(as_resampling(rsmp("holdout"), clone2 = TRUE), "Received the following")
})

#' @title Insample Resampling
#'
#' @name mlr_resamplings_insample
#' @include Resampling.R
#'
#' @description
#' Uses all observations as training and as test set.
#'
#' @templateVar id insample
#' @template resampling
#'
#' @template seealso_resampling
#' @export
#' @examples
#' # Create a task with 10 observations
#' task = tsk("penguins")
#' task$filter(1:10)
#'
#' # Instantiate Resampling
#' insample = rsmp("insample")
#' insample$instantiate(task)
#'
#' # Train set equal to test set:
#' setequal(insample$train_set(1), insample$test_set(1))
#'
#' # Internal storage:
#' insample$instance # just row ids
ResamplingInsample = R6Class(
  "ResamplingInsample",
  inherit = Resampling,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      super$initialize(id = "insample", label = "Insample Resampling", man = "mlr3::mlr_resamplings_insample")
    }
  ),
  active = list(
    #' @template field_iters
    iters = function(rhs) {
      assert_ro_binding(rhs)
      1L
    }
  ),

  private = list(
    .sample = function(ids, ...) {
      ids
    },

    .get_train = function(i) {
      self$instance
    },

    .get_test = function(i) {
      self$instance
    },

    .combine = function(instances) {
      do.call(c, instances)
    }
  )
)

#' @include mlr_resamplings.R
mlr_resamplings$add("insample", function() ResamplingInsample$new())

#' @title Wisconsin Breast Cancer Classification Task
#'
#' @name mlr_tasks_breast_cancer
#' @format [R6::R6Class] inheriting from [TaskClassif].
#' @include mlr_tasks.R
#'
#' @description
#' A classification task for the [mlbench::BreastCancer] data set.
#'
#' * Column `"Id"` has been removed.
#' * Column names have been converted to `snake_case`.
#' * Positive class is set to `"malignant"`.
#' * 16 incomplete cases have been removed from the data set.
#' * All factor features have been converted to ordered factors.
#'
#' @templateVar id breast_cancer
#' @template task
#'
#' @template seealso_task
NULL

load_task_breast_cancer = function(id = "breast_cancer") {
  tab = load_dataset("BreastCancer", "mlbench")
  names(tab) = tolower(chartr(".", "_", colnames(tab)))
  for (cn in c("bare_nuclei", "bl_cromatin", "normal_nucleoli", "mitoses")) {
    tab[[cn]] = ordered(tab[[cn]])
  }
  b = as_data_backend(remove_named(tab[stats::complete.cases(tab), ], "id"))

  task = TaskClassif$new(id, b, target = "class", positive = "malignant", label = "Wisconsin Breast Cancer")
  b$hash = task$man = "mlr3::mlr_tasks_breast_cancer"
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("breast_cancer", load_task_breast_cancer)

#' @title Convert to an Unsupervised Task
#'
#' @description
#' Convert object to a [TaskUnsupervised] or a list of [TaskUnsupervised].
#'
#' @inheritParams as_task
#'
#' @export
as_task_unsupervised = function(x, ...) {
  UseMethod("as_task_unsupervised")
}

#' @rdname as_task_unsupervised
#' @export
# nolint next
as_task_unsupervised.Task = function(x, clone = FALSE, ...) {
  if (clone) x$clone() else x
}

#' @rdname as_task_unsupervised
#' @param id (`character(1)`)\cr
#'   Id for the new task.
#'   Defaults to the (deparsed and substituted) name of the data argument.
#' @template param_label
#' @export
# nolint next
as_task_unsupervised.data.frame = function(x, id = deparse1(substitute(x)), label = NA_character_, ...) {
  force(id)

  ii = which(map_lgl(keep(x, is.double), anyInfinite))
  if (length(ii)) {
    warning_input("Detected columns with unsupported Inf values in data: %s", str_collapse(names(ii)))
  }

  TaskUnsupervised$new(id = id, backend = x, label = label)
}

#' @rdname as_task_unsupervised
#' @export
# nolint next
as_task_unsupervised.DataBackend = function(x, id = deparse1(substitute(x)), label = NA_character_, ...) {
  force(id)

  TaskUnsupervised$new(id = id, backend = x, label = label)
}

#' @rdname as_task_unsupervised
#' @export
as_tasks_unsupervised = function(x, ...) {
  UseMethod("as_tasks_unsupervised")
}

#' @rdname as_task_unsupervised
#' @param clone (`logical(1)`)\cr
#'   If `TRUE`, ensures that the returned object is not the same as the input `x`.
#' @export
# nolint next
as_tasks_unsupervised.list = function(x, clone = FALSE, ...) {
  lapply(x, as_task_unsupervised, clone = clone, ...)
}

#' @rdname as_task_unsupervised
#' @export
# nolint next
as_tasks_unsupervised.Task = function(x, clone = FALSE, ...) {
  list(if (clone) x$clone() else x)
}

test_that("as_tasks_unsupervised dispatches correctly", {
  task = as_task_unsupervised(data.frame(x = 1:3))
  expect_identical(as_tasks_unsupervised(task), list(task))
  expect_false(identical(as_tasks_unsupervised(task, clone = TRUE)[[1]], task))
})

#' @title Out-of-bag Error Measure
#'
#' @name mlr_measures_oob_error
#' @include Measure.R
#'
#' @description
#' Returns the out-of-bag error of the [Learner] for learners that support it
#' (learners with property `"oob_error"`).
#' Returns `NA` for unsupported learners.
#'
#' @templateVar id oob_error
#' @template measure
#'
#' @template seealso_measure
#' @export
MeasureOOBError = R6Class(
  "MeasureOOBError",
  inherit = Measure,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      super$initialize(
        id = "oob_error",
        task_type = NA_character_,
        properties = c("na_score", "requires_learner", "requires_no_prediction"),
        predict_sets = NULL,
        predict_type = NA_character_,
        range = c(-Inf, Inf),
        minimize = TRUE,
        label = "Out-of-bag Error",
        man = "mlr3::mlr_measures_oob_error"
      )
    }
  ),

  private = list(
    .score = function(prediction, learner, ...) {
      learner = learner$base_learner()
      if ("oob_error" %nin% learner$properties) {
        return(NA_real_)
      }
      # prefer value captured during training in learner$state to avoid requiring stored models
      if (!is.null(learner$state$oob_error)) {
        return(learner$state$oob_error)
      }

      learner$oob_error()
    }
  )
)

#' @include mlr_measures.R
mlr_measures$add("oob_error", function() MeasureOOBError$new())

#' @title Dictionary of Resampling Strategies
#'
#' @usage NULL
#' @format [R6::R6Class] object inheriting from [mlr3misc::Dictionary].
#'
#' @description
#' A simple [mlr3misc::Dictionary] storing objects of class [Resampling].
#' Each resampling has an associated help page, see `mlr_resamplings_[id]`.
#'
#' This dictionary can get populated with additional resampling strategies by add-on packages.
#'
#' For a more convenient way to retrieve and construct resampling strategies, see [rsmp()]/[rsmps()].
#'
#' @section Methods:
#' See [mlr3misc::Dictionary].
#'
#' @section S3 methods:
#' * `as.data.table(dict, ..., objects = FALSE)`\cr
#'   [mlr3misc::Dictionary] -> [data.table::data.table()]\cr
#'   Returns a [data.table::data.table()] with columns "key", "label", "params", and "iters".
#'   If `objects` is set to `TRUE`, the constructed objects are returned in the list column named `object`.
#'
#' @family Dictionary
#' @family Resampling
#' @seealso
#' Sugar functions: [rsmp()], [rsmps()]
#' @export
#' @examples
#' as.data.table(mlr_resamplings)
#' mlr_resamplings$get("cv")
#' rsmp("subsampling")
mlr_resamplings = R6Class("DictionaryResampling", inherit = Dictionary, cloneable = FALSE, )$new()

#' @export
# nolint next
as.data.table.DictionaryResampling = function(x, ..., objects = FALSE) {
  assert_flag(objects)

  setkeyv(
    map_dtr(
      x$keys(),
      function(key) {
        r = tryCatch(x$get(key), missingDefaultError = function(e) NULL)
        if (is.null(r)) {
          return(list(key = key))
        }

        insert_named(
          list(key = key, label = r$label, params = list(r$param_set$ids()), iters = r$iters),
          if (objects) list(object = list(r))
        )
      },
      .fill = TRUE
    ),
    "key"
  )[]
}

#' @title (Un)marshal a Learner
#'
#' @name marshaling
#'
#' @description
#' Marshaling is the process of processing the model of a trained [`Learner`] so it can be successfully serialized and
#' deserialized. The naming is inspired by the [marshal package](https://github.com/futureverse/marshal) and we
#' plan to fully migrate to this package once it is on CRAN.
#' The current implementation should therefore be considered as a temporary solution and is likely
#' to change in the future.
#'
#' The central functions (and the only methods that are used by `mlr3` internally) are:
#' * the S3 generic `marshal_model(model, inplace, ...)`.
#'   Which takes in a model and returns it in marshaled form.
#'   This means, that the resulting object can be serialized and de-serialized without loss of information.
#'   If a model is serializable anyway, nothing has to be implemented and the generic will fall back to the
#'   default implementation of `marshal_model`, which is to return the object as-is.
#'   Otherwise, the marshaled object should be a list with named elements `marshaled` and
#'   `packages`, where the former contains
#'   the marshaled object, and the latter the package that contains the packages required to unmarshal.
#'   Most importantly, this list should contain the package that contains the `unmarshal_model` method.
#'   The returned object should have the classes of the original object with the suffix `"_marshaled"` appended and the
#'   root class should be set to `"marshaled"`.
#' * the S3 generic `unmarshal_model(model, inplace ...)`.
#'   Which takes in the marshaled model and returns it in unmarshaled form.
#'   The generic takes care that the packages specified during `"marshal"` are loaded,
#'   and errs if they are not available.
#'   Calling this function on a marshaled model should reconstruct the original model, i.e.
#'   `unmarshal_model(marshal_model(x))` should return `x`.
#'   The default implementation of this generic returns `x` as-is.
#' * the function `is_marshaled_model(model)`.
#'   This (helper) function returns `TRUE` if the model inherits from class `"marshaled"` and `FALSE` otherwise.
#'   Note that it is not guaranteed that `is_marshaled_model(marshal_model(x))` returns `TRUE`.
#'   This is because the default `marshal_model(x)` returns `x` as-is.
#'
#' For both `marshal_model` and `unmarshal_model`, the `inplace` argument determines whether in-place marshaling
#' should be performed. This is especially relevant in the context of references semantics.
#' If `inplace` is `FALSE`, the original input should not be modified, otherwise this is allowed.
#' Note that the input and output can still share references, even when `inplace` is `FALSE`.
#'
#' @section Implementing Marshaling:
#'
#' In order to implement marshaling for a Learner, you need to overload the `marshal_model` and `unmarshal_model`
#' methods for the class of the learner's model and tag the learner with the `"marshal"` property.
#' To make marshaling accessible in an R6-manner, you should also add the public methods `$marshal()`, `$unmarshal()`
#' and the active binding `$marshaled`.
#' To make this as convenient as possible, the functions `learner_marshal(.learner, ...)`,
#' `learner_unmarshal(.learner, ...)`
#' and `learner_marshaled(.learner)` are provided and can be called from the public methods.
#'
#' You can verify whether you have correctly implemented marshaling by using the internal test helper
#' `expect_marshalable_learner(learner, task)`. This is also run by `expect_learner()` if a task is provided.
#'
#' For a concrete example on how to implement marshaling, see [`LearnerClassifDebug`].
#'
#' @param .learner [`Learner`]\cr
#'   The learner.
#' @param ... (any)\cr
#'   Additional parameters, currently unused.
#' @param model (any)\cr
#'   Model to marshal.
#' @param inplace (`logical(1)`)\cr
#'   Whether to marshal in-place.
#'
#' @export
learner_unmarshal = function(.learner, ...) {
  # this will do nothing if the model was not marshaled
  .learner$model = unmarshal_model(model = .learner$model, inplace = TRUE, ...)
  invisible(.learner)
}

#' @rdname marshaling
#' @export
learner_marshal = function(.learner, ...) {
  # this will do nothing if the model was already marshaled
  .learner$model = marshal_model(model = .learner$model, inplace = TRUE, ...)
  invisible(.learner)
}

#' @rdname marshaling
#' @export
learner_marshaled = function(.learner) {
  is_marshaled_model(model = .learner$model)
}

#' @rdname marshaling
#' @export
marshal_model = function(model, inplace = FALSE, ...) {
  assert_flag(inplace)
  UseMethod("marshal_model")
}

#' @rdname marshaling
#' @export
unmarshal_model = function(model, inplace = FALSE, ...) {
  if (is_marshaled_model(model) && is.character(model$packages)) {
    require_namespaces(model$packages)
  }
  assert_flag(inplace)
  UseMethod("unmarshal_model")
}

#' @rdname marshaling
#' @export
is_marshaled_model = function(model) {
  inherits(model, "marshaled")
}

#' @export
marshal_model.default = function(model, ...) {
  model
}

#' @export
unmarshal_model.default = function(model, inplace = FALSE, ...) {
  model
}

marshal_state_if_model = function(.state, inplace, ...) {
  if (!is.null(.state$model)) {
    .state$model = marshal_model(.state$model, inplace, ...)
  }
  .state
}

unmarshal_state_if_model = function(.state, inplace, ...) {
  if (!is.null(.state$model)) {
    .state$model = unmarshal_model(model = .state$model, inplace = inplace, ...)
  }
  .state
}

#' @export
print.marshaled = function(x, ...) {
  catn(sprintf("<%s>", class(x)[[1L]]))
}

#' @title Unsupervised Task
#'
#' @include Task.R
#'
#' @description
#' This is the abstract base class for unsupervised tasks such as cluster tasks in
#' \CRANpkg{mlr3cluster} and \CRANpkg{mlr3spatial}.
#'
#' @template param_id
#' @template param_task_type
#' @template param_backend
#' @template param_label
#' @template param_extra_args
#'
#' @template seealso_task
#' @keywords internal
#' @export
#' @examples
#' TaskUnsupervised$new("penguins", task_type = "unsupervised", backend = palmerpenguins::penguins)
TaskUnsupervised = R6Class(
  "TaskUnsupervised",
  inherit = Task,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function(id, task_type = "unsupervised", backend, label = NA_character_, extra_args = list()) {
      super$initialize(id = id, task_type = task_type, backend = backend, label = label, extra_args = extra_args)
    }
  )
)

#' @title Prediction Object for Regression
#'
#' @include Prediction.R
#'
#' @description
#' This object wraps the predictions returned by a learner of class [LearnerRegr], i.e.
#' the predicted response and standard error.
#' Additionally, probability distributions implemented in package `distr6` are supported.
#'
#' @template seealso_prediction
#' @export
#' @examples
#' task = tsk("california_housing")
#' learner = lrn("regr.featureless", predict_type = "se")
#' p = learner$train(task)$predict(task)
#' p$predict_types
#' head(as.data.table(p))
PredictionRegr = R6Class(
  "PredictionRegr",
  inherit = Prediction,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param task ([TaskRegr])\cr
    #'   Task, used to extract defaults for `row_ids` and `truth`.
    #'
    #' @param row_ids (`integer()`)\cr
    #'   Row ids of the predicted observations, i.e. the row ids of the test set.
    #'
    #' @param truth (`numeric()`)\cr
    #'   True (observed) response.
    #'
    #' @param response (`numeric()`)\cr
    #'   Vector of numeric response values.
    #'   One element for each observation in the test set.
    #'
    #' @param se (`numeric()`)\cr
    #'   Numeric vector of predicted standard errors.
    #'   One element for each observation in the test set.
    #'
    #' @param quantiles (`matrix()`)\cr
    #'   Numeric matrix of predicted quantiles. One row per observation, one column per quantile.
    #'
    #' @param distr (`VectorDistribution`)\cr
    #'   `VectorDistribution` from package distr6 (in repository \url{https://raphaels1.r-universe.dev}).
    #'   Each individual distribution in the vector represents the random variable 'survival time'
    #'   for an individual observation.
    #'
    #' @param weights (`numeric()`)\cr
    #'   Vector of measure weights for each observation. Should be constructed from
    #'   the `Task`'s `weights_measure` column.
    #'
    #' @param check (`logical(1)`)\cr
    #'   If `TRUE`, performs some argument checks and predict type conversions.
    #'
    #' @param extra (`list()`)\cr
    #'   List of extra data to be stored in the prediction object.
    #'
    #' @param raw (any)\cr
    #'   Raw prediction object from the upstream model. Stored as-is without validation.
    initialize = function(
      task = NULL,
      row_ids = task$row_ids,
      truth = task$truth(),
      response = NULL,
      se = NULL,
      quantiles = NULL,
      distr = NULL,
      weights = NULL,
      check = TRUE,
      extra = NULL,
      raw = NULL
    ) {
      pdata = new_prediction_data(
        list(
          row_ids = row_ids,
          truth = truth,
          response = response,
          se = se,
          quantiles = quantiles,
          distr = distr,
          weights = weights,
          extra = extra,
          raw = raw
        ),
        task_type = "regr"
      )

      if (check) {
        pdata = check_prediction_data(pdata)
      }
      self$task_type = "regr"
      self$man = "mlr3::PredictionRegr"
      self$data = pdata
      predict_types = intersect(names(mlr_reflections$learner_predict_types[["regr"]]), names(pdata))
      # response is in saved in quantiles matrix
      if ("quantiles" %chin% predict_types) {
        predict_types = union(predict_types, "response")
      }
      self$predict_types = predict_types
      if (is.null(pdata$response)) private$.quantile_response = attr(quantiles, "response")
    }
  ),

  active = list(
    #' @field response (`numeric()`)\cr
    #' Access the stored predicted response.
    response = function(rhs) {
      assert_ro_binding(rhs)
      if (!is.null(private$.quantile_response)) {
        return(self$data$quantiles[, private$.quantile_response])
      }
      self$data$response %??% rep(NA_real_, length(self$data$row_ids))
    },

    #' @field se (`numeric()`)\cr
    #' Access the stored standard error.
    se = function(rhs) {
      assert_ro_binding(rhs)
      self$data$se %??% rep(NA_real_, length(self$data$row_ids))
    },

    #' @field quantiles (`matrix()`)\cr
    #' Matrix of predicted quantiles. Observations are in rows, quantile (in ascending order) in columns.
    quantiles = function(rhs) {
      assert_ro_binding(rhs)
      self$data$quantiles
    },

    #' @field distr (`VectorDistribution`)\cr
    #' Access the stored vector distribution.
    #' Requires package `distr6`(in repository \url{https://raphaels1.r-universe.dev}) .
    distr = function() {
      if ("distr" %chin% self$predict_types) {
        require_namespaces("distr6", msg = "To predict probability distributions, please install %s")
      }
      self$data$distr
    }
  ),

  private = list(
    .quantile_response = NULL
  )
)


#' @export
# nolint next
as.data.table.PredictionRegr = function(x, ...) {
  tab = as.data.table(x$data[c("row_ids", "truth", "response", "se")])

  if ("quantiles" %chin% x$predict_types) {
    tab = rcbind(tab, as.data.table(x$data$quantiles))
    set(tab, j = "response", value = x$response)
  }

  if ("distr" %chin% x$predict_types) {
    require_namespaces("distr6", msg = "To predict probability distributions, please install %s")
    tab$distr = list(x$distr)
  }

  if (!is.null(x$data$weights)) {
    tab$weights = x$data$weights
  }

  if (!is.null(x$data$extra)) {
    tab = cbind(tab, as.data.table(x$data$extra))
  }

  tab
}

#' @title Median House Value in California
#'
#' @name california_housing
#' @format [R6::R6Class] inheriting from [TaskRegr].
#' @aliases mlr_tasks_california_housing
#'
#' @description
#' A regression task to predict the median house value in California.
#'
#' Contains 9 features and 20640 observations.
#' Target column is `"median_house_value"`.
#'
#' @section Construction:
#' ```
#' mlr_tasks$get("california_housing")
#' tsk("california_housing")
#' ```
#'
#' @section Meta Information:
#' `r rd_info(tsk("california_housing"))`
#'
#' @source https://www.kaggle.com/datasets/camnugent/california-housing-prices
#'
#' @template seealso_task
NULL

load_task_california_housing = function(id = "california_housing") {
  b = as_data_backend(readRDS(system.file("extdata", "california_housing.rds", package = "mlr3")))
  task = mlr3::TaskRegr$new(id, b, target = "median_house_value", label = "California House Value")
  b$hash = task$man = "mlr3::mlr_tasks_california_housing"
  task
}

#' @include mlr_tasks.R
mlr_tasks$add("california_housing", load_task_california_housing)

#' @title Set the Number of Threads
#'
#' @description
#' Control the parallelism via threading while calling external packages from \CRANpkg{mlr3}.
#'
#' For example, the random forest implementation in package \CRANpkg{ranger} (connected
#' via \CRANpkg{mlr3learners}) supports threading via OpenMP.
#' The number of threads to use can be set via hyperparameter `num.threads`, and
#' defaults to 1. By calling `set_threads(x, 4)` with `x` being a ranger learner, the
#' hyperparameter is changed so that 4 cores are used.
#'
#' If the object `x` does not support threading, `x` is returned as-is.
#' If applied to a list, recurses through all list elements.
#'
#' Note that threading is incompatible with other parallelization techniques such as forking
#' via the [future::plan] [future::multicore]. For this reason all learners connected to \CRANpkg{mlr3}
#' have threading disabled in their defaults.
#'
#' @param x (any)\cr
#'   Object to set threads for, e.g. a [Learner].
#'   This object is modified in-place.
#' @param n (`integer(1)`)\cr
#'   Number of threads to use. Defaults to [parallelly::availableCores()].
#' @param ... (any)\cr
#'   Additional arguments.
#'
#' @return Same object as input `x` (changed in-place),
#'   with possibly updated parameter values.
#' @export
set_threads = function(x, n = availableCores(), ...) {
  UseMethod("set_threads")
}

#' @rdname set_threads
#' @export
# nolint next
set_threads.default = function(x, n = availableCores(), ...) {
  x
}

#' @rdname set_threads
#' @export
# nolint next
set_threads.R6 = function(x, n = availableCores(), ...) {
  if (exists("param_set", envir = x)) {
    id = x$param_set$ids(tags = "threads")
    if (length(id)) {
      x$param_set$values = insert_named(x$param_set$values, named_list(id, n))
    }
  }
  x
}

#' @rdname set_threads
#' @export
# nolint next
set_threads.list = function(x, n = availableCores(), ...) {
  lapply(x, set_threads, n = n)
}

#' @title Model Extractor Callback
#'
#' @include CallbackResample.R
#' @name mlr3.model_extractor
#'
#' @description
#' This [CallbackResample] extracts information from the model after training with a user-defined function.
#' This way information can be extracted from the model without saving the model (`store_models = FALSE`).
#' The `fun` must be a function that takes a learner as input and returns the extracted information
#' as named list (see example).
#' The callback is very helpful to call `$selected_features()`, `$importance()`, `$oob_error()` on the learner.
#'
#' @param fun (`function(learner)`)\cr
#'   Function to extract information from the learner.
#'   The function must have the argument `learner`.
#'   The function must return a named list.
#'
#' @examples
#' task = tsk("pima")
#' learner = lrn("classif.rpart")
#' resampling = rsmp("cv", folds = 3)
#'
#' # define function to extract selected features
#' selected_features = function(learner) list(selected_features = learner$selected_features())
#'
#' # create callback
#' callback = clbk("mlr3.model_extractor", fun = selected_features)
#'
#' rr = resample(task, learner, resampling = resampling, store_models = FALSE, callbacks = callback)
#'
#' rr$data_extra
NULL

load_callback_model_extractor = function() {
  callback_resample(
    "mlr3.model_extractor",
    label = "Model Extractor Callback",
    man = "mlr3::mlr3.model_extractor",

    on_resample_end = function(callback, context) {
      assert_function(callback$state$fun, args = "learner")
      context$data_extra = invoke(callback$state$fun, learner = context$learner)
    }
  )
}

#' @title Callback Holdout Task
#'
#' @include CallbackResample.R
#' @name mlr3.holdout_task
#'
#' @description
#' This [CallbackResample] predicts on an additional holdout task after training.
#'
#' @param task ([Task])\cr
#'  The holdout task.
#'
#' @examples
#' task = tsk("pima")
#' task_holdout = task$clone()
#' learner = lrn("classif.rpart")
#' resampling = rsmp("cv", folds = 3)
#' splits = partition(task, 0.7)
#'
#' task$filter(splits$train)
#' task_holdout$filter(splits$test)
#'
#' callback = clbk("mlr3.holdout_task", task = task_holdout)
#'
#' rr = resample(task, learner, resampling = resampling, callbacks = callback)
#'
#' rr$data_extra
NULL

load_callback_holdout_task = function() {
  callback_resample(
    "mlr3.holdout_task",
    label = "Callback Holdout Task",
    man = "mlr3::mlr3.holdout_task",

    on_resample_before_predict = function(callback, context) {
      pred = context$learner$predict(callback$state$task)
      context$data_extra = list(prediction_holdout = pred)
    }
  )
}

#' @title Featureless Classification Learner
#'
#' @name mlr_learners_classif.featureless
#' @include LearnerClassif.R
#'
#' @description
#' A simple [LearnerClassif] which only analyzes the labels during train, ignoring all features.
#' Hyperparameter `method` determines the mode of operation during prediction:
#' \describe{
#'   \item{mode:}{
#'     Predicts the most frequent label. If there are two or more labels tied, randomly selects one per prediction.
#'     Probabilities correspond to the relative frequency of the class labels in the training set.
#'     For weighted data, the label(s) with the highest weighted frequency are selected.
#'   }
#'   \item{sample:}{
#'     Randomly predict a label uniformly.
#'     Probabilities correspond to a uniform distribution of class labels, i.e. 1 divided by the number of classes.
#'     Weights are ignored, if present.
#'   }
#'   \item{weighted.sample:}{
#'     Randomly predict a label, with probability estimated from the training distribution.
#'     For consistency, probabilities are 1 for the sampled label and 0 for all other labels.
#'     For weighted data, sample weights are used to weight the class labels.
#'   }
#' }
#'
#' @templateVar id classif.featureless
#' @template learner
#'
#' @template seealso_learner
#' @export
LearnerClassifFeatureless = R6Class(
  "LearnerClassifFeatureless",
  inherit = LearnerClassif,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        method = p_fct(c("mode", "sample", "weighted.sample"), default = "mode", tags = "predict")
      )
      ps$set_values(method = "mode")
      super$initialize(
        id = "classif.featureless",
        feature_types = mlr_reflections$task_feature_types,
        predict_types = c("response", "prob"),
        param_set = ps,
        properties = c(
          "featureless",
          "twoclass",
          "multiclass",
          "missings",
          "importance",
          "selected_features",
          "weights"
        ),
        label = "Featureless Classification Learner",
        man = "mlr3::mlr_learners_classif.featureless",
      )
    },

    #' @description
    #' All features have a score of `0` for this learner.
    #' @return Named `numeric()`.
    importance = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      fn = self$model$features
      named_vector(fn, 0)
    },

    #' @description
    #' Selected features are always the empty set for this learner.
    #' @return `character(0)`.
    selected_features = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      character()
    }
  ),

  private = list(
    .train = function(task) {
      weights = NULL
      counts_table = data.table(truth = task$truth(), weights = private$.get_weights(task, 1.0))[,
        list(weights = sum(weights)),
        by = "truth"
      ]
      tab = set_names(counts_table$weights, counts_table$truth)
      set_class(list(tab = tab, features = task$feature_names), "classif.featureless_model")
    },

    .predict = function(task) {
      pv = self$param_set$get_values(tags = "predict")
      tab = self$model$tab
      n = task$nrow
      labels = names(tab)
      response = prob = NULL

      if (self$predict_type == "response") {
        response = switch(
          pv$method,
          mode = sample(labels[tab == max(tab)], n, replace = TRUE),
          sample = sample(labels, n, replace = TRUE),
          weighted.sample = sample(labels, n, replace = TRUE, prob = tab)
        )
      } else if (self$predict_type == "prob") {
        prob = switch(
          pv$method,
          mode = matrix(unname(tab) / sum(tab), nrow = n, ncol = length(tab), byrow = TRUE),
          sample = matrix(1 / length(tab), nrow = n, ncol = length(tab)),
          weighted.sample = diag(length(tab))[sample(seq_along(tab), n, replace = TRUE, prob = tab), , drop = FALSE]
        )
        colnames(prob) = labels
      }

      list(response = response, prob = prob)
    }
  )
)

#' @include mlr_learners.R
mlr_learners$add("classif.featureless", function() LearnerClassifFeatureless$new())

#' @title Regression Tree Learner
#'
#' @name mlr_learners_regr.rpart
#' @include LearnerRegr.R
#'
#' @description
#' A [LearnerRegr] for a regression tree implemented in [rpart::rpart()] in package \CRANpkg{rpart}.
#'
#' @section Initial parameter values:
#' * Parameter `xval` is initialized to 0 in order to save some computation time.
#'
#' @section Custom mlr3 parameters:
#' * Parameter `model` has been renamed to `keep_model`.
#'
#' @templateVar id regr.rpart
#' @template learner
#'
#' @references
#' `r format_bib("breiman_1984")`
#'
#' @template seealso_learner
#' @export
LearnerRegrRpart = R6Class(
  "LearnerRegrRpart",
  inherit = LearnerRegr,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      # fmt: skip
      ps = ps(
        cp             = p_dbl(0, 1, default = 0.01, tags = "train"),
        keep_model     = p_lgl(default = FALSE, tags = "train"),
        maxcompete     = p_int(0L, default = 4L, tags = "train"),
        maxdepth       = p_int(1L, 30L, default = 30L, tags = "train"),
        maxsurrogate   = p_int(0L, default = 5L, tags = "train"),
        minbucket      = p_int(1L, tags = "train"),
        minsplit       = p_int(1L, default = 20L, tags = "train"),
        surrogatestyle = p_int(0L, 1L, default = 0L, tags = "train"),
        usesurrogate   = p_int(0L, 2L, default = 2L, tags = "train"),
        xval           = p_int(0L, default = 10L, init = 0L, tags = "train")
      )

      super$initialize(
        id = "regr.rpart",
        feature_types = c("logical", "integer", "numeric", "factor", "ordered"),
        predict_types = "response",
        packages = "rpart",
        param_set = ps,
        properties = c("weights", "missings", "importance", "selected_features"),
        label = "Regression Tree",
        man = "mlr3::mlr_learners_regr.rpart"
      )
    },

    #' @description
    #' The importance scores are extracted from the model slot `variable.importance`.
    #' @return Named `numeric()`.
    importance = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      # importance is only present if there is at least one split
      sort(self$model$variable.importance %??% set_names(numeric()), decreasing = TRUE)
    },

    #' @description
    #' Selected features are extracted from the model slot `frame$var`.
    #' @return `character()`.
    selected_features = function() {
      if (is.null(self$model)) {
        error_learner("No model stored")
      }
      setdiff(self$model$frame$var, "<leaf>")
    }
  ),

  private = list(
    .train = function(task) {
      pv = self$param_set$get_values(tags = "train")
      names(pv) = replace(names(pv), names(pv) == "keep_model", "model")
      pv$weights = private$.get_weights(task)
      invoke(rpart::rpart, formula = task$formula(), data = task$data(), .args = pv, .opts = allow_partial_matching)
    },

    .predict = function(task) {
      pv = self$param_set$get_values(tags = "predict")
      newdata = task$data(cols = task$feature_names)
      response = invoke(predict, self$model, newdata = newdata, .opts = allow_partial_matching, .args = pv)
      result = list(response = unname(response))
      if (self$predict_raw) {
        result$raw = response
      }
      result
    }
  )
)

#' @export
# nolint next
default_values.LearnerRegrRpart = function(x, search_space, task, ...) {
  special_defaults = list(
    minbucket = round(20 / 3)
  )
  defaults = insert_named(default_values(x$param_set), special_defaults)
  defaults[search_space$ids()]
}

#' @include mlr_learners.R
mlr_learners$add("regr.rpart", function() LearnerRegrRpart$new())

#' @title Moons Classification Task Generator
#'
#' @name mlr_task_generators_moons
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] creating two interleaving half circles ("moons") as
#' binary classification problem.
#'
#' @templateVar id moons
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("moons")
#' plot(generator, n = 200)
#'
#' task = generator$generate(200)
#' str(task$data())
TaskGeneratorMoons = R6Class(
  "TaskGeneratorMoons",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        sigma = p_dbl(0, tags = "required")
      )
      ps$set_values(sigma = 1)

      super$initialize(
        id = "moons",
        task_type = "classif",
        param_set = ps,
        label = "Moons Classification",
        man = "mlr3::mlr_task_generators_moons"
      )
    },

    #' @description
    #' Creates a simple plot of generated data.
    #' @param n (`integer(1)`)\cr
    #'   Number of samples to draw for the plot. Default is 200.
    #' @param pch (`integer(1)`)\cr
    #'   Point char. Passed to [plot()].
    #' @param ... (any)\cr
    #'   Additional arguments passed to [plot()].
    plot = function(n = 200L, pch = 19L, ...) {
      tab = private$.generate_obj(n)
      plot(tab$x1, tab$x2, pch = pch, col = tab$y, ...)
    }
  ),

  private = list(
    .generate_obj = function(n) {
      sigma = self$param_set$values$sigma

      n1 = n %/% 2L
      n2 = n - n1
      mu = c(rep(-2.5, n1), rep(2.5, n2))
      x = c(runif(n1, 0, pi), runif(n2, pi, 2 * pi))

      data.table(
        y = factor(rep(c("A", "B"), c(n1, n2)), levels = c("A", "B")),
        x1 = 5 * cos(x) + rnorm(n, mean = mu, sd = sigma),
        x2 = 10 * sin(x) + rnorm(n, mean = mu, sd = sigma)
      )
    },

    .generate = function(n) {
      tab = private$.generate_obj(n)
      TaskClassif$new(sprintf("%s_%i", self$id, n), tab, target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("moons", function() TaskGeneratorMoons$new())

#' @title Cassini Classification Task Generator
#'
#' @name mlr_task_generators_cassini
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] for the cassini task in [mlbench::mlbench.cassini()].
#'
#' @templateVar id cassini
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("cassini")
#' plot(generator, n = 200)
#'
#' task = generator$generate(200)
#' str(task$data())
TaskGeneratorCassini = R6Class(
  "TaskGeneratorCassini",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        relsize1 = p_int(1L, default = 2L),
        relsize2 = p_int(1L, default = 2L),
        relsize3 = p_int(1L, default = 1L)
      )

      super$initialize(
        id = "cassini",
        "classif",
        "mlbench",
        ps,
        label = "Cassini Classification",
        man = "mlr3::mlr_task_generators_cassini"
      )
    },

    #' @description
    #' Creates a simple plot of generated data.
    #' @param n (`integer(1)`)\cr
    #'   Number of samples to draw for the plot. Default is 200.
    #' @param pch (`integer(1)`)\cr
    #'   Point char. Passed to [plot()].
    #' @param ... (any)\cr
    #'   Additional arguments passed to [plot()].
    plot = function(n = 200L, pch = 19L, ...) {
      plot(private$.generate_obj(n), pch = pch, ...)
    }
  ),

  private = list(
    .generate_obj = function(n) {
      pv = self$param_set$values
      relsize = c(pv$relsize1 %??% 2L, pv$relsize2 %??% 2L, pv$relsize3 %??% 1L)
      invoke(mlbench::mlbench.cassini, n = n, .args = list(relsize = relsize), .opts = allow_partial_matching)
    },

    .generate = function(n) {
      obj = private$.generate_obj(n)
      TaskClassif$new(sprintf("%s_%i", self$id, n), convert_mlbench(obj), target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("cassini", function() TaskGeneratorCassini$new())

#' @title Convert to a Prediction
#'
#' @description
#' Convert object to a [Prediction] or a list of [Prediction].
#'
#' @inheritParams as_prediction_data
#'
#' @return [Prediction].
#' @export
as_prediction = function(x, check = FALSE, ...) {
  if (is.null(x)) {
    return(list())
  }

  UseMethod("as_prediction")
}

#' @rdname as_prediction
#' @export
# nolint next
as_prediction.Prediction = function(x, check = FALSE, ...) {
  x
}


#' @rdname as_prediction
#' @export
# nolint next
as_prediction.PredictionDataClassif = function(x, check = FALSE, ...) {
  invoke(PredictionClassif$new, check = check, .args = x)
}


#' @rdname as_prediction
#' @export
# nolint next
as_prediction.PredictionDataRegr = function(x, check = FALSE, ...) {
  invoke(PredictionRegr$new, check = check, .args = x)
}


#' @rdname as_prediction
#' @template param_predict_sets
#' @export
as_predictions = function(x, predict_sets = "test", ...) {
  UseMethod("as_predictions")
}

#' @rdname as_prediction
#' @export
# nolint next
as_predictions.list = function(x, predict_sets = "test", ...) {
  result = replicate(length(x), list())
  ii = lengths(x) > 0L
  result[ii] = map(x[ii], function(li) {
    assert_list(li, "PredictionData")
    li = discard(li[predict_sets], is.null)
    if (length(li) == 0L) {
      return(list())
    }

    if (length(li) == 1L) {
      combined = li[[1L]]
    } else {
      combined = do.call(c, li)
    }
    as_prediction(combined, check = FALSE)
  })
  result
}

#' @title Resampling Class
#'
#' @description
#' This is the abstract base class for resampling objects like [ResamplingCV] and [ResamplingBootstrap].
#'
#' The objects of this class define how a task is partitioned for resampling (e.g., in [resample()] or [benchmark()]),
#' using a set of hyperparameters such as the number of folds in cross-validation.
#'
#' Resampling objects can be instantiated on a [Task], which applies the strategy on the task and manifests in a
#' fixed partition of `row_ids` of the [Task].
#'
#' Predefined resamplings are stored in the [dictionary][mlr3misc::Dictionary] [mlr_resamplings],
#' e.g. [`cv`][mlr_resamplings_cv] or [`bootstrap`][mlr_resamplings_bootstrap].
#'
#'
#' @template param_id
#' @template param_param_set
#' @template param_label
#' @template param_man
#'
#' @section Stochasticity & Reproducibility:
#' The [`Resampling`] class only defines an abstract resampling strategy.
#' Concrete data splits are obtained by calling `$instantiate()` on a [`Task`].
#' To ensure reproducibility of results, you need to call `set.seed` before doing so.
#' Note that [`benchmark_grid`] internally does instantiate resamplings, so you need to set
#' the seed before calling it.
#'
#' @section Stratification:
#' All derived classes support stratified sampling.
#' The stratification variables are assumed to be discrete and must be stored in the [Task]
#' with column role `"stratum"`.
#' In case of multiple stratification variables,
#' each combination of the values of the stratification variables forms a strata.
#'
#' First, the observations are divided into subpopulations based one or multiple stratification
#' variables (assumed to be discrete), c.f. `task$strata`.
#'
#' Second, the sampling is performed in each of the `k` subpopulations separately.
#' Each subgroup is divided into `iter` training sets and `iter` test sets by the derived `Resampling`.
#' These sets are merged based on their iteration number:
#' all training sets from all subpopulations with iteration 1 are combined,
#' then all training sets with iteration 2, and so on.
#' Same is done for all test sets.
#' The merged sets can be accessed via `$train_set(i)` and `$test_set(i)`, respectively.
#' Note that this procedure can lead to set sizes that are slightly different from those
#' without stratification.
#'
#'
#' @section Grouping / Blocking:
#' All derived classes support grouping of observations.
#' The grouping variable is assumed to be discrete and must be stored in the [Task] with column role `"group"`.
#'
#' Observations in the same group are treated like a "block" of observations which must be kept together.
#' These observations either all go together into the training set or together into the test set.
#'
#' The sampling is performed by the derived [Resampling] on the grouping variable.
#' Next, the grouping information is replaced with the respective row ids to generate training and test sets.
#' The sets can be accessed via `$train_set(i)` and `$test_set(i)`, respectively.
#'
#' @section Inheriting:
#' It is possible to overwrite both `private$.get_instance()` to have full control,
#' or only `private$.sample()` when one wants to use the pre-defined mechanism for stratification and grouping.
#'
#' @template seealso_resampling
#' @export
#' @examples
#' r = rsmp("subsampling")
#'
#' # Default parametrization
#' r$param_set$values
#'
#' # Do only 3 repeats on 10% of the data
#' r$param_set$set_values(ratio = 0.1, repeats = 3)
#' r$param_set$values
#'
#' # Instantiate on penguins task
#' task = tsk("penguins")
#' r$instantiate(task)
#'
#' # Extract train/test sets
#' train_set = r$train_set(1)
#' print(train_set)
#' intersect(train_set, r$test_set(1))
#'
#' # Another example: 10-fold CV
#' r = rsmp("cv")$instantiate(task)
#' r$train_set(1)
#'
#' # Stratification
#' task = tsk("pima")
#' prop.table(table(task$truth())) # moderately unbalanced
#' task$col_roles$stratum = task$target_names
#'
#' r = rsmp("subsampling")
#' r$instantiate(task)
#' prop.table(table(task$truth(r$train_set(1)))) # roughly same proportion
Resampling = R6Class(
  "Resampling",
  public = list(
    #' @field instance (any)\cr
    #'   During `instantiate()`, the instance is stored in this slot in an arbitrary format.
    #'   Note that if a grouping variable is present in the [Task], a [Resampling] may operate on the
    #'   group ids internally instead of the row ids (which may lead to confusion).
    #'
    #'   It is advised to not work directly with the `instance`, but instead only use the getters
    #'   `$train_set()` and `$test_set()`.
    instance = NULL,

    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param duplicated_ids (`logical(1)`)\cr
    #'   Set to `TRUE` if this resampling strategy may have duplicated row ids in a single training set or test set.
    #'
    #' Note that this object is typically constructed via a derived classes, e.g. [ResamplingCV] or [ResamplingHoldout].
    initialize = function(id, param_set = ps(), duplicated_ids = FALSE, label = NA_character_, man = NA_character_) {
      private$.id = assert_id(id)
      private$.label = assert_string(label, na.ok = TRUE)
      private$.param_set = assert_param_set(param_set)
      private$.duplicated_ids = assert_flag(duplicated_ids)
      private$.man = assert_string(man, na.ok = TRUE)
      private$.task_hash = NA_character_
      private$.task_row_hash = NA_character_
      private$.task_nrow = NA_integer_
    },

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s>", class(self)[1L])
    },

    #' @description
    #' Printer.
    #' @param ... (ignored).
    print = function(...) {
      msg_h = if (is.null(self$label) || is.na(self$label)) "" else paste0(": ", self$label)
      cat_cli({
        cli_h1("{.cls {class(self)[1L]}} {msg_h}")
        cli_li("Iterations: {.val {self$iters}}")
        cli_li("Instantiated: {.val {self$is_instantiated}}")
        cli_li("Parameters: {as_short_string(self$param_set$values, 1000L)}")
      })
    },

    #' @description
    #' Opens the corresponding help page referenced by field `$man`.
    help = function() {
      open_help(self$man)
    },

    #' @description
    #' Materializes fixed training and test splits for a given task and stores them in `r$instance`
    #' in an arbitrary format.
    #'
    #' @param task ([Task])\cr
    #'   Task used for instantiation.
    #'
    #' @return
    #' Returns the object itself, but modified **by reference**.
    #' You need to explicitly `$clone()` the object beforehand if you want to keep
    #' the object in its previous state.
    #' @examples
    #' task = tsk("penguins")
    #' resampling = rsmp("holdout")
    #' resampling$instantiate(task)
    instantiate = function(task) {
      task = assert_task(as_task(task))
      private$.hash = NULL
      self$instance = private$.get_instance(task)
      private$.task_hash = task$hash
      private$.task_row_hash = task$row_hash
      private$.task_nrow = task$nrow
      invisible(self)
    },

    #' @description
    #' Returns the row ids of the i-th training set.
    #'
    #' @param i (`integer(1)`)\cr
    #'   Iteration.
    #' @return (`integer()`) of row ids.
    #' @examples
    #' task = tsk("penguins")
    #' resampling = rsmp("holdout")$instantiate(task)
    #' resampling$train_set(1)
    train_set = function(i) {
      private$.get_set(private$.get_train, i)
    },

    #' @description
    #' Returns the row ids of the i-th test set.
    #'
    #' @param i (`integer(1)`)\cr
    #'   Iteration.
    #' @return (`integer()`) of row ids.
    #' @examples
    #' task = tsk("penguins")
    #' resampling = rsmp("holdout")$instantiate(task)
    #' resampling$test_set(1)
    test_set = function(i) {
      private$.get_set(private$.get_test, i)
    }
  ),

  active = list(
    #' @template field_id
    id = function(rhs) {
      if (missing(rhs)) {
        return(private$.id)
      }

      private$.hash = NULL
      private$.id = assert_id(rhs)
    },

    #' @field is_instantiated (`logical(1)`)\cr
    #'   Is `TRUE` if the resampling has been instantiated.
    is_instantiated = function(rhs) {
      assert_ro_binding(rhs)
      !is.null(self$instance)
    },

    #' @field hash (`character(1)`)\cr
    #' Hash (unique identifier) for this object.
    #' If the object has not been instantiated yet, `NA_character_` is returned.
    #' The hash is calculated based on the class name, the id, the parameter set, and the instance.
    hash = function(rhs) {
      assert_ro_binding(rhs)
      if (!self$is_instantiated) {
        return(NA_character_)
      }

      if (is.null(private$.hash)) {
        private$.hash = calculate_hash(list(class(self), self$id, self$param_set$values, self$instance))
      }

      private$.hash
    },

    #' @template field_label
    label = function(rhs) {
      if (missing(rhs)) {
        return(private$.label)
      }
      private$.label = assert_string(rhs, na.ok = TRUE)
    },

    #' @template field_param_set
    param_set = function(rhs) {
      if (!missing(rhs) && !identical(rhs, private$.param_set)) {
        error_input("param_set is read-only.")
      }
      private$.param_set
    },

    #' @field task_hash (`character(1)`)\cr
    #'   The hash of the [Task] which was passed to `r$instantiate()`.
    task_hash = function(rhs) {
      if (missing(rhs)) {
        return(private$.task_hash)
      }
      private$.task_hash = assert_string(rhs, na.ok = TRUE)
    },

    #' @field task_row_hash (`character(1)`)\cr
    #'   The hash of the row ids of the [Task] which was passed to `r$instantiate()`.
    task_row_hash = function(rhs) {
      if (missing(rhs)) {
        return(private$.task_row_hash)
      }
      private$.task_row_hash = assert_string(rhs, na.ok = TRUE)
    },

    #' @field task_nrow (`integer(1)`)\cr
    #'   The number of observations of the [Task] which was passed to `r$instantiate()`.
    task_nrow = function(rhs) {
      if (missing(rhs)) {
        return(private$.task_nrow)
      }
      private$.task_nrow = assert_int(rhs, na.ok = TRUE, coerce = TRUE)
    },

    #' @field duplicated_ids (`logical(1)`)\cr
    #'   If `TRUE`, duplicated rows can occur within a single training set or within a single test set.
    #'   E.g., this is `TRUE` for Bootstrap, and `FALSE` for cross-validation.
    #'   Only used internally.
    duplicated_ids = function(rhs) {
      if (missing(rhs)) {
        return(private$.duplicated_ids)
      }
      private$.duplicated_ids = assert_flag(rhs)
    },

    #' @template field_man
    man = function(rhs) {
      if (missing(rhs)) {
        return(private$.man)
      }
      private$.man = assert_string(rhs, na.ok = TRUE)
    }
  ),

  private = list(
    .primary_iters = NULL,
    .id = NULL,
    .label = NULL,
    .param_set = NULL,
    .task_hash = NULL,
    .task_row_hash = NULL,
    .task_nrow = NULL,
    .duplicated_ids = NULL,
    .man = NULL,
    .hash = NULL,
    .groups = NULL,

    .get_instance = function(task) {
      strata = task$strata
      groups = task$groups
      if (is.null(strata)) {
        if (is.null(groups)) {
          private$.sample(task$row_ids, task = task)
        } else {
          private$.groups = groups
          private$.sample(unique(groups$group), task = task)
        }
      } else {
        if (!is.null(groups)) {
          error_input("Cannot combine stratification with grouping")
        }
        private$.combine(lapply(strata$row_id, private$.sample, task = task))
      }
    },

    .get_set = function(getter, i) {
      if (!self$is_instantiated) {
        error_input("Resampling '%s' has not been instantiated yet", self$id)
      }
      i = assert_int(i, lower = 1L, upper = self$iters, coerce = TRUE)
      ids = getter(i)

      if (is.null(private$.groups)) {
        return(ids)
      }

      private$.groups[list(ids), on = "group", allow.cartesian = TRUE][[1L]]
    },

    deep_clone = function(name, value) {
      if (name == ".param_set") value$clone(deep = TRUE) else value
    }
  )
)


#' @export
# nolint next
as.data.table.Resampling = function(x, ...) {
  assert_resampling(x, instantiated = TRUE)
  iterations = seq_len(x$iters)

  tab = rbindlist(
    list(
      map_dtr(iterations, function(i) list(row_id = x$train_set(i)), .idcol = "iteration"),
      map_dtr(iterations, function(i) list(row_id = x$test_set(i)), .idcol = "iteration")
    ),
    idcol = "set"
  )
  set(tab, j = "set", value = factor(c("train", "test")[tab$set], levels = c("train", "test")))
  setkeyv(tab, c("set", "iteration"))[]
}

test_that("insample has no duplicated ids", {
  r = rsmp("insample")
  expect_identical(r$duplicated_ids, FALSE)
})

test_that("stratification", {
  data = data.table(
    y = factor(rep(letters[1:2], times = c(90, 10))),
    x1 = runif(100),
    x2 = rep(LETTERS[1:2], times = c(50, 50))
  )
  b = as_data_backend(data)
  task = TaskClassif$new("stratify_data", b, target = "y")
  task$col_roles$stratum = task$target_names

  r = rsmp("insample")
  r$instantiate(task)

  i = 1L
  expect_set_equal(r$train_set(i), task$row_ids)
  expect_set_equal(r$test_set(i), task$row_ids)
})

test_that("grouping", {
  r = rsmp("insample")
  expect_grouping_works(r)
})

test_that("loo has no duplicated ids", {
  r = rsmp("loo")
  expect_identical(r$duplicated_ids, FALSE)
})

test_that("stratification", {
  data = data.table(
    y = factor(rep(letters[1:2], times = c(90, 10))),
    x1 = runif(100),
    x2 = rep(LETTERS[1:2], times = c(50, 50))
  )
  b = as_data_backend(data)
  task = TaskClassif$new("stratify_data", b, target = "y")
  task$col_roles$stratum = task$target_names

  r = rsmp("loo")
  r$instantiate(task)

  i = 1L
  expect_data_table(task$data(r$train_set(i)), nrows = 99)
  expect_data_table(task$data(r$test_set(i)), nrows = 1)
})

test_that("grouping", {
  r = rsmp("loo")
  expect_grouping_works(r)
})

test_that("subsampling has no duplicated ids", {
  r = rsmp("subsampling")
  expect_identical(r$duplicated_ids, FALSE)
})

test_that("stratification", {
  data = data.table(
    y = factor(rep(letters[1:2], times = c(90, 10))),
    x1 = runif(100),
    x2 = rep(LETTERS[1:2], times = c(50, 50))
  )
  b = as_data_backend(data)
  task = TaskClassif$new("stratify_data", b, target = "y")
  task$col_roles$stratum = task$target_names

  r = rsmp("subsampling", ratio = 0.5, repeats = 3)
  r$instantiate(task)

  for (i in seq_len(r$iters)) {
    expect_equal(task$data(r$train_set(i))[y == "a", .N], 45)
    expect_equal(task$data(r$train_set(i))[y == "b", .N], 5)
    expect_equal(task$data(r$test_set(i))[y == "a", .N], 45)
    expect_equal(task$data(r$test_set(i))[y == "b", .N], 5)
  }
})

test_that("grouping", {
  r = rsmp("subsampling", ratio = 0.5, repeats = 3)
  expect_grouping_works(r)
})

test_that("bootstrap has duplicated ids", {
  r = rsmp("bootstrap")
  expect_identical(r$duplicated_ids, TRUE)
})

test_that("stratification", {
  data = data.table(
    y = factor(rep(letters[1:2], times = c(90, 10))),
    x1 = runif(100),
    x2 = rep(LETTERS[1:2], times = c(50, 50))
  )
  b = as_data_backend(data)
  task = TaskClassif$new("stratify_data", b, target = "y")
  task$col_roles$stratum = task$target_names

  r = rsmp("bootstrap", ratio = 1, repeats = 3)
  r$instantiate(task)

  for (i in seq_len(r$iters)) {
    expect_equal(task$data(r$train_set(i))[y == "a", .N], 90)
    expect_equal(task$data(r$train_set(i))[y == "b", .N], 10)
  }
})

test_that("grouping", {
  r = rsmp("bootstrap", ratio = 1, repeats = 3)
  expect_grouping_works(r)
})

test_that("#518", {
  # nolint next
  with_seed(36L, {
    test_df = data.frame(
      feat1 = runif(3000),
      feat2 = runif(3000),
      target = runif(3000),
      group = sample(paste0("g", 1:100), 3000, replace = TRUE)
    )
    task = TaskRegr$new(id = "test", backend = test_df, target = "target")
    task$col_roles$group = "group"
    task$col_roles$feature = c("feat1", "feat2")
    rb = rsmp("bootstrap", repeats = 10, ratio = 1)
    rb$instantiate(task)
  })

  min.len = if (getRversion() >= "3.6.0") 3001L else NULL
  expect_integer(rb$train_set(2), min.len = min.len)
})

test_that("repeated cv has no duplicated ids", {
  r = rsmp("repeated_cv")
  expect_identical(r$duplicated_ids, FALSE)
})

test_that("folds first, then repetitions", {
  task = tsk("iris")
  rrcv = rsmp("repeated_cv", repeats = 2, folds = 3)
  rrcv$instantiate(task)

  expect_set_equal(c(rrcv$train_set(1), rrcv$test_set(1)), task$row_ids)
  expect_integer(intersect(intersect(rrcv$test_set(1), rrcv$test_set(2)), rrcv$test_set(3)), len = 0L)

  expect_equal(rrcv$folds(seq_len(rrcv$iters)), rep(1:3, 2))
  expect_equal(rrcv$repeats(seq_len(rrcv$iters)), rep(1:2, each = 3))
})

test_that("stratification", {
  data = data.table(
    y = factor(rep(letters[1:2], times = c(90, 10))),
    x1 = runif(100),
    x2 = rep(LETTERS[1:2], times = c(50, 50))
  )
  task = TaskClassif$new("stratify_data", data, target = "y")
  task$col_roles$stratum = task$target_names

  r = rsmp("repeated_cv", folds = 5, repeats = 2)
  r$instantiate(task)

  for (i in seq_len(r$iters)) {
    expect_equal(task$data(r$train_set(i))[y == "a", .N], 72)
    expect_equal(task$data(r$train_set(i))[y == "b", .N], 8)
    expect_equal(task$data(r$test_set(i))[y == "a", .N], 18)
    expect_equal(task$data(r$test_set(i))[y == "b", .N], 2)
  }
})

test_that("grouping", {
  r = rsmp("repeated_cv", folds = 5, repeats = 2)
  expect_grouping_works(r)
})

test_that("cv has no duplicated ids", {
  r = rsmp("cv")
  expect_identical(r$duplicated_ids, FALSE)
})

test_that("split into evenly sized groups", {
  task = tsk("iris")
  r = rsmp("cv", folds = 3)
  r$instantiate(task)
  expect_equal(uniqueN(lengths(lapply(1:3, r$train_set))), 1L)
})

test_that("stratification", {
  data = data.table(
    y = factor(rep(letters[1:2], times = c(90, 10))),
    x1 = runif(100),
    x2 = rep(LETTERS[1:2], times = c(50, 50))
  )
  b = as_data_backend(data)
  task = TaskClassif$new("stratify_data", b, target = "y")
  task$col_roles$stratum = task$target_names

  r = rsmp("cv", folds = 5L)
  r$instantiate(task)

  for (i in seq_len(r$iters)) {
    expect_equal(task$data(r$train_set(i))[y == "a", .N], 72)
    expect_equal(task$data(r$train_set(i))[y == "b", .N], 8)
    expect_equal(task$data(r$test_set(i))[y == "a", .N], 18)
    expect_equal(task$data(r$test_set(i))[y == "b", .N], 2)
  }
})

test_that("grouping", {
  r = rsmp("cv", folds = 3)
  expect_grouping_works(r)
})

#' @title Measure Class
#'
#' @include mlr_reflections.R
#'
#' @description
#' This is the abstract base class for measures like [MeasureClassif] and [MeasureRegr].
#'
#' Measures are classes tailored around two functions doing the work:
#'
#' 1. A function `$score()` which quantifies the performance on a [Prediction] object, so a set
#' of predicted observation via a scalar number -- usually an aggregate of losses on the contained observations,
#' by comparing the truth and prediction columns in the prediction object.
#' 2. A function `$aggregator()` which combines multiple performance scores returned by
#' `$score()` obtained in different resampling iterations to a scalar performance value associated
#' with the complete resampling -- usually by averaging or summing.
#'
#' In addition to these two functions, meta-information about the performance measure is stored.
#'
#' Predefined measures are stored in the [dictionary][mlr3misc::Dictionary] [mlr_measures],
#' e.g. [`classif.auc`][mlr_measures_classif.auc] or [`time_train`][mlr_measures_time_train].
#' Many of the measures in \pkg{mlr3} are implemented in \CRANpkg{mlr3measures} as ordinary functions.
#'
#' A guide on how to extend \CRANpkg{mlr3} with custom measures can be found in the
#' [mlr3book](https://mlr3book.mlr-org.com).
#' @section Inheriting:
#' For some measures (such as confidence intervals from `mlr3inferr`) it is necessary that a measure
#' returns more than one value.
#' In such cases it is necessary to overwrite the public methods `$aggregate()` and/or `$score()`
#' to return a named `numeric()`
#' where at least one of its names corresponds to the `id` of the measure itself.
#'
#' @section Weights:
#'
#' Many measures support observation weights, indicated by their property `"weights"`.
#' The weights are stored in the [Task] where the column role `weights_measure` needs to be assigned
#' to a single numeric column.
#' The weights are automatically used if found in the task,
#' this can be disabled by setting the field `use_weights` to `"ignore"`.
#' See the description of `use_weights` for more information.
#'
#' If the measure is set-up to use weights but the task does not have a designated `weights_measure` column,
#' an unweighted version is calculated instead.
#' The weights do not necessarily need to sum up to 1, they are normalized by the measure if necessary.
#'
#' Most measures are so-called decomposable loss functions where a point-wise loss is computed
#' and then either mean-aggregated or summed over the test set.
#' For measures that do mean-aggregation, weights are typically used to compute the weighted mean,
#' which normalizes weights to sum to 1.
#' Measures that use sum-aggregation do not normalize weights and instead multiply individual losses
#' with the given weights.
#' See the documentation of specific measures for more details.
#'
#' @section Missing Values during Scoring:
#'
#' Many measurements cannot be calculated if the test set or predictions are unfortunate,
#' for example because a denominator is 0.
#' This typically occurs during (binary) classification if some entries of the confusion matrix are 0.
#' For this reason, many measures which originate in \CRANpkg{mlr3measures} allow to change the
#' default missing value (`NaN`) via the field `na_value`.
#'
#' If you encounter missing values in a compound object like a [ResampleResult] or [BenchmarkResult]
#' during scoring or aggregating, simply removing iterations with missing values is statistically
#' arguable (but technically possible by providing a custom aggregation function which handles
#' missing values, e.g. `function(x) mean(x, na.rm = TRUE)`).
#' Instead, consider stratification on the target of the [Task] to work around missing values.
#' Switching to micro averaging in the [Resampling] can also be a solution here.
#'
#' @template param_id
#' @template param_param_set
#' @template param_range
#' @template param_minimize
#' @template param_average
#' @template param_aggregator
#' @template param_task_type
#' @template param_predict_type
#' @template param_measure_properties
#' @template param_predict_sets
#' @template param_task_properties
#' @template param_packages
#' @template param_label
#' @template param_man
#' @param trafo (`list()` | `NULL`)\cr
#'   An optional list with two elements, containing the transformation `"fn"` and its derivative `"deriv"`.
#'   The transformation function is the function that is applied after aggregating the pointwise losses, i.e.
#'   this requires an `$obs_loss` to be present. An example is `sqrt` for RMSE.
#'
#' @template seealso_measure
#' @export
Measure = R6Class(
  "Measure",
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' Note that this object is typically constructed via a derived classes, e.g. [MeasureClassif] or [MeasureRegr].
    initialize = function(
      id,
      task_type = NA,
      param_set = ps(),
      range = c(-Inf, Inf),
      minimize = NA,
      average = "macro",
      aggregator = NULL,
      properties = character(),
      predict_type = "response",
      predict_sets = "test",
      task_properties = character(),
      packages = character(),
      label = NA_character_,
      man = NA_character_,
      trafo = NULL
    ) {
      private$.id = assert_id(id)
      private$.label = assert_string(label, na.ok = TRUE)
      private$.task_type = task_type
      private$.param_set = assert_param_set(param_set)
      private$.range = assert_range(range)
      private$.minimize = assert_flag(minimize, na.ok = TRUE)
      private$.average = assert_choice(average, c("micro", "macro", "custom", "macro_weighted"))
      private$.aggregator = assert_function(aggregator, null.ok = TRUE)
      private$.trafo_fun = assert_list(trafo, len = 2L, types = "function", null.ok = TRUE)
      if (!is.null(private$.trafo_fun)) {
        assert_permutation(names(trafo), c("fn", "deriv"))
      }

      if (!is_scalar_na(task_type)) {
        assert_choice(task_type, mlr_reflections$task_types$type)
        assert_subset(properties, mlr_reflections$measure_properties[[task_type]])
        assert_choice(predict_type, names(mlr_reflections$learner_predict_types[[task_type]]))
        assert_subset(task_properties, mlr_reflections$task_properties[[task_type]])
      } else {
        assert_subset(properties, unique(unlist(mlr_reflections$measure_properties, use.names = FALSE)))
      }

      private$.properties = unique(properties)

      if ("weights" %in% properties) {
        private$.use_weights = "use"
      } else if ("requires_no_prediction" %in% properties) {
        private$.use_weights = "ignore"
      } else {
        private$.use_weights = "error"
      }

      private$.predict_type = predict_type
      private$.predict_sets = assert_subset(
        predict_sets,
        mlr_reflections$predict_sets,
        empty.ok = "requires_no_prediction" %chin% properties
      )
      private$.task_properties = task_properties
      private$.packages = union("mlr3", assert_character(packages, any.missing = FALSE, min.chars = 1L))
      private$.man = assert_string(man, na.ok = TRUE)
      private$.check_prerequisites = "warn"

      check_packages_installed(packages, msg = sprintf("Package '%%s' required but not installed for Measure '%s'", id))
    },

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s:%s>", class(self)[1L], self$id)
    },

    #' @description
    #' Printer.
    #' @param ... (ignored).
    print = function(...) {
      msg_h = if (is.null(self$label) || is.na(self$label)) "" else paste0(": ", self$label)
      msg_properties = if (length(self$properties)) self$properties else "-"
      cat_cli({
        cli_h1("{.cls {class(self)[1L]}} ({self$id}){msg_h}")
        cli_li("Packages: {.pkg {self$packages}}")
        cli_li("Range: [{self$range[1L]}, {self$range[2L]}]")
        cli_li("Minimize: {.val {self$minimize}}")
        cli_li("Average: {self$average}")
        cli_li("Parameters: {as_short_string(self$param_set$values, 1000L)}")
        cli_li("Properties: {msg_properties}")
        cli_li("Predict type: {self$predict_type}")
        cli_li("Predict sets: {self$predict_sets}")
        cli_li("Aggregator: {if (is.null(self$aggregator)) 'mean()' else '[user-defined]'}")
      })
    },

    #' @description
    #' Opens the corresponding help page referenced by field `$man`.
    help = function() {
      open_help(self$man)
    },

    #' @description
    #' Takes a [Prediction] and calculates a numeric score.
    #' If the measure if flagged with the properties `"requires_task"`, `"requires_learner"`,
    #' `"requires_model"` or `"requires_train_set"`, you must additionally
    #' pass the respective [Task], the (trained) [Learner] or the training set indices.
    #' This is handled internally during [resample()]/[benchmark()].
    #'
    #' @param prediction ([Prediction]).
    #'
    #' @param task ([Task]).
    #'
    #' @param learner ([Learner]).
    #'
    #' @param train_set (`integer()`).
    #'
    #' @return `numeric(1)`.
    #' @examples
    #' task = tsk("penguins")
    #' learner = lrn("classif.rpart")$train(task)
    #' prediction = learner$predict(task)
    #' msr("classif.ce")$score(prediction)
    score = function(prediction, task = NULL, learner = NULL, train_set = NULL) {
      assert_scorable(self, task = task, learner = learner, prediction = prediction)
      properties = self$properties
      assert_prediction(prediction, null.ok = "requires_no_prediction" %nin% properties)

      # check should be added to assert_measure()
      # except when the checks are superfluous for rr$score() and bmr$score()
      # these checks should be added below
      if ("requires_task" %chin% properties && is.null(task)) {
        error_input("Measure '%s' requires a task", self$id)
      }

      if ("requires_learner" %chin% properties && is.null(learner)) {
        error_input("Measure '%s' requires a learner", self$id)
      }

      if (!is.null(prediction) && !is_scalar_na(self$task_type) && self$task_type != prediction$task_type) {
        error_input("Measure '%s' incompatible with task type '%s'", self$id, prediction$task_type)
      }

      if ("requires_train_set" %chin% properties && is.null(train_set)) {
        error_input("Measure '%s' requires the train_set", self$id)
      }

      score_single_measure(self, task, learner, train_set, prediction)
    },

    #' @description
    #' Aggregates multiple performance scores into a single score, e.g. by using the `aggregator`
    #' function of the measure.
    #'
    #' @param rr [ResampleResult].
    #'
    #' @return `numeric(1)`.
    #' @examples
    #' task = tsk("penguins")
    #' learner = lrn("classif.rpart")
    #' rr = resample(task, learner, rsmp("holdout"))
    #' msr("classif.ce")$aggregate(rr)
    aggregate = function(rr) {
      switch(
        self$average,
        "macro_weighted" = {
          aggregator = self$aggregator %??% weighted.mean
          tab = score_measures(
            rr,
            list(self),
            reassemble = FALSE,
            view = get_private(rr)$.view,
            iters = get_private(rr$resampling)$.primary_iters
          )
          # score_measures constructs both .weights and .samples, since it can work with multiple measures and hence
          # does not depend on an individual measure's `use_weights` setting.
          set_names(aggregator(tab[[self$id]], if (self$use_weights == "use") tab$.weights else tab$.samples), self$id)
        },
        "macro" = {
          aggregator = self$aggregator %??% mean
          tab = score_measures(
            rr,
            list(self),
            reassemble = FALSE,
            view = get_private(rr)$.view,
            iters = get_private(rr$resampling)$.primary_iters
          )
          set_names(aggregator(tab[[self$id]]), self$id)
        },
        "micro" = {
          if (is.null(get_private(rr$resampling)$.primary_iters)) {
            self$score(rr$prediction(self$predict_sets), task = rr$task, learner = rr$learner)
          } else {
            prediction = do.call(c, rr$predictions(self$predict_sets)[get_private(rr$resampling)$.primary_iters])
            self$score(prediction, task = rr$task, learner = rr$learner)
          }
        },
        "custom" = {
          if (
            !is.null(get_private(rr$resampling)$.primary_iters) &&
              "primary_iters" %nin% self$properties &&
              !test_permutation(get_private(rr$resampling)$.primary_iters, seq_len(rr$resampling$iters))
          ) {
            error_input("Resample result has non-NULL primary_iters, but measure '%s' cannot handle them", self$id)
          }
          private$.aggregator(rr)
        }
      )
    },

    #' @description
    #' Calculates the observation-wise loss.
    #' Returns a `numeric()` with one element for each row in the [Prediction].
    #' If there is no observation-wise loss function for the measure, `NA_real_` values are returned.
    #'
    #' @param prediction ([Prediction]).
    #' @param task ([Task]).
    #' @param learner ([Learner]).
    #'
    #' @return `numeric()` with one element for each row in the [Prediction].
    #' @examples
    #' task = tsk("penguins")
    #' learner = lrn("classif.rpart")
    #' learner$train(task)
    #' prediction = learner$predict(task)
    #' msr("classif.ce")$obs_loss(prediction)
    obs_loss = function(prediction, task = NULL, learner = NULL) {
      if (!is_scalar_na(self$task_type) && self$task_type != prediction$task_type) {
        error_input("Measure '%s' incompatible with task type '%s'", self$id, prediction$task_type)
      }

      if ("obs_loss" %nin% self$properties) {
        return(rep(NA_real_, length(prediction$row_ids)))
      }

      private$.obs_loss(prediction, task, learner)
    }
  ),

  active = list(
    #' @template field_predict_sets
    predict_sets = function(rhs) {
      if (!missing(rhs)) {
        private$.predict_sets = assert_subset(
          rhs,
          mlr_reflections$predict_sets,
          empty.ok = "requires_no_prediction" %chin% self$properties
        )
      }
      private$.predict_sets
    },

    #' @field hash (`character(1)`)\cr
    #' Hash (unique identifier) for this object.
    #' The hash is calculated based on the id, the parameter settings, predict sets and the `$score`,
    #' `$average`, `$aggregator`, `$obs_loss`, `$trafo` method.
    #' Measure can define additional fields to be included in the hash by setting the field `$.extra_hash`.
    hash = function(rhs) {
      assert_ro_binding(rhs)
      calculate_hash(
        class(self),
        self$id,
        self$param_set$values,
        private$.score,
        private$.average,
        private$.aggregator,
        private$.obs_loss,
        self$trafo,
        self$predict_sets,
        mget(private$.extra_hash, envir = self),
        private$.use_weights
      )
    },

    #' @field properties (`character()`)\cr
    #' Properties of this measure.
    properties = function(rhs) {
      if (!missing(rhs)) {
        props = if (is.na(self$task_type)) {
          unique(unlist(mlr_reflections$measure_properties, use.names = FALSE))
        } else {
          mlr_reflections$measure_properties[[self$task_type]]
        }
        private$.properties = assert_subset(rhs, props)
      } else {
        private$.properties
      }
    },

    #' @field average (`character(1)`)\cr
    #' Method for aggregation:
    #'
    #' * `"micro"`:
    #'   All predictions from multiple resampling iterations are first combined into a single [Prediction] object.
    #'   Next, the scoring function of the measure is applied on this combined object, yielding a single numeric score.
    #' * `"macro"`:
    #'   The scoring function is applied on the [Prediction] object of each resampling iterations,
    #'   each yielding a single numeric score.
    #'   Next, the scores are combined with the `aggregator` function to a single numerical score.
    #' * `"macro_weighted"`:
    #'   The scoring function is applied on the [Prediction] object of each resampling iterations,
    #'   each yielding a single numeric score.
    #'   Next, the scores are combined with the `aggregator` function to a single numerical score.
    #'   The scores are weighted by the total sample weights (if present, and if `$use_weights` is set to `"use"`),
    #'   or the number of samples in each resampling iteration.
    #' * `"custom"`:
    #'   The measure comes with a custom aggregation method which directly operates on a [ResampleResult].
    average = function(rhs) {
      if (!missing(rhs)) {
        private$.average = assert_choice(rhs, c("micro", "macro", "custom", "macro_weighted"))
      } else {
        private$.average
      }
    },

    #' @field aggregator (`function()`)\cr
    #' Function to aggregate scores computed on different resampling iterations.
    aggregator = function(rhs) {
      if (!missing(rhs)) {
        private$.aggregator = assert_function(rhs, null.ok = TRUE)
      } else {
        private$.aggregator
      }
    },

    #' @field use_weights (`character(1)`)\cr
    #' How to handle weights.
    #' Settings are `"use"`, `"ignore"`, and `"error"`.
    #'
    #' * `"use"`: Weights are used automatically if found in the task, as supported by the measure.
    #' * `"ignore"`: Weights are ignored.
    #' * `"error"`: throw an error if weights are present in the training `Task`.
    #'
    #' For measures with the property `"weights"`, this is initialized as `"use"`.
    #' For measures with the property `"requires_no_prediction"`, this is initialized as `"ignore"`.
    #' For measures that have neither of the properties, this is initialized as `"error"`.
    #' The latter behavior is to avoid cases where a user erroneously assumes that a measure
    #' supports weights when it does not.
    #' For measures that do not support weights,
    #' `use_weights` needs to be set to `"ignore"` if tasks with weights should be handled (by dropping the weights).
    use_weights = function(rhs) {
      if (!missing(rhs)) {
        private$.use_weights = assert_choice(rhs, c(if ("weights" %chin% self$properties) "use", "ignore", "error"))
      } else {
        private$.use_weights
      }
    },

    #' @template field_id
    id = function(rhs) {
      if (missing(rhs)) {
        return(private$.id)
      }
      private$.id = assert_id(rhs)
    },

    #' @template field_label
    label = function(rhs) {
      if (missing(rhs)) {
        return(private$.label)
      }
      private$.label = assert_string(rhs, na.ok = TRUE)
    },

    #' @template field_task_type
    task_type = function(rhs) {
      if (missing(rhs)) {
        return(private$.task_type)
      }
      if (!is_scalar_na(rhs)) {
        assert_choice(rhs, mlr_reflections$task_types$type)
      }
      private$.task_type = rhs
    },

    #' @template field_param_set
    param_set = function(rhs) {
      if (!missing(rhs) && !identical(rhs, private$.param_set)) {
        error_input("param_set is read-only.")
      }
      private$.param_set
    },

    #' @field predict_type (`character(1)`)\cr
    #' Required predict type of the [Learner].
    predict_type = function(rhs) {
      if (missing(rhs)) {
        return(private$.predict_type)
      }
      if (!is_scalar_na(self$task_type)) {
        assert_choice(rhs, names(mlr_reflections$learner_predict_types[[self$task_type]]))
      }
      private$.predict_type = rhs
    },

    #' @field check_prerequisites (`character(1)`)\cr
    #' How to proceed if one of the following prerequisites is not met:
    #'
    #' * wrong predict type (e.g., probabilities required, but only labels available).
    #' * wrong predict set (e.g., learner predicted on training set, but predictions of test set required).
    #' * task properties not satisfied (e.g., binary classification measure on multiclass task).
    #'
    #' Possible values are `"ignore"` (just return `NaN`) and `"warn"` (default,
    #' raise a warning before returning `NaN`).
    check_prerequisites = function(rhs) {
      if (missing(rhs)) {
        return(private$.check_prerequisites)
      }
      private$.check_prerequisites = assert_choice(rhs, c("ignore", "warn"))
    },

    #' @field task_properties (`character()`)\cr
    #' Required properties of the [Task].
    task_properties = function(rhs) {
      if (missing(rhs)) {
        return(private$.task_properties)
      }
      if (!is_scalar_na(self$task_type)) {
        assert_subset(rhs, mlr_reflections$task_properties[[self$task_type]])
      }
      private$.task_properties = rhs
    },

    #' @field range (`numeric(2)`)\cr
    #' Lower and upper bound of possible performance scores.
    range = function(rhs) {
      if (missing(rhs)) {
        return(private$.range)
      }
      private$.range = assert_range(rhs)
    },

    #' @field minimize (`logical(1)`)\cr
    #' If `TRUE`, good predictions correspond to small values of performance scores.
    minimize = function(rhs) {
      if (missing(rhs)) {
        return(private$.minimize)
      }
      private$.minimize = assert_flag(rhs, na.ok = TRUE)
    },

    #' @template field_packages
    packages = function(rhs) {
      if (missing(rhs)) {
        return(private$.packages)
      }
      private$.packages = union("mlr3", assert_character(rhs, any.missing = FALSE, min.chars = 1L))
    },

    #' @template field_man
    man = function(rhs) {
      if (missing(rhs)) {
        return(private$.man)
      }
      private$.man = assert_string(rhs, na.ok = TRUE)
    },

    #' @field trafo (`list()` | `NULL`)\cr
    #' `NULL` or a list with two elements:
    #' * `fn`: the transformation function applied after aggregating
    #'   observation-wise losses (e.g. `sqrt` for RMSE)
    #' * `deriv`: The derivative of the `fn`.
    trafo = function(rhs) {
      if (missing(rhs)) {
        return(private$.trafo_fun)
      }
      rhs = assert_list(rhs, len = 2L, types = "function", null.ok = TRUE)
      if (!is.null(rhs)) {
        assert_permutation(names(rhs), c("fn", "deriv"))
      }
      private$.trafo_fun = rhs
    }
  ),

  private = list(
    .id = NULL,
    .label = NULL,
    .task_type = NULL,
    .param_set = NULL,
    .predict_type = NULL,
    .check_prerequisites = NULL,
    .task_properties = NULL,
    .range = NULL,
    .minimize = NULL,
    .packages = NULL,
    .man = NULL,
    # NB: .trafo is already taken by mlr3inferr
    .trafo_fun = NULL,
    .properties = character(),
    .predict_sets = NULL,
    .extra_hash = character(),
    .average = NULL,
    .aggregator = NULL,
    .use_weights = NULL,
    .score = function(prediction, task, weights, ...) {
      error_mlr3("abstract method")
    },
    .obs_loss = function(prediction, task, ...) {
      error_mlr3("abstract method")
    }
  )
)


#' @title Workhorse function to calculate a single score
#'
#' @description
#' Assumes that learner is reassembled (if required).
#' It is usually a good idea to exploit lazy evaluation while calling this function
#' to avoid unnecessary allocations.
#'
#' @param measure ([Measure]).
#' @param task ([Task]).
#' @param learner ([Learner]).
#' @param train_set (`integer()`).
#' @param prediction ([Prediction] | [PredictionData]).
#'
#' @return (`numeric()`).
#' @noRd
score_single_measure = function(measure, task, learner, train_set, prediction) {
  if (!length(measure$predict_sets)) {
    score = get_private(measure)$.score(
      prediction = NULL,
      task = task,
      learner = learner,
      train_set = train_set
    )
    return(score)
  }

  # merge multiple predictions (on different predict sets) to a single one
  if (is.list(prediction)) {
    ii = match(measure$predict_sets, names(prediction))
    if (anyMissing(ii)) {
      # TODO lgr$debug()
      return(NaN)
    }
    prediction = do.call(c, prediction[ii])
  }

  # convert pdata to regular prediction
  prediction = as_prediction(prediction, check = FALSE)

  if (is.null(prediction) && length(measure$predict_sets)) {
    return(NaN)
  }

  if (!is_scalar_na(measure$predict_type) && measure$predict_type %nin% prediction$predict_types) {
    # TODO lgr$debug()
    return(NaN)
  }

  if (!is.null(task) && any(measure$task_properties %nin% task$properties)) {
    # TODO lgr$debug()
    return(NaN)
  }

  get_private(measure)$.score(
    prediction = prediction,
    task = task,
    learner = learner,
    train_set = train_set,
    weights = if (measure$use_weights == "use") prediction$weights
  )
}

#' @title Workhorse function to calculate multiple scores
#'
#' @description
#' Converts `obj` from [ResampleResult] or [BenchmarkResult] to a `data.table`.
#' Automatically reassembles learners if needed.
#' Uses lazy evaluation to avoid querying the `train_set` while calling
#' `score_single_measure()`.
#'
#' @param obj ([ResampleResult] | [BenchmarkResult]).
#' @param measures (list of [Measure]).
#' @param iters (`integer()` | `NULL`)\cr
#' To which iterations of the resample result to restrict the scoring.
#'
#' @return (`data.table()`) with added score columns.
#'
#' @noRd
score_measures = function(obj, measures, reassemble = TRUE, view = NULL, iters = NULL) {
  reassemble_learners = reassemble ||
    some(measures, function(m) any(c("requires_learner", "requires_model") %chin% m$properties))
  tab = get_private(obj)$.data$as_data_table(
    view = view,
    reassemble_learners = reassemble_learners,
    convert_predictions = FALSE
  )

  set(tab, j = ".samples", value = map_dbl(tab$prediction, function(x) length(x$test$row_ids)))
  if (length(tab$prediction) && "weights" %chin% names(tab$prediction[[1L]]$test)) {
    set(tab, j = ".weights", value = map_dbl(tab$prediction, function(x) sum(x$test$weights)))
  } else {
    set(tab, j = ".weights", value = tab$.samples)
  }

  if (!is.null(iters)) {
    tab = tab[list(iters), on = "iteration"]
  }

  tmp = unique(tab, by = c("task_hash", "learner_hash"))[, c("task", "learner"), with = FALSE]

  for (measure in measures) {
    pmap(tmp, assert_scorable, measure = measure)

    score = pmap_dbl(
      tab[, c("task", "learner", "resampling", "iteration", "prediction"), with = FALSE],
      function(task, learner, resampling, iteration, prediction) {
        score_single_measure(measure, task, learner, train_set = resampling$train_set(iteration), prediction)
      }
    )
    set(tab, j = measure$id, value = score)
  }

  tab[]
}


#' @export
# nolint next
rd_info.Measure = function(obj, ...) {
  x = c(
    "",
    sprintf("* Task type: %s", rd_format_string(obj$task_type)),
    sprintf("* Range: %s", rd_format_range(obj$range[1L], obj$range[2L])),
    sprintf("* Minimize: %s", obj$minimize),
    sprintf("* Average: %s", obj$average),
    sprintf("* Required Prediction: %s", rd_format_string(obj$predict_type)),
    sprintf("* Required Packages: %s", rd_format_packages(obj$packages))
  )
  paste(x, collapse = "\n")
}

#' @title Bayesian Information Criterion Measure
#'
#' @name mlr_measures_bic
#' @include Measure.R
#'
#' @description
#' Calculates the Bayesian Information Criterion (BIC) which is a
#' trade-off between goodness of fit (measured in terms of
#' log-likelihood) and model complexity (measured in terms of number
#' of included features).
#' Internally, [stats::BIC()] is called.
#' Requires the learner property `"loglik"`, `NA` is returned for unsupported learners.
#'
#' @templateVar id bic
#' @template measure
#'
#' @template seealso_measure
#' @export
MeasureBIC = R6Class(
  "MeasureBIC",
  inherit = Measure,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      super$initialize(
        id = "bic",
        task_type = NA_character_,
        properties = c("na_score", "requires_learner", "requires_model", "requires_no_prediction"),
        predict_sets = NULL,
        predict_type = NA_character_,
        minimize = TRUE,
        label = "Bayesian Information Criterion",
        man = "mlr3::mlr_measures_bic"
      )
    }
  ),

  private = list(
    .score = function(prediction, learner, ...) {
      learner = learner$base_learner()

      tryCatch(
        {
          stats::BIC(stats::logLik(learner$model))
        },
        error = function(e) {
          warning_config("Learner '%s' does not support BIC calculation", learner$id)
          NA_real_
        }
      )
    }
  )
)

#' @include mlr_measures.R
mlr_measures$add("bic", function() MeasureBIC$new())

#' @title Cost-sensitive Classification Measure
#'
#' @name mlr_measures_classif.costs
#' @include MeasureClassif.R
#'
#' @description
#' Uses a cost matrix to create a classification measure.
#' True labels must be arranged in columns, predicted labels must be arranged  in rows.
#' The cost matrix is stored as slot `$costs`.
#'
#' For calculation of the score, the confusion matrix is multiplied element-wise with the cost matrix.
#' The costs are then summed up (and potentially divided by the number of observations
#' if `normalize` is set to `TRUE` (default)).
#'
#' @templateVar id classif.costs
#' @template measure
#'
#' @template seealso_measure
#' @family classification measures
#' @family multiclass classification measures
#' @export
#' @examples
#' # get a cost sensitive task
#' task = tsk("german_credit")
#'
#' # cost matrix as given on the UCI page of the german credit data set
#' # https://archive.ics.uci.edu/ml/datasets/statlog+(german+credit+data)
#' costs = matrix(c(0, 5, 1, 0), nrow = 2)
#' dimnames(costs) = list(truth = task$class_names, predicted = task$class_names)
#' print(costs)
#'
#' # mlr3 needs truth in columns, predictions in rows
#' costs = t(costs)
#'
#' # create a cost measure which calculates the absolute costs
#' m = msr("classif.costs", id = "german_credit_costs", costs = costs, normalize = FALSE)
#'
#' # fit models and evaluate with the cost measure
#' learner = lrn("classif.rpart")
#' rr = resample(task, learner, rsmp("cv", folds = 3))
#' rr$aggregate(m)
MeasureClassifCosts = R6Class(
  "MeasureClassifCosts",
  inherit = MeasureClassif,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      param_set = ps(normalize = p_lgl(tags = "required"))
      param_set$set_values(normalize = TRUE)

      super$initialize(
        id = "classif.costs",
        param_set = param_set,
        range = c(-Inf, Inf),
        minimize = TRUE,
        properties = "weights",
        label = "Cost-sensitive Classification",
        man = "mlr3::mlr_measures_classif.costs"
      )
    }
  ),

  active = list(
    #' @field costs (numeric `matrix()`)\cr
    #' Matrix of costs (truth in columns, predicted response in rows).
    costs = function(rhs) {
      if (missing(rhs)) {
        return(private$.costs)
      }

      assert_matrix(rhs, mode = "numeric", any.missing = FALSE, col.names = "unique", row.names = "unique")
      assert_set_equal(rownames(rhs), colnames(rhs))
      private$.costs = rhs

      self$range = c(-Inf, Inf)
      if (min(rhs) >= 0) {
        self$range[1L] = 0
      }
      if (max(rhs) <= 0) {
        self$range[2L] = 0
      }
    }
  ),

  private = list(
    .costs = NULL,

    .score = function(prediction, weights, ...) {
      costs = self$costs
      if (is.null(costs)) {
        error_config("Costs matrix must be set before scoring, use `$costs` to set it")
      }
      lvls = levels(prediction$truth)
      assert_set_equal(lvls, colnames(costs))

      if (is.null(weights)) {
        confusion = table(response = prediction$response, truth = prediction$truth, useNA = "ifany")
      } else {
        confusion = tapply(
          weights,
          list(response = addNA(prediction$response, ifany = TRUE), truth = addNA(prediction$truth, ifany = TRUE)),
          sum,
          default = 0
        )
      }

      # reorder rows / cols if necessary
      ii = reorder_vector(rownames(confusion), rownames(costs))
      jj = reorder_vector(colnames(confusion), colnames(costs))
      if (is.unsorted(ii) || is.unsorted(jj)) {
        confusion = confusion[ii, jj]
      }

      perf = sum(confusion * costs)
      if (self$param_set$values$normalize) {
        perf = perf / sum(confusion)
      }

      perf
    },

    .extra_hash = "costs"
  )
)

#' @include mlr_measures.R
mlr_measures$add("classif.costs", function() MeasureClassifCosts$new())

#' @title TaskGenerator Class
#'
#' @include mlr_reflections.R
#'
#' @description
#' Creates a [Task] of arbitrary size.
#' Predefined task generators are stored in the [dictionary][mlr3misc::Dictionary] [mlr_task_generators],
#' e.g. [`xor`][mlr_task_generators_xor].
#'
#' @template param_id
#' @template param_param_set
#' @template param_task_type
#' @template param_packages
#' @template param_man
#' @template param_label
#'
#' @template seealso_task_generator
#' @export
TaskGenerator = R6Class(
  "TaskGenerator",
  public = list(
    #' @template field_id
    id = NULL,

    #' @template field_label
    label = NULL,

    #' @template field_task_type
    task_type = NULL,

    #' @template field_param_set
    param_set = NULL,

    #' @template field_packages
    packages = NULL,

    #' @template field_man
    man = NULL,

    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function(
      id,
      task_type,
      packages = character(),
      param_set = ps(),
      label = NA_character_,
      man = NA_character_
    ) {
      self$id = assert_id(id)
      self$param_set = assert_param_set(param_set)
      self$packages = union("mlr3", assert_character(packages, any.missing = FALSE, min.chars = 1L))
      self$task_type = assert_choice(task_type, mlr_reflections$task_types$type)
      self$label = assert_string(label, na.ok = TRUE)
      self$man = assert_string(man, na.ok = TRUE)

      check_packages_installed(
        packages,
        msg = sprintf("Package '%%s' required but not installed for TaskGenerator '%s'", id)
      )
    },

    #' @description
    #' Helper for print outputs.
    #' @param ... (ignored).
    format = function(...) {
      sprintf("<%s:%s>", class(self)[1L], self$id)
    },

    #' @description
    #' Printer.
    #' @param ... (ignored).
    print = function(...) {
      cat_cli({
        msg_h = if (is.null(self$label) || is.na(self$label)) "" else paste0(": ", self$label)
        cli_h1("{.cls {class(self)[1L]}} ({self$id}){msg_h}")
        cli_li("Task type: {self$task_type}")
        cli_li("Packages: {.pkg {self$packages}}")
        cli_li("Parameters: {as_short_string(self$param_set$values, 1000L)}")
        cli_li("Manual: {.help {self$man}}")
      })
    },

    #' @description
    #' Creates a task of type `task_type` with `n` observations,
    #' possibly using additional settings stored in `param_set`.
    #'
    #' @param n (`integer(1)`)\cr
    #'   Number of rows to generate.
    #' @return [Task].
    generate = function(n) {
      n = assert_count(n, coerce = TRUE)
      require_namespaces(self$packages)
      private$.generate(n)
    }
  )
)

convert_mlbench = function(obj) {
  y = factor(LETTERS[as.integer(obj$classes)], levels = LETTERS[seq_len(uniqueN(obj$classes))])
  X = set_col_names(obj$x, sprintf("x%i", seq_col(obj$x)))
  insert_named(as.data.table(X), list(y = y))
}

#' @title Repeated Cross-Validation Resampling
#'
#' @name mlr_resamplings_repeated_cv
#' @include Resampling.R
#'
#' @description
#' Splits data `repeats` (default: 10) times using a `folds`-fold (default: 10) cross-validation.
#'
#' The iteration counter translates to `repeats` blocks of `folds`
#' cross-validations, i.e., the first `folds` iterations belong to
#' a single cross-validation.
#'
#' Iteration numbers can be translated into folds or repeats with provided methods.
#'
#' @templateVar id repeated_cv
#' @template resampling
#'
#' @section Parameters:
#' * `repeats` (`integer(1)`)\cr
#'   Number of repetitions.
#' * `folds` (`integer(1)`)\cr
#'   Number of folds.
#'
#' @references
#' `r format_bib("bischl_2012")`
#'
#' @template seealso_resampling
#' @export
#' @examples
#' # Create a task with 10 observations
#' task = tsk("penguins")
#' task$filter(1:10)
#'
#' # Instantiate Resampling
#' repeated_cv = rsmp("repeated_cv", repeats = 2, folds = 3)
#' repeated_cv$instantiate(task)
#' repeated_cv$iters
#' repeated_cv$folds(1:6)
#' repeated_cv$repeats(1:6)
#'
#' # Individual sets:
#' repeated_cv$train_set(1)
#' repeated_cv$test_set(1)
#'
#' # Disjunct sets:
#' intersect(repeated_cv$train_set(1), repeated_cv$test_set(1))
#'
#' # Internal storage:
#' repeated_cv$instance # table
ResamplingRepeatedCV = R6Class(
  "ResamplingRepeatedCV",
  inherit = Resampling,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        folds = p_int(2L, tags = "required"),
        repeats = p_int(1L)
      )
      ps$set_values(repeats = 10L, folds = 10L)
      super$initialize(
        id = "repeated_cv",
        param_set = ps,
        label = "Repeated Cross-Validation",
        man = "mlr3::mlr_resamplings_repeated_cv"
      )
    },

    #' @description
    #' Translates iteration numbers to fold numbers.
    #' @param iters (`integer()`)\cr
    #'   Iteration number.
    #' @return `integer()` of fold numbers.
    folds = function(iters) {
      iters = assert_integerish(iters, lower = 1L, upper = self$iters, any.missing = FALSE, coerce = TRUE)
      ((iters - 1L) %% as.integer(self$param_set$values$folds)) + 1L
    },

    #' @description
    #' Translates iteration numbers to repetition numbers.
    #' @param iters (`integer()`)\cr
    #'   Iteration number.
    #' @return `integer()` of repetition numbers.
    repeats = function(iters) {
      iters = assert_integerish(iters, lower = 1L, upper = self$iters, any.missing = FALSE, coerce = TRUE)
      ((iters - 1L) %/% as.integer(self$param_set$values$folds)) + 1L
    }
  ),

  active = list(
    #' @template field_iters
    iters = function(rhs) {
      assert_ro_binding(rhs)
      pv = self$param_set$values
      as.integer(pv$repeats) * as.integer(pv$folds)
    }
  ),

  private = list(
    .sample = function(ids, ...) {
      pv = self$param_set$values
      n = length(ids)
      folds = as.integer(pv$folds)
      map_dtr(seq_len(pv$repeats), function(i) {
        data.table(row_id = ids, rep = i, fold = shuffle(seq_len0(n) %% folds + 1L))
      })
    },

    .get_train = function(i) {
      i = as.integer(i) - 1L
      folds = as.integer(self$param_set$values$folds)
      rep = i %/% folds + 1L
      fold = i %% folds + 1L
      ii = data.table(rep = rep, fold = seq_len(folds)[-fold])
      self$instance[ii, "row_id", on = names(ii), nomatch = NULL][[1L]]
    },

    .get_test = function(i) {
      i = as.integer(i) - 1L
      folds = as.integer(self$param_set$values$folds)
      rep = i %/% folds + 1L
      fold = i %% folds + 1L
      ii = data.table(rep = rep, fold = fold)
      self$instance[ii, "row_id", on = names(ii), nomatch = NULL][[1L]]
    },

    .combine = function(instances) {
      rbindlist(instances, use.names = TRUE)
    },

    deep_clone = function(name, value) {
      # fmt: skip
      switch(name,
        "instance" = copy(value),
        ".param_set" = value$clone(deep = TRUE),
        value
      )
    }
  )
)

#' @include mlr_resamplings.R
mlr_resamplings$add("repeated_cv", function() ResamplingRepeatedCV$new())

#' @title Convert to a Measure
#'
#' @description
#' Convert object to a [Measure] or a list of [Measure].
#'
#' @inheritParams as_task
#' @param task_type (`character(1)`)\cr
#'   Used if `x` is `NULL` to construct a default measure for the respective task type.
#'   The default measures are stored in [`mlr_reflections$default_measures`][mlr_reflections].
#'
#' @return [Measure].
#' @export
# nolint next
as_measure = function(x, task_type = NULL, clone = FALSE, ...) {
  UseMethod("as_measure")
}

#' @export
#' @rdname as_measure
# nolint next
as_measure.NULL = function(x, task_type = NULL, clone = FALSE, ...) {
  assert_empty_ellipsis(...)
  default_measures(task_type)[[1L]]
}

#' @export
#' @rdname as_measure
# nolint next
as_measure.Measure = function(x, task_type = NULL, clone = FALSE, ...) {
  assert_empty_ellipsis(...)
  if (isTRUE(clone)) x$clone() else x
}

#' @export
#' @rdname as_measure
# nolint next
as_measures = function(x, task_type = NULL, clone = FALSE, ...) {
  UseMethod("as_measures")
}

#' @export
#' @rdname as_measure
# nolint next
as_measures.default = function(x, task_type = NULL, clone = FALSE, ...) {
  list(as_measure(x, task_type = task_type, clone = clone, ...))
}

#' @export
#' @rdname as_measure
# nolint next
as_measures.NULL = function(x, task_type = NULL, clone = FALSE, ...) {
  default_measures(task_type)
}

#' @export
#' @rdname as_measure
# nolint next
as_measures.list = function(x, task_type = NULL, clone = FALSE, ...) {
  lapply(x, as_measure, task_type = task_type, clone = clone, ...)
}

test_that("as_measure conversion", {
  measure = msr("classif.ce")
  converted = as_measure(measure)

  expect_class(converted, "Measure")
  expect_same_address(measure, converted)

  expect_list(as_measures(measure), types = "Measure")
  expect_list(as_measures(list(measure)), types = "Measure")

  default = as_measure(NULL, task_type = "classif")
  expect_class(default, "Measure")

  default = as_measures(NULL, task_type = "classif")
  expect_list(default, types = "Measure")
})

test_that("error when arguments are misspelled", {
  expect_error(as_measure(msr("classif.acc"), clone2 = TRUE), "Received the following")
})

#' @title Install (Missing) Packages
#'
#' @description
#' `extract_pkgs()` extracts required package from various objects, including
#' [TaskGenerator], [Learner], [Measure] and objects from
#' extension packages such as \CRANpkg{mlr3pipelines} or \CRANpkg{mlr3filters}.
#' If applied on a list, the function is called recursively on all elements.
#'
#' `install_pkgs()` calls `extract_pkgs()` internally and proceeds with the
#' installation of extracted packages.
#'
#' @details
#' If a package contains a forward slash ('/'), it is assumed to be a package hosted
#' on GitHub in `"<user>/<repo>"` format, and the string will be passed to
#' [remotes::install_github()].
#' Otherwise, the package name will be passed to [remotes::install_cran()].
#'
#' @param x (any)\cr
#'   Object with package information (or a list of such objects).
#' @param ... (any)\cr
#'   Additional arguments passed down to [remotes::install_cran()] or
#'   [remotes::install_github()].
#'   Arguments `force` and `upgrade` are often important in this context.
#'
#' @return `extract_pkgs()` returns a `character()` of package strings,
#'   `install_pkgs()` returns the names of extracted packages invisibly.
#'
#' @export
#' @examples
#' extract_pkgs(lrns(c("regr.rpart", "regr.featureless")))
install_pkgs = function(x, ...) {
  require_namespaces("remotes")
  pkg = NULL

  # build initial table
  tab = data.table(pkg = extract_pkgs(x))
  set(tab, j = "name", value = map_chr(strsplit(tab$pkg, "/", fixed = TRUE), tail, 1L))

  # filter duplicates, keep github versions
  set(tab, j = "github", value = grepl("/", tab$pkg, fixed = TRUE))
  setkeyv(tab, c("github", "name"))
  tab = unique(tab, by = "name", fromLast = TRUE)

  # never update self
  tab = tab[get("name") != "mlr3"]

  # install cran packages first
  pkgs = tab[list(FALSE), pkg, on = "github", nomatch = NULL]
  remotes::install_cran(pkgs, ...)

  # install github packages
  pkgs = tab[list(TRUE), pkg, on = "github", nomatch = NULL]
  remotes::install_github(pkgs, ...)

  # return names of extracted packages
  invisible(tab$name)
}

#' @rdname install_pkgs
#' @export
extract_pkgs = function(x) {
  UseMethod("extract_pkgs")
}

#' @rdname install_pkgs
#' @export
extract_pkgs.character = function(x) {
  x
}

#' @rdname install_pkgs
#' @export
# nolint next
extract_pkgs.R6 = function(x) {
  get0("packages", envir = x, inherits = FALSE, ifnotfound = character())
}

#' @rdname install_pkgs
#' @export
# nolint next
extract_pkgs.list = function(x) {
  unique(unlist(lapply(x, extract_pkgs), recursive = FALSE, use.names = FALSE))
}

#' @rdname install_pkgs
#' @export
# nolint next
extract_pkgs.ResampleResult = function(x) {
  extract_pkgs(x$learner)
}

#' @rdname install_pkgs
#' @export
# nolint next
extract_pkgs.BenchmarkResult = function(x) {
  extract_pkgs(x$learners$learner)
}

#' @title 2D Normals Classification Task Generator
#'
#' @name mlr_task_generators_2dnormals
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] for the 2d normals task in [mlbench::mlbench.2dnormals()].
#'
#' @templateVar id 2dnormals
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("2dnormals")
#' plot(generator, n = 200)
#'
#' task = generator$generate(200)
#' str(task$data())
TaskGenerator2DNormals = R6Class(
  "TaskGenerator2DNormals",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        cl = p_int(2L),
        r = p_dbl(1L),
        sd = p_dbl(0L)
      )

      super$initialize(
        id = "2dnormals",
        "classif",
        "mlbench",
        ps,
        label = "2D Normals Classification",
        man = "mlr3::mlr_task_generators_2dnormals"
      )
    },

    #' @description
    #' Creates a simple plot of generated data.
    #' @param n (`integer(1)`)\cr
    #'   Number of samples to draw for the plot. Default is 200.
    #' @param pch (`integer(1)`)\cr
    #'   Point char. Passed to [plot()].
    #' @param ... (any)\cr
    #'   Additional arguments passed to [plot()].
    plot = function(n = 200L, pch = 19L, ...) {
      plot(private$.generate_obj(n), pch = pch, ...)
    }
  ),

  private = list(
    .generate_obj = function(n) {
      invoke(mlbench::mlbench.2dnormals, n = n, .args = self$param_set$values)
    },

    .generate = function(n) {
      obj = private$.generate_obj(n)
      TaskClassif$new(sprintf("%s_%i", self$id, n), convert_mlbench(obj), target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("2dnormals", function() TaskGenerator2DNormals$new())

#' @title Regression Measure
#'
#' @include Measure.R
#'
#' @description
#' This measure specializes [Measure] for regression problems:
#'
#' * `task_type` is set to `"regr"`.
#' * Possible values for `predict_type` are `"response"`, `"se"` and `"distr"`.
#'
#' Predefined measures can be found in the [dictionary][mlr3misc::Dictionary] [mlr_measures].
#' The default measure for regression is [`regr.mse`][mlr_measures_regr.mse].
#'
#' @template param_id
#' @template param_param_set
#' @template param_range
#' @template param_minimize
#' @template param_average
#' @template param_aggregator
#' @template param_predict_type
#' @template param_measure_properties
#' @template param_predict_sets
#' @template param_task_properties
#' @template param_packages
#' @template param_label
#' @template param_man
#'
#' @template seealso_measure
#' @export
MeasureRegr = R6Class(
  "MeasureRegr",
  inherit = Measure,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function(
      id,
      param_set = ps(),
      range,
      minimize = NA,
      average = "macro",
      aggregator = NULL,
      properties = character(),
      predict_type = "response",
      predict_sets = "test",
      task_properties = character(),
      packages = character(),
      label = NA_character_,
      man = NA_character_
    ) {
      super$initialize(
        id,
        task_type = "regr",
        param_set = param_set,
        range = range,
        minimize = minimize,
        average = average,
        aggregator = aggregator,
        properties = properties,
        predict_type = predict_type,
        predict_sets = predict_sets,
        task_properties = task_properties,
        packages = packages,
        label = label,
        man = man
      )
    }
  )
)

#' @title Cross-Validation Resampling
#'
#' @name mlr_resamplings_cv
#' @include Resampling.R
#'
#' @description
#' Splits data using a `folds`-folds (default: 10 folds) cross-validation.
#'
#' @templateVar id cv
#' @template resampling
#'
#' @section Parameters:
#' * `folds` (`integer(1)`)\cr
#'   Number of folds.
#'
#' @references
#' `r format_bib("bischl_2012")`
#'
#' @template seealso_resampling
#' @export
#' @examples
#' # Create a task with 10 observations
#' task = tsk("penguins")
#' task$filter(1:10)
#'
#' # Instantiate Resampling
#' cv = rsmp("cv", folds = 3)
#' cv$instantiate(task)
#'
#' # Individual sets:
#' cv$train_set(1)
#' cv$test_set(1)
#'
#' # Disjunct sets:
#' intersect(cv$train_set(1), cv$test_set(1))
#'
#' # Internal storage:
#' cv$instance # table
ResamplingCV = R6Class(
  "ResamplingCV",
  inherit = Resampling,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        folds = p_int(2L, tags = "required")
      )
      ps$set_values(folds = 10L)

      super$initialize(id = "cv", param_set = ps, label = "Cross-Validation", man = "mlr3::mlr_resamplings_cv")
    }
  ),

  active = list(
    #' @template field_iters
    iters = function(rhs) {
      assert_ro_binding(rhs)
      as.integer(self$param_set$values$folds)
    }
  ),

  private = list(
    .sample = function(ids, ...) {
      data.table(
        row_id = ids,
        fold = shuffle(seq_along0(ids) %% as.integer(self$param_set$values$folds) + 1L),
        key = "fold"
      )
    },

    .get_train = function(i) {
      self$instance[!list(i), "row_id", on = "fold"][[1L]]
    },

    .get_test = function(i) {
      self$instance[list(i), "row_id", on = "fold"][[1L]]
    },

    .combine = function(instances) {
      rbindlist(instances, use.names = TRUE)
    },

    deep_clone = function(name, value) {
      # fmt: skip
      switch(name,
        "instance" = copy(value),
        ".param_set" = value$clone(deep = TRUE),
        value
      )
    }
  )
)

#' @include mlr_resamplings.R
mlr_resamplings$add("cv", function() ResamplingCV$new())

#' @title Dictionary of Tasks
#'
#' @usage NULL
#' @format [R6::R6Class] object inheriting from [mlr3misc::Dictionary].
#' @description
#' A simple [mlr3misc::Dictionary] storing objects of class [Task].
#' Each task has an associated help page, see `mlr_tasks_[id]`.
#'
#' This dictionary can get populated with additional tasks by add-on packages,
#' e.g. \CRANpkg{mlr3data}, \CRANpkg{mlr3proba} or \CRANpkg{mlr3cluster}.
#' \CRANpkg{mlr3oml} allows to interact with [OpenML](https://www.openml.org).
#'
#' For a more convenient way to retrieve and construct tasks, see [tsk()]/[tsks()].
#'
#' @section Methods:
#' See [mlr3misc::Dictionary].
#'
#' @section S3 methods:
#' * `as.data.table(dict, ..., objects = FALSE)`\cr
#'   [mlr3misc::Dictionary] -> [data.table::data.table()]\cr
#'   Returns a [data.table::data.table()] with columns "key", "label", "task_type", "nrow", "ncol", "properties",
#'   and the number of features of type "lgl", "int", "dbl", "chr", "fct" and "ord", respectively.
#'   If `objects` is set to `TRUE`, the constructed objects are returned in the list column named `object`.
#'
#' @family Dictionary
#' @family Task
#' @seealso
#' Sugar functions: [tsk()], [tsks()]
#'
#' Extension Packages: \CRANpkg{mlr3data}
#' @export
#' @examples
#' as.data.table(mlr_tasks)
#' task = mlr_tasks$get("penguins") # same as tsk("penguins")
#' head(task$data())
#'
#' # Add a new task, based on a subset of penguins:
#' data = palmerpenguins::penguins
#' data$species = factor(ifelse(data$species == "Adelie", "1", "0"))
#' task = TaskClassif$new("penguins.binary", data, target = "species", positive = "1")
#'
#' # add to dictionary
#' mlr_tasks$add("penguins.binary", task)
#'
#' # list available tasks
#' mlr_tasks$keys()
#'
#' # retrieve from dictionary
#' mlr_tasks$get("penguins.binary")
#'
#' # remove task again
#' mlr_tasks$remove("penguins.binary")
mlr_tasks = R6Class("DictionaryTask", inherit = Dictionary, cloneable = FALSE)$new()


#' @export
as.data.table.DictionaryTask = function(x, ..., objects = FALSE) {
  assert_flag(objects)

  setkeyv(
    map_dtr(
      x$keys(),
      function(key) {
        t = tryCatch(x$get(key), missingDefaultError = function(e) NULL)
        if (is.null(t)) {
          return(list(key = key))
        }

        feats = translate_types(t$feature_types$type)
        insert_named(
          c(
            list(
              key = key,
              label = t$label,
              task_type = t$task_type,
              nrow = t$nrow,
              ncol = t$ncol,
              properties = list(t$properties)
            ),
            table(feats)
          ),
          if (objects) list(object = list(t))
        )
      },
      .fill = TRUE
    ),
    "key"
  )[]
}

#' @include MeasureClassif.R
MeasureBinarySimple = R6Class(
  "MeasureBinarySimple",
  inherit = MeasureClassif,
  public = list(
    fun = NULL,
    fun_obs_loss = NULL,
    na_value = NaN,
    initialize = function(name, param_set = ps()) {
      info = mlr3measures::measures[[name]]
      weights = info$sample_weights
      properties = if (weights) "weights" else character()

      if (!is.na(info$obs_loss)) {
        properties = c(properties, "obs_loss")
        self$fun_obs_loss = get(info$obs_loss, envir = asNamespace("mlr3measures"), mode = "function")
      }

      super$initialize(
        id = paste0("classif.", name),
        param_set = param_set,
        range = c(info$lower, info$upper),
        minimize = info$minimize,
        properties = properties,
        predict_type = info$predict_type,
        task_properties = "twoclass",
        packages = "mlr3measures",
        label = info$title,
        man = paste0("mlr3::mlr_measures_classif.", name)
      )

      self$fun = get(name, envir = asNamespace("mlr3measures"), mode = "function")

      if (test_list(info$trafo)) {
        self$trafo = info$trafo
      }
    }
  ),

  private = list(
    .score = function(prediction, task, weights = NULL, ...) {
      truth = prediction$truth
      positive = levels(truth)[1L]
      invoke(
        self$fun,
        .args = self$param_set$get_values(),
        truth = truth,
        response = prediction$response,
        prob = prediction$prob[, positive],
        positive = positive,
        na_value = self$na_value,
        sample_weights = weights
      )
    },

    .extra_hash = c("fun", "fun_obs_loss", "na_value"),

    .obs_loss = function(prediction, ...) {
      truth = prediction$truth
      positive = levels(truth)[1L]
      invoke(
        self$fun_obs_loss,
        .args = self$param_set$get_values(),
        truth = truth,
        response = prediction$response,
        prob = prediction$prob[, positive],
        positive = positive
      )
    }
  )
)

#' @include MeasureClassif.R
MeasureClassifSimple = R6Class(
  "MeasureClassifSimple",
  inherit = MeasureClassif,
  public = list(
    fun = NULL,
    fun_obs_loss = NULL,
    na_value = NaN,
    initialize = function(name, param_set = ps()) {
      info = mlr3measures::measures[[name]]
      weights = info$sample_weights
      properties = if (weights) "weights" else character()
      if (!is.na(info$obs_loss)) {
        properties = c(properties, "obs_loss")
        self$fun_obs_loss = get(info$obs_loss, envir = asNamespace("mlr3measures"), mode = "function")
      }

      super$initialize(
        id = paste0("classif.", name),
        param_set = param_set,
        range = c(info$lower, info$upper),
        minimize = info$minimize,
        properties = properties,
        predict_type = info$predict_type,
        packages = "mlr3measures",
        label = info$title,
        man = paste0("mlr3::mlr_measures_classif.", name)
      )
      self$fun = get(name, envir = asNamespace("mlr3measures"), mode = "function")
      if (test_list(info$trafo)) {
        self$trafo = info$trafo
      }
    }
  ),

  private = list(
    .score = function(prediction, task, weights = NULL, ...) {
      invoke(
        self$fun,
        .args = self$param_set$get_values(),
        truth = prediction$truth,
        response = prediction$response,
        prob = prediction$prob,
        na_value = self$na_value,
        sample_weights = weights
      )
    },

    .extra_hash = c("fun", "fun_obs_loss", "na_value"),

    .obs_loss = function(prediction, ...) {
      invoke(
        self$fun_obs_loss,
        .args = self$param_set$get_values(),
        truth = prediction$truth,
        response = prediction$response,
        prob = prediction$prob
      )
    }
  )
)

#' @include MeasureRegr.R
MeasureRegrSimple = R6Class(
  "MeasureRegrSimple",
  inherit = MeasureRegr,
  public = list(
    fun = NULL,
    fun_obs_loss = NULL,
    na_value = NaN,
    initialize = function(name, param_set = NULL) {
      if (is.null(param_set)) {
        param_set = ps()
      } else {
        # cloning required because the param set lives in the
        # dictionary mlr_measures
        param_set = param_set$clone()
      }

      info = mlr3measures::measures[[name]]
      weights = info$sample_weights
      properties = if (weights) "weights" else character()
      if (!is.na(info$obs_loss)) {
        properties = c(properties, "obs_loss")
        self$fun_obs_loss = get(info$obs_loss, envir = asNamespace("mlr3measures"), mode = "function")
      }

      super$initialize(
        id = paste0("regr.", name),
        param_set = param_set,
        range = c(info$lower, info$upper),
        minimize = info$minimize,
        properties = properties,
        predict_type = info$predict_type,
        packages = "mlr3measures",
        label = info$title,
        man = paste0("mlr3::mlr_measures_regr.", name)
      )
      self$fun = get(name, envir = asNamespace("mlr3measures"), mode = "function")

      if (test_list(info$trafo)) {
        self$trafo = info$trafo
      }
    }
  ),

  private = list(
    .score = function(prediction, task, weights = NULL, ...) {
      invoke(
        self$fun,
        .args = self$param_set$get_values(),
        truth = prediction$truth,
        response = prediction$response,
        se = prediction$se,
        na_value = self$na_value,
        sample_weights = weights
      )
    },

    .extra_hash = c("fun", "fun_obs_loss", "na_value"),

    .obs_loss = function(prediction, ...) {
      invoke(
        self$fun_obs_loss,
        .args = self$param_set$get_values(),
        truth = prediction$truth,
        response = prediction$response,
        se = prediction$se
      )
    }
  )
)

#' @include MeasureSimilarity.R
MeasureSimilaritySimple = R6Class(
  "MeasureSimilaritySimple",
  inherit = MeasureSimilarity,
  public = list(
    fun = NULL,
    na_value = NaN,
    initialize = function(name, param_set = NULL) {
      if (is.null(param_set)) {
        param_set = ps()
      } else {
        # cloning required because the param set lives in the
        # dictionary mlr_measures
        param_set = param_set$clone()
      }

      info = mlr3measures::measures[[name]]
      self$fun = get(name, envir = asNamespace("mlr3measures"), mode = "function")

      agg = function(rr) {
        sets = map(rr$learners, function(l) l$selected_features())
        self$fun(sets, p = length(rr$task$feature_names))
      }

      super$initialize(
        id = paste0("sim.", name),
        param_set = param_set,
        range = c(info$lower, info$upper),
        minimize = info$minimize,
        aggregator = agg,
        predict_type = NA_character_,
        packages = "mlr3measures",
        label = info$title,
        man = paste0("mlr3::mlr_measures_sim.", name),
      )
    }
  ),

  private = list(
    .extra_hash = c("fun", "na_value")
  )
)


### binary classification measures

#' @templateVar id auc
#' @template measure_binary
mlr_measures$add("classif.auc", function() MeasureBinarySimple$new(name = "auc"))

#' @templateVar id bbrier
#' @template measure_binary
mlr_measures$add("classif.bbrier", function() MeasureBinarySimple$new(name = "bbrier"))

#' @templateVar id dor
#' @template measure_binary
mlr_measures$add("classif.dor", function() MeasureBinarySimple$new(name = "dor"))

#' @templateVar id fbeta
#' @template measure_binary
mlr_measures$add("classif.fbeta", function() {
  MeasureBinarySimple$new(name = "fbeta", param_set = ps(beta = p_int(lower = 0L)))
})

#' @templateVar id fdr
#' @template measure_binary
mlr_measures$add("classif.fdr", function() MeasureBinarySimple$new(name = "fdr"))

#' @templateVar id fn
#' @template measure_binary
mlr_measures$add("classif.fn", function() MeasureBinarySimple$new(name = "fn"))

#' @templateVar id fnr
#' @template measure_binary
mlr_measures$add("classif.fnr", function() MeasureBinarySimple$new(name = "fnr"))

#' @templateVar id fomr
#' @template measure_binary
mlr_measures$add("classif.fomr", function() MeasureBinarySimple$new(name = "fomr"))

#' @templateVar id fp
#' @template measure_binary
mlr_measures$add("classif.fp", function() MeasureBinarySimple$new(name = "fp"))

#' @templateVar id fpr
#' @template measure_binary
mlr_measures$add("classif.fpr", function() MeasureBinarySimple$new(name = "fpr"))

#' @templateVar id npv
#' @template measure_binary
mlr_measures$add("classif.npv", function() MeasureBinarySimple$new(name = "npv"))

#' @templateVar id ppv
#' @template measure_binary
mlr_measures$add("classif.ppv", function() MeasureBinarySimple$new(name = "ppv"))

#' @templateVar id prauc
#' @template measure_binary
mlr_measures$add("classif.prauc", function() MeasureBinarySimple$new(name = "prauc"))

#' @templateVar id precision
#' @template measure_binary
mlr_measures$add("classif.precision", function() MeasureBinarySimple$new(name = "precision"))

#' @templateVar id tn
#' @template measure_binary
mlr_measures$add("classif.tn", function() MeasureBinarySimple$new(name = "tn"))

#' @templateVar id tnr
#' @template measure_binary
mlr_measures$add("classif.tnr", function() MeasureBinarySimple$new(name = "tnr"))

#' @templateVar id specificity
#' @template measure_binary
mlr_measures$add("classif.specificity", function() MeasureBinarySimple$new(name = "specificity"))

#' @templateVar id tp
#' @template measure_binary
mlr_measures$add("classif.tp", function() MeasureBinarySimple$new(name = "tp"))

#' @templateVar id tpr
#' @template measure_binary
mlr_measures$add("classif.tpr", function() MeasureBinarySimple$new(name = "tpr"))

#' @templateVar id recall
#' @template measure_binary
mlr_measures$add("classif.recall", function() MeasureBinarySimple$new(name = "recall"))

#' @templateVar id sensitivity
#' @template measure_binary
mlr_measures$add("classif.sensitivity", function() MeasureBinarySimple$new(name = "sensitivity"))

### multiclass classification measures

#' @templateVar id acc
#' @template measure_classif
mlr_measures$add("classif.acc", function() MeasureClassifSimple$new(name = "acc"))

#' @templateVar id bacc
#' @template measure_classif
mlr_measures$add("classif.bacc", function() MeasureClassifSimple$new(name = "bacc"))

#' @templateVar id mbrier
#' @template measure_classif
mlr_measures$add("classif.mbrier", function() MeasureClassifSimple$new(name = "mbrier"))

#' @templateVar id ce
#' @template measure_classif
mlr_measures$add("classif.ce", function() MeasureClassifSimple$new(name = "ce"))

#' @templateVar id logloss
#' @template measure_classif
mlr_measures$add("classif.logloss", function() MeasureClassifSimple$new(name = "logloss"))

#' @templateVar id mauc_aunu
#' @template measure_classif
mlr_measures$add("classif.mauc_aunu", function() MeasureClassifSimple$new(name = "mauc_aunu"))

#' @templateVar id mauc_aunp
#' @template measure_classif
mlr_measures$add("classif.mauc_aunp", function() MeasureClassifSimple$new(name = "mauc_aunp"))

#' @templateVar id mauc_au1u
#' @template measure_classif
mlr_measures$add("classif.mauc_au1u", function() MeasureClassifSimple$new(name = "mauc_au1u"))

#' @templateVar id mauc_au1p
#' @template measure_classif
mlr_measures$add("classif.mauc_au1p", function() MeasureClassifSimple$new(name = "mauc_au1p"))

#' @templateVar id mauc_mu
#' @template measure_classif
mlr_measures$add("classif.mauc_mu", function() MeasureClassifSimple$new(name = "mauc_mu"))

#' @templateVar id mcc
#' @template measure_classif
mlr_measures$add("classif.mcc", function() MeasureClassifSimple$new(name = "mcc"))


### regression measures

#' @templateVar id bias
#' @template measure_regr
mlr_measures$add("regr.bias", function() MeasureRegrSimple$new(name = "bias"))

#' @templateVar id ktau
#' @template measure_regr
mlr_measures$add("regr.ktau", function() MeasureRegrSimple$new(name = "ktau"))

#' @templateVar id mae
#' @template measure_regr
mlr_measures$add("regr.mae", function() MeasureRegrSimple$new(name = "mae"))

#' @templateVar id mape
#' @template measure_regr
mlr_measures$add("regr.mape", function() MeasureRegrSimple$new(name = "mape"))

#' @templateVar id maxae
#' @template measure_regr
mlr_measures$add("regr.maxae", function() MeasureRegrSimple$new(name = "maxae"))

#' @templateVar id medae
#' @template measure_regr
mlr_measures$add("regr.medae", function() MeasureRegrSimple$new(name = "medae"))

#' @templateVar id medse
#' @template measure_regr
mlr_measures$add("regr.medse", function() MeasureRegrSimple$new(name = "medse"))

#' @templateVar id mse
#' @template measure_regr
mlr_measures$add("regr.mse", function() MeasureRegrSimple$new(name = "mse"))

#' @templateVar id msle
#' @template measure_regr
mlr_measures$add("regr.msle", function() MeasureRegrSimple$new(name = "msle"))

#' @templateVar id pbias
#' @template measure_regr
mlr_measures$add("regr.pbias", function() MeasureRegrSimple$new(name = "pbias"))

#' @templateVar id rmse
#' @template measure_regr
mlr_measures$add("regr.rmse", function() MeasureRegrSimple$new(name = "rmse"))

#' @templateVar id rmsle
#' @template measure_regr
mlr_measures$add("regr.rmsle", function() MeasureRegrSimple$new(name = "rmsle"))

#' @templateVar id sae
#' @template measure_regr
mlr_measures$add("regr.sae", function() MeasureRegrSimple$new(name = "sae"))

#' @templateVar id smape
#' @template measure_regr
mlr_measures$add("regr.smape", function() MeasureRegrSimple$new(name = "smape"))

#' @templateVar id srho
#' @template measure_regr
mlr_measures$add("regr.srho", function() MeasureRegrSimple$new(name = "srho"))

#' @templateVar id sse
#' @template measure_regr
mlr_measures$add("regr.sse", function() MeasureRegrSimple$new(name = "sse"))


### similarity measures

#' @templateVar id jaccard
#' @template measure_similarity
mlr_measures$add("sim.jaccard", function() MeasureSimilaritySimple$new(name = "jaccard"))

#' @templateVar id phi
#' @template measure_similarity
mlr_measures$add("sim.phi", function() {
  MeasureSimilaritySimple$new(name = "phi", param_set = ps(p = p_int(lower = 1L)))
})

#' @title Bootstrap Resampling
#'
#' @name mlr_resamplings_bootstrap
#' @include Resampling.R
#'
#' @description
#' Splits data into bootstrap samples (sampling with replacement).
#' Hyperparameters are the number of bootstrap iterations (`repeats`, default: 30)
#' and the ratio of observations to draw per iteration (`ratio`, default: 1) for the training set.
#'
#' @templateVar id bootstrap
#' @template resampling
#'
#' @section Parameters:
#' * `repeats` (`integer(1)`)\cr
#'   Number of repetitions.
#' * `ratio` (`numeric(1)`)\cr
#'   Ratio of observations to put into the training set.
#'
#' @references
#' `r format_bib("bischl_2012")`
#'
#' @template seealso_resampling
#' @export
#' @examples
#' # Create a task with 10 observations
#' task = tsk("penguins")
#' task$filter(1:10)
#'
#' # Instantiate Resampling
#' bootstrap = rsmp("bootstrap", repeats = 2, ratio = 1)
#' bootstrap$instantiate(task)
#'
#' # Individual sets:
#' bootstrap$train_set(1)
#' bootstrap$test_set(1)
#'
#' # Disjunct sets:
#' intersect(bootstrap$train_set(1), bootstrap$test_set(1))
#'
#' # Internal storage:
#' bootstrap$instance$M # Matrix of counts
ResamplingBootstrap = R6Class(
  "ResamplingBootstrap",
  inherit = Resampling,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        ratio = p_dbl(0, upper = 1, tags = "required"),
        repeats = p_int(1L, tags = "required")
      )
      ps$set_values(ratio = 1, repeats = 30L)

      super$initialize(
        id = "bootstrap",
        param_set = ps,
        duplicated_ids = TRUE,
        label = "Bootstrap",
        man = "mlr3::mlr_resamplings_bootstrap"
      )
    }
  ),

  active = list(
    #' @template field_iters
    iters = function(rhs) {
      assert_ro_binding(rhs)
      as.integer(self$param_set$values$repeats)
    }
  ),

  private = list(
    .sample = function(ids, ...) {
      pv = self$param_set$values
      nr = round(length(ids) * pv$ratio)
      x = factor(seq_along(ids))
      M = replicate(pv$repeats, table(sample(x, nr, replace = TRUE)), simplify = "array")
      rownames(M) = NULL
      list(row_ids = ids, M = M)
    },

    .get_train = function(i) {
      rep(self$instance$row_ids, times = self$instance$M[, i])
    },

    .get_test = function(i) {
      self$instance$row_ids[self$instance$M[, i] == 0L]
    },

    .combine = function(instances) {
      list(row_ids = do.call(c, map(instances, "row_ids")), M = do.call(rbind, map(instances, "M")))
    }
  )
)

#' @include mlr_resamplings.R
mlr_resamplings$add("bootstrap", function() ResamplingBootstrap$new())

#' @title Resample Callback
#'
#' @description
#' Specialized [mlr3misc::Callback] to customize the behavior of [resample()] and [benchmark()] in mlr3.
#' For example, callbacks can be used to extract information from models on the worker
#' or to store intermediate results to disk.
#' The [callback_resample()] function is used to create instances of this class.
#' Predefined callbacks are stored in the [dictionary][mlr3misc::Dictionary] [mlr_callbacks]
#' and can be retrieved with [clbk()].
#' For more information on callbacks, see the [callback_resample()] documentation.
#'
#' @export
CallbackResample = R6Class(
  "CallbackResample",
  inherit = Callback,
  public = list(
    #' @field on_resample_begin (`function()`)\cr
    #' Stage called at the beginning of the resampling iteration.
    #' Called in `workhorse()` (internal).
    on_resample_begin = NULL,

    #' @field on_resample_before_train (`function()`)\cr
    #' Stage called before training the learner.
    #' Called in `workhorse()` (internal).
    on_resample_before_train = NULL,

    #' @field on_resample_before_predict (`function()`)\cr
    #' Stage called before predicting.
    #' Called in `workhorse()` (internal).
    on_resample_before_predict = NULL,

    #' @field on_resample_end (`function()`)\cr
    #' Stage called at the end of the resample iteration.
    #' Called in `workhorse()` (internal).
    on_resample_end = NULL
  )
)

#' @title Create Evaluation Callback
#'
#' @description
#' Function to create a [CallbackResample].
#' Predefined callbacks are stored in the [dictionary][mlr3misc::Dictionary] [mlr_callbacks]
#' and can be retrieved with [clbk()].
#'
#' Evaluation callbacks are called at different stages of the resampling process.
#' Each stage is called once per resampling iteration.
#' The stages are prefixed with `on_resample_*`.
#' The text in brackets indicates what happens between the stages in the internal `workhorse()` function
#' and which accesses to the [ContextResample] (`ctx`) are typical for the stage.
#'
#' ```
#' Start Resampling Iteration on Worker
#'  - on_resample_begin
#'    (Split `ctx$task` into training and test set with `ctx$resampling` and `ctx$iteration`)
#'  - on_resample_before_train
#'    (Train the learner `ctx$learner` on training data)
#'  - on_resample_before_predict
#'    (Predict on predict sets and store prediction data `ctx$pdatas`)
#'  - on_resample_end
#'    (Erase model `ctx$learner$model` if requested and return results)
#' End Resampling Iteration on Worker
#' ```
#'
#' The callback can store data in `ctx$learner$state` or `ctx$data_extra`.
#' The data in `ctx$data_extra` is stored in the [ResampleResult] or [BenchmarkResult].
#' See also the section on parameters for more information on the stages.
#
#' @details
#' When implementing a callback, each function must have two arguments named `callback` and `context`.
#' A callback can write data to the state (`$state`), e.g. settings that affect the callback itself.
#' We highly discourage changing the task, learner and resampling objects via the callback.
#'
#' @param id (`character(1)`)\cr
#'   Identifier for the new instance.
#' @param label (`character(1)`)\cr
#'   Label for the new instance.
#' @param man (`character(1)`)\cr
#'   String in the format `[pkg]::[topic]` pointing to a manual page for this object.
#'   The referenced help package can be opened via method `$help()`.
#' @param on_resample_begin (`function()`)\cr
#'   Stage called at the beginning of an evaluation.
#'   Called in `workhorse()` (internal).
#' @param on_resample_before_train (`function()`)\cr
#'   Stage called before training the learner.
#'   Called in `workhorse()` (internal).
#' @param on_resample_before_predict (`function()`)\cr
#'   Stage called before predicting.
#'   Called in `workhorse()` (internal).
#' @param on_resample_end (`function()`)\cr
#'   Stage called at the end of an evaluation.
#'   Called in `workhorse()` (internal).
#'
#' @export
#' @examples
#' learner = lrn("classif.rpart")
#' task = tsk("pima")
#' resampling = rsmp("cv", folds = 3)
#'
#' # save selected features callback
#' callback = callback_resample("selected_features",
#'  on_resample_end = function(callback, context) {
#'     context$learner$state$selected_features = context$learner$selected_features()
#'   }
#' )
#'
#' rr = resample(task, learner, resampling, callbacks = callback)
#' rr$learners[[1]]$state$selected_features
#'
#' # holdout task callback
#' callback = callback_resample("holdout_task",
#'   on_resample_before_predict = function(callback, context) {
#'     pred = context$learner$predict(callback$state$task)
#'     context$data_extra = list(prediction_holdout = pred)
#'   }
#' )
#'
#' task_holdout = tsk("pima")
#' splits = partition(task, 0.7)
#' task$filter(splits$train)
#' task_holdout$filter(splits$test)
#'
#' callback$state$task = task_holdout
#'
#' rr = resample(task, learner, resampling, callbacks = callback)
#' rr$data_extra
callback_resample = function(
  id,
  label = NA_character_,
  man = NA_character_,
  on_resample_begin = NULL,
  on_resample_before_train = NULL,
  on_resample_before_predict = NULL,
  on_resample_end = NULL
) {
  stages = discard(
    set_names(
      list(
        on_resample_begin,
        on_resample_before_train,
        on_resample_before_predict,
        on_resample_end
      ),
      c(
        "on_resample_begin",
        "on_resample_before_train",
        "on_resample_before_predict",
        "on_resample_end"
      )
    ),
    is.null
  )

  stages = map(stages, function(stage) crate(assert_function(stage, args = c("callback", "context"))))
  callback = CallbackResample$new(id, label, man)
  iwalk(stages, function(stage, name) callback[[name]] = stage)
  callback
}

#' @title Assertions for Callbacks
#'
#' @description
#' Assertions for [CallbackResample] class.
#'
#' @param callback ([CallbackResample]).
#' @param null_ok (`logical(1)`)\cr
#'   If `TRUE`, `NULL` is allowed.
#'
#' @return [CallbackResample] | List of [CallbackResample]s.
#' @export
assert_resample_callback = function(callback, null_ok = FALSE) {
  assert_class(callback, "CallbackResample", null.ok = null_ok)
  invisible(callback)
}

#' @export
#' @param callbacks (list of [CallbackResample]).
#' @rdname assert_resample_callback
assert_resample_callbacks = function(callbacks, null_ok = FALSE) {
  assert_list(callbacks, null.ok = null_ok)
  if (null_ok && is.null(callbacks)) {
    return(invisible(NULL))
  }
  invisible(lapply(callbacks, assert_resample_callback))
}

#' @title Generate a Benchmark Grid Design
#'
#' @description
#' Takes a lists of [Task], a list of [Learner] and a list of [Resampling] to
#' generate a design in an [expand.grid()] fashion (a.k.a. cross join or Cartesian product).
#'
#' There are two modes of operation, depending on the flag `paired`.
#'
#' * With `paired` set to `FALSE` (default), resampling strategies are not allowed to be
#'   instantiated, and instead will be instantiated per task internally.
#'   The only exception to this rule applies if all tasks have exactly the same row ids,
#'   and the resamplings are all instantiated for such tasks.
#'   The grid will be generated based on the Cartesian product of tasks, learners,
#'   and resamplings.
#'   Because the resamplings are instantiated on the tasks, reproducibility requires a seed
#'   to be set **before** calling this function, as this process is stochastic.
#' * With `paired` set to `TRUE`, tasks and resamplings are treated as pairs.
#'   This means that you must provide as many tasks as corresponding instantiated resamplings.
#'   The grid will be generated based on the Cartesian product of learners and pairs.
#'
#' @section Errors and Warnings:
#' * `Mlr3WarningVaryingPredictTypes`: This warning will be thrown if the learners have different `predict_type`s.
#'
#' @param tasks (list of [Task]).
#' @param learners (list of [Learner]).
#' @param resamplings (list of [Resampling]).
#' @template param_param_values
#' @param paired (`logical(1)`)\cr
#'   Set this to `TRUE` if the resamplings are instantiated on the tasks, i.e., the tasks and resamplings are paired.
#'   You need to provide the same number of tasks and instantiated resamplings.
#'
#' @return ([data.table::data.table()]) with the cross product of the input vectors.
#'
#' @template seealso_benchmark
#' @export
#' @examples
#' tasks = list(tsk("penguins"), tsk("sonar"))
#' learners = list(lrn("classif.featureless"), lrn("classif.rpart"))
#' resamplings = list(rsmp("cv"), rsmp("subsampling"))
#'
#' # Set a seed to ensure reproducibility of the resampling instantiation
#' set.seed(123)
#' grid = benchmark_grid(tasks, learners, resamplings)
#' # the resamplings are now instantiated
#' head(grid$resampling[[1]]$instance)
#' print(grid)
#' \dontrun{
#' benchmark(grid)
#' }
#'
#' # paired
#' learner = lrn("classif.rpart")
#' task1 = tsk("penguins")
#' task2 = tsk("german_credit")
#' res1 = rsmp("holdout")
#' res2 = rsmp("holdout")
#' res1$instantiate(task1)
#' res2$instantiate(task2)
#' design = benchmark_grid(list(task1, task2), learner, list(res1, res2), paired = TRUE)
#' print(design)
#'
#' # manual construction of the grid with data.table::CJ()
#' grid = data.table::CJ(
#'  task = tasks,
#'   learner = learners,
#'   resampling = resamplings,
#'   sorted = FALSE
#' )
#'
#' # manual instantiation (not suited for a fair comparison of learners!)
#' Map(function(task, resampling) {
#'   resampling$instantiate(task)
#' }, task = grid$task, resampling = grid$resampling)
#' \dontrun{
#' benchmark(grid)
#' }
#'
benchmark_grid = function(tasks, learners, resamplings, param_values = NULL, paired = FALSE) {
  tasks = assert_tasks(as_tasks(tasks))
  learners = assert_learners(as_learners(learners), unique_ids = TRUE)
  resamplings = assert_resamplings(as_resamplings(resamplings))
  if (!is.null(param_values)) {
    assert_param_values(param_values, n_learners = length(learners))
  }
  if (length(unique(map_chr(unique(learners), "predict_type"))) > 1L) {
    warning_config(
      "Multiple predict types detected, this will mean that you cannot evaluate the same measures on all learners.",
      class = "Mlr3WarningVaryingPredictTypes"
    )
  }

  if (assert_flag(paired)) {
    if (length(tasks) != length(resamplings)) {
      error_input("If `paired` is `TRUE`, you need to provide the same number of tasks and instantiated resamplings")
    }

    for (i in seq_along(tasks)) {
      task = tasks[[i]]
      resampling = resamplings[[i]]
      if (!resamplings[[i]]$is_instantiated) {
        error_input("Resampling #%i ('%s' for task '%s') is not instantiated", i, resampling$id, task$id)
      }
      if (resampling$task_row_hash != task$row_hash) {
        error_input(
          "Resampling #%i ('%s' for task '%s') is not instantiated on the corresponding task",
          i,
          resampling$id,
          task$id
        )
      }
    }

    grid = CJ(task = seq_along(tasks), learner = seq_along(learners))
    tab = data.table(task = tasks[grid$task], learner = learners[grid$learner], resampling = resamplings[grid$task])
  } else {
    grid = CJ(task = seq_along(tasks), resampling = seq_along(resamplings))
    is_instantiated = map_lgl(resamplings, "is_instantiated")

    if (any(is_instantiated) && !all(is_instantiated)) {
      # prevent that some resamplings are instantiated and others are not
      error_input("All resamplings must be instantiated, or none at all")
    } else if (all(is_instantiated)) {
      # check that all row ids of the resamplings are present in the tasks
      pwalk(grid, function(task, resampling) {
        if (
          !is.null(resamplings[[resampling]]$task_row_hash) &&
            resamplings[[resampling]]$task_row_hash != tasks[[task]]$row_hash
        ) {
          error_input(
            "Resampling '%s' is not instantiated on task '%s'",
            resamplings[[resampling]]$id,
            tasks[[task]]$id
          )
        }
      })

      # clone resamplings for each task and update task hashes
      instances = pmap(grid, function(task, resampling) resampling = resamplings[[resampling]]$clone())
    } else {
      instances = pmap(grid, function(task, resampling) resamplings[[resampling]]$clone()$instantiate(tasks[[task]]))
    }

    grid$instance = seq_row(grid)
    grid = grid[CJ(task = seq_along(tasks), learner = seq_along(learners)), on = "task", allow.cartesian = TRUE]

    tab = data.table(task = tasks[grid$task], learner = learners[grid$learner], resampling = instances[grid$instance])
  }

  if (!is.null(param_values)) {
    set(tab, j = "param_values", value = list(param_values[grid$learner]))
  }

  set_data_table_class(tab, "benchmark_grid")
  return(tab)
}

#' @export
print.benchmark_grid = function(x, ...) {
  print(data.table(
    task = ids(x$task),
    learner = ids(x$learner),
    resampling = ids(x$resampling),
    param_values = if (is.null(x$param_values)) NULL else sprintf("<%i>", lengths(x$param_values))
  ))
}

#' @title Average Pinball Loss
#'
#' @name mlr_measures_regr.pinball
#' @include Measure.R
#'
#' @description
#' Measure to compare true observed response with predicted response in regression tasks.
#'
#' @details
#' The pinball loss for quantile regression is defined as \deqn{
#' \text{Average Pinball Loss} = \frac{1}{n} \sum_{i=1}^{n} w_{i}
#' \begin{cases}
#' q \cdot (t_i - r_i) & \text{if } t_i \geq r_i \\
#'(1 - q) \cdot (r_i - t_i) & \text{if } t_i < r_i
#' \end{cases}
#' }
#' where \eqn{q} is the quantile and \eqn{w_i} are normalized sample weights.
#'
#' @param alpha `numeric(1)`\cr
#' The quantile to compute the pinball loss.
#' Must be one of the quantiles that the Learner was trained on.
#
#' @templateVar id regr.pinball
#' @template measure
#'
#' @template seealso_measure
#' @export
MeasureRegrPinball = R6Class(
  "MeasureRegrPinball",
  inherit = MeasureRegr,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function(alpha = 0.5) {
      param_set = ps(alpha = p_dbl(lower = 0, upper = 1))
      param_set$set_values(alpha = alpha)
      super$initialize(
        id = "regr.pinball",
        param_set = param_set,
        predict_type = "quantiles",
        properties = "weights",
        minimize = TRUE,
        range = c(-Inf, Inf),
        label = "Average Pinball Loss",
        man = "mlr3::mlr_measures_regr.pinball"
      )
    }
  ),

  private = list(
    .score = function(prediction, weights = NULL, ...) {
      alpha = self$param_set$values$alpha
      probs = attr(prediction$data$quantiles, "probs")
      assert_choice(alpha, probs)

      mlr3measures::pinball(
        truth = prediction$truth,
        response = prediction$data$quantiles[, which(probs == alpha)],
        alpha = alpha,
        sample_weights = weights
      )
    }
  )
)

#' @include mlr_measures.R
mlr_measures$add("regr.pinball", MeasureRegrPinball)

#' @title Simplex Classification Task Generator
#'
#' @name mlr_task_generators_simplex
#' @include TaskGenerator.R
#'
#' @description
#' A [TaskGenerator] for the simplex task in [mlbench::mlbench.simplex()].
#'
#' Note that the generator implemented in \CRANpkg{mlbench} returns
#' fewer samples than requested.
#'
#' @templateVar id simplex
#' @template task_generator
#'
#' @template seealso_task_generator
#' @export
#' @examples
#' generator = tgen("simplex")
#' plot(generator, n = 200)
#'
#' task = generator$generate(200)
#' str(task$data())
TaskGeneratorSimplex = R6Class(
  "TaskGeneratorSimplex",
  inherit = TaskGenerator,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        center = p_lgl(default = TRUE),
        d = p_int(1L, default = 3L),
        sd = p_dbl(0, default = 0.1),
        sides = p_int(1L, default = 1L)
      )

      super$initialize(
        id = "simplex",
        "classif",
        "mlbench",
        ps,
        label = "Simplex Classification",
        man = "mlr3::mlr_task_generators_simplex"
      )
    },

    #' @description
    #' Creates a simple plot of generated data.
    #' @param n (`integer(1)`)\cr
    #'   Number of samples to draw for the plot. Default is 200.
    #' @param pch (`integer(1)`)\cr
    #'   Point char. Passed to [plot()].
    #' @param ... (any)\cr
    #'   Additional arguments passed to [plot()].
    plot = function(n = 200L, pch = 19L, ...) {
      plot(private$.generate_obj(n), pch = pch, ...)
    }
  ),

  private = list(
    .generate_obj = function(n) {
      invoke(mlbench::mlbench.simplex, n = n, .args = self$param_set$values, .opts = allow_partial_matching)
    },

    .generate = function(n) {
      obj = private$.generate_obj(n)
      TaskClassif$new(sprintf("%s_%i", self$id, n), convert_mlbench(obj), target = "y")
    }
  )
)

#' @include mlr_task_generators.R
mlr_task_generators$add("simplex", function() TaskGeneratorSimplex$new())

#' @title R-Squared
#'
#' @name mlr_measures_regr.rsq
#' @include Measure.R
#'
#' @description
#' Measure to compare true observed response with predicted response in regression tasks.
#'
#' @details
#' R Squared is defined as \deqn{
#'   1 - \frac{\sum_{i=1}^n w_i \left( t_i - r_i \right)^2}{\sum_{i=1}^n w_i \left( t_i - \bar{t} \right)^2},
#' }{
#'   1 - sum(w * (t - r)^2) / sum(w * (t - mean(t))^2),
#' }
#' where \eqn{\bar{t} = \frac{1}{n} \sum_{i=1}^n t_i} and \eqn{w_i} are weights.
#'
#' Also known as coefficient of determination or explained variation.
#' It compares the squared error of the predictions relative to a naive model predicting the mean.
#'
#' Note that weights are used to scale the squared error of individual predictions
#' (both in the numerator and in the denominator),
#' but the "plug in" value \eqn{\bar{t}} is computed without weights.
#'
#' This measure is undefined for constant \eqn{t}.
#'
#' @param pred_set_mean `logical(1)`\cr
#' If `TRUE`, the mean of the true values is calculated on the prediction set.
#' If `FALSE`, the mean of the true values is calculated on the training set.
#'
#' @templateVar id regr.rsq
#' @template measure
#'
#' @template seealso_measure
#' @export
MeasureRegrRSQ = R6Class(
  "MeasureRegrRSQ",
  inherit = MeasureRegr,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function(pred_set_mean = TRUE) {
      # this is not included in the paramset as this flag influences properties of the learner
      # so this flag should not be "dynamic state"
      private$.pred_set_mean = assert_flag(pred_set_mean)

      super$initialize(
        id = "regr.rsq",
        properties = c(if (!private$.pred_set_mean) c("requires_task", "requires_train_set"), "weights"),
        predict_type = "response",
        minimize = FALSE,
        range = c(-Inf, 1),
        man = "mlr3::mlr_measures_regr.rsq"
      )
    }
  ),

  active = list(
    #' @field pred_set_mean (`logical(1)`)\cr
    #' Whether the mean of the true values is calculated on the prediction set.
    #' Set during construction.
    pred_set_mean = function(rhs) {
      assert_ro_binding(rhs)
      private$.pred_set_mean
    }
  ),

  private = list(
    # this is not included in the paramset as this flag influences properties of the learner
    # so this flag should not be "dynamic state"
    .pred_set_mean = NULL,

    .extra_hash = "pred_set_mean",

    .score = function(prediction, task = NULL, train_set = NULL, weights = NULL, ...) {
      if (is.null(weights)) {
        mu = if (private$.pred_set_mean) mean(prediction$truth) else mean(task$truth(train_set))
        1 - sum((prediction$truth - prediction$response)^2) / sum((prediction$truth - mu)^2)
      } else {
        # Don't use weighted mean here,
        # since resampling weights only concern weighting of loss of individual predictions.
        # One could argue that we should use weights when .pred_set.mean is TRUE,
        # or we could use weights_learner from the task, but currently we decided against it.
        mu = if (private$.pred_set_mean) {
          mean(prediction$truth)
        } else {
          mean(task$truth(train_set))
        }
        1 - sum(weights * (prediction$truth - prediction$response)^2) / sum(weights * (prediction$truth - mu)^2)
      }
    }
  )
)

#' @include mlr_measures.R
mlr_measures$add("regr.rsq", MeasureRegrRSQ)

#' @title Convert to a Classification Task
#'
#' @description
#' Convert object to a [TaskClassif].
#' This is a S3 generic. mlr3 ships with methods for the following objects:
#'
#' 1. [TaskClassif]: returns the object as-is, possibly cloned.
#' 2. [`formula`], [data.frame()], [matrix()], and [DataBackend]: provides an alternative to the
#'    constructor of [TaskClassif].
#' 3. [TaskRegr]: Calls [convert_task()].
#'
#' Note that the target column will be converted to a `factor()`, if possible.
#'
#' @inheritParams as_task
#'
#' @return [TaskClassif].
#' @export
#' @examples
#' as_task_classif(palmerpenguins::penguins, target = "species")
as_task_classif = function(x, ...) {
  UseMethod("as_task_classif")
}

#' @rdname as_task_classif
#' @export
# nolint next
as_task_classif.TaskClassif = function(x, clone = FALSE, ...) {
  if (clone) x$clone() else x
}

#' @rdname as_task_classif
#' @template param_target
#' @param id (`character(1)`)\cr
#'   Id for the new task.
#'   Defaults to the (deparsed and substituted) name of the data argument.
#' @param positive (`character(1)`)\cr
#'   Level of the positive class. See [TaskClassif].
#' @template param_label
#' @export
# nolint next
as_task_classif.data.frame = function(
  x,
  target,
  id = deparse1(substitute(x)),
  positive = NULL,
  label = NA_character_,
  ...
) {
  force(id)

  assert_data_frame(x, min.rows = 1L, min.cols = 1L, col.names = "unique")
  assert_choice(target, names(x))

  ii = which(map_lgl(keep(x, is.double), anyInfinite))
  if (length(ii)) {
    warning_input("Detected columns with unsupported Inf values in data: %s", str_collapse(names(ii)))
  }

  if (anyMissing(x[[target]])) {
    messagef("Target column '%s' contains missing values. This will be an error in the future.", target)
    # error_input("Target column '%s' must not contain missing values", target)
  }

  y = x[[target]]
  if (!is.factor(y)) {
    x[[target]] = as.factor(y)
  }

  TaskClassif$new(id = id, backend = x, target = target, positive = positive, label = label)
}

#' @rdname as_task_classif
#' @export
# nolint next
as_task_classif.matrix = function(x, target, id = deparse1(substitute(x)), label = NA_character_, ...) {
  force(id)

  assert_matrix(x, col.names = "unique", min.rows = 1L, min.cols = 1L)
  assert_choice(target, colnames(x))

  as_task_classif(as.data.table(x), target = target, id = id, label = label, ...)
}

#' @rdname as_task_classif
#' @export
# nolint next
as_task_classif.DataBackend = function(
  x,
  target,
  id = deparse1(substitute(x)),
  positive = NULL,
  label = NA_character_,
  ...
) {
  force(id)

  assert_choice(target, x$colnames)

  TaskClassif$new(id = id, backend = x, target = target, positive = positive, label = label, ...)
}

#' @rdname as_task_classif
#' @inheritParams convert_task
#' @export
# nolint next
as_task_classif.TaskRegr = function(x, target, drop_original_target = FALSE, drop_levels = TRUE, ...) {
  convert_task(intask = x, target = target, new_type = "classif", drop_original_target = drop_original_target, drop_levels = drop_levels)
}

#' @rdname as_task_classif
#' @param data (`data.frame()`)\cr
#'   Data frame containing all columns referenced in formula `x`.
#' @export
# nolint next
as_task_classif.formula = function(
  x,
  data,
  id = deparse1(substitute(data)),
  positive = NULL,
  label = NA_character_,
  ...
) {
  force(id)

  assert_data_frame(data)
  assert_subset(all.vars(x), c(names(data), "."), .var.name = "formula")

  if (!attr(terms(x, data = data), "response")) {
    error_input("Formula %s is missing a response", format(x))
  }
  tab = model.frame(x, data, na.action = "na.pass")
  setattr(tab, "terms", NULL)
  setattr(tab, "na.action", NULL)
  target = all.vars(x)[1L]

  as_task_classif(tab, target = target, id = id, positive = positive, label = label, ...)
}

#' @title Selected Features Measure
#'
#' @name mlr_measures_selected_features
#' @include Measure.R
#'
#' @description
#' Measures the number of selected features by extracting it from learners with property `"selected_features"`.
#' If parameter `normalize` is set to `TRUE`, the relative number of features instead of the absolute
#' number of features is returned.
#' Note that the models must be stored to be able to extract this information.
#' If the learner does not support the extraction of used features, `NA` is returned.
#'
#' This measure requires the [Task] and the [Learner] for scoring.
#'
#' @templateVar id selected_features
#' @template measure
#'
#' @template seealso_measure
#' @export
#' @examples
#' task = tsk("german_credit")
#' learner = lrn("classif.rpart")
#' rr = resample(task, learner, rsmp("cv", folds = 3), store_models = TRUE)
#'
#' scores = rr$score(msr("selected_features"))
#' scores[, c("iteration", "selected_features")]
MeasureSelectedFeatures = R6Class(
  "MeasureSelectedFeatures",
  inherit = Measure,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      param_set = ps(normalize = p_lgl(tags = "required"))
      param_set$set_values(normalize = FALSE)

      super$initialize(
        id = "selected_features",
        param_set = param_set,
        task_type = NA_character_,
        properties = c("requires_task", "requires_learner", "requires_model", "requires_no_prediction"),
        predict_sets = NULL,
        predict_type = NA_character_,
        range = c(0, Inf),
        minimize = TRUE,
        label = "Absolute or Relative Frequency of Selected Features",
        man = "mlr3::mlr_measures_selected_features"
      )
    }
  ),

  private = list(
    .score = function(prediction, task, learner, ...) {
      learner = learner$base_learner()
      if ("selected_features" %nin% learner$properties) {
        return(NA_integer_)
      }

      n = length(learner$selected_features())
      if (self$param_set$get_values()$normalize) {
        n = n / length(task$feature_names)
      }

      n
    }
  )
)

#' @include mlr_measures.R
mlr_measures$add("selected_features", function() MeasureSelectedFeatures$new())

#' @title Leave-One-Out Cross-Validation
#'
#' @name mlr_resamplings_loo
#' @include Resampling.R
#'
#' @description
#' Splits data using leave-one-observation-out.
#' This is identical to cross-validation with the number of folds set
#' to the number of observations.
#'
#' If this resampling is combined with the grouping features of tasks,
#' it is possible to create custom splits based on an arbitrary factor variable,
#' see the examples.
#'
#' @templateVar id loo
#' @template resampling
#'
#' @references
#' `r format_bib("bischl_2012")`
#'
#' @template seealso_resampling
#' @export
#' @examples
#' # Create a task with 10 observations
#' task = tsk("penguins")
#' task$filter(1:10)
#'
#' # Instantiate Resampling
#' loo = rsmp("loo")
#' loo$instantiate(task)
#'
#' # Individual sets:
#' loo$train_set(1)
#' loo$test_set(1)
#'
#' # Disjunct sets:
#' intersect(loo$train_set(1), loo$test_set(1))
#'
#' # Internal storage:
#' loo$instance # vector
#'
#' # Combine with group feature of tasks:
#' task = tsk("penguins")
#' task$set_col_roles("island", add_to = "group")
#' loo$instantiate(task)
#' loo$iters # one fold for each level of "island"
ResamplingLOO = R6Class(
  "ResamplingLOO",
  inherit = Resampling,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      super$initialize(id = "loo", label = "Leave-One-Out", man = "mlr3::mlr_resamplings_loo")
    }
  ),

  active = list(
    #' @field iters (`integer(1)`)\cr
    #' Returns the number of resampling iterations which is the number of rows of the task
    #' provided to instantiate. Is `NA` if the resampling has not been instantiated.
    iters = function(rhs) {
      assert_ro_binding(rhs)
      if (is.null(self$instance)) NA_integer_ else length(self$instance)
    }
  ),

  private = list(
    .sample = function(ids, ...) {
      shuffle(ids)
    },

    .get_train = function(i) {
      self$instance[-i]
    },

    .get_test = function(i) {
      self$instance[i]
    },

    .combine = function(instances) {
      do.call(c, instances)
    }
  )
)

#' @include mlr_resamplings.R
mlr_resamplings$add("loo", function() ResamplingLOO$new())

#' @title Holdout Resampling
#'
#' @name mlr_resamplings_holdout
#' @include Resampling.R
#'
#' @description
#' Splits data into a training set and a test set.
#' Parameter `ratio` determines the ratio of observation going into the training set (default: 2/3).
#'
#' @templateVar id holdout
#' @template resampling
#'
#' @section Parameters:
#' * `ratio` (`numeric(1)`)\cr
#'   Ratio of observations to put into the training set.
#'
#' @references
#' `r format_bib("bischl_2012")`
#'
#' @template seealso_resampling
#' @export
#' @examples
#' # Create a task with 10 observations
#' task = tsk("penguins")
#' task$filter(1:10)
#'
#' # Instantiate Resampling
#' holdout = rsmp("holdout", ratio = 0.5)
#' holdout$instantiate(task)
#'
#' # Individual sets:
#' holdout$train_set(1)
#' holdout$test_set(1)
#'
#' # Disjunct sets:
#' intersect(holdout$train_set(1), holdout$test_set(1))
#'
#' # Internal storage:
#' holdout$instance # simple list
ResamplingHoldout = R6Class(
  "ResamplingHoldout",
  inherit = Resampling,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function() {
      ps = ps(
        ratio = p_dbl(0, 1, tags = "required")
      )
      ps$set_values(ratio = 2 / 3)

      super$initialize(id = "holdout", param_set = ps, label = "Holdout", man = "mlr3::mlr_resamplings_holdout")
    }
  ),
  active = list(
    #' @template field_iters
    iters = function(rhs) {
      assert_ro_binding(rhs)
      1L
    }
  ),
  private = list(
    .sample = function(ids, ...) {
      n = length(ids)
      in_train = logical(n)
      in_train[sample.int(n, round(n * self$param_set$values$ratio))] = TRUE
      list(train = ids[in_train], test = ids[!in_train])
    },

    .get_train = function(i) {
      self$instance$train
    },

    .get_test = function(i) {
      self$instance$test
    },

    .combine = function(instances) {
      list(train = do.call(c, map(instances, "train")), test = do.call(c, map(instances, "test")))
    }
  )
)

#' @include mlr_resamplings.R
mlr_resamplings$add("holdout", function() ResamplingHoldout$new())

#' @title Classification Learner
#'
#' @include Learner.R
#'
#' @description
#' This Learner specializes [Learner] for classification problems:
#'
#' * `task_type` is set to `"classif"`.
#' * Creates [Prediction]s of class [PredictionClassif].
#' * Possible values for `predict_types` are:
#'   - `"response"`: Predicts a class label for each observation in the test set.
#'   - `"prob"`: Predicts the posterior probability for each class for each observation in the test set.
#' * Additional learner properties include:
#'   - `"twoclass"`: The learner works on binary classification problems.
#'   - `"multiclass"`: The learner works on multiclass classification problems.
#'
#' Predefined learners can be found in the [dictionary][mlr3misc::Dictionary] [mlr_learners].
#' Essential classification learners can be found in this dictionary after loading \CRANpkg{mlr3learners}.
#' Additional learners are implemented in the Github package \url{https://github.com/mlr-org/mlr3extralearners}.
#'
#' @template param_id
#' @template param_task_type
#' @template param_param_set
#' @template param_predict_types
#' @template param_feature_types
#' @template param_learner_properties
#' @template param_packages
#' @template param_label
#' @template param_man
#'
#' @template seealso_learner
#' @export
#' @examples
#' # get all classification learners from mlr_learners:
#' lrns = mlr_learners$mget(mlr_learners$keys("^classif"))
#' names(lrns)
#'
#' # get a specific learner from mlr_learners:
#' lrn = lrn("classif.rpart")
#' print(lrn)
#'
#' # train the learner:
#' task = tsk("penguins")
#' lrn$train(task, 1:200)
#'
#' # predict on new observations:
#' lrn$predict(task, 201:344)$confusion
LearnerClassif = R6Class(
  "LearnerClassif",
  inherit = Learner,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    initialize = function(
      id,
      task_type = "classif",
      param_set = ps(),
      predict_types = "response",
      feature_types = character(),
      properties = character(),
      packages = character(),
      label = NA_character_,
      man = NA_character_
    ) {
      super$initialize(
        id = id,
        task_type = task_type,
        param_set = param_set,
        predict_types = predict_types,
        feature_types = feature_types,
        properties = properties,
        packages = packages,
        label = label,
        man = man
      )

      if (getOption("mlr3.prob_as_default", FALSE) && "prob" %in% self$predict_types) {
        self$predict_type = "prob"
      }
    },

    #' @description
    #' Predicts outcomes for new data in `newdata` using the model fitted during `$train()`.
    #' This method is faster than `$predict_newdata()` as it skips assertions,
    #' type conversions, encapsulation, and logging.
    #' The learner must be trained before calling this method.
    #'
    #' Unlike `$predict_newdata()`, this method does not return a [Prediction] object.
    #' Instead, it returns a list with either a `"response"` or `"prob"` element, depending on the prediction type.
    #'
    #' Note that `state$predict_time` and `state$log` will remain empty after using this method.
    #' Some learners may not support this method and may fail when it is called.
    #' Prefer `$predict_newdata()` unless performance is critical.
    #'
    #' If the model was trained via [resample()] or [benchmark()],
    #' you must pass the associated task object stored in the corresponding [ResampleResult] or [BenchmarkResult].
    #'
    #' @param newdata [`data.table::data.table()`]\cr
    #'   New data to predict on.
    #' @param task ([Task]).
    #'
    #' @return `list()` with elements `"response"` or `"prob"` depending on the predict type.
    predict_newdata_fast = function(newdata, task = NULL) {
      if (is.null(task) && is.null(self$state$train_task)) {
        error_input("No task stored, and no task provided")
      }
      feature_names = self$state$train_task$feature_names %??% task$feature_names
      class_names = self$state$train_task$class_names %??% task$class_names

      # add data and most common used meta data
      fake_task = list(
        # some learners require that newdata contains the features names in the same order as the training data
        data = function(...) newdata[, feature_names, with = FALSE],
        class_names = class_names,
        feature_names = feature_names,
        nrow = nrow(newdata)
      )

      # train failed, use fallback
      if (is.null(self$model) && !is.null(self$state$fallback_state$model)) {
        return(self$fallback$predict_newdata_fast(newdata))
      }
      pred = get_private(self)$.predict(fake_task)

      # predict missing predictions with fallback
      miss = logical(fake_task$nrow)
      if (!is.null(pred$response)) {
        miss = is.na(pred$response)
      }

      if (!is.null(pred$prob)) {
        miss = miss | row_any_na(pred$prob)
      }

      miss_ids = which(miss)
      if (length(miss_ids) && !is.null(self$state$fallback_state$model)) {
        pred_miss = self$fallback$predict_newdata_fast(newdata[miss_ids, ])

        if (!is.null(pred$response)) {
          pred$response[miss_ids] = pred_miss$response
        }

        if (!is.null(pred$prob)) {
          pred$prob[miss_ids, ] = pred_miss$prob
        }
      }

      return(pred)
    }
  )
)

#' @include DataBackend.R
DataBackendCbind = R6Class(
  "DataBackendCbind",
  inherit = DataBackend,
  cloneable = FALSE,
  public = list(
    initialize = function(b1, b2) {
      assert_backend(b1)
      assert_backend(b2)
      pk = b1$primary_key

      if (pk != b2$primary_key) {
        error_input("All backends to cbind must have the primary_key '%s'", pk)
      }

      super$initialize(list(b1 = b1, b2 = b2), pk)
    },

    data = function(rows, cols) {
      pk = self$primary_key
      qrows = unique(assert_numeric(rows))
      qcols = union(assert_names(cols, type = "unique"), pk)

      data = private$.data$b2$data(qrows, qcols)
      if (ncol(data) < length(qcols)) {
        qcols = c(setdiff(cols, names(data)), pk)
        tmp = private$.data$b1$data(qrows, qcols)
        data = merge(data, tmp, by = pk, all = TRUE, sort = TRUE)
      }

      # duplicate rows / reorder columns
      ijoin(data, rows, intersect(cols, names(data)), pk)
    },

    head = function(n = 6L) {
      rows = head(self$rownames, n)
      self$data(rows = rows, cols = self$colnames)
    },

    distinct = function(rows, cols, na_rm = TRUE) {
      d2 = private$.data$b2$distinct(rows, cols, na_rm = na_rm)
      d1 = private$.data$b1$distinct(rows, setdiff(cols, names(d2)), na_rm = na_rm)
      res = c(d1, d2)
      res[reorder_vector(names(res), cols)]
    },

    missings = function(rows, cols) {
      m2 = private$.data$b2$missings(rows, cols)
      m1 = private$.data$b1$missings(rows, setdiff(cols, names(m2)))
      if (length(m2) == 0L && length(m1) == 0L) {
        return(set_names(integer()))
      }

      res = c(m1, m2)
      res[reorder_vector(names(res), cols)]
    }
  ),

  active = list(
    rownames = function() {
      union(private$.data$b1$rownames, private$.data$b2$rownames)
    },

    colnames = function() {
      union(private$.data$b1$colnames, private$.data$b2$colnames)
    },

    nrow = function() {
      uniqueN(c(private$.data$b1$rownames, private$.data$b2$rownames))
    },

    ncol = function() {
      uniqueN(c(private$.data$b1$colnames, private$.data$b2$colnames))
    },

    col_hashes = function() {
      insert_named(private$.data$b1$col_hashes, private$.data$b2$col_hashes)
    }
  ),

  private = list(
    .calculate_hash = function() {
      data = private$.data
      calculate_hash(data$b1$hash, data$b2$hash)
    }
  )
)

#' @title Resample a Learner on a Task
#'
#' @description
#' Runs a resampling (possibly in parallel):
#' Repeatedly apply [Learner] `learner` on a training set of [Task] `task` to train a model,
#' then use the trained model to predict observations of a test set.
#' Training and test sets are defined by the [Resampling] `resampling`.
#'
#' @section Stochasticity:
#' Note that uninstantiated [`Resampling`]s are instantiated on the task, making
#' the procedure stochastic even in case of a deterministic learner.
#'
#' @param task ([Task]).
#' @param learner ([Learner]).
#' @param resampling ([Resampling]).
#' @template param_store_models
#' @template param_store_backends
#' @template param_encapsulate
#' @template param_allow_hotstart
#' @template param_clone
#' @template param_unmarshal
#' @template param_callbacks
#' @return [ResampleResult].
#'
#' @template section_predict_sets
#' @template section_parallelization
#' @template section_progress_bars
#' @template section_logging
#'
#' @note
#' The fitted models are discarded after the predictions have been computed in order to reduce memory consumption.
#' If you need access to the models for later analysis, set `store_models` to `TRUE`.
#'
#' @template seealso_resample
#' @export
#' @examples
#' task = tsk("penguins")
#' learner = lrn("classif.rpart")
#' resampling = rsmp("cv")
#'
#' # Explicitly instantiate the resampling for this task for reproducibility
#' set.seed(123)
#' resampling$instantiate(task)
#'
#' rr = resample(task, learner, resampling)
#' print(rr)
#'
#' # Retrieve performance
#' rr$score(msr("classif.ce"))
#' rr$aggregate(msr("classif.ce"))
#'
#' # merged prediction objects of all resampling iterations
#' pred = rr$prediction()
#' pred$confusion
#'
#' # Repeat resampling with featureless learner
#' rr_featureless = resample(task, lrn("classif.featureless"), resampling)
#'
#' # Convert results to BenchmarkResult, then combine them
#' bmr1 = as_benchmark_result(rr)
#' bmr2 = as_benchmark_result(rr_featureless)
#' print(bmr1$combine(bmr2))
resample = function(
  task,
  learner,
  resampling,
  store_models = FALSE,
  store_backends = TRUE,
  encapsulate = NA_character_,
  allow_hotstart = FALSE,
  clone = c("task", "learner", "resampling"),
  unmarshal = TRUE,
  callbacks = NULL
) {
  lg$debug("Start resampling")

  assert_subset(clone, c("task", "learner", "resampling"))
  task = assert_task(as_task(task, clone = "task" %chin% clone))
  learner = assert_learner(as_learner(learner, clone = "learner" %chin% clone, discard_state = TRUE))
  resampling = assert_resampling(as_resampling(resampling, clone = "resampling" %chin% clone))
  assert_flag(store_models)
  assert_flag(store_backends)
  # this does not check the internal validation task as it might not be set yet
  assert_learnable(task, learner)
  assert_flag(unmarshal)
  callbacks = assert_callbacks(as_callbacks(callbacks))

  set_encapsulation(list(learner), encapsulate)
  if (!resampling$is_instantiated) {
    lg$info("Resampling '%s' is being instantiated on task '%s'", resampling$id, task$id)
    resampling = resampling$instantiate(task)
  }

  if (!is.null(resampling$task_row_hash) && resampling$task_row_hash != task$row_hash) {
    error_input("Resampling '%s' is not instantiated on task '%s'", resampling$id, task$id)
  }

  n = resampling$iters
  pb = if (isNamespaceLoaded("progressr")) {
    # NB: the progress bar needs to be created in this env
    pb = progressr::progressor(steps = n)
  } else {
    NULL
  }

  grid = if (allow_hotstart) {
    lg$debug("Resampling with hotstart enabled.")

    hotstart_grid = map_dtr(seq_len(n), function(iteration) {
      if (!is.null(learner$hotstart_stack)) {
        # search for hotstart learner
        task_hashes = resampling_task_hashes(task, resampling, learner)
        start_learner = get_private(learner$hotstart_stack)$.start_learner(learner$clone(), task_hashes[iteration])
      }
      if (is.null(learner$hotstart_stack) || is.null(start_learner)) {
        # no hotstart learners stored or no adaptable model found
        lg$debug("Resampling with hotstarting not possible. No start learner found.")
        mode = "train"
      } else {
        # hotstart learner found
        lg$debug("Resampling with hotstarting.")
        start_learner$param_set$values = insert_named(start_learner$param_set$values, learner$param_set$values)
        learner = start_learner
        mode = "hotstart"
      }
      data.table(learner = list(learner), mode = mode)
    })

    # null hotstart stack to reduce overhead in parallelization
    walk(hotstart_grid$learner, function(learner) {
      learner$hotstart_stack = NULL
      learner
    })
    hotstart_grid
  } else {
    data.table(learner = replicate(n, learner), mode = "train")
  }

  res = future_map(
    n,
    workhorse,
    iteration = seq_len(n),
    learner = grid$learner,
    mode = grid$mode,
    MoreArgs = list(
      task = task,
      resampling = resampling,
      store_models = store_models,
      lgr_index = lgr::logger_index(),
      pb = pb,
      unmarshal = unmarshal,
      callbacks = callbacks
    )
  )

  data = data.table(
    task = list(task),
    learner = grid$learner,
    learner_state = map(res, "learner_state"),
    resampling = list(resampling),
    iteration = seq_len(n),
    prediction = map(res, "prediction"),
    uhash = UUIDgenerate(),
    param_values = map(res, "param_values"),
    learner_hash = map_chr(res, "learner_hash")
  )

  data_extra = if (length(callbacks) && any(map_lgl(res, function(x) !is.null(x$data_extra)))) map(res, "data_extra")

  result_data = ResultData$new(data, data_extra, store_backends = store_backends)

  # the worker already ensures that models are sent back in marshaled form if unmarshal = FALSE, so we don't have
  # to do anything in this case. This allows us to minimize the amount of marshaling in those situations where
  # the model is available in both states on the worker
  if (unmarshal && store_models) {
    result_data$unmarshal()
  }

  ResampleResult$new(result_data)
}

# nolint start

#' @include mlr_reflections.R
mlr_reflections$auto_converters = ee = new.env(parent = emptyenv())

## from: logical
ee[["logical___integer"]] =
  function(value, type, levels) {
    as.integer(value)
  }
ee[["logical___numeric"]] =
  function(value, type, levels) {
    as.double(value)
  }
ee[["logical___character"]] =
  function(value, type, levels) {
    as.character(value)
  }
ee[["logical___factor"]] =
  function(value, type, levels) {
    factor(value, levels = union(levels, value), ordered = FALSE)
  }
ee[["logical___ordered"]] =
  function(value, type, levels) {
    if (all(value %in% c(NA_character_, levels))) factor(value, levels = levels, ordered = TRUE) else NULL
  }
ee[["logical___POSIXct"]] =
  function(value, type, levels) {
    if (allMissing(value)) .POSIXct(value, tz = "") else value
  }
ee[["logical___Date"]] =
  function(value, type, levels) {
    if (allMissing(value)) as.Date(value) else value
  }

## from: integer
ee[["integer___logical"]] =
  function(value, type, levels) {
    if (test_integerish(value, lower = 0L, upper = 1L)) as.logical(value) else value
  }
ee[["integer___numeric"]] =
  ee[["logical___numeric"]]
ee[["integer___character"]] =
  ee[["logical___character"]]
ee[["integer___factor"]] =
  ee[["logical___factor"]]
ee[["integer___ordered"]] =
  ee[["logical___ordered"]]
ee[["integer___POSIXct"]] =
  ee[["logical___POSIXct"]]
ee[["integer___Date"]] =
  ee[["logical___Date"]]

## from: numeric
ee[["numeric___logical"]] =
  ee[["integer___logical"]]
ee[["numeric___integer"]] =
  function(value, type, levels) {
    if (test_integerish(value)) as.integer(value) else value
  }
ee[["numeric___character"]] =
  ee[["logical___character"]]
ee[["numeric___factor"]] =
  ee[["logical___factor"]]
ee[["numeric___ordered"]] =
  ee[["logical___ordered"]]
ee[["numeric___POSIXct"]] =
  ee[["logical___POSIXct"]]
ee[["numeric___Date"]] =
  ee[["logical___Date"]]

## from: character
ee[["character___logical"]] =
  function(value, type, levels) {
    if (all(value %in% c(NA_character_, "TRUE", "FALSE", "true", "false", "T", "F"))) as.logical(value) else value
  }
ee[["character___integer"]] =
  function(value, type, levels) {
    if (allMissing(value)) as.integer(value) else value
  }
ee[["character___numeric"]] =
  function(value, type, levels) {
    if (allMissing(value)) as.double(value) else value
  }
ee[["character___factor"]] =
  ee[["logical___factor"]]
ee[["character___ordered"]] =
  ee[["logical___ordered"]]
ee[["character___POSIXct"]] =
  function(value, type, levels) {
    x = try(as.POSIXct(value, ""), silent = TRUE)
    if (inherits(x, "try-error")) value else x
  }
ee[["character___Date"]] =
  function(value, type, levels) {
    x = try(as.Date(value), silent = TRUE)
    if (inherits(x, "try-error")) value else x
  }

## from: factor
ee[["factor___logical"]] =
  ee[["character___logical"]]
ee[["factor___character"]] =
  ee[["logical___character"]]
ee[["factor___factor"]] =
  function(value, type, levels) {
    factor(value, levels = union(levels, levels(value)), ordered = FALSE)
  }
ee[["factor___ordered"]] =
  function(value, type, levels) {
    if (all(levels(value) %in% levels)) factor(value, levels = levels, ordered = TRUE) else NULL
  }
ee[["factor___POSIXct"]] =
  ee[["character___POSIXct"]]
ee[["factor___Date"]] =
  ee[["character___Date"]]

## from: ordered
ee[["ordered___character"]] =
  ee[["logical___character"]]
ee[["ordered___factor"]] =
  ee[["factor___factor"]]
ee[["ordered___ordered"]] =
  ee[["ordered___ordered"]]
ee[["ordered___POSIXct"]] =
  ee[["character___POSIXct"]]
ee[["ordered___Date"]] =
  ee[["character___Date"]]

## from: POSIXct
ee[["POSIXct___Date"]] =
  function(value, type, levels) {
    as.Date(value)
  }

## from: Date
ee[["Date___POSIXct"]] =
  function(value, type, levels) {
    as.POSIXct(value)
  }

rm(ee)
# nolint end

#' @title Column Auto-Converter
#'
#' @description
#' Set of rules to automatically convert column types.
#' This is used during `rbind`-ing of [Task]s, but also in some pipe operators in
#' \CRANpkg{mlr3pipelines}.
#'
#' All rules are stored as functions in [mlr_reflections$auto_converters][mlr_reflections].
#'
#' @param value (any)\cr
#'   New values to convert in order to match `type`.
#' @param id (`character(1)`)\cr
#'   Name of the column, used in error messages.
#' @param type (`character(1)`)\cr
#'   Type to convert `values` to.
#' @param levels (`character()` | `NULL`)\cr
#'   Levels to use for conversion to `factor` or `ordered`.
#'
#' @return Vector `value` converted to type `type`.
#'
#' @keywords internal
#' @export
auto_convert = function(value, id, type, levels) {
  cl = class(value)[1L]
  fun = get0(sprintf("%s___%s", cl, type), envir = mlr_reflections$auto_converters, inherits = FALSE)

  if (!is.null(fun)) {
    value = fun(value, type, levels)
  }

  if (class(value)[1L] != type) {
    error_input(
      "Incompatible types during auto-converting column '%s': failed to convert from class '%s' to class '%s'",
      id,
      cl,
      type
    )
  }

  value
}

test_that("logical", {
  x = c(TRUE, NA, FALSE)

  expect_identical(
    auto_convert(x, "x", "logical", character()),
    c(TRUE, NA, FALSE)
  )

  expect_identical(
    auto_convert(x, "x", "integer", character()),
    c(1L, NA, 0L)
  )

  expect_identical(
    auto_convert(x, "x", "numeric", character()),
    c(1, NA, 0)
  )

  expect_identical(
    auto_convert(x, "x", "character", character()),
    c("TRUE", NA, "FALSE")
  )

  expect_identical(
    auto_convert(x, "x", "factor", c("TRUE", "FALSE")),
    factor(c("TRUE", NA, "FALSE"), levels = c("TRUE", "FALSE"))
  )

  expect_identical(
    auto_convert(x, "x", "ordered", c("TRUE", "FALSE")),
    ordered(c("TRUE", NA, "FALSE"), levels = c("TRUE", "FALSE"))
  )
})


test_that("integer", {
  x = c(1L, NA, 0L)

  expect_identical(
    auto_convert(x, "x", "logical", character()),
    c(TRUE, NA, FALSE)
  )

  expect_identical(
    auto_convert(x, "x", "integer", character()),
    c(1L, NA, 0L)
  )

  expect_identical(
    auto_convert(x, "x", "numeric", character()),
    c(1, NA, 0)
  )

  expect_identical(
    auto_convert(x, "x", "character", character()),
    c("1", NA, "0")
  )

  expect_identical(
    auto_convert(x, "x", "factor", c("1", "0")),
    factor(c("1", NA, "0"), levels = c("1", "0"))
  )

  expect_identical(
    auto_convert(x, "x", "ordered", c("1", "0")),
    ordered(c("1", NA, "0"), levels = c("1", "0"))
  )
})

test_that("numeric", {
  x = c(1, NA, 0)

  expect_identical(
    auto_convert(x, "x", "logical", character()),
    c(TRUE, NA, FALSE)
  )

  expect_identical(
    auto_convert(x, "x", "integer", character()),
    c(1L, NA, 0L)
  )

  expect_identical(
    auto_convert(x, "x", "numeric", character()),
    c(1, NA, 0)
  )

  expect_identical(
    auto_convert(x, "x", "character", character()),
    c("1", NA, "0")
  )

  expect_identical(
    auto_convert(x, "x", "factor", c("1", "0")),
    factor(c("1", NA, "0"), levels = c("1", "0"))
  )

  expect_identical(
    auto_convert(x, "x", "ordered", c("1", "0")),
    ordered(c("1", NA, "0"), levels = c("1", "0"))
  )
})

test_that("character", {
  x = c("1", NA, "0")

  expect_error(
    auto_convert(x, "x", "logical", character()),
    "[Ii]ncompatible type"
  )

  expect_error(
    auto_convert(x, "x", "integer", character()),
    "[Ii]ncompatible type"
  )

  expect_error(
    auto_convert(x, "x", "numeric", character()),
    "[Ii]ncompatible type"
  )

  expect_identical(
    auto_convert(as.character(x == "1"), "x", "logical", character()),
    c(TRUE, NA, FALSE)
  )

  expect_identical(
    auto_convert(x, "x", "character", character()),
    c("1", NA, "0")
  )

  expect_identical(
    auto_convert(x, "x", "factor", c("1", "0")),
    factor(c("1", NA, "0"), levels = c("1", "0"))
  )

  expect_identical(
    auto_convert(x, "x", "ordered", c("1", "0")),
    ordered(c("1", NA, "0"), levels = c("1", "0"))
  )
})

test_that("factor", {
  x = factor(c("1", NA, "0"), levels = c("1", "0"))

  expect_error(
    auto_convert(x, "x", "logical", character()),
    "[Ii]ncompatible type"
  )

  expect_error(
    auto_convert(x, "x", "integer", character()),
    "[Ii]ncompatible type"
  )

  expect_error(
    auto_convert(x, "x", "numeric", character()),
    "[Ii]ncompatible type"
  )
  expect_identical(
    auto_convert(x, "x", "character", character()),
    c("1", NA, "0")
  )

  expect_identical(
    auto_convert(x, "x", "factor", c("1", "0")),
    factor(c("1", NA, "0"), levels = c("1", "0"))
  )

  expect_identical(
    auto_convert(x, "x", "ordered", c("1", "0")),
    ordered(c("1", NA, "0"), levels = c("1", "0"))
  )
})

test_that("ordered", {
  x = ordered(c("1", NA, "0"), levels = c("1", "0"))

  expect_error(
    auto_convert(x, "x", "logical", character()),
    "[Ii]ncompatible type"
  )

  expect_error(
    auto_convert(x, "x", "integer", character()),
    "[Ii]ncompatible type"
  )

  expect_error(
    auto_convert(x, "x", "numeric", character()),
    "[Ii]ncompatible type"
  )

  expect_identical(
    auto_convert(x, "x", "character", character()),
    c("1", NA, "0")
  )

  expect_identical(
    auto_convert(x, "x", "factor", c("1", "0")),
    factor(c("1", NA, "0"), levels = c("1", "0"))
  )

  expect_identical(
    auto_convert(x, "x", "ordered", c("1", "0")),
    ordered(c("1", NA, "0"), levels = c("1", "0"))
  )
})

test_that("POSIXct", {
  # these tests are disabled, R-devel is changing stuff too frequently
  # expect_identical(
  #   auto_convert(NA, "x", "POSIXct", character()),
  #   .POSIXct(NA, "")
  # )

  # expect_identical(
  #   auto_convert(NA_integer_, "x", "POSIXct", character()),
  #   .POSIXct(NA_integer_, "")
  # )

  # expect_identical(
  #   auto_convert(NA_real_, "x", "POSIXct", character()),
  #   .POSIXct(NA_real_, "")
  # )

  # expect_identical(
  #   auto_convert(NA_character_, "x", "POSIXct", character()),
  #   as.POSIXct(NA_character_)
  # )

  # expect_identical(
  #   auto_convert(factor(NA_character_, levels = "a"), "x", "POSIXct", "a"),
  #   as.POSIXct(NA_character_)
  # )

  # expect_identical(
  #   auto_convert(ordered(NA_character_, levels = "a"), "x", "POSIXct", "a"),
  #   as.POSIXct(NA_character_)
  # )

  expect_identical(
    auto_convert("2020-01-20 10:00:00", "x", "POSIXct", character()),
    as.POSIXct("2020-01-20 10:00:00", "")
  )

  expect_date(auto_convert(Sys.time(), "x", "Date", character()))
})

test_that("Date", {
  expect_date(auto_convert("2021-04-21", "x", "Date", character()))
})

fix_factor_levels = function(data, levels, ...) {
  UseMethod("fix_factor_levels")
}

#' @export
# nolint next
fix_factor_levels.data.table = function(data, levels, ...) {
  levels = levels[intersect(names(levels), names(data))]
  iwalk(
    levels,
    function(lvls, id, data) {
      x = data[[id]]
      if (!identical(levels(x), lvls)) {
        set(data, j = id, value = factor(x, levels = lvls, ordered = is.ordered(x)))
      }
    },
    data = data
  )
  data[]
}

#' @title Reflections for mlr3
#'
#' @format [environment].
#' @description
#' Environment which stores various information to allow objects to examine and introspect their
#' structure and properties
#' (c.f. [Reflections](https://en.wikipedia.org/wiki/Reflective_programming)).
#'
#' This environment be modified by third-party packages, e.g. by adding information about new task types
#' or by extending the set of allowed feature types.
#'
#' Third-party packages that modify the reflections must register themselves in the `loaded_packages` field.
#'
#' The following objects are set by \CRANpkg{mlr3}:
#'
#' * `task_types` (`data.table()`)\cr
#'   Table with task type (`"type"`), the implementing package (`"pkg"`), and the names of the generators
#'   of the corresponding [Task] (`"task"`), [Learner] (`"learner"`),
#'   [Prediction] (`"prediction"`), [Measure] (`"measure"`) and fallback [Learner].
#'   The column `"type"` must be unique for each row.
#'
#' * `task_feature_types` (named `character()`)\cr
#'   Vector of base R types supported as [Task] features, named with a 3 letter abbreviation.
#'
#' * `task_row_roles` (`character()`)\cr
#'   Vector of supported row roles for a [Task].
#'
#' * `task_col_roles` (list of `character()`)\cr
#'   List of vectors of supported column roles for a [Task], named by their task type.
#'
#' * `task_properties` (list of `character()`)\cr
#'   List of vectors of supported [Task] properties, named by their task type.
#'
#' * `task_mandatory_properties` (list of `character()`)\cr
#'   List of vectors of [Task] properties which necessarily must be supported by the [Learner].
#'   I.e., if the task property is not found in the set of the learner properties, an exception
#'   is raised.
#'
#' * `task_print_col_roles` (list of named `character()`)\cr
#'   Vector of column roles to print via `print(task)` if the role is not empty, either
#'   before or after the task's target, properties and features.
#'   The names of the column roles are the values, the names correspond to the labels to use in the printer.
#'
#' * `learner_properties` (list of `character()`)\cr
#'   List of vectors of supported [Learner] properties, named by their task type.
#'
#' * `learner_predict_types` (list of list of `character()`)\cr
#'   List of lists of supported [Learner] predict_types, named by their task type.
#'   The inner list translates the `"predict_type"` to all predict types returned, e.g.
#'   predict type `"prob"` for a [LearnerClassif] provides the probabilities as well as the
#'   predicted labels, therefore `"prob"` maps to `c("response", "prob")`.
#'
#' * `learner_param_tags` (`character()`)\cr
#'   Character vector of allowed 'tags' for the [paradox::ParamSet]s of a [Learner].
#'
#' * `predict_sets` (`character()`)\cr
#'   Vector of possible predict sets. Currently supported are `"train"`, `"test"` and `"internal_valid"`.
#'
#' * `measure_properties` (list of `character()`)\cr
#'   List of vectors of supported [Measure] properties, named by their task type.
#'
#' * `default_measures` (list of `character()`)\cr
#'   List of keys for the default [Measure]s, named by their task type.
#'
#' * `rr_names` (`character()`)\cr
#'   Names of the objects stored in a [ResampleResult].
#'
#' * `auto_converters` (`environment()`)\cr
#'   Environment of converter functions used for `rbind`-ing data to tasks.
#'   Functions are named using the pattern `"[from_type]___[to_type]"`.
#'   Can be extended by third-party with additional converters.
#'
#' @keywords internal
#' @export
#' @examples
#' ls.str(mlr_reflections)
mlr_reflections = new.env(parent = emptyenv())

local({
  ### Task
  # task types + constructors
  # nolint start
  # fmt: skip
  mlr_reflections$task_types = rowwise_table(.key = "type",
    ~type,          ~package, ~task,              ~learner,         ~prediction,          ~prediction_data,         ~measure,
    "regr",         "mlr3",   "TaskRegr",         "LearnerRegr",    "PredictionRegr",     "PredictionDataRegr",     "MeasureRegr",
    "classif",      "mlr3",   "TaskClassif",      "LearnerClassif", "PredictionClassif",  "PredictionDataClassif",  "MeasureClassif",
    "unsupervised", "mlr3",   "TaskUnsupervised", "Learner",        NA_character_,        NA_character_,            NA_character_
  )
  # nolint end

  mlr_reflections$task_feature_types = c(
    lgl = "logical",
    int = "integer",
    dbl = "numeric",
    chr = "character",
    fct = "factor",
    ord = "ordered",
    pxc = "POSIXct",
    dte = "Date"
  )

  mlr_reflections$task_row_roles = c(
    "use"
  )

  tmp = c("feature", "target", "name", "order", "stratum", "group", "offset", "weights_learner", "weights_measure")
  mlr_reflections$task_col_roles = list(
    regr = tmp,
    classif = tmp,
    unsupervised = c("feature", "name", "order")
  )

  tmp = c("strata", "groups", "offset", "weights_learner", "weights_measure")
  mlr_reflections$task_properties = list(
    classif = c(tmp, "twoclass", "multiclass"),
    regr = tmp,
    unsupervised = character()
  )

  mlr_reflections$task_mandatory_properties = list(
    classif = c("twoclass", "multiclass")
  )

  mlr_reflections$task_print_col_roles = list(
    before = character(),
    after = c(
      "Order by" = "order",
      "Strata" = "stratum",
      "Groups" = "group",
      "Offset" = "offset",
      "Weights/Learner" = "weights_learner",
      "Weights/Measure" = "weights_measure"
    )
  )

  tmp = c("weights_learner", "weights_measure", "name", "order", "stratum", "group")
  mlr_reflections$task_col_roles_optional_newdata = list(
    classif = tmp,
    regr = tmp
  )

  ### Learner
  tmp = c(
    "featureless",
    "missings",
    "weights",
    "importance",
    "selected_features",
    "oob_error",
    "hotstart_forward",
    "hotstart_backward",
    "validation",
    "internal_tuning",
    "marshal",
    "offset",
    "new_levels"
  )
  mlr_reflections$learner_properties = list(
    classif = c(tmp, "twoclass", "multiclass"),
    regr = tmp
  )

  mlr_reflections$learner_predict_types = list(
    classif = list(response = "response", prob = c("response", "prob")),
    regr = list(
      response = "response",
      se = c("response", "se"),
      quantiles = c("response", "quantiles"),
      distr = c("response", "se", "distr")
    ),
    unsupervised = NULL
  )

  # Allowed tags for parameters
  mlr_reflections$learner_param_tags = c(
    "train",
    "predict",
    "hotstart",
    "importance",
    "threads",
    "required",
    "internal_tuning"
  )

  ### Prediction
  mlr_reflections$predict_sets = c("train", "test", "internal_valid")

  ### Resampling
  mlr_reflections$resampling_properties = c("duplicated_ids", "weights")

  ### Measures
  tmp = c(
    "na_score",
    "requires_task",
    "requires_learner",
    "requires_model",
    "requires_train_set",
    "weights",
    "primary_iters",
    "requires_no_prediction",
    "obs_loss"
  )
  mlr_reflections$measure_properties = list(
    classif = tmp,
    regr = tmp
  )

  mlr_reflections$default_measures = list(
    classif = "classif.ce",
    regr = "regr.mse",
    unsupervised = NA_character_
  )

  ### ResampleResult
  mlr_reflections$rr_names = c("task", "learner", "resampling", "iteration")

  ### Cached package version
  mlr_reflections$package_version = packageVersion("mlr3")

  ### Loaded packages
  mlr_reflections$loaded_packages = "mlr3"
})

test_that("warn_deprecated_old works as expected", {
  oldopts = options(mlr3.warn_deprecated = TRUE)
  expect_warning(warn_deprecated_old("test"), "^test is deprecated and will be removed in the future\\.$")
  expect_no_warning(warn_deprecated_old("test")) # no second warning

  oldopts = options(mlr3.warn_deprecated = FALSE)
  expect_no_warning(warn_deprecated_old("test2")) # no warning when options disallow it

  options(oldopts)
})

test_that("deprecated_binding_old works as expected", {
  oldopts = options(mlr3.warn_deprecated = TRUE)
  MyClass = R6::R6Class(
    "MyClass",
    public = list(val = 1),
    active = list(
      foo = deprecated_binding_old("MyClass$foo", "bar"),
      foo2 = deprecated_binding_old("MyClass$foo2", self$val)
    ),
  )
  mco = MyClass$new()
  expect_warning(
    {
      fooval = mco$foo
    },
    "^MyClass\\$foo is deprecated and will be removed in the future\\.$"
  )
  expect_equal(fooval, "bar")
  expect_no_warning({
    fooval = mco$foo
  }) # no second warning
  expect_equal(fooval, "bar")

  oldopts = options(mlr3.warn_deprecated = FALSE)
  expect_no_warning({
    foo2val = mco$foo2
  }) # no warning when options disallow it
  expect_equal(foo2val, 1)
  mco$val = 2
  expect_equal(mco$foo2, 2)

  options(oldopts)
})

test_that("expecations do not use globals", {
  skip_if_not_installed("codetools")

  ee = new.env()
  path = system.file("testthat", "helper_expectations.R", package = "mlr3")
  sys.source(path, envir = ee)

  testthat::expect_silent(codetools::checkUsageEnv(ee))
})

# $characteristics works

    Code
      task
    Output
      
      -- <TaskClassif> (4601x58): HP Spam Detection ----------------------------------
      * Target: type
      * Properties: twoclass
      * Features (57):
        * dbl (57): address, addresses, all, business, capitalAve, capitalLong,
        capitalTotal, charDollar, charExclamation, charHash, charRoundbracket,
        charSemicolon, charSquarebracket, conference, credit, cs, data, direct, edu,
        email, font, free, george, hp, hpl, internet, lab, labs, mail, make, meeting,
        money, num000, num1999, num3d, num415, num650, num85, num857, order,
        original, our, over, parts, people, pm, project, re, receive, remove, report,
        table, technology, telnet, will, you, your
      * Characteristics: foo=1, bar=a
      * Target classes: spam (positive class, 39%), nonspam (61%)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskClassif_german_credit.R
\name{mlr_tasks_german_credit}
\alias{mlr_tasks_german_credit}
\title{German Credit Classification Task}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link{TaskClassif}.
}
\source{
Data set originally published on \href{http://archive.ics.uci.edu/dataset/144/statlog+german+credit+data}{UCI}.
This is the preprocessed version taken from package \CRANpkg{rchallenge} with
factors instead of dummy variables, and corrected as proposed by Ulrike
Grömping.

Donor:
Professor Dr. Hans Hofmann\cr
Institut für Statistik und Ökonometrie\cr
Universität Hamburg\cr
FB Wirtschaftswissenschaften\cr
Von-Melle-Park 5\cr
2000 Hamburg 13
}
\description{
A classification task for the German credit data set.
The aim is to predict creditworthiness, labeled as "good" and "bad".
Positive class is set to label "good".

See the example for the creation of a \link{MeasureClassifCosts} with the described misclassification costs.
}
\section{Dictionary}{

This \link{Task} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_tasks} or with the associated sugar function \code{\link[=tsk]{tsk()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("german_credit")
tsk("german_credit")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Dimensions: 1000x21
\item Properties: \dQuote{twoclass}
\item Has Missings: \code{FALSE}
\item Target: \dQuote{credit_risk}
\item Features: \dQuote{age}, \dQuote{amount}, \dQuote{credit_history}, \dQuote{duration}, \dQuote{employment_duration}, \dQuote{foreign_worker}, \dQuote{housing}, \dQuote{installment_rate}, \dQuote{job}, \dQuote{number_credits}, \dQuote{other_debtors}, \dQuote{other_installment_plans}, \dQuote{people_liable}, \dQuote{personal_status_sex}, \dQuote{present_residence}, \dQuote{property}, \dQuote{purpose}, \dQuote{savings}, \dQuote{status}, \dQuote{telephone}
}
}

\examples{
task = tsk("german_credit")
costs = matrix(c(0, 1, 5, 0), nrow = 2)
dimnames(costs) = list(predicted = task$class_names, truth = task$class_names)
measure = msr("classif.costs", id = "german_credit_costs", costs = costs)
print(measure)
}
\references{
Grömping U (2019).
\dQuote{South German Credit Data: Correcting a Widely Used Data Set.}
Reports in Mathematics, Physics and Chemistry 4, Department II, Beuth University of Applied Sciences Berlin.
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{Task}},
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskRegr_mtcars.R
\name{mlr_tasks_mtcars}
\alias{mlr_tasks_mtcars}
\title{Motor Trend Regression Task}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link{TaskRegr}.
}
\description{
A regression task for the \link[datasets:mtcars]{datasets::mtcars} data set.
Target variable is \code{mpg} (Miles/(US) gallon).
Rownames are stored as variable \code{"model"} with column role \code{"name"}.
}
\section{Construction}{


\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("mtcars")
tsk("mtcars")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{regr}
\item Dimensions: 32x11
\item Properties: -
\item Has Missings: \code{FALSE}
\item Target: \dQuote{mpg}
\item Features: \dQuote{am}, \dQuote{carb}, \dQuote{cyl}, \dQuote{disp}, \dQuote{drat}, \dQuote{gear}, \dQuote{hp}, \dQuote{qsec}, \dQuote{vs}, \dQuote{wt}
}
}

\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{Task}},
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_prediction.R
\name{as_prediction}
\alias{as_prediction}
\alias{as_prediction.Prediction}
\alias{as_prediction.PredictionDataClassif}
\alias{as_prediction.PredictionDataRegr}
\alias{as_predictions}
\alias{as_predictions.list}
\title{Convert to a Prediction}
\usage{
as_prediction(x, check = FALSE, ...)

\method{as_prediction}{Prediction}(x, check = FALSE, ...)

\method{as_prediction}{PredictionDataClassif}(x, check = FALSE, ...)

\method{as_prediction}{PredictionDataRegr}(x, check = FALSE, ...)

as_predictions(x, predict_sets = "test", ...)

\method{as_predictions}{list}(x, predict_sets = "test", ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{check}{(\code{logical(1)})\cr
Perform argument checks and type conversions?}

\item{...}{(any)\cr
Additional arguments.}

\item{predict_sets}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
}
\value{
\link{Prediction}.
}
\description{
Convert object to a \link{Prediction} or a list of \link{Prediction}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ResamplingCustom.R
\name{mlr_resamplings_custom}
\alias{mlr_resamplings_custom}
\alias{ResamplingCustom}
\title{Custom Resampling}
\description{
Splits data into training and test sets using manually provided indices.
}
\section{Dictionary}{

This \link{Resampling} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_resamplings} or with the associated sugar function \code{\link[=rsmp]{rsmp()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_resamplings$get("custom")
rsmp("custom")
}\if{html}{\out{</div>}}
}

\examples{
# Create a task with 10 observations
task = tsk("penguins")
task$filter(1:10)

# Instantiate Resampling
custom = rsmp("custom")
train_sets = list(1:5, 5:10)
test_sets = list(5:10, 1:5)
custom$instantiate(task, train_sets, test_sets)

custom$train_set(1)
custom$test_set(1)
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter3/evaluation_and_benchmarking.html#sec-resampling}
\item Package \CRANpkg{mlr3spatiotempcv} for spatio-temporal resamplings.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Resampling]{Resamplings}: \link{mlr_resamplings}
\item \code{as.data.table(mlr_resamplings)} for a table of available \link[=Resampling]{Resamplings} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3spatiotempcv} for additional \link{Resampling}s for spatio-temporal
tasks.
}

Other Resampling:
\code{\link{Resampling}},
\code{\link{mlr_resamplings}},
\code{\link{mlr_resamplings_bootstrap}},
\code{\link{mlr_resamplings_custom_cv}},
\code{\link{mlr_resamplings_cv}},
\code{\link{mlr_resamplings_holdout}},
\code{\link{mlr_resamplings_insample}},
\code{\link{mlr_resamplings_loo}},
\code{\link{mlr_resamplings_repeated_cv}},
\code{\link{mlr_resamplings_subsampling}}
}
\concept{Resampling}
\section{Super class}{
\code{\link[mlr3:Resampling]{Resampling}} -> \code{ResamplingCustom}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{iters}}{(\code{integer(1)})\cr
Returns the number of resampling iterations, depending on the values stored in the \code{param_set}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ResamplingCustom-initialize}{\code{ResamplingCustom$new()}}
    \item \href{#method-ResamplingCustom-instantiate}{\code{ResamplingCustom$instantiate()}}
    \item \href{#method-ResamplingCustom-clone}{\code{ResamplingCustom$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="format"><a href='../../mlr3/html/Resampling.html#method-Resampling-format'><code>Resampling$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="help"><a href='../../mlr3/html/Resampling.html#method-Resampling-help'><code>Resampling$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="print"><a href='../../mlr3/html/Resampling.html#method-Resampling-print'><code>Resampling$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="test_set"><a href='../../mlr3/html/Resampling.html#method-Resampling-test_set'><code>Resampling$test_set()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="train_set"><a href='../../mlr3/html/Resampling.html#method-Resampling-train_set'><code>Resampling$train_set()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResamplingCustom-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ResamplingCustom-initialize}{}}}
\subsection{\code{ResamplingCustom$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResamplingCustom$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResamplingCustom-instantiate"></a>}}
\if{latex}{\out{\hypertarget{method-ResamplingCustom-instantiate}{}}}
\subsection{\code{ResamplingCustom$instantiate()}}{
  Instantiate this \link{Resampling} with custom splits into training and test set.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResamplingCustom$instantiate(task, train_sets, test_sets)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{\link{Task}\cr
Mainly used to check if \code{train_sets} and \code{test_sets} are feasible.}
      \item{\code{train_sets}}{(list of \code{integer()})\cr
List with row ids for training, one list element per iteration.
Must have the same length as \code{test_sets}.}
      \item{\code{test_sets}}{(list of \code{integer()})\cr
List with row ids for testing, one list element per iteration.
Must have the same length as \code{train_sets}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResamplingCustom-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ResamplingCustom-clone}{}}}
\subsection{\code{ResamplingCustom$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResamplingCustom$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/default_fallback.R
\name{default_fallback}
\alias{default_fallback}
\alias{default_fallback.Learner}
\alias{default_fallback.LearnerClassif}
\alias{default_fallback.LearnerRegr}
\title{Create a Fallback Learner}
\usage{
default_fallback(learner, ...)

\method{default_fallback}{Learner}(learner, ...)

\method{default_fallback}{LearnerClassif}(learner, ...)

\method{default_fallback}{LearnerRegr}(learner, ...)
}
\arguments{
\item{learner}{\link{Learner}\cr
The learner for which a fallback learner should be created.}

\item{...}{\code{any}\cr
ignored.}
}
\value{
\link{Learner}
}
\description{
Create a fallback learner for a given learner.
The function searches for a suitable fallback learner based on the task type.
Additional checks are performed to ensure that the fallback learner supports the predict type.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSimple.R
\name{mlr_measures_sim.jaccard}
\alias{mlr_measures_sim.jaccard}
\title{Jaccard Similarity Index}
\description{
Measure to compare two or more sets w.r.t. their similarity.
}
\details{
For two sets \eqn{A} and \eqn{B}, the Jaccard Index is defined as
\deqn{
  J(A, B) = \frac{|A \cap B|}{|A \cup B|}.
}{
  J(A, B) = length(intersect(A, B)) / length(union(A, B)).
}
If more than two sets are provided, the mean of all pairwise scores
is calculated.

This measure is undefined if two or more sets are empty.
}
\note{
This measure requires learners with property \code{"selected_features"}.
The extracted feature sets are passed to \code{\link[mlr3measures:jaccard]{mlr3measures::jaccard()}} from
package \CRANpkg{mlr3measures}.

If the measure is undefined for the input, \code{NaN} is returned.
This can be customized by setting the field \code{na_value}.
}
\section{Parameters}{

Empty ParamSet
}

\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("sim.jaccard")
msr("sim.jaccard")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{


\itemize{
\item Type: \code{"similarity"}
\item Range: \eqn{[0, 1]}{[0, 1]}
\item Minimize: \code{FALSE}
}

}

\seealso{
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}

\code{as.data.table(mlr_measures)} for a complete table of all (also dynamically created) \link{Measure} implementations.

Other similarity measures:
\code{\link{mlr_measures_sim.phi}}
}
\concept{similarity measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSimple.R
\name{mlr_measures_classif.mcc}
\alias{mlr_measures_classif.mcc}
\title{Matthews Correlation Coefficient}
\description{
Measure to compare true observed labels with predicted
labels
in multiclass classification tasks.
}
\details{
In the binary case, the Matthews Correlation Coefficient is defined as \deqn{
   \frac{\mathrm{TP} \cdot \mathrm{TN} - \mathrm{FP} \cdot \mathrm{FN}}{\sqrt{(\mathrm{TP} + \mathrm{FP})
   (\mathrm{TP} + \mathrm{FN}) (\mathrm{TN} + \mathrm{FP}) (\mathrm{TN} + \mathrm{FN})}},
}{
   (TP * TN - FP * FN) / sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)),
}
where \eqn{TP}, \eqn{FP}, \eqn{TN}, \eqn{TP} are the number of true positives, false positives,
true negatives, and false negatives respectively.

In the multi-class case, the Matthews Correlation Coefficient is defined
for a multi-class confusion matrix \eqn{C} with \eqn{K} classes: \deqn{
   \frac{c \cdot s - \sum_k^K p_k \cdot t_k}{\sqrt{(s^2 - \sum_k^K p_k^2) \cdot (s^2 - \sum_k^K t_k^2)}},
}{
   (c * s - sum(pk * tk)) / sqrt((s^2 - sum(pk^2)) * (s^2 - sum(tk^2))),
}
where
\itemize{
\item \eqn{s = \sum_i^K \sum_j^K C_{ij}}: total number of samples
\item \eqn{c = \sum_k^K C_{kk}}: total number of correctly predicted samples
\item \eqn{t_k = \sum_i^K C_{ik}}: number of predictions for each class \eqn{k}
\item \eqn{p_k = \sum_j^K C_{kj}}: number of true occurrences for each class \eqn{k}.
}

The above formula is undefined if any of the four sums in the denominator is 0 in the binary case,
and more generally if either \eqn{s^2 - \sum_k^K p_k^2} or \eqn{s^2 - \sum_k^K t_k^2)} is equal to 0.
The denominator is then set to 1.

When there are more than two classes, the MCC will no longer range between -1 and +1.
Instead, the minimum value will be between -1 and 0 depending on the true distribution.
The maximum value is always +1.
}
\note{
The score function calls \code{\link[mlr3measures:mcc]{mlr3measures::mcc()}} from package \CRANpkg{mlr3measures}.

If the measure is undefined for the input, \code{NaN} is returned.
This can be customized by setting the field \code{na_value}.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("classif.mcc")
msr("classif.mcc")
}\if{html}{\out{</div>}}
}

\section{Parameters}{

Empty ParamSet
}

\section{Meta Information}{


\itemize{
\item Type: \code{"classif"}
\item Range: \eqn{[-1, 1]}{[-1, 1]}
\item Minimize: \code{FALSE}
\item Required prediction: \code{response}
}

}

\seealso{
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}

\code{as.data.table(mlr_measures)} for a complete table of all (also dynamically created) \link{Measure} implementations.

Other classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.auc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.bbrier}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.dor}},
\code{\link{mlr_measures_classif.fbeta}},
\code{\link{mlr_measures_classif.fdr}},
\code{\link{mlr_measures_classif.fn}},
\code{\link{mlr_measures_classif.fnr}},
\code{\link{mlr_measures_classif.fomr}},
\code{\link{mlr_measures_classif.fp}},
\code{\link{mlr_measures_classif.fpr}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}},
\code{\link{mlr_measures_classif.npv}},
\code{\link{mlr_measures_classif.ppv}},
\code{\link{mlr_measures_classif.prauc}},
\code{\link{mlr_measures_classif.precision}},
\code{\link{mlr_measures_classif.recall}},
\code{\link{mlr_measures_classif.sensitivity}},
\code{\link{mlr_measures_classif.specificity}},
\code{\link{mlr_measures_classif.tn}},
\code{\link{mlr_measures_classif.tnr}},
\code{\link{mlr_measures_classif.tp}},
\code{\link{mlr_measures_classif.tpr}}

Other multiclass classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}}
}
\concept{classification measures}
\concept{multiclass classification measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task_converters.R
\name{convert_task}
\alias{convert_task}
\title{Convert a Task from One Type to Another}
\usage{
convert_task(
  intask,
  target = NULL,
  new_type = NULL,
  drop_original_target = FALSE,
  drop_levels = TRUE
)
}
\arguments{
\item{intask}{(\link{Task})\cr
A \link{Task} to be converted.}

\item{target}{(\code{character(1)})\cr
New target to be set, must be a column in the \code{intask} data.
If \code{NULL}, no new target is set, and task is converted as-is.}

\item{new_type}{(\code{character(1)})\cr
The new task type. Must be in \code{\link[=mlr_reflections]{mlr_reflections$task_types}}].
If \code{NULL} (default), a new task with the same task_type is created.}

\item{drop_original_target}{(\code{logical(1)})\cr
If \code{FALSE} (default), the original target is added as a feature.
Otherwise the original target is dropped.}

\item{drop_levels}{(\code{logical(1)})\cr
If \code{TRUE} (default), unused levels of the new target variable are dropped.}
}
\value{
\link{Task} of requested type.
}
\description{
The task's target is replaced by a different column from the data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Prediction.R
\name{Prediction}
\alias{Prediction}
\title{Abstract Prediction Object}
\description{
This is the abstract base class for task objects like \link{PredictionClassif} or \link{PredictionRegr}.

Prediction objects store the following information:
\enumerate{
\item The row ids of the test set
\item The corresponding true (observed) response.
\item The corresponding predicted response.
\item Additional predictions based on the class and \code{predict_type}.
E.g., the class probabilities for classification or the estimated standard error for regression.
}

Note that this object is usually constructed via a derived classes, e.g. \link{PredictionClassif} or \link{PredictionRegr}.
}
\section{S3 Methods}{

\itemize{
\item \code{as.data.table(rr)}\cr
\link{Prediction} -> \code{\link[data.table:data.table]{data.table::data.table()}}\cr
Converts the data to a \code{\link[data.table:data.table]{data.table::data.table()}}.
\item \code{c(..., keep_duplicates = TRUE)}\cr
(\link{Prediction}, \link{Prediction}, ...) -> \link{Prediction}\cr
Combines multiple \code{Prediction}s to a single \code{Prediction}.
If \code{keep_duplicates} is \code{FALSE} and there are duplicated row ids,
the data of the former passed objects get overwritten by the data of the later passed objects.
}
}

\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Prediction:
\code{\link{PredictionClassif}},
\code{\link{PredictionRegr}}
}
\concept{Prediction}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{data}}{(named \code{list()})\cr
Internal data structure.}

    \item{\code{task_type}}{(\code{character(1)})\cr
Required type of the \link{Task}.}

    \item{\code{task_properties}}{(\code{character()})\cr
Required properties of the \link{Task}.}

    \item{\code{predict_types}}{(\code{character()})\cr
Set of predict types this object stores.}

    \item{\code{man}}{(\code{character(1)} | \code{NULL})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
Defaults to \code{NA}, but can be set by child classes.}
  }
  \if{html}{\out{</div>}}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{row_ids}}{(\code{integer()})\cr
Vector of row ids for which predictions are stored.}

    \item{\code{truth}}{(any)\cr
True (observed) outcome.}

    \item{\code{missing}}{(\code{integer()})\cr
Returns \code{row_ids} for which the predictions are missing or incomplete.}

    \item{\code{weights}}{(\code{numeric()})\cr
Vector of measure weights, obtained from the \code{weights_measure} column of the \link{Task} if present.
This is \code{NULL} if no weights are present.}

    \item{\code{extra}}{(\code{list()})\cr
List of extra data stored in the prediction object.}

    \item{\code{raw}}{(any)\cr
Raw prediction object from the upstream model.
Only available if the learner's \code{predict_raw} flag was set to \code{TRUE} and the learner implementation
supports it.
The object is stored as-is, without validation or subsetting during filtering.
When multiple predictions are combined via \code{c()}, the individual raw objects are collected into a \code{list()}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Prediction-format}{\code{Prediction$format()}}
    \item \href{#method-Prediction-print}{\code{Prediction$print()}}
    \item \href{#method-Prediction-help}{\code{Prediction$help()}}
    \item \href{#method-Prediction-score}{\code{Prediction$score()}}
    \item \href{#method-Prediction-obs_loss}{\code{Prediction$obs_loss()}}
    \item \href{#method-Prediction-filter}{\code{Prediction$filter()}}
    \item \href{#method-Prediction-clone}{\code{Prediction$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Prediction-format"></a>}}
\if{latex}{\out{\hypertarget{method-Prediction-format}{}}}
\subsection{\code{Prediction$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Prediction$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Prediction-print"></a>}}
\if{latex}{\out{\hypertarget{method-Prediction-print}{}}}
\subsection{\code{Prediction$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Prediction$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Prediction-help"></a>}}
\if{latex}{\out{\hypertarget{method-Prediction-help}{}}}
\subsection{\code{Prediction$help()}}{
  Opens the corresponding help page referenced by field \verb{$man}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Prediction$help()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Prediction-score"></a>}}
\if{latex}{\out{\hypertarget{method-Prediction-score}{}}}
\subsection{\code{Prediction$score()}}{
  Calculates the performance for all provided measures
\link{Task} and \link{Learner} may be \code{NULL} for most measures, but some measures need to extract information
from these objects.
Note that the \code{predict_sets} of the \code{measures} are ignored by this method,
instead all predictions are used.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Prediction$score(
  measures = NULL,
  task = NULL,
  learner = NULL,
  train_set = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{measures}}{(\link{Measure} | list of \link{Measure} | \code{NULL})\cr
Measure(s) to calculate.}
      \item{\code{task}}{(\link{Task}).}
      \item{\code{learner}}{(\link{Learner}).}
      \item{\code{train_set}}{(\code{integer()}).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \link{Prediction}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Prediction-obs_loss"></a>}}
\if{latex}{\out{\hypertarget{method-Prediction-obs_loss}{}}}
\subsection{\code{Prediction$obs_loss()}}{
  Calculates the observation-wise loss via the \link{Measure}'s \code{obs_loss} method.
Returns a \code{data.table()} with the columns of the matching \link{Prediction} object plus one additional numeric column
for each measure, named with the respective measure id.
If there is no observation-wise loss function for the measure, the column is filled with \code{NA_real_} values.
Note that some measures such as RMSE, do have an \verb{$obs_loss},
but they require an additional transformation after aggregation,
in this example taking the square-root.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Prediction$obs_loss(measures = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{measures}}{(\link{Measure} | list of \link{Measure} | \code{NULL})\cr
Measure(s) to calculate.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Prediction-filter"></a>}}
\if{latex}{\out{\hypertarget{method-Prediction-filter}{}}}
\subsection{\code{Prediction$filter()}}{
  Filters the \link{Prediction}, keeping only predictions for the provided row_ids.
This changes the object in-place, you need to create a clone to preserve
the original \link{Prediction}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Prediction$filter(row_ids)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{row_ids}}{\code{integer()}\cr
Row indices.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{self}, modified.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Prediction-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Prediction-clone}{}}}
\subsection{\code{Prediction$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Prediction$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskRegr_california_housing.R
\name{california_housing}
\alias{california_housing}
\alias{mlr_tasks_california_housing}
\title{Median House Value in California}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link{TaskRegr}.
}
\source{
https://www.kaggle.com/datasets/camnugent/california-housing-prices
}
\description{
A regression task to predict the median house value in California.

Contains 9 features and 20640 observations.
Target column is \code{"median_house_value"}.
}
\section{Construction}{


\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("california_housing")
tsk("california_housing")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{regr}
\item Dimensions: 20640x10
\item Properties: -
\item Has Missings: \code{TRUE}
\item Target: \dQuote{median_house_value}
\item Features: \dQuote{households}, \dQuote{housing_median_age}, \dQuote{latitude}, \dQuote{longitude}, \dQuote{median_income}, \dQuote{ocean_proximity}, \dQuote{population}, \dQuote{total_bedrooms}, \dQuote{total_rooms}
}
}

\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{Task}},
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ResampleResult.R
\name{ResampleResult}
\alias{ResampleResult}
\title{Container for Results of \code{resample()}}
\description{
This is the result container object returned by \code{\link[=resample]{resample()}}.

Note that all stored objects are accessed by reference.
Do not modify any object without cloning it first.

\link{ResampleResult}s can be visualized via \CRANpkg{mlr3viz}'s \code{autoplot()} function.
}
\section{S3 Methods}{

\itemize{
\item \code{as.data.table(rr, reassemble_learners = TRUE, convert_predictions = TRUE, predict_sets = "test")}\cr
\link{ResampleResult} -> \code{\link[data.table:data.table]{data.table::data.table()}}\cr
Returns a tabular view of the internal data.
\item \code{c(...)}\cr
(\link{ResampleResult}, ...) -> \link{BenchmarkResult}\cr
Combines multiple objects convertible to \link{BenchmarkResult} into a new \link{BenchmarkResult}.
}
}

\examples{
task = tsk("penguins")
learner = lrn("classif.rpart")
resampling = rsmp("cv", folds = 3)
rr = resample(task, learner, resampling)
print(rr)

# combined predictions and predictions for each fold separately
rr$prediction()
rr$predictions()

# folds scored separately, then aggregated (macro)
rr$aggregate(msr("classif.acc"))

# predictions first combined, then scored (micro)
rr$prediction()$score(msr("classif.acc"))

# check for warnings and errors
rr$warnings
rr$errors

## ------------------------------------------------
## Method `ResampleResult$prediction()`
## ------------------------------------------------

rr$prediction()

## ------------------------------------------------
## Method `ResampleResult$predictions()`
## ------------------------------------------------

rr$predictions()

## ------------------------------------------------
## Method `ResampleResult$score()`
## ------------------------------------------------

rr$score(msr("classif.acc"))

## ------------------------------------------------
## Method `ResampleResult$obs_loss()`
## ------------------------------------------------

rr$obs_loss(msr("classif.acc"))

## ------------------------------------------------
## Method `ResampleResult$aggregate()`
## ------------------------------------------------

rr$aggregate(msr("classif.acc"))

## ------------------------------------------------
## Method `ResampleResult$filter()`
## ------------------------------------------------

rr$filter(1L)

## ------------------------------------------------
## Method `ResampleResult$marshal()`
## ------------------------------------------------

rr$marshal()

## ------------------------------------------------
## Method `ResampleResult$unmarshal()`
## ------------------------------------------------

rr$unmarshal()

## ------------------------------------------------
## Method `ResampleResult$set_threshold()`
## ------------------------------------------------

learner = lrn("classif.rpart", predict_type = "prob")
rr = resample(tsk("sonar"), learner, rsmp("cv", folds = 3))
rr$set_threshold(0.6)
}
\seealso{
\itemize{
\item \code{\link[=as_benchmark_result]{as_benchmark_result()}} to convert to a \link{BenchmarkResult}.
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter3/evaluation_and_benchmarking.html#sec-resampling}
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
}

Other resample:
\code{\link[=resample]{resample()}}
}
\concept{resample}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{task_type}}{(\code{character(1)})\cr
Task type of objects in the \code{ResampleResult}, e.g. \code{"classif"} or \code{"regr"}.
This is \code{NA} for empty \link{ResampleResult}s.}

    \item{\code{uhash}}{(\code{character(1)})\cr
Unique hash for this object.}

    \item{\code{iters}}{(\code{integer(1)})\cr
Number of resampling iterations stored in the \code{ResampleResult}.}

    \item{\code{task}}{(\link{Task})\cr
The task \code{\link[=resample]{resample()}} operated on.}

    \item{\code{learner}}{(\link{Learner})\cr
Learner prototype \code{\link[=resample]{resample()}} operated on.
For a list of \strong{trained} learners, see methods \verb{$learners()}.}

    \item{\code{resampling}}{(\link{Resampling})\cr
Instantiated \link{Resampling} object which stores the splits into training and test.}

    \item{\code{learners}}{(list of \link{Learner})\cr
List of trained learners, sorted by resampling iteration.}

    \item{\code{data_extra}}{(list())\cr
Additional data stored in the \link{ResampleResult}.}

    \item{\code{warnings}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
A table with all warning messages.
Column names are \code{"iteration"} and \code{"condition"}.
Note that there can be multiple rows per resampling iteration if multiple warnings have been recorded.}

    \item{\code{errors}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
A table with all error messages.
Column names are \code{"iteration"} and \code{"condition"}.
Note that there can be multiple rows per resampling iteration if multiple errors have been recorded.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ResampleResult-initialize}{\code{ResampleResult$new()}}
    \item \href{#method-ResampleResult-format}{\code{ResampleResult$format()}}
    \item \href{#method-ResampleResult-print}{\code{ResampleResult$print()}}
    \item \href{#method-ResampleResult-help}{\code{ResampleResult$help()}}
    \item \href{#method-ResampleResult-prediction}{\code{ResampleResult$prediction()}}
    \item \href{#method-ResampleResult-predictions}{\code{ResampleResult$predictions()}}
    \item \href{#method-ResampleResult-score}{\code{ResampleResult$score()}}
    \item \href{#method-ResampleResult-obs_loss}{\code{ResampleResult$obs_loss()}}
    \item \href{#method-ResampleResult-aggregate}{\code{ResampleResult$aggregate()}}
    \item \href{#method-ResampleResult-filter}{\code{ResampleResult$filter()}}
    \item \href{#method-ResampleResult-discard}{\code{ResampleResult$discard()}}
    \item \href{#method-ResampleResult-marshal}{\code{ResampleResult$marshal()}}
    \item \href{#method-ResampleResult-unmarshal}{\code{ResampleResult$unmarshal()}}
    \item \href{#method-ResampleResult-set_threshold}{\code{ResampleResult$set_threshold()}}
    \item \href{#method-ResampleResult-clone}{\code{ResampleResult$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-initialize}{}}}
\subsection{\code{ResampleResult$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
An alternative construction method is provided by \code{\link[=as_resample_result]{as_resample_result()}}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$new(data = ResultData$new(), view = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{data}}{(\link{ResultData} | \code{\link[=data.table]{data.table()}})\cr
An object of type \link{ResultData}, either extracted from another \link{ResampleResult}, another
\link{BenchmarkResult}, or manually constructed with \code{\link[=as_result_data]{as_result_data()}}.}
      \item{\code{view}}{(\code{character()})\cr
Single \code{uhash} of the \link{ResultData} to operate on.
Used internally for optimizations.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-format"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-format}{}}}
\subsection{\code{ResampleResult$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-print"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-print}{}}}
\subsection{\code{ResampleResult$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-help"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-help}{}}}
\subsection{\code{ResampleResult$help()}}{
  Opens the corresponding help page referenced by field \verb{$man}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$help()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-prediction"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-prediction}{}}}
\subsection{\code{ResampleResult$prediction()}}{
  Combined \link{Prediction} of all individual resampling iterations, and all provided predict sets.
Note that, per default, most performance measures do not operate on this object directly,
but instead on the prediction objects from the resampling iterations separately, and then combine
the performance scores with the aggregate function of the respective \link{Measure} (macro averaging).

If you calculate the performance on this prediction object directly, this is called micro averaging.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$prediction(predict_sets = "test")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{predict_sets}}{(\code{character()})\cr
Subset of \verb{\{"train", "test"\}}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \link{Prediction} or empty \code{list()} if no predictions are available.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{rr$prediction()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-predictions"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-predictions}{}}}
\subsection{\code{ResampleResult$predictions()}}{
  List of prediction objects, sorted by resampling iteration.
If multiple sets are given, these are combined to a single one for each iteration.

If you evaluate the performance on all of the returned prediction objects and then average them, this
is called macro averaging. For micro averaging, operate on the combined prediction object as returned by
\verb{$prediction()}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$predictions(predict_sets = "test")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{predict_sets}}{(\code{character()})\cr
Subset of \verb{\{"train", "test", "internal_valid"\}}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    List of \link{Prediction} objects, one per element in \code{predict_sets}.
Or list of empty \code{list()}s if no predictions are available.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{rr$predictions()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-score"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-score}{}}}
\subsection{\code{ResampleResult$score()}}{
  Returns a table with one row for each resampling iteration, including all involved objects:
\link{Task}, \link{Learner}, \link{Resampling}, iteration number (\code{integer(1)}), and (if enabled)
one \link{Prediction} for each predict set of the \link{Learner}.
Additionally, a column with the individual (per resampling iteration) performance is added
for each \link{Measure} in \code{measures}, named with the id of the respective measure id.
If \code{measures} is \code{NULL}, \code{measures} defaults to the return value of \code{\link[=default_measures]{default_measures()}}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$score(
  measures = NULL,
  ids = TRUE,
  conditions = FALSE,
  predictions = TRUE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{measures}}{(\link{Measure} | list of \link{Measure} | \code{NULL})\cr
Measure(s) to calculate.}
      \item{\code{ids}}{(\code{logical(1)})\cr
If \code{ids} is \code{TRUE}, extra columns with the ids of objects (\code{"task_id"}, \code{"learner_id"}, \code{"resampling_id"})
are added to the returned table.
These allow to subset more conveniently.}
      \item{\code{conditions}}{(\code{logical(1)})\cr
Adds condition messages (\code{"warnings"}, \code{"errors"}) as extra
list columns of character vectors to the returned table}
      \item{\code{predictions}}{(\code{logical(1)})\cr
Additionally return prediction objects, one column for each \code{predict_set} of the learner.
Columns are named \code{"prediction_train"}, \code{"prediction_test"} and \code{"prediction_internal_valid"},
if present.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{\link[data.table:data.table]{data.table::data.table()}}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{rr$score(msr("classif.acc"))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-obs_loss"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-obs_loss}{}}}
\subsection{\code{ResampleResult$obs_loss()}}{
  Calculates the observation-wise loss via the \link{Measure}'s \code{obs_loss} method.
Returns a \code{data.table()} with an \code{iteration} column plus one numeric column for each measure,
named with the respective measure id.
If there is no observation-wise loss function for the measure, the column is filled with \code{NA_real_} values.
Note that some measures such as RMSE, do have an \verb{$obs_loss},
but they require an additional transformation after aggregation,
in this example taking the square-root.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$obs_loss(measures = NULL, predict_sets = "test")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{measures}}{(\link{Measure} | list of \link{Measure} | \code{NULL})\cr
Measure(s) to calculate.}
      \item{\code{predict_sets}}{(\code{character()})\cr
The predict sets.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{rr$obs_loss(msr("classif.acc"))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-aggregate"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-aggregate}{}}}
\subsection{\code{ResampleResult$aggregate()}}{
  Calculates and aggregates performance values for all provided measures, according to the
respective aggregation function in \link{Measure}.
If \code{measures} is \code{NULL}, \code{measures} defaults to the return value of \code{\link[=default_measures]{default_measures()}}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$aggregate(measures = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{measures}}{(\link{Measure} | list of \link{Measure} | \code{NULL})\cr
Measure(s) to calculate.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Named \code{numeric()}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{rr$aggregate(msr("classif.acc"))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-filter"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-filter}{}}}
\subsection{\code{ResampleResult$filter()}}{
  Subsets the \link{ResampleResult}, reducing it to only keep the iterations specified in \code{iters}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$filter(iters)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{iters}}{(\code{integer()})\cr
Resampling iterations to keep.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{rr$filter(1L)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-discard"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-discard}{}}}
\subsection{\code{ResampleResult$discard()}}{
  Shrinks the \link{ResampleResult} by discarding parts of the internally stored data.
Note that certain operations might stop work, e.g. extracting
importance values from learners or calculating measures requiring the task's data.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$discard(backends = FALSE, models = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{backends}}{(\code{logical(1)})\cr
If \code{TRUE}, the \link{DataBackend} is removed from all stored \link{Task}s.}
      \item{\code{models}}{(\code{logical(1)})\cr
If \code{TRUE}, the stored model is removed from all \link{Learner}s.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-marshal"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-marshal}{}}}
\subsection{\code{ResampleResult$marshal()}}{
  Marshals all stored models.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$marshal(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=marshal_model]{marshal_model()}}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{rr$marshal()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-unmarshal"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-unmarshal}{}}}
\subsection{\code{ResampleResult$unmarshal()}}{
  Unmarshals all stored models.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$unmarshal(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=unmarshal_model]{unmarshal_model()}}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{rr$unmarshal()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-set_threshold"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-set_threshold}{}}}
\subsection{\code{ResampleResult$set_threshold()}}{
  Sets the threshold for the response prediction of classification learners, given they have
output a probability prediction for a binary classification task.
This modifies the object in-place.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$set_threshold(threshold, ties_method = "random")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{threshold}}{(\code{numeric(1)})\cr
Threshold value.}
      \item{\code{ties_method}}{(\code{character(1)})\cr
Method to handle ties in probabilities when selecting a class label.
Must be one of \code{"random"}, \code{"first"} or \code{"last"} (corresponding to the same options in \code{\link[=max.col]{max.col()}}).
\itemize{
\item \code{"random"}: Randomly select one of the tied class labels (default).
\item \code{"first"}: Select the first class label among tied values.
\item \code{"last"}: Select the last class label among tied values.
}}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{learner = lrn("classif.rpart", predict_type = "prob")
rr = resample(tsk("sonar"), learner, rsmp("cv", folds = 3))
rr$set_threshold(0.6)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResampleResult-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ResampleResult-clone}{}}}
\subsection{\code{ResampleResult$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResampleResult$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BenchmarkResult.R
\name{BenchmarkResult}
\alias{BenchmarkResult}
\title{Container for Benchmarking Results}
\description{
This is the result container object returned by \code{\link[=benchmark]{benchmark()}}.
A \link{BenchmarkResult} consists of the data of multiple \link{ResampleResult}s.
The contents of a \code{BenchmarkResult} and \link{ResampleResult} are almost identical
and the stored \link{ResampleResult}s can be extracted via the \verb{$resample_result(i)} method,
where i is the index of the performed resample experiment.
This allows us to investigate the extracted \link{ResampleResult}
and individual resampling iterations, as well as the predictions and models from each fold.

\link{BenchmarkResult}s can be visualized via \CRANpkg{mlr3viz}'s \code{autoplot()} function.

For statistical analysis of benchmark results and more advanced plots, see \CRANpkg{mlr3benchmark}.
}
\note{
All stored objects are accessed by reference.
Do not modify any extracted object without cloning it first.
}
\section{S3 Methods}{

\itemize{
\item \code{as.data.table(rr, ..., reassemble_learners = TRUE, convert_predictions = TRUE, predict_sets = "test", task_characteristics = FALSE)}\cr
\link{BenchmarkResult} -> \code{\link[data.table:data.table]{data.table::data.table()}}\cr
Returns a tabular view of the internal data.
\item \code{c(...)}\cr
(\link{BenchmarkResult}, ...) -> \link{BenchmarkResult}\cr
Combines multiple objects convertible to \link{BenchmarkResult} into a new \link{BenchmarkResult}.
}
}

\examples{
set.seed(123)
learners = list(
  lrn("classif.featureless", predict_type = "prob"),
  lrn("classif.rpart", predict_type = "prob")
)

design = benchmark_grid(
  tasks = list(tsk("sonar"), tsk("penguins")),
  learners = learners,
  resamplings = rsmp("cv", folds = 3)
)
print(design)

bmr = benchmark(design)
print(bmr)

bmr$tasks
bmr$learners

# first 5 resampling iterations
head(as.data.table(bmr, measures = c("classif.acc", "classif.auc")), 5)

# aggregate results
bmr$aggregate()

# aggregate results with hyperparameters as separate columns
mlr3misc::unnest(bmr$aggregate(params = TRUE), "params")

# extract resample result for classif.rpart
rr = bmr$aggregate()[learner_id == "classif.rpart", resample_result][[1]]
print(rr)

# access the confusion matrix of the first resampling iteration
rr$predictions()[[1]]$confusion

# reduce to subset with task id "sonar"
bmr$filter(task_ids = "sonar")
print(bmr)

## ------------------------------------------------
## Method `BenchmarkResult$marshal()`
## ------------------------------------------------

bmr$marshal()

## ------------------------------------------------
## Method `BenchmarkResult$unmarshal()`
## ------------------------------------------------

bmr$unmarshal()

## ------------------------------------------------
## Method `BenchmarkResult$score()`
## ------------------------------------------------

bmr$score(msr("classif.acc"))

## ------------------------------------------------
## Method `BenchmarkResult$obs_loss()`
## ------------------------------------------------

bmr$obs_loss(msr("classif.acc"))

## ------------------------------------------------
## Method `BenchmarkResult$aggregate()`
## ------------------------------------------------

bmr$aggregate()

## ------------------------------------------------
## Method `BenchmarkResult$filter()`
## ------------------------------------------------

design = benchmark_grid(
  tsks(c("iris", "sonar")),
  lrns(c("classif.debug", "classif.featureless")),
  rsmp("holdout")
)
bmr = benchmark(design)
bmr
bmr2 = bmr$clone(deep = TRUE)
bmr2$filter(learner_ids = "classif.featureless")
bmr2

## ------------------------------------------------
## Method `BenchmarkResult$resample_result()`
## ------------------------------------------------

design = benchmark_grid(
  tsk("iris"),
  lrns(c("classif.debug", "classif.featureless")),
  rsmp("holdout")
)
bmr = benchmark(design)
bmr$resample_result(learner_id = "classif.featureless")
bmr$resample_result(i = 1)
bmr$resample_result(uhash = uhashes(bmr, learner_id = "classif.debug"))

## ------------------------------------------------
## Method `BenchmarkResult$discard()`
## ------------------------------------------------

bmr$discard(models = TRUE)

## ------------------------------------------------
## Method `BenchmarkResult$set_threshold()`
## ------------------------------------------------

design = benchmark_grid(
  tsk("sonar"),
  lrns(c("classif.debug", "classif.featureless"), predict_type = "prob"),
  rsmp("holdout")
)
bmr = benchmark(design)
bmr$set_threshold(0.8, learner_ids = "classif.featureless")
bmr$set_threshold(0.3, i = 2)
bmr$set_threshold(0.7, uhashes = uhashes(bmr, learner_ids = "classif.featureless"))
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter3/evaluation_and_benchmarking.html#sec-benchmarking}
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \CRANpkg{mlr3benchmark} for post-hoc analysis of benchmark results.
}

Other benchmark:
\code{\link[=benchmark]{benchmark()}},
\code{\link[=benchmark_grid]{benchmark_grid()}}
}
\concept{benchmark}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{task_type}}{(\code{character(1)})\cr
Task type of objects in the \code{BenchmarkResult}.
All stored objects (\link{Task}, \link{Learner}, \link{Prediction}) in a single \code{BenchmarkResult} are
required to have the same task type, e.g., \code{"classif"} or \code{"regr"}.
This is \code{NA} for empty \link{BenchmarkResult}s.}

    \item{\code{tasks}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Table of included \link{Task}s with three columns:
\itemize{
\item \code{"task_hash"} (\code{character(1)}),
\item \code{"task_id"} (\code{character(1)}), and
\item \code{"task"} (\link{Task}).
}}

    \item{\code{learners}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Table of included \link{Learner}s with three columns:
\itemize{
\item \code{"learner_hash"} (\code{character(1)}),
\item \code{"learner_id"} (\code{character(1)}), and
\item \code{"learner"} (\link{Learner}).
}

Note that it is not feasible to access learned models via this field, as the training task would be ambiguous.
For this reason the returned learners are reset before they are returned.
Instead, select a row from the table returned by \verb{$score()}.}

    \item{\code{resamplings}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Table of included \link{Resampling}s with three columns:
\itemize{
\item \code{"resampling_hash"} (\code{character(1)}),
\item \code{"resampling_id"} (\code{character(1)}), and
\item \code{"resampling"} (\link{Resampling}).
}}

    \item{\code{resample_results}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Returns a table with three columns:
\itemize{
\item \code{uhash} (\code{character()}).
\item \code{resample_result} (\link{ResampleResult}).
}}

    \item{\code{n_resample_results}}{(\code{integer(1)})\cr
Returns the total number of stored \link{ResampleResult}s.}

    \item{\code{uhashes}}{(\code{character()})\cr
Set of (unique) hashes of all included \link{ResampleResult}s.}

    \item{\code{uhash_table}}{(\link[data.table:data.table]{data.table::data.table})\cr
Table with columns \code{uhash}, \code{learner_id}, \code{task_id} and \code{resampling_id}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-BenchmarkResult-initialize}{\code{BenchmarkResult$new()}}
    \item \href{#method-BenchmarkResult-help}{\code{BenchmarkResult$help()}}
    \item \href{#method-BenchmarkResult-format}{\code{BenchmarkResult$format()}}
    \item \href{#method-BenchmarkResult-print}{\code{BenchmarkResult$print()}}
    \item \href{#method-BenchmarkResult-combine}{\code{BenchmarkResult$combine()}}
    \item \href{#method-BenchmarkResult-marshal}{\code{BenchmarkResult$marshal()}}
    \item \href{#method-BenchmarkResult-unmarshal}{\code{BenchmarkResult$unmarshal()}}
    \item \href{#method-BenchmarkResult-score}{\code{BenchmarkResult$score()}}
    \item \href{#method-BenchmarkResult-obs_loss}{\code{BenchmarkResult$obs_loss()}}
    \item \href{#method-BenchmarkResult-aggregate}{\code{BenchmarkResult$aggregate()}}
    \item \href{#method-BenchmarkResult-filter}{\code{BenchmarkResult$filter()}}
    \item \href{#method-BenchmarkResult-resample_result}{\code{BenchmarkResult$resample_result()}}
    \item \href{#method-BenchmarkResult-discard}{\code{BenchmarkResult$discard()}}
    \item \href{#method-BenchmarkResult-set_threshold}{\code{BenchmarkResult$set_threshold()}}
    \item \href{#method-BenchmarkResult-clone}{\code{BenchmarkResult$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-initialize}{}}}
\subsection{\code{BenchmarkResult$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$new(data = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{data}}{(\code{ResultData})\cr
An object of type \code{ResultData}, either extracted from another \link{ResampleResult}, another
\link{BenchmarkResult}, or manually constructed with \code{\link[=as_result_data]{as_result_data()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-help"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-help}{}}}
\subsection{\code{BenchmarkResult$help()}}{
  Opens the help page for this object.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$help()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-format"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-format}{}}}
\subsection{\code{BenchmarkResult$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-print"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-print}{}}}
\subsection{\code{BenchmarkResult$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$print()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-combine"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-combine}{}}}
\subsection{\code{BenchmarkResult$combine()}}{
  Fuses a second \link{BenchmarkResult} into itself, mutating the \link{BenchmarkResult} in-place.
If the second \link{BenchmarkResult} \code{bmr} is \code{NULL}, simply returns \code{self}.
Note that you can alternatively use the combine function \code{\link[=c]{c()}} which calls this method internally.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$combine(bmr)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{bmr}}{(\link{BenchmarkResult})\cr
A second \link{BenchmarkResult} object.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-marshal"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-marshal}{}}}
\subsection{\code{BenchmarkResult$marshal()}}{
  Marshals all stored models.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$marshal(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=marshal_model]{marshal_model()}}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{bmr$marshal()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-unmarshal"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-unmarshal}{}}}
\subsection{\code{BenchmarkResult$unmarshal()}}{
  Unmarshals all stored models.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$unmarshal(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=unmarshal_model]{unmarshal_model()}}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{bmr$unmarshal()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-score"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-score}{}}}
\subsection{\code{BenchmarkResult$score()}}{
  Returns a table with one row for each resampling iteration, including
all involved objects: \link{Task}, \link{Learner}, \link{Resampling}, iteration number
(\code{integer(1)}), and \link{Prediction}. If \code{ids} is set to \code{TRUE}, character
column of extracted ids are added to the table for convenient
filtering: \code{"task_id"}, \code{"learner_id"}, and \code{"resampling_id"}.

Additionally calculates the provided performance measures and binds the
performance scores as extra columns. These columns are named using the id of
the respective \link{Measure}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$score(
  measures = NULL,
  ids = TRUE,
  conditions = FALSE,
  predictions = TRUE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{measures}}{(\link{Measure} | list of \link{Measure} | \code{NULL})\cr
Measure(s) to calculate.}
      \item{\code{ids}}{(\code{logical(1)})\cr
Adds object ids (\code{"task_id"}, \code{"learner_id"}, \code{"resampling_id"}) as
extra character columns to the returned table.}
      \item{\code{conditions}}{(\code{logical(1)})\cr
Adds condition messages (\code{"warnings"}, \code{"errors"}) as extra
list columns of character vectors to the returned table}
      \item{\code{predictions}}{(\code{logical(1)})\cr
Additionally return prediction objects, one column for each \code{predict_set} of all learners combined.
Columns are named \code{"prediction_train"}, \code{"prediction_test"} and \code{"prediction_internal_valid"},
if present.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{\link[data.table:data.table]{data.table::data.table()}}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{bmr$score(msr("classif.acc"))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-obs_loss"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-obs_loss}{}}}
\subsection{\code{BenchmarkResult$obs_loss()}}{
  Calculates the observation-wise loss via the \link{Measure}'s \code{obs_loss} method.
Returns a \code{data.table()} with columns from the predictions (e.g., \code{row_ids}, \code{truth}, \code{response}, etc.),
plus one numeric column for each measure, named with the respective measure id, and a \code{resample_result} column.
If there is no observation-wise loss function for the measure, the column is filled with \code{NA_real_} values.
Note that some measures such as RMSE, do have an \verb{$obs_loss},
but they require an additional transformation after aggregation,
in this example taking the square-root.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$obs_loss(measures = NULL, predict_sets = "test")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{measures}}{(\link{Measure} | list of \link{Measure} | \code{NULL})\cr
Measure(s) to calculate.}
      \item{\code{predict_sets}}{(\code{character()})\cr
The predict sets.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{bmr$obs_loss(msr("classif.acc"))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-aggregate"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-aggregate}{}}}
\subsection{\code{BenchmarkResult$aggregate()}}{
  Returns a result table where resampling iterations are combined into
\link{ResampleResult}s. A column with the aggregated performance score is
added for each \link{Measure}, named with the id of the respective measure.

The method for aggregation is controlled by the \link{Measure}, e.g. micro
aggregation, macro aggregation or custom aggregation. Most measures
default to macro aggregation.

Note that the aggregated performances just give a quick impression which
approaches work well and which approaches are probably underperforming.
However, the aggregates do not account for variance and cannot replace
a statistical test.
See \CRANpkg{mlr3viz} to get a better impression via boxplots or
\CRANpkg{mlr3benchmark} for critical difference plots and
significance tests.

For convenience, different flags can be set to extract more
information from the returned \link{ResampleResult}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$aggregate(
  measures = NULL,
  ids = TRUE,
  uhashes = FALSE,
  params = FALSE,
  conditions = FALSE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{measures}}{(\link{Measure} | list of \link{Measure} | \code{NULL})\cr
Measure(s) to calculate.}
      \item{\code{ids}}{(\code{logical(1)})\cr
Adds object ids (\code{"task_id"}, \code{"learner_id"}, \code{"resampling_id"}) as
extra character columns for convenient subsetting.}
      \item{\code{uhashes}}{(\code{logical(1)})\cr
Adds the uhash values of the \link{ResampleResult} as extra character
column \code{"uhash"}.}
      \item{\code{params}}{(\code{logical(1)})\cr
Adds the hyperparameter values as extra list column \code{"params"}. You
can unnest them with \code{\link[mlr3misc:unnest]{mlr3misc::unnest()}}.}
      \item{\code{conditions}}{(\code{logical(1)})\cr
Adds the number of resampling iterations with at least one warning as
extra integer column \code{"warnings"}, and the number of resampling
iterations with errors as extra integer column \code{"errors"}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{\link[data.table:data.table]{data.table::data.table()}}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{bmr$aggregate()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-filter"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-filter}{}}}
\subsection{\code{BenchmarkResult$filter()}}{
  Subsets the benchmark result.
You can either directly provide the row IDs or the uhashes of the resample results to keep,
or use the \code{learner_ids}, \code{task_ids} and \code{resampling_ids} arguments to filter for learner,
task and resampling IDs.
The three options are mutually exclusive.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$filter(
  i = NULL,
  uhashes = NULL,
  learner_ids = NULL,
  task_ids = NULL,
  resampling_ids = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{i}}{(\code{integer()} | \code{NULL})\cr
The iteration values to filter for.}
      \item{\code{uhashes}}{(\code{character()} | \code{NULL})\cr
The uhashes of the resample results to filter for.}
      \item{\code{learner_ids}}{(\code{character()} | \code{NULL})\cr
The learner IDs to filter for.}
      \item{\code{task_ids}}{(\code{character()} | \code{NULL})\cr
The task IDs to filter for.}
      \item{\code{resampling_ids}}{(\code{character()} | \code{NULL})\cr
The resampling IDs to filter for.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{design = benchmark_grid(
  tsks(c("iris", "sonar")),
  lrns(c("classif.debug", "classif.featureless")),
  rsmp("holdout")
)
bmr = benchmark(design)
bmr
bmr2 = bmr$clone(deep = TRUE)
bmr2$filter(learner_ids = "classif.featureless")
bmr2
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-resample_result"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-resample_result}{}}}
\subsection{\code{BenchmarkResult$resample_result()}}{
  Retrieve the i-th \link{ResampleResult}, by position, by unique hash \code{uhash} or by learner,
task and resampling IDs.
All three options are mutually exclusive.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$resample_result(
  i = NULL,
  uhash = NULL,
  task_id = NULL,
  learner_id = NULL,
  resampling_id = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{i}}{(\code{integer(1)} | \code{NULL})\cr
The iteration value to filter for.}
      \item{\code{uhash}}{(\code{character(1)} | \code{NULL})\cr
The unique identifier to filter for.}
      \item{\code{task_id}}{(\code{character(1)} | \code{NULL})\cr
The task ID to filter for.}
      \item{\code{learner_id}}{(\code{character(1)} | \code{NULL})\cr
The learner ID to filter for.}
      \item{\code{resampling_id}}{(\code{character(1)} | \code{NULL})\cr
The resampling ID to filter for.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \link{ResampleResult}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{design = benchmark_grid(
  tsk("iris"),
  lrns(c("classif.debug", "classif.featureless")),
  rsmp("holdout")
)
bmr = benchmark(design)
bmr$resample_result(learner_id = "classif.featureless")
bmr$resample_result(i = 1)
bmr$resample_result(uhash = uhashes(bmr, learner_id = "classif.debug"))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-discard"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-discard}{}}}
\subsection{\code{BenchmarkResult$discard()}}{
  Shrinks the \link{BenchmarkResult} by discarding parts of the internally stored data.
Note that certain operations might stop work, e.g. extracting
importance values from learners or calculating measures requiring the task's data.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$discard(backends = FALSE, models = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{backends}}{(\code{logical(1)})\cr
If \code{TRUE}, the \link{DataBackend} is removed from all stored \link{Task}s.}
      \item{\code{models}}{(\code{logical(1)})\cr
If \code{TRUE}, the stored model is removed from all \link{Learner}s.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{bmr$discard(models = TRUE)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-set_threshold"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-set_threshold}{}}}
\subsection{\code{BenchmarkResult$set_threshold()}}{
  Sets the threshold for the response prediction of classification learners, given they have
output a probability prediction for a binary classification task.

The resample results for which to change the threshold can either be specified directly
via \code{uhashes}, by selecting the specific iterations (\code{i}) or by filtering according to
learner, task and resampling IDs.

If none of the three options is specified, the threshold is set for all resample results.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$set_threshold(
  threshold,
  i = NULL,
  uhashes = NULL,
  learner_ids = NULL,
  task_ids = NULL,
  resampling_ids = NULL,
  ties_method = "random"
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{threshold}}{(\code{numeric(1)})\cr
Threshold value.}
      \item{\code{i}}{(\code{integer()} | \code{NULL})\cr
The iteration values to filter for.}
      \item{\code{uhashes}}{(\code{character()} | \code{NULL})\cr
The unique identifiers of the \link{ResampleResult}s for which the threshold should be set.}
      \item{\code{learner_ids}}{(\code{character()} | \code{NULL})\cr
The learner IDs for which the threshold should be set.}
      \item{\code{task_ids}}{(\code{character()} | \code{NULL})\cr
The task IDs for which the threshold should be set.}
      \item{\code{resampling_ids}}{(\code{character()} | \code{NULL})\cr
The resampling IDs for which the threshold should be set.}
      \item{\code{ties_method}}{(\code{character(1)})\cr
Method to handle ties in probabilities when selecting a class label.
Must be one of \code{"random"}, \code{"first"} or \code{"last"} (corresponding to the same options in \code{\link[=max.col]{max.col()}}).
\itemize{
\item \code{"random"}: Randomly select one of the tied class labels (default).
\item \code{"first"}: Select the first class label among tied values.
\item \code{"last"}: Select the last class label among tied values.
}}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{design = benchmark_grid(
  tsk("sonar"),
  lrns(c("classif.debug", "classif.featureless"), predict_type = "prob"),
  rsmp("holdout")
)
bmr = benchmark(design)
bmr$set_threshold(0.8, learner_ids = "classif.featureless")
bmr$set_threshold(0.3, i = 2)
bmr$set_threshold(0.7, uhashes = uhashes(bmr, learner_ids = "classif.featureless"))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BenchmarkResult-clone"></a>}}
\if{latex}{\out{\hypertarget{method-BenchmarkResult-clone}{}}}
\subsection{\code{BenchmarkResult$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{BenchmarkResult$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Task.R
\name{task_check_col_roles}
\alias{task_check_col_roles}
\alias{task_check_col_roles.Task}
\alias{task_check_col_roles.TaskClassif}
\alias{task_check_col_roles.TaskRegr}
\alias{task_check_col_roles.TaskSupervised}
\alias{task_check_col_roles.TaskUnsupervised}
\title{Check Column Roles}
\usage{
task_check_col_roles(task, new_roles, ...)

\method{task_check_col_roles}{Task}(task, new_roles, ...)

\method{task_check_col_roles}{TaskClassif}(task, new_roles, ...)

\method{task_check_col_roles}{TaskRegr}(task, new_roles, ...)

\method{task_check_col_roles}{TaskSupervised}(task, new_roles, ...)

\method{task_check_col_roles}{TaskUnsupervised}(task, new_roles, ...)
}
\arguments{
\item{task}{(\link{Task})\cr
Task.}

\item{new_roles}{(\code{list()})\cr
Column roles.}
}
\description{
Internal function to check column roles.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_reflections.R
\name{mlr_reflections}
\alias{mlr_reflections}
\title{Reflections for mlr3}
\format{
\link{environment}.
}
\usage{
mlr_reflections
}
\description{
Environment which stores various information to allow objects to examine and introspect their
structure and properties
(c.f. \href{https://en.wikipedia.org/wiki/Reflective_programming}{Reflections}).

This environment be modified by third-party packages, e.g. by adding information about new task types
or by extending the set of allowed feature types.

Third-party packages that modify the reflections must register themselves in the \code{loaded_packages} field.

The following objects are set by \CRANpkg{mlr3}:
\itemize{
\item \code{task_types} (\code{data.table()})\cr
Table with task type (\code{"type"}), the implementing package (\code{"pkg"}), and the names of the generators
of the corresponding \link{Task} (\code{"task"}), \link{Learner} (\code{"learner"}),
\link{Prediction} (\code{"prediction"}), \link{Measure} (\code{"measure"}) and fallback \link{Learner}.
The column \code{"type"} must be unique for each row.
\item \code{task_feature_types} (named \code{character()})\cr
Vector of base R types supported as \link{Task} features, named with a 3 letter abbreviation.
\item \code{task_row_roles} (\code{character()})\cr
Vector of supported row roles for a \link{Task}.
\item \code{task_col_roles} (list of \code{character()})\cr
List of vectors of supported column roles for a \link{Task}, named by their task type.
\item \code{task_properties} (list of \code{character()})\cr
List of vectors of supported \link{Task} properties, named by their task type.
\item \code{task_mandatory_properties} (list of \code{character()})\cr
List of vectors of \link{Task} properties which necessarily must be supported by the \link{Learner}.
I.e., if the task property is not found in the set of the learner properties, an exception
is raised.
\item \code{task_print_col_roles} (list of named \code{character()})\cr
Vector of column roles to print via \code{print(task)} if the role is not empty, either
before or after the task's target, properties and features.
The names of the column roles are the values, the names correspond to the labels to use in the printer.
\item \code{learner_properties} (list of \code{character()})\cr
List of vectors of supported \link{Learner} properties, named by their task type.
\item \code{learner_predict_types} (list of list of \code{character()})\cr
List of lists of supported \link{Learner} predict_types, named by their task type.
The inner list translates the \code{"predict_type"} to all predict types returned, e.g.
predict type \code{"prob"} for a \link{LearnerClassif} provides the probabilities as well as the
predicted labels, therefore \code{"prob"} maps to \code{c("response", "prob")}.
\item \code{learner_param_tags} (\code{character()})\cr
Character vector of allowed 'tags' for the \link[paradox:ParamSet]{paradox::ParamSet}s of a \link{Learner}.
\item \code{predict_sets} (\code{character()})\cr
Vector of possible predict sets. Currently supported are \code{"train"}, \code{"test"} and \code{"internal_valid"}.
\item \code{measure_properties} (list of \code{character()})\cr
List of vectors of supported \link{Measure} properties, named by their task type.
\item \code{default_measures} (list of \code{character()})\cr
List of keys for the default \link{Measure}s, named by their task type.
\item \code{rr_names} (\code{character()})\cr
Names of the objects stored in a \link{ResampleResult}.
\item \code{auto_converters} (\code{environment()})\cr
Environment of converter functions used for \code{rbind}-ing data to tasks.
Functions are named using the pattern \code{"[from_type]___[to_type]"}.
Can be extended by third-party with additional converters.
}
}
\examples{
ls.str(mlr_reflections)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskUnsupervised.R
\name{TaskUnsupervised}
\alias{TaskUnsupervised}
\title{Unsupervised Task}
\description{
This is the abstract base class for unsupervised tasks such as cluster tasks in
\CRANpkg{mlr3cluster} and \CRANpkg{mlr3spatial}.
}
\examples{
TaskUnsupervised$new("penguins", task_type = "unsupervised", backend = palmerpenguins::penguins)
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{Task}},
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}
\keyword{internal}
\section{Super class}{
\code{\link[mlr3:Task]{Task}} -> \code{TaskUnsupervised}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-TaskUnsupervised-initialize}{\code{TaskUnsupervised$new()}}
    \item \href{#method-TaskUnsupervised-clone}{\code{TaskUnsupervised$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="add_strata"><a href='../../mlr3/html/Task.html#method-Task-add_strata'><code>Task$add_strata()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="cbind"><a href='../../mlr3/html/Task.html#method-Task-cbind'><code>Task$cbind()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="data"><a href='../../mlr3/html/Task.html#method-Task-data'><code>Task$data()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="droplevels"><a href='../../mlr3/html/Task.html#method-Task-droplevels'><code>Task$droplevels()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="filter"><a href='../../mlr3/html/Task.html#method-Task-filter'><code>Task$filter()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="format"><a href='../../mlr3/html/Task.html#method-Task-format'><code>Task$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="formula"><a href='../../mlr3/html/Task.html#method-Task-formula'><code>Task$formula()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="head"><a href='../../mlr3/html/Task.html#method-Task-head'><code>Task$head()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="help"><a href='../../mlr3/html/Task.html#method-Task-help'><code>Task$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="levels"><a href='../../mlr3/html/Task.html#method-Task-levels'><code>Task$levels()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="materialize_view"><a href='../../mlr3/html/Task.html#method-Task-materialize_view'><code>Task$materialize_view()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="missings"><a href='../../mlr3/html/Task.html#method-Task-missings'><code>Task$missings()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="print"><a href='../../mlr3/html/Task.html#method-Task-print'><code>Task$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="rbind"><a href='../../mlr3/html/Task.html#method-Task-rbind'><code>Task$rbind()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="rename"><a href='../../mlr3/html/Task.html#method-Task-rename'><code>Task$rename()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="select"><a href='../../mlr3/html/Task.html#method-Task-select'><code>Task$select()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="set_col_roles"><a href='../../mlr3/html/Task.html#method-Task-set_col_roles'><code>Task$set_col_roles()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="set_levels"><a href='../../mlr3/html/Task.html#method-Task-set_levels'><code>Task$set_levels()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="set_row_roles"><a href='../../mlr3/html/Task.html#method-Task-set_row_roles'><code>Task$set_row_roles()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskUnsupervised-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-TaskUnsupervised-initialize}{}}}
\subsection{\code{TaskUnsupervised$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskUnsupervised$new(
  id,
  task_type = "unsupervised",
  backend,
  label = NA_character_,
  extra_args = list()
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier for the new instance.}
      \item{\code{task_type}}{(\code{character(1)})\cr
Type of task, e.g. \code{"regr"} or \code{"classif"}.
Must be an element of \link[=mlr_reflections]{mlr_reflections$task_types$type}.}
      \item{\code{backend}}{(\link{DataBackend})\cr
Either a \link{DataBackend}, or any object which is convertible to a \link{DataBackend} with \code{as_data_backend()}.
E.g., a \code{data.frame()} will be converted to a \link{DataBackendDataTable}.}
      \item{\code{label}}{(\code{character(1)})\cr
Label for the new instance.}
      \item{\code{extra_args}}{(named \code{list()})\cr
Named list of constructor arguments, required for converting task types
via \code{\link[=convert_task]{convert_task()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskUnsupervised-clone"></a>}}
\if{latex}{\out{\hypertarget{method-TaskUnsupervised-clone}{}}}
\subsection{\code{TaskUnsupervised$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskUnsupervised$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ResamplingCustomCV.R
\name{mlr_resamplings_custom_cv}
\alias{mlr_resamplings_custom_cv}
\alias{ResamplingCustomCV}
\title{Custom Cross-Validation}
\description{
Splits data into training and test sets in a cross-validation fashion based
on a user-provided categorical vector.
This vector can be passed during instantiation either via an arbitrary factor \code{f}
with the same length as \code{task$nrow}, or via a single string \code{col} referring to a
column in the task.

An alternative but equivalent approach using leave-one-out resampling is
showcased in the examples of \link{mlr_resamplings_loo}.
}
\section{Dictionary}{

This \link{Resampling} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_resamplings} or with the associated sugar function \code{\link[=rsmp]{rsmp()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_resamplings$get("custom_cv")
rsmp("custom_cv")
}\if{html}{\out{</div>}}
}

\examples{
# Create a task with 10 observations
task = tsk("penguins")
task$filter(1:10)

# Instantiate Resampling:
custom_cv = rsmp("custom_cv")
f = factor(c(rep(letters[1:3], each = 3), NA))
custom_cv$instantiate(task, f = f)
custom_cv$iters # 3 folds

# Individual sets:
custom_cv$train_set(1)
custom_cv$test_set(1)

# Disjunct sets:
intersect(custom_cv$train_set(1), custom_cv$test_set(1))
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter3/evaluation_and_benchmarking.html#sec-resampling}
\item Package \CRANpkg{mlr3spatiotempcv} for spatio-temporal resamplings.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Resampling]{Resamplings}: \link{mlr_resamplings}
\item \code{as.data.table(mlr_resamplings)} for a table of available \link[=Resampling]{Resamplings} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3spatiotempcv} for additional \link{Resampling}s for spatio-temporal
tasks.
}

Other Resampling:
\code{\link{Resampling}},
\code{\link{mlr_resamplings}},
\code{\link{mlr_resamplings_bootstrap}},
\code{\link{mlr_resamplings_custom}},
\code{\link{mlr_resamplings_cv}},
\code{\link{mlr_resamplings_holdout}},
\code{\link{mlr_resamplings_insample}},
\code{\link{mlr_resamplings_loo}},
\code{\link{mlr_resamplings_repeated_cv}},
\code{\link{mlr_resamplings_subsampling}}
}
\concept{Resampling}
\section{Super class}{
\code{\link[mlr3:Resampling]{Resampling}} -> \code{ResamplingCustomCV}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{iters}}{(\code{integer(1)})\cr
Returns the number of resampling iterations, depending on the values stored in the \code{param_set}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ResamplingCustomCV-initialize}{\code{ResamplingCustomCV$new()}}
    \item \href{#method-ResamplingCustomCV-instantiate}{\code{ResamplingCustomCV$instantiate()}}
    \item \href{#method-ResamplingCustomCV-clone}{\code{ResamplingCustomCV$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="format"><a href='../../mlr3/html/Resampling.html#method-Resampling-format'><code>Resampling$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="help"><a href='../../mlr3/html/Resampling.html#method-Resampling-help'><code>Resampling$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="print"><a href='../../mlr3/html/Resampling.html#method-Resampling-print'><code>Resampling$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="test_set"><a href='../../mlr3/html/Resampling.html#method-Resampling-test_set'><code>Resampling$test_set()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="train_set"><a href='../../mlr3/html/Resampling.html#method-Resampling-train_set'><code>Resampling$train_set()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResamplingCustomCV-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ResamplingCustomCV-initialize}{}}}
\subsection{\code{ResamplingCustomCV$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResamplingCustomCV$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResamplingCustomCV-instantiate"></a>}}
\if{latex}{\out{\hypertarget{method-ResamplingCustomCV-instantiate}{}}}
\subsection{\code{ResamplingCustomCV$instantiate()}}{
  Instantiate this \link{Resampling} as cross-validation with custom splits.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResamplingCustomCV$instantiate(task, f = NULL, col = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{\link{Task}\cr
Used to extract row ids.}
      \item{\code{f}}{(\code{factor()} | \code{character()})\cr
Vector of type factor or character with the same length as \code{task$nrow}.
Row ids are split on this vector, each distinct value results in a fold.
Empty factor levels are dropped and row ids corresponding to missing values are removed,
c.f. \code{\link[=split]{split()}}.}
      \item{\code{col}}{(\code{character(1)})\cr
Name of the task column to use for splitting.
Alternative and mutually exclusive to providing the factor levels as a vector via
parameter \code{f}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResamplingCustomCV-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ResamplingCustomCV-clone}{}}}
\subsection{\code{ResamplingCustomCV$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResamplingCustomCV$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskGeneratorSpirals.R
\name{mlr_task_generators_spirals}
\alias{mlr_task_generators_spirals}
\alias{TaskGeneratorSpirals}
\title{Spiral Classification Task Generator}
\description{
A \link{TaskGenerator} for the spirals task in \code{\link[mlbench:mlbench.spirals]{mlbench::mlbench.spirals()}}.
}
\section{Dictionary}{

This \link{TaskGenerator} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_task_generators} or with the associated sugar function \code{\link[=tgen]{tgen()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_task_generators$get("spirals")
tgen("spirals")
}\if{html}{\out{</div>}}
}

\section{Parameters}{
\tabular{llll}{
   Id \tab Type \tab Default \tab Range \cr
   cycles \tab integer \tab 1 \tab \eqn{[1, \infty)}{[1, Inf)} \cr
   sd \tab numeric \tab 0 \tab \eqn{[0, \infty)}{[0, Inf)} \cr
}
}

\examples{
generator = tgen("spirals")
plot(generator, n = 200)

task = generator$generate(200)
str(task$data())
}
\seealso{
\itemize{
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=TaskGenerator]{TaskGenerators}: \link{mlr_task_generators}
\item \code{as.data.table(mlr_task_generators)}  for a table of available \link[=TaskGenerator]{TaskGenerators} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other TaskGenerator:
\code{\link{TaskGenerator}},
\code{\link{mlr_task_generators}},
\code{\link{mlr_task_generators_2dnormals}},
\code{\link{mlr_task_generators_cassini}},
\code{\link{mlr_task_generators_circle}},
\code{\link{mlr_task_generators_friedman1}},
\code{\link{mlr_task_generators_moons}},
\code{\link{mlr_task_generators_peak}},
\code{\link{mlr_task_generators_simplex}},
\code{\link{mlr_task_generators_smiley}},
\code{\link{mlr_task_generators_xor}}
}
\concept{TaskGenerator}
\section{Super class}{
\code{\link[mlr3:TaskGenerator]{TaskGenerator}} -> \code{TaskGeneratorSpirals}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-TaskGeneratorSpirals-initialize}{\code{TaskGeneratorSpirals$new()}}
    \item \href{#method-TaskGeneratorSpirals-plot}{\code{TaskGeneratorSpirals$plot()}}
    \item \href{#method-TaskGeneratorSpirals-clone}{\code{TaskGeneratorSpirals$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="TaskGenerator" data-id="format"><a href='../../mlr3/html/TaskGenerator.html#method-TaskGenerator-format'><code>TaskGenerator$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="TaskGenerator" data-id="generate"><a href='../../mlr3/html/TaskGenerator.html#method-TaskGenerator-generate'><code>TaskGenerator$generate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="TaskGenerator" data-id="print"><a href='../../mlr3/html/TaskGenerator.html#method-TaskGenerator-print'><code>TaskGenerator$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGeneratorSpirals-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGeneratorSpirals-initialize}{}}}
\subsection{\code{TaskGeneratorSpirals$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGeneratorSpirals$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGeneratorSpirals-plot"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGeneratorSpirals-plot}{}}}
\subsection{\code{TaskGeneratorSpirals$plot()}}{
  Creates a simple plot of generated data.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGeneratorSpirals$plot(n = 200L, pch = 19L, ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{(\code{integer(1)})\cr
Number of samples to draw for the plot. Default is 200.}
      \item{\code{pch}}{(\code{integer(1)})\cr
Point char. Passed to \code{\link[=plot]{plot()}}.}
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=plot]{plot()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGeneratorSpirals-clone"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGeneratorSpirals-clone}{}}}
\subsection{\code{TaskGeneratorSpirals$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGeneratorSpirals$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataBackend.R
\name{DataBackend}
\alias{DataBackend}
\title{DataBackend}
\description{
This is the abstract base class for data backends.

Data backends provide a layer of abstraction for various data storage systems.
It is not recommended to work directly with the DataBackend.
Instead, all data access is handled transparently via the \link{Task}.

This package currently ships with one implementation for backends:
\itemize{
\item \link{DataBackendDataTable} which stores the data as \code{\link[data.table:data.table]{data.table::data.table()}}.
}

To connect to out-of-memory database management systems such as SQL servers,
see the extension package \CRANpkg{mlr3db}.
}
\details{
The required set of fields and methods to implement a custom \code{DataBackend} is
listed in the respective sections (see \link{DataBackendDataTable}).
}
\examples{
data = data.table::data.table(id = 1:5, x = runif(5),
  y = sample(letters[1:3], 5, replace = TRUE))

b = DataBackendDataTable$new(data, primary_key = "id")
print(b)
b$head(2)
b$data(rows = 1:2, cols = "x")
b$distinct(rows = b$rownames, "y")
b$missings(rows = b$rownames, cols = names(data))
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter10/advanced_technical_aspects_of_mlr3.html#sec-backends}
\item Package \CRANpkg{mlr3db} to interface out-of-memory data,
e.g. SQL servers or \CRANpkg{duckdb}.
}

Other DataBackend:
\code{\link{DataBackendDataTable}},
\code{\link[=as_data_backend]{as_data_backend()}}
}
\concept{DataBackend}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{primary_key}}{(\code{character(1)})\cr
Column name of the primary key column of positive and unique integer row ids.}
  }
  \if{html}{\out{</div>}}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{hash}}{(\code{character(1)})\cr
Hash (unique identifier) for this object.}

    \item{\code{col_hashes}}{(named \code{character})\cr
Hash (unique identifier) for all columns except the \code{primary_key}: A \code{character} vector, named by the columns that each element refers to.\cr
Columns of different \code{\link{Task}}s or \code{\link{DataBackend}}s that have agreeing \code{col_hashes} always represent the same data, given that the same \code{row}s are selected.
The reverse is not necessarily true: There can be columns with the same content that have different \code{col_hashes}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-DataBackend-initialize}{\code{DataBackend$new()}}
    \item \href{#method-DataBackend-format}{\code{DataBackend$format()}}
    \item \href{#method-DataBackend-print}{\code{DataBackend$print()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataBackend-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-DataBackend-initialize}{}}}
\subsection{\code{DataBackend$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.

Note: This object is typically constructed via a derived classes, e.g.
\link{DataBackendDataTable}, or via the S3 method
\code{\link[=as_data_backend]{as_data_backend()}}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{DataBackend$new(data, primary_key)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{data}}{(any)\cr
The format of the input data depends on the specialization. E.g.,
\link{DataBackendDataTable} expects a \code{\link[data.table:data.table]{data.table::data.table()}}.}
      \item{\code{primary_key}}{(\code{character(1)})\cr
Each DataBackend needs a way to address rows, which is done via a
column of unique integer values, referenced here by \code{primary_key}. The
use of this variable may differ between backends.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataBackend-format"></a>}}
\if{latex}{\out{\hypertarget{method-DataBackend-format}{}}}
\subsection{\code{DataBackend$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{DataBackend$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataBackend-print"></a>}}
\if{latex}{\out{\hypertarget{method-DataBackend-print}{}}}
\subsection{\code{DataBackend$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{DataBackend$print()}
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_prediction_regr.R
\name{as_prediction_regr}
\alias{as_prediction_regr}
\alias{as_prediction_regr.PredictionRegr}
\alias{as_prediction_regr.data.frame}
\title{Convert to a Regression Prediction}
\usage{
as_prediction_regr(x, ...)

\method{as_prediction_regr}{PredictionRegr}(x, ...)

\method{as_prediction_regr}{data.frame}(x, ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{...}{(any)\cr
Additional arguments.}
}
\value{
\link{PredictionRegr}.
}
\description{
Convert object to a \link{PredictionRegr}.
}
\examples{
# create a prediction object
task = tsk("mtcars")
learner = lrn("regr.rpart")
learner$train(task)
p = learner$predict(task)

# convert to a data.table
tab = as.data.table(p)

# convert back to a Prediction
as_prediction_regr(tab)

# split data.table into a list of data.tables
tabs = split(tab, cut(tab$truth, 3))

# convert back to list of predictions
preds = lapply(tabs, as_prediction_regr)

# calculate performance in each group
sapply(preds, function(p) p$score())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_callbacks.R
\name{mlr3.holdout_task}
\alias{mlr3.holdout_task}
\title{Callback Holdout Task}
\arguments{
\item{task}{(\link{Task})\cr
The holdout task.}
}
\description{
This \link{CallbackResample} predicts on an additional holdout task after training.
}
\examples{
task = tsk("pima")
task_holdout = task$clone()
learner = lrn("classif.rpart")
resampling = rsmp("cv", folds = 3)
splits = partition(task, 0.7)

task$filter(splits$train)
task_holdout$filter(splits$test)

callback = clbk("mlr3.holdout_task", task = task_holdout)

rr = resample(task, learner, resampling = resampling, callbacks = callback)

rr$data_extra
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LearnerClassifDebug.R
\name{mlr_learners_classif.debug}
\alias{mlr_learners_classif.debug}
\alias{LearnerClassifDebug}
\title{Classification Learner for Debugging}
\description{
A simple \link{LearnerClassif} used primarily in the unit tests and for debugging purposes.
If no hyperparameter is set, it simply constantly predicts a randomly selected label.
The following hyperparameters trigger the following actions:
\describe{
\item{error_predict:}{Probability to raise an exception during predict.}
\item{error_train:}{Probability to raises an exception during train.}
\item{message_predict:}{Probability to output a message during predict.}
\item{message_train:}{Probability to output a message during train.}
\item{predict_missing:}{Ratio of predictions which will be NA.}
\item{predict_missing_type:}{How to encode missingness.
\dQuote{na} will insert NA values,
\dQuote{omit} will just return fewer predictions than requested.}
\item{save_tasks:}{Saves input task in \code{model} slot during training and prediction.}
\item{segfault_predict:}{Probability to provokes a segfault during predict.}
\item{segfault_train:}{Probability to provokes a segfault during train.}
\item{sleep_train:}{Function returning a single number determining how many seconds
to sleep during \verb{$train()}.}
\item{sleep_predict:}{Function returning a single number determining how many seconds
to sleep during \verb{$predict()}.}
\item{threads:}{Number of threads to use. Has no effect.}
\item{warning_predict:}{Probability to signal a warning during predict.}
\item{warning_train:}{Probability to signal a warning during train.}
\item{x:}{Numeric tuning parameter. Has no effect.}
\item{iter:}{Integer parameter for testing hotstarting.}
\item{count_marshaling:}{If \code{TRUE}, \code{marshal_model} will increase the \code{marshal_count}
by 1 each time it is called. The default is \code{FALSE}.}
\item{check_pid:}{If \code{TRUE}, the \verb{$predict()} function will throw an error if the model
was not unmarshaled in the same session that is used for prediction.)}
}
Note that segfaults may not be triggered reliably on your operating system.
Also note that if they work as intended, they will tear down your R session immediately!
}
\section{Dictionary}{

This \link{Learner} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_learners} or with the associated sugar function \code{\link[=lrn]{lrn()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_learners$get("classif.debug")
lrn("classif.debug")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Predict Types: \dQuote{response}, \dQuote{prob}
\item Feature Types: \dQuote{logical}, \dQuote{integer}, \dQuote{numeric}, \dQuote{character}, \dQuote{factor}, \dQuote{ordered}
\item Required Packages: \CRANpkg{mlr3}
}
}

\section{Parameters}{
\tabular{lllll}{
   Id \tab Type \tab Default \tab Levels \tab Range \cr
   error_predict \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   error_train \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   message_predict \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   message_train \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   predict_missing \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   predict_missing_type \tab character \tab na \tab na, omit \tab - \cr
   save_tasks \tab logical \tab FALSE \tab TRUE, FALSE \tab - \cr
   segfault_predict \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   segfault_train \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   sleep_train \tab untyped \tab - \tab  \tab - \cr
   sleep_predict \tab untyped \tab - \tab  \tab - \cr
   threads \tab integer \tab - \tab  \tab \eqn{[1, \infty)}{[1, Inf)} \cr
   warning_predict \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   warning_train \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   x \tab numeric \tab - \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   iter \tab integer \tab 1 \tab  \tab \eqn{[1, \infty)}{[1, Inf)} \cr
   early_stopping \tab logical \tab FALSE \tab TRUE, FALSE \tab - \cr
   count_marshaling \tab logical \tab FALSE \tab TRUE, FALSE \tab - \cr
   check_pid \tab logical \tab TRUE \tab TRUE, FALSE \tab - \cr
   config_error \tab logical \tab FALSE \tab TRUE, FALSE \tab - \cr
}
}

\examples{
learner = lrn("classif.debug")
learner$param_set$set_values(message_train = 1, save_tasks = TRUE)

# this should signal a message
task = tsk("penguins")
learner$train(task)
learner$predict(task)

# task_train and task_predict are the input tasks for train() and predict()
names(learner$model)
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-learners}
\item Package \CRANpkg{mlr3learners} for a solid collection of essential learners.
\item Package \href{https://github.com/mlr-org/mlr3extralearners}{mlr3extralearners} for more learners.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Learner]{Learners}: \link{mlr_learners}
\item \code{as.data.table(mlr_learners)} for a table of available \link[=Learner]{Learners} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3pipelines} to combine learners with pre- and postprocessing steps.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
\item \CRANpkg{mlr3tuning} for tuning of hyperparameters, \CRANpkg{mlr3tuningspaces}
for established default tuning spaces.
}

Other Learner:
\code{\link{Learner}},
\code{\link{LearnerClassif}},
\code{\link{LearnerRegr}},
\code{\link{mlr_learners}},
\code{\link{mlr_learners_classif.featureless}},
\code{\link{mlr_learners_classif.rpart}},
\code{\link{mlr_learners_regr.debug}},
\code{\link{mlr_learners_regr.featureless}},
\code{\link{mlr_learners_regr.rpart}}
}
\concept{Learner}
\section{Super classes}{
\code{\link[mlr3:Learner]{Learner}} -> \code{\link[mlr3:LearnerClassif]{LearnerClassif}} -> \code{LearnerClassifDebug}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{marshaled}}{(\code{logical(1)})\cr
Whether the learner has been marshaled.}

    \item{\code{internal_valid_scores}}{Retrieves the internal validation scores as a named \code{list()}.
Returns \code{NULL} if learner is not trained yet.}

    \item{\code{internal_tuned_values}}{Retrieves the internally tuned values as a named \code{list()}.
Returns \code{NULL} if learner is not trained yet.}

    \item{\code{validate}}{How to construct the internal validation data. This parameter can be either \code{NULL},
a ratio in $(0, 1)$, \code{"test"}, or \code{"predefined"}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-LearnerClassifDebug-initialize}{\code{LearnerClassifDebug$new()}}
    \item \href{#method-LearnerClassifDebug-marshal}{\code{LearnerClassifDebug$marshal()}}
    \item \href{#method-LearnerClassifDebug-unmarshal}{\code{LearnerClassifDebug$unmarshal()}}
    \item \href{#method-LearnerClassifDebug-importance}{\code{LearnerClassifDebug$importance()}}
    \item \href{#method-LearnerClassifDebug-selected_features}{\code{LearnerClassifDebug$selected_features()}}
    \item \href{#method-LearnerClassifDebug-clone}{\code{LearnerClassifDebug$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="base_learner"><a href='../../mlr3/html/Learner.html#method-Learner-base_learner'><code>Learner$base_learner()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="configure"><a href='../../mlr3/html/Learner.html#method-Learner-configure'><code>Learner$configure()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="encapsulate"><a href='../../mlr3/html/Learner.html#method-Learner-encapsulate'><code>Learner$encapsulate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="format"><a href='../../mlr3/html/Learner.html#method-Learner-format'><code>Learner$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="help"><a href='../../mlr3/html/Learner.html#method-Learner-help'><code>Learner$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict"><a href='../../mlr3/html/Learner.html#method-Learner-predict'><code>Learner$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict_newdata"><a href='../../mlr3/html/Learner.html#method-Learner-predict_newdata'><code>Learner$predict_newdata()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="print"><a href='../../mlr3/html/Learner.html#method-Learner-print'><code>Learner$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="reset"><a href='../../mlr3/html/Learner.html#method-Learner-reset'><code>Learner$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="train"><a href='../../mlr3/html/Learner.html#method-Learner-train'><code>Learner$train()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="LearnerClassif" data-id="predict_newdata_fast"><a href='../../mlr3/html/LearnerClassif.html#method-LearnerClassif-predict_newdata_fast'><code>LearnerClassif$predict_newdata_fast()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifDebug-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifDebug-initialize}{}}}
\subsection{\code{LearnerClassifDebug$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifDebug$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifDebug-marshal"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifDebug-marshal}{}}}
\subsection{\code{LearnerClassifDebug$marshal()}}{
  Marshal the learner's model.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifDebug$marshal(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=marshal_model]{marshal_model()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifDebug-unmarshal"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifDebug-unmarshal}{}}}
\subsection{\code{LearnerClassifDebug$unmarshal()}}{
  Unmarshal the learner's model.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifDebug$unmarshal(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=unmarshal_model]{unmarshal_model()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifDebug-importance"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifDebug-importance}{}}}
\subsection{\code{LearnerClassifDebug$importance()}}{
  Returns 0 for each feature seen in training.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifDebug$importance()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Named \code{numeric()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifDebug-selected_features"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifDebug-selected_features}{}}}
\subsection{\code{LearnerClassifDebug$selected_features()}}{
  Always returns character(0).
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifDebug$selected_features()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{character()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifDebug-clone"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifDebug-clone}{}}}
\subsection{\code{LearnerClassifDebug$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifDebug$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSimple.R
\name{mlr_measures_regr.rmsle}
\alias{mlr_measures_regr.rmsle}
\title{Root Mean Squared Log Error}
\description{
Measure to compare true observed response with predicted response in regression tasks.
}
\details{
The Root Mean Squared Log Error is defined as \deqn{
  \sqrt{\frac{1}{n} \sum_{i=1}^n w_i \left( \ln (1 + t_i) - \ln (1 + r_i) \right)^2},
}{
  sqrt(weighted.mean((log(1 + t) - log(1 + r))^2, w)),
}
where \eqn{w_i} are normalized sample weights.

This measure is undefined if any element of \eqn{t} or \eqn{r} is less than or equal to \eqn{-1}.
}
\note{
The score function calls \code{\link[mlr3measures:rmsle]{mlr3measures::rmsle()}} from package \CRANpkg{mlr3measures}.

If the measure is undefined for the input, \code{NaN} is returned.
This can be customized by setting the field \code{na_value}.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("regr.rmsle")
msr("regr.rmsle")
}\if{html}{\out{</div>}}
}

\section{Parameters}{

Empty ParamSet
}

\section{Meta Information}{


\itemize{
\item Type: \code{"regr"}
\item Range: \eqn{[0, \infty)}{[0, Inf)}
\item Minimize: \code{TRUE}
\item Required prediction: \code{response}
}

}

\seealso{
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}

\code{as.data.table(mlr_measures)} for a complete table of all (also dynamically created) \link{Measure} implementations.

Other regression measures:
\code{\link{mlr_measures_regr.bias}},
\code{\link{mlr_measures_regr.ktau}},
\code{\link{mlr_measures_regr.mae}},
\code{\link{mlr_measures_regr.mape}},
\code{\link{mlr_measures_regr.maxae}},
\code{\link{mlr_measures_regr.medae}},
\code{\link{mlr_measures_regr.medse}},
\code{\link{mlr_measures_regr.mse}},
\code{\link{mlr_measures_regr.msle}},
\code{\link{mlr_measures_regr.pbias}},
\code{\link{mlr_measures_regr.rmse}},
\code{\link{mlr_measures_regr.sae}},
\code{\link{mlr_measures_regr.smape}},
\code{\link{mlr_measures_regr.srho}},
\code{\link{mlr_measures_regr.sse}}
}
\concept{regression measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/default_measures.R
\name{default_measures}
\alias{default_measures}
\title{Get the Default Measure}
\usage{
default_measures(task_type)
}
\arguments{
\item{task_type}{(\code{character(1)})\cr
Get the default measure for the task type \code{task_type}, e.g., \code{"classif"} or \code{"regr"}.
If \code{task_type} is \code{NULL}, an empty list is returned.}
}
\value{
list of \link{Measure}.
}
\description{
Gets the default measures using the information in \link[=mlr_reflections]{mlr_reflections$default_measures}:
\itemize{
\item \code{\link[=mlr_measures_classif.ce]{"classif.ce"}} for classification (\code{"classif"}).
\item \code{\link[=mlr_measures_regr.mse]{"regr.mse"}} for regression (\code{"regr"}).
\item Add-on package may register additional default measures for their own task types.
}
}
\examples{
default_measures("classif")
default_measures("regr")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_callbacks.R
\name{mlr3.model_extractor}
\alias{mlr3.model_extractor}
\title{Model Extractor Callback}
\arguments{
\item{fun}{(\verb{function(learner)})\cr
Function to extract information from the learner.
The function must have the argument \code{learner}.
The function must return a named list.}
}
\description{
This \link{CallbackResample} extracts information from the model after training with a user-defined function.
This way information can be extracted from the model without saving the model (\code{store_models = FALSE}).
The \code{fun} must be a function that takes a learner as input and returns the extracted information
as named list (see example).
The callback is very helpful to call \verb{$selected_features()}, \verb{$importance()}, \verb{$oob_error()} on the learner.
}
\examples{
task = tsk("pima")
learner = lrn("classif.rpart")
resampling = rsmp("cv", folds = 3)

# define function to extract selected features
selected_features = function(learner) list(selected_features = learner$selected_features())

# create callback
callback = clbk("mlr3.model_extractor", fun = selected_features)

rr = resample(task, learner, resampling = resampling, store_models = FALSE, callbacks = callback)

rr$data_extra
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install_pkgs.R
\name{install_pkgs}
\alias{install_pkgs}
\alias{extract_pkgs}
\alias{extract_pkgs.character}
\alias{extract_pkgs.R6}
\alias{extract_pkgs.list}
\alias{extract_pkgs.ResampleResult}
\alias{extract_pkgs.BenchmarkResult}
\title{Install (Missing) Packages}
\usage{
install_pkgs(x, ...)

extract_pkgs(x)

\method{extract_pkgs}{character}(x)

\method{extract_pkgs}{R6}(x)

\method{extract_pkgs}{list}(x)

\method{extract_pkgs}{ResampleResult}(x)

\method{extract_pkgs}{BenchmarkResult}(x)
}
\arguments{
\item{x}{(any)\cr
Object with package information (or a list of such objects).}

\item{...}{(any)\cr
Additional arguments passed down to \code{\link[remotes:install_cran]{remotes::install_cran()}} or
\code{\link[remotes:install_github]{remotes::install_github()}}.
Arguments \code{force} and \code{upgrade} are often important in this context.}
}
\value{
\code{extract_pkgs()} returns a \code{character()} of package strings,
\code{install_pkgs()} returns the names of extracted packages invisibly.
}
\description{
\code{extract_pkgs()} extracts required package from various objects, including
\link{TaskGenerator}, \link{Learner}, \link{Measure} and objects from
extension packages such as \CRANpkg{mlr3pipelines} or \CRANpkg{mlr3filters}.
If applied on a list, the function is called recursively on all elements.

\code{install_pkgs()} calls \code{extract_pkgs()} internally and proceeds with the
installation of extracted packages.
}
\details{
If a package contains a forward slash ('/'), it is assumed to be a package hosted
on GitHub in \code{"<user>/<repo>"} format, and the string will be passed to
\code{\link[remotes:install_github]{remotes::install_github()}}.
Otherwise, the package name will be passed to \code{\link[remotes:install_cran]{remotes::install_cran()}}.
}
\examples{
extract_pkgs(lrns(c("regr.rpart", "regr.featureless")))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HotstartStack.R
\name{HotstartStack}
\alias{HotstartStack}
\title{Stack for Hot Start Learners}
\description{
This class stores learners for hot starting training, i.e. resuming or
continuing from an already fitted model.
We assume that hot starting is only possible if a single hyperparameter
(also called the fidelity parameter, usually controlling the complexity or
expensiveness) is altered and all other hyperparameters are identical.

The \code{HotstartStack} stores trained learners which can be potentially used to
hot start a learner. Learner automatically hot start while training if a
stack is attached to the \verb{$hotstart_stack} field and the stack contains a
suitable learner.

For example, if you want to train a random forest learner with 1000 trees but
already have a random forest learner with 500 trees (hot start learner),
you can add the hot start learner to the \code{HotstartStack} of the expensive learner
with 1000 trees. If you now call the \code{train()} method (or \code{\link[=resample]{resample()}} or
\code{\link[=benchmark]{benchmark()}}), a random forest with 500 trees will be fitted and combined
with the 500 trees of the hotstart learner, effectively saving you to
fit 500 trees.

Hot starting is only supported by learners which have the property
\code{"hotstart_forward"} or \code{"hotstart_backward"}. For example, an \code{xgboost} model
(in \CRANpkg{mlr3learners}) can hot start forward by adding more boosting
iterations, and a random forest can go backwards by removing trees.
The fidelity parameters are tagged with \code{"hotstart"} in learner's parameter set.
}
\examples{
# train learner on pima task
task = tsk("pima")
learner = lrn("classif.debug", iter = 1)
learner$train(task)

# initialize stack with previously fitted learner
hot = HotstartStack$new(list(learner))

# retrieve learner with increased fidelity parameter
learner = lrn("classif.debug", iter = 2)

# calculate cost of hot starting
hot$start_cost(learner, task$hash)

# add stack with hot start learner
learner$hotstart_stack = hot

# train automatically uses hot start learner while fitting the model
learner$train(task)
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{stack}}{\code{\link[data.table:data.table]{data.table::data.table()}}\cr
Stores hot start learners.}

    \item{\code{hotstart_threshold}}{(named \code{numeric(1)})\cr
Threshold for storing learners in the stack.
If the value of the hotstart parameter is below this threshold, the learner is not added to the stack.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-HotstartStack-initialize}{\code{HotstartStack$new()}}
    \item \href{#method-HotstartStack-add}{\code{HotstartStack$add()}}
    \item \href{#method-HotstartStack-start_cost}{\code{HotstartStack$start_cost()}}
    \item \href{#method-HotstartStack-format}{\code{HotstartStack$format()}}
    \item \href{#method-HotstartStack-print}{\code{HotstartStack$print()}}
    \item \href{#method-HotstartStack-clone}{\code{HotstartStack$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HotstartStack-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-HotstartStack-initialize}{}}}
\subsection{\code{HotstartStack$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{HotstartStack$new(learners = NULL, hotstart_threshold = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{learners}}{(List of \link{Learner}s)\cr
Learners are added to the hotstart stack. If \code{NULL} (default), empty
stack is created.}
      \item{\code{hotstart_threshold}}{(named \code{numeric(1)})\cr
Threshold for storing learners in the stack.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HotstartStack-add"></a>}}
\if{latex}{\out{\hypertarget{method-HotstartStack-add}{}}}
\subsection{\code{HotstartStack$add()}}{
  Add learners to hot start stack.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{HotstartStack$add(learners)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{learners}}{(List of \link{Learner}s).
Learners are added to the hotstart stack.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    self (invisibly).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HotstartStack-start_cost"></a>}}
\if{latex}{\out{\hypertarget{method-HotstartStack-start_cost}{}}}
\subsection{\code{HotstartStack$start_cost()}}{
  Calculates the cost for each learner of the stack to hot start the target \code{learner}.

The following cost values can be returned:
\itemize{
\item \code{NA_real_}: Learner is unsuitable to hot start target \code{learner}.
\item \code{-1}: Hotstart learner in the stack and target \code{learner} are identical.
\item \code{0} Cost for hot starting backwards is always 0.
\item \verb{> 0} Cost for hot starting forward.
}
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{HotstartStack$start_cost(learner, task_hash)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{learner}}{\link{Learner}\cr
Target learner.}
      \item{\code{task_hash}}{\link{Task}\cr
Hash of the task on which the target learner is trained.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HotstartStack-format"></a>}}
\if{latex}{\out{\hypertarget{method-HotstartStack-format}{}}}
\subsection{\code{HotstartStack$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{HotstartStack$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HotstartStack-print"></a>}}
\if{latex}{\out{\hypertarget{method-HotstartStack-print}{}}}
\subsection{\code{HotstartStack$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{HotstartStack$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HotstartStack-clone"></a>}}
\if{latex}{\out{\hypertarget{method-HotstartStack-clone}{}}}
\subsection{\code{HotstartStack$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{HotstartStack$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSimple.R
\name{mlr_measures_classif.tp}
\alias{mlr_measures_classif.tp}
\title{True Positives}
\description{
Measure to compare true observed labels with predicted
labels
in binary classification tasks.
}
\details{
This measure counts the true positives, i.e. the number of
predictions correctly indicating a positive class label. This is sometimes also called a "hit".
}
\note{
The score function calls \code{\link[mlr3measures:tp]{mlr3measures::tp()}} from package \CRANpkg{mlr3measures}.

If the measure is undefined for the input, \code{NaN} is returned.
This can be customized by setting the field \code{na_value}.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("classif.tp")
msr("classif.tp")
}\if{html}{\out{</div>}}
}

\section{Parameters}{

Empty ParamSet
}

\section{Meta Information}{


\itemize{
\item Type: \code{"binary"}
\item Range: \eqn{[0, \infty)}{[0, Inf)}
\item Minimize: \code{FALSE}
\item Required prediction: \code{response}
}

}

\seealso{
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}

\code{as.data.table(mlr_measures)} for a complete table of all (also dynamically created) \link{Measure} implementations.

Other classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.auc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.bbrier}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.dor}},
\code{\link{mlr_measures_classif.fbeta}},
\code{\link{mlr_measures_classif.fdr}},
\code{\link{mlr_measures_classif.fn}},
\code{\link{mlr_measures_classif.fnr}},
\code{\link{mlr_measures_classif.fomr}},
\code{\link{mlr_measures_classif.fp}},
\code{\link{mlr_measures_classif.fpr}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}},
\code{\link{mlr_measures_classif.mcc}},
\code{\link{mlr_measures_classif.npv}},
\code{\link{mlr_measures_classif.ppv}},
\code{\link{mlr_measures_classif.prauc}},
\code{\link{mlr_measures_classif.precision}},
\code{\link{mlr_measures_classif.recall}},
\code{\link{mlr_measures_classif.sensitivity}},
\code{\link{mlr_measures_classif.specificity}},
\code{\link{mlr_measures_classif.tn}},
\code{\link{mlr_measures_classif.tnr}},
\code{\link{mlr_measures_classif.tpr}}

Other binary classification measures:
\code{\link{mlr_measures_classif.auc}},
\code{\link{mlr_measures_classif.bbrier}},
\code{\link{mlr_measures_classif.dor}},
\code{\link{mlr_measures_classif.fbeta}},
\code{\link{mlr_measures_classif.fdr}},
\code{\link{mlr_measures_classif.fn}},
\code{\link{mlr_measures_classif.fnr}},
\code{\link{mlr_measures_classif.fomr}},
\code{\link{mlr_measures_classif.fp}},
\code{\link{mlr_measures_classif.fpr}},
\code{\link{mlr_measures_classif.npv}},
\code{\link{mlr_measures_classif.ppv}},
\code{\link{mlr_measures_classif.prauc}},
\code{\link{mlr_measures_classif.precision}},
\code{\link{mlr_measures_classif.recall}},
\code{\link{mlr_measures_classif.sensitivity}},
\code{\link{mlr_measures_classif.specificity}},
\code{\link{mlr_measures_classif.tn}},
\code{\link{mlr_measures_classif.tnr}},
\code{\link{mlr_measures_classif.tpr}}
}
\concept{binary classification measures}
\concept{classification measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSimple.R
\name{mlr_measures_classif.logloss}
\alias{mlr_measures_classif.logloss}
\title{Log Loss}
\description{
Measure to compare true observed labels with predicted
probabilities
in multiclass classification tasks.
}
\details{
The Log Loss (a.k.a Bernoulli Loss, Logistic Loss, Cross-Entropy Loss) is defined as
\deqn{
  -\frac{1}{n} \sum_{i=1}^n w_i \log \left(  p_i \right )
}{
  -weighted.mean(log(p), w)
}
where \eqn{p_i}{p} is the probability for the true class of observation \eqn{i},
and \eqn{w_i} are normalized weights for each observation \eqn{x_i}.
}
\note{
The score function calls \code{\link[mlr3measures:logloss]{mlr3measures::logloss()}} from package \CRANpkg{mlr3measures}.

If the measure is undefined for the input, \code{NaN} is returned.
This can be customized by setting the field \code{na_value}.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("classif.logloss")
msr("classif.logloss")
}\if{html}{\out{</div>}}
}

\section{Parameters}{

Empty ParamSet
}

\section{Meta Information}{


\itemize{
\item Type: \code{"classif"}
\item Range: \eqn{[0, \infty)}{[0, Inf)}
\item Minimize: \code{TRUE}
\item Required prediction: \code{prob}
}

}

\seealso{
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}

\code{as.data.table(mlr_measures)} for a complete table of all (also dynamically created) \link{Measure} implementations.

Other classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.auc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.bbrier}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.dor}},
\code{\link{mlr_measures_classif.fbeta}},
\code{\link{mlr_measures_classif.fdr}},
\code{\link{mlr_measures_classif.fn}},
\code{\link{mlr_measures_classif.fnr}},
\code{\link{mlr_measures_classif.fomr}},
\code{\link{mlr_measures_classif.fp}},
\code{\link{mlr_measures_classif.fpr}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}},
\code{\link{mlr_measures_classif.mcc}},
\code{\link{mlr_measures_classif.npv}},
\code{\link{mlr_measures_classif.ppv}},
\code{\link{mlr_measures_classif.prauc}},
\code{\link{mlr_measures_classif.precision}},
\code{\link{mlr_measures_classif.recall}},
\code{\link{mlr_measures_classif.sensitivity}},
\code{\link{mlr_measures_classif.specificity}},
\code{\link{mlr_measures_classif.tn}},
\code{\link{mlr_measures_classif.tnr}},
\code{\link{mlr_measures_classif.tp}},
\code{\link{mlr_measures_classif.tpr}}

Other multiclass classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}},
\code{\link{mlr_measures_classif.mcc}}
}
\concept{classification measures}
\concept{multiclass classification measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CallbackResample.R
\name{callback_resample}
\alias{callback_resample}
\title{Create Evaluation Callback}
\usage{
callback_resample(
  id,
  label = NA_character_,
  man = NA_character_,
  on_resample_begin = NULL,
  on_resample_before_train = NULL,
  on_resample_before_predict = NULL,
  on_resample_end = NULL
)
}
\arguments{
\item{id}{(\code{character(1)})\cr
Identifier for the new instance.}

\item{label}{(\code{character(1)})\cr
Label for the new instance.}

\item{man}{(\code{character(1)})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
The referenced help package can be opened via method \verb{$help()}.}

\item{on_resample_begin}{(\verb{function()})\cr
Stage called at the beginning of an evaluation.
Called in \code{workhorse()} (internal).}

\item{on_resample_before_train}{(\verb{function()})\cr
Stage called before training the learner.
Called in \code{workhorse()} (internal).}

\item{on_resample_before_predict}{(\verb{function()})\cr
Stage called before predicting.
Called in \code{workhorse()} (internal).}

\item{on_resample_end}{(\verb{function()})\cr
Stage called at the end of an evaluation.
Called in \code{workhorse()} (internal).}
}
\description{
Function to create a \link{CallbackResample}.
Predefined callbacks are stored in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_callbacks}
and can be retrieved with \code{\link[=clbk]{clbk()}}.

Evaluation callbacks are called at different stages of the resampling process.
Each stage is called once per resampling iteration.
The stages are prefixed with \verb{on_resample_*}.
The text in brackets indicates what happens between the stages in the internal \code{workhorse()} function
and which accesses to the \link{ContextResample} (\code{ctx}) are typical for the stage.

\if{html}{\out{<div class="sourceCode">}}\preformatted{Start Resampling Iteration on Worker
 - on_resample_begin
   (Split `ctx$task` into training and test set with `ctx$resampling` and `ctx$iteration`)
 - on_resample_before_train
   (Train the learner `ctx$learner` on training data)
 - on_resample_before_predict
   (Predict on predict sets and store prediction data `ctx$pdatas`)
 - on_resample_end
   (Erase model `ctx$learner$model` if requested and return results)
End Resampling Iteration on Worker
}\if{html}{\out{</div>}}

The callback can store data in \code{ctx$learner$state} or \code{ctx$data_extra}.
The data in \code{ctx$data_extra} is stored in the \link{ResampleResult} or \link{BenchmarkResult}.
See also the section on parameters for more information on the stages.
}
\details{
When implementing a callback, each function must have two arguments named \code{callback} and \code{context}.
A callback can write data to the state (\verb{$state}), e.g. settings that affect the callback itself.
We highly discourage changing the task, learner and resampling objects via the callback.
}
\examples{
learner = lrn("classif.rpart")
task = tsk("pima")
resampling = rsmp("cv", folds = 3)

# save selected features callback
callback = callback_resample("selected_features",
 on_resample_end = function(callback, context) {
    context$learner$state$selected_features = context$learner$selected_features()
  }
)

rr = resample(task, learner, resampling, callbacks = callback)
rr$learners[[1]]$state$selected_features

# holdout task callback
callback = callback_resample("holdout_task",
  on_resample_before_predict = function(callback, context) {
    pred = context$learner$predict(callback$state$task)
    context$data_extra = list(prediction_holdout = pred)
  }
)

task_holdout = tsk("pima")
splits = partition(task, 0.7)
task$filter(splits$train)
task_holdout$filter(splits$test)

callback$state$task = task_holdout

rr = resample(task, learner, resampling, callbacks = callback)
rr$data_extra
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resample.R
\name{resample}
\alias{resample}
\title{Resample a Learner on a Task}
\usage{
resample(
  task,
  learner,
  resampling,
  store_models = FALSE,
  store_backends = TRUE,
  encapsulate = NA_character_,
  allow_hotstart = FALSE,
  clone = c("task", "learner", "resampling"),
  unmarshal = TRUE,
  callbacks = NULL
)
}
\arguments{
\item{task}{(\link{Task}).}

\item{learner}{(\link{Learner}).}

\item{resampling}{(\link{Resampling}).}

\item{store_models}{(\code{logical(1)})\cr
Store the fitted model in the resulting object=
Set to \code{TRUE} if you want to further analyse the models or want to
extract information like variable importance.}

\item{store_backends}{(\code{logical(1)})\cr
Keep the \link{DataBackend} of the \link{Task} in the \link{ResampleResult}?
Set to \code{TRUE} if your performance measures require a \link{Task},
or to analyse results more conveniently.
Set to \code{FALSE} to reduce the file size and memory footprint
after serialization.
The current default is \code{TRUE}, but this eventually will be changed
in a future release.}

\item{encapsulate}{(\code{character(1)})\cr
If not \code{NA}, enables encapsulation by setting the field
\code{Learner$encapsulate} to one of the supported values:
\code{"none"} (disable encapsulation),
\code{"try"} (captures errors but output is printed to the console and not logged),
\code{"evaluate"} (execute via \CRANpkg{evaluate}) and
\code{"callr"} (start in external session via \CRANpkg{callr}).
If \code{NA}, encapsulation is not changed, i.e. the settings of the
individual learner are active.
Additionally, if encapsulation is set to \code{"evaluate"} or \code{"callr"},
the fallback learner is set to the featureless learner if the learner
does not already have a fallback configured.}

\item{allow_hotstart}{(\code{logical(1)})\cr
Determines if learner(s) are hot started with trained models in
\verb{$hotstart_stack}. See also \link{HotstartStack}.}

\item{clone}{(\code{character()})\cr
Select the input objects to be cloned before proceeding by
providing a set with possible values \code{"task"}, \code{"learner"} and
\code{"resampling"} for \link{Task}, \link{Learner} and \link{Resampling}, respectively.
Per default, all input objects are cloned.}

\item{unmarshal}{\code{\link{Learner}}\cr
Whether to unmarshal learners that were marshaled during the execution.
If \code{TRUE} all models are stored in unmarshaled form.
If \code{FALSE}, all learners (that need marshaling) are stored in marshaled form.}

\item{callbacks}{(list of \link[mlr3misc:Callback]{mlr3misc::Callback} | \code{NULL})\cr
Callbacks to be executed during the resampling process.
See \link{CallbackResample} and \link{ContextResample} for details.}
}
\value{
\link{ResampleResult}.
}
\description{
Runs a resampling (possibly in parallel):
Repeatedly apply \link{Learner} \code{learner} on a training set of \link{Task} \code{task} to train a model,
then use the trained model to predict observations of a test set.
Training and test sets are defined by the \link{Resampling} \code{resampling}.
}
\note{
The fitted models are discarded after the predictions have been computed in order to reduce memory consumption.
If you need access to the models for later analysis, set \code{store_models} to \code{TRUE}.
}
\section{Stochasticity}{

Note that uninstantiated \code{\link{Resampling}}s are instantiated on the task, making
the procedure stochastic even in case of a deterministic learner.
}

\section{Predict Sets}{

If you want to compare the performance of a learner on the training with the performance
on the test set, you have to configure the \link{Learner} to predict on multiple sets by
setting the field \code{predict_sets} to \code{c("train", "test")} (default is \code{"test"}).
Each set yields a separate \link{Prediction} object during resampling.
In the next step, you have to configure the measures to operate on the respective Prediction object:

\if{html}{\out{<div class="sourceCode">}}\preformatted{m1 = msr("classif.ce", id = "ce.train", predict_sets = "train")
m2 = msr("classif.ce", id = "ce.test", predict_sets = "test")
}\if{html}{\out{</div>}}

The (list of) created measures can finally be passed to \verb{$aggregate()} or \verb{$score()}.
}

\section{Parallelization}{


This function can be parallelized with the \CRANpkg{future} or \CRANpkg{mirai} package.
One job is one resampling iteration.
All jobs are send to an apply function from \CRANpkg{future.apply} or \code{mirai::mirai_map()} in a single batch.
To select a parallel backend, use \code{\link[future:plan]{future::plan()}}.
To use \code{mirai}, call \code{mirai::daemons(.compute = "mlr3_parallelization")} before calling this function.
The \code{future} package guarantees reproducible results independent of the parallel backend.
The results of \code{mirai} will not be the same but can be made reproducible by setting a \code{seed} when calling \code{mirai::daemons()}.
More on parallelization can be found in the book:
\url{https://mlr3book.mlr-org.com/chapters/chapter10/advanced_technical_aspects_of_mlr3.html}
}

\section{Progress Bars}{

This function supports progress bars via the package \CRANpkg{progressr}.
Simply wrap the function call in \code{\link[progressr:with_progress]{progressr::with_progress()}} to enable them.
Alternatively, call \code{\link[progressr:handlers]{progressr::handlers()}} with \code{global = TRUE} to enable progress bars
globally.
We recommend the \CRANpkg{progress} package as backend which can be enabled with
\code{progressr::handlers("progress")}.
}

\section{Logging}{


The \CRANpkg{mlr3} uses the \CRANpkg{lgr} package for logging.
\CRANpkg{lgr} supports multiple log levels which can be queried with
\code{getOption("lgr.log_levels")}.

To suppress output and reduce verbosity, you can lower the log from the
default level \code{"info"} to \code{"warn"}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{lgr::get_logger("mlr3")$set_threshold("warn")
}\if{html}{\out{</div>}}

To get additional log output for debugging, increase the log level to \code{"debug"}
or \code{"trace"}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{lgr::get_logger("mlr3")$set_threshold("debug")
}\if{html}{\out{</div>}}

To obtain a more informative traceback when an iteration fails, set
\code{options(mlr3.debug = TRUE)}. This disables parallelization via \CRANpkg{future}
so that errors are raised in the main process.
Note that results computed in debug mode use a different seeding mechanism and are \strong{not reproducible}.

To log to a file or a data base, see the documentation of \link[lgr:lgr-package]{lgr::lgr-package}.
}

\examples{
task = tsk("penguins")
learner = lrn("classif.rpart")
resampling = rsmp("cv")

# Explicitly instantiate the resampling for this task for reproducibility
set.seed(123)
resampling$instantiate(task)

rr = resample(task, learner, resampling)
print(rr)

# Retrieve performance
rr$score(msr("classif.ce"))
rr$aggregate(msr("classif.ce"))

# merged prediction objects of all resampling iterations
pred = rr$prediction()
pred$confusion

# Repeat resampling with featureless learner
rr_featureless = resample(task, lrn("classif.featureless"), resampling)

# Convert results to BenchmarkResult, then combine them
bmr1 = as_benchmark_result(rr)
bmr2 = as_benchmark_result(rr_featureless)
print(bmr1$combine(bmr2))
}
\seealso{
\itemize{
\item \code{\link[=as_benchmark_result]{as_benchmark_result()}} to convert to a \link{BenchmarkResult}.
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter3/evaluation_and_benchmarking.html#sec-resampling}
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
}

Other resample:
\code{\link{ResampleResult}}
}
\concept{resample}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_prediction_data.R
\name{as_prediction_data}
\alias{as_prediction_data}
\alias{as_prediction_data.Prediction}
\alias{as_prediction_data.PredictionData}
\alias{as_prediction_data.list}
\title{PredictionData}
\usage{
as_prediction_data(x, task, row_ids = task$row_ids, check = TRUE, ...)

\method{as_prediction_data}{Prediction}(x, task, row_ids = task$row_ids, check = TRUE, ...)

\method{as_prediction_data}{PredictionData}(x, task, row_ids = task$row_ids, check = TRUE, ...)

\method{as_prediction_data}{list}(
  x,
  task,
  row_ids = task$row_ids,
  check = TRUE,
  ...,
  train_task
)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{task}{(\link{Task}).}

\item{row_ids}{\code{integer()}\cr
Row indices.}

\item{check}{(\code{logical(1)})\cr
Perform argument checks and type conversions?}

\item{...}{(any)\cr
Additional arguments.}

\item{train_task}{(\link{Task})\cr
Task used for training the learner.}
}
\value{
\link{PredictionData}.
}
\description{
Convert object to a \link{PredictionData} or a list of \link{PredictionData}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSimple.R
\name{mlr_measures_classif.auc}
\alias{mlr_measures_classif.auc}
\title{Area Under the ROC Curve}
\description{
Measure to compare true observed labels with predicted
probabilities
in binary classification tasks.
}
\details{
Computes the area under the Receiver Operator Characteristic (ROC) curve.
The AUC can be interpreted as the probability that a randomly chosen positive observation
has a higher predicted probability than a randomly chosen negative observation.

For \eqn{n^+} positive and \eqn{n^-} negative observations with \eqn{R_i^+} the rank of the
\eqn{i}-th positive observation's predicted probability (average ranks for ties), the AUC
is estimated as \deqn{
  \widehat{\operatorname{AUC}} = \frac{1}{n^+ n^-} \left( \sum_{i=1}^{n^+} R_i^+ \; - \; \frac{n^+(n^+ + 1)}{2} \right).
}{
  AUC = (sum(R_pos) - n_pos * (n_pos + 1) / 2) / (n_pos * n_neg).
}

If \code{sample_weights} are provided, let \eqn{w_i^+} be the weight of the \eqn{i}-th positive observation with
predicted probability \eqn{p_i^+}, \eqn{W^+ = \sum_i w_i^+}, and \eqn{W^-} the total weight of the
negative observations. Define the weighted Mann-Whitney contribution of positive observation \eqn{i} as
\eqn{U_i^w = W_{< p_i^+}^- + \tfrac{1}{2} W_{= p_i^+}^-},
i.e. the total weight of negative observations with a smaller predicted probability plus half the weight
of negatives tied with \eqn{p_i^+}. The weighted AUC is then \deqn{
  \widehat{\operatorname{AUC}}_w = \frac{1}{W^+ W^-} \sum_{i=1}^{n^+} w_i^+ U_i^w.
}{
  AUC_w = sum(w_pos * U_pos) / (W_pos * W_neg).
}

This measure is undefined if the true values are either all positive or
all negative.
}
\note{
The score function calls \code{\link[mlr3measures:auc]{mlr3measures::auc()}} from package \CRANpkg{mlr3measures}.

If the measure is undefined for the input, \code{NaN} is returned.
This can be customized by setting the field \code{na_value}.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("classif.auc")
msr("classif.auc")
}\if{html}{\out{</div>}}
}

\section{Parameters}{

Empty ParamSet
}

\section{Meta Information}{


\itemize{
\item Type: \code{"binary"}
\item Range: \eqn{[0, 1]}{[0, 1]}
\item Minimize: \code{FALSE}
\item Required prediction: \code{prob}
}

}

\seealso{
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}

\code{as.data.table(mlr_measures)} for a complete table of all (also dynamically created) \link{Measure} implementations.

Other classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.bbrier}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.dor}},
\code{\link{mlr_measures_classif.fbeta}},
\code{\link{mlr_measures_classif.fdr}},
\code{\link{mlr_measures_classif.fn}},
\code{\link{mlr_measures_classif.fnr}},
\code{\link{mlr_measures_classif.fomr}},
\code{\link{mlr_measures_classif.fp}},
\code{\link{mlr_measures_classif.fpr}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}},
\code{\link{mlr_measures_classif.mcc}},
\code{\link{mlr_measures_classif.npv}},
\code{\link{mlr_measures_classif.ppv}},
\code{\link{mlr_measures_classif.prauc}},
\code{\link{mlr_measures_classif.precision}},
\code{\link{mlr_measures_classif.recall}},
\code{\link{mlr_measures_classif.sensitivity}},
\code{\link{mlr_measures_classif.specificity}},
\code{\link{mlr_measures_classif.tn}},
\code{\link{mlr_measures_classif.tnr}},
\code{\link{mlr_measures_classif.tp}},
\code{\link{mlr_measures_classif.tpr}}

Other binary classification measures:
\code{\link{mlr_measures_classif.bbrier}},
\code{\link{mlr_measures_classif.dor}},
\code{\link{mlr_measures_classif.fbeta}},
\code{\link{mlr_measures_classif.fdr}},
\code{\link{mlr_measures_classif.fn}},
\code{\link{mlr_measures_classif.fnr}},
\code{\link{mlr_measures_classif.fomr}},
\code{\link{mlr_measures_classif.fp}},
\code{\link{mlr_measures_classif.fpr}},
\code{\link{mlr_measures_classif.npv}},
\code{\link{mlr_measures_classif.ppv}},
\code{\link{mlr_measures_classif.prauc}},
\code{\link{mlr_measures_classif.precision}},
\code{\link{mlr_measures_classif.recall}},
\code{\link{mlr_measures_classif.sensitivity}},
\code{\link{mlr_measures_classif.specificity}},
\code{\link{mlr_measures_classif.tn}},
\code{\link{mlr_measures_classif.tnr}},
\code{\link{mlr_measures_classif.tp}},
\code{\link{mlr_measures_classif.tpr}}
}
\concept{binary classification measures}
\concept{classification measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_tasks.R
\name{mlr_tasks}
\alias{mlr_tasks}
\title{Dictionary of Tasks}
\format{
\link[R6:R6Class]{R6::R6Class} object inheriting from \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}
\description{
A simple \link[mlr3misc:Dictionary]{mlr3misc::Dictionary} storing objects of class \link{Task}.
Each task has an associated help page, see \code{mlr_tasks_[id]}.

This dictionary can get populated with additional tasks by add-on packages,
e.g. \CRANpkg{mlr3data}, \CRANpkg{mlr3proba} or \CRANpkg{mlr3cluster}.
\CRANpkg{mlr3oml} allows to interact with \href{https://www.openml.org}{OpenML}.

For a more convenient way to retrieve and construct tasks, see \code{\link[=tsk]{tsk()}}/\code{\link[=tsks]{tsks()}}.
}
\section{Methods}{

See \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}

\section{S3 methods}{

\itemize{
\item \code{as.data.table(dict, ..., objects = FALSE)}\cr
\link[mlr3misc:Dictionary]{mlr3misc::Dictionary} -> \code{\link[data.table:data.table]{data.table::data.table()}}\cr
Returns a \code{\link[data.table:data.table]{data.table::data.table()}} with columns "key", "label", "task_type", "nrow", "ncol", "properties",
and the number of features of type "lgl", "int", "dbl", "chr", "fct" and "ord", respectively.
If \code{objects} is set to \code{TRUE}, the constructed objects are returned in the list column named \code{object}.
}
}

\examples{
as.data.table(mlr_tasks)
task = mlr_tasks$get("penguins") # same as tsk("penguins")
head(task$data())

# Add a new task, based on a subset of penguins:
data = palmerpenguins::penguins
data$species = factor(ifelse(data$species == "Adelie", "1", "0"))
task = TaskClassif$new("penguins.binary", data, target = "species", positive = "1")

# add to dictionary
mlr_tasks$add("penguins.binary", task)

# list available tasks
mlr_tasks$keys()

# retrieve from dictionary
mlr_tasks$get("penguins.binary")

# remove task again
mlr_tasks$remove("penguins.binary")
}
\seealso{
Sugar functions: \code{\link[=tsk]{tsk()}}, \code{\link[=tsks]{tsks()}}

Extension Packages: \CRANpkg{mlr3data}

Other Dictionary:
\code{\link{mlr_learners}},
\code{\link{mlr_measures}},
\code{\link{mlr_resamplings}},
\code{\link{mlr_task_generators}}

Other Task:
\code{\link{Task}},
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Dictionary}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_task_regr.R
\name{as_task_regr}
\alias{as_task_regr}
\alias{as_task_regr.TaskRegr}
\alias{as_task_regr.data.frame}
\alias{as_task_regr.matrix}
\alias{as_task_regr.DataBackend}
\alias{as_task_regr.TaskClassif}
\alias{as_task_regr.formula}
\title{Convert to a Regression Task}
\usage{
as_task_regr(x, ...)

\method{as_task_regr}{TaskRegr}(x, clone = FALSE, ...)

\method{as_task_regr}{data.frame}(
  x,
  target,
  id = deparse1(substitute(x)),
  label = NA_character_,
  ...
)

\method{as_task_regr}{matrix}(
  x,
  target,
  id = deparse1(substitute(x)),
  label = NA_character_,
  ...
)

\method{as_task_regr}{DataBackend}(
  x,
  target,
  id = deparse1(substitute(x)),
  label = NA_character_,
  ...
)

\method{as_task_regr}{TaskClassif}(x, target, drop_original_target = FALSE, drop_levels = TRUE, ...)

\method{as_task_regr}{formula}(
  x,
  data,
  id = deparse1(substitute(data)),
  label = NA_character_,
  ...
)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{...}{(any)\cr
Additional arguments.}

\item{clone}{(\code{logical(1)})\cr
If \code{TRUE}, ensures that the returned object is not the same as the input \code{x}.}

\item{target}{(\code{character(1)})\cr
Name of the target column.}

\item{id}{(\code{character(1)})\cr
Id for the new task.
Defaults to the (deparsed and substituted) name of the data argument.}

\item{label}{(\code{character(1)})\cr
Label for the new instance.}

\item{drop_original_target}{(\code{logical(1)})\cr
If \code{FALSE} (default), the original target is added as a feature.
Otherwise the original target is dropped.}

\item{drop_levels}{(\code{logical(1)})\cr
If \code{TRUE} (default), unused levels of the new target variable are dropped.}

\item{data}{(\code{data.frame()})\cr
Data frame containing all columns referenced in formula \code{x}.}
}
\value{
\link{TaskRegr}.
}
\description{
Convert object to a \link{TaskRegr}.
This is a S3 generic. mlr3 ships with methods for the following objects:
\enumerate{
\item \link{TaskRegr}: returns the object as-is, possibly cloned.
\item \code{\link{formula}}, \code{\link[=data.frame]{data.frame()}}, \code{\link[=matrix]{matrix()}}, and \link{DataBackend}: provides an alternative to the
constructor of \link{TaskRegr}.
\item \link{TaskClassif}: Calls \code{\link[=convert_task]{convert_task()}}.
}
}
\examples{
as_task_regr(datasets::mtcars, target = "mpg")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PredictionClassif.R
\name{PredictionClassif}
\alias{PredictionClassif}
\title{Prediction Object for Classification}
\description{
This object wraps the predictions returned by a learner of class \link{LearnerClassif}, i.e.
the predicted response and class probabilities.

If the response is not provided during construction, but class probabilities are,
the response is calculated from the probabilities: the class label with the highest
probability is chosen. In case of ties, a label is selected randomly.
}
\note{
If this object is constructed manually, make sure that the factor levels for \code{truth}
have the same levels as the task, in the same order.
In case of binary classification tasks, the positive class label must be the first level.
}
\section{Thresholding}{

If probabilities are stored, it is possible to change the threshold which determines the predicted class label.
Usually, the label of the class with the highest predicted probability is selected.
For binary classification problems, such an threshold defaults to 0.5.
For cost-sensitive or imbalanced classification problems, manually adjusting the threshold can increase
the predictive performance.
\itemize{
\item For binary problems only a single threshold value can be set.
If the probability exceeds the threshold, the positive class is predicted.
If the probability equals the threshold, the label is selected randomly.
\item For binary and multi-class problems, a named numeric vector of thresholds can be set.
The length and names must correspond to the number of classes and class names, respectively.
To determine the class label, the probabilities are divided by the threshold.
This results in a ratio > 1 if the probability exceeds the threshold, and a ratio < 1 otherwise.
Note that it is possible that either none or multiple ratios are greater than 1 at the same time.
Anyway, the class label with maximum ratio is selected.
In case of ties in the ratio, one of the tied class labels is selected randomly.

Note that there are the following edge cases for threshold equal to \code{0} which are handled specially:
\enumerate{
\item With threshold 0 the resulting ratio gets \code{Inf} and thus gets always selected.
If there are multiple ratios with value \code{Inf}, one is selected according to \code{ties_method}
(randomly per default).
\item If additionally the predicted probability is also 0, the ratio \code{0/0} results in \code{NaN} values.
These are simply replaced by \code{0} and thus will never get selected.
}
}
}

\examples{
task = tsk("penguins")
learner = lrn("classif.rpart", predict_type = "prob")
learner$train(task)
p = learner$predict(task)
p$predict_types
head(as.data.table(p))

# confusion matrix
p$confusion

# change threshold
th = c(0.05, 0.9, 0.05)
names(th) = task$class_names

# new predictions
p$set_threshold(th)$response
p$score(measures = msr("classif.ce"))
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Prediction:
\code{\link{Prediction}},
\code{\link{PredictionRegr}}
}
\concept{Prediction}
\section{Super class}{
\code{\link[mlr3:Prediction]{Prediction}} -> \code{PredictionClassif}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{response}}{(\code{factor()})\cr
Access to the stored predicted class labels.}

    \item{\code{prob}}{(\code{matrix()})\cr
Access to the stored probabilities.}

    \item{\code{confusion}}{(\code{matrix()})\cr
Confusion matrix, as resulting from the comparison of truth and response.
Truth is in columns, predicted response is in rows.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PredictionClassif-initialize}{\code{PredictionClassif$new()}}
    \item \href{#method-PredictionClassif-set_threshold}{\code{PredictionClassif$set_threshold()}}
    \item \href{#method-PredictionClassif-clone}{\code{PredictionClassif$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="filter"><a href='../../mlr3/html/Prediction.html#method-Prediction-filter'><code>Prediction$filter()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="format"><a href='../../mlr3/html/Prediction.html#method-Prediction-format'><code>Prediction$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="help"><a href='../../mlr3/html/Prediction.html#method-Prediction-help'><code>Prediction$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="obs_loss"><a href='../../mlr3/html/Prediction.html#method-Prediction-obs_loss'><code>Prediction$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="print"><a href='../../mlr3/html/Prediction.html#method-Prediction-print'><code>Prediction$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="score"><a href='../../mlr3/html/Prediction.html#method-Prediction-score'><code>Prediction$score()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PredictionClassif-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PredictionClassif-initialize}{}}}
\subsection{\code{PredictionClassif$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PredictionClassif$new(
  task = NULL,
  row_ids = task$row_ids,
  truth = task$truth(),
  response = NULL,
  prob = NULL,
  weights = NULL,
  check = TRUE,
  extra = NULL,
  raw = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link{TaskClassif})\cr
Task, used to extract defaults for \code{row_ids} and \code{truth}.}
      \item{\code{row_ids}}{(\code{integer()})\cr
Row ids of the predicted observations, i.e. the row ids of the test set.}
      \item{\code{truth}}{(\code{factor()})\cr
True (observed) labels. See the note on manual construction.}
      \item{\code{response}}{(\code{character()} | \code{factor()})\cr
Vector of predicted class labels.
One element for each observation in the test set.
Character vectors are automatically converted to factors.
See the note on manual construction.}
      \item{\code{prob}}{(\code{matrix()})\cr
Numeric matrix of posterior class probabilities with one column for each class
and one row for each observation in the test set.
Columns must be named with class labels, row names are automatically removed.
If \code{prob} is provided, but \code{response} is not, the class labels are calculated from
the probabilities using \code{\link[=max.col]{max.col()}} with \code{ties.method} set to \code{"random"}.}
      \item{\code{weights}}{(\code{numeric()})\cr
Vector of measure weights for each observation. Should be constructed from
the \code{Task}'s \code{weights_measure} column.}
      \item{\code{check}}{(\code{logical(1)})\cr
If \code{TRUE}, performs some argument checks and predict type conversions.}
      \item{\code{extra}}{(\code{list()})\cr
List of extra data to be stored in the prediction object.}
      \item{\code{raw}}{(any)\cr
Raw prediction object from the upstream model. Stored as-is without validation.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PredictionClassif-set_threshold"></a>}}
\if{latex}{\out{\hypertarget{method-PredictionClassif-set_threshold}{}}}
\subsection{\code{PredictionClassif$set_threshold()}}{
  Sets the prediction response based on the provided threshold.
See the section on thresholding for more information.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PredictionClassif$set_threshold(threshold, ties_method = "random")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{threshold}}{(\code{numeric()}).}
      \item{\code{ties_method}}{(\code{character(1)})\cr
One of \code{"random"}, \code{"first"} or \code{"last"} (c.f. \code{\link[=max.col]{max.col()}}) to determine how to deal with
tied probabilities.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PredictionClassif-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PredictionClassif-clone}{}}}
\subsection{\code{PredictionClassif$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PredictionClassif$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSimple.R
\name{mlr_measures_classif.mbrier}
\alias{mlr_measures_classif.mbrier}
\title{Multiclass Brier Score}
\description{
Measure to compare true observed labels with predicted
probabilities
in multiclass classification tasks.
}
\details{
Brier score for multi-class classification problems with \eqn{k} labels defined as \deqn{
   \frac{1}{n} \sum_{i=1}^n \sum_{j=1}^k (I_{ij} - p_{ij})^2.
}{
}
\eqn{I_{ij}}{I_ij} is 1 if observation \eqn{x_i} has true label \eqn{j}, and 0 otherwise.
\eqn{p_{ij}}{p_ij} is the probability that observation \eqn{x_i} belongs to class \eqn{j}.

Note that there also is the more common definition of the Brier score for binary
classification problems in \code{\link[mlr3measures:bbrier]{bbrier()}}.
}
\note{
The score function calls \code{\link[mlr3measures:mbrier]{mlr3measures::mbrier()}} from package \CRANpkg{mlr3measures}.

If the measure is undefined for the input, \code{NaN} is returned.
This can be customized by setting the field \code{na_value}.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("classif.mbrier")
msr("classif.mbrier")
}\if{html}{\out{</div>}}
}

\section{Parameters}{

Empty ParamSet
}

\section{Meta Information}{


\itemize{
\item Type: \code{"classif"}
\item Range: \eqn{[0, 2]}{[0, 2]}
\item Minimize: \code{TRUE}
\item Required prediction: \code{prob}
}

}

\seealso{
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}

\code{as.data.table(mlr_measures)} for a complete table of all (also dynamically created) \link{Measure} implementations.

Other classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.auc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.bbrier}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.dor}},
\code{\link{mlr_measures_classif.fbeta}},
\code{\link{mlr_measures_classif.fdr}},
\code{\link{mlr_measures_classif.fn}},
\code{\link{mlr_measures_classif.fnr}},
\code{\link{mlr_measures_classif.fomr}},
\code{\link{mlr_measures_classif.fp}},
\code{\link{mlr_measures_classif.fpr}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mcc}},
\code{\link{mlr_measures_classif.npv}},
\code{\link{mlr_measures_classif.ppv}},
\code{\link{mlr_measures_classif.prauc}},
\code{\link{mlr_measures_classif.precision}},
\code{\link{mlr_measures_classif.recall}},
\code{\link{mlr_measures_classif.sensitivity}},
\code{\link{mlr_measures_classif.specificity}},
\code{\link{mlr_measures_classif.tn}},
\code{\link{mlr_measures_classif.tnr}},
\code{\link{mlr_measures_classif.tp}},
\code{\link{mlr_measures_classif.tpr}}

Other multiclass classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mcc}}
}
\concept{classification measures}
\concept{multiclass classification measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskGeneratorSimplex.R
\name{mlr_task_generators_simplex}
\alias{mlr_task_generators_simplex}
\alias{TaskGeneratorSimplex}
\title{Simplex Classification Task Generator}
\description{
A \link{TaskGenerator} for the simplex task in \code{\link[mlbench:mlbench.simplex]{mlbench::mlbench.simplex()}}.

Note that the generator implemented in \CRANpkg{mlbench} returns
fewer samples than requested.
}
\section{Dictionary}{

This \link{TaskGenerator} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_task_generators} or with the associated sugar function \code{\link[=tgen]{tgen()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_task_generators$get("simplex")
tgen("simplex")
}\if{html}{\out{</div>}}
}

\section{Parameters}{
\tabular{lllll}{
   Id \tab Type \tab Default \tab Levels \tab Range \cr
   center \tab logical \tab TRUE \tab TRUE, FALSE \tab - \cr
   d \tab integer \tab 3 \tab  \tab \eqn{[1, \infty)}{[1, Inf)} \cr
   sd \tab numeric \tab 0.1 \tab  \tab \eqn{[0, \infty)}{[0, Inf)} \cr
   sides \tab integer \tab 1 \tab  \tab \eqn{[1, \infty)}{[1, Inf)} \cr
}
}

\examples{
generator = tgen("simplex")
plot(generator, n = 200)

task = generator$generate(200)
str(task$data())
}
\seealso{
\itemize{
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=TaskGenerator]{TaskGenerators}: \link{mlr_task_generators}
\item \code{as.data.table(mlr_task_generators)}  for a table of available \link[=TaskGenerator]{TaskGenerators} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other TaskGenerator:
\code{\link{TaskGenerator}},
\code{\link{mlr_task_generators}},
\code{\link{mlr_task_generators_2dnormals}},
\code{\link{mlr_task_generators_cassini}},
\code{\link{mlr_task_generators_circle}},
\code{\link{mlr_task_generators_friedman1}},
\code{\link{mlr_task_generators_moons}},
\code{\link{mlr_task_generators_peak}},
\code{\link{mlr_task_generators_smiley}},
\code{\link{mlr_task_generators_spirals}},
\code{\link{mlr_task_generators_xor}}
}
\concept{TaskGenerator}
\section{Super class}{
\code{\link[mlr3:TaskGenerator]{TaskGenerator}} -> \code{TaskGeneratorSimplex}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-TaskGeneratorSimplex-initialize}{\code{TaskGeneratorSimplex$new()}}
    \item \href{#method-TaskGeneratorSimplex-plot}{\code{TaskGeneratorSimplex$plot()}}
    \item \href{#method-TaskGeneratorSimplex-clone}{\code{TaskGeneratorSimplex$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="TaskGenerator" data-id="format"><a href='../../mlr3/html/TaskGenerator.html#method-TaskGenerator-format'><code>TaskGenerator$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="TaskGenerator" data-id="generate"><a href='../../mlr3/html/TaskGenerator.html#method-TaskGenerator-generate'><code>TaskGenerator$generate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="TaskGenerator" data-id="print"><a href='../../mlr3/html/TaskGenerator.html#method-TaskGenerator-print'><code>TaskGenerator$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGeneratorSimplex-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGeneratorSimplex-initialize}{}}}
\subsection{\code{TaskGeneratorSimplex$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGeneratorSimplex$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGeneratorSimplex-plot"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGeneratorSimplex-plot}{}}}
\subsection{\code{TaskGeneratorSimplex$plot()}}{
  Creates a simple plot of generated data.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGeneratorSimplex$plot(n = 200L, pch = 19L, ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{(\code{integer(1)})\cr
Number of samples to draw for the plot. Default is 200.}
      \item{\code{pch}}{(\code{integer(1)})\cr
Point char. Passed to \code{\link[=plot]{plot()}}.}
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=plot]{plot()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGeneratorSimplex-clone"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGeneratorSimplex-clone}{}}}
\subsection{\code{TaskGeneratorSimplex$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGeneratorSimplex$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ResamplingSubsampling.R
\name{mlr_resamplings_subsampling}
\alias{mlr_resamplings_subsampling}
\alias{ResamplingSubsampling}
\title{Subsampling Resampling}
\description{
Splits data \code{repeats} (default: 30) times into training and test set
with a ratio of \code{ratio} (default: 2/3) observations going into the training set.
}
\section{Dictionary}{

This \link{Resampling} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_resamplings} or with the associated sugar function \code{\link[=rsmp]{rsmp()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_resamplings$get("subsampling")
rsmp("subsampling")
}\if{html}{\out{</div>}}
}

\section{Parameters}{

\itemize{
\item \code{repeats} (\code{integer(1)})\cr
Number of repetitions.
\item \code{ratio} (\code{numeric(1)})\cr
Ratio of observations to put into the training set.
}
}

\examples{
# Create a task with 10 observations
task = tsk("penguins")
task$filter(1:10)

# Instantiate Resampling
subsampling = rsmp("subsampling", repeats = 2, ratio = 0.5)
subsampling$instantiate(task)

# Individual sets:
subsampling$train_set(1)
subsampling$test_set(1)

# Disjunct sets:
intersect(subsampling$train_set(1), subsampling$test_set(1))

# Internal storage:
subsampling$instance$train # list of index vectors
}
\references{
Bischl B, Mersmann O, Trautmann H, Weihs C (2012).
\dQuote{Resampling Methods for Meta-Model Validation with Recommendations for Evolutionary Computation.}
\emph{Evolutionary Computation}, \bold{20}(2), 249--275.
\doi{10.1162/evco_a_00069}.
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter3/evaluation_and_benchmarking.html#sec-resampling}
\item Package \CRANpkg{mlr3spatiotempcv} for spatio-temporal resamplings.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Resampling]{Resamplings}: \link{mlr_resamplings}
\item \code{as.data.table(mlr_resamplings)} for a table of available \link[=Resampling]{Resamplings} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3spatiotempcv} for additional \link{Resampling}s for spatio-temporal
tasks.
}

Other Resampling:
\code{\link{Resampling}},
\code{\link{mlr_resamplings}},
\code{\link{mlr_resamplings_bootstrap}},
\code{\link{mlr_resamplings_custom}},
\code{\link{mlr_resamplings_custom_cv}},
\code{\link{mlr_resamplings_cv}},
\code{\link{mlr_resamplings_holdout}},
\code{\link{mlr_resamplings_insample}},
\code{\link{mlr_resamplings_loo}},
\code{\link{mlr_resamplings_repeated_cv}}
}
\concept{Resampling}
\section{Super class}{
\code{\link[mlr3:Resampling]{Resampling}} -> \code{ResamplingSubsampling}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{iters}}{(\code{integer(1)})\cr
Returns the number of resampling iterations, depending on the values stored in the \code{param_set}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ResamplingSubsampling-initialize}{\code{ResamplingSubsampling$new()}}
    \item \href{#method-ResamplingSubsampling-clone}{\code{ResamplingSubsampling$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="format"><a href='../../mlr3/html/Resampling.html#method-Resampling-format'><code>Resampling$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="help"><a href='../../mlr3/html/Resampling.html#method-Resampling-help'><code>Resampling$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="instantiate"><a href='../../mlr3/html/Resampling.html#method-Resampling-instantiate'><code>Resampling$instantiate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="print"><a href='../../mlr3/html/Resampling.html#method-Resampling-print'><code>Resampling$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="test_set"><a href='../../mlr3/html/Resampling.html#method-Resampling-test_set'><code>Resampling$test_set()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Resampling" data-id="train_set"><a href='../../mlr3/html/Resampling.html#method-Resampling-train_set'><code>Resampling$train_set()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResamplingSubsampling-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ResamplingSubsampling-initialize}{}}}
\subsection{\code{ResamplingSubsampling$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResamplingSubsampling$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResamplingSubsampling-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ResamplingSubsampling-clone}{}}}
\subsection{\code{ResamplingSubsampling$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResamplingSubsampling$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSimple.R
\name{mlr_measures_sim.phi}
\alias{mlr_measures_sim.phi}
\title{Phi Coefficient Similarity}
\description{
Measure to compare two or more sets w.r.t. their similarity.
}
\details{
The Phi Coefficient is defined as the Pearson correlation between the binary
representation of two sets \eqn{A} and \eqn{B}.
The binary representation for \eqn{A} is a logical vector of
length \eqn{p} with the i-th element being 1 if the corresponding
element is in \eqn{A}, and 0 otherwise.

If more than two sets are provided, the mean of all pairwise scores
is calculated.

This measure is undefined if one set contains none or all possible elements.
}
\note{
This measure requires learners with property \code{"selected_features"}.
The extracted feature sets are passed to \code{\link[mlr3measures:phi]{mlr3measures::phi()}} from
package \CRANpkg{mlr3measures}.

If the measure is undefined for the input, \code{NaN} is returned.
This can be customized by setting the field \code{na_value}.
}
\section{Parameters}{
\tabular{llll}{
   Id \tab Type \tab Default \tab Range \cr
   p \tab integer \tab - \tab \eqn{[1, \infty)}{[1, Inf)} \cr
}
}

\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("sim.phi")
msr("sim.phi")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{


\itemize{
\item Type: \code{"similarity"}
\item Range: \eqn{[-1, 1]}{[-1, 1]}
\item Minimize: \code{FALSE}
}

}

\seealso{
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}

\code{as.data.table(mlr_measures)} for a complete table of all (also dynamically created) \link{Measure} implementations.

Other similarity measures:
\code{\link{mlr_measures_sim.jaccard}}
}
\concept{similarity measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSelectedFeatures.R
\name{mlr_measures_selected_features}
\alias{mlr_measures_selected_features}
\alias{MeasureSelectedFeatures}
\title{Selected Features Measure}
\description{
Measures the number of selected features by extracting it from learners with property \code{"selected_features"}.
If parameter \code{normalize} is set to \code{TRUE}, the relative number of features instead of the absolute
number of features is returned.
Note that the models must be stored to be able to extract this information.
If the learner does not support the extraction of used features, \code{NA} is returned.

This measure requires the \link{Task} and the \link{Learner} for scoring.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("selected_features")
msr("selected_features")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{NA}
\item Range: \eqn{[0, \infty)}{[0, Inf)}
\item Minimize: TRUE
\item Average: macro
\item Required Prediction: \dQuote{NA}
\item Required Packages: \CRANpkg{mlr3}
}
}

\section{Parameters}{
\tabular{llll}{
   Id \tab Type \tab Default \tab Levels \cr
   normalize \tab logical \tab - \tab TRUE, FALSE \cr
}
}

\examples{
task = tsk("german_credit")
learner = lrn("classif.rpart")
rr = resample(task, learner, rsmp("cv", folds = 3), store_models = TRUE)

scores = rr$score(msr("selected_features"))
scores[, c("iteration", "selected_features")]
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-eval}
\item Package \CRANpkg{mlr3measures} for the scoring functions.
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}
\code{as.data.table(mlr_measures)} for a table of available \link[=Measure]{Measures} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Measure:
\code{\link{Measure}},
\code{\link{MeasureClassif}},
\code{\link{MeasureRegr}},
\code{\link{MeasureSimilarity}},
\code{\link{mlr_measures}},
\code{\link{mlr_measures_aic}},
\code{\link{mlr_measures_bic}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_debug_classif}},
\code{\link{mlr_measures_elapsed_time}},
\code{\link{mlr_measures_internal_valid_score}},
\code{\link{mlr_measures_oob_error}},
\code{\link{mlr_measures_regr.pinball}},
\code{\link{mlr_measures_regr.rqr}},
\code{\link{mlr_measures_regr.rsq}}
}
\concept{Measure}
\section{Super class}{
\code{\link[mlr3:Measure]{Measure}} -> \code{MeasureSelectedFeatures}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MeasureSelectedFeatures-initialize}{\code{MeasureSelectedFeatures$new()}}
    \item \href{#method-MeasureSelectedFeatures-clone}{\code{MeasureSelectedFeatures$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="aggregate"><a href='../../mlr3/html/Measure.html#method-Measure-aggregate'><code>Measure$aggregate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="format"><a href='../../mlr3/html/Measure.html#method-Measure-format'><code>Measure$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="help"><a href='../../mlr3/html/Measure.html#method-Measure-help'><code>Measure$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="obs_loss"><a href='../../mlr3/html/Measure.html#method-Measure-obs_loss'><code>Measure$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="print"><a href='../../mlr3/html/Measure.html#method-Measure-print'><code>Measure$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="score"><a href='../../mlr3/html/Measure.html#method-Measure-score'><code>Measure$score()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureSelectedFeatures-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureSelectedFeatures-initialize}{}}}
\subsection{\code{MeasureSelectedFeatures$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureSelectedFeatures$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureSelectedFeatures-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureSelectedFeatures-clone}{}}}
\subsection{\code{MeasureSelectedFeatures$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureSelectedFeatures$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LearnerRegrDebug.R
\name{mlr_learners_regr.debug}
\alias{mlr_learners_regr.debug}
\alias{LearnerRegrDebug}
\title{Regression Learner for Debugging}
\description{
A simple \link{LearnerRegr} used primarily in the unit tests and for debugging purposes.
If no hyperparameter is set, it simply constantly predicts the mean value of the training data.
The following hyperparameters trigger the following actions:
\describe{
\item{predict_missing:}{Ratio of predictions which will be NA.}
\item{predict_missing_type:}{How to encode missingness.
\dQuote{na} will insert NA values,
\dQuote{omit} will just return fewer predictions than requested.}
\item{save_tasks:}{Saves input task in \code{model} slot during training and prediction.}
\item{threads:}{Number of threads to use. Has no effect.}
\item{x:}{Numeric tuning parameter. Has no effect.}
}
}
\section{Dictionary}{

This \link{Learner} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_learners} or with the associated sugar function \code{\link[=lrn]{lrn()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_learners$get("regr.debug")
lrn("regr.debug")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{regr}
\item Predict Types: \dQuote{response}, \dQuote{se}, \dQuote{quantiles}
\item Feature Types: \dQuote{logical}, \dQuote{integer}, \dQuote{numeric}, \dQuote{character}, \dQuote{factor}, \dQuote{ordered}
\item Required Packages: \CRANpkg{mlr3}, 'stats'
}
}

\section{Parameters}{
\tabular{lllll}{
   Id \tab Type \tab Default \tab Levels \tab Range \cr
   error_predict \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   error_train \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   message_predict \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   message_train \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   predict_missing \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   predict_missing_type \tab character \tab na \tab na, omit \tab - \cr
   save_tasks \tab logical \tab FALSE \tab TRUE, FALSE \tab - \cr
   segfault_predict \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   segfault_train \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   threads \tab integer \tab - \tab  \tab \eqn{[1, \infty)}{[1, Inf)} \cr
   warning_predict \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   warning_train \tab numeric \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   x \tab numeric \tab - \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
}
}

\examples{
task = tsk("mtcars")
learner = lrn("regr.debug", save_tasks = TRUE)
learner$train(task, row_ids = 1:20)
prediction = learner$predict(task, row_ids = 21:32)

learner$model$task_train
learner$model$task_predict
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-learners}
\item Package \CRANpkg{mlr3learners} for a solid collection of essential learners.
\item Package \href{https://github.com/mlr-org/mlr3extralearners}{mlr3extralearners} for more learners.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Learner]{Learners}: \link{mlr_learners}
\item \code{as.data.table(mlr_learners)} for a table of available \link[=Learner]{Learners} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3pipelines} to combine learners with pre- and postprocessing steps.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
\item \CRANpkg{mlr3tuning} for tuning of hyperparameters, \CRANpkg{mlr3tuningspaces}
for established default tuning spaces.
}

Other Learner:
\code{\link{Learner}},
\code{\link{LearnerClassif}},
\code{\link{LearnerRegr}},
\code{\link{mlr_learners}},
\code{\link{mlr_learners_classif.debug}},
\code{\link{mlr_learners_classif.featureless}},
\code{\link{mlr_learners_classif.rpart}},
\code{\link{mlr_learners_regr.featureless}},
\code{\link{mlr_learners_regr.rpart}}
}
\concept{Learner}
\section{Super classes}{
\code{\link[mlr3:Learner]{Learner}} -> \code{\link[mlr3:LearnerRegr]{LearnerRegr}} -> \code{LearnerRegrDebug}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-LearnerRegrDebug-initialize}{\code{LearnerRegrDebug$new()}}
    \item \href{#method-LearnerRegrDebug-importance}{\code{LearnerRegrDebug$importance()}}
    \item \href{#method-LearnerRegrDebug-selected_features}{\code{LearnerRegrDebug$selected_features()}}
    \item \href{#method-LearnerRegrDebug-clone}{\code{LearnerRegrDebug$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="base_learner"><a href='../../mlr3/html/Learner.html#method-Learner-base_learner'><code>Learner$base_learner()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="configure"><a href='../../mlr3/html/Learner.html#method-Learner-configure'><code>Learner$configure()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="encapsulate"><a href='../../mlr3/html/Learner.html#method-Learner-encapsulate'><code>Learner$encapsulate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="format"><a href='../../mlr3/html/Learner.html#method-Learner-format'><code>Learner$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="help"><a href='../../mlr3/html/Learner.html#method-Learner-help'><code>Learner$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict"><a href='../../mlr3/html/Learner.html#method-Learner-predict'><code>Learner$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict_newdata"><a href='../../mlr3/html/Learner.html#method-Learner-predict_newdata'><code>Learner$predict_newdata()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="print"><a href='../../mlr3/html/Learner.html#method-Learner-print'><code>Learner$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="reset"><a href='../../mlr3/html/Learner.html#method-Learner-reset'><code>Learner$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="train"><a href='../../mlr3/html/Learner.html#method-Learner-train'><code>Learner$train()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="LearnerRegr" data-id="predict_newdata_fast"><a href='../../mlr3/html/LearnerRegr.html#method-LearnerRegr-predict_newdata_fast'><code>LearnerRegr$predict_newdata_fast()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerRegrDebug-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerRegrDebug-initialize}{}}}
\subsection{\code{LearnerRegrDebug$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerRegrDebug$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerRegrDebug-importance"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerRegrDebug-importance}{}}}
\subsection{\code{LearnerRegrDebug$importance()}}{
  Returns 0 for each feature seen in training.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerRegrDebug$importance()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Named \code{numeric()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerRegrDebug-selected_features"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerRegrDebug-selected_features}{}}}
\subsection{\code{LearnerRegrDebug$selected_features()}}{
  Always returns character(0).
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerRegrDebug$selected_features()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{character()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerRegrDebug-clone"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerRegrDebug-clone}{}}}
\subsection{\code{LearnerRegrDebug$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerRegrDebug$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Learner.R
\name{Learner}
\alias{Learner}
\title{Learner Class}
\description{
This is the abstract base class for learner objects like \link{LearnerClassif} and \link{LearnerRegr}.

Learners are built around the three following key parts:
\itemize{
\item Methods \verb{$train()} and \verb{$predict()} which call internal methods or private methods \verb{$.train()}/\verb{$.predict()}).
\item A \link[paradox:ParamSet]{paradox::ParamSet} which stores meta-information about available hyperparameters,
and also stores hyperparameter settings.
\item Meta-information about the requirements and capabilities of the learner.
\item The fitted model stored in field \verb{$model}, available after calling \verb{$train()}.
}

Predefined learners are stored in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_learners},
e.g. \code{\link[=mlr_learners_classif.rpart]{classif.rpart}} or \code{\link[=mlr_learners_regr.rpart]{regr.rpart}}.

More classification and regression learners are implemented in the add-on package \CRANpkg{mlr3learners}.
Learners for survival analysis (or more general, for probabilistic regression) can be found in \CRANpkg{mlr3proba}.
Unsupervised cluster algorithms are implemented in \CRANpkg{mlr3cluster}.
The dictionary \link{mlr_learners} gets automatically populated with the new learners
as soon as the respective packages are loaded.

More (experimental) learners can be found in the GitHub repository:
\url{https://github.com/mlr-org/mlr3extralearners}.
A guide on how to extend \CRANpkg{mlr3} with custom learners can be found in the
\href{https://mlr3book.mlr-org.com}{mlr3book}.

To combine the learner with preprocessing operations like factor encoding, \CRANpkg{mlr3pipelines} is recommended.
Hyperparameters stored in the \code{param_set} can be tuned with \CRANpkg{mlr3tuning}.
}
\section{Optional Extractors}{


Specific learner implementations are free to implement additional getters to ease the access of certain parts
of the model in the inherited subclasses.

For the following operations, extractors are standardized:
\itemize{
\item \code{importance(...)}: Returns the feature importance score as numeric vector.
The higher the score, the more important the variable.
The returned vector is named with feature names and sorted in decreasing order.
Note that the model might omit features it has not used at all.
The learner must be tagged with property \code{"importance"}.
To filter variables using the importance scores, see package \CRANpkg{mlr3filters}.
\item \code{selected_features(...)}: Returns a subset of selected features as \code{character()}.
The learner must be tagged with property \code{"selected_features"}.
\item \code{oob_error(...)}: Returns the out-of-bag error of the model as \code{numeric(1)}.
The learner must be tagged with property \code{"oob_error"}.
\item \code{internal_valid_scores}: Returns the internal validation score(s) of the model as a named \code{list()}.
Only available for \code{\link{Learner}}s with the \code{"validation"} property.
If the learner is not trained yet, this returns \code{NULL}.
\item \code{internal_tuned_values}: Returns the internally tuned hyperparameters of the model as a named \code{list()}.
Only available for \code{\link{Learner}}s with the \code{"internal_tuning"} property.
If the learner is not trained yet, this returns \code{NULL}.
}
}

\section{Weights}{


Many learners support observation weights, indicated by their property \code{"weights"}.
The weights are stored in the \link{Task} where the column role \code{weights_learner} needs to be assigned
to a single numeric column.
If a task has weights and the learner supports them, they are used automatically.
If a task has weights but the learner does not support them, an error is thrown by default.
Both of these behaviors can be disabled by setting the \code{use_weights} field to \code{"ignore"}.
See the description of \code{use_weights} for more information.

If the learner is set-up to use weights but the task does not have a designated weight column,
samples are considered to have equal weight.
When weights are being used, they are passed down to the learner directly;
the effect of weights depends on the specific learner.
Generally, weights do not need to sum up to 1.

When implementing a Learner that uses weights, the \code{"weights"} property should be set.
The \verb{$.train()}-method should then call the \verb{$.get_weights()}-method to retrieve the weights from the task.
\verb{$.get_weights()} will automatically discard weights when \code{use_weights} is set to \code{"ignore"};
}

\section{Setting Hyperparameters}{


All information about hyperparameters is stored in the slot \code{param_set} which is a \link[paradox:ParamSet]{paradox::ParamSet}.
The printer gives an overview about the ids of available hyperparameters, their storage type, lower and upper bounds,
possible levels (for factors), default values and assigned values.
To set hyperparameters, call the \code{set_values()} method on the \code{param_set}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{lrn = lrn("classif.rpart")
lrn$param_set$set_values(minsplit = 3, cp = 0.01)
}\if{html}{\out{</div>}}

Note that this operation replaces all previously set hyperparameter values.
If you only intend to change one specific hyperparameter value and leave the others as-is,
you can use the helper function \code{\link[mlr3misc:insert_named]{mlr3misc::insert_named()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{lrn$param_set$values = mlr3misc::insert_named(lrn$param_set$values, list(cp = 0.001))
}\if{html}{\out{</div>}}

If the learner has additional hyperparameters which are not encoded in the
\link[paradox:ParamSet]{ParamSet}, you can easily extend the learner.
Here, we add a factor hyperparameter with id \code{"foo"} and possible levels \code{"a"} and \code{"b"}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{lrn$param_set$add(paradox::ParamFct$new("foo", levels = c("a", "b")))
}\if{html}{\out{</div>}}
}

\section{Implementing Validation}{

Some Learners, such as \code{XGBoost}, other boosting algorithms, or deep learning models (\code{mlr3torch}),
utilize validation data during the training to prevent overfitting or to log the validation performance.
It is possible to configure learners to be able to receive such an independent validation set during training.
To do so, one must:
\itemize{
\item annotate the learner with the \code{"validation"} property
\item implement the active binding \verb{$internal_valid_scores} (see section \emph{Optional Extractors}), as well as the
private method \verb{$.extract_internal_valid_scores()} which returns the (final) internal validation scores from the
model of the \code{\link{Learner}} and returns them as a named \code{list()} of \code{numeric(1)}.
If the model is not trained yet, this method should return \code{NULL}.
\item Add the \code{validate} parameter, which can be either \code{NULL}, a ratio in $(0, 1)$, \code{"test"}, or \code{"predefined"}:
\itemize{
\item \code{NULL}: no validation
\item \code{ratio}: only proportion \code{1 - ratio} of the task is used for training and \code{ratio} is used for validation.
\item \code{"test"} means that the \code{"test"} task is used.
\strong{Warning}: This can lead to biased performance estimation.
This option is only available if the learner is being trained via \code{\link[=resample]{resample()}}, \code{\link[=benchmark]{benchmark()}} or functions that
internally use them, e.g. \code{tune()} of \CRANpkg{mlr3tuning} or \code{batchmark()} of \CRANpkg{mlr3batchmark}.
This is especially useful for hyperparameter tuning, where one might e.g. want to use the same validation data
for early stopping and model evaluation.
\item \code{"predefined"} means that the task's (manually set) \verb{$internal_valid_task} is used.
See the \code{\link{Task}} documentation for more information.
}
}

For an example how to do this, see \code{\link{LearnerClassifDebug}}.
Note that in \code{.train()}, the \verb{$internal_valid_task} will only be present if the \verb{$validate} field of the \code{Learner}
is set to a non-\code{NULL} value.
}

\section{Implementing Internal Tuning}{

Some learners such as \code{XGBoost} or \code{cv.glmnet} can internally tune hyperparameters.
XGBoost, for example, can tune the number of boosting rounds based on the validation performance.
CV Glmnet, on the other hand, can tune the regularization parameter based on an internal cross-validation.
Internal tuning \emph{can} therefore rely on the internal validation data, but does not necessarily do so.

In order to be able to combine this internal hyperparameter tuning with the standard hyperparameter optimization
implemented via \CRANpkg{mlr3tuning}, one must:
\itemize{
\item annotate the learner with the \code{"internal_tuning"} property
\item implement the active binding \verb{$internal_tuned_values} (see section \emph{Optional Extractors}) as well as the
private method \verb{$.extract_internal_tuned_values()} which extracts the internally tuned values
from the \code{\link{Learner}}'s
model and returns them as a named \code{list()}.
If the model is not trained yet, this method should return \code{NULL}.
\item Have at least one parameter tagged with \code{"internal_tuning"}, which requires to also provide a \code{in_tune_fn} and
\code{disable_tune_fn}, and \emph{should} also include a default aggregation function (\code{aggr}).
}

For an example how to do this, see \code{\link{LearnerClassifDebug}}.
}

\section{Implementing Marshaling}{

Some \code{\link{Learner}}s have models that cannot be serialized as they e.g. contain external pointers.
In order to still be able to save them, use them with parallelization or callr encapsulation it is necessary
to implement how they should be (un)-marshaled. See \code{\link{marshaling}} for how to do this.
}

\section{Implementing Out-of-Bag Error}{

Some \code{\link{Learner}}s can compute the out-of-bag error during training.
In order to do this, the learner must:
\itemize{
\item annotate the learner with the \code{"oob_error"} property
\item implement the private method \verb{$.extract_oob_error()} which extracts the out-of-bag error
from the \code{\link{Learner}}'s model and returns it as a \code{numeric(1)}.
}
}

\examples{

## ------------------------------------------------
## Method `Learner$train()`
## ------------------------------------------------

task   = tsk("penguins")
learner = lrn("classif.rpart")
learner$train(task)

## ------------------------------------------------
## Method `Learner$predict()`
## ------------------------------------------------

task = tsk("penguins")
learner = lrn("classif.rpart")$train(task)
learner$predict(task)

## ------------------------------------------------
## Method `Learner$predict_newdata()`
## ------------------------------------------------

task = tsk("penguins")
learner = lrn("classif.rpart")$train(task)
learner$predict_newdata(task$data(rows = 1:5))

## ------------------------------------------------
## Method `Learner$reset()`
## ------------------------------------------------

task = tsk("penguins")
learner = lrn("classif.rpart")$train(task)
learner$reset()

## ------------------------------------------------
## Method `Learner$encapsulate()`
## ------------------------------------------------

learner = lrn("classif.rpart")
fallback = lrn("classif.featureless")
learner$encapsulate("try", fallback = fallback)

## ------------------------------------------------
## Method `Learner$configure()`
## ------------------------------------------------

learner = lrn("classif.rpart")
learner$configure(minsplit = 3, parallel_predict = FALSE)
learner$configure(.values = list(cp = 0.005))
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-learners}
\item Package \CRANpkg{mlr3learners} for a solid collection of essential learners.
\item Package \href{https://github.com/mlr-org/mlr3extralearners}{mlr3extralearners} for more learners.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Learner]{Learners}: \link{mlr_learners}
\item \code{as.data.table(mlr_learners)} for a table of available \link[=Learner]{Learners} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3pipelines} to combine learners with pre- and postprocessing steps.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
\item \CRANpkg{mlr3tuning} for tuning of hyperparameters, \CRANpkg{mlr3tuningspaces}
for established default tuning spaces.
}

Other Learner:
\code{\link{LearnerClassif}},
\code{\link{LearnerRegr}},
\code{\link{mlr_learners}},
\code{\link{mlr_learners_classif.debug}},
\code{\link{mlr_learners_classif.featureless}},
\code{\link{mlr_learners_classif.rpart}},
\code{\link{mlr_learners_regr.debug}},
\code{\link{mlr_learners_regr.featureless}},
\code{\link{mlr_learners_regr.rpart}}
}
\concept{Learner}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{state}}{(named \code{list()} | \code{NULL})\cr
Current (internal) state of the learner.
Contains all information gathered during \code{train()} and \code{predict()}.
It is not recommended to access elements from \code{state} directly.
This is an internal data structure which may change in the future.}
  }
  \if{html}{\out{</div>}}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{use_weights}}{(\code{character(1)})\cr
How weights should be handled.
Settings are \code{"use"} \code{"ignore"}, and \code{"error"}.
\itemize{
\item \code{"use"}: use weights, as supported by the underlying \code{Learner}.
Only available for \code{Learner}s with the property \code{"weights"}.
\item \code{"ignore"}: do not use weights.
\item \code{"error"}: throw an error if weights are present in the training \code{Task}.
}

For \code{Learner}s with the property \code{"weights"}, this is initialized as \code{"use"}.
For \code{Learner}s that do not support weights, i.e. without the \code{"weights"} property,
this is initialized as \code{"error"}.
The latter behavior is to avoid cases where a user erroneously assumes that a \code{Learner}
supports weights when it does not.
For \code{Learner}s that do not support weights, \code{use_weights} needs to be set to \code{"ignore"}
if tasks with weights should be handled (by dropping the weights).
See Section 'weights' for more details.}

    \item{\code{model}}{(any)\cr
The fitted model. Only available after \verb{$train()} has been called.}

    \item{\code{native_model}}{(any)\cr
The native model object from the upstream package.
For most learners, this is identical to \verb{$model}.
However, some learners store additional information beyond the model from the upstream package.
In such cases, \verb{$model} contains a named list with the native model stored in element \code{model}
along with additional information.
The \verb{$native_model} field can be overwritten by the learner to return the actual model object
from the upstream package.
The default returns \verb{$model}.}

    \item{\code{timings}}{(named \code{numeric(2)})\cr
Elapsed time in seconds for the steps \code{"train"} and \code{"predict"}.

When predictions for multiple predict sets were made during \code{\link[=resample]{resample()}} or \code{\link[=benchmark]{benchmark()}},
the predict time shows the cumulative duration of all predictions.
If \code{learner$predict()} is called manually, the last predict time gets overwritten.

Measured via \code{\link[mlr3misc:encapsulate]{mlr3misc::encapsulate()}}.}

    \item{\code{log}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Returns the output (including warning and errors) as table with columns
\itemize{
\item \code{"stage"} ("train" or "predict"),
\item \code{"class"} ("output", "warning", or "error"), and
\item \code{"condition"} (\code{condition}).
}}

    \item{\code{warnings}}{(\code{list()})\cr
Logged warnings condition objects.}

    \item{\code{errors}}{(\code{list()})\cr
Logged error condition objects.}

    \item{\code{hash}}{(\code{character(1)})\cr
Hash (unique identifier) for this object.
The hash is calculated based on the learner id, the parameter settings,
the predict type, the fallback hash, the parallel predict setting,
the validate setting, the predict sets, and the predict raw setting.}

    \item{\code{phash}}{(\code{character(1)})\cr
Hash (unique identifier) for this partial object, excluding some components
which are varied systematically during tuning (parameter values).}

    \item{\code{predict_type}}{(\code{character(1)})\cr
Stores the currently active predict type, e.g. \code{"response"}.
Must be an element of \verb{$predict_types}.
A few learners already use the predict type during training.
So there is no guarantee that changing the predict type after training will have any effect
or does not lead to errors.}

    \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}

    \item{\code{fallback}}{(\link{Learner})\cr
Returns the fallback learner set with \verb{$encapsulate()}.}

    \item{\code{encapsulation}}{(\code{character(2)})\cr
Returns the encapsulation settings set with \verb{$encapsulate()}.}

    \item{\code{hotstart_stack}}{(\link{HotstartStack})\cr.
Stores \code{HotstartStack}.}

    \item{\code{selected_features_impute}}{(\code{character(1)})\cr
Controls the behavior if the learner does not support feature selection.
If set to \code{"error"}, an error is thrown.
If set to \code{"all"} the complete feature set is returned.}

    \item{\code{predict_types}}{(\code{character()})\cr
Stores the possible predict types the learner is capable of.
A complete list of candidate predict types, grouped by task type,
is stored in \code{\link[=mlr_reflections]{mlr_reflections$learner_predict_types}}.
This field is read-only.}

    \item{\code{id}}{(\code{character(1)})\cr
Identifier of the object.
Used in tables, plot and text output.}

    \item{\code{label}}{(\code{character(1)})\cr
Label for this object.
Can be used in tables, plot and text output instead of the ID.}

    \item{\code{task_type}}{(\code{character(1)})\cr
Task type, e.g. \code{"classif"} or \code{"regr"}.

For a complete list of possible task types (depending on the loaded packages),
see \code{\link[=mlr_reflections]{mlr_reflections$task_types$type}}.}

    \item{\code{feature_types}}{(\code{character()})\cr
Stores the feature types the learner can handle, e.g. \code{"logical"}, \code{"numeric"}, or \code{"factor"}.
A complete list of candidate feature types, grouped by task type,
is stored in \code{\link[=mlr_reflections]{mlr_reflections$task_feature_types}}.}

    \item{\code{properties}}{(\code{character()})\cr
Stores a set of properties/capabilities the learner has.
A complete list of candidate properties, grouped by task type,
is stored in \code{\link[=mlr_reflections]{mlr_reflections$learner_properties}}.}

    \item{\code{packages}}{(\code{character(1)})\cr
Set of required packages.
These packages are loaded, but not attached.}

    \item{\code{predict_sets}}{(\code{character()})\cr
During \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}, a \link{Learner} can predict on multiple sets.
Per default, a learner only predicts observations in the test set (\code{predict_sets == "test"}).
To change this behavior, set \code{predict_sets} to a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
The \code{"train"} predict set contains the train ids from the resampling. This means that if a learner does validation and
sets \verb{$validate} to a ratio (creating the validation data from the training data), the train predictions
will include the predictions for the validation data.
Each set yields a separate \link{Prediction} object.
Those can be combined via getters in \link{ResampleResult}/\link{BenchmarkResult}, or \link{Measure}s can be configured
to operate on specific subsets of the calculated prediction sets.}

    \item{\code{parallel_predict}}{(\code{logical(1)})\cr
If set to \code{TRUE}, use \CRANpkg{future} to calculate predictions in parallel (default: \code{FALSE}).
The row ids of the \code{task} will be split into \code{\link[future:nbrOfWorkers]{future::nbrOfWorkers()}} chunks,
and predictions are evaluated according to the active \code{\link[future:plan]{future::plan()}}.
This currently only works for methods \code{Learner$predict()} and \code{Learner$predict_newdata()},
and has no effect during \code{\link[=resample]{resample()}} or \code{\link[=benchmark]{benchmark()}} where you have other means
to parallelize.

Note that the recorded time required for prediction is not properly defined and
depends on the parallelization backend.}

    \item{\code{predict_raw}}{(\code{logical(1)})\cr
If set to \code{TRUE}, the raw prediction object from the upstream model is stored in the
\link{Prediction} object (default: \code{FALSE}).
The raw prediction is stored as-is, without validation or subsetting during filtering.
When multiple predictions are combined via \code{c()}, the individual raw objects are collected into a \code{list()}.
Individual learner implementations must support this flag by including a \code{raw} element in the list returned
by \verb{$.predict()}.}

    \item{\code{timeout}}{(named \code{numeric(2)})\cr
Timeout for the learner's train and predict steps, in seconds.
This works differently for different encapsulation methods, see
\code{\link[mlr3misc:encapsulate]{mlr3misc::encapsulate()}}.
Default is \code{c(train = Inf, predict = Inf)}.
Also see the section on error handling in the mlr3book:
\url{https://mlr3book.mlr-org.com/chapters/chapter10/advanced_technical_aspects_of_mlr3.html#sec-error-handling}}

    \item{\code{deadline}}{(named \code{POSIXct(2)})\cr
Deadline for the learner's train and predict steps.
This uses the same mechanism as \code{timeout}, but with absolute time (as
date-time) instead of a relative time in seconds.
Default is \code{as.POSIXct(c(train = Inf, predict = Inf))}.}

    \item{\code{man}}{(\code{character(1)} | \code{NULL})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
Defaults to \code{NA}, but can be set by child classes.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Learner-initialize}{\code{Learner$new()}}
    \item \href{#method-Learner-format}{\code{Learner$format()}}
    \item \href{#method-Learner-print}{\code{Learner$print()}}
    \item \href{#method-Learner-help}{\code{Learner$help()}}
    \item \href{#method-Learner-train}{\code{Learner$train()}}
    \item \href{#method-Learner-predict}{\code{Learner$predict()}}
    \item \href{#method-Learner-predict_newdata}{\code{Learner$predict_newdata()}}
    \item \href{#method-Learner-reset}{\code{Learner$reset()}}
    \item \href{#method-Learner-base_learner}{\code{Learner$base_learner()}}
    \item \href{#method-Learner-encapsulate}{\code{Learner$encapsulate()}}
    \item \href{#method-Learner-configure}{\code{Learner$configure()}}
    \item \href{#method-Learner-selected_features}{\code{Learner$selected_features()}}
    \item \href{#method-Learner-clone}{\code{Learner$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-initialize}{}}}
\subsection{\code{Learner$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.

Note that this object is typically constructed via a derived classes, e.g. \link{LearnerClassif} or \link{LearnerRegr}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$new(
  id,
  task_type,
  param_set = ps(),
  predict_types = character(),
  feature_types = character(),
  properties = character(),
  packages = character(),
  label = NA_character_,
  man = NA_character_
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier for the new instance.}
      \item{\code{task_type}}{(\code{character(1)})\cr
Type of task, e.g. \code{"regr"} or \code{"classif"}.
Must be an element of \link[=mlr_reflections]{mlr_reflections$task_types$type}.}
      \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}
      \item{\code{predict_types}}{(\code{character()})\cr
Supported predict types. Must be a subset of \code{\link[=mlr_reflections]{mlr_reflections$learner_predict_types}}.}
      \item{\code{feature_types}}{(\code{character()})\cr
Feature types the learner operates on. Must be a subset of \code{\link[=mlr_reflections]{mlr_reflections$task_feature_types}}.}
      \item{\code{properties}}{(\code{character()})\cr
Set of properties of the \link{Learner}.
Must be a subset of \code{\link[=mlr_reflections]{mlr_reflections$learner_properties}}.
The following properties are currently standardized and understood by learners in \CRANpkg{mlr3}:
\itemize{
\item \code{"missings"}: The learner can handle missing values in the data.
\item \code{"weights"}: The learner supports observation weights.
\item \code{"offset"}: The learner can incorporate offset values to adjust predictions.
\item \code{"importance"}: The learner supports extraction of importance scores, i.e. comes with an \verb{$importance()} extractor function (see section on optional extractors in \link{Learner}).
\item \code{"selected_features"}: The learner supports extraction of the set of selected features, i.e. comes with a \verb{$selected_features()} extractor function (see section on optional extractors in \link{Learner}).
\item \code{"oob_error"}: The learner supports extraction of estimated out of bag error, i.e. comes with a \code{oob_error()} extractor function (see section on optional extractors in \link{Learner}).
\item \code{"validation"}: The learner can use a validation task during training.
\item \code{"internal_tuning"}: The learner is able to internally optimize hyperparameters (those are also tagged with \code{"internal_tuning"}).
\item \code{"marshal"}: To save learners with this property, you need to call \verb{$marshal()} first.
If a learner is in a marshaled state, you call first need to call \verb{$unmarshal()} to use its model, e.g. for prediction.
\item \code{"hotstart_forward"}: The learner supports to hotstart a model forward.
\item \code{"hotstart_backward"}: The learner supports hotstarting a model backward.
\item `"featureless": The learner does not use features.
}}
      \item{\code{packages}}{(\code{character()})\cr
Set of required packages.
A warning is signaled by the constructor if at least one of the packages is not installed,
but loaded (not attached) later on-demand via \code{\link[=requireNamespace]{requireNamespace()}}.}
      \item{\code{label}}{(\code{character(1)})\cr
Label for the new instance.}
      \item{\code{man}}{(\code{character(1)})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
The referenced help package can be opened via method \verb{$help()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-format"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-format}{}}}
\subsection{\code{Learner$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-print"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-print}{}}}
\subsection{\code{Learner$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-help"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-help}{}}}
\subsection{\code{Learner$help()}}{
  Opens the corresponding help page referenced by field \verb{$man}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$help()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-train"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-train}{}}}
\subsection{\code{Learner$train()}}{
  Train the learner on a set of observations of the provided \code{task}.
Mutates the learner by reference, i.e. stores the model alongside other information in field \verb{$state}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$train(task, row_ids = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link{Task}).}
      \item{\code{row_ids}}{(\code{integer()})\cr
Vector of training indices as subset of \code{task$row_ids}.
For a simple split into training and test set, see \code{\link[=partition]{partition()}}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task   = tsk("penguins")
learner = lrn("classif.rpart")
learner$train(task)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-predict"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-predict}{}}}
\subsection{\code{Learner$predict()}}{
  Uses the fitted model stored in \verb{$state} to generate predictions for a set of observations
from the provided \code{task}.
This method requires that the learner has been previously trained using \verb{$train()}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$predict(task, row_ids = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link{Task})\cr
The task containing the observations to predict on.
Must be compatible with the learner's task type and feature types.
Unlike \verb{$predict_newdata()}, no type conversion is done.}
      \item{\code{row_ids}}{(\code{integer()})\cr
Vector of row indices from \code{task$row_ids} to predict on.
If \code{NULL} (default), predictions are made for all rows in the task.
For a simple train-test split, see \code{\link[=partition]{partition()}}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \link{Prediction} object containing the predictions for the specified observations.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
learner = lrn("classif.rpart")$train(task)
learner$predict(task)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-predict_newdata"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-predict_newdata}{}}}
\subsection{\code{Learner$predict_newdata()}}{
  Uses the model fitted during \verb{$train()} to create a new \link{Prediction} based on the new data in \code{newdata}.
Object \code{task} is the task used during \verb{$train()} and required for conversion of \code{newdata}.
If the learner's \verb{$train()} method has been called, there is a (size reduced) version
of the training task stored in the learner.
If the learner has been fitted via \code{\link[=resample]{resample()}} or \code{\link[=benchmark]{benchmark()}},
you need to pass the corresponding task stored in the \link{ResampleResult} or
\link{BenchmarkResult}, respectively.
Further, \code{\link{auto_convert}} is used for type-conversions to ensure compatibility
of features between \verb{$train()} and \verb{$predict()}.

If the stored training task has a \code{weights_measure} column,
\emph{and} if \code{newdata} contains a column with the same name,
that column must be numeric with no missing values and is used as measure weights column.
Otherwise, no measure weights are used.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$predict_newdata(newdata, task = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{newdata}}{(any object supported by \code{\link[=as_data_backend]{as_data_backend()}})\cr
New data to predict on.
All data formats convertible by \code{\link[=as_data_backend]{as_data_backend()}} are supported, e.g.
\code{data.frame()} or \link{DataBackend}.
If a \link{DataBackend} is provided as \code{newdata}, the row ids are preserved,
otherwise they are set to the sequence \code{1:nrow(newdata)}.}
      \item{\code{task}}{(\link{Task}).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \link{Prediction}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
learner = lrn("classif.rpart")$train(task)
learner$predict_newdata(task$data(rows = 1:5))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-reset"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-reset}{}}}
\subsection{\code{Learner$reset()}}{
  Reset the learner, i.e. un-train by resetting the \code{state}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$reset()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
learner = lrn("classif.rpart")$train(task)
learner$reset()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-base_learner"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-base_learner}{}}}
\subsection{\code{Learner$base_learner()}}{
  Extracts the base learner from nested learner objects like
\code{GraphLearner} in \CRANpkg{mlr3pipelines} or \code{AutoTuner} in
\CRANpkg{mlr3tuning}.
Returns the \link{Learner} itself for regular learners.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$base_learner(recursive = Inf)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{recursive}}{(\code{integer(1)})\cr
Depth of recursion for multiple nested objects.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \link{Learner}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-encapsulate"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-encapsulate}{}}}
\subsection{\code{Learner$encapsulate()}}{
  Sets the encapsulation method and fallback learner for the train and predict steps.
There are currently four different methods implemented:
\itemize{
\item \code{"none"}: Just runs the learner in the current session and measures the elapsed time.
Does not keep a log, output is printed directly to the console.
Works well together with \code{\link[=traceback]{traceback()}}.
\item \code{"try"}: Similar to \code{"none"}, but catches error.
Output is printed to the console and not logged.
\item \code{"evaluate"}: Uses the package \CRANpkg{evaluate} to call the learner, measure time and do the logging.
\item \code{"callr"}: Uses the package \CRANpkg{callr} to call the learner, measure time and do the logging.
This encapsulation spawns a separate R session in which the learner is called.
While this comes with a considerable overhead,
it also guards your session from being torn down by segfaults.
\item \code{"mirai"}: Uses the package \CRANpkg{mirai} to call the learner,
measure time and do the logging.
This encapsulation calls the function in a \code{mirai} on a \code{daemon}.
The \code{daemon} can be pre-started via \code{daemons(1, .compute = "mlr3_encapsulation")},
otherwise a new R session will be created for each encapsulated call.
If a \code{daemon} is already running with compute profile \code{"mlr3_encapsulation"},
it will be used to execute all calls.
Using \code{mirai} is similarly safe as \code{callr} but much faster if several learners
are encapsulated one after the other on the same daemon.
}

The fallback learner is fitted to create valid predictions in case that either the
model fitting or the prediction of the original learner fails.
If the training step or the predict step of the original learner fails,
the fallback is used to make the predictions.
If the original learner only partially fails during predict step
(usually in the form of missing to predict some observations or producing
some \code{NA} predictions),
these missing predictions are imputed by the fallback.
Note that the fallback is always trained, as we do not know in advance whether prediction will fail.
If the training step fails, the \verb{$model} field of the original learner is \code{NULL}.
The results are reproducible across the different encapsulation methods.

Note that for errors of class \code{Mlr3ErrorConfig}, the function always errs and no fallback learner
is trained.

Also see the section on error handling in the mlr3book:
\url{https://mlr3book.mlr-org.com/chapters/chapter10/advanced_technical_aspects_of_mlr3.html#sec-error-handling}
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$encapsulate(method, fallback = NULL, when = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{method}}{\code{character(1)}\cr
One of \code{"none"}, \code{"try"}, \code{"evaluate"}, \code{"callr"} or \code{"mirai"}.
See the description for details.}
      \item{\code{fallback}}{\link{Learner}\cr
The fallback learner for failed predictions.}
      \item{\code{when}}{(\verb{function(cond, stage)})\cr
Function that takes in the condition (\code{cond}) and the stage (\code{"train"} or \code{"predict"}) and
returns \code{logical(1)} indicating whether to run the fallback learner.

If \code{NULL} (default), the fallback is always used, except for errors of class \code{Mlr3ErrorConfig}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{self} (invisibly).
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{learner = lrn("classif.rpart")
fallback = lrn("classif.featureless")
learner$encapsulate("try", fallback = fallback)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-configure"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-configure}{}}}
\subsection{\code{Learner$configure()}}{
  Sets parameter values and fields of the learner.
All arguments whose names match the name of a parameter of the \link[paradox:ParamSet]{paradox::ParamSet} are set as parameters.
All remaining arguments are assumed to be regular fields.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$configure(..., .values = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(named \code{any})\cr
Named arguments to set parameter values and fields.}
      \item{\code{.values}}{(named \code{any})\cr
Named list of parameter values and fields.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{learner = lrn("classif.rpart")
learner$configure(minsplit = 3, parallel_predict = FALSE)
learner$configure(.values = list(cp = 0.005))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-selected_features"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-selected_features}{}}}
\subsection{\code{Learner$selected_features()}}{
  Returns the features selected by the model.
The field \code{selected_features_impute} controls the behavior if the learner does not support feature selection.
If set to \code{"error"}, an error is thrown, otherwise all features are returned.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$selected_features()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Learner-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Learner-clone}{}}}
\subsection{\code{Learner$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Learner$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskClassif_breast_cancer.R
\name{mlr_tasks_breast_cancer}
\alias{mlr_tasks_breast_cancer}
\title{Wisconsin Breast Cancer Classification Task}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link{TaskClassif}.
}
\description{
A classification task for the \link[mlbench:BreastCancer]{mlbench::BreastCancer} data set.
\itemize{
\item Column \code{"Id"} has been removed.
\item Column names have been converted to \code{snake_case}.
\item Positive class is set to \code{"malignant"}.
\item 16 incomplete cases have been removed from the data set.
\item All factor features have been converted to ordered factors.
}
}
\section{Dictionary}{

This \link{Task} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_tasks} or with the associated sugar function \code{\link[=tsk]{tsk()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("breast_cancer")
tsk("breast_cancer")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Dimensions: 683x10
\item Properties: \dQuote{twoclass}
\item Has Missings: \code{FALSE}
\item Target: \dQuote{class}
\item Features: \dQuote{bare_nuclei}, \dQuote{bl_cromatin}, \dQuote{cell_shape}, \dQuote{cell_size}, \dQuote{cl_thickness}, \dQuote{epith_c_size}, \dQuote{marg_adhesion}, \dQuote{mitoses}, \dQuote{normal_nucleoli}
}
}

\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{Task}},
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_task_generators.R
\name{mlr_task_generators}
\alias{mlr_task_generators}
\title{Dictionary of Task Generators}
\format{
\link[R6:R6Class]{R6::R6Class} object inheriting from \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}
\description{
A simple \link[mlr3misc:Dictionary]{mlr3misc::Dictionary} storing objects of class \link{TaskGenerator}.
Each task generator has an associated help page, see \code{mlr_task_generators_[id]}.

This dictionary can get populated with additional task generators by add-on packages.

For a more convenient way to retrieve and construct task generators, see \code{\link[=tgen]{tgen()}}/\code{\link[=tgens]{tgens()}}.
}
\section{Methods}{

See \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}

\section{S3 methods}{

\itemize{
\item \code{as.data.table(dict, ..., objects = FALSE)}\cr
\link[mlr3misc:Dictionary]{mlr3misc::Dictionary} -> \code{\link[data.table:data.table]{data.table::data.table()}}\cr
Returns a \code{\link[data.table:data.table]{data.table::data.table()}} with fields "key", "label", "task_type", "params", and "packages" as columns.
If \code{objects} is set to \code{TRUE}, the constructed objects are returned in the list column named \code{object}.
}
}

\examples{
mlr_task_generators$get("smiley")
tgen("2dnormals")
}
\seealso{
Sugar functions: \code{\link[=tgen]{tgen()}}, \code{\link[=tgens]{tgens()}}

Other Dictionary:
\code{\link{mlr_learners}},
\code{\link{mlr_measures}},
\code{\link{mlr_resamplings}},
\code{\link{mlr_tasks}}

Other TaskGenerator:
\code{\link{TaskGenerator}},
\code{\link{mlr_task_generators_2dnormals}},
\code{\link{mlr_task_generators_cassini}},
\code{\link{mlr_task_generators_circle}},
\code{\link{mlr_task_generators_friedman1}},
\code{\link{mlr_task_generators_moons}},
\code{\link{mlr_task_generators_peak}},
\code{\link{mlr_task_generators_simplex}},
\code{\link{mlr_task_generators_smiley}},
\code{\link{mlr_task_generators_spirals}},
\code{\link{mlr_task_generators_xor}}
}
\concept{Dictionary}
\concept{TaskGenerator}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/warn_deprecated.R
\name{deprecated_binding_old}
\alias{deprecated_binding_old}
\title{Create an Active Binding that Generates a Deprecation Warning}
\usage{
deprecated_binding_old(what, value)
}
\arguments{
\item{what}{(character(1))\cr
A description of the deprecated binding. Should be of the form \code{"Class$field"}.}

\item{value}{(any)\cr
The value of the active binding. This should be an expression that will
be evaluated in the context of the active binding. It could, for example,
refer to \code{self}.}
}
\description{
Creates an active binding that generates a warning when accessed, using
\code{warn_deprecated_old()}. The active binding will otherwise be read-only.
}
\examples{
MyClass = R6::R6Class("MyClass", public = list(),
  active = list(
    foo = deprecated_binding_old("MyClass$foo", "bar")
  )
)
mco = MyClass$new()
mco$foo
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Task.R
\name{col_info}
\alias{col_info}
\alias{col_info.data.table}
\alias{col_info.DataBackend}
\title{Column Information for Backend}
\usage{
col_info(x, ...)

\method{col_info}{data.table}(x, primary_key = character(), ...)

\method{col_info}{DataBackend}(x, ...)
}
\arguments{
\item{x}{(any)\cr
A backend-like object for which to retrieve column information.}

\item{...}{(any)\cr
Additional arguments.}

\item{primary_key}{(\code{character()})\cr
The primary key of the backend.}
}
\description{
Collects column information for backend.

Currently, this includes:
\itemize{
\item storage type
\item levels (factor / ordered), but not for the primary key column
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_learner.R
\name{as_learner}
\alias{as_learner}
\alias{as_learner.Learner}
\alias{as_learners}
\alias{as_learners.default}
\alias{as_learners.list}
\title{Convert to a Learner}
\usage{
as_learner(x, ...)

\method{as_learner}{Learner}(x, clone = FALSE, discard_state = FALSE, ...)

as_learners(x, ...)

\method{as_learners}{default}(x, ...)

\method{as_learners}{list}(x, ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{...}{(any)\cr
Additional arguments.}

\item{clone}{(\code{logical(1)})\cr
If \code{TRUE}, ensures that the returned object is not the same as the input \code{x}.}

\item{discard_state}{(\code{logical(1)})
Whether to discard the state.}
}
\value{
\link{Learner}.
}
\description{
Convert object to a \link{Learner} or a list of \link{Learner}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_learners.R
\name{mlr_learners}
\alias{mlr_learners}
\title{Dictionary of Learners}
\format{
\link[R6:R6Class]{R6::R6Class} object inheriting from \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}
\description{
A simple \link[mlr3misc:Dictionary]{mlr3misc::Dictionary} storing objects of class \link{Learner}.
Each learner has an associated help page, see \code{mlr_learners_[id]}.

This dictionary can get populated with additional learners by add-on packages.
For an opinionated set of solid classification and regression learners,
install and load the \CRANpkg{mlr3learners} package.
More learners are connected via \url{https://github.com/mlr-org/mlr3extralearners}.

For a more convenient way to retrieve and construct learners, see \code{\link[=lrn]{lrn()}}/\code{\link[=lrns]{lrns()}}.
}
\section{Methods}{

See \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}

\section{S3 methods}{

\itemize{
\item \code{as.data.table(dict, ..., objects = FALSE)}\cr
\link[mlr3misc:Dictionary]{mlr3misc::Dictionary} -> \code{\link[data.table:data.table]{data.table::data.table()}}\cr
Returns a \code{\link[data.table:data.table]{data.table::data.table()}} with fields "key", "label", "task_type", "feature_types", "packages",
"properties", and "predict_types" as columns.
If \code{objects} is set to \code{TRUE}, the constructed objects are returned in the list column named \code{object}.
}
}

\examples{
as.data.table(mlr_learners)
mlr_learners$get("classif.featureless")
lrn("classif.rpart")
}
\seealso{
Sugar functions: \code{\link[=lrn]{lrn()}}, \code{\link[=lrns]{lrns()}}

Extension Packages: \CRANpkg{mlr3learners}

Other Dictionary:
\code{\link{mlr_measures}},
\code{\link{mlr_resamplings}},
\code{\link{mlr_task_generators}},
\code{\link{mlr_tasks}}

Other Learner:
\code{\link{Learner}},
\code{\link{LearnerClassif}},
\code{\link{LearnerRegr}},
\code{\link{mlr_learners_classif.debug}},
\code{\link{mlr_learners_classif.featureless}},
\code{\link{mlr_learners_classif.rpart}},
\code{\link{mlr_learners_regr.debug}},
\code{\link{mlr_learners_regr.featureless}},
\code{\link{mlr_learners_regr.rpart}}
}
\concept{Dictionary}
\concept{Learner}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureRegrRSQ.R
\name{mlr_measures_regr.rsq}
\alias{mlr_measures_regr.rsq}
\alias{MeasureRegrRSQ}
\title{R-Squared}
\description{
Measure to compare true observed response with predicted response in regression tasks.
}
\details{
R Squared is defined as \deqn{
  1 - \frac{\sum_{i=1}^n w_i \left( t_i - r_i \right)^2}{\sum_{i=1}^n w_i \left( t_i - \bar{t} \right)^2},
}{
  1 - sum(w * (t - r)^2) / sum(w * (t - mean(t))^2),
}
where \eqn{\bar{t} = \frac{1}{n} \sum_{i=1}^n t_i} and \eqn{w_i} are weights.

Also known as coefficient of determination or explained variation.
It compares the squared error of the predictions relative to a naive model predicting the mean.

Note that weights are used to scale the squared error of individual predictions
(both in the numerator and in the denominator),
but the "plug in" value \eqn{\bar{t}} is computed without weights.

This measure is undefined for constant \eqn{t}.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("regr.rsq")
msr("regr.rsq")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{regr}
\item Range: \eqn{(-\infty, 1]}{(-Inf, 1]}
\item Minimize: FALSE
\item Average: macro
\item Required Prediction: \dQuote{response}
\item Required Packages: \CRANpkg{mlr3}
}
}

\section{Parameters}{

Empty ParamSet
}

\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-eval}
\item Package \CRANpkg{mlr3measures} for the scoring functions.
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}
\code{as.data.table(mlr_measures)} for a table of available \link[=Measure]{Measures} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Measure:
\code{\link{Measure}},
\code{\link{MeasureClassif}},
\code{\link{MeasureRegr}},
\code{\link{MeasureSimilarity}},
\code{\link{mlr_measures}},
\code{\link{mlr_measures_aic}},
\code{\link{mlr_measures_bic}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_debug_classif}},
\code{\link{mlr_measures_elapsed_time}},
\code{\link{mlr_measures_internal_valid_score}},
\code{\link{mlr_measures_oob_error}},
\code{\link{mlr_measures_regr.pinball}},
\code{\link{mlr_measures_regr.rqr}},
\code{\link{mlr_measures_selected_features}}
}
\concept{Measure}
\section{Super classes}{
\code{\link[mlr3:Measure]{Measure}} -> \code{\link[mlr3:MeasureRegr]{MeasureRegr}} -> \code{MeasureRegrRSQ}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{pred_set_mean}}{(\code{logical(1)})\cr
Whether the mean of the true values is calculated on the prediction set.
Set during construction.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MeasureRegrRSQ-initialize}{\code{MeasureRegrRSQ$new()}}
    \item \href{#method-MeasureRegrRSQ-clone}{\code{MeasureRegrRSQ$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="aggregate"><a href='../../mlr3/html/Measure.html#method-Measure-aggregate'><code>Measure$aggregate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="format"><a href='../../mlr3/html/Measure.html#method-Measure-format'><code>Measure$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="help"><a href='../../mlr3/html/Measure.html#method-Measure-help'><code>Measure$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="obs_loss"><a href='../../mlr3/html/Measure.html#method-Measure-obs_loss'><code>Measure$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="print"><a href='../../mlr3/html/Measure.html#method-Measure-print'><code>Measure$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="score"><a href='../../mlr3/html/Measure.html#method-Measure-score'><code>Measure$score()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureRegrRSQ-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureRegrRSQ-initialize}{}}}
\subsection{\code{MeasureRegrRSQ$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureRegrRSQ$new(pred_set_mean = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{pred_set_mean}}{\code{logical(1)}\cr
If \code{TRUE}, the mean of the true values is calculated on the prediction set.
If \code{FALSE}, the mean of the true values is calculated on the training set.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureRegrRSQ-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureRegrRSQ-clone}{}}}
\subsection{\code{MeasureRegrRSQ$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureRegrRSQ$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LearnerClassifFeatureless.R
\name{mlr_learners_classif.featureless}
\alias{mlr_learners_classif.featureless}
\alias{LearnerClassifFeatureless}
\title{Featureless Classification Learner}
\description{
A simple \link{LearnerClassif} which only analyzes the labels during train, ignoring all features.
Hyperparameter \code{method} determines the mode of operation during prediction:
\describe{
\item{mode:}{
Predicts the most frequent label. If there are two or more labels tied, randomly selects one per prediction.
Probabilities correspond to the relative frequency of the class labels in the training set.
For weighted data, the label(s) with the highest weighted frequency are selected.
}
\item{sample:}{
Randomly predict a label uniformly.
Probabilities correspond to a uniform distribution of class labels, i.e. 1 divided by the number of classes.
Weights are ignored, if present.
}
\item{weighted.sample:}{
Randomly predict a label, with probability estimated from the training distribution.
For consistency, probabilities are 1 for the sampled label and 0 for all other labels.
For weighted data, sample weights are used to weight the class labels.
}
}
}
\section{Dictionary}{

This \link{Learner} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_learners} or with the associated sugar function \code{\link[=lrn]{lrn()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_learners$get("classif.featureless")
lrn("classif.featureless")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Predict Types: \dQuote{response}, \dQuote{prob}
\item Feature Types: \dQuote{logical}, \dQuote{integer}, \dQuote{numeric}, \dQuote{character}, \dQuote{factor}, \dQuote{ordered}, \dQuote{POSIXct}, \dQuote{Date}
\item Required Packages: \CRANpkg{mlr3}
}
}

\section{Parameters}{
\tabular{llll}{
   Id \tab Type \tab Default \tab Levels \cr
   method \tab character \tab mode \tab mode, sample, weighted.sample \cr
}
}

\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-learners}
\item Package \CRANpkg{mlr3learners} for a solid collection of essential learners.
\item Package \href{https://github.com/mlr-org/mlr3extralearners}{mlr3extralearners} for more learners.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Learner]{Learners}: \link{mlr_learners}
\item \code{as.data.table(mlr_learners)} for a table of available \link[=Learner]{Learners} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3pipelines} to combine learners with pre- and postprocessing steps.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
\item \CRANpkg{mlr3tuning} for tuning of hyperparameters, \CRANpkg{mlr3tuningspaces}
for established default tuning spaces.
}

Other Learner:
\code{\link{Learner}},
\code{\link{LearnerClassif}},
\code{\link{LearnerRegr}},
\code{\link{mlr_learners}},
\code{\link{mlr_learners_classif.debug}},
\code{\link{mlr_learners_classif.rpart}},
\code{\link{mlr_learners_regr.debug}},
\code{\link{mlr_learners_regr.featureless}},
\code{\link{mlr_learners_regr.rpart}}
}
\concept{Learner}
\section{Super classes}{
\code{\link[mlr3:Learner]{Learner}} -> \code{\link[mlr3:LearnerClassif]{LearnerClassif}} -> \code{LearnerClassifFeatureless}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-LearnerClassifFeatureless-initialize}{\code{LearnerClassifFeatureless$new()}}
    \item \href{#method-LearnerClassifFeatureless-importance}{\code{LearnerClassifFeatureless$importance()}}
    \item \href{#method-LearnerClassifFeatureless-selected_features}{\code{LearnerClassifFeatureless$selected_features()}}
    \item \href{#method-LearnerClassifFeatureless-clone}{\code{LearnerClassifFeatureless$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="base_learner"><a href='../../mlr3/html/Learner.html#method-Learner-base_learner'><code>Learner$base_learner()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="configure"><a href='../../mlr3/html/Learner.html#method-Learner-configure'><code>Learner$configure()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="encapsulate"><a href='../../mlr3/html/Learner.html#method-Learner-encapsulate'><code>Learner$encapsulate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="format"><a href='../../mlr3/html/Learner.html#method-Learner-format'><code>Learner$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="help"><a href='../../mlr3/html/Learner.html#method-Learner-help'><code>Learner$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict"><a href='../../mlr3/html/Learner.html#method-Learner-predict'><code>Learner$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict_newdata"><a href='../../mlr3/html/Learner.html#method-Learner-predict_newdata'><code>Learner$predict_newdata()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="print"><a href='../../mlr3/html/Learner.html#method-Learner-print'><code>Learner$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="reset"><a href='../../mlr3/html/Learner.html#method-Learner-reset'><code>Learner$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="train"><a href='../../mlr3/html/Learner.html#method-Learner-train'><code>Learner$train()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="LearnerClassif" data-id="predict_newdata_fast"><a href='../../mlr3/html/LearnerClassif.html#method-LearnerClassif-predict_newdata_fast'><code>LearnerClassif$predict_newdata_fast()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifFeatureless-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifFeatureless-initialize}{}}}
\subsection{\code{LearnerClassifFeatureless$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifFeatureless$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifFeatureless-importance"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifFeatureless-importance}{}}}
\subsection{\code{LearnerClassifFeatureless$importance()}}{
  All features have a score of \code{0} for this learner.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifFeatureless$importance()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Named \code{numeric()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifFeatureless-selected_features"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifFeatureless-selected_features}{}}}
\subsection{\code{LearnerClassifFeatureless$selected_features()}}{
  Selected features are always the empty set for this learner.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifFeatureless$selected_features()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{character(0)}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifFeatureless-clone"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifFeatureless-clone}{}}}
\subsection{\code{LearnerClassifFeatureless$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifFeatureless$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_task_unsupervised.R
\name{as_task_unsupervised}
\alias{as_task_unsupervised}
\alias{as_task_unsupervised.Task}
\alias{as_task_unsupervised.data.frame}
\alias{as_task_unsupervised.DataBackend}
\alias{as_tasks_unsupervised}
\alias{as_tasks_unsupervised.list}
\alias{as_tasks_unsupervised.Task}
\title{Convert to an Unsupervised Task}
\usage{
as_task_unsupervised(x, ...)

\method{as_task_unsupervised}{Task}(x, clone = FALSE, ...)

\method{as_task_unsupervised}{data.frame}(
  x,
  id = deparse1(substitute(x)),
  label = NA_character_,
  ...
)

\method{as_task_unsupervised}{DataBackend}(
  x,
  id = deparse1(substitute(x)),
  label = NA_character_,
  ...
)

as_tasks_unsupervised(x, ...)

\method{as_tasks_unsupervised}{list}(x, clone = FALSE, ...)

\method{as_tasks_unsupervised}{Task}(x, clone = FALSE, ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{...}{(any)\cr
Additional arguments.}

\item{clone}{(\code{logical(1)})\cr
If \code{TRUE}, ensures that the returned object is not the same as the input \code{x}.}

\item{id}{(\code{character(1)})\cr
Id for the new task.
Defaults to the (deparsed and substituted) name of the data argument.}

\item{label}{(\code{character(1)})\cr
Label for the new instance.}
}
\description{
Convert object to a \link{TaskUnsupervised} or a list of \link{TaskUnsupervised}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_measures.R
\name{mlr_measures}
\alias{mlr_measures}
\title{Dictionary of Performance Measures}
\format{
\link[R6:R6Class]{R6::R6Class} object inheriting from \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}
\description{
A simple \link[mlr3misc:Dictionary]{mlr3misc::Dictionary} storing objects of class \link{Measure}.
Each measure has an associated help page, see \code{mlr_measures_[id]}.

This dictionary can get populated with additional measures by add-on packages.
E.g., \CRANpkg{mlr3proba} adds survival measures and \CRANpkg{mlr3cluster} adds
cluster analysis measures.

For a more convenient way to retrieve and construct measures, see \code{\link[=msr]{msr()}}/\code{\link[=msrs]{msrs()}}.
}
\section{Methods}{

See \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}

\section{S3 methods}{

\itemize{
\item \code{as.data.table(dict, ..., objects = FALSE)}\cr
\link[mlr3misc:Dictionary]{mlr3misc::Dictionary} -> \code{\link[data.table:data.table]{data.table::data.table()}}\cr
Returns a \code{\link[data.table:data.table]{data.table::data.table()}} with fields "key", "label", "task_type", "packages",
"predict_type", and "task_properties" as columns.
If \code{objects} is set to \code{TRUE}, the constructed objects are returned in the list column named \code{object}.
}
}

\examples{
as.data.table(mlr_measures)
mlr_measures$get("classif.ce")
msr("regr.mse")
}
\seealso{
Sugar functions: \code{\link[=msr]{msr()}}, \code{\link[=msrs]{msrs()}}

Implementation of most measures: \CRANpkg{mlr3measures}

Other Dictionary:
\code{\link{mlr_learners}},
\code{\link{mlr_resamplings}},
\code{\link{mlr_task_generators}},
\code{\link{mlr_tasks}}

Other Measure:
\code{\link{Measure}},
\code{\link{MeasureClassif}},
\code{\link{MeasureRegr}},
\code{\link{MeasureSimilarity}},
\code{\link{mlr_measures_aic}},
\code{\link{mlr_measures_bic}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_debug_classif}},
\code{\link{mlr_measures_elapsed_time}},
\code{\link{mlr_measures_internal_valid_score}},
\code{\link{mlr_measures_oob_error}},
\code{\link{mlr_measures_regr.pinball}},
\code{\link{mlr_measures_regr.rqr}},
\code{\link{mlr_measures_regr.rsq}},
\code{\link{mlr_measures_selected_features}}
}
\concept{Dictionary}
\concept{Measure}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskClassif_sonar.R
\name{mlr_tasks_sonar}
\alias{mlr_tasks_sonar}
\title{Sonar Classification Task}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link{TaskClassif}.
}
\description{
A classification task for the \link[mlbench:Sonar]{mlbench::Sonar} data set.
Positive class is set to "M" (Mine).
}
\section{Dictionary}{

This \link{Task} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_tasks} or with the associated sugar function \code{\link[=tsk]{tsk()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("sonar")
tsk("sonar")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Dimensions: 208x61
\item Properties: \dQuote{twoclass}
\item Has Missings: \code{FALSE}
\item Target: \dQuote{Class}
\item Features: \dQuote{V1}, \dQuote{V10}, \dQuote{V11}, \dQuote{V12}, \dQuote{V13}, \dQuote{V14}, \dQuote{V15}, \dQuote{V16}, \dQuote{V17}, \dQuote{V18}, \dQuote{V19}, \dQuote{V2}, \dQuote{V20}, \dQuote{V21}, \dQuote{V22}, \dQuote{V23}, \dQuote{V24}, \dQuote{V25}, \dQuote{V26}, \dQuote{V27}, \dQuote{V28}, \dQuote{V29}, \dQuote{V3}, \dQuote{V30}, \dQuote{V31}, \dQuote{V32}, \dQuote{V33}, \dQuote{V34}, \dQuote{V35}, \dQuote{V36}, \dQuote{V37}, \dQuote{V38}, \dQuote{V39}, \dQuote{V4}, \dQuote{V40}, \dQuote{V41}, \dQuote{V42}, \dQuote{V43}, \dQuote{V44}, \dQuote{V45}, \dQuote{V46}, \dQuote{V47}, \dQuote{V48}, \dQuote{V49}, \dQuote{V5}, \dQuote{V50}, \dQuote{V51}, \dQuote{V52}, \dQuote{V53}, \dQuote{V54}, \dQuote{V55}, \dQuote{V56}, \dQuote{V57}, \dQuote{V58}, \dQuote{V59}, \dQuote{V6}, \dQuote{V60}, \dQuote{V7}, \dQuote{V8}, \dQuote{V9}
}
}

\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{Task}},
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskGenerator.R
\name{TaskGenerator}
\alias{TaskGenerator}
\title{TaskGenerator Class}
\description{
Creates a \link{Task} of arbitrary size.
Predefined task generators are stored in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_task_generators},
e.g. \code{\link[=mlr_task_generators_xor]{xor}}.
}
\seealso{
\itemize{
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=TaskGenerator]{TaskGenerators}: \link{mlr_task_generators}
\item \code{as.data.table(mlr_task_generators)}  for a table of available \link[=TaskGenerator]{TaskGenerators} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other TaskGenerator:
\code{\link{mlr_task_generators}},
\code{\link{mlr_task_generators_2dnormals}},
\code{\link{mlr_task_generators_cassini}},
\code{\link{mlr_task_generators_circle}},
\code{\link{mlr_task_generators_friedman1}},
\code{\link{mlr_task_generators_moons}},
\code{\link{mlr_task_generators_peak}},
\code{\link{mlr_task_generators_simplex}},
\code{\link{mlr_task_generators_smiley}},
\code{\link{mlr_task_generators_spirals}},
\code{\link{mlr_task_generators_xor}}
}
\concept{TaskGenerator}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{id}}{(\code{character(1)})\cr
Identifier of the object.
Used in tables, plot and text output.}

    \item{\code{label}}{(\code{character(1)})\cr
Label for this object.
Can be used in tables, plot and text output instead of the ID.}

    \item{\code{task_type}}{(\code{character(1)})\cr
Task type, e.g. \code{"classif"} or \code{"regr"}.

For a complete list of possible task types (depending on the loaded packages),
see \code{\link[=mlr_reflections]{mlr_reflections$task_types$type}}.}

    \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}

    \item{\code{packages}}{(\code{character(1)})\cr
Set of required packages.
These packages are loaded, but not attached.}

    \item{\code{man}}{(\code{character(1)} | \code{NULL})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
Defaults to \code{NA}, but can be set by child classes.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-TaskGenerator-initialize}{\code{TaskGenerator$new()}}
    \item \href{#method-TaskGenerator-format}{\code{TaskGenerator$format()}}
    \item \href{#method-TaskGenerator-print}{\code{TaskGenerator$print()}}
    \item \href{#method-TaskGenerator-generate}{\code{TaskGenerator$generate()}}
    \item \href{#method-TaskGenerator-clone}{\code{TaskGenerator$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGenerator-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGenerator-initialize}{}}}
\subsection{\code{TaskGenerator$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGenerator$new(
  id,
  task_type,
  packages = character(),
  param_set = ps(),
  label = NA_character_,
  man = NA_character_
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier for the new instance.}
      \item{\code{task_type}}{(\code{character(1)})\cr
Type of task, e.g. \code{"regr"} or \code{"classif"}.
Must be an element of \link[=mlr_reflections]{mlr_reflections$task_types$type}.}
      \item{\code{packages}}{(\code{character()})\cr
Set of required packages.
A warning is signaled by the constructor if at least one of the packages is not installed,
but loaded (not attached) later on-demand via \code{\link[=requireNamespace]{requireNamespace()}}.}
      \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}
      \item{\code{label}}{(\code{character(1)})\cr
Label for the new instance.}
      \item{\code{man}}{(\code{character(1)})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
The referenced help package can be opened via method \verb{$help()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGenerator-format"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGenerator-format}{}}}
\subsection{\code{TaskGenerator$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGenerator$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGenerator-print"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGenerator-print}{}}}
\subsection{\code{TaskGenerator$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGenerator$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGenerator-generate"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGenerator-generate}{}}}
\subsection{\code{TaskGenerator$generate()}}{
  Creates a task of type \code{task_type} with \code{n} observations,
possibly using additional settings stored in \code{param_set}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGenerator$generate(n)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{(\code{integer(1)})\cr
Number of rows to generate.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \link{Task}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGenerator-clone"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGenerator-clone}{}}}
\subsection{\code{TaskGenerator$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGenerator$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureClassifCosts.R
\name{mlr_measures_classif.costs}
\alias{mlr_measures_classif.costs}
\alias{MeasureClassifCosts}
\title{Cost-sensitive Classification Measure}
\description{
Uses a cost matrix to create a classification measure.
True labels must be arranged in columns, predicted labels must be arranged  in rows.
The cost matrix is stored as slot \verb{$costs}.

For calculation of the score, the confusion matrix is multiplied element-wise with the cost matrix.
The costs are then summed up (and potentially divided by the number of observations
if \code{normalize} is set to \code{TRUE} (default)).
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("classif.costs")
msr("classif.costs")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Range: \eqn{(-\infty, \infty)}{(-Inf, Inf)}
\item Minimize: TRUE
\item Average: macro
\item Required Prediction: \dQuote{response}
\item Required Packages: \CRANpkg{mlr3}
}
}

\section{Parameters}{
\tabular{llll}{
   Id \tab Type \tab Default \tab Levels \cr
   normalize \tab logical \tab - \tab TRUE, FALSE \cr
}
}

\examples{
# get a cost sensitive task
task = tsk("german_credit")

# cost matrix as given on the UCI page of the german credit data set
# https://archive.ics.uci.edu/ml/datasets/statlog+(german+credit+data)
costs = matrix(c(0, 5, 1, 0), nrow = 2)
dimnames(costs) = list(truth = task$class_names, predicted = task$class_names)
print(costs)

# mlr3 needs truth in columns, predictions in rows
costs = t(costs)

# create a cost measure which calculates the absolute costs
m = msr("classif.costs", id = "german_credit_costs", costs = costs, normalize = FALSE)

# fit models and evaluate with the cost measure
learner = lrn("classif.rpart")
rr = resample(task, learner, rsmp("cv", folds = 3))
rr$aggregate(m)
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-eval}
\item Package \CRANpkg{mlr3measures} for the scoring functions.
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}
\code{as.data.table(mlr_measures)} for a table of available \link[=Measure]{Measures} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Measure:
\code{\link{Measure}},
\code{\link{MeasureClassif}},
\code{\link{MeasureRegr}},
\code{\link{MeasureSimilarity}},
\code{\link{mlr_measures}},
\code{\link{mlr_measures_aic}},
\code{\link{mlr_measures_bic}},
\code{\link{mlr_measures_debug_classif}},
\code{\link{mlr_measures_elapsed_time}},
\code{\link{mlr_measures_internal_valid_score}},
\code{\link{mlr_measures_oob_error}},
\code{\link{mlr_measures_regr.pinball}},
\code{\link{mlr_measures_regr.rqr}},
\code{\link{mlr_measures_regr.rsq}},
\code{\link{mlr_measures_selected_features}}

Other classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.auc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.bbrier}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.dor}},
\code{\link{mlr_measures_classif.fbeta}},
\code{\link{mlr_measures_classif.fdr}},
\code{\link{mlr_measures_classif.fn}},
\code{\link{mlr_measures_classif.fnr}},
\code{\link{mlr_measures_classif.fomr}},
\code{\link{mlr_measures_classif.fp}},
\code{\link{mlr_measures_classif.fpr}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}},
\code{\link{mlr_measures_classif.mcc}},
\code{\link{mlr_measures_classif.npv}},
\code{\link{mlr_measures_classif.ppv}},
\code{\link{mlr_measures_classif.prauc}},
\code{\link{mlr_measures_classif.precision}},
\code{\link{mlr_measures_classif.recall}},
\code{\link{mlr_measures_classif.sensitivity}},
\code{\link{mlr_measures_classif.specificity}},
\code{\link{mlr_measures_classif.tn}},
\code{\link{mlr_measures_classif.tnr}},
\code{\link{mlr_measures_classif.tp}},
\code{\link{mlr_measures_classif.tpr}}

Other multiclass classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1p}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}},
\code{\link{mlr_measures_classif.mcc}}
}
\concept{Measure}
\concept{classification measures}
\concept{multiclass classification measures}
\section{Super classes}{
\code{\link[mlr3:Measure]{Measure}} -> \code{\link[mlr3:MeasureClassif]{MeasureClassif}} -> \code{MeasureClassifCosts}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{costs}}{(numeric \code{matrix()})\cr
Matrix of costs (truth in columns, predicted response in rows).}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MeasureClassifCosts-initialize}{\code{MeasureClassifCosts$new()}}
    \item \href{#method-MeasureClassifCosts-clone}{\code{MeasureClassifCosts$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="aggregate"><a href='../../mlr3/html/Measure.html#method-Measure-aggregate'><code>Measure$aggregate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="format"><a href='../../mlr3/html/Measure.html#method-Measure-format'><code>Measure$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="help"><a href='../../mlr3/html/Measure.html#method-Measure-help'><code>Measure$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="obs_loss"><a href='../../mlr3/html/Measure.html#method-Measure-obs_loss'><code>Measure$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="print"><a href='../../mlr3/html/Measure.html#method-Measure-print'><code>Measure$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="score"><a href='../../mlr3/html/Measure.html#method-Measure-score'><code>Measure$score()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureClassifCosts-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureClassifCosts-initialize}{}}}
\subsection{\code{MeasureClassifCosts$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureClassifCosts$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureClassifCosts-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureClassifCosts-clone}{}}}
\subsection{\code{MeasureClassifCosts$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureClassifCosts$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskClassif_spam.R
\name{mlr_tasks_spam}
\alias{mlr_tasks_spam}
\title{Spam Classification Task}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link{TaskClassif}.
}
\source{
Creators:
Mark Hopkins, Erik Reeber, George Forman, Jaap Suermondt.
Hewlett-Packard Labs, 1501 Page Mill Rd., Palo Alto, CA 94304

Donor:
George Forman (gforman at nospam hpl.hp.com) 650-857-7835

Preprocessing:
Columns have been renamed. Preprocessed data taken from the \CRANpkg{kernlab} package.
}
\description{
Spam data set from the UCI machine learning repository (\url{http://archive.ics.uci.edu/dataset/94/spambase}).
Data set collected at Hewlett-Packard Labs to classify emails as spam or non-spam.
57 variables indicate the frequency of certain words and characters in the e-mail.
The positive class is set to "spam".
}
\section{Dictionary}{

This \link{Task} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_tasks} or with the associated sugar function \code{\link[=tsk]{tsk()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("spam")
tsk("spam")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Dimensions: 4601x58
\item Properties: \dQuote{twoclass}
\item Has Missings: \code{FALSE}
\item Target: \dQuote{type}
\item Features: \dQuote{address}, \dQuote{addresses}, \dQuote{all}, \dQuote{business}, \dQuote{capitalAve}, \dQuote{capitalLong}, \dQuote{capitalTotal}, \dQuote{charDollar}, \dQuote{charExclamation}, \dQuote{charHash}, \dQuote{charRoundbracket}, \dQuote{charSemicolon}, \dQuote{charSquarebracket}, \dQuote{conference}, \dQuote{credit}, \dQuote{cs}, \dQuote{data}, \dQuote{direct}, \dQuote{edu}, \dQuote{email}, \dQuote{font}, \dQuote{free}, \dQuote{george}, \dQuote{hp}, \dQuote{hpl}, \dQuote{internet}, \dQuote{lab}, \dQuote{labs}, \dQuote{mail}, \dQuote{make}, \dQuote{meeting}, \dQuote{money}, \dQuote{num000}, \dQuote{num1999}, \dQuote{num3d}, \dQuote{num415}, \dQuote{num650}, \dQuote{num85}, \dQuote{num857}, \dQuote{order}, \dQuote{original}, \dQuote{our}, \dQuote{over}, \dQuote{parts}, \dQuote{people}, \dQuote{pm}, \dQuote{project}, \dQuote{re}, \dQuote{receive}, \dQuote{remove}, \dQuote{report}, \dQuote{table}, \dQuote{technology}, \dQuote{telnet}, \dQuote{will}, \dQuote{you}, \dQuote{your}
}
}

\references{
Dua, Dheeru, Graff, Casey (2017).
\dQuote{UCI Machine Learning Repository.}
\url{http://archive.ics.uci.edu/datasets}.
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{Task}},
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_resample_result.R
\name{as_resample_result}
\alias{as_resample_result}
\alias{as_resample_result.ResampleResult}
\alias{as_resample_result.ResultData}
\alias{as_resample_result.list}
\title{Convert to ResampleResult}
\usage{
as_resample_result(x, ...)

\method{as_resample_result}{ResampleResult}(x, ...)

\method{as_resample_result}{ResultData}(x, view = NULL, ...)

\method{as_resample_result}{list}(x, task, learners, resampling, store_backends = TRUE, ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{...}{(any)\cr
Currently not used.}

\item{view}{(\code{character()})\cr
See construction argument \code{view} of \code{\link{ResampleResult}}.}

\item{task}{(\link{Task}).}

\item{learners}{(list of trained \link{Learner}s).}

\item{resampling}{(\link{Resampling}).}

\item{store_backends}{(\code{logical(1)})\cr
If set to \code{FALSE}, the backends of the \link{Task}s provided in \code{data} are
removed.}
}
\value{
(\link{ResampleResult}).
}
\description{
Convert object to a \link{ResampleResult}.

The S3 method for \code{list} expects argument \code{x} to be a list of \link{Prediction} objects and
all other relevant objects (\link{Task}, \link{Learner}s, and instantiated \link{Resampling}) must
be provided, too.
A more flexible way to manually create a \link{ResampleResult} is implemented in \code{\link[=as_result_data]{as_result_data()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmark_grid.R
\name{benchmark_grid}
\alias{benchmark_grid}
\title{Generate a Benchmark Grid Design}
\usage{
benchmark_grid(
  tasks,
  learners,
  resamplings,
  param_values = NULL,
  paired = FALSE
)
}
\arguments{
\item{tasks}{(list of \link{Task}).}

\item{learners}{(list of \link{Learner}).}

\item{resamplings}{(list of \link{Resampling}).}

\item{param_values}{(\code{list()} | \code{NULL})\cr
If you want to try many parameter settings for learners, you can pass them through the design
which is optimized to be faster than creating learners for each setting.

A list of lists of named lists, from outer to inner:
\enumerate{
\item One list element for each \link{Learner}.
\item One list element for each hyperparameter configuration to try.
\item Named list of hyperparameter settings to set in the Learner, possibly overwriting
already set set hyperparameters in the \link{Learner}.
}}

\item{paired}{(\code{logical(1)})\cr
Set this to \code{TRUE} if the resamplings are instantiated on the tasks, i.e., the tasks and resamplings are paired.
You need to provide the same number of tasks and instantiated resamplings.}
}
\value{
(\code{\link[data.table:data.table]{data.table::data.table()}}) with the cross product of the input vectors.
}
\description{
Takes a lists of \link{Task}, a list of \link{Learner} and a list of \link{Resampling} to
generate a design in an \code{\link[=expand.grid]{expand.grid()}} fashion (a.k.a. cross join or Cartesian product).

There are two modes of operation, depending on the flag \code{paired}.
\itemize{
\item With \code{paired} set to \code{FALSE} (default), resampling strategies are not allowed to be
instantiated, and instead will be instantiated per task internally.
The only exception to this rule applies if all tasks have exactly the same row ids,
and the resamplings are all instantiated for such tasks.
The grid will be generated based on the Cartesian product of tasks, learners,
and resamplings.
Because the resamplings are instantiated on the tasks, reproducibility requires a seed
to be set \strong{before} calling this function, as this process is stochastic.
\item With \code{paired} set to \code{TRUE}, tasks and resamplings are treated as pairs.
This means that you must provide as many tasks as corresponding instantiated resamplings.
The grid will be generated based on the Cartesian product of learners and pairs.
}
}
\section{Errors and Warnings}{

\itemize{
\item \code{Mlr3WarningVaryingPredictTypes}: This warning will be thrown if the learners have different \code{predict_type}s.
}
}

\examples{
tasks = list(tsk("penguins"), tsk("sonar"))
learners = list(lrn("classif.featureless"), lrn("classif.rpart"))
resamplings = list(rsmp("cv"), rsmp("subsampling"))

# Set a seed to ensure reproducibility of the resampling instantiation
set.seed(123)
grid = benchmark_grid(tasks, learners, resamplings)
# the resamplings are now instantiated
head(grid$resampling[[1]]$instance)
print(grid)
\dontrun{
benchmark(grid)
}

# paired
learner = lrn("classif.rpart")
task1 = tsk("penguins")
task2 = tsk("german_credit")
res1 = rsmp("holdout")
res2 = rsmp("holdout")
res1$instantiate(task1)
res2$instantiate(task2)
design = benchmark_grid(list(task1, task2), learner, list(res1, res2), paired = TRUE)
print(design)

# manual construction of the grid with data.table::CJ()
grid = data.table::CJ(
 task = tasks,
  learner = learners,
  resampling = resamplings,
  sorted = FALSE
)

# manual instantiation (not suited for a fair comparison of learners!)
Map(function(task, resampling) {
  resampling$instantiate(task)
}, task = grid$task, resampling = grid$resampling)
\dontrun{
benchmark(grid)
}

}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter3/evaluation_and_benchmarking.html#sec-benchmarking}
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \CRANpkg{mlr3benchmark} for post-hoc analysis of benchmark results.
}

Other benchmark:
\code{\link{BenchmarkResult}},
\code{\link[=benchmark]{benchmark()}}
}
\concept{benchmark}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskGeneratorCassini.R
\name{mlr_task_generators_cassini}
\alias{mlr_task_generators_cassini}
\alias{TaskGeneratorCassini}
\title{Cassini Classification Task Generator}
\description{
A \link{TaskGenerator} for the cassini task in \code{\link[mlbench:mlbench.cassini]{mlbench::mlbench.cassini()}}.
}
\section{Dictionary}{

This \link{TaskGenerator} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_task_generators} or with the associated sugar function \code{\link[=tgen]{tgen()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_task_generators$get("cassini")
tgen("cassini")
}\if{html}{\out{</div>}}
}

\section{Parameters}{
\tabular{llll}{
   Id \tab Type \tab Default \tab Range \cr
   relsize1 \tab integer \tab 2 \tab \eqn{[1, \infty)}{[1, Inf)} \cr
   relsize2 \tab integer \tab 2 \tab \eqn{[1, \infty)}{[1, Inf)} \cr
   relsize3 \tab integer \tab 1 \tab \eqn{[1, \infty)}{[1, Inf)} \cr
}
}

\examples{
generator = tgen("cassini")
plot(generator, n = 200)

task = generator$generate(200)
str(task$data())
}
\seealso{
\itemize{
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=TaskGenerator]{TaskGenerators}: \link{mlr_task_generators}
\item \code{as.data.table(mlr_task_generators)}  for a table of available \link[=TaskGenerator]{TaskGenerators} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other TaskGenerator:
\code{\link{TaskGenerator}},
\code{\link{mlr_task_generators}},
\code{\link{mlr_task_generators_2dnormals}},
\code{\link{mlr_task_generators_circle}},
\code{\link{mlr_task_generators_friedman1}},
\code{\link{mlr_task_generators_moons}},
\code{\link{mlr_task_generators_peak}},
\code{\link{mlr_task_generators_simplex}},
\code{\link{mlr_task_generators_smiley}},
\code{\link{mlr_task_generators_spirals}},
\code{\link{mlr_task_generators_xor}}
}
\concept{TaskGenerator}
\section{Super class}{
\code{\link[mlr3:TaskGenerator]{TaskGenerator}} -> \code{TaskGeneratorCassini}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-TaskGeneratorCassini-initialize}{\code{TaskGeneratorCassini$new()}}
    \item \href{#method-TaskGeneratorCassini-plot}{\code{TaskGeneratorCassini$plot()}}
    \item \href{#method-TaskGeneratorCassini-clone}{\code{TaskGeneratorCassini$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="TaskGenerator" data-id="format"><a href='../../mlr3/html/TaskGenerator.html#method-TaskGenerator-format'><code>TaskGenerator$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="TaskGenerator" data-id="generate"><a href='../../mlr3/html/TaskGenerator.html#method-TaskGenerator-generate'><code>TaskGenerator$generate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="TaskGenerator" data-id="print"><a href='../../mlr3/html/TaskGenerator.html#method-TaskGenerator-print'><code>TaskGenerator$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGeneratorCassini-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGeneratorCassini-initialize}{}}}
\subsection{\code{TaskGeneratorCassini$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGeneratorCassini$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGeneratorCassini-plot"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGeneratorCassini-plot}{}}}
\subsection{\code{TaskGeneratorCassini$plot()}}{
  Creates a simple plot of generated data.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGeneratorCassini$plot(n = 200L, pch = 19L, ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{(\code{integer(1)})\cr
Number of samples to draw for the plot. Default is 200.}
      \item{\code{pch}}{(\code{integer(1)})\cr
Point char. Passed to \code{\link[=plot]{plot()}}.}
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=plot]{plot()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskGeneratorCassini-clone"></a>}}
\if{latex}{\out{\hypertarget{method-TaskGeneratorCassini-clone}{}}}
\subsection{\code{TaskGeneratorCassini$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskGeneratorCassini$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureSimple.R
\name{mlr_measures_classif.mauc_au1p}
\alias{mlr_measures_classif.mauc_au1p}
\title{Multiclass AUC Scores}
\description{
Measure to compare true observed labels with predicted
probabilities
in multiclass classification tasks.
}
\details{
Multiclass AUC measures.
\itemize{
\item \emph{AUNU}: AUC of each class against the rest, using the uniform class
distribution. Computes the AUC treating a \code{c}-dimensional classifier
as \code{c} two-dimensional 1-vs-rest classifiers, where classes are assumed to have
uniform distribution, in order to have a measure which is independent
of class distribution change (Fawcett 2001).
\item \emph{AUNP}: AUC of each class against the rest, using the a-priori class
distribution. Computes the AUC treating a \code{c}-dimensional classifier as \code{c}
two-dimensional 1-vs-rest classifiers, taking into account the prior probability of
each class (Fawcett 2001).
\item \emph{AU1U}: AUC of each class against each other, using the uniform class
distribution. Computes something like the AUC of \code{c(c - 1)} binary classifiers
(all possible pairwise combinations). See Hand (2001) for details.
\item \emph{AU1P}: AUC of each class against each other, using the a-priori class
distribution. Computes something like AUC of \code{c(c - 1)} binary classifiers
while considering the a-priori distribution of the classes as suggested
in Ferri (2009). Note we deviate from the definition in
Ferri (2009) by a factor of \code{c}.
\item \emph{MU}: Multiclass AUC as defined in Kleinman and Page (2019).
This measure is an average of the pairwise AUCs between all classes.
The measure was tested against the Python implementation by \href{https://github.com/kleimanr/auc_mu}{Ross Kleinman}.
}
}
\note{
The score function calls \code{\link[mlr3measures:mauc_au1p]{mlr3measures::mauc_au1p()}} from package \CRANpkg{mlr3measures}.

If the measure is undefined for the input, \code{NaN} is returned.
This can be customized by setting the field \code{na_value}.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("classif.mauc_au1p")
msr("classif.mauc_au1p")
}\if{html}{\out{</div>}}
}

\section{Parameters}{

Empty ParamSet
}

\section{Meta Information}{


\itemize{
\item Type: \code{"classif"}
\item Range: \eqn{[0, 1]}{[0, 1]}
\item Minimize: \code{FALSE}
\item Required prediction: \code{prob}
}

}

\seealso{
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}

\code{as.data.table(mlr_measures)} for a complete table of all (also dynamically created) \link{Measure} implementations.

Other classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.auc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.bbrier}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.dor}},
\code{\link{mlr_measures_classif.fbeta}},
\code{\link{mlr_measures_classif.fdr}},
\code{\link{mlr_measures_classif.fn}},
\code{\link{mlr_measures_classif.fnr}},
\code{\link{mlr_measures_classif.fomr}},
\code{\link{mlr_measures_classif.fp}},
\code{\link{mlr_measures_classif.fpr}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}},
\code{\link{mlr_measures_classif.mcc}},
\code{\link{mlr_measures_classif.npv}},
\code{\link{mlr_measures_classif.ppv}},
\code{\link{mlr_measures_classif.prauc}},
\code{\link{mlr_measures_classif.precision}},
\code{\link{mlr_measures_classif.recall}},
\code{\link{mlr_measures_classif.sensitivity}},
\code{\link{mlr_measures_classif.specificity}},
\code{\link{mlr_measures_classif.tn}},
\code{\link{mlr_measures_classif.tnr}},
\code{\link{mlr_measures_classif.tp}},
\code{\link{mlr_measures_classif.tpr}}

Other multiclass classification measures:
\code{\link{mlr_measures_classif.acc}},
\code{\link{mlr_measures_classif.bacc}},
\code{\link{mlr_measures_classif.ce}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_classif.logloss}},
\code{\link{mlr_measures_classif.mauc_au1u}},
\code{\link{mlr_measures_classif.mauc_aunp}},
\code{\link{mlr_measures_classif.mauc_aunu}},
\code{\link{mlr_measures_classif.mauc_mu}},
\code{\link{mlr_measures_classif.mbrier}},
\code{\link{mlr_measures_classif.mcc}}
}
\concept{classification measures}
\concept{multiclass classification measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_resamplings.R
\name{mlr_resamplings}
\alias{mlr_resamplings}
\title{Dictionary of Resampling Strategies}
\format{
\link[R6:R6Class]{R6::R6Class} object inheriting from \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}
\description{
A simple \link[mlr3misc:Dictionary]{mlr3misc::Dictionary} storing objects of class \link{Resampling}.
Each resampling has an associated help page, see \code{mlr_resamplings_[id]}.

This dictionary can get populated with additional resampling strategies by add-on packages.

For a more convenient way to retrieve and construct resampling strategies, see \code{\link[=rsmp]{rsmp()}}/\code{\link[=rsmps]{rsmps()}}.
}
\section{Methods}{

See \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}.
}

\section{S3 methods}{

\itemize{
\item \code{as.data.table(dict, ..., objects = FALSE)}\cr
\link[mlr3misc:Dictionary]{mlr3misc::Dictionary} -> \code{\link[data.table:data.table]{data.table::data.table()}}\cr
Returns a \code{\link[data.table:data.table]{data.table::data.table()}} with columns "key", "label", "params", and "iters".
If \code{objects} is set to \code{TRUE}, the constructed objects are returned in the list column named \code{object}.
}
}

\examples{
as.data.table(mlr_resamplings)
mlr_resamplings$get("cv")
rsmp("subsampling")
}
\seealso{
Sugar functions: \code{\link[=rsmp]{rsmp()}}, \code{\link[=rsmps]{rsmps()}}

Other Dictionary:
\code{\link{mlr_learners}},
\code{\link{mlr_measures}},
\code{\link{mlr_task_generators}},
\code{\link{mlr_tasks}}

Other Resampling:
\code{\link{Resampling}},
\code{\link{mlr_resamplings_bootstrap}},
\code{\link{mlr_resamplings_custom}},
\code{\link{mlr_resamplings_custom_cv}},
\code{\link{mlr_resamplings_cv}},
\code{\link{mlr_resamplings_holdout}},
\code{\link{mlr_resamplings_insample}},
\code{\link{mlr_resamplings_loo}},
\code{\link{mlr_resamplings_repeated_cv}},
\code{\link{mlr_resamplings_subsampling}}
}
\concept{Dictionary}
\concept{Resampling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CallbackResample.R
\name{CallbackResample}
\alias{CallbackResample}
\title{Resample Callback}
\description{
Specialized \link[mlr3misc:Callback]{mlr3misc::Callback} to customize the behavior of \code{\link[=resample]{resample()}} and \code{\link[=benchmark]{benchmark()}} in mlr3.
For example, callbacks can be used to extract information from models on the worker
or to store intermediate results to disk.
The \code{\link[=callback_resample]{callback_resample()}} function is used to create instances of this class.
Predefined callbacks are stored in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_callbacks}
and can be retrieved with \code{\link[=clbk]{clbk()}}.
For more information on callbacks, see the \code{\link[=callback_resample]{callback_resample()}} documentation.
}
\section{Super class}{
\code{\link[mlr3misc:Callback]{mlr3misc::Callback}} -> \code{CallbackResample}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{on_resample_begin}}{(\verb{function()})\cr
Stage called at the beginning of the resampling iteration.
Called in \code{workhorse()} (internal).}

    \item{\code{on_resample_before_train}}{(\verb{function()})\cr
Stage called before training the learner.
Called in \code{workhorse()} (internal).}

    \item{\code{on_resample_before_predict}}{(\verb{function()})\cr
Stage called before predicting.
Called in \code{workhorse()} (internal).}

    \item{\code{on_resample_end}}{(\verb{function()})\cr
Stage called at the end of the resample iteration.
Called in \code{workhorse()} (internal).}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-CallbackResample-clone}{\code{CallbackResample$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3misc" data-topic="Callback" data-id="call"><a href='../../mlr3misc/html/Callback.html#method-Callback-call'><code>mlr3misc::Callback$call()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3misc" data-topic="Callback" data-id="format"><a href='../../mlr3misc/html/Callback.html#method-Callback-format'><code>mlr3misc::Callback$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3misc" data-topic="Callback" data-id="help"><a href='../../mlr3misc/html/Callback.html#method-Callback-help'><code>mlr3misc::Callback$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3misc" data-topic="Callback" data-id="initialize"><a href='../../mlr3misc/html/Callback.html#method-Callback-initialize'><code>mlr3misc::Callback$initialize()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3misc" data-topic="Callback" data-id="print"><a href='../../mlr3misc/html/Callback.html#method-Callback-print'><code>mlr3misc::Callback$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CallbackResample-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CallbackResample-clone}{}}}
\subsection{\code{CallbackResample$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{CallbackResample$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LearnerClassif.R
\name{LearnerClassif}
\alias{LearnerClassif}
\title{Classification Learner}
\description{
This Learner specializes \link{Learner} for classification problems:
\itemize{
\item \code{task_type} is set to \code{"classif"}.
\item Creates \link{Prediction}s of class \link{PredictionClassif}.
\item Possible values for \code{predict_types} are:
\itemize{
\item \code{"response"}: Predicts a class label for each observation in the test set.
\item \code{"prob"}: Predicts the posterior probability for each class for each observation in the test set.
}
\item Additional learner properties include:
\itemize{
\item \code{"twoclass"}: The learner works on binary classification problems.
\item \code{"multiclass"}: The learner works on multiclass classification problems.
}
}

Predefined learners can be found in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_learners}.
Essential classification learners can be found in this dictionary after loading \CRANpkg{mlr3learners}.
Additional learners are implemented in the Github package \url{https://github.com/mlr-org/mlr3extralearners}.
}
\examples{
# get all classification learners from mlr_learners:
lrns = mlr_learners$mget(mlr_learners$keys("^classif"))
names(lrns)

# get a specific learner from mlr_learners:
lrn = lrn("classif.rpart")
print(lrn)

# train the learner:
task = tsk("penguins")
lrn$train(task, 1:200)

# predict on new observations:
lrn$predict(task, 201:344)$confusion
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-learners}
\item Package \CRANpkg{mlr3learners} for a solid collection of essential learners.
\item Package \href{https://github.com/mlr-org/mlr3extralearners}{mlr3extralearners} for more learners.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Learner]{Learners}: \link{mlr_learners}
\item \code{as.data.table(mlr_learners)} for a table of available \link[=Learner]{Learners} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3pipelines} to combine learners with pre- and postprocessing steps.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
\item \CRANpkg{mlr3tuning} for tuning of hyperparameters, \CRANpkg{mlr3tuningspaces}
for established default tuning spaces.
}

Other Learner:
\code{\link{Learner}},
\code{\link{LearnerRegr}},
\code{\link{mlr_learners}},
\code{\link{mlr_learners_classif.debug}},
\code{\link{mlr_learners_classif.featureless}},
\code{\link{mlr_learners_classif.rpart}},
\code{\link{mlr_learners_regr.debug}},
\code{\link{mlr_learners_regr.featureless}},
\code{\link{mlr_learners_regr.rpart}}
}
\concept{Learner}
\section{Super class}{
\code{\link[mlr3:Learner]{Learner}} -> \code{LearnerClassif}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-LearnerClassif-initialize}{\code{LearnerClassif$new()}}
    \item \href{#method-LearnerClassif-predict_newdata_fast}{\code{LearnerClassif$predict_newdata_fast()}}
    \item \href{#method-LearnerClassif-clone}{\code{LearnerClassif$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="base_learner"><a href='../../mlr3/html/Learner.html#method-Learner-base_learner'><code>Learner$base_learner()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="configure"><a href='../../mlr3/html/Learner.html#method-Learner-configure'><code>Learner$configure()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="encapsulate"><a href='../../mlr3/html/Learner.html#method-Learner-encapsulate'><code>Learner$encapsulate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="format"><a href='../../mlr3/html/Learner.html#method-Learner-format'><code>Learner$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="help"><a href='../../mlr3/html/Learner.html#method-Learner-help'><code>Learner$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict"><a href='../../mlr3/html/Learner.html#method-Learner-predict'><code>Learner$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict_newdata"><a href='../../mlr3/html/Learner.html#method-Learner-predict_newdata'><code>Learner$predict_newdata()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="print"><a href='../../mlr3/html/Learner.html#method-Learner-print'><code>Learner$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="reset"><a href='../../mlr3/html/Learner.html#method-Learner-reset'><code>Learner$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="selected_features"><a href='../../mlr3/html/Learner.html#method-Learner-selected_features'><code>Learner$selected_features()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="train"><a href='../../mlr3/html/Learner.html#method-Learner-train'><code>Learner$train()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassif-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassif-initialize}{}}}
\subsection{\code{LearnerClassif$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassif$new(
  id,
  task_type = "classif",
  param_set = ps(),
  predict_types = "response",
  feature_types = character(),
  properties = character(),
  packages = character(),
  label = NA_character_,
  man = NA_character_
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier for the new instance.}
      \item{\code{task_type}}{(\code{character(1)})\cr
Type of task, e.g. \code{"regr"} or \code{"classif"}.
Must be an element of \link[=mlr_reflections]{mlr_reflections$task_types$type}.}
      \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}
      \item{\code{predict_types}}{(\code{character()})\cr
Supported predict types. Must be a subset of \code{\link[=mlr_reflections]{mlr_reflections$learner_predict_types}}.}
      \item{\code{feature_types}}{(\code{character()})\cr
Feature types the learner operates on. Must be a subset of \code{\link[=mlr_reflections]{mlr_reflections$task_feature_types}}.}
      \item{\code{properties}}{(\code{character()})\cr
Set of properties of the \link{Learner}.
Must be a subset of \code{\link[=mlr_reflections]{mlr_reflections$learner_properties}}.
The following properties are currently standardized and understood by learners in \CRANpkg{mlr3}:
\itemize{
\item \code{"missings"}: The learner can handle missing values in the data.
\item \code{"weights"}: The learner supports observation weights.
\item \code{"offset"}: The learner can incorporate offset values to adjust predictions.
\item \code{"importance"}: The learner supports extraction of importance scores, i.e. comes with an \verb{$importance()} extractor function (see section on optional extractors in \link{Learner}).
\item \code{"selected_features"}: The learner supports extraction of the set of selected features, i.e. comes with a \verb{$selected_features()} extractor function (see section on optional extractors in \link{Learner}).
\item \code{"oob_error"}: The learner supports extraction of estimated out of bag error, i.e. comes with a \code{oob_error()} extractor function (see section on optional extractors in \link{Learner}).
\item \code{"validation"}: The learner can use a validation task during training.
\item \code{"internal_tuning"}: The learner is able to internally optimize hyperparameters (those are also tagged with \code{"internal_tuning"}).
\item \code{"marshal"}: To save learners with this property, you need to call \verb{$marshal()} first.
If a learner is in a marshaled state, you call first need to call \verb{$unmarshal()} to use its model, e.g. for prediction.
\item \code{"hotstart_forward"}: The learner supports to hotstart a model forward.
\item \code{"hotstart_backward"}: The learner supports hotstarting a model backward.
\item `"featureless": The learner does not use features.
}}
      \item{\code{packages}}{(\code{character()})\cr
Set of required packages.
A warning is signaled by the constructor if at least one of the packages is not installed,
but loaded (not attached) later on-demand via \code{\link[=requireNamespace]{requireNamespace()}}.}
      \item{\code{label}}{(\code{character(1)})\cr
Label for the new instance.}
      \item{\code{man}}{(\code{character(1)})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
The referenced help package can be opened via method \verb{$help()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassif-predict_newdata_fast"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassif-predict_newdata_fast}{}}}
\subsection{\code{LearnerClassif$predict_newdata_fast()}}{
  Predicts outcomes for new data in \code{newdata} using the model fitted during \verb{$train()}.
This method is faster than \verb{$predict_newdata()} as it skips assertions,
type conversions, encapsulation, and logging.
The learner must be trained before calling this method.

Unlike \verb{$predict_newdata()}, this method does not return a \link{Prediction} object.
Instead, it returns a list with either a \code{"response"} or \code{"prob"} element, depending on the prediction type.

Note that \code{state$predict_time} and \code{state$log} will remain empty after using this method.
Some learners may not support this method and may fail when it is called.
Prefer \verb{$predict_newdata()} unless performance is critical.

If the model was trained via \code{\link[=resample]{resample()}} or \code{\link[=benchmark]{benchmark()}},
you must pass the associated task object stored in the corresponding \link{ResampleResult} or \link{BenchmarkResult}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassif$predict_newdata_fast(newdata, task = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{newdata}}{\code{\link[data.table:data.table]{data.table::data.table()}}\cr
New data to predict on.}
      \item{\code{task}}{(\link{Task}).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{list()} with elements \code{"response"} or \code{"prob"} depending on the predict type.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassif-clone"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassif-clone}{}}}
\subsection{\code{LearnerClassif$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassif$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/partition.R
\name{partition}
\alias{partition}
\alias{partition.Task}
\title{Manually Partition into Training, Test and Validation Set}
\usage{
partition(task, ratio = 0.67)

\method{partition}{Task}(task, ratio = 0.67)
}
\arguments{
\item{task}{(\link{Task})\cr
Task to operate on.}

\item{ratio}{(\code{numeric()})\cr
Ratio of observations to put into the training set.
If a 2 element vector is provided, the first element is the ratio for the training set,
the second element is the ratio for the test set.
The validation set will contain the remaining observations.}
}
\description{
Creates a split of the row ids of a \link{Task} into a training and a test set, and optionally a validation set.
}
\examples{
# regression task partitioned into training and test set
task = tsk("california_housing")
split = partition(task, ratio = 0.5)
data = data.frame(
  y = c(task$truth(split$train), task$truth(split$test)),
  split = rep(c("train", "predict"), lengths(split[c("train", "test")]))
)
boxplot(y ~ split, data = data)

# classification task partitioned into training, test and validation set
task = tsk("pima")
split = partition(task, c(0.66, 0.14))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/score_roc_measures.R
\name{score_roc_measures}
\alias{score_roc_measures}
\title{Calculate ROC Measures}
\usage{
score_roc_measures(pred)
}
\arguments{
\item{pred}{(\link{PredictionClassif})\cr
The prediction object.}
}
\value{
\code{list()}\cr
A list containing two elements \code{confusion_matrix} which is the 2 times 2 confusion matrix
of absolute frequencies and \code{measures}, a list of the above mentioned measures.
}
\description{
Calculate a set of roc performance measures based on the confusion matrix.
\itemize{
\item \code{tpr} True positive rate (Sensitivity, Recall)
\item \code{fpr} False positive rate (Fall-out)
\item \code{fnr} False negative rate (Miss rate)
\item \code{tnr} True negative rate (Specificity)
\item \code{ppv} Positive predictive value (Precision)
\item \code{fomr} False omission rate
\item \code{lrp} Positive likelihood ratio (LR+)
\item \code{fdr} False discovery rate
\item \code{npv} Negative predictive value
\item \code{acc} Accuracy
\item \code{lrm} Negative likelihood ratio (LR-)
\item \code{dor} Diagnostic odds ratio
}
}
\examples{
learner = lrn("classif.rpart", predict_type = "prob")
splits = partition(task = tsk("pima"), ratio = 0.7)
task = tsk("pima")
learner$train(task)
pred = learner$predict(task)
score_roc_measures(pred)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.Learner}
\alias{predict.Learner}
\title{Predict Method for Learners}
\usage{
\method{predict}{Learner}(object, newdata, predict_type = NULL, ...)
}
\arguments{
\item{object}{(\link{Learner})\cr
Any \link{Learner}.}

\item{newdata}{(\code{\link[=data.frame]{data.frame()}})\cr
New data to predict on.}

\item{predict_type}{(\code{character(1)})\cr
The predict type to return.
Set to \verb{<Prediction>} to retrieve the complete \link{Prediction} object.
If set to \code{NULL} (default), the first predict type for the respective class of the \link{Learner}
as stored in \link{mlr_reflections} is used.}

\item{...}{(any)\cr
Hyperparameters to pass down to the \link{Learner}.}
}
\description{
Extends the generic \code{\link[stats:predict]{stats::predict()}} with a method for \link{Learner}.
Note that this function is intended as glue code to be used in third party packages.
We recommend to work with the \link{Learner} directly, i.e. calling \code{learner$predict()}
or \code{learner$predict_newdata()} directly.

Performs the following steps:
\itemize{
\item Sets additional hyperparameters passed to this function.
\item Creates a \link{Prediction} object by calling \code{learner$predict_newdata()}.
\item Returns (subset of) \link{Prediction}.
}
}
\examples{
task = tsk("spam")

learner = lrn("classif.rpart", predict_type = "prob")
learner$train(task)
predict(learner, task$data(1:3), predict_type = "response")
predict(learner, task$data(1:3), predict_type = "prob")
predict(learner, task$data(1:3), predict_type = "<Prediction>")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CallbackResample.R
\name{assert_resample_callback}
\alias{assert_resample_callback}
\alias{assert_resample_callbacks}
\title{Assertions for Callbacks}
\usage{
assert_resample_callback(callback, null_ok = FALSE)

assert_resample_callbacks(callbacks, null_ok = FALSE)
}
\arguments{
\item{callback}{(\link{CallbackResample}).}

\item{null_ok}{(\code{logical(1)})\cr
If \code{TRUE}, \code{NULL} is allowed.}

\item{callbacks}{(list of \link{CallbackResample}).}
}
\value{
\link{CallbackResample} | List of \link{CallbackResample}s.
}
\description{
Assertions for \link{CallbackResample} class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Measure.R
\name{Measure}
\alias{Measure}
\title{Measure Class}
\description{
This is the abstract base class for measures like \link{MeasureClassif} and \link{MeasureRegr}.

Measures are classes tailored around two functions doing the work:
\enumerate{
\item A function \verb{$score()} which quantifies the performance on a \link{Prediction} object, so a set
of predicted observation via a scalar number -- usually an aggregate of losses on the contained observations,
by comparing the truth and prediction columns in the prediction object.
\item A function \verb{$aggregator()} which combines multiple performance scores returned by
\verb{$score()} obtained in different resampling iterations to a scalar performance value associated
with the complete resampling -- usually by averaging or summing.
}

In addition to these two functions, meta-information about the performance measure is stored.

Predefined measures are stored in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures},
e.g. \code{\link[=mlr_measures_classif.auc]{classif.auc}} or \code{\link[=mlr_measures_time_train]{time_train}}.
Many of the measures in \pkg{mlr3} are implemented in \CRANpkg{mlr3measures} as ordinary functions.

A guide on how to extend \CRANpkg{mlr3} with custom measures can be found in the
\href{https://mlr3book.mlr-org.com}{mlr3book}.
}
\section{Inheriting}{

For some measures (such as confidence intervals from \code{mlr3inferr}) it is necessary that a measure
returns more than one value.
In such cases it is necessary to overwrite the public methods \verb{$aggregate()} and/or \verb{$score()}
to return a named \code{numeric()}
where at least one of its names corresponds to the \code{id} of the measure itself.
}

\section{Weights}{


Many measures support observation weights, indicated by their property \code{"weights"}.
The weights are stored in the \link{Task} where the column role \code{weights_measure} needs to be assigned
to a single numeric column.
The weights are automatically used if found in the task,
this can be disabled by setting the field \code{use_weights} to \code{"ignore"}.
See the description of \code{use_weights} for more information.

If the measure is set-up to use weights but the task does not have a designated \code{weights_measure} column,
an unweighted version is calculated instead.
The weights do not necessarily need to sum up to 1, they are normalized by the measure if necessary.

Most measures are so-called decomposable loss functions where a point-wise loss is computed
and then either mean-aggregated or summed over the test set.
For measures that do mean-aggregation, weights are typically used to compute the weighted mean,
which normalizes weights to sum to 1.
Measures that use sum-aggregation do not normalize weights and instead multiply individual losses
with the given weights.
See the documentation of specific measures for more details.
}

\section{Missing Values during Scoring}{


Many measurements cannot be calculated if the test set or predictions are unfortunate,
for example because a denominator is 0.
This typically occurs during (binary) classification if some entries of the confusion matrix are 0.
For this reason, many measures which originate in \CRANpkg{mlr3measures} allow to change the
default missing value (\code{NaN}) via the field \code{na_value}.

If you encounter missing values in a compound object like a \link{ResampleResult} or \link{BenchmarkResult}
during scoring or aggregating, simply removing iterations with missing values is statistically
arguable (but technically possible by providing a custom aggregation function which handles
missing values, e.g. \code{function(x) mean(x, na.rm = TRUE)}).
Instead, consider stratification on the target of the \link{Task} to work around missing values.
Switching to micro averaging in the \link{Resampling} can also be a solution here.
}

\examples{

## ------------------------------------------------
## Method `Measure$score()`
## ------------------------------------------------

task = tsk("penguins")
learner = lrn("classif.rpart")$train(task)
prediction = learner$predict(task)
msr("classif.ce")$score(prediction)

## ------------------------------------------------
## Method `Measure$aggregate()`
## ------------------------------------------------

task = tsk("penguins")
learner = lrn("classif.rpart")
rr = resample(task, learner, rsmp("holdout"))
msr("classif.ce")$aggregate(rr)

## ------------------------------------------------
## Method `Measure$obs_loss()`
## ------------------------------------------------

task = tsk("penguins")
learner = lrn("classif.rpart")
learner$train(task)
prediction = learner$predict(task)
msr("classif.ce")$obs_loss(prediction)
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-eval}
\item Package \CRANpkg{mlr3measures} for the scoring functions.
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}
\code{as.data.table(mlr_measures)} for a table of available \link[=Measure]{Measures} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Measure:
\code{\link{MeasureClassif}},
\code{\link{MeasureRegr}},
\code{\link{MeasureSimilarity}},
\code{\link{mlr_measures}},
\code{\link{mlr_measures_aic}},
\code{\link{mlr_measures_bic}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_debug_classif}},
\code{\link{mlr_measures_elapsed_time}},
\code{\link{mlr_measures_internal_valid_score}},
\code{\link{mlr_measures_oob_error}},
\code{\link{mlr_measures_regr.pinball}},
\code{\link{mlr_measures_regr.rqr}},
\code{\link{mlr_measures_regr.rsq}},
\code{\link{mlr_measures_selected_features}}
}
\concept{Measure}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{predict_sets}}{(\code{character()})\cr
During \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}, a \link{Learner} can predict on multiple sets.
Per default, a learner only predicts observations in the test set (\code{predict_sets == "test"}).
To change this behavior, set \code{predict_sets} to a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
The \code{"train"} predict set contains the train ids from the resampling. This means that if a learner does validation and
sets \verb{$validate} to a ratio (creating the validation data from the training data), the train predictions
will include the predictions for the validation data.
Each set yields a separate \link{Prediction} object.
Those can be combined via getters in \link{ResampleResult}/\link{BenchmarkResult}, or \link{Measure}s can be configured
to operate on specific subsets of the calculated prediction sets.}

    \item{\code{hash}}{(\code{character(1)})\cr
Hash (unique identifier) for this object.
The hash is calculated based on the id, the parameter settings, predict sets and the \verb{$score},
\verb{$average}, \verb{$aggregator}, \verb{$obs_loss}, \verb{$trafo} method.
Measure can define additional fields to be included in the hash by setting the field \verb{$.extra_hash}.}

    \item{\code{properties}}{(\code{character()})\cr
Properties of this measure.}

    \item{\code{average}}{(\code{character(1)})\cr
Method for aggregation:
\itemize{
\item \code{"micro"}:
All predictions from multiple resampling iterations are first combined into a single \link{Prediction} object.
Next, the scoring function of the measure is applied on this combined object, yielding a single numeric score.
\item \code{"macro"}:
The scoring function is applied on the \link{Prediction} object of each resampling iterations,
each yielding a single numeric score.
Next, the scores are combined with the \code{aggregator} function to a single numerical score.
\item \code{"macro_weighted"}:
The scoring function is applied on the \link{Prediction} object of each resampling iterations,
each yielding a single numeric score.
Next, the scores are combined with the \code{aggregator} function to a single numerical score.
The scores are weighted by the total sample weights (if present, and if \verb{$use_weights} is set to \code{"use"}),
or the number of samples in each resampling iteration.
\item \code{"custom"}:
The measure comes with a custom aggregation method which directly operates on a \link{ResampleResult}.
}}

    \item{\code{aggregator}}{(\verb{function()})\cr
Function to aggregate scores computed on different resampling iterations.}

    \item{\code{use_weights}}{(\code{character(1)})\cr
How to handle weights.
Settings are \code{"use"}, \code{"ignore"}, and \code{"error"}.
\itemize{
\item \code{"use"}: Weights are used automatically if found in the task, as supported by the measure.
\item \code{"ignore"}: Weights are ignored.
\item \code{"error"}: throw an error if weights are present in the training \code{Task}.
}

For measures with the property \code{"weights"}, this is initialized as \code{"use"}.
For measures with the property \code{"requires_no_prediction"}, this is initialized as \code{"ignore"}.
For measures that have neither of the properties, this is initialized as \code{"error"}.
The latter behavior is to avoid cases where a user erroneously assumes that a measure
supports weights when it does not.
For measures that do not support weights,
\code{use_weights} needs to be set to \code{"ignore"} if tasks with weights should be handled (by dropping the weights).}

    \item{\code{id}}{(\code{character(1)})\cr
Identifier of the object.
Used in tables, plot and text output.}

    \item{\code{label}}{(\code{character(1)})\cr
Label for this object.
Can be used in tables, plot and text output instead of the ID.}

    \item{\code{task_type}}{(\code{character(1)})\cr
Task type, e.g. \code{"classif"} or \code{"regr"}.

For a complete list of possible task types (depending on the loaded packages),
see \code{\link[=mlr_reflections]{mlr_reflections$task_types$type}}.}

    \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}

    \item{\code{predict_type}}{(\code{character(1)})\cr
Required predict type of the \link{Learner}.}

    \item{\code{check_prerequisites}}{(\code{character(1)})\cr
How to proceed if one of the following prerequisites is not met:
\itemize{
\item wrong predict type (e.g., probabilities required, but only labels available).
\item wrong predict set (e.g., learner predicted on training set, but predictions of test set required).
\item task properties not satisfied (e.g., binary classification measure on multiclass task).
}

Possible values are \code{"ignore"} (just return \code{NaN}) and \code{"warn"} (default,
raise a warning before returning \code{NaN}).}

    \item{\code{task_properties}}{(\code{character()})\cr
Required properties of the \link{Task}.}

    \item{\code{range}}{(\code{numeric(2)})\cr
Lower and upper bound of possible performance scores.}

    \item{\code{minimize}}{(\code{logical(1)})\cr
If \code{TRUE}, good predictions correspond to small values of performance scores.}

    \item{\code{packages}}{(\code{character(1)})\cr
Set of required packages.
These packages are loaded, but not attached.}

    \item{\code{man}}{(\code{character(1)} | \code{NULL})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
Defaults to \code{NA}, but can be set by child classes.}

    \item{\code{trafo}}{(\code{list()} | \code{NULL})\cr
\code{NULL} or a list with two elements:
\itemize{
\item \code{fn}: the transformation function applied after aggregating
observation-wise losses (e.g. \code{sqrt} for RMSE)
\item \code{deriv}: The derivative of the \code{fn}.
}}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Measure-initialize}{\code{Measure$new()}}
    \item \href{#method-Measure-format}{\code{Measure$format()}}
    \item \href{#method-Measure-print}{\code{Measure$print()}}
    \item \href{#method-Measure-help}{\code{Measure$help()}}
    \item \href{#method-Measure-score}{\code{Measure$score()}}
    \item \href{#method-Measure-aggregate}{\code{Measure$aggregate()}}
    \item \href{#method-Measure-obs_loss}{\code{Measure$obs_loss()}}
    \item \href{#method-Measure-clone}{\code{Measure$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Measure-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Measure-initialize}{}}}
\subsection{\code{Measure$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.

Note that this object is typically constructed via a derived classes, e.g. \link{MeasureClassif} or \link{MeasureRegr}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Measure$new(
  id,
  task_type = NA,
  param_set = ps(),
  range = c(-Inf, Inf),
  minimize = NA,
  average = "macro",
  aggregator = NULL,
  properties = character(),
  predict_type = "response",
  predict_sets = "test",
  task_properties = character(),
  packages = character(),
  label = NA_character_,
  man = NA_character_,
  trafo = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier for the new instance.}
      \item{\code{task_type}}{(\code{character(1)})\cr
Type of task, e.g. \code{"regr"} or \code{"classif"}.
Must be an element of \link[=mlr_reflections]{mlr_reflections$task_types$type}.}
      \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}
      \item{\code{range}}{(\code{numeric(2)})\cr
Feasible range for this measure as \code{c(lower_bound, upper_bound)}.
Both bounds may be infinite.}
      \item{\code{minimize}}{(\code{logical(1)})\cr
Set to \code{TRUE} if good predictions correspond to small values,
and to \code{FALSE} if good predictions correspond to large values.
If set to \code{NA} (default), tuning this measure is not possible.}
      \item{\code{average}}{(\code{character(1)})\cr
How to average multiple \link{Prediction}s from a \link{ResampleResult}.

The default, \code{"macro"}, calculates the individual performances scores for each \link{Prediction} and then uses the
function defined in \verb{$aggregator} to average them to a single number.

\code{"macro_weighted"} is similar to \code{"macro"}, but uses weighted averages.
Weights are taken from the \code{weights_measure} column of the resampled \link{Task} if present.
Note that \code{"macro_weighted"} can differ from \code{"macro"} even if no weights are present or if \verb{$use_weights} is set to \code{"ignore"},
since then aggregation is done using \emph{uniform sample weights}, which result in non-uniform weights for \link{Prediction}s if they contain different
numbers of samples.

If set to \code{"micro"}, the individual \link{Prediction} objects are first combined into a single new \link{Prediction} object which is then used to assess the performance.
The function in \verb{$aggregator} is not used in this case.}
      \item{\code{aggregator}}{(\verb{function()} | \code{NULL})\cr
Function to aggregate over multiple iterations. The role of this function depends on
the value of field \code{"average"}:
\itemize{
\item \code{"macro"}: A numeric vector of scores (one per iteration) is passed.
The aggregate function defaults to \code{\link[=mean]{mean()}} in this case.
\item \code{"micro"}: The \code{aggregator} function is not used.
Instead, predictions from multiple iterations are first combined and then
scored in one go.
\item \code{"custom"}: A \link{ResampleResult} is passed to the aggregate function.
}}
      \item{\code{properties}}{(\code{character()})\cr
Properties of the measure.
Must be a subset of \link[=mlr_reflections]{mlr_reflections$measure_properties}.
Supported by \code{mlr3}:
\itemize{
\item \code{"requires_task"} (requires the complete \link{Task}),
\item \code{"requires_learner"} (requires the trained \link{Learner}),
\item \code{"requires_model"} (requires the trained \link{Learner}, including the fitted model),
\item \code{"requires_train_set"} (requires the training indices from the \link{Resampling}),
\item \code{"na_score"} (the measure is expected to occasionally return \code{NA} or \code{NaN}),
\item \code{"weights"} (support weighted scoring using sample weights from task, column role \code{weights_measure}), and
\item \code{"primary_iters"} (the measure explicitly handles resamplings
that only use a subset of their iterations for the point estimate)
\item \code{"requires_no_prediction"} (No prediction is required;
This usually means that the measure extracts some information from the learner state.).
}}
      \item{\code{predict_type}}{(\code{character(1)})\cr
Required predict type of the \link{Learner}.
Possible values are stored in \link[=mlr_reflections]{mlr_reflections$learner_predict_types}.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
      \item{\code{task_properties}}{(\code{character()})\cr
Required task properties, see \link{Task}.}
      \item{\code{packages}}{(\code{character()})\cr
Set of required packages.
A warning is signaled by the constructor if at least one of the packages is not installed,
but loaded (not attached) later on-demand via \code{\link[=requireNamespace]{requireNamespace()}}.}
      \item{\code{label}}{(\code{character(1)})\cr
Label for the new instance.}
      \item{\code{man}}{(\code{character(1)})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
The referenced help package can be opened via method \verb{$help()}.}
      \item{\code{trafo}}{(\code{list()} | \code{NULL})\cr
An optional list with two elements, containing the transformation \code{"fn"} and its derivative \code{"deriv"}.
The transformation function is the function that is applied after aggregating the pointwise losses, i.e.
this requires an \verb{$obs_loss} to be present. An example is \code{sqrt} for RMSE.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Measure-format"></a>}}
\if{latex}{\out{\hypertarget{method-Measure-format}{}}}
\subsection{\code{Measure$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Measure$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Measure-print"></a>}}
\if{latex}{\out{\hypertarget{method-Measure-print}{}}}
\subsection{\code{Measure$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Measure$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Measure-help"></a>}}
\if{latex}{\out{\hypertarget{method-Measure-help}{}}}
\subsection{\code{Measure$help()}}{
  Opens the corresponding help page referenced by field \verb{$man}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Measure$help()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Measure-score"></a>}}
\if{latex}{\out{\hypertarget{method-Measure-score}{}}}
\subsection{\code{Measure$score()}}{
  Takes a \link{Prediction} and calculates a numeric score.
If the measure if flagged with the properties \code{"requires_task"}, \code{"requires_learner"},
\code{"requires_model"} or \code{"requires_train_set"}, you must additionally
pass the respective \link{Task}, the (trained) \link{Learner} or the training set indices.
This is handled internally during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Measure$score(prediction, task = NULL, learner = NULL, train_set = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{prediction}}{(\link{Prediction}).}
      \item{\code{task}}{(\link{Task}).}
      \item{\code{learner}}{(\link{Learner}).}
      \item{\code{train_set}}{(\code{integer()}).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{numeric(1)}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
learner = lrn("classif.rpart")$train(task)
prediction = learner$predict(task)
msr("classif.ce")$score(prediction)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Measure-aggregate"></a>}}
\if{latex}{\out{\hypertarget{method-Measure-aggregate}{}}}
\subsection{\code{Measure$aggregate()}}{
  Aggregates multiple performance scores into a single score, e.g. by using the \code{aggregator}
function of the measure.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Measure$aggregate(rr)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rr}}{\link{ResampleResult}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{numeric(1)}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
learner = lrn("classif.rpart")
rr = resample(task, learner, rsmp("holdout"))
msr("classif.ce")$aggregate(rr)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Measure-obs_loss"></a>}}
\if{latex}{\out{\hypertarget{method-Measure-obs_loss}{}}}
\subsection{\code{Measure$obs_loss()}}{
  Calculates the observation-wise loss.
Returns a \code{numeric()} with one element for each row in the \link{Prediction}.
If there is no observation-wise loss function for the measure, \code{NA_real_} values are returned.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Measure$obs_loss(prediction, task = NULL, learner = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{prediction}}{(\link{Prediction}).}
      \item{\code{task}}{(\link{Task}).}
      \item{\code{learner}}{(\link{Learner}).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{numeric()} with one element for each row in the \link{Prediction}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
learner = lrn("classif.rpart")
learner$train(task)
prediction = learner$predict(task)
msr("classif.ce")$obs_loss(prediction)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Measure-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Measure-clone}{}}}
\subsection{\code{Measure$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Measure$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskClassif_penguins.R
\name{mlr_tasks_penguins}
\alias{mlr_tasks_penguins}
\title{Palmer Penguins Data Set}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link{TaskClassif}.
}
\source{
\CRANpkg{palmerpenguins}
}
\description{
Classification data to predict the species of penguins from the \CRANpkg{palmerpenguins} package,
see \link[palmerpenguins:penguins]{palmerpenguins::penguins}.
A better alternative to the \link[=iris]{iris data set}.
}
\section{Dictionary}{

This \link{Task} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_tasks} or with the associated sugar function \code{\link[=tsk]{tsk()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_tasks$get("penguins")
tsk("penguins")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Dimensions: 344x8
\item Properties: \dQuote{multiclass}
\item Has Missings: \code{TRUE}
\item Target: \dQuote{species}
\item Features: \dQuote{bill_depth}, \dQuote{bill_length}, \dQuote{body_mass}, \dQuote{flipper_length}, \dQuote{island}, \dQuote{sex}, \dQuote{year}
}
}

\section{Pre-processing}{

\itemize{
\item The unit of measurement have been removed from the column names.
Lengths are given in millimeters (mm), weight in gram (g).
}
}

\references{
Gorman KB, Williams TD, Fraser WR (2014).
\dQuote{Ecological Sexual Dimorphism and Environmental Variability within a Community of Antarctic Penguins (Genus Pygoscelis).}
\emph{PLoS ONE}, \bold{9}(3), e90081.
\doi{10.1371/journal.pone.0090081}.

\url{https://github.com/allisonhorst/palmerpenguins}
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{Task}},
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureBIC.R
\name{mlr_measures_bic}
\alias{mlr_measures_bic}
\alias{MeasureBIC}
\title{Bayesian Information Criterion Measure}
\description{
Calculates the Bayesian Information Criterion (BIC) which is a
trade-off between goodness of fit (measured in terms of
log-likelihood) and model complexity (measured in terms of number
of included features).
Internally, \code{\link[stats:BIC]{stats::BIC()}} is called.
Requires the learner property \code{"loglik"}, \code{NA} is returned for unsupported learners.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("bic")
msr("bic")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{NA}
\item Range: \eqn{(-\infty, \infty)}{(-Inf, Inf)}
\item Minimize: TRUE
\item Average: macro
\item Required Prediction: \dQuote{NA}
\item Required Packages: \CRANpkg{mlr3}
}
}

\section{Parameters}{

Empty ParamSet
}

\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-eval}
\item Package \CRANpkg{mlr3measures} for the scoring functions.
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}
\code{as.data.table(mlr_measures)} for a table of available \link[=Measure]{Measures} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Measure:
\code{\link{Measure}},
\code{\link{MeasureClassif}},
\code{\link{MeasureRegr}},
\code{\link{MeasureSimilarity}},
\code{\link{mlr_measures}},
\code{\link{mlr_measures_aic}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_debug_classif}},
\code{\link{mlr_measures_elapsed_time}},
\code{\link{mlr_measures_internal_valid_score}},
\code{\link{mlr_measures_oob_error}},
\code{\link{mlr_measures_regr.pinball}},
\code{\link{mlr_measures_regr.rqr}},
\code{\link{mlr_measures_regr.rsq}},
\code{\link{mlr_measures_selected_features}}
}
\concept{Measure}
\section{Super class}{
\code{\link[mlr3:Measure]{Measure}} -> \code{MeasureBIC}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MeasureBIC-initialize}{\code{MeasureBIC$new()}}
    \item \href{#method-MeasureBIC-clone}{\code{MeasureBIC$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="aggregate"><a href='../../mlr3/html/Measure.html#method-Measure-aggregate'><code>Measure$aggregate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="format"><a href='../../mlr3/html/Measure.html#method-Measure-format'><code>Measure$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="help"><a href='../../mlr3/html/Measure.html#method-Measure-help'><code>Measure$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="obs_loss"><a href='../../mlr3/html/Measure.html#method-Measure-obs_loss'><code>Measure$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="print"><a href='../../mlr3/html/Measure.html#method-Measure-print'><code>Measure$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="score"><a href='../../mlr3/html/Measure.html#method-Measure-score'><code>Measure$score()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureBIC-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureBIC-initialize}{}}}
\subsection{\code{MeasureBIC$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureBIC$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureBIC-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureBIC-clone}{}}}
\subsection{\code{MeasureBIC$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureBIC$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marshal.R
\name{marshaling}
\alias{marshaling}
\alias{learner_unmarshal}
\alias{learner_marshal}
\alias{learner_marshaled}
\alias{marshal_model}
\alias{unmarshal_model}
\alias{is_marshaled_model}
\title{(Un)marshal a Learner}
\usage{
learner_unmarshal(.learner, ...)

learner_marshal(.learner, ...)

learner_marshaled(.learner)

marshal_model(model, inplace = FALSE, ...)

unmarshal_model(model, inplace = FALSE, ...)

is_marshaled_model(model)
}
\arguments{
\item{.learner}{\code{\link{Learner}}\cr
The learner.}

\item{...}{(any)\cr
Additional parameters, currently unused.}

\item{model}{(any)\cr
Model to marshal.}

\item{inplace}{(\code{logical(1)})\cr
Whether to marshal in-place.}
}
\description{
Marshaling is the process of processing the model of a trained \code{\link{Learner}} so it can be successfully serialized and
deserialized. The naming is inspired by the \href{https://github.com/futureverse/marshal}{marshal package} and we
plan to fully migrate to this package once it is on CRAN.
The current implementation should therefore be considered as a temporary solution and is likely
to change in the future.

The central functions (and the only methods that are used by \code{mlr3} internally) are:
\itemize{
\item the S3 generic \code{marshal_model(model, inplace, ...)}.
Which takes in a model and returns it in marshaled form.
This means, that the resulting object can be serialized and de-serialized without loss of information.
If a model is serializable anyway, nothing has to be implemented and the generic will fall back to the
default implementation of \code{marshal_model}, which is to return the object as-is.
Otherwise, the marshaled object should be a list with named elements \code{marshaled} and
\code{packages}, where the former contains
the marshaled object, and the latter the package that contains the packages required to unmarshal.
Most importantly, this list should contain the package that contains the \code{unmarshal_model} method.
The returned object should have the classes of the original object with the suffix \code{"_marshaled"} appended and the
root class should be set to \code{"marshaled"}.
\item the S3 generic \verb{unmarshal_model(model, inplace ...)}.
Which takes in the marshaled model and returns it in unmarshaled form.
The generic takes care that the packages specified during \code{"marshal"} are loaded,
and errs if they are not available.
Calling this function on a marshaled model should reconstruct the original model, i.e.
\code{unmarshal_model(marshal_model(x))} should return \code{x}.
The default implementation of this generic returns \code{x} as-is.
\item the function \code{is_marshaled_model(model)}.
This (helper) function returns \code{TRUE} if the model inherits from class \code{"marshaled"} and \code{FALSE} otherwise.
Note that it is not guaranteed that \code{is_marshaled_model(marshal_model(x))} returns \code{TRUE}.
This is because the default \code{marshal_model(x)} returns \code{x} as-is.
}

For both \code{marshal_model} and \code{unmarshal_model}, the \code{inplace} argument determines whether in-place marshaling
should be performed. This is especially relevant in the context of references semantics.
If \code{inplace} is \code{FALSE}, the original input should not be modified, otherwise this is allowed.
Note that the input and output can still share references, even when \code{inplace} is \code{FALSE}.
}
\section{Implementing Marshaling}{


In order to implement marshaling for a Learner, you need to overload the \code{marshal_model} and \code{unmarshal_model}
methods for the class of the learner's model and tag the learner with the \code{"marshal"} property.
To make marshaling accessible in an R6-manner, you should also add the public methods \verb{$marshal()}, \verb{$unmarshal()}
and the active binding \verb{$marshaled}.
To make this as convenient as possible, the functions \code{learner_marshal(.learner, ...)},
\code{learner_unmarshal(.learner, ...)}
and \code{learner_marshaled(.learner)} are provided and can be called from the public methods.

You can verify whether you have correctly implemented marshaling by using the internal test helper
\code{expect_marshalable_learner(learner, task)}. This is also run by \code{expect_learner()} if a task is provided.

For a concrete example on how to implement marshaling, see \code{\link{LearnerClassifDebug}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ResultData.R
\name{ResultData}
\alias{ResultData}
\title{ResultData}
\description{
Internal object to store results in list of data.tables, arranged in a star schema.
It is advised to not directly work on this data structure as it may be changed in the future
without further warnings.

The main motivation of this data structure is the necessity to avoid storing duplicated \link[R6:R6Class]{R6} objects.
While this is usually no problem in a single R session, serialization via \code{\link[=serialize]{serialize()}} (which is
used in \code{\link[=save]{save()}}/\code{\link[=saveRDS]{saveRDS()}} or during parallelization) leads to objects with unreasonable memory
requirements.
}
\examples{
# table overview
print(ResultData$new()$data)
}
\keyword{internal}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{data}}{(\code{list()})\cr
List of \code{\link[data.table:data.table]{data.table::data.table()}}, arranged in a star schema.
Do not operate directly on this list.}
  }
  \if{html}{\out{</div>}}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{task_type}}{(\code{character(1)})\cr
Returns the task type of stored objects, e.g. \code{"classif"} or \code{"regr"}.
Returns \code{NULL} if the \link{ResultData} is empty.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ResultData-initialize}{\code{ResultData$new()}}
    \item \href{#method-ResultData-uhashes}{\code{ResultData$uhashes()}}
    \item \href{#method-ResultData-uhash_table}{\code{ResultData$uhash_table()}}
    \item \href{#method-ResultData-iterations}{\code{ResultData$iterations()}}
    \item \href{#method-ResultData-tasks}{\code{ResultData$tasks()}}
    \item \href{#method-ResultData-learners}{\code{ResultData$learners()}}
    \item \href{#method-ResultData-learner_states}{\code{ResultData$learner_states()}}
    \item \href{#method-ResultData-resamplings}{\code{ResultData$resamplings()}}
    \item \href{#method-ResultData-predictions}{\code{ResultData$predictions()}}
    \item \href{#method-ResultData-prediction}{\code{ResultData$prediction()}}
    \item \href{#method-ResultData-data_extra}{\code{ResultData$data_extra()}}
    \item \href{#method-ResultData-combine}{\code{ResultData$combine()}}
    \item \href{#method-ResultData-sweep}{\code{ResultData$sweep()}}
    \item \href{#method-ResultData-marshal}{\code{ResultData$marshal()}}
    \item \href{#method-ResultData-unmarshal}{\code{ResultData$unmarshal()}}
    \item \href{#method-ResultData-discard}{\code{ResultData$discard()}}
    \item \href{#method-ResultData-as_data_table}{\code{ResultData$as_data_table()}}
    \item \href{#method-ResultData-logs}{\code{ResultData$logs()}}
    \item \href{#method-ResultData-set_threshold}{\code{ResultData$set_threshold()}}
    \item \href{#method-ResultData-clone}{\code{ResultData$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-initialize}{}}}
\subsection{\code{ResultData$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
An alternative construction method is provided by \code{\link[=as_result_data]{as_result_data()}}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$new(data = NULL, data_extra = NULL, store_backends = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{data}}{(\code{\link[data.table:data.table]{data.table::data.table()}}) | \code{NULL})\cr
Do not initialize this object yourself, use \code{\link[=as_result_data]{as_result_data()}} instead.}
      \item{\code{data_extra}}{(\code{list()})\cr
Additional data to store.
This can be used to store additional information for each iteration.}
      \item{\code{store_backends}}{(\code{logical(1)})\cr
If set to \code{FALSE}, the backends of the \link{Task}s provided in \code{data} are removed.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-uhashes"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-uhashes}{}}}
\subsection{\code{ResultData$uhashes()}}{
  Returns all unique hashes (\code{uhash} values) of all included \link{ResampleResult}s.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$uhashes(view = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{character()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-uhash_table"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-uhash_table}{}}}
\subsection{\code{ResultData$uhash_table()}}{
  Returns a \code{\link{data.table}} with columns \code{uhash}, \code{learner_id}, \code{task_id} and \code{resampling_id}
for the given view.
The \code{uhash} uniquely identifies an individual \code{\link{ResampleResult}}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$uhash_table(view = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{data.table()}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-iterations"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-iterations}{}}}
\subsection{\code{ResultData$iterations()}}{
  Returns the number of recorded iterations / experiments.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$iterations(view = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{integer(1)}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-tasks"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-tasks}{}}}
\subsection{\code{ResultData$tasks()}}{
  Returns a table of included \link{Task}s.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$tasks(view = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{data.table()} with columns \code{"task_hash"} (\code{character()}) and \code{"task"} (\link{Task}).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-learners"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-learners}{}}}
\subsection{\code{ResultData$learners()}}{
  Returns a table of included \link{Learner}s.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$learners(view = NULL, states = TRUE, reassemble = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{states}}{(\code{logical(1)})\cr
If \code{TRUE}, returns a learner for each iteration/experiment in the \link{ResultData} object.
If \code{FALSE}, returns an exemplary learner (without state) for each \link{ResampleResult}.}
      \item{\code{reassemble}}{(\code{logical(1)})\cr
Reassemble the learners, i.e. re-set the \code{state} and the hyperparameters which are stored separately before
returning the learners.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{data.table()} with columns \code{"learner_hash"} (\code{character()}) and \code{"learner"} (\link{Learner}).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-learner_states"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-learner_states}{}}}
\subsection{\code{ResultData$learner_states()}}{
  Returns a list of states of included \link{Learner}s without reassembling the learners.

@return list of \code{list()}
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$learner_states(view = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-resamplings"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-resamplings}{}}}
\subsection{\code{ResultData$resamplings()}}{
  Returns a table of included \link{Resampling}s.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$resamplings(view = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{data.table()} with columns \code{"resampling_hash"} (\code{character()}) and \code{"resampling"} (\link{Resampling}).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-predictions"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-predictions}{}}}
\subsection{\code{ResultData$predictions()}}{
  Returns a list of \link{Prediction} objects.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$predictions(view = NULL, predict_sets = "test")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{list()} of \link{Prediction}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-prediction"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-prediction}{}}}
\subsection{\code{ResultData$prediction()}}{
  Returns a combined \link{Prediction} objects.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$prediction(view = NULL, predict_sets = "test")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \link{Prediction}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-data_extra"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-data_extra}{}}}
\subsection{\code{ResultData$data_extra()}}{
  Returns additional data stored.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$data_extra(view = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{\link[data.table:data.table]{data.table::data.table()}}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-combine"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-combine}{}}}
\subsection{\code{ResultData$combine()}}{
  Combines multiple \link{ResultData} objects, modifying \code{self} in-place.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$combine(rdata)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rdata}}{(\link{ResultData}).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{self} (invisibly).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-sweep"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-sweep}{}}}
\subsection{\code{ResultData$sweep()}}{
  Updates the \link{ResultData} object, removing rows from all tables which are not referenced
by the fact table anymore.
E.g., can be called after filtering/subsetting the fact table.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$sweep()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified \code{self} (invisibly).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-marshal"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-marshal}{}}}
\subsection{\code{ResultData$marshal()}}{
  Marshals all stored learner models.
This will do nothing to models that are already marshaled.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$marshal(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=marshal_model]{marshal_model()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-unmarshal"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-unmarshal}{}}}
\subsection{\code{ResultData$unmarshal()}}{
  Unmarshals all stored learner models.
This will do nothing to models which are not marshaled.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$unmarshal(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(any)\cr
Additional arguments passed to \code{\link[=unmarshal_model]{unmarshal_model()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-discard"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-discard}{}}}
\subsection{\code{ResultData$discard()}}{
  Shrinks the object by discarding parts of the stored data.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$discard(backends = FALSE, models = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{backends}}{(\code{logical(1)})\cr
If \code{TRUE}, the \link{DataBackend} is removed from all stored \link{Task}s.}
      \item{\code{models}}{(\code{logical(1)})\cr
If \code{TRUE}, the stored model is removed from all \link{Learner}s.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified \code{self} (invisibly).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-as_data_table"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-as_data_table}{}}}
\subsection{\code{ResultData$as_data_table()}}{
  Combines internal tables into a single flat \code{\link[=data.table]{data.table()}}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$as_data_table(
  view = NULL,
  reassemble_learners = TRUE,
  convert_predictions = TRUE,
  predict_sets = "test"
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{reassemble_learners}}{(\code{logical(1)})\cr
Reassemble the tasks?}
      \item{\code{convert_predictions}}{(\code{logical(1)})\cr
Convert \link{PredictionData} to \link{Prediction}?}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-logs"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-logs}{}}}
\subsection{\code{ResultData$logs()}}{
  Get a table of recorded learner logs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$logs(view = NULL, condition)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{condition}}{(\code{character(1)})
The condition to extract. One of \code{"message"}, \code{"warning"} or \code{"error"}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{\link[data.table:data.table]{data.table::data.table()}}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-set_threshold"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-set_threshold}{}}}
\subsection{\code{ResultData$set_threshold()}}{
  Sets the threshold for the response prediction of classification learners,
given they have output a probability prediction.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$set_threshold(view = NULL, threshold, ties_method = "random")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{view}}{(\code{character(1)} | \code{NULL})\cr
Single \code{uhash} to restrict the results to.}
      \item{\code{threshold}}{(\code{numeric(1)})\cr
Threshold value.}
      \item{\code{ties_method}}{(\code{character(1)})\cr
Method to handle ties in probabilities when selecting a class label.
Must be one of \code{"random"}, \code{"first"} or \code{"last"} (corresponding to the same options in \code{\link[=max.col]{max.col()}}).
\itemize{
\item \code{"random"}: Randomly select one of the tied class labels (default).
\item \code{"first"}: Select the first class label among tied values.
\item \code{"last"}: Select the last class label among tied values.
}}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ResultData-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ResultData-clone}{}}}
\subsection{\code{ResultData$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ResultData$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auto_convert.R
\name{auto_convert}
\alias{auto_convert}
\title{Column Auto-Converter}
\usage{
auto_convert(value, id, type, levels)
}
\arguments{
\item{value}{(any)\cr
New values to convert in order to match \code{type}.}

\item{id}{(\code{character(1)})\cr
Name of the column, used in error messages.}

\item{type}{(\code{character(1)})\cr
Type to convert \code{values} to.}

\item{levels}{(\code{character()} | \code{NULL})\cr
Levels to use for conversion to \code{factor} or \code{ordered}.}
}
\value{
Vector \code{value} converted to type \code{type}.
}
\description{
Set of rules to automatically convert column types.
This is used during \code{rbind}-ing of \link{Task}s, but also in some pipe operators in
\CRANpkg{mlr3pipelines}.

All rules are stored as functions in \link[=mlr_reflections]{mlr_reflections$auto_converters}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_threads.R
\name{set_threads}
\alias{set_threads}
\alias{set_threads.default}
\alias{set_threads.R6}
\alias{set_threads.list}
\title{Set the Number of Threads}
\usage{
set_threads(x, n = availableCores(), ...)

\method{set_threads}{default}(x, n = availableCores(), ...)

\method{set_threads}{R6}(x, n = availableCores(), ...)

\method{set_threads}{list}(x, n = availableCores(), ...)
}
\arguments{
\item{x}{(any)\cr
Object to set threads for, e.g. a \link{Learner}.
This object is modified in-place.}

\item{n}{(\code{integer(1)})\cr
Number of threads to use. Defaults to \code{\link[parallelly:availableCores]{parallelly::availableCores()}}.}

\item{...}{(any)\cr
Additional arguments.}
}
\value{
Same object as input \code{x} (changed in-place),
with possibly updated parameter values.
}
\description{
Control the parallelism via threading while calling external packages from \CRANpkg{mlr3}.

For example, the random forest implementation in package \CRANpkg{ranger} (connected
via \CRANpkg{mlr3learners}) supports threading via OpenMP.
The number of threads to use can be set via hyperparameter \code{num.threads}, and
defaults to 1. By calling \code{set_threads(x, 4)} with \code{x} being a ranger learner, the
hyperparameter is changed so that 4 cores are used.

If the object \code{x} does not support threading, \code{x} is returned as-is.
If applied to a list, recurses through all list elements.

Note that threading is incompatible with other parallelization techniques such as forking
via the \link[future:plan]{future::plan} \link[future:multicore]{future::multicore}. For this reason all learners connected to \CRANpkg{mlr3}
have threading disabled in their defaults.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R, R/helper.R
\name{mlr_assertions}
\alias{mlr_assertions}
\alias{assert_backend}
\alias{assert_task}
\alias{assert_tasks}
\alias{assert_learner}
\alias{assert_learners}
\alias{assert_learnable}
\alias{assert_predictable}
\alias{assert_measure}
\alias{assert_scorable}
\alias{assert_measures}
\alias{assert_resampling}
\alias{assert_resamplings}
\alias{assert_prediction}
\alias{assert_resample_result}
\alias{assert_benchmark_result}
\alias{assert_row_ids}
\alias{assert_has_backend}
\alias{assert_quantiles}
\alias{assert_validate}
\title{Assertion for mlr3 Objects}
\usage{
assert_backend(b, .var.name = vname(b))

assert_task(
  task,
  task_type = NULL,
  feature_types = NULL,
  task_properties = NULL,
  .var.name = vname(task)
)

assert_tasks(
  tasks,
  task_type = NULL,
  feature_types = NULL,
  task_properties = NULL,
  .var.name = vname(tasks)
)

assert_learner(
  learner,
  task = NULL,
  task_type = NULL,
  properties = character(),
  .var.name = vname(learner)
)

assert_learners(
  learners,
  task = NULL,
  task_type = NULL,
  properties = character(),
  unique_ids = FALSE,
  .var.name = vname(learners)
)

assert_learnable(task, learner, param_values = NULL)

assert_predictable(task, learner)

assert_measure(
  measure,
  task = NULL,
  learner = NULL,
  prediction = NULL,
  .var.name = vname(measure)
)

assert_scorable(
  measure,
  task,
  learner,
  prediction = NULL,
  .var.name = vname(measure)
)

assert_measures(
  measures,
  task = NULL,
  learner = NULL,
  .var.name = vname(measures)
)

assert_resampling(
  resampling,
  instantiated = NULL,
  .var.name = vname(resampling)
)

assert_resamplings(
  resamplings,
  instantiated = NULL,
  .var.name = vname(resamplings)
)

assert_prediction(prediction, .var.name = vname(prediction), null.ok = FALSE)

assert_resample_result(rr, .var.name = vname(rr))

assert_benchmark_result(bmr, .var.name = vname(bmr))

assert_row_ids(
  row_ids,
  task = NULL,
  null.ok = FALSE,
  .var.name = vname(row_ids)
)

assert_has_backend(task)

assert_quantiles(learner, quantile_response = FALSE)

assert_validate(x)
}
\arguments{
\item{b}{(\link{DataBackend}).}

\item{task}{(\link{Task}).}

\item{task_type}{(\code{character(1)}).}

\item{feature_types}{(\code{character()})\cr
Feature types the learner operates on. Must be a subset of \code{\link[=mlr_reflections]{mlr_reflections$task_feature_types}}.}

\item{task_properties}{(\code{character()})\cr
Set of required task properties.}

\item{tasks}{(list of \link{Task}).}

\item{learner}{(\link{Learner}).}

\item{learners}{(list of \link{Learner}).}

\item{param_values}{(\code{list()})\cr
TuneToken are not allowed in the parameter set of the learner.
If the \code{param_values} overwrite the TuneToken, the assertion will pass.}

\item{measure}{(\link{Measure}).}

\item{prediction}{(\link{Prediction}).}

\item{measures}{(list of \link{Measure}).}

\item{resampling}{(\link{Resampling}).}

\item{resamplings}{(list of \link{Resampling}).}

\item{rr}{(\link{ResampleResult}).}

\item{bmr}{(\link{BenchmarkResult}).}

\item{row_ids}{\code{integer()}\cr
Row indices.}

\item{quantile_response}{(\code{logical(1)})\cr
Whether to check if the quantile response is set.
If \code{TRUE}, the learner must have the \verb{$quantile_response} field set.}

\item{x}{(any)\cr
The input to check.}
}
\description{
Functions intended to be used in packages extending \pkg{mlr3}.
Most assertion functions ensure the right class attribute, and optionally additional properties.
Additionally, the following compound assertions are implemented:
\itemize{
\item \code{assert_learnable(task, learner)}\cr
(\link{Task}, \link{Learner}) -> \code{NULL}\cr
Checks if the learner is applicable to the task.
This includes type checks on the type, the feature types, and properties.
}

If an assertion fails, an exception is raised.
Otherwise, the input object is returned invisibly.

Asserts whether the input is a valid value for the \verb{$validate} field of a \code{\link{Learner}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BenchmarkResult.R
\name{uhashes}
\alias{uhashes}
\alias{uhash}
\title{Obtain specific uhashes from a \link{BenchmarkResult}}
\usage{
uhashes(bmr, learner_ids = NULL, task_ids = NULL, resampling_ids = NULL)

uhash(bmr, learner_id = NULL, task_id = NULL, resampling_id = NULL)
}
\arguments{
\item{bmr}{(\code{BenchmarkResult})\cr
Benchmark result.}

\item{learner_ids}{(\code{character()} | \code{NULL})\cr
Learner IDs.}

\item{task_ids}{(\code{character()} | \code{NULL})\cr
Task IDs.}

\item{resampling_ids}{(\code{character()} | \code{NULL})\cr
Resampling IDs.}

\item{learner_id}{(\code{character(1)} | \code{NULL})\cr
Learner ID.}

\item{task_id}{(\code{character(1)} | \code{NULL})\cr
Task ID.}

\item{resampling_id}{(\code{character(1)} | \code{NULL})\cr
Resampling ID.}
}
\description{
In a \code{\link{BenchmarkResult}}, each \link{ResampleResult} is uniquely identified by a \emph{hash} (\emph{uhash}).
Operations that select specific \link{ResampleResult}s from a \link{BenchmarkResult} operate using
these hashes.
This function allows to obtain uhashes for specific learners, tasks, and resamplings.

If you want more control, you can also directly obtain the uhash table from the \code{\link{BenchmarkResult}}
via the field \verb{$uhash_table}.
}
\examples{
design = benchmark_grid(
  tsks(c("sonar", "iris")),
  lrns(c("classif.debug", "classif.featureless", "classif.rpart")),
  rsmp("holdout")
)
bmr = benchmark(design)
bmr
bmr$uhashes
uhash(bmr, learner_id = "classif.debug", task_id = "sonar", resampling_id = "holdout")
uhashes(bmr, learner_ids = c("classif.debug", "classif.featureless"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataBackendDataTable.R
\name{DataBackendDataTable}
\alias{DataBackendDataTable}
\title{DataBackend for data.table}
\description{
\link{DataBackend} for \CRANpkg{data.table} which serves as an efficient in-memory data base.
}
\examples{
data = as.data.table(palmerpenguins::penguins)
data$id = seq_len(nrow(palmerpenguins::penguins))
b = DataBackendDataTable$new(data = data, primary_key = "id")
print(b)
b$head()
b$data(rows = 100:101, cols = "species")

b$nrow
head(b$rownames)

b$ncol
b$colnames

# alternative construction
as_data_backend(palmerpenguins::penguins)
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter10/advanced_technical_aspects_of_mlr3.html#sec-backends}
\item Package \CRANpkg{mlr3db} to interface out-of-memory data,
e.g. SQL servers or \CRANpkg{duckdb}.
}

Other DataBackend:
\code{\link{DataBackend}},
\code{\link[=as_data_backend]{as_data_backend()}}
}
\concept{DataBackend}
\section{Super class}{
\code{\link[mlr3:DataBackend]{DataBackend}} -> \code{DataBackendDataTable}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{compact_seq}}{\code{logical(1)}\cr
If \code{TRUE}, row ids are a natural sequence from 1 to \code{nrow(data)} (determined internally).
In this case, row lookup uses faster positional indices instead of equi joins.}
  }
  \if{html}{\out{</div>}}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{rownames}}{(\code{integer()})\cr
Returns vector of all distinct row identifiers, i.e. the contents of the primary key column.}

    \item{\code{colnames}}{(\code{character()})\cr
Returns vector of all column names, including the primary key column.}

    \item{\code{nrow}}{(\code{integer(1)})\cr
Number of rows (observations).}

    \item{\code{ncol}}{(\code{integer(1)})\cr
Number of columns (variables), including the primary key column.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-DataBackendDataTable-initialize}{\code{DataBackendDataTable$new()}}
    \item \href{#method-DataBackendDataTable-data}{\code{DataBackendDataTable$data()}}
    \item \href{#method-DataBackendDataTable-head}{\code{DataBackendDataTable$head()}}
    \item \href{#method-DataBackendDataTable-distinct}{\code{DataBackendDataTable$distinct()}}
    \item \href{#method-DataBackendDataTable-missings}{\code{DataBackendDataTable$missings()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="DataBackend" data-id="format"><a href='../../mlr3/html/DataBackend.html#method-DataBackend-format'><code>DataBackend$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="DataBackend" data-id="print"><a href='../../mlr3/html/DataBackend.html#method-DataBackend-print'><code>DataBackend$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataBackendDataTable-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-DataBackendDataTable-initialize}{}}}
\subsection{\code{DataBackendDataTable$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.

Note that \code{DataBackendDataTable} does not copy the input data,
while \code{as_data_backend()} calls \code{\link[data.table:copy]{data.table::copy()}}.
\code{as_data_backend()} also takes care about casting to a \code{data.table()} and adds a primary key column if necessary.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{DataBackendDataTable$new(data, primary_key)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{data}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
The input \code{\link[=data.table]{data.table()}}.}
      \item{\code{primary_key}}{(\code{character(1)} | \code{integer()})\cr
Name of the primary key column, or integer vector of row ids.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataBackendDataTable-data"></a>}}
\if{latex}{\out{\hypertarget{method-DataBackendDataTable-data}{}}}
\subsection{\code{DataBackendDataTable$data()}}{
  Returns a slice of the data.
The rows must be addressed as vector of primary key values, columns must be referred to via column names.
Queries for rows with no matching row id and queries for columns with no matching column name are
silently ignored.
Rows are guaranteed to be returned in the same order as \code{rows}, columns may be returned in an arbitrary order.
Duplicated row ids result in duplicated rows, duplicated column names lead to an exception.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{DataBackendDataTable$data(rows, cols)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rows}}{(positive \code{integer()} | \code{NULL})\cr
Vector or row indices.
Always refers to the complete data set, even after filtering.}
      \item{\code{cols}}{(\code{character()} | \code{NULL})\cr
Vector of column names.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataBackendDataTable-head"></a>}}
\if{latex}{\out{\hypertarget{method-DataBackendDataTable-head}{}}}
\subsection{\code{DataBackendDataTable$head()}}{
  Retrieve the first \code{n} rows.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{DataBackendDataTable$head(n = 6L)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{(\code{integer(1)})\cr
Number of rows.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{\link[data.table:data.table]{data.table::data.table()}} of the first \code{n} rows.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataBackendDataTable-distinct"></a>}}
\if{latex}{\out{\hypertarget{method-DataBackendDataTable-distinct}{}}}
\subsection{\code{DataBackendDataTable$distinct()}}{
  Returns a named list of vectors of distinct values for each column
specified. If \code{na_rm} is \code{TRUE}, missing values are removed from the
returned vectors of distinct values. Non-existing rows and columns are
silently ignored.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{DataBackendDataTable$distinct(rows, cols, na_rm = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rows}}{(positive \code{integer()} | \code{NULL})\cr
Vector or row indices.
Always refers to the complete data set, even after filtering.}
      \item{\code{cols}}{(\code{character()} | \code{NULL})\cr
Vector of column names.}
      \item{\code{na_rm}}{\code{logical(1)}\cr
Whether to remove NAs or not.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Named \code{list()} of distinct values.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataBackendDataTable-missings"></a>}}
\if{latex}{\out{\hypertarget{method-DataBackendDataTable-missings}{}}}
\subsection{\code{DataBackendDataTable$missings()}}{
  Returns the number of missing values per column in the specified slice
of data. Non-existing rows and columns are silently ignored.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{DataBackendDataTable$missings(rows, cols)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rows}}{(positive \code{integer()} | \code{NULL})\cr
Vector or row indices.
Always refers to the complete data set, even after filtering.}
      \item{\code{cols}}{(\code{character()} | \code{NULL})\cr
Vector of column names.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Total of missing values per column (named \code{numeric()}).
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TaskClassif.R
\name{TaskClassif}
\alias{TaskClassif}
\title{Classification Task}
\description{
This task specializes \link{Task} and \link{TaskSupervised} for classification problems.
The target column is assumed to be a factor or ordered factor.
The \code{task_type} is set to \code{"classif"}.

Additional task properties include:
\itemize{
\item \code{"twoclass"}: The task is a binary classification problem.
\item \code{"multiclass"}: The task is a multiclass classification problem.
}

It is recommended to use \code{\link[=as_task_classif]{as_task_classif()}} for construction.
Predefined tasks are stored in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_tasks}.
}
\examples{
data("Sonar", package = "mlbench")
task = as_task_classif(Sonar, target = "Class", positive = "M")

task$task_type
task$formula()
task$truth()
task$class_names
task$positive
task$data(rows = 1:3, cols = task$feature_names[1:2])
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{Task}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}
\section{Super classes}{
\code{\link[mlr3:Task]{Task}} -> \code{\link[mlr3:TaskSupervised]{TaskSupervised}} -> \code{TaskClassif}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{class_names}}{(\code{character()})\cr
Returns all class labels of the target column.}

    \item{\code{positive}}{(\code{character(1)})\cr
Stores the positive class for binary classification tasks, and \code{NA} for multiclass tasks.
To switch the positive class, assign a level to this field.}

    \item{\code{negative}}{(\code{character(1)})\cr
Stores the negative class for binary classification tasks, and \code{NA} for multiclass tasks.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-TaskClassif-initialize}{\code{TaskClassif$new()}}
    \item \href{#method-TaskClassif-print}{\code{TaskClassif$print()}}
    \item \href{#method-TaskClassif-truth}{\code{TaskClassif$truth()}}
    \item \href{#method-TaskClassif-droplevels}{\code{TaskClassif$droplevels()}}
    \item \href{#method-TaskClassif-clone}{\code{TaskClassif$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="add_strata"><a href='../../mlr3/html/Task.html#method-Task-add_strata'><code>Task$add_strata()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="cbind"><a href='../../mlr3/html/Task.html#method-Task-cbind'><code>Task$cbind()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="data"><a href='../../mlr3/html/Task.html#method-Task-data'><code>Task$data()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="filter"><a href='../../mlr3/html/Task.html#method-Task-filter'><code>Task$filter()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="format"><a href='../../mlr3/html/Task.html#method-Task-format'><code>Task$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="formula"><a href='../../mlr3/html/Task.html#method-Task-formula'><code>Task$formula()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="head"><a href='../../mlr3/html/Task.html#method-Task-head'><code>Task$head()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="help"><a href='../../mlr3/html/Task.html#method-Task-help'><code>Task$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="levels"><a href='../../mlr3/html/Task.html#method-Task-levels'><code>Task$levels()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="materialize_view"><a href='../../mlr3/html/Task.html#method-Task-materialize_view'><code>Task$materialize_view()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="missings"><a href='../../mlr3/html/Task.html#method-Task-missings'><code>Task$missings()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="rbind"><a href='../../mlr3/html/Task.html#method-Task-rbind'><code>Task$rbind()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="rename"><a href='../../mlr3/html/Task.html#method-Task-rename'><code>Task$rename()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="select"><a href='../../mlr3/html/Task.html#method-Task-select'><code>Task$select()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="set_col_roles"><a href='../../mlr3/html/Task.html#method-Task-set_col_roles'><code>Task$set_col_roles()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="set_levels"><a href='../../mlr3/html/Task.html#method-Task-set_levels'><code>Task$set_levels()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Task" data-id="set_row_roles"><a href='../../mlr3/html/Task.html#method-Task-set_row_roles'><code>Task$set_row_roles()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskClassif-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-TaskClassif-initialize}{}}}
\subsection{\code{TaskClassif$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
The function \code{\link[=as_task_classif]{as_task_classif()}} provides an alternative way to construct classification tasks.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskClassif$new(
  id,
  backend,
  target,
  positive = NULL,
  label = NA_character_,
  extra_args = list()
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier for the new instance.}
      \item{\code{backend}}{(\link{DataBackend})\cr
Either a \link{DataBackend}, or any object which is convertible to a \link{DataBackend} with \code{as_data_backend()}.
E.g., a \code{data.frame()} will be converted to a \link{DataBackendDataTable}.}
      \item{\code{target}}{(\code{character(1)})\cr
Name of the target column.}
      \item{\code{positive}}{(\code{character(1)})\cr
Only for binary classification: Name of the positive class.
The levels of the target columns are reordered accordingly, so that the first element of \verb{$class_names} is the
positive class, and the second element is the negative class.}
      \item{\code{label}}{(\code{character(1)})\cr
Label for the new instance.}
      \item{\code{extra_args}}{(named \code{list()})\cr
Named list of constructor arguments, required for converting task types
via \code{\link[=convert_task]{convert_task()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskClassif-print"></a>}}
\if{latex}{\out{\hypertarget{method-TaskClassif-print}{}}}
\subsection{\code{TaskClassif$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskClassif$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskClassif-truth"></a>}}
\if{latex}{\out{\hypertarget{method-TaskClassif-truth}{}}}
\subsection{\code{TaskClassif$truth()}}{
  True response for specified \code{row_ids}. Format depends on the task type.
Defaults to all rows with role \code{"use"}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskClassif$truth(rows = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rows}}{(positive \code{integer()} | \code{NULL})\cr
Vector or row indices.
Always refers to the complete data set, even after filtering.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{factor()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskClassif-droplevels"></a>}}
\if{latex}{\out{\hypertarget{method-TaskClassif-droplevels}{}}}
\subsection{\code{TaskClassif$droplevels()}}{
  Updates the cache of stored factor levels, removing all levels not present in the current set of active rows.
\code{cols} defaults to all columns with storage type "factor" or "ordered".
Also updates the task property \code{"twoclass"}/\code{"multiclass"}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskClassif$droplevels(cols = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{cols}}{(\code{character()} | \code{NULL})\cr
Vector of column names.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified \code{self}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-TaskClassif-clone"></a>}}
\if{latex}{\out{\hypertarget{method-TaskClassif-clone}{}}}
\subsection{\code{TaskClassif$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{TaskClassif$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_data_backend.R
\name{as_data_backend}
\alias{as_data_backend}
\alias{as_data_backend.data.frame}
\title{Create a Data Backend}
\usage{
as_data_backend(data, primary_key = NULL, ...)

\method{as_data_backend}{data.frame}(data, primary_key = NULL, keep_rownames = FALSE, ...)
}
\arguments{
\item{data}{(\code{\link[=data.frame]{data.frame()}})\cr
The input \code{\link[=data.frame]{data.frame()}}.
Automatically converted to a \code{\link[data.table:data.table]{data.table::data.table()}}.}

\item{primary_key}{(\code{character(1)} | \code{integer()})\cr
Name of the primary key column, or integer vector of row ids.}

\item{...}{(any)\cr
Additional arguments passed to the respective \link{DataBackend} method.}

\item{keep_rownames}{(\code{logical(1)} | \code{character(1)})\cr
If \code{TRUE} or a single string, keeps the row names of \code{data} as a new column.
The column is named like the provided string, defaulting to \code{"..rownames"} for \code{keep_rownames == TRUE}.
Note that the created column will be used as a regular feature by the task
unless you manually change the column role.
Also see \code{\link[data.table:as.data.table]{data.table::as.data.table()}}.}
}
\value{
\link{DataBackend}.
}
\description{
Wraps a \link{DataBackend} around data.
\CRANpkg{mlr3} ships with methods for \code{data.frame} (converted to a \link{DataBackendDataTable}.

Additional methods are implemented in the package \CRANpkg{mlr3db}, e.g. to connect
to real DBMS like PostgreSQL (via \CRANpkg{dbplyr}) or DuckDB (via \CRANpkg{DBI}/\CRANpkg{duckdb}).
}
\examples{
# create a new backend using the penguins data:
as_data_backend(palmerpenguins::penguins)
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter10/advanced_technical_aspects_of_mlr3.html#sec-backends}
\item Package \CRANpkg{mlr3db} to interface out-of-memory data,
e.g. SQL servers or \CRANpkg{duckdb}.
}

Other DataBackend:
\code{\link{DataBackend}},
\code{\link{DataBackendDataTable}}
}
\concept{DataBackend}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ContextResample.R
\name{ContextResample}
\alias{ContextResample}
\title{Resample Context}
\description{
A \link{CallbackResample} accesses and modifies data during \code{\link[=resample]{resample()}} and \code{\link[=benchmark]{benchmark()}} via the \code{ContextResample}.
See the section on fields for a list of modifiable objects.
See \code{\link[=callback_resample]{callback_resample()}} for a list of stages that access \code{ContextResample}.
}
\section{Super class}{
\code{\link[mlr3misc:Context]{mlr3misc::Context}} -> \code{ContextResample}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{task}}{(\link{Task})\cr
The task to be evaluated.
The task is unchanged during the evaluation.
The task is read-only.}

    \item{\code{learner}}{(\link{Learner})\cr
The learner to be evaluated.
The learner contains the models after stage \code{on_resample_before_train}.}

    \item{\code{resampling}}{\link{Resampling}\cr
The resampling strategy to be used.
The resampling is unchanged during the evaluation.
The resampling is read-only.}

    \item{\code{iteration}}{(\code{integer()})\cr
The current iteration.
The iteration is read-only.}

    \item{\code{pdatas}}{(List of \link{PredictionData})\cr
The prediction data.
The data is available on stage \code{on_resample_end}.}

    \item{\code{data_extra}}{(list())\cr
Data saved in the \link{ResampleResult} or \link{BenchmarkResult}.
Use this field to save results.
Must be a \code{list()}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ContextResample-initialize}{\code{ContextResample$new()}}
    \item \href{#method-ContextResample-clone}{\code{ContextResample$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3misc" data-topic="Context" data-id="format"><a href='../../mlr3misc/html/Context.html#method-Context-format'><code>mlr3misc::Context$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3misc" data-topic="Context" data-id="print"><a href='../../mlr3misc/html/Context.html#method-Context-print'><code>mlr3misc::Context$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ContextResample-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ContextResample-initialize}{}}}
\subsection{\code{ContextResample$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ContextResample$new(task, learner, resampling, iteration)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link{Task})\cr
The task to be evaluated.}
      \item{\code{learner}}{(\link{Learner})\cr
The learner to be evaluated.}
      \item{\code{resampling}}{(\link{Resampling})\cr
The resampling strategy to be used.}
      \item{\code{iteration}}{(\code{integer()})\cr
The current iteration.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ContextResample-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ContextResample-clone}{}}}
\subsection{\code{ContextResample$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ContextResample$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_test_helpers.R
\name{mlr_test_helpers}
\alias{mlr_test_helpers}
\title{Documentation of mlr3 test helpers}
\description{
The mlr3 package contains various helper functions to test the validity of objects such as learners.
These functions are not contained in the mlr3 namespaces and are instead located in the \code{inst/testthat}
directory of the source package or the \code{testthat} directory of the installed package.

These files can be sourced with the following line of code:

\if{html}{\out{<div class="sourceCode">}}\preformatted{lapply(list.files(system.file("testthat", package = "mlr3"),
  pattern = "^helper.*\\\\.[rR]", full.names = TRUE), source)
}\if{html}{\out{</div>}}

Other extension packages such as \code{mlr3proba} have similar files that can be sourced accordingly.

This manual page documents the most important helper functions that are relevant when users implement their own
custom learners.
}
\section{run_autotest()}{

This function runs a Learner's automatic test suite.

During the autotests, multiple tasks are generated depending on the properties of the learner.
The \code{run_autotest()} function then trains the learner on each task and predicts with all supported predict types.
(see argument \code{predict_types}).
To debug, simply run \code{result = run_autotest(learner)} and proceed with investigating
the task, learner and prediction of the returned \code{result}.

For example usages you can look at the autotests in various mlr3 source repositories such as mlr3learners.
More information can be found in the \code{inst/testthat/autotest.R} file.

\strong{Parameters}:
\itemize{
\item \code{learner} (\code{\link{Learner}})\cr
The learner to check.
\item \code{N} (\code{integer(1)})\cr
The number of rows of the generated tasks.
\item \code{exclude} (\code{character()})\cr
Each task on which the learner is trained has an id.
If for some reason, one or more such tests ought to be disabled, this argument takes in a regular expression
that disables all tasks whose id matches the regular expression.
\item \code{predict_types} (\code{character()})\cr
The predict types of the learner to check.
Defaults to all predict types of the learner.
\item \code{check_replicable} (\code{logical(1)})\cr
Whether to check that running the learner twice with the same seed should result in identical predictions.
Default is \code{TRUE}.
\item \code{configure_learner} (\verb{function(learner, task)})\cr
Before running a \code{learner} on a \code{task}, this function allows to change its parameter values
depending on the input task.
}

To speed up test execution, start an encapsulation daemon in advance by calling
\code{mirai::daemons(1, .compute = "mlr3_encapsulation")} in \code{tests/testthat/setup.R}.
After the tests complete, stop them with
\code{mirai::daemons(0, .compute = "mlr3_encapsulation")} in \code{tests/testthat/teardown.R}.
}

\section{run_paramtest()}{


\strong{Description}:

Checks parameters of mlr3 Learners against parameters defined in the upstream functions of the respective learner.
The goal is to detect if parameters have been dropped or added in the upstream implementation.
Some learners do not have all of their parameters stored within the learner function that is called during training.
Sometimes learners come with a "control" function, e.g. \code{glmnet.control()} from package \CRANpkg{glmnet}.
Such learners need to be checked as well since they make up the full ParamSet of the respective learner.

To work nicely with the defined ParamSet, certain parameters need to be
excluded because these are only present in either the "control" object or the
actual top-level function call. Such exclusions should go into argument
\code{exclude} with a comment for the reason of the exclusion. See examples for
more information.

For example usages you can look at the parameter tests in various mlr3 source repositories
such as \CRANpkg{mlr3learners}.

\strong{Parameters}:
\itemize{
\item \code{learner} (\code{Learner})\cr
The learner whose parameter set is being checked.
\item \code{fun} (\verb{function()} or list of \code{functions()}s)\cr
The function(s) containing the parameters that must be implemented by the learner.
\item \code{exclude} (\code{character()})\cr
Argument names that specified through this argument are exempt from checking.
This can be used when parameters that are available in the \code{fun} function(s) are not implemented in the learner,
or when the learner implements additional parameters that are not available in the \code{fun} function(s).
\item \code{tag} (\code{character(1)})\cr
Only parameters that are tagged with this tag are being checked.
If \code{NULL} (default), all parameters are checked.
}
}

\section{expect_learner()}{


Checks various properties that learners have to satisfy.
Used for testing learner implementations, especially if all methods and fields are implemented as documented.

\strong{Parameters}
\itemize{
\item \code{lrn} :: (\code{\link{Learner}})\cr
The learner whose properties are being verified.
\item \code{tsk} :: (\code{\link{Task}})\cr
Optional argument (default is \code{NULL}).
If provided, some additional checks are being run that check the compatibility of the learner and task.
\item \code{check_man} :: (\code{logical(1)})\cr
Whether to check if the learner has a man page.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_resampling.R
\name{as_resampling}
\alias{as_resampling}
\alias{as_resampling.Resampling}
\alias{as_resamplings}
\alias{as_resamplings.default}
\alias{as_resamplings.list}
\title{Convert to a Resampling}
\usage{
as_resampling(x, ...)

\method{as_resampling}{Resampling}(x, clone = FALSE, ...)

as_resamplings(x, ...)

\method{as_resamplings}{default}(x, ...)

\method{as_resamplings}{list}(x, ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{...}{(any)\cr
Additional arguments.}

\item{clone}{(\code{logical(1)})\cr
If \code{TRUE}, ensures that the returned object is not the same as the input \code{x}.}
}
\description{
Convert object to a \link{Resampling} or a list of \link{Resampling}.
This method e.g. allows to convert an \code{OMLTask} of \CRANpkg{mlr3oml} to a \code{\link{Resampling}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_task.R
\name{as_task}
\alias{as_task}
\alias{as_task.Task}
\alias{as_tasks}
\alias{as_tasks.default}
\alias{as_tasks.list}
\title{Convert to a Task}
\usage{
as_task(x, ...)

\method{as_task}{Task}(x, clone = FALSE, ...)

as_tasks(x, ...)

\method{as_tasks}{default}(x, ...)

\method{as_tasks}{list}(x, ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{...}{(any)\cr
Additional arguments.}

\item{clone}{(\code{logical(1)})\cr
If \code{TRUE}, ensures that the returned object is not the same as the input \code{x}.}
}
\description{
Convert object to a \link{Task} or a list of \link{Task}.

The function supports:
\itemize{
\item Converting existing \link{Task} objects (with optional cloning)
\item Converting objects from other packages (e.g., \code{OMLTask} from \CRANpkg{mlr3oml})
\item Converting lists of objects to lists of tasks
}

For constructing tasks from data frames, use the dedicated converters:
\itemize{
\item \code{\link[=as_task_classif]{as_task_classif()}} for classification tasks
\item \code{\link[=as_task_regr]{as_task_regr()}} for regression tasks
\item \code{\link[=as_task_unsupervised]{as_task_unsupervised()}} for unsupervised tasks
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureRegrRQR.R
\name{mlr_measures_regr.rqr}
\alias{mlr_measures_regr.rqr}
\alias{MeasureRegrRQR}
\title{R-Squared for Quantile Regression}
\description{
Measure to compare true observed response with predicted quantiles in regression tasks.
}
\details{
\eqn{R^1(\alpha)} is defined as \deqn{
  1 - \frac{\sum_{i=1}^n \rho_\alpha \left( t_i - r_i(\alpha) \right)}{
    \sum_{i=1}^n \rho_\alpha \left( t_i - q_{\alpha} \right)},
}{
  1 - sum(pinball(t - r(alpha))) / sum(pinball(t - quantile(t, alpha))),
}
where for a quantile \eqn{\alpha}, \eqn{\rho_\alpha} is the pinball function, \eqn{r_i(\alpha)} are the predictions
for the quantile and \eqn{q_{\alpha}} is the empirical \eqn{\alpha}-quantile of the test or training data.

\eqn{R^1(\alpha)} is analogous to \eqn{R^2} for regression tasks.
It compares the pinball function of the predictions relative to a naive model predicting the empirical quantile.

This measure is undefined for constant \eqn{t}.
}
\section{Dictionary}{

This \link{Measure} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures} or with the associated sugar function \code{\link[=msr]{msr()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_measures$get("regr.rqr")
msr("regr.rqr")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{regr}
\item Range: \eqn{(-\infty, 1]}{(-Inf, 1]}
\item Minimize: FALSE
\item Average: macro
\item Required Prediction: \dQuote{quantiles}
\item Required Packages: \CRANpkg{mlr3}
}
}

\section{Parameters}{
\tabular{llll}{
   Id \tab Type \tab Default \tab Range \cr
   alpha \tab numeric \tab - \tab \eqn{[0, 1]}{[0, 1]} \cr
}
}

\references{
Koenker, Roger, Machado, F. JA (1999).
\dQuote{Goodness of Fit and Related Inference Processes for Quantile Regression.}
\emph{Journal of the American Statistical Association}, \bold{94}(448), 1296--1310.
\doi{10.1080/01621459.1999.10473882}.
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-eval}
\item Package \CRANpkg{mlr3measures} for the scoring functions.
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}
\code{as.data.table(mlr_measures)} for a table of available \link[=Measure]{Measures} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Measure:
\code{\link{Measure}},
\code{\link{MeasureClassif}},
\code{\link{MeasureRegr}},
\code{\link{MeasureSimilarity}},
\code{\link{mlr_measures}},
\code{\link{mlr_measures_aic}},
\code{\link{mlr_measures_bic}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_debug_classif}},
\code{\link{mlr_measures_elapsed_time}},
\code{\link{mlr_measures_internal_valid_score}},
\code{\link{mlr_measures_oob_error}},
\code{\link{mlr_measures_regr.pinball}},
\code{\link{mlr_measures_regr.rsq}},
\code{\link{mlr_measures_selected_features}}
}
\concept{Measure}
\section{Super classes}{
\code{\link[mlr3:Measure]{Measure}} -> \code{\link[mlr3:MeasureRegr]{MeasureRegr}} -> \code{MeasureRegrRQR}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{pred_set_mean}}{(\code{logical(1)})\cr
Whether the empirical quantile is calculated on the prediction set.
Set during construction.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MeasureRegrRQR-initialize}{\code{MeasureRegrRQR$new()}}
    \item \href{#method-MeasureRegrRQR-clone}{\code{MeasureRegrRQR$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="aggregate"><a href='../../mlr3/html/Measure.html#method-Measure-aggregate'><code>Measure$aggregate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="format"><a href='../../mlr3/html/Measure.html#method-Measure-format'><code>Measure$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="help"><a href='../../mlr3/html/Measure.html#method-Measure-help'><code>Measure$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="obs_loss"><a href='../../mlr3/html/Measure.html#method-Measure-obs_loss'><code>Measure$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="print"><a href='../../mlr3/html/Measure.html#method-Measure-print'><code>Measure$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="score"><a href='../../mlr3/html/Measure.html#method-Measure-score'><code>Measure$score()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureRegrRQR-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureRegrRQR-initialize}{}}}
\subsection{\code{MeasureRegrRQR$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureRegrRQR$new(alpha = 0.5, pred_set_mean = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{alpha}}{\code{numeric(1)}\cr
The quantile for which to compute the measure.
Must be one of the quantiles that the Learner was trained on.}
      \item{\code{pred_set_mean}}{\code{logical(1)}\cr
If \code{TRUE}, the mean of the true values is calculated on the prediction set.
If \code{FALSE}, the mean of the true values is calculated on the training set.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureRegrRQR-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureRegrRQR-clone}{}}}
\subsection{\code{MeasureRegrRQR$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureRegrRQR$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_benchmark_result.R
\name{as_benchmark_result}
\alias{as_benchmark_result}
\alias{as_benchmark_result.BenchmarkResult}
\alias{as_benchmark_result.ResampleResult}
\title{Convert to BenchmarkResult}
\usage{
as_benchmark_result(x, ...)

\method{as_benchmark_result}{BenchmarkResult}(x, ...)

\method{as_benchmark_result}{ResampleResult}(x, ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{...}{(any)\cr
Additional arguments.}
}
\value{
(\link{BenchmarkResult}).
}
\description{
Convert object to a \link{BenchmarkResult}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Resampling.R
\name{Resampling}
\alias{Resampling}
\title{Resampling Class}
\description{
This is the abstract base class for resampling objects like \link{ResamplingCV} and \link{ResamplingBootstrap}.

The objects of this class define how a task is partitioned for resampling (e.g., in \code{\link[=resample]{resample()}} or \code{\link[=benchmark]{benchmark()}}),
using a set of hyperparameters such as the number of folds in cross-validation.

Resampling objects can be instantiated on a \link{Task}, which applies the strategy on the task and manifests in a
fixed partition of \code{row_ids} of the \link{Task}.

Predefined resamplings are stored in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_resamplings},
e.g. \code{\link[=mlr_resamplings_cv]{cv}} or \code{\link[=mlr_resamplings_bootstrap]{bootstrap}}.
}
\section{Stochasticity & Reproducibility}{

The \code{\link{Resampling}} class only defines an abstract resampling strategy.
Concrete data splits are obtained by calling \verb{$instantiate()} on a \code{\link{Task}}.
To ensure reproducibility of results, you need to call \code{set.seed} before doing so.
Note that \code{\link{benchmark_grid}} internally does instantiate resamplings, so you need to set
the seed before calling it.
}

\section{Stratification}{

All derived classes support stratified sampling.
The stratification variables are assumed to be discrete and must be stored in the \link{Task}
with column role \code{"stratum"}.
In case of multiple stratification variables,
each combination of the values of the stratification variables forms a strata.

First, the observations are divided into subpopulations based one or multiple stratification
variables (assumed to be discrete), c.f. \code{task$strata}.

Second, the sampling is performed in each of the \code{k} subpopulations separately.
Each subgroup is divided into \code{iter} training sets and \code{iter} test sets by the derived \code{Resampling}.
These sets are merged based on their iteration number:
all training sets from all subpopulations with iteration 1 are combined,
then all training sets with iteration 2, and so on.
Same is done for all test sets.
The merged sets can be accessed via \verb{$train_set(i)} and \verb{$test_set(i)}, respectively.
Note that this procedure can lead to set sizes that are slightly different from those
without stratification.
}

\section{Grouping / Blocking}{

All derived classes support grouping of observations.
The grouping variable is assumed to be discrete and must be stored in the \link{Task} with column role \code{"group"}.

Observations in the same group are treated like a "block" of observations which must be kept together.
These observations either all go together into the training set or together into the test set.

The sampling is performed by the derived \link{Resampling} on the grouping variable.
Next, the grouping information is replaced with the respective row ids to generate training and test sets.
The sets can be accessed via \verb{$train_set(i)} and \verb{$test_set(i)}, respectively.
}

\section{Inheriting}{

It is possible to overwrite both \code{private$.get_instance()} to have full control,
or only \code{private$.sample()} when one wants to use the pre-defined mechanism for stratification and grouping.
}

\examples{
r = rsmp("subsampling")

# Default parametrization
r$param_set$values

# Do only 3 repeats on 10\% of the data
r$param_set$set_values(ratio = 0.1, repeats = 3)
r$param_set$values

# Instantiate on penguins task
task = tsk("penguins")
r$instantiate(task)

# Extract train/test sets
train_set = r$train_set(1)
print(train_set)
intersect(train_set, r$test_set(1))

# Another example: 10-fold CV
r = rsmp("cv")$instantiate(task)
r$train_set(1)

# Stratification
task = tsk("pima")
prop.table(table(task$truth())) # moderately unbalanced
task$col_roles$stratum = task$target_names

r = rsmp("subsampling")
r$instantiate(task)
prop.table(table(task$truth(r$train_set(1)))) # roughly same proportion

## ------------------------------------------------
## Method `Resampling$instantiate()`
## ------------------------------------------------

task = tsk("penguins")
resampling = rsmp("holdout")
resampling$instantiate(task)

## ------------------------------------------------
## Method `Resampling$train_set()`
## ------------------------------------------------

task = tsk("penguins")
resampling = rsmp("holdout")$instantiate(task)
resampling$train_set(1)

## ------------------------------------------------
## Method `Resampling$test_set()`
## ------------------------------------------------

task = tsk("penguins")
resampling = rsmp("holdout")$instantiate(task)
resampling$test_set(1)
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter3/evaluation_and_benchmarking.html#sec-resampling}
\item Package \CRANpkg{mlr3spatiotempcv} for spatio-temporal resamplings.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Resampling]{Resamplings}: \link{mlr_resamplings}
\item \code{as.data.table(mlr_resamplings)} for a table of available \link[=Resampling]{Resamplings} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3spatiotempcv} for additional \link{Resampling}s for spatio-temporal
tasks.
}

Other Resampling:
\code{\link{mlr_resamplings}},
\code{\link{mlr_resamplings_bootstrap}},
\code{\link{mlr_resamplings_custom}},
\code{\link{mlr_resamplings_custom_cv}},
\code{\link{mlr_resamplings_cv}},
\code{\link{mlr_resamplings_holdout}},
\code{\link{mlr_resamplings_insample}},
\code{\link{mlr_resamplings_loo}},
\code{\link{mlr_resamplings_repeated_cv}},
\code{\link{mlr_resamplings_subsampling}}
}
\concept{Resampling}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{instance}}{(any)\cr
During \code{instantiate()}, the instance is stored in this slot in an arbitrary format.
Note that if a grouping variable is present in the \link{Task}, a \link{Resampling} may operate on the
group ids internally instead of the row ids (which may lead to confusion).

It is advised to not work directly with the \code{instance}, but instead only use the getters
\verb{$train_set()} and \verb{$test_set()}.}
  }
  \if{html}{\out{</div>}}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{id}}{(\code{character(1)})\cr
Identifier of the object.
Used in tables, plot and text output.}

    \item{\code{is_instantiated}}{(\code{logical(1)})\cr
Is \code{TRUE} if the resampling has been instantiated.}

    \item{\code{hash}}{(\code{character(1)})\cr
Hash (unique identifier) for this object.
If the object has not been instantiated yet, \code{NA_character_} is returned.
The hash is calculated based on the class name, the id, the parameter set, and the instance.}

    \item{\code{label}}{(\code{character(1)})\cr
Label for this object.
Can be used in tables, plot and text output instead of the ID.}

    \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}

    \item{\code{task_hash}}{(\code{character(1)})\cr
The hash of the \link{Task} which was passed to \code{r$instantiate()}.}

    \item{\code{task_row_hash}}{(\code{character(1)})\cr
The hash of the row ids of the \link{Task} which was passed to \code{r$instantiate()}.}

    \item{\code{task_nrow}}{(\code{integer(1)})\cr
The number of observations of the \link{Task} which was passed to \code{r$instantiate()}.}

    \item{\code{duplicated_ids}}{(\code{logical(1)})\cr
If \code{TRUE}, duplicated rows can occur within a single training set or within a single test set.
E.g., this is \code{TRUE} for Bootstrap, and \code{FALSE} for cross-validation.
Only used internally.}

    \item{\code{man}}{(\code{character(1)} | \code{NULL})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
Defaults to \code{NA}, but can be set by child classes.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Resampling-initialize}{\code{Resampling$new()}}
    \item \href{#method-Resampling-format}{\code{Resampling$format()}}
    \item \href{#method-Resampling-print}{\code{Resampling$print()}}
    \item \href{#method-Resampling-help}{\code{Resampling$help()}}
    \item \href{#method-Resampling-instantiate}{\code{Resampling$instantiate()}}
    \item \href{#method-Resampling-train_set}{\code{Resampling$train_set()}}
    \item \href{#method-Resampling-test_set}{\code{Resampling$test_set()}}
    \item \href{#method-Resampling-clone}{\code{Resampling$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Resampling-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Resampling-initialize}{}}}
\subsection{\code{Resampling$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Resampling$new(
  id,
  param_set = ps(),
  duplicated_ids = FALSE,
  label = NA_character_,
  man = NA_character_
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier for the new instance.}
      \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}
      \item{\code{duplicated_ids}}{(\code{logical(1)})\cr
Set to \code{TRUE} if this resampling strategy may have duplicated row ids in a single training set or test set.

Note that this object is typically constructed via a derived classes, e.g. \link{ResamplingCV} or \link{ResamplingHoldout}.}
      \item{\code{label}}{(\code{character(1)})\cr
Label for the new instance.}
      \item{\code{man}}{(\code{character(1)})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
The referenced help package can be opened via method \verb{$help()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Resampling-format"></a>}}
\if{latex}{\out{\hypertarget{method-Resampling-format}{}}}
\subsection{\code{Resampling$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Resampling$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Resampling-print"></a>}}
\if{latex}{\out{\hypertarget{method-Resampling-print}{}}}
\subsection{\code{Resampling$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Resampling$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Resampling-help"></a>}}
\if{latex}{\out{\hypertarget{method-Resampling-help}{}}}
\subsection{\code{Resampling$help()}}{
  Opens the corresponding help page referenced by field \verb{$man}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Resampling$help()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Resampling-instantiate"></a>}}
\if{latex}{\out{\hypertarget{method-Resampling-instantiate}{}}}
\subsection{\code{Resampling$instantiate()}}{
  Materializes fixed training and test splits for a given task and stores them in \code{r$instance}
in an arbitrary format.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Resampling$instantiate(task)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link{Task})\cr
Task used for instantiation.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
resampling = rsmp("holdout")
resampling$instantiate(task)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Resampling-train_set"></a>}}
\if{latex}{\out{\hypertarget{method-Resampling-train_set}{}}}
\subsection{\code{Resampling$train_set()}}{
  Returns the row ids of the i-th training set.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Resampling$train_set(i)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{i}}{(\code{integer(1)})\cr
Iteration.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    (\code{integer()}) of row ids.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
resampling = rsmp("holdout")$instantiate(task)
resampling$train_set(1)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Resampling-test_set"></a>}}
\if{latex}{\out{\hypertarget{method-Resampling-test_set}{}}}
\subsection{\code{Resampling$test_set()}}{
  Returns the row ids of the i-th test set.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Resampling$test_set(i)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{i}}{(\code{integer(1)})\cr
Iteration.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    (\code{integer()}) of row ids.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
resampling = rsmp("holdout")$instantiate(task)
resampling$test_set(1)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Resampling-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Resampling-clone}{}}}
\subsection{\code{Resampling$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Resampling$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/score_roc_measures.R
\name{print.roc_measures}
\alias{print.roc_measures}
\title{Print ROC Measures}
\usage{
\method{print}{roc_measures}(x, abbreviations = TRUE, digits = 2L, ...)
}
\arguments{
\item{x}{(\code{roc_measures})\cr
The object returned by \code{score_roc_measures}.}

\item{abbreviations}{(\code{logical(1)})\cr
If \code{TRUE}, print a list of abbreviations for the measures.}

\item{digits}{(\code{integer(1)})\cr
Number of digits to round the measures to.}

\item{...}{(\code{any})\cr
Additional parameters, currently unused.}
}
\description{
Print the confusion matrix and a set of roc performance measures.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reexports.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{as.data.table}
\alias{data.table}
\alias{mlr_callbacks}
\alias{clbk}
\alias{clbks}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{data.table}{\code{\link[data.table:as.data.table]{as.data.table()}}, \code{\link[data.table:data.table]{data.table()}}}

  \item{mlr3misc}{\code{\link[mlr3misc:clbk]{clbk()}}, \code{\link[mlr3misc:clbks]{clbks()}}, \code{\link[mlr3misc:mlr_callbacks]{mlr_callbacks}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Task.R
\name{Task}
\alias{Task}
\title{Task Class}
\description{
This is the abstract base class for \link{TaskSupervised} and \link{TaskUnsupervised}.
\link{TaskClassif} and \link{TaskRegr} inherit from \link{TaskSupervised}.
More supervised tasks are implemented in \CRANpkg{mlr3proba}, unsupervised cluster tasks
in package \CRANpkg{mlr3cluster}.

Tasks serve two purposes:
\enumerate{
\item Tasks wrap a \link{DataBackend}, an object to transparently interface different data storage types.
\item Tasks store meta-information, such as the role of the individual columns in the \link{DataBackend}.
For example, for a classification task a single column must be marked as target column, and others as features.
}

Predefined (toy) tasks are stored in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_tasks},
e.g. \code{\link[=mlr_tasks_penguins]{penguins}} or \code{\link[=mlr_tasks_california_housing]{california_housing}}.
More toy tasks can be found in the dictionary after loading \CRANpkg{mlr3data}.
}
\section{S3 methods}{

\itemize{
\item \code{as.data.table(t)}\cr
\link{Task} -> \code{\link[data.table:data.table]{data.table::data.table()}}\cr
Returns the complete data as \code{\link[data.table:data.table]{data.table::data.table()}}.
\item \code{head(t)}\cr
Calls \code{\link[=head]{head()}} on the task's data.
\item \code{summary(t)}\cr
Calls \code{\link[=summary]{summary()}} on the task's data.
}
}

\section{Task mutators}{

The following methods change the task in-place:
\itemize{
\item Any modification of the lists \verb{$col_roles} or \verb{$row_roles}.
This provides a different "view" on the data without altering the data itself.
This may affect, e.g., \verb{$data}, \verb{$nrow}, \verb{$ncol}, \code{n_features}, \code{row_ids}, and \verb{$feature_names}.
Altering \verb{$col_roles} may affect, e.g., \verb{$data}, \verb{$ncol}, \verb{$n_features}, and \verb{$feature_names}.
Altering \verb{$row_roles} may affect, e.g., \verb{$data}, \verb{$nrow}, and \verb{$row_ids}.
\item Modification of column or row roles via \verb{$set_col_roles()} or \verb{$set_row_roles()}, respectively.
They are an alternative to directly accessing \verb{$col_roles} or \verb{$row_roles}, with the same side effects.
\item \verb{$select()} and \verb{$filter()} subset the set of active features or rows in \verb{$col_roles} or \verb{$row_roles},
respectively.
\item \verb{$cbind()} and \verb{$rbind()} change the task in-place by binding new columns or rows to the data.
\item \verb{$rename()} changes column names.
\item \verb{$set_levels()} and \verb{$droplevels()} update the field \verb{$col_info()} to automatically repair
factor levels while querying data with \verb{$data()}.
\item \verb{$materialize_view()} creates a new \link{DataBackendDataTable} which keeps only the data in the currently active view
possibly freeing some memory consumed by the \link{DataBackend} stored in the \code{Task}.
}
}

\examples{
# We use the inherited class TaskClassif here,
# because the base class `Task` is not intended for direct use
task = TaskClassif$new("penguins", palmerpenguins::penguins, target = "species")

task$nrow
task$ncol
task$feature_names
task$formula()

# de-select "year"
task$select(setdiff(task$feature_names, "year"))

task$feature_names

# Add new column "foo"
task$cbind(data.frame(foo = 1:344))
head(task)

## ------------------------------------------------
## Method `Task$data()`
## ------------------------------------------------

task = tsk("penguins")
task$data(rows = 1:5, cols = c("species", "sex"))

## ------------------------------------------------
## Method `Task$formula()`
## ------------------------------------------------

task = tsk("penguins")
task$formula()

## ------------------------------------------------
## Method `Task$head()`
## ------------------------------------------------

task = tsk("penguins")
task$head(3)

## ------------------------------------------------
## Method `Task$levels()`
## ------------------------------------------------

task = tsk("penguins")
task$levels()

## ------------------------------------------------
## Method `Task$missings()`
## ------------------------------------------------

task = tsk("penguins")
task$missings()

## ------------------------------------------------
## Method `Task$filter()`
## ------------------------------------------------

task = tsk("penguins")
task$filter(1:10)
task$nrow

## ------------------------------------------------
## Method `Task$select()`
## ------------------------------------------------

task = tsk("penguins")
task$select(c("bill_length", "bill_depth"))
task$feature_names

## ------------------------------------------------
## Method `Task$rbind()`
## ------------------------------------------------

task = tsk("penguins")
extra = task$data(rows = 1:2)
task$rbind(extra)

## ------------------------------------------------
## Method `Task$cbind()`
## ------------------------------------------------

task = tsk("penguins")
task$cbind(data.table(extra_col = seq_len(task$nrow)))
head(task$data(cols = "extra_col"))

## ------------------------------------------------
## Method `Task$rename()`
## ------------------------------------------------

task = tsk("penguins")
task$rename("body_mass", "mass")
task$feature_names

## ------------------------------------------------
## Method `Task$set_row_roles()`
## ------------------------------------------------

task = tsk("penguins")
task$set_row_roles(1:5, remove_from = "use")

## ------------------------------------------------
## Method `Task$set_col_roles()`
## ------------------------------------------------

task = tsk("penguins")
task$set_col_roles("sex", roles = "stratum")
task$col_roles$stratum

## ------------------------------------------------
## Method `Task$set_levels()`
## ------------------------------------------------

task = tsk("penguins")
task$set_levels(list(sex = c("male", "female", "unknown")))
task$levels("sex")

## ------------------------------------------------
## Method `Task$droplevels()`
## ------------------------------------------------

task = tsk("penguins")
task$set_levels(list(sex = c("male", "female", "unknown")))
task$levels("sex")

## ------------------------------------------------
## Method `Task$add_strata()`
## ------------------------------------------------

task = tsk("penguins")
task$add_strata("flipper_length", bins = 4)

## ------------------------------------------------
## Method `Task$materialize_view()`
## ------------------------------------------------

task = tsk("iris")
task$backend$nrow
task$filter(1:120)
task$backend$nrow
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3data} for more toy tasks.
\item Package \CRANpkg{mlr3oml} for downloading tasks from \url{https://www.openml.org}.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Task]{Tasks}: \link{mlr_tasks}
\item \code{as.data.table(mlr_tasks)} for a table of available \link[=Task]{Tasks} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3fselect} and \CRANpkg{mlr3filters} for feature selection and feature filtering.
\item Extension packages for additional task types:
\itemize{
\item Unsupervised clustering: \CRANpkg{mlr3cluster}
\item Probabilistic supervised regression and survival analysis: \url{https://mlr3proba.mlr-org.com/}.
}
}

Other Task:
\code{\link{TaskClassif}},
\code{\link{TaskRegr}},
\code{\link{TaskSupervised}},
\code{\link{TaskUnsupervised}},
\code{\link{california_housing}},
\code{\link{mlr_tasks}},
\code{\link{mlr_tasks_breast_cancer}},
\code{\link{mlr_tasks_german_credit}},
\code{\link{mlr_tasks_iris}},
\code{\link{mlr_tasks_mtcars}},
\code{\link{mlr_tasks_penguins}},
\code{\link{mlr_tasks_pima}},
\code{\link{mlr_tasks_sonar}},
\code{\link{mlr_tasks_spam}},
\code{\link{mlr_tasks_wine}},
\code{\link{mlr_tasks_zoo}}
}
\concept{Task}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{id}}{(\code{character(1)})\cr
Identifier of the object.
Used in tables, plot and text output.}

    \item{\code{internal_valid_task}}{(\code{Task} | \code{integer()} | \code{NULL})\cr
Optional validation task that can, e.g., be used for early stopping with learners such as XGBoost.
See also the \verb{$validate} field of \code{\link{Learner}}.
If integers are assigned they are removed from the primary task and an internal validation task
with those ids is created from the primary task using only those ids.
When assigning a new task, it is always cloned.}

    \item{\code{hash}}{(\code{character(1)})\cr
Hash (unique identifier) for this object.
The hash is calculated based on the complete task object and \verb{$row_ids}.
If an internal validation task is set, the hash is recalculated.}

    \item{\code{row_hash}}{(\code{character(1)})\cr
Hash (unique identifier) calculated based on the row ids.}

    \item{\code{row_ids}}{(positive \code{integer()})\cr
Returns the row ids of the \link{DataBackend} for observations with role "use".}

    \item{\code{row_names}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Returns a table with two columns:
\itemize{
\item \code{"row_id"} (\code{integer()}), and
\item \code{"row_name"} (\code{character()}).
}}

    \item{\code{feature_names}}{(\code{character()})\cr
Returns all column names with \code{role == "feature"}.

Note that this vector determines the default order of columns for \code{task$data(cols = NULL, ...)}.
However, it is recommended to \strong{not} rely on the order of columns, but instead always
address columns by their name. The default order is not well defined after some
operations, e.g. after \code{task$cbind()} or after processing via \CRANpkg{mlr3pipelines}.}

    \item{\code{target_names}}{(\code{character()})\cr
Returns all column names with role "target".}

    \item{\code{properties}}{(\code{character()})\cr
Set of task properties.
Possible properties are stored in \link[=mlr_reflections]{mlr_reflections$task_properties}.
The following properties are currently standardized and understood by tasks in \CRANpkg{mlr3}:
\itemize{
\item \code{"strata"}: The task is resampled using one or more stratification variables (role \code{"stratum"}).
\item \code{"groups"}: The task comes with grouping/blocking information (role \code{"group"}).
\item \code{"weights_learner"}: If the task has observation weights with this role,
they are passed to the \link{Learner} during train.
The use of weights can be disabled by setting the learner's hyperparameter \code{use_weights} to \code{FALSE}.
\item \code{"weights_measure"}: If the task has observation weights with this role,
they are passed to the \link{Measure} for weighted scoring.
The use of weights can be disabled by setting the measure's hyperparameter \code{use_weights} to \code{FALSE}.
\item \code{"offset"}: The task includes one or more offset columns specifying fixed adjustments for model training
and possibly for prediction (role \code{"offset"}).
\item \code{"ordered"}: The task has columns which define the row order (role \code{"order"}).
}

Note that above listed properties are calculated from the \verb{$col_roles}, and may not be set explicitly.}

    \item{\code{row_roles}}{(named \code{list()})\cr
Each row (observation) can have an arbitrary number of roles in the learning task:
\itemize{
\item \code{"use"}: Use in train / predict / resampling.
}

\code{row_roles} is a named list whose elements are named by row role and each element is an \code{integer()} vector of
row ids.
To alter the roles, just modify the list,
e.g. with \R's set functions (\code{\link[=intersect]{intersect()}}, \code{\link[=setdiff]{setdiff()}}, \code{\link[=union]{union()}}, \ldots).}

    \item{\code{col_roles}}{(named \code{list()})\cr
Each column can be in one or more of the following groups to fulfill different roles:
\itemize{
\item \code{"feature"}: Regular feature used in the model fitting process.
\item \code{"target"}: Target variable. Most tasks only accept a single target column.
\item \code{"name"}: Row names / observation labels. To be used in plots. Can be queried with \verb{$row_names}.
Not more than a single column can be associated with this role.
\item \code{"order"}: Data returned by \verb{$data()} is ordered by this column (or these columns).
Columns must be sortable with \code{\link[=order]{order()}}.
\item \code{"group"}: During resampling, observations with the same value of the variable with role "group" are marked
as "belonging together".
For each resampling iteration, observations of the same group will be exclusively assigned to be either in
the training set or in the test set.
Not more than a single column can be associated with this role.
\item \code{"stratum"}: Stratification variables. Multiple discrete columns may have this role.
\item \code{"weights_learner"}: If the task has observation weights with this role,
they are passed to the \link{Learner} during train.
The use of weights can be disabled by setting the learner's hyperparameter \code{use_weights} to \code{FALSE}.
\item \code{"weights_measure"}: If the task has observation weights with this role,
they are passed to the \link{Measure} for weighted scoring.
The use of weights can be disabled by setting the measure's hyperparameter \code{use_weights} to \code{FALSE}.
\item \code{"offset"}: Numeric columns used to specify fixed adjustments for model training.
Some models use offsets to simply shift predictions,
while others incorporate them to boost predictions from a baseline model.
For learners supporting offsets in multiclass settings,
an offset column must be provided for each target class.
These columns must follow the naming convention \code{"offset_{target_class_name}"}.
For an example of a learner that supports offsets, see \code{LearnerClassifXgboost}  of \CRANpkg{mlr3learners}.
}

\code{col_roles} is a named list whose elements are named by column role and each element is a \code{character()} vector
of column names.
To alter the roles, just modify the list,
e.g. with \R's set functions (\code{\link[=intersect]{intersect()}}, \code{\link[=setdiff]{setdiff()}}, \code{\link[=union]{union()}}, \ldots).
The method \verb{$set_col_roles} provides a convenient alternative to assign columns to roles.

The roles \code{weights_learner} and \code{weights_measure} may only point to a single numeric column, but they can
all point to the same column or different columns.
Weights must be non-negative numerics with at least one weight being > 0.
They don't necessarily need to sum up to 1.}

    \item{\code{nrow}}{(\code{integer(1)})\cr
Returns the total number of rows with role "use".}

    \item{\code{ncol}}{(\code{integer(1)})\cr
Returns the total number of columns with role "target" or "feature".}

    \item{\code{n_features}}{(\code{integer(1)})\cr
Returns the total number of columns with role "feature" (i.e. the number of "active" features in the task).}

    \item{\code{feature_types}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Returns a table with columns \code{id} and \code{type} where \code{id} are the column names of "active"
features of the task and \code{type} is the storage type.}

    \item{\code{strata}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
If the task has columns designated with role \code{"stratum"},
returns a table with one subpopulation per row and two columns:
\itemize{
\item \code{N} (\code{integer()}) with the number of observations in the subpopulation, and
\item \code{row_id} (list of \code{integer()}) as list column with the row ids in the respective subpopulation.
Returns \code{NULL} if there is no stratification variable.
See \link{Resampling} for more information on stratification.
}}

    \item{\code{groups}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
If the task has a column with designated role \code{"group"}, a table with two columns:
\itemize{
\item \code{row_id} (\code{integer()}), and
\item grouping variable \code{group} (\code{vector()}).
}

Returns \code{NULL} if there is no grouping column.
See \link{Resampling} for more information on grouping.}

    \item{\code{order}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
If the task has at least one column with designated role \code{"order"}, a table with two columns:
\itemize{
\item \code{row_id} (\code{integer()}), and
\item ordering vector \code{order} (\code{integer()}).
}

Returns \code{NULL} if there is no order column.}

    \item{\code{weights}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Deprecated, use \verb{$weights_learner} instead.}

    \item{\code{weights_learner}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Returns the observation weights used for training a \link{Learner} (column role \code{weights_learner})
as a \code{data.table} with the following columns:
\itemize{
\item \code{row_id} (\code{integer()}), and
\item \code{weight} (\code{numeric()}).
}

Returns \code{NULL} if there is no column with the designated role.}

    \item{\code{weights_measure}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Returns the observation weights used for scoring a prediction with a \link{Measure} (column role \code{weights_measure})
as a \code{data.table} with the following columns:
\itemize{
\item \code{row_id} (\code{integer()}), and
\item \code{weight} (\code{numeric()}).
}

Returns \code{NULL} if there is no column with the designated role.}

    \item{\code{offset}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
If the task has a column with designated role \code{"offset"}, a table with two or more columns:
\itemize{
\item \code{row_id} (\code{integer()}), and
\item offset variable(s) (\code{numeric()}).
}

For regression or binary classification tasks, there will be only a single-column offset.
For multiclass tasks, it may return multiple offset columns, one for each target class.
If there is only one offset column, it will be named as \code{offset}.

If there are no columns with the \code{"offset"} role, \code{NULL} is returned.}

    \item{\code{labels}}{(named \code{character()})\cr
Retrieve \code{labels} (prettier formatted names) from columns.
Internally queries the column \code{label} of the table in field \code{col_info}.
Columns ids referenced by the name of the vector, the labels are the actual string values.

Assigning to this column update the task by reference.
You have to provide a character vector of labels, named with column ids.
To remove a label, set it to \code{NA}.
Alternatively, you can provide a \code{\link[=data.frame]{data.frame()}} with the two columns
\code{"id"} and \code{"label"}.}

    \item{\code{col_hashes}}{(named \code{character})\cr
Hash (unique identifier) for all columns except the \code{primary_key}: A \code{character} vector, named by the columns that each element refers to.\cr
Columns of different \code{\link{Task}}s or \code{\link{DataBackend}}s that have agreeing \code{col_hashes} always represent the same data, given that the same \code{row}s are selected.
The reverse is not necessarily true: There can be columns with the same content that have different \code{col_hashes}.}

    \item{\code{characteristics}}{(\code{list()})\cr
List of characteristics of the task, e.g. \code{list(n = 5, p = 7)}.}

    \item{\code{row_ids_backend}}{(\code{integer()})\cr
Returns all row ids from the backend, regardless of their roles.
This is different from \verb{$row_ids} which only returns rows with role "use".}

    \item{\code{label}}{(\code{character(1)})\cr
Label for this object.
Can be used in tables, plot and text output instead of the ID.}

    \item{\code{task_type}}{(\code{character(1)})\cr
Task type, e.g. \code{"classif"} or \code{"regr"}.

For a complete list of possible task types (depending on the loaded packages),
see \code{\link[=mlr_reflections]{mlr_reflections$task_types$type}}.}

    \item{\code{man}}{(\code{character(1)} | \code{NULL})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
Defaults to \code{NA}, but can be set by child classes.}

    \item{\code{extra_args}}{(named \code{list()})\cr
Additional arguments set during construction.
Required for \code{\link[=convert_task]{convert_task()}}.}

    \item{\code{backend}}{(\link{DataBackend})\cr
Abstract interface to the data of the task.}

    \item{\code{col_info}}{(\code{\link[data.table:data.table]{data.table::data.table()}})\cr
Table with 4 columns, mainly for internal purposes:
\itemize{
\item \code{"id"} (\code{character()}) stores the name of the column.
\item \code{"type"} (\code{character()}) holds the storage type of the variable, e.g. \code{integer}, \code{numeric} or \code{character}.
See \link[=mlr_reflections]{mlr_reflections$task_feature_types} for a complete list of allowed types.
\item \code{"levels"} (\code{list()}) stores a vector of distinct values (levels) for ordered and unordered factor variables.
\item \code{"label"} (\code{character()}) stores a vector of prettier, formatted column names.
\item \code{"fix_factor_levels"} (\code{logical()}) stores flags which determine if the levels of the respective variable
need to be reordered after querying the data from the \link{DataBackend}.
}

Note that all columns of the \link{DataBackend}, also columns which are not selected or have any role, are listed
in this table.}

    \item{\code{mlr3_version}}{(\code{package_version})\cr
Package version of \code{mlr3} used to create the task.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Task-initialize}{\code{Task$new()}}
    \item \href{#method-Task-help}{\code{Task$help()}}
    \item \href{#method-Task-format}{\code{Task$format()}}
    \item \href{#method-Task-print}{\code{Task$print()}}
    \item \href{#method-Task-data}{\code{Task$data()}}
    \item \href{#method-Task-formula}{\code{Task$formula()}}
    \item \href{#method-Task-head}{\code{Task$head()}}
    \item \href{#method-Task-levels}{\code{Task$levels()}}
    \item \href{#method-Task-missings}{\code{Task$missings()}}
    \item \href{#method-Task-filter}{\code{Task$filter()}}
    \item \href{#method-Task-select}{\code{Task$select()}}
    \item \href{#method-Task-rbind}{\code{Task$rbind()}}
    \item \href{#method-Task-cbind}{\code{Task$cbind()}}
    \item \href{#method-Task-rename}{\code{Task$rename()}}
    \item \href{#method-Task-set_row_roles}{\code{Task$set_row_roles()}}
    \item \href{#method-Task-set_col_roles}{\code{Task$set_col_roles()}}
    \item \href{#method-Task-set_levels}{\code{Task$set_levels()}}
    \item \href{#method-Task-droplevels}{\code{Task$droplevels()}}
    \item \href{#method-Task-add_strata}{\code{Task$add_strata()}}
    \item \href{#method-Task-materialize_view}{\code{Task$materialize_view()}}
    \item \href{#method-Task-clone}{\code{Task$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Task-initialize}{}}}
\subsection{\code{Task$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.

Note that this object is typically constructed via a derived classes, e.g. \link{TaskClassif} or \link{TaskRegr}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$new(id, task_type, backend, label = NA_character_, extra_args = list())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier for the new instance.}
      \item{\code{task_type}}{(\code{character(1)})\cr
Type of task, e.g. \code{"regr"} or \code{"classif"}.
Must be an element of \link[=mlr_reflections]{mlr_reflections$task_types$type}.}
      \item{\code{backend}}{(\link{DataBackend})\cr
Either a \link{DataBackend}, or any object which is convertible to a \link{DataBackend} with \code{as_data_backend()}.
E.g., a \code{data.frame()} will be converted to a \link{DataBackendDataTable}.}
      \item{\code{label}}{(\code{character(1)})\cr
Label for the new instance.}
      \item{\code{extra_args}}{(named \code{list()})\cr
Named list of constructor arguments, required for converting task types
via \code{\link[=convert_task]{convert_task()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-help"></a>}}
\if{latex}{\out{\hypertarget{method-Task-help}{}}}
\subsection{\code{Task$help()}}{
  Opens the corresponding help page referenced by field \verb{$man}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$help()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-format"></a>}}
\if{latex}{\out{\hypertarget{method-Task-format}{}}}
\subsection{\code{Task$format()}}{
  Helper for print outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$format(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-print"></a>}}
\if{latex}{\out{\hypertarget{method-Task-print}{}}}
\subsection{\code{Task$print()}}{
  Printer.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{(ignored).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-data"></a>}}
\if{latex}{\out{\hypertarget{method-Task-data}{}}}
\subsection{\code{Task$data()}}{
  Returns a slice of the data from the \link{DataBackend} as a \code{data.table}.
Rows default to observations with role \code{"use"},
and columns default to features with roles \code{"target"} or \code{"feature"}.
Rows must be a subset of \verb{$row_ids}.
If \code{rows} or \code{cols} are specified which do not exist in the \link{DataBackend}, an exception is raised.

Rows and columns are returned in the order specified via the arguments \code{rows} and \code{cols}.
If \code{rows} is \code{NULL}, rows are returned in the order of \code{task$row_ids}.
If \code{cols} is \code{NULL}, the column order defaults to \code{c(task$target_names, task$feature_names)}.
Note that it is recommended to \strong{not} rely on the order of columns,
and instead always address columns with their respective column name.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$data(rows = NULL, cols = NULL, ordered = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rows}}{(positive \code{integer()} | \code{NULL})\cr
Vector or row indices.
Always refers to the complete data set, even after filtering.}
      \item{\code{cols}}{(\code{character()} | \code{NULL})\cr
Vector of column names.}
      \item{\code{ordered}}{(\code{logical(1)})\cr
If \code{TRUE}, data is ordered according to the columns with column role \code{"order"}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Depending on the \link{DataBackend}, but usually a \code{\link[data.table:data.table]{data.table::data.table()}}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$data(rows = 1:5, cols = c("species", "sex"))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-formula"></a>}}
\if{latex}{\out{\hypertarget{method-Task-formula}{}}}
\subsection{\code{Task$formula()}}{
  Constructs a \code{\link[=formula]{formula()}}, e.g. \verb{[target] ~ [feature_1] + [feature_2] + ... + [feature_k]},
using the features provided in argument \code{rhs}
(defaults to all columns with role \code{"feature"}, symbolized by \code{"."}).

Note that it is currently not possible to change the formula.
However, \CRANpkg{mlr3pipelines} provides a pipe operator interfacing \code{\link[stats:model.matrix]{stats::model.matrix()}} for this
purpose: \code{"modelmatrix"}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$formula(rhs = ".")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rhs}}{(\code{character(1)})\cr
Right hand side of the formula. Defaults to \code{"."} (all features of the task).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{\link[=formula]{formula()}}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$formula()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-head"></a>}}
\if{latex}{\out{\hypertarget{method-Task-head}{}}}
\subsection{\code{Task$head()}}{
  Get the first \code{n} observations with role \code{"use"} of all columns with role \code{"target"} or \code{"feature"}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$head(n = 6L)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{(\code{integer(1)}).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{\link[data.table:data.table]{data.table::data.table()}} with \code{n} rows.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$head(3)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-levels"></a>}}
\if{latex}{\out{\hypertarget{method-Task-levels}{}}}
\subsection{\code{Task$levels()}}{
  Returns the distinct values for columns referenced in \code{cols} with storage type "factor" or "ordered".
Argument \code{cols} defaults to all such columns with role \code{"target"} or \code{"feature"}.

Note that this function ignores the row roles, it returns all levels available in the \link{DataBackend}.
To update the stored level information, e.g. after subsetting a task with \verb{$filter()}, call \verb{$droplevels()}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$levels(cols = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{cols}}{(\code{character()} | \code{NULL})\cr
Vector of column names.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    named \code{list()}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$levels()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-missings"></a>}}
\if{latex}{\out{\hypertarget{method-Task-missings}{}}}
\subsection{\code{Task$missings()}}{
  Returns the number of missing observations for columns referenced in \code{cols}.
Considers only active rows with row role \code{"use"}.
Argument \code{cols} defaults to all columns with role "target" or "feature".
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$missings(cols = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{cols}}{(\code{character()} | \code{NULL})\cr
Vector of column names.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Named \code{integer()}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$missings()
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-filter"></a>}}
\if{latex}{\out{\hypertarget{method-Task-filter}{}}}
\subsection{\code{Task$filter()}}{
  Subsets the task, keeping only the rows specified via row ids \code{rows}.

This operation mutates the task in-place.
See the section on task mutators for more information.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$filter(rows)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rows}}{(positive \code{integer()} | \code{NULL})\cr
Vector or row indices.
Always refers to the complete data set, even after filtering.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$filter(1:10)
task$nrow
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-select"></a>}}
\if{latex}{\out{\hypertarget{method-Task-select}{}}}
\subsection{\code{Task$select()}}{
  Subsets the task, keeping only the features specified via column names \code{cols}.
Note that you cannot deselect the target column, for obvious reasons.

This operation mutates the task in-place.
See the section on task mutators for more information.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$select(cols)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{cols}}{(\code{character()} | \code{NULL})\cr
Vector of column names.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$select(c("bill_length", "bill_depth"))
task$feature_names
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-rbind"></a>}}
\if{latex}{\out{\hypertarget{method-Task-rbind}{}}}
\subsection{\code{Task$rbind()}}{
  Adds additional rows to the \link{DataBackend} stored in \verb{$backend}.
New row ids are automatically created, unless \code{data} has a column whose name matches
the primary key of the \link{DataBackend} (\code{task$backend$primary_key}).
In case of name clashes of row ids, rows in \code{data} have higher precedence
and virtually overwrite the rows in the \link{DataBackend}.

All columns roles \code{"target"}, \code{"feature"}, \code{"weights_learner"}, \code{"weights_measure"},
\code{"group"}, \code{"stratum"}, and \code{"order"} must be present in \code{data}.
Columns only present in \code{data} but not in the \link{DataBackend} of \code{task} will be discarded.

This operation mutates the task in-place.
See the section on task mutators for more information.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$rbind(data)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{data}}{(\code{data.frame()}).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
extra = task$data(rows = 1:2)
task$rbind(extra)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-cbind"></a>}}
\if{latex}{\out{\hypertarget{method-Task-cbind}{}}}
\subsection{\code{Task$cbind()}}{
  Adds additional columns to the \link{DataBackend} stored in \verb{$backend}.

The row ids must be provided as column in \code{data}
(with column name matching the primary key name of the \link{DataBackend}).
If this column is missing, it is assumed that the rows are exactly in the order of \verb{$row_ids}.
In case of name clashes of column names in \code{data} and \link{DataBackend}, columns in \code{data} have higher precedence
and virtually overwrite the columns in the \link{DataBackend}.

This operation mutates the task in-place.
See the section on task mutators for more information.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$cbind(data)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{data}}{(\code{data.frame()}).}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$cbind(data.table(extra_col = seq_len(task$nrow)))
head(task$data(cols = "extra_col"))
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-rename"></a>}}
\if{latex}{\out{\hypertarget{method-Task-rename}{}}}
\subsection{\code{Task$rename()}}{
  Renames columns by mapping column names in \code{old} to new column names in \code{new} (element-wise).

This operation mutates the task in-place.
See the section on task mutators for more information.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$rename(old, new)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{old}}{(\code{character()})\cr
Old names.}
      \item{\code{new}}{(\code{character()})\cr
New names.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$rename("body_mass", "mass")
task$feature_names
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-set_row_roles"></a>}}
\if{latex}{\out{\hypertarget{method-Task-set_row_roles}{}}}
\subsection{\code{Task$set_row_roles()}}{
  Modifies the roles in \verb{$row_roles} \strong{in-place}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$set_row_roles(rows, roles = NULL, add_to = NULL, remove_from = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{rows}}{(\code{integer()})\cr
Row ids for which to change the roles for.}
      \item{\code{roles}}{(\code{character()})\cr
Exclusively set rows to the specified \code{roles} (remove from other roles).}
      \item{\code{add_to}}{(\code{character()})\cr
Add rows with row ids \code{rows} to roles specified in \code{add_to}.
Rows keep their previous roles.}
      \item{\code{remove_from}}{(\code{character()})\cr
Remove rows with row ids \code{rows} from roles specified in \code{remove_from}.
Other row roles are preserved.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Details}{
    Roles are first set exclusively (argument \code{roles}), then added (argument \code{add_to}) and finally
removed (argument \code{remove_from}) from different roles.
Duplicated row ids are explicitly allowed, so you can replicate an observation by repeating its
\code{row_id}.
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$set_row_roles(1:5, remove_from = "use")
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-set_col_roles"></a>}}
\if{latex}{\out{\hypertarget{method-Task-set_col_roles}{}}}
\subsection{\code{Task$set_col_roles()}}{
  Modifies the roles in \verb{$col_roles} \strong{in-place}.
See \verb{$col_roles} for a list of possible roles.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$set_col_roles(cols, roles = NULL, add_to = NULL, remove_from = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{cols}}{(\code{character()})\cr
Column names for which to change the roles for.}
      \item{\code{roles}}{(\code{character()})\cr
Exclusively set columns to the specified \code{roles} (remove from other roles).}
      \item{\code{add_to}}{(\code{character()})\cr
Add columns with column names \code{cols} to roles specified in \code{add_to}.
Columns keep their previous roles.}
      \item{\code{remove_from}}{(\code{character()})\cr
Remove columns with columns names \code{cols} from roles specified in \code{remove_from}.
Other column roles are preserved.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Details}{
    Roles are first set exclusively (argument \code{roles}), then added (argument \code{add_to}) and finally removed
(argument \code{remove_from}) from different roles.
Duplicated columns are removed from the same role.
For tasks that only allow one target, the target column cannot be set with \verb{$set_col_roles()}.
Use the \verb{$col_roles} field to swap the target column.
  }
  \subsection{Returns}{
    Returns the object itself, but modified \strong{by reference}.
You need to explicitly \verb{$clone()} the object beforehand if you want to keep
the object in its previous state.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$set_col_roles("sex", roles = "stratum")
task$col_roles$stratum
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-set_levels"></a>}}
\if{latex}{\out{\hypertarget{method-Task-set_levels}{}}}
\subsection{\code{Task$set_levels()}}{
  Set levels for columns of type \code{factor} and \code{ordered} in field \code{col_info}.
You can add, remove or reorder the levels, affecting the data returned by
\verb{$data()} and \verb{$levels()}.
If you just want to remove unused levels, use \verb{$droplevels()} instead.

Note that factor levels which are present in the data but not listed in the task as
valid levels are converted to missing values.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$set_levels(levels)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{levels}}{(named \code{list()} of \code{character()})\cr
List of character vectors of new levels, named by column names.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified \code{self}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$set_levels(list(sex = c("male", "female", "unknown")))
task$levels("sex")
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-droplevels"></a>}}
\if{latex}{\out{\hypertarget{method-Task-droplevels}{}}}
\subsection{\code{Task$droplevels()}}{
  Updates the cache of stored factor levels, removing all levels not present in the current set of active rows.
\code{cols} defaults to all columns with storage type "factor" or "ordered".
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$droplevels(cols = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{cols}}{(\code{character()} | \code{NULL})\cr
Vector of column names.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified \code{self}.
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$set_levels(list(sex = c("male", "female", "unknown")))
task$levels("sex")
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-add_strata"></a>}}
\if{latex}{\out{\hypertarget{method-Task-add_strata}{}}}
\subsection{\code{Task$add_strata()}}{
  Cuts numeric variables into new factors columns which are added to the task with role
\code{"stratum"}.
This ensures that all training and test splits contain observations from all bins.
The columns are named \code{"..stratum_[col_name]"}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$add_strata(cols, bins = 3L)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{cols}}{(\code{character()})\cr
Names of columns to operate on.}
      \item{\code{bins}}{(\code{integer()})\cr
Number of bins to cut into (passed to \code{\link[=cut]{cut()}} as \code{breaks}).
Replicated to have the same length as \code{cols}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    self (invisibly).
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("penguins")
task$add_strata("flipper_length", bins = 4)
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-materialize_view"></a>}}
\if{latex}{\out{\hypertarget{method-Task-materialize_view}{}}}
\subsection{\code{Task$materialize_view()}}{
  Certain operations change the view on the data, e.g., \verb{$filter()} or \verb{$select()}.
This operation queries the \link{DataBackend} for all data required in the active view and
replaces the internal \link{DataBackend} with the new one. In some scenarios this helps to
free up memory or speeds up accesses to the data, especially after several \verb{$rbind()}
and \verb{$cbind()} operations.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$materialize_view(internal_valid_task = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{internal_valid_task}}{(\code{logical(1)})\cr
Also materialize the internal validation task. Default is \code{TRUE}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Details}{
    For tasks containing the same observation more than once (duplicates in \verb{$row_ids}),
the resulting backend contains it only once.
  }
  \subsection{Returns}{
    self (invisibly).
  }
  \subsection{Examples}{
    \if{html}{\out{<div class="r example copy">}}
    \preformatted{task = tsk("iris")
task$backend$nrow
task$filter(1:120)
task$backend$nrow
}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Task-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Task-clone}{}}}
\subsection{\code{Task$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Task$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LearnerRegrFeatureless.R
\name{mlr_learners_regr.featureless}
\alias{mlr_learners_regr.featureless}
\alias{LearnerRegrFeatureless}
\title{Featureless Regression Learner}
\description{
A simple \link{LearnerRegr} which only analyzes the response during train, ignoring all features.
If hyperparameter \code{robust} is \code{FALSE} (default), constantly predicts \code{mean(y)} as response
and \code{sd(y)} as standard error.
If \code{robust} is \code{TRUE}, \code{\link[=median]{median()}} and \code{\link[=mad]{mad()}} are used instead of \code{\link[=mean]{mean()}} and \code{\link[=sd]{sd()}},
respectively.

For weighted data, the response is the weighted mean (weighted median for robust regression).
The predicted standard error is the square root of the weighted variance estimator with bias correction
based on effective degrees of freedom:

\if{html}{\out{<div class="sourceCode">}}\preformatted{sd(y, weights) = sqrt(
  sum(weights * (y - weighted.mean(y, weights))^2) /
    (sum(weights) - sum(weights ^2) / sum(weights))
)
}\if{html}{\out{</div>}}

If \code{robust} is \code{TRUE}, the weighted median absolute deviation is used, adjusted by a factor of 1.4826
for consistency with \code{\link[=mad]{mad()}}.
}
\section{Dictionary}{

This \link{Learner} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_learners} or with the associated sugar function \code{\link[=lrn]{lrn()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_learners$get("regr.featureless")
lrn("regr.featureless")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{regr}
\item Predict Types: \dQuote{response}, \dQuote{se}, \dQuote{quantiles}
\item Feature Types: \dQuote{logical}, \dQuote{integer}, \dQuote{numeric}, \dQuote{character}, \dQuote{factor}, \dQuote{ordered}, \dQuote{POSIXct}, \dQuote{Date}
\item Required Packages: \CRANpkg{mlr3}, 'stats'
}
}

\section{Parameters}{
\tabular{llll}{
   Id \tab Type \tab Default \tab Levels \cr
   robust \tab logical \tab TRUE \tab TRUE, FALSE \cr
}
}

\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-learners}
\item Package \CRANpkg{mlr3learners} for a solid collection of essential learners.
\item Package \href{https://github.com/mlr-org/mlr3extralearners}{mlr3extralearners} for more learners.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Learner]{Learners}: \link{mlr_learners}
\item \code{as.data.table(mlr_learners)} for a table of available \link[=Learner]{Learners} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3pipelines} to combine learners with pre- and postprocessing steps.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
\item \CRANpkg{mlr3tuning} for tuning of hyperparameters, \CRANpkg{mlr3tuningspaces}
for established default tuning spaces.
}

Other Learner:
\code{\link{Learner}},
\code{\link{LearnerClassif}},
\code{\link{LearnerRegr}},
\code{\link{mlr_learners}},
\code{\link{mlr_learners_classif.debug}},
\code{\link{mlr_learners_classif.featureless}},
\code{\link{mlr_learners_classif.rpart}},
\code{\link{mlr_learners_regr.debug}},
\code{\link{mlr_learners_regr.rpart}}
}
\concept{Learner}
\section{Super classes}{
\code{\link[mlr3:Learner]{Learner}} -> \code{\link[mlr3:LearnerRegr]{LearnerRegr}} -> \code{LearnerRegrFeatureless}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-LearnerRegrFeatureless-initialize}{\code{LearnerRegrFeatureless$new()}}
    \item \href{#method-LearnerRegrFeatureless-importance}{\code{LearnerRegrFeatureless$importance()}}
    \item \href{#method-LearnerRegrFeatureless-selected_features}{\code{LearnerRegrFeatureless$selected_features()}}
    \item \href{#method-LearnerRegrFeatureless-clone}{\code{LearnerRegrFeatureless$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="base_learner"><a href='../../mlr3/html/Learner.html#method-Learner-base_learner'><code>Learner$base_learner()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="configure"><a href='../../mlr3/html/Learner.html#method-Learner-configure'><code>Learner$configure()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="encapsulate"><a href='../../mlr3/html/Learner.html#method-Learner-encapsulate'><code>Learner$encapsulate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="format"><a href='../../mlr3/html/Learner.html#method-Learner-format'><code>Learner$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="help"><a href='../../mlr3/html/Learner.html#method-Learner-help'><code>Learner$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict"><a href='../../mlr3/html/Learner.html#method-Learner-predict'><code>Learner$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict_newdata"><a href='../../mlr3/html/Learner.html#method-Learner-predict_newdata'><code>Learner$predict_newdata()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="print"><a href='../../mlr3/html/Learner.html#method-Learner-print'><code>Learner$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="reset"><a href='../../mlr3/html/Learner.html#method-Learner-reset'><code>Learner$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="train"><a href='../../mlr3/html/Learner.html#method-Learner-train'><code>Learner$train()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="LearnerRegr" data-id="predict_newdata_fast"><a href='../../mlr3/html/LearnerRegr.html#method-LearnerRegr-predict_newdata_fast'><code>LearnerRegr$predict_newdata_fast()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerRegrFeatureless-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerRegrFeatureless-initialize}{}}}
\subsection{\code{LearnerRegrFeatureless$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerRegrFeatureless$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerRegrFeatureless-importance"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerRegrFeatureless-importance}{}}}
\subsection{\code{LearnerRegrFeatureless$importance()}}{
  All features have a score of \code{0} for this learner.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerRegrFeatureless$importance()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Named \code{numeric()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerRegrFeatureless-selected_features"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerRegrFeatureless-selected_features}{}}}
\subsection{\code{LearnerRegrFeatureless$selected_features()}}{
  Selected features are always the empty set for this learner.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerRegrFeatureless$selected_features()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{character(0)}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerRegrFeatureless-clone"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerRegrFeatureless-clone}{}}}
\subsection{\code{LearnerRegrFeatureless$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerRegrFeatureless$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MeasureClassif.R
\name{MeasureClassif}
\alias{MeasureClassif}
\title{Classification Measure}
\description{
This measure specializes \link{Measure} for classification problems:
\itemize{
\item \code{task_type} is set to \code{"classif"}.
\item Possible values for \code{predict_type} are \code{"response"} and \code{"prob"}.
}

Predefined measures can be found in the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_measures}.
The default measure for classification is \code{\link[=mlr_measures_classif.ce]{classif.ce}}.
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-eval}
\item Package \CRANpkg{mlr3measures} for the scoring functions.
\link[mlr3misc:Dictionary]{Dictionary} of \link[=Measure]{Measures}: \link{mlr_measures}
\code{as.data.table(mlr_measures)} for a table of available \link[=Measure]{Measures} in the running session (depending on the loaded packages).
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Measure:
\code{\link{Measure}},
\code{\link{MeasureRegr}},
\code{\link{MeasureSimilarity}},
\code{\link{mlr_measures}},
\code{\link{mlr_measures_aic}},
\code{\link{mlr_measures_bic}},
\code{\link{mlr_measures_classif.costs}},
\code{\link{mlr_measures_debug_classif}},
\code{\link{mlr_measures_elapsed_time}},
\code{\link{mlr_measures_internal_valid_score}},
\code{\link{mlr_measures_oob_error}},
\code{\link{mlr_measures_regr.pinball}},
\code{\link{mlr_measures_regr.rqr}},
\code{\link{mlr_measures_regr.rsq}},
\code{\link{mlr_measures_selected_features}}
}
\concept{Measure}
\section{Super class}{
\code{\link[mlr3:Measure]{Measure}} -> \code{MeasureClassif}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MeasureClassif-initialize}{\code{MeasureClassif$new()}}
    \item \href{#method-MeasureClassif-clone}{\code{MeasureClassif$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="aggregate"><a href='../../mlr3/html/Measure.html#method-Measure-aggregate'><code>Measure$aggregate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="format"><a href='../../mlr3/html/Measure.html#method-Measure-format'><code>Measure$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="help"><a href='../../mlr3/html/Measure.html#method-Measure-help'><code>Measure$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="obs_loss"><a href='../../mlr3/html/Measure.html#method-Measure-obs_loss'><code>Measure$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="print"><a href='../../mlr3/html/Measure.html#method-Measure-print'><code>Measure$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Measure" data-id="score"><a href='../../mlr3/html/Measure.html#method-Measure-score'><code>Measure$score()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureClassif-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureClassif-initialize}{}}}
\subsection{\code{MeasureClassif$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureClassif$new(
  id,
  param_set = ps(),
  range,
  minimize = NA,
  average = "macro",
  aggregator = NULL,
  properties = character(),
  predict_type = "response",
  predict_sets = "test",
  task_properties = character(),
  packages = character(),
  label = NA_character_,
  man = NA_character_
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{id}}{(\code{character(1)})\cr
Identifier for the new instance.}
      \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet})\cr
Set of hyperparameters.}
      \item{\code{range}}{(\code{numeric(2)})\cr
Feasible range for this measure as \code{c(lower_bound, upper_bound)}.
Both bounds may be infinite.}
      \item{\code{minimize}}{(\code{logical(1)})\cr
Set to \code{TRUE} if good predictions correspond to small values,
and to \code{FALSE} if good predictions correspond to large values.
If set to \code{NA} (default), tuning this measure is not possible.}
      \item{\code{average}}{(\code{character(1)})\cr
How to average multiple \link{Prediction}s from a \link{ResampleResult}.

The default, \code{"macro"}, calculates the individual performances scores for each \link{Prediction} and then uses the
function defined in \verb{$aggregator} to average them to a single number.

\code{"macro_weighted"} is similar to \code{"macro"}, but uses weighted averages.
Weights are taken from the \code{weights_measure} column of the resampled \link{Task} if present.
Note that \code{"macro_weighted"} can differ from \code{"macro"} even if no weights are present or if \verb{$use_weights} is set to \code{"ignore"},
since then aggregation is done using \emph{uniform sample weights}, which result in non-uniform weights for \link{Prediction}s if they contain different
numbers of samples.

If set to \code{"micro"}, the individual \link{Prediction} objects are first combined into a single new \link{Prediction} object which is then used to assess the performance.
The function in \verb{$aggregator} is not used in this case.}
      \item{\code{aggregator}}{(\verb{function()} | \code{NULL})\cr
Function to aggregate over multiple iterations. The role of this function depends on
the value of field \code{"average"}:
\itemize{
\item \code{"macro"}: A numeric vector of scores (one per iteration) is passed.
The aggregate function defaults to \code{\link[=mean]{mean()}} in this case.
\item \code{"micro"}: The \code{aggregator} function is not used.
Instead, predictions from multiple iterations are first combined and then
scored in one go.
\item \code{"custom"}: A \link{ResampleResult} is passed to the aggregate function.
}}
      \item{\code{properties}}{(\code{character()})\cr
Properties of the measure.
Must be a subset of \link[=mlr_reflections]{mlr_reflections$measure_properties}.
Supported by \code{mlr3}:
\itemize{
\item \code{"requires_task"} (requires the complete \link{Task}),
\item \code{"requires_learner"} (requires the trained \link{Learner}),
\item \code{"requires_model"} (requires the trained \link{Learner}, including the fitted model),
\item \code{"requires_train_set"} (requires the training indices from the \link{Resampling}),
\item \code{"na_score"} (the measure is expected to occasionally return \code{NA} or \code{NaN}),
\item \code{"weights"} (support weighted scoring using sample weights from task, column role \code{weights_measure}), and
\item \code{"primary_iters"} (the measure explicitly handles resamplings
that only use a subset of their iterations for the point estimate)
\item \code{"requires_no_prediction"} (No prediction is required;
This usually means that the measure extracts some information from the learner state.).
}}
      \item{\code{predict_type}}{(\code{character(1)})\cr
Required predict type of the \link{Learner}.
Possible values are stored in \link[=mlr_reflections]{mlr_reflections$learner_predict_types}.}
      \item{\code{predict_sets}}{(\code{character()})\cr
Prediction sets to operate on, used in \code{aggregate()} to extract the matching \code{predict_sets} from the \link{ResampleResult}.
Multiple predict sets are calculated by the respective \link{Learner} during \code{\link[=resample]{resample()}}/\code{\link[=benchmark]{benchmark()}}.
Must be a non-empty subset of \verb{\{"train", "test", "internal_valid"\}}.
If multiple sets are provided, these are first combined to a single prediction object.
Default is \code{"test"}.}
      \item{\code{task_properties}}{(\code{character()})\cr
Required task properties, see \link{Task}.}
      \item{\code{packages}}{(\code{character()})\cr
Set of required packages.
A warning is signaled by the constructor if at least one of the packages is not installed,
but loaded (not attached) later on-demand via \code{\link[=requireNamespace]{requireNamespace()}}.}
      \item{\code{label}}{(\code{character(1)})\cr
Label for the new instance.}
      \item{\code{man}}{(\code{character(1)})\cr
String in the format \verb{[pkg]::[topic]} pointing to a manual page for this object.
The referenced help package can be opened via method \verb{$help()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MeasureClassif-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MeasureClassif-clone}{}}}
\subsection{\code{MeasureClassif$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MeasureClassif$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_result_data.R
\name{as_result_data}
\alias{as_result_data}
\title{Convert to ResultData}
\usage{
as_result_data(
  task,
  learners,
  resampling,
  iterations,
  predictions,
  learner_states = NULL,
  data_extra = NULL,
  store_backends = TRUE
)
}
\arguments{
\item{task}{(\link{Task}).}

\item{learners}{(list of trained \link{Learner}s).}

\item{resampling}{(\link{Resampling}).}

\item{iterations}{(\code{integer()}).}

\item{predictions}{(list of list of \link{Prediction}s).}

\item{learner_states}{(\code{list()})\cr
Learner states. If not provided, the states of \code{learners} are automatically extracted.}

\item{data_extra}{(\code{list()})\cr
Additional data for each iteration.}

\item{store_backends}{(\code{logical(1)})\cr
If set to \code{FALSE}, the backends of the \link{Task}s provided in \code{data} are
removed.}
}
\value{
\code{ResultData} object which can be passed to the constructor of \link{ResampleResult}.
}
\description{
This function allows to construct or convert to a \link{ResultData} object, the
result container used by \link{ResampleResult} and \link{BenchmarkResult}.
A \link{ResampleResult} or \link{BenchmarkResult} can be initialized with the returned object.
Note that \link{ResampleResult}s can be converted to a \link{BenchmarkResult} with \code{\link[=as_benchmark_result]{as_benchmark_result()}}
and multiple \link{BenchmarkResult}s can be combined to a larger \link{BenchmarkResult} with the
\verb{$combine()} method of \link{BenchmarkResult}.
}
\examples{
task = tsk("penguins")
learner = lrn("classif.rpart")
resampling = rsmp("cv", folds = 2)$instantiate(task)
iterations = seq_len(resampling$iters)

# manually train two learners.
# store learners and predictions
learners = list()
predictions = list()
for (i in iterations) {
  l = learner$clone(deep = TRUE)
  learners[[i]] = l$train(task, row_ids = resampling$train_set(i))
  predictions[[i]] = list(test = l$predict(task, row_ids = resampling$test_set(i)))
}

rdata = as_result_data(task, learners, resampling, iterations, predictions)
ResampleResult$new(rdata)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_empty_ellipsis}
\alias{assert_empty_ellipsis}
\title{Assert Empty Ellipsis}
\usage{
assert_empty_ellipsis(...)
}
\arguments{
\item{...}{(any)\cr
Ellipsis arguments to check.}
}
\value{
\code{NULL}
}
\description{
Assert that \code{...} arguments are empty.
Use this function in S3-methods to ensure that misspelling of arguments does not go unnoticed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_measure.R
\name{as_measure}
\alias{as_measure}
\alias{as_measure.NULL}
\alias{as_measure.Measure}
\alias{as_measures}
\alias{as_measures.default}
\alias{as_measures.NULL}
\alias{as_measures.list}
\title{Convert to a Measure}
\usage{
as_measure(x, task_type = NULL, clone = FALSE, ...)

\method{as_measure}{`NULL`}(x, task_type = NULL, clone = FALSE, ...)

\method{as_measure}{Measure}(x, task_type = NULL, clone = FALSE, ...)

as_measures(x, task_type = NULL, clone = FALSE, ...)

\method{as_measures}{default}(x, task_type = NULL, clone = FALSE, ...)

\method{as_measures}{`NULL`}(x, task_type = NULL, clone = FALSE, ...)

\method{as_measures}{list}(x, task_type = NULL, clone = FALSE, ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{task_type}{(\code{character(1)})\cr
Used if \code{x} is \code{NULL} to construct a default measure for the respective task type.
The default measures are stored in \code{\link[=mlr_reflections]{mlr_reflections$default_measures}}.}

\item{clone}{(\code{logical(1)})\cr
If \code{TRUE}, ensures that the returned object is not the same as the input \code{x}.}

\item{...}{(any)\cr
Additional arguments.}
}
\value{
\link{Measure}.
}
\description{
Convert object to a \link{Measure} or a list of \link{Measure}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_prediction_classif.R
\name{as_prediction_classif}
\alias{as_prediction_classif}
\alias{as_prediction_classif.PredictionClassif}
\alias{as_prediction_classif.data.frame}
\title{Convert to a Classification Prediction}
\usage{
as_prediction_classif(x, ...)

\method{as_prediction_classif}{PredictionClassif}(x, ...)

\method{as_prediction_classif}{data.frame}(x, ...)
}
\arguments{
\item{x}{(any)\cr
Object to convert.}

\item{...}{(any)\cr
Additional arguments.}
}
\value{
\link{PredictionClassif}.
}
\description{
Convert object to a \link{PredictionClassif}.
}
\examples{
# create a prediction object
task = tsk("penguins")
learner = lrn("classif.rpart", predict_type = "prob")
learner$train(task)
p = learner$predict(task)

# convert to a data.table
tab = as.data.table(p)

# convert back to a Prediction
as_prediction_classif(tab)

# split data.table into a list of data.tables
tabs = split(tab, tab$truth)

# convert back to list of predictions
preds = lapply(tabs, as_prediction_classif)

# calculate performance in each group
sapply(preds, function(p) p$score())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{package}
\name{mlr3-package}
\alias{mlr3}
\alias{mlr3-package}
\title{mlr3: Machine Learning in R - Next Generation}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

Efficient, object-oriented programming on the building blocks of machine learning. Provides 'R6' objects for tasks, learners, resamplings, and measures. The package is geared towards scalability and larger datasets by supporting parallelization and out-of-memory data-backends like databases. While 'mlr3' focuses on the core computational operations, add-on packages provide additional functionality.
}
\section{Learn mlr3}{

\itemize{
\item Book on mlr3: \url{https://mlr3book.mlr-org.com}
\item Use cases and examples gallery: \url{https://mlr3gallery.mlr-org.com}
\item Cheat Sheets: \url{https://github.com/mlr-org/mlr3cheatsheets}
}
}

\section{mlr3 extensions}{

\itemize{
\item Preprocessing and machine learning pipelines: \CRANpkg{mlr3pipelines}
\item Analysis of benchmark experiments: \CRANpkg{mlr3benchmark}
\item More classification and regression tasks: \CRANpkg{mlr3data}
\item Connector to \href{https://www.openml.org}{OpenML}: \CRANpkg{mlr3oml}
\item Solid selection of good classification and regression learners: \CRANpkg{mlr3learners}
\item Even more learners: \url{https://github.com/mlr-org/mlr3extralearners}
\item Tuning of hyperparameters: \CRANpkg{mlr3tuning}
\item Hyperband tuner: \CRANpkg{mlr3hyperband}
\item Visualizations for many \pkg{mlr3} objects: \CRANpkg{mlr3viz}
\item Survival analysis and probabilistic regression: \CRANpkg{mlr3proba}
\item Cluster analysis: \CRANpkg{mlr3cluster}
\item Feature selection filters: \CRANpkg{mlr3filters}
\item Feature selection wrappers: \CRANpkg{mlr3fselect}
\item Interface to real (out-of-memory) data bases: \CRANpkg{mlr3db}
\item Performance measures as plain functions: \CRANpkg{mlr3measures}
\item Resampling methods for spatiotemporal data: \CRANpkg{mlr3spatiotempcv}
\item Data storage and prediction support for spatial objects: \CRANpkg{mlr3spatial}
}
}

\section{Suggested packages}{

\itemize{
\item Parallelization framework: \CRANpkg{future}
\item Progress bars: \CRANpkg{progressr}
\item Encapsulated evaluation: \CRANpkg{evaluate}, \CRANpkg{callr} (external process)
}
}

\section{Package Options}{

\itemize{
\item \code{"mlr3.exec_random"}: Randomize the order of execution in \code{\link[=resample]{resample()}} and \code{\link[=benchmark]{benchmark()}} during
parallelization with \CRANpkg{future}. Defaults to \code{TRUE}.
Note that this does not affect the order of results.
\item \code{"mlr3.exec_chunk_size"}: Number of iterations to perform in a single \code{\link[future:future]{future::future()}} during
parallelization with \CRANpkg{future}. Defaults to 1.
\item \code{"mlr3.exec_chunk_bins"}: Number of bins to split the iterations into. If set, \code{"mlr3.exec_chunk_size"} is ignored.
\item \code{"mlr3.debug"}: If set to \code{TRUE}, parallelization via \CRANpkg{future} is disabled to simplify
debugging and provide more concise tracebacks.
Note that results computed in debug mode use a different seeding mechanism and are \strong{not reproducible}.
\item \code{"mlr3.warn_version_mismatch"}: Set to \code{FALSE} to silence warnings raised during predict if a learner has been
trained with a different version version of mlr3.
\item \code{"mlr3.prob_as_default"}: Set to \code{TRUE} to set the predict type of classification learners to
\code{"prob"} by default (if they support it).
\item \code{"mlr3.mirai_parallelization"}: Compute profile to use for parallelization with \CRANpkg{mirai}.
Defaults to \code{"mlr3_parallelization"}.
\item \code{"mlr3.mirai_encapsulation"}: Compute profile to use for encapsulation with \CRANpkg{mirai}.
Defaults to \code{"mlr3_encapsulation"}.
\item \code{"mlr3.print_class_ratio_threshold"}: Maximum number of rows for which class ratios are computed
when printing a classification task. For tasks with more rows, only class names are shown.
Defaults to \code{1000000}.
}
}

\section{Error Classes}{

\itemize{
\item \code{Mlr3Error}: The base mlr3 error class.
\item \code{Mlr3ErrorConfig}: This error signals that the user has misconfigured something.
By default, this error is not caught when the learner is encapsulated.
\item \code{Mlr3ErrorInput}: This error signals that the input to the function is invalid.
\item \code{Mlr3ErrorLearner}: The base error class for errors related to the learner.
\item \code{Mlr3ErrorLearnerTrain}: This error signals that the learner failed to train the model.
\item \code{Mlr3ErrorLearnerPredict}: This error signals that something went wrong during prediction.
\item \code{Mlr3TimeoutError}: This error signals that the encapsulation during train or predict timed out.
}
}

\section{Warning Classes}{

\itemize{
\item \code{Mlr3Warning}: The base mlr3 warning class.
\item \code{Mlr3WarningConfig}: This warning signals that the user has misconfigured something.
\item \code{Mlr3WarningInput}: This warning signals that the input to the function is invalid.
}
}

\references{
Lang M, Binder M, Richter J, Schratz P, Pfisterer F, Coors S, Au Q, Casalicchio G, Kotthoff L, Bischl B (2019).
\dQuote{mlr3: A modern object-oriented machine learning framework in R.}
\emph{Journal of Open Source Software}.
\doi{10.21105/joss.01903}.
\url{https://joss.theoj.org/papers/10.21105/joss.01903}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://mlr3.mlr-org.com}
  \item \url{https://github.com/mlr-org/mlr3}
  \item Report bugs at \url{https://github.com/mlr-org/mlr3/issues}
}

}
\author{
\strong{Maintainer}: Marc Becker \email{marcbecker@posteo.de} (\href{https://orcid.org/0000-0002-8115-0400}{ORCID})

Authors:
\itemize{
  \item Marc Becker \email{marcbecker@posteo.de} (\href{https://orcid.org/0000-0002-8115-0400}{ORCID})
  \item Michel Lang \email{michellang@gmail.com} (\href{https://orcid.org/0000-0001-9754-0393}{ORCID})
  \item Bernd Bischl \email{bernd_bischl@gmx.net} (\href{https://orcid.org/0000-0001-6002-6980}{ORCID})
  \item Jakob Richter \email{jakob1richter@gmail.com} (\href{https://orcid.org/0000-0003-4481-5554}{ORCID})
  \item Patrick Schratz \email{patrick.schratz@gmail.com} (\href{https://orcid.org/0000-0003-0748-6624}{ORCID})
  \item Martin Binder \email{mlr.developer@mb706.com}
  \item Florian Pfisterer \email{pfistererf@googlemail.com} (\href{https://orcid.org/0000-0001-8867-762X}{ORCID})
  \item Raphael Sonabend \email{raphaelsonabend@gmail.com} (\href{https://orcid.org/0000-0001-9225-4654}{ORCID})
  \item Sebastian Fischer \email{sebf.fischer@gmail.com} (\href{https://orcid.org/0000-0002-9609-3197}{ORCID})
}

Other contributors:
\itemize{
  \item Giuseppe Casalicchio \email{giuseppe.casalicchio@stat.uni-muenchen.de} (\href{https://orcid.org/0000-0001-5324-5966}{ORCID}) [contributor]
  \item Stefan Coors \email{mail@stefancoors.de} (\href{https://orcid.org/0000-0002-7465-2146}{ORCID}) [contributor]
  \item Quay Au \email{quayau@gmail.com} (\href{https://orcid.org/0000-0002-5252-8902}{ORCID}) [contributor]
  \item Lennart Schneider \email{lennart.sch@web.de} (\href{https://orcid.org/0000-0003-4152-5308}{ORCID}) [contributor]
  \item Lona Koers \email{lona.koers@gmail.com} [contributor]
  \item John Zobolas \email{bblodfon@gmail.com} (\href{https://orcid.org/0000-0002-3609-8674}{ORCID}) [contributor]
  \item Maximilian Mücke \email{muecke.maximilian@gmail.com} (\href{https://orcid.org/0009-0000-9432-9795}{ORCID}) [contributor]
  \item Keno Mersmann \email{keno.mersmann@gmail.com} [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PredictionRegr.R
\name{PredictionRegr}
\alias{PredictionRegr}
\title{Prediction Object for Regression}
\description{
This object wraps the predictions returned by a learner of class \link{LearnerRegr}, i.e.
the predicted response and standard error.
Additionally, probability distributions implemented in package \code{distr6} are supported.
}
\examples{
task = tsk("california_housing")
learner = lrn("regr.featureless", predict_type = "se")
p = learner$train(task)$predict(task)
p$predict_types
head(as.data.table(p))
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html}
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
}

Other Prediction:
\code{\link{Prediction}},
\code{\link{PredictionClassif}}
}
\concept{Prediction}
\section{Super class}{
\code{\link[mlr3:Prediction]{Prediction}} -> \code{PredictionRegr}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{response}}{(\code{numeric()})\cr
Access the stored predicted response.}

    \item{\code{se}}{(\code{numeric()})\cr
Access the stored standard error.}

    \item{\code{quantiles}}{(\code{matrix()})\cr
Matrix of predicted quantiles. Observations are in rows, quantile (in ascending order) in columns.}

    \item{\code{distr}}{(\code{VectorDistribution})\cr
Access the stored vector distribution.
Requires package \code{distr6}(in repository \url{https://raphaels1.r-universe.dev}) .}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PredictionRegr-initialize}{\code{PredictionRegr$new()}}
    \item \href{#method-PredictionRegr-clone}{\code{PredictionRegr$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="filter"><a href='../../mlr3/html/Prediction.html#method-Prediction-filter'><code>Prediction$filter()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="format"><a href='../../mlr3/html/Prediction.html#method-Prediction-format'><code>Prediction$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="help"><a href='../../mlr3/html/Prediction.html#method-Prediction-help'><code>Prediction$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="obs_loss"><a href='../../mlr3/html/Prediction.html#method-Prediction-obs_loss'><code>Prediction$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="print"><a href='../../mlr3/html/Prediction.html#method-Prediction-print'><code>Prediction$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Prediction" data-id="score"><a href='../../mlr3/html/Prediction.html#method-Prediction-score'><code>Prediction$score()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PredictionRegr-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PredictionRegr-initialize}{}}}
\subsection{\code{PredictionRegr$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PredictionRegr$new(
  task = NULL,
  row_ids = task$row_ids,
  truth = task$truth(),
  response = NULL,
  se = NULL,
  quantiles = NULL,
  distr = NULL,
  weights = NULL,
  check = TRUE,
  extra = NULL,
  raw = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link{TaskRegr})\cr
Task, used to extract defaults for \code{row_ids} and \code{truth}.}
      \item{\code{row_ids}}{(\code{integer()})\cr
Row ids of the predicted observations, i.e. the row ids of the test set.}
      \item{\code{truth}}{(\code{numeric()})\cr
True (observed) response.}
      \item{\code{response}}{(\code{numeric()})\cr
Vector of numeric response values.
One element for each observation in the test set.}
      \item{\code{se}}{(\code{numeric()})\cr
Numeric vector of predicted standard errors.
One element for each observation in the test set.}
      \item{\code{quantiles}}{(\code{matrix()})\cr
Numeric matrix of predicted quantiles. One row per observation, one column per quantile.}
      \item{\code{distr}}{(\code{VectorDistribution})\cr
\code{VectorDistribution} from package distr6 (in repository \url{https://raphaels1.r-universe.dev}).
Each individual distribution in the vector represents the random variable 'survival time'
for an individual observation.}
      \item{\code{weights}}{(\code{numeric()})\cr
Vector of measure weights for each observation. Should be constructed from
the \code{Task}'s \code{weights_measure} column.}
      \item{\code{check}}{(\code{logical(1)})\cr
If \code{TRUE}, performs some argument checks and predict type conversions.}
      \item{\code{extra}}{(\code{list()})\cr
List of extra data to be stored in the prediction object.}
      \item{\code{raw}}{(any)\cr
Raw prediction object from the upstream model. Stored as-is without validation.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PredictionRegr-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PredictionRegr-clone}{}}}
\subsection{\code{PredictionRegr$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PredictionRegr$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlr_sugar.R, R/set_validate.R
\name{mlr_sugar}
\alias{mlr_sugar}
\alias{tsk}
\alias{tsks}
\alias{tgen}
\alias{tgens}
\alias{lrn}
\alias{lrns}
\alias{rsmp}
\alias{rsmps}
\alias{msr}
\alias{msrs}
\alias{set_validate}
\title{Syntactic Sugar for Object Construction}
\usage{
tsk(.key, ...)

tsks(.keys, ...)

tgen(.key, ...)

tgens(.keys, ...)

lrn(.key, ...)

lrns(.keys, ...)

rsmp(.key, ...)

rsmps(.keys, ...)

msr(.key, ...)

msrs(.keys, ...)

set_validate(learner, validate, ...)
}
\arguments{
\item{.key}{(\code{character(1)})\cr
Key passed to the respective \link[mlr3misc:Dictionary]{dictionary} to retrieve the object.}

\item{...}{(any)\cr
Additional arguments.}

\item{.keys}{(\code{character()})\cr
Keys passed to the respective \link[mlr3misc:Dictionary]{dictionary} to retrieve multiple objects.}

\item{learner}{(any)\cr
The learner.}

\item{validate}{(\code{numeric(1)} | \code{"predefined"} | \code{"test"} | \code{NULL})\cr
Which validation set to use.}
}
\value{
\link[R6:R6Class]{R6::R6Class} object of the respective type,
or a list of \link[R6:R6Class]{R6::R6Class} objects for the plural versions.

Modified \code{\link{Learner}}
}
\description{
Functions to retrieve objects, set hyperparameters and assign to fields in one go.
Relies on \code{\link[mlr3misc:dictionary_sugar_get]{mlr3misc::dictionary_sugar_get()}} to extract objects from the respective \link[mlr3misc:Dictionary]{mlr3misc::Dictionary}:
\itemize{
\item \code{tsk()} for a \link{Task} from \link{mlr_tasks}.
\item \code{tsks()} for a list of \link{Task}s from \link{mlr_tasks}.
\item \code{tgen()} for a \link{TaskGenerator} from \link{mlr_task_generators}.
\item \code{tgens()} for a list of \link{TaskGenerator}s from \link{mlr_task_generators}.
\item \code{lrn()} for a \link{Learner} from \link{mlr_learners}.
\item \code{lrns()} for a list of \link{Learner}s from \link{mlr_learners}.
\item \code{rsmp()} for a \link{Resampling} from \link{mlr_resamplings}.
\item \code{rsmps()} for a list of \link{Resampling}s from \link{mlr_resamplings}.
\item \code{msr()} for a \link{Measure} from \link{mlr_measures}.
\item \code{msrs()} for a list of \link{Measure}s from \link{mlr_measures}.
}

Helper function to configure the \verb{$validate} field(s) of a \code{\link{Learner}}.

This is especially useful for learners such as \code{AutoTuner} of \CRANpkg{mlr3tuning}
or \code{GraphLearner} of \CRANpkg{mlr3pipelines}
which have multiple levels of \verb{$validate} fields,
where the \verb{$validate} fields need to be configured on multiple levels.
}
\examples{
# penguins task with new id
tsk("penguins", id = "penguins2")

# classification tree with different hyperparameters
# and predict type set to predict probabilities
lrn("classif.rpart", cp = 0.1, predict_type = "prob")

# multiple learners with predict type 'prob'
lrns(c("classif.featureless", "classif.rpart"), predict_type = "prob")
learner = lrn("classif.debug")
set_validate(learner, 0.2)
learner$validate
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PredictionData.R, R/PredictionDataClassif.R,
%   R/PredictionDataRegr.R
\name{PredictionData}
\alias{PredictionData}
\alias{create_empty_prediction_data}
\alias{check_prediction_data}
\alias{is_missing_prediction_data}
\alias{filter_prediction_data}
\alias{check_prediction_data.PredictionDataClassif}
\alias{is_missing_prediction_data.PredictionDataClassif}
\alias{c.PredictionDataClassif}
\alias{check_prediction_data.PredictionDataRegr}
\alias{is_missing_prediction_data.PredictionDataRegr}
\alias{c.PredictionDataRegr}
\title{Convert to PredictionData}
\usage{
create_empty_prediction_data(task, learner)

check_prediction_data(pdata, ...)

is_missing_prediction_data(pdata, ...)

filter_prediction_data(pdata, row_ids, ...)

\method{check_prediction_data}{PredictionDataClassif}(pdata, train_task, ...)

\method{is_missing_prediction_data}{PredictionDataClassif}(pdata, ...)

\method{c}{PredictionDataClassif}(..., keep_duplicates = TRUE)

\method{check_prediction_data}{PredictionDataRegr}(pdata, ...)

\method{is_missing_prediction_data}{PredictionDataRegr}(pdata, ...)

\method{c}{PredictionDataRegr}(..., keep_duplicates = TRUE)
}
\arguments{
\item{task}{(\link{Task}).}

\item{learner}{(\link{Learner}).}

\item{pdata}{(\link{PredictionData})\cr
Named list inheriting from \code{"PredictionData"}.}

\item{...}{(one or more \link{PredictionData} objects).}

\item{row_ids}{\code{integer()}\cr
Row indices.}

\item{train_task}{(\link{Task})\cr
Task used for training the learner.}

\item{keep_duplicates}{(\code{logical(1)})
If \code{TRUE}, the combined \link{PredictionData} object is filtered for duplicated row ids (starting from last).}
}
\description{
Objects of type \code{PredictionData} serve as a intermediate representation for objects of type \link{Prediction}.
It is an internal data structure, implemented to optimize runtime and solve some issues
emerging while serializing R6 objects.
End-users typically do not need to worry about the details,
package developers are advised to continue reading for some technical information.

Unlike most other \CRANpkg{mlr3} objects, \code{PredictionData} relies on the S3 class system.
The following operations must be supported to extend mlr3 for new task types:
\itemize{
\item \code{\link[=as_prediction_data]{as_prediction_data()}} converts objects to class \code{PredictionData}, e.g. objects of type \link{Prediction}.
\item \code{\link[=as_prediction]{as_prediction()}} converts objects to class \link{Prediction}, e.g. objects of type \code{PredictionData}.
\item \code{check_prediction_data()} is called on the return value of the predict method of a \link{Learner}
to perform assertions and type conversions.
Returns an update object of class \code{PredictionData}.
\item \code{is_missing_prediction_data()} is used for the fallback learner (see \link{Learner}) to impute
missing predictions.
Returns vector with row ids which need imputation.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LearnerClassifRpart.R
\name{mlr_learners_classif.rpart}
\alias{mlr_learners_classif.rpart}
\alias{LearnerClassifRpart}
\title{Classification Tree Learner}
\description{
A \link{LearnerClassif} for a classification tree implemented in \code{\link[rpart:rpart]{rpart::rpart()}} in package \CRANpkg{rpart}.
}
\section{Initial parameter values}{

\itemize{
\item Parameter \code{xval} is initialized to 0 in order to save some computation time.
}
}

\section{Custom mlr3 parameters}{

\itemize{
\item Parameter \code{model} has been renamed to \code{keep_model}.
}
}

\section{Dictionary}{

This \link{Learner} can be instantiated via the \link[mlr3misc:Dictionary]{dictionary} \link{mlr_learners} or with the associated sugar function \code{\link[=lrn]{lrn()}}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{mlr_learners$get("classif.rpart")
lrn("classif.rpart")
}\if{html}{\out{</div>}}
}

\section{Meta Information}{

\itemize{
\item Task type: \dQuote{classif}
\item Predict Types: \dQuote{response}, \dQuote{prob}
\item Feature Types: \dQuote{logical}, \dQuote{integer}, \dQuote{numeric}, \dQuote{factor}, \dQuote{ordered}
\item Required Packages: \CRANpkg{mlr3}, \CRANpkg{rpart}
}
}

\section{Parameters}{
\tabular{lllll}{
   Id \tab Type \tab Default \tab Levels \tab Range \cr
   cp \tab numeric \tab 0.01 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   keep_model \tab logical \tab FALSE \tab TRUE, FALSE \tab - \cr
   maxcompete \tab integer \tab 4 \tab  \tab \eqn{[0, \infty)}{[0, Inf)} \cr
   maxdepth \tab integer \tab 30 \tab  \tab \eqn{[1, 30]}{[1, 30]} \cr
   maxsurrogate \tab integer \tab 5 \tab  \tab \eqn{[0, \infty)}{[0, Inf)} \cr
   minbucket \tab integer \tab - \tab  \tab \eqn{[1, \infty)}{[1, Inf)} \cr
   minsplit \tab integer \tab 20 \tab  \tab \eqn{[1, \infty)}{[1, Inf)} \cr
   surrogatestyle \tab integer \tab 0 \tab  \tab \eqn{[0, 1]}{[0, 1]} \cr
   usesurrogate \tab integer \tab 2 \tab  \tab \eqn{[0, 2]}{[0, 2]} \cr
   xval \tab integer \tab 10 \tab  \tab \eqn{[0, \infty)}{[0, Inf)} \cr
}
}

\references{
Breiman L, Friedman JH, Olshen RA, Stone CJ (1984).
\emph{Classification And Regression Trees}.
Routledge.
\doi{10.1201/9781315139470}.
}
\seealso{
\itemize{
\item Chapter in the \href{https://mlr3book.mlr-org.com/}{mlr3book}:
\url{https://mlr3book.mlr-org.com/chapters/chapter2/data_and_basic_modeling.html#sec-learners}
\item Package \CRANpkg{mlr3learners} for a solid collection of essential learners.
\item Package \href{https://github.com/mlr-org/mlr3extralearners}{mlr3extralearners} for more learners.
\item \link[mlr3misc:Dictionary]{Dictionary} of \link[=Learner]{Learners}: \link{mlr_learners}
\item \code{as.data.table(mlr_learners)} for a table of available \link[=Learner]{Learners} in the running session (depending on the loaded packages).
\item \CRANpkg{mlr3pipelines} to combine learners with pre- and postprocessing steps.
\item Package \CRANpkg{mlr3viz} for some generic visualizations.
\item Extension packages for additional task types:
\itemize{
\item \CRANpkg{mlr3proba} for probabilistic supervised regression and survival analysis.
\item \CRANpkg{mlr3cluster} for unsupervised clustering.
}
\item \CRANpkg{mlr3tuning} for tuning of hyperparameters, \CRANpkg{mlr3tuningspaces}
for established default tuning spaces.
}

Other Learner:
\code{\link{Learner}},
\code{\link{LearnerClassif}},
\code{\link{LearnerRegr}},
\code{\link{mlr_learners}},
\code{\link{mlr_learners_classif.debug}},
\code{\link{mlr_learners_classif.featureless}},
\code{\link{mlr_learners_regr.debug}},
\code{\link{mlr_learners_regr.featureless}},
\code{\link{mlr_learners_regr.rpart}}
}
\concept{Learner}
\section{Super classes}{
\code{\link[mlr3:Learner]{Learner}} -> \code{\link[mlr3:LearnerClassif]{LearnerClassif}} -> \code{LearnerClassifRpart}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-LearnerClassifRpart-initialize}{\code{LearnerClassifRpart$new()}}
    \item \href{#method-LearnerClassifRpart-importance}{\code{LearnerClassifRpart$importance()}}
    \item \href{#method-LearnerClassifRpart-selected_features}{\code{LearnerClassifRpart$selected_features()}}
    \item \href{#method-LearnerClassifRpart-clone}{\code{LearnerClassifRpart$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="base_learner"><a href='../../mlr3/html/Learner.html#method-Learner-base_learner'><code>Learner$base_learner()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="configure"><a href='../../mlr3/html/Learner.html#method-Learner-configure'><code>Learner$configure()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="encapsulate"><a href='../../mlr3/html/Learner.html#method-Learner-encapsulate'><code>Learner$encapsulate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="format"><a href='../../mlr3/html/Learner.html#method-Learner-format'><code>Learner$format()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="help"><a href='../../mlr3/html/Learner.html#method-Learner-help'><code>Learner$help()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict"><a href='../../mlr3/html/Learner.html#method-Learner-predict'><code>Learner$predict()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="predict_newdata"><a href='../../mlr3/html/Learner.html#method-Learner-predict_newdata'><code>Learner$predict_newdata()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="print"><a href='../../mlr3/html/Learner.html#method-Learner-print'><code>Learner$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="reset"><a href='../../mlr3/html/Learner.html#method-Learner-reset'><code>Learner$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="Learner" data-id="train"><a href='../../mlr3/html/Learner.html#method-Learner-train'><code>Learner$train()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="mlr3" data-topic="LearnerClassif" data-id="predict_newdata_fast"><a href='../../mlr3/html/LearnerClassif.html#method-LearnerClassif-predict_newdata_fast'><code>LearnerClassif$predict_newdata_fast()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifRpart-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifRpart-initialize}{}}}
\subsection{\code{LearnerClassifRpart$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifRpart$new()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifRpart-importance"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifRpart-importance}{}}}
\subsection{\code{LearnerClassifRpart$importance()}}{
  The importance scores are extracted from the model slot \code{variable.importance}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifRpart$importance()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Named \code{numeric()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifRpart-selected_features"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifRpart-selected_features}{}}}
\subsection{\code{LearnerClassifRpart$selected_features()}}{
  Selected features are extracted from the model slot \code{frame$var}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifRpart$selected_features()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    \code{character()}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LearnerClassifRpart-clone"></a>}}
\if{latex}{\out{\hypertarget{method-LearnerClassifRpart-clone}{}}}
\subsection{\code{LearnerClassifRpart$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LearnerClassifRpart$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}
